/* $Header: /home/xplore/cvs/xpl4gr/axis.C,v 1.3 2002/08/16 12:31:37 sigbert Exp $ */

#include "object.h"
#include "axe.h"
#include "x4math2.h"
#include "xnan.h"


Achse::Achse(const int dim, const float from, const float to):
	line_out(0, 0, 0, 1, 1), mark_out(0, 0, 0, 11, 8), visible(True),
	label_visible(True),
	ticktext_visible(True)
{

#ifdef _BLABLA_
	if (dim!=0)
		cout<<"Acse::Achse(const int dim, const float from, const float to) aufgerufen"<<endl;
#endif

	dimension=dim;
	position[0]=from;
	position[1]=to;

   if (XisInf(to))
   {
   	position[1] = (XisPosInf(to)) ? MAXFLOAT : -MAXFLOAT;
   	globalError.Set (ERR_INTERN, "Axis Error: One of limits is infinity (>1e37)!\nRescale the data.");
   }
   if (XisInf(from))
   {
   	position[0] = (XisPosInf(from)) ? MAXFLOAT : -MAXFLOAT;
   	globalError.Set (ERR_INTERN, "Axis Error: One of limits is infinity (>1e37)!\nRescale the data.");
   }
	axis_text=new char [16];
        // PC 20020320 - texts for axes should be 1, 2, 3
	switch(dim)
	{	case 1:
		sprintf(axis_text, "X");
		break;
		
		case 2:
		sprintf(axis_text, "Y");
		break;
		
		default:
		sprintf(axis_text, "Z");
		// sprintf(axis_text, "%d%s", dim, ". column");
		break;
	}

	//cout<<axis_text<<endl;
	font_number=1;
	ComputeTicks();
}	

Achse::Achse():line_out(0, 0, 0, 1, 1), mark_out(0, 0, 0, 11, 8), visible(True),
	label_visible(True), 
	ticktext_visible(True)
{

#ifdef _BLABLA_
	//cout<<"Acse::Achse() aufgerufen\n";
#endif

	dimension=0;
	position[0]=0;
	position[1]=1;
	axis_text=new char [16];
	sprintf(axis_text, "%d%s", dimension, ". column");
	font_number=1;
	ComputeTicks();
}	

Achse::Achse(const Achse & old):dimension(old.dimension),
	line_out(old.line_out), mark_out(old.mark_out),
	 font_number(old.font_number), visible(old.visible),
	label_visible(old.label_visible), 
	ticktext_visible(old.ticktext_visible)

{

#ifdef _BLABLA_
	cout<<"Acse::Achse(const Achse & old) aufgerufen"<<endl;
#endif

	position[0]=old.position[0];
	position[1]=old.position[1];
	if (strlen(old.axis_text)==0)
	{	globalError.Set (ERR_INTERN, "No text at all!");
		globalError.ShowError();
		axis_text=strdup("something");
	}
	else
		axis_text=strdup(old.axis_text);
	tickmarkpos[0]=old.tickmarkpos[0];
	tickmarkpos[1]=old.tickmarkpos[1];
	tickmarkpos[2]=old.tickmarkpos[2];
	tickmarkvalue[0]=old.tickmarkvalue[0];
	tickmarkvalue[1]=old.tickmarkvalue[1];
}	

Achse::Achse(const Achse * old):dimension(old->dimension),
	line_out(old->line_out), mark_out(old->mark_out),
	 font_number(old->font_number), visible(old->visible),
	label_visible(old->label_visible), 
	ticktext_visible(old->ticktext_visible)
{

#ifdef _BLABLA_
	cout<<"Acse::Achse(const Achse * old) aufgerufen"<<endl;
#endif

	position[0]=old->position[0];
	position[1]=old->position[1];
	if (strlen(old->axis_text)==0)
	{	globalError.Set (ERR_INTERN, "No text at all!");
		globalError.ShowError();
		axis_text=strdup("something");
	}
	else
		axis_text=strdup(old->axis_text);
	tickmarkpos[0]=old->tickmarkpos[0];
	tickmarkpos[1]=old->tickmarkpos[1];
	tickmarkpos[2]=old->tickmarkpos[2];
	tickmarkvalue[0]=old->tickmarkvalue[0];
	tickmarkvalue[1]=old->tickmarkvalue[1];
}	

Achse::~Achse()
{

#ifdef _BLABLA_
	//cout<<"Acse::~Achse) aufgerufen"<<endl;
#endif

	if (axis_text==NULL || strlen(axis_text)==0)
	{	globalError.Set (ERR_INTERN, "No text at all!");
		globalError.ShowError();
	}
	else
		delete axis_text;
}
	
Achse & Achse::operator= (const Achse &ach)
{	if (strlen(ach.axis_text)==0)
	{	globalError.Set (ERR_INTERN, "No text at all!");
		globalError.ShowError();
	}
	else
	{	delete axis_text;
		axis_text=strdup(ach.axis_text);
	}
	tickmarkpos[0]=ach.tickmarkpos[0];
	tickmarkpos[1]=ach.tickmarkpos[1];
	tickmarkpos[2]=ach.tickmarkpos[2];
	tickmarkvalue[0]=ach.tickmarkvalue[0];
	tickmarkvalue[1]=ach.tickmarkvalue[1];
	dimension=ach.dimension;
	position[0]=ach.position[0];
	position[1]=ach.position[1];
	visible=ach.visible;
	label_visible=ach.label_visible;
	ticktext_visible=ach.ticktext_visible;
	return(*this);
}	

int Achse::SetPosition(const float from, const float to)
	/* set position[] und automatic (?) tickmarkpos und values*/
{	position[0]=from;
	position[1]=to;
	if (tickmarkpos[1]==0)
		ComputeTicks();
	return(0);
}
	
int Achse::SetTicks(const float origin, const float major, const float minor)
	/* 	sets the tickmarks;
		if minor==0 then minor =major/5		*/
{	tickmarkpos[0]=origin;	
	if (major==-1)
		return(0);
	if (major==0)
	{	ComputeTicks();
		return(0);
	}
	tickmarkpos[1]=major;
	tickmarkpos[2]=minor;
	if (tickmarkpos[2]==0.0)
		tickmarkpos[2]=tickmarkpos[1]/5.0;
	return(0);
}	

int Achse::SetValues(const float m, const float k)
{	tickmarkvalue[0]=m;
	tickmarkvalue[1]=k;
	return(0);
}

int Achse::SetString(const char *string)
{	delete axis_text;
	axis_text=strdup(string);
	return(0);
}	

Vector <float> Achse::GetContent() const
{	Vector <float> res;
	res.haengeAn(position[0]);
	res.haengeAn(position[1]);
	res.haengeAn(tickmarkpos[0]);
	res.haengeAn(tickmarkpos[1]);
	res.haengeAn(tickmarkpos[2]);
	res.haengeAn(tickmarkvalue[0]);
	res.haengeAn(tickmarkvalue[1]);
	return(res);
}
	
char *Achse::GetString()	const
{	return ( strdup(axis_text));
}
		
int Achse::GetDimension() const
{	return(dimension);
}

void Achse::Out()
{	cout<<"ACHSE: dimension="<<dimension<<", from "<<position[0]<<
" to "<<position[1]<<", tickmarks="<<tickmarkpos[0]<<", "<<tickmarkpos[1]<<", "<<tickmarkpos[2]<<", string= "<<axis_text<<endl;
}
int Achse::ComputeTicks()
{ double range, crange, step, pot = 1.0, realrange, links, rechts;
  int minorstep, lrange, mfrac;
  char mout[32];

  realrange  = position[1] - position[0];
  
  if (dimension==0)	// Achse ach(0, 0, 0)
  	return(1);
  
  if (realrange==0)
  {	/*
  	globalError.Set (ERR_INTERN, "range too small");
 		globalError.ShowError();
 	cerr<<"  position[1] - position[0]<=0.0!!"<<endl;
	cerr<<"position[0]: "<<position[0]<<endl;
	cerr<<"position[1]: "<<position[1]<<endl;
	*/
	tickmarkpos[0]=0;
	tickmarkpos[1]=1;
	tickmarkpos[2]=1;
	tickmarkvalue[0]=0;
	tickmarkvalue[1]=1;
	
	return(1);
	
  }
 range=realrange>0 ? realrange  : -realrange; 
  links=realrange>0 ? position[0]   : position[1];
  rechts=realrange>0 ? position[1] : position[0];
  
   tickmarkpos[0]=0;
  lrange = (int) floor(x4log10(range*(1-1.0e-10)));
  while (lrange)
  { if (lrange<0)
    { pot /= 10;
      lrange++;
    }
    else
    { pot *= 10;
      lrange--;
    }
  }     
  crange = range / pot;
  if (crange<3.5)
  { tickmarkvalue[1] = pot/10;
    tickmarkpos[1]   = 5*tickmarkvalue[1];
    minorstep = 5;
  }
  else if (crange<6.5)
  { tickmarkvalue[1] = pot;
    tickmarkpos[1]   = tickmarkvalue[1];
    minorstep = 5;
  }
  else 
  { tickmarkvalue[1] = pot;
    tickmarkpos[1]   = 2.0 * tickmarkvalue[1];
    minorstep = 4;
  }



  if (((links-0.1*range<=0.0) && 
       (rechts>=0.0)) || 
      ((links<=0.0) &&
       (rechts+0.1*range>=0.0)))
  { tickmarkvalue[0] = 0.0;
    mfrac = -1;
  }
  else if (rechts<0.0)
  { tickmarkvalue[0] = 
      tickmarkpos[1] * ceil(rechts/tickmarkpos[1]);
    mfrac = -(int) floor(x4log10(tickmarkpos[1]));
    sprintf (mout, "%.*f", mfrac, tickmarkvalue[0]);
    tickmarkvalue[0] = atof(mout);
  } 
  else 
  { tickmarkvalue[0] = 
      tickmarkpos[1] * floor(links /tickmarkpos[1]);
    mfrac = -(int) floor(x4log10(tickmarkpos[1]));
    sprintf (mout, "%.*f", mfrac, tickmarkvalue[0]);
    tickmarkvalue[0] = atof(mout);
  }
  tickmarkpos[2] = tickmarkpos[1]/minorstep;
  
	//	ATTENTION!
  if (tickmarkvalue[1]>0.01 &&tickmarkvalue[1]<100)
  {	tickmarkvalue[0]=0.0;
  	tickmarkvalue[1]=1;
  }
  

  return(0);
}  
