/* $Header: /home/xplore/cvs/xpl4gr/chi2.c,v 1.3 2002/08/16 12:31:39 sigbert Exp $ */

#include <math.h>

#include "chi2.h"

#define pow_dd(a,b) pow(*a, *b)
#define min(a,b)    (a<b?a:b)
#define EPSILON 1.0e-12

/**** Borland C++  */
#ifdef __BORLANDC__ 
  static double dnan = 0.0/0.0, dinf = 1.0/0.0, lgdnan = 0.0/0.0;
#endif

#ifdef __GNUC__
  static double lgdnan = 0.0/0.0, dinf = 1.0/0.0;
#endif

void SetNaN (double    *X) { *X = sqrt(-1); }

double sqfc(double v, double p)
{
  double a, vv, g;
  int ifault, i;
  if (v<1.0)
  { SetNaN (&a);
    return (a);
  }
  if ((p<0.0) || (p>1.0))
  { SetNaN (&a);
    return (a);
  }
  if ((p<EPSILON) && (p>=0))
  { a = 0.0; }
  else if ((1-p>=0) && (1-p<EPSILON))
  { a = dinf; }
  else
  { vv=v/2.0;
    g=alngam_(&vv, &ifault); 
    a=ppchi2_(&p, &v, &g, &ifault);
  }
  return a;   
}

/* 239.f -- translated by f2c (version 19950110).
   You must link the resulting object file with the libraries:
	-lf2c -lm   (in that order)
*/


/* Table of constant values */

static doublereal c_b2 = .33333333333333331;
static logical c_false = FALSE_;

doublereal gammad_(x, p, ifault)
doublereal *x, *p;
integer *ifault;
{
    /* System generated locals */
    doublereal ret_val, d__1, d__2, d__3;

    /* Builtin functions */
 //   double sqrt(), pow_dd(), log(), exp();

    /* Local variables */
    static doublereal a, b, c, an, rn;
    extern doublereal alogam_(), alnorm_();
    static doublereal pn1, pn2, pn3, pn4, pn5, pn6, arg;


/* 	ALGORITHM AS239  APPL. STATIST. (1988) VOL. 37, NO. 3 */

/* 	Computation of the Incomplete Gamma Integral */

/* 	Auxiliary functions required: ALOGAM = logarithm of the gamma */
/* 	function, and ALNORM = algorithm AS66 */


    ret_val = 0.;

/* 	Check that we have valid values for X and P */

    if (*p <= 0. || *x < 0.) {
	*ifault = 1;
	return ret_val;
    }
    *ifault = 0;
    if (*x == 0.) {
	return ret_val;
    }

/* 	Use a normal approximation if P > PLIMIT */

    if (*p > 1e3) {
	d__1 = *x / *p;
	pn1 = sqrt(*p) * 3. * (pow_dd(&d__1, &c_b2) + 1. / (*p * 9.) - 1.);
	ret_val = alnorm_(&pn1, &c_false);
	return ret_val;
    }

/* 	If X is extremely large compared to P then set GAMMAD = 1 */

    if (*x > 1e8) {
	ret_val = 1.;
	return ret_val;
    }

    if (*x <= 1. || *x < *p) {

/* 	Use Pearson's series expansion. */
/* 	(Note that P is not large enough to force overflow in ALOGAM). */
/* 	No need to test IFAULT on exit since P > 0. */

	d__1 = *p + 1.;
	arg = *p * log(*x) - *x - alngam_(&d__1, ifault);  // d__1=P + ONE
	c = 1.;
	ret_val = 1.;
	a = *p;
L40:
	a += 1.;
	c = c * *x / a;
	ret_val += c;
	if (c > 1e-14) {
	    goto L40;
	}
	arg += log(ret_val);
	ret_val = 0.;
	if (arg >= -88.) {
	    ret_val = exp(arg);
	}

    } else {

/* 	Use a continued fraction expansion */

	arg = *p * log(*x) - *x - alngam_(p, ifault);
	a = 1. - *p;
	b = a + *x + 1.;
	c = 0.;
	pn1 = 1.;
	pn2 = *x;
	pn3 = *x + 1.;
	pn4 = *x * b;
	ret_val = pn3 / pn4;
L60:
	a += 1.;
	b += 2.;
	c += 1.;
	an = a * c;
	pn5 = b * pn3 - an * pn1;
	pn6 = b * pn4 - an * pn2;
	if (abs(pn6) > 0.) {
	    rn = pn5 / pn6;
/* Computing MIN */
	    d__2 = 1e-14, d__3 = rn * 1e-14;
	    if ((d__1 = ret_val - rn, abs(d__1)) <= min(d__2,d__3)) {
		goto L80;
	    }
	    ret_val = rn;
	}

	pn1 = pn3;
	pn2 = pn4;
	pn3 = pn5;
	pn4 = pn6;
	if (abs(pn5) >= 1e37) {

/* 	Re-scale terms in continued fraction if terms are large */

	    pn1 /= 1e37;
	    pn2 /= 1e37;
	    pn3 /= 1e37;
	    pn4 /= 1e37;
	}
	goto L60;
L80:
	arg += log(ret_val);
	ret_val = 1.;
	if (arg >= -88.) {
	    ret_val = 1. - exp(arg);
	}
    }

    return ret_val;
} /* gammad_ */

/* 241.f -- translated by f2c (version 19950110).
   You must link the resulting object file with the libraries:
	-lf2c -lm   (in that order)
*/

doublereal ppnd7_(p, ifault)
real *p;
integer *ifault;
{
    /* System generated locals */
    real ret_val;

    /* Builtin functions */
    double log(), sqrt();

    /* Local variables */
    static real q, r;


/* 	ALGORITHM AS241  APPL. STATIST. (1988) VOL. 37, NO. 3, 477- */
/* 	484. */

/* 	Produces the normal deviate Z corresponding to a given lower */
/* 	tail area of P; Z is accurate to about 1 part in 10**7. */

/* 	The hash sums below are the sums of the mantissas of the */
/* 	coefficients.   They are included for use in checking */
/* 	transcription. */


/* 	Coefficients for P close to 0.5 */

/* 	HASH SUM AB    32.31845 77772 */

/* 	Coefficients for P not close to 0, 0.5 or 1. */

/* 	HASH SUM CD    15.76149 29821 */

/* 	Coefficients for P near 0 or 1. */

/* 	HASH SUM EF    19.40529 10204 */

    *ifault = 0;
    q = *p - (float).5;
    if (dabs(q) <= (float).425) {
	r = (float).180625 - q * q;
	ret_val = q * (((r * (float)59.10937472 + (float)159.29113202) * r + (
		float)50.434271938) * r + (float)3.3871327179) / (((r * (
		float)67.1875636 + (float)78.757757664) * r + (float)
		17.895169469) * r + (float)1.);
	return ret_val;
    } else {
	if (q < (float)0.) {
	    r = *p;
	} else {
	    r = (float)1. - *p;
	}
	if (r <= (float)0.) {
	    *ifault = 1;
	    ret_val = (float)0.;
	    return ret_val;
	}
	r = sqrt(-(doublereal)log(r));
	if (r <= (float)5.) {
	    r += (float)-1.6;
	    ret_val = (((r * (float).17023821103 + (float)1.3067284816) * r + 
		    (float)2.75681539) * r + (float)1.4234372777) / ((r * (
		    float).12021132975 + (float).7370016425) * r + (float)1.);
	} else {
	    r += (float)-5.;
	    ret_val = (((r * (float).017337203997 + (float).42868294337) * r 
		    + (float)3.081226386) * r + (float)6.657905115) / ((r * (
		    float).012258202635 + (float).24197894225) * r + (float)
		    1.);
	}
	if (q < (float)0.) {
	    ret_val = -(doublereal)ret_val;
	}
	return ret_val;
    }
    return ret_val;
} /* ppnd7_ */



doublereal ppnd16_(p, ifault)
doublereal *p;
integer *ifault;
{
    /* System generated locals */
    doublereal ret_val;

    /* Builtin functions */
    double log(), sqrt();

    /* Local variables */
    static doublereal q, r;


/* 	ALGORITHM AS241  APPL. STATIST. (1988) VOL. 37, NO. 3 */

/* 	Produces the normal deviate Z corresponding to a given lower */
/* 	tail area of P; Z is accurate to about 1 part in 10**16. */

/* 	The hash sums below are the sums of the mantissas of the */
/* 	coefficients.   They are included for use in checking */
/* 	transcription. */


/* 	Coefficients for P close to 0.5 */

/* 	HASH SUM AB    55.88319 28806 14901 4439 */

/* 	Coefficients for P not close to 0, 0.5 or 1. */

/* 	HASH SUM CD    49.33206 50330 16102 89036 */

/* 	Coefficients for P near 0 or 1. */

/* 	HASH SUM EF    47.52583 31754 92896 71629 */

    *ifault = 0;
    q = *p - .5;
    if (abs(q) <= .425) {
	r = .180625 - q * q;
	ret_val = q * (((((((r * 2509.0809287301226727 + 
		33430.575583588128105) * r + 67265.770927008700853) * r + 
		45921.953931549871457) * r + 13731.693765509461125) * r + 
		1971.5909503065514427) * r + 133.14166789178437745) * r + 
		3.387132872796366608) / (((((((r * 5226.495278852854561 + 
		28729.085735721942674) * r + 39307.89580009271061) * r + 
		21213.794301586595867) * r + 5394.1960214247511077) * r + 
		687.1870074920579083) * r + 42.313330701600911252) * r + 1.);
	return ret_val;
    } else {
	if (q < 0.) {
	    r = *p;
	} else {
	    r = 1. - *p;
	}
	if (r <= 0.) {
	    *ifault = 1;
	    ret_val = 0.;
	    return ret_val;
	}
	r = sqrt(-log(r));
	if (r <= 5.) {
	    r += -1.6;
	    ret_val = (((((((r * 7.7454501427834140764e-4 + 
		    .0227238449892691845833) * r + .24178072517745061177) * r 
		    + 1.27045825245236838258) * r + 3.64784832476320460504) * 
		    r + 5.7694972214606914055) * r + 4.6303378461565452959) * 
		    r + 1.42343711074968357734) / (((((((r * 
		    1.05075007164441684324e-9 + 5.475938084995344946e-4) * r 
		    + .0151986665636164571966) * r + .14810397642748007459) * 
		    r + .68976733498510000455) * r + 1.6763848301838038494) * 
		    r + 2.05319162663775882187) * r + 1.);
	} else {
	    r += -5.;
	    ret_val = (((((((r * 2.01033439929228813265e-7 + 
		    2.71155556874348757815e-5) * r + .0012426609473880784386) 
		    * r + .026532189526576123093) * r + .29656057182850489123)
		     * r + 1.7848265399172913358) * r + 5.4637849111641143699)
		     * r + 6.6579046435011037772) / (((((((r * 
		    2.04426310338993978564e-15 + 1.4215117583164458887e-7) * 
		    r + 1.8463183175100546818e-5) * r + 
		    7.868691311456132591e-4) * r + .0148753612908506148525) * 
		    r + .13692988092273580531) * r + .59983220655588793769) * 
		    r + 1.);
	}
	if (q < 0.) {
	    ret_val = -ret_val;
	}
	return ret_val;
    }
    return ret_val;
} /* ppnd16_ */

/* 245.f -- translated by f2c (version 19950110).
   You must link the resulting object file with the libraries:
	-lf2c -lm   (in that order)
*/

/*This file contains two algorithms for the logarithm of the gamma function.*/
/*Algorithm AS 245 is the faster (but longer) and gives an accuracy of about*/
/*10-12 significant decimal digits except for small regions around X = 1 and*/
/* X = 2, where the function goes to zero. */
/*The second algorithm is not part of the AS algorithms.   It is slower but*/
/*gives 14 or more significant decimal digits accuracy, except around X = 1*/
/* and X = 2.   The Lanczos series from which this algorithm is derived is */
/*interesting in that it is a convergent series approximation for the gamma*/
/*function, whereas the familiar series due to De Moivre (and usually wrongly
*/
/*called Stirling's approximation) is only an asymptotic approximation, as*/
/* is the true and preferable approximation due to Stirling. */



doublereal alngam_(xvalue, ifault)
doublereal *xvalue;
integer *ifault;
{
    /* Initialized data */

    static doublereal r1[9] = { -2.66685511495,-24.4387534237,-21.9698958928,
	    11.1667541262,3.13060547623,.607771387771,11.9400905721,
	    31.4690115749,15.234687407 };
    static doublereal twelve = 12.;
    static doublereal zero = 0.;
    static doublereal xlge = 5.1e6;
    static doublereal xlgst = 1e305;
    static doublereal r2[9] = { -78.3359299449,-142.046296688,137.519416416,
	    78.6994924154,4.16438922228,47.066876606,313.399215894,
	    263.505074721,43.3400022514 };
    static doublereal r3[9] = { -212159.572323,230661.510616,27464.7644705,
	    -40262.1119975,-2296.6072978,-116328.495004,-146025.937511,
	    -24235.7409629,-570.691009324 };
    static doublereal r4[5] = { .279195317918525,.4917317610505968,
	    .0692910599291889,3.350343815022304,6.012459259764103 };
    static doublereal alr2pi = .918938533204673;
    static doublereal four = 4.;
    static doublereal half = .5;
    static doublereal one = 1.;
    static doublereal onep5 = 1.5;

    /* System generated locals */
    doublereal ret_val;

    /* Builtin functions */
    double log();

    /* Local variables */
    static doublereal x, y, x1, x2;


/*     ALGORITHM AS245  APPL. STATIST. (1989) VOL. 38, NO. 2 */

/*     Calculation of the logarithm of the gamma function */


/*     Coefficients of rational functions */


/*     Fixed constants */


/*     Machine-dependant constants. */
/*     A table of values is given at the top of page 399 of the paper. */
/*     These values are for the IEEE double-precision format for which */
/*     B = 2, t = 53 and U = 1023 in the notation of the paper. */


    x = *xvalue;
    ret_val = zero;

/*     Test for valid function argument */

    *ifault = 2;
    if (x >= xlgst) {
	return ret_val;
    }
    *ifault = 1;
    if (x <= zero) {
	return ret_val;
    }
    *ifault = 0;

/*     Calculation for 0 < X < 0.5 and 0.5 <= X < 1.5 combined */

    if (x < onep5) {
	if (x < half) {
	    ret_val = -log(x);
	    y = x + one;

/*     Test whether X < machine epsilon */

	    if (y == one) {
		return ret_val;
	    }
	} else {
	    ret_val = zero;
	    y = x;
	    x = x - half - half;
	}
	ret_val += x * ((((r1[4] * y + r1[3]) * y + r1[2]) * y + r1[1]) * y + 
		r1[0]) / ((((y + r1[8]) * y + r1[7]) * y + r1[6]) * y + r1[5])
		;
	return ret_val;
    }

/*     Calculation for 1.5 <= X < 4.0 */

    if (x < four) {
	y = x - one - one;
	ret_val = y * ((((r2[4] * x + r2[3]) * x + r2[2]) * x + r2[1]) * x + 
		r2[0]) / ((((x + r2[8]) * x + r2[7]) * x + r2[6]) * x + r2[5])
		;
	return ret_val;
    }

/*     Calculation for 4.0 <= X < 12.0 */

    if (x < twelve) {
	ret_val = ((((r3[4] * x + r3[3]) * x + r3[2]) * x + r3[1]) * x + r3[0]
		) / ((((x + r3[8]) * x + r3[7]) * x + r3[6]) * x + r3[5]);
	return ret_val;
    }

/*     Calculation for X >= 12.0 */

    y = log(x);
    ret_val = x * (y - one) - half * y + alr2pi;
    if (x > xlge) {
	return ret_val;
    }
    x1 = one / x;
    x2 = x1 * x1;
    ret_val += x1 * ((r4[2] * x2 + r4[1]) * x2 + r4[0]) / ((x2 + r4[4]) * x2 
	    + r4[3]);
    return ret_val;
} /* alngam_ */





doublereal lngamma_(z, ier)
doublereal *z;
integer *ier;
{
    /* Initialized data */

    static doublereal a[9] = { .9999999999995183,676.5203681218835,
	    -1259.139216722289,771.3234287757674,-176.6150291498386,
	    12.50734324009056,-.1385710331296526,9.934937113930748e-6,
	    1.659470187408462e-7 };
    static doublereal lnsqrt2pi = .9189385332046727;

    /* System generated locals */
    doublereal ret_val;

    /* Builtin functions */
    double log();

    /* Local variables */
    static integer j;
    static doublereal tmp;


/*       Uses Lanczos-type approximation to ln(gamma) for z > 0. */
/*       Reference: */
/*            Lanczos, C. 'A precision approximation of the gamma */
/*                    function', J. SIAM Numer. Anal., B, 1, 86-96, 1964. 
*/
/*       Accuracy: About 14 significant digits except for small regions */
/*                 in the vicinity of 1 and 2. */

/*       Programmer: Alan Miller */
/*                   CSIRO Division of Mathematics & Statistics */

/* 	N.B. It is assumed that the Fortran compiler supports long */
/* 	     variable names, including the underline character.   Some */
/* 	     compilers will not accept the 'implicit none' statement */
/* 	     below. */

/*       Latest revision - 17 April 1988 */



    if (*z <= 0.) {
	*ier = 1;
	return ret_val;
    }
    *ier = 0;

    ret_val = 0.;
    tmp = *z + 7.;
    for (j = 9; j >= 2; --j) {
	ret_val += a[j - 1] / tmp;
	tmp += -1.;
/* L10: */
    }
    ret_val += a[0];
    ret_val = log(ret_val) + lnsqrt2pi - (*z + 6.5) + (*z - .5) * log(*z + 
	    6.5);
    return ret_val;
} /* lngamma_ */

/* 66.f -- translated by f2c (version 19950110).
   You must link the resulting object file with the libraries:
	-lf2c -lm   (in that order)
*/

/*This file includes the Applied Statistics algorithm AS 66 for calculating*/
/*the tail area under the normal curve, and two alternative routines which*/
/*give higher accuracy.   The latter have been contributed by Alan Miller of*/
/* CSIRO Division of Mathematics & Statistics, Clayton, Victoria.   Notice */
/* that each function or routine has different call arguments. */


doublereal alnorm_(x, upper)
doublereal *x;
logical *upper;
{
    /* Initialized data */

    static doublereal zero = 0.;
    static doublereal a1 = 5.75885480458;
    static doublereal a2 = 2.62433121679;
    static doublereal a3 = 5.92885724438;
    static doublereal b1 = -29.8213557807;
    static doublereal b2 = 48.6959930692;
    static doublereal c1 = -3.8052e-8;
    static doublereal c2 = 3.98064794e-4;
    static doublereal c3 = -.151679116635;
    static doublereal c4 = 4.8385912808;
    static doublereal c5 = .742380924027;
    static doublereal one = 1.;
    static doublereal c6 = 3.99019417011;
    static doublereal d1 = 1.00000615302;
    static doublereal d2 = 1.98615381364;
    static doublereal d3 = 5.29330324926;
    static doublereal d4 = -15.1508972451;
    static doublereal d5 = 30.789933034;
    static doublereal half = .5;
    static doublereal ltone = 7.;
    static doublereal utzero = 18.66;
    static doublereal con = 1.28;
    static doublereal p = .398942280444;
    static doublereal q = .39990348504;
    static doublereal r = .398942280385;

    /* System generated locals */
    doublereal ret_val;

    /* Builtin functions */
    double exp();

    /* Local variables */
    static doublereal y, z;
    static logical up;


/*         Algorithm AS66 Applied Statistics (1973) vol22 no.3 */

/*       Evaluates the tail area of the standardised normal curve */
/*       from x to infinity if upper is .true. or */
/*       from minus infinity to x if upper is .false. */

/* *** machine dependent constants */

    up = *upper;
    z = *x;
    if (z >= zero) {
	goto L10;
    }
    up = ! up;
    z = -z;
L10:
    if (z <= ltone || up && z <= utzero) {
	goto L20;
    }
    ret_val = zero;
    goto L40;
L20:
    y = half * z * z;
    if (z > con) {
	goto L30;
    }

    ret_val = half - z * (p - q * y / (y + a1 + b1 / (y + a2 + b2 / (y + a3)))
	    );
    goto L40;
L30:
    ret_val = r * exp(-y) / (z + c1 + d1 / (z + c2 + d2 / (z + c3 + d3 / (z + 
	    c4 + d4 / (z + c5 + d5 / (z + c6))))));
L40:
    if (! up) {
	ret_val = one - ret_val;
    }
    return ret_val;
} /* alnorm_ */




/* Subroutine */ int normp_(z, p, q, pdf)
doublereal *z, *p, *q, *pdf;
{
    /* Initialized data */

    static doublereal p0 = 220.2068679123761;
    static doublereal p1 = 221.2135961699311;
    static doublereal p2 = 112.0792914978709;
    static doublereal p3 = 33.912866078383;
    static doublereal p4 = 6.37396220353165;
    static doublereal p5 = .7003830644436881;
    static doublereal p6 = .03526249659989109;
    static doublereal q0 = 440.4137358247522;
    static doublereal q1 = 793.8265125199484;
    static doublereal q2 = 637.3336333788311;
    static doublereal q3 = 296.5642487796737;
    static doublereal q4 = 86.78073220294608;
    static doublereal q5 = 16.06417757920695;
    static doublereal q6 = 1.755667163182642;
    static doublereal q7 = .08838834764831844;
    static doublereal cutoff = 7.071;
    static doublereal root2pi = 2.506628274631001;

    /* System generated locals */
    doublereal d__1;

    /* Builtin functions */
    double exp();

    /* Local variables */
    static doublereal zabs, expntl;


/* 	Normal distribution probabilities accurate to 1.e-15. */
/* 	Z = no. of standard deviations from the mean. */
/* 	P, Q = probabilities to the left & right of Z.   P + Q = 1. */
/*       PDF = the probability density. */

/*       Based upon algorithm 5666 for the error function, from: */
/*       Hart, J.F. et al, 'Computer Approximations', Wiley 1968 */

/*       Programmer: Alan Miller */

/* 	Latest revision - 30 March 1986 */


    zabs = abs(*z);

/* 	|Z| > 37. */

    if (zabs > 37.) {
	*pdf = 0.;
	if (*z > 0.) {
	    *p = 1.;
	    *q = 0.;
	} else {
	    *p = 0.;
	    *q = 1.;
	}
	return 0;
    }

/* 	|Z| <= 37. */

/* Computing 2nd power */
    d__1 = zabs;
    expntl = exp(d__1 * d__1 * -.5);
    *pdf = expntl / root2pi;

/* 	|Z| < CUTOFF = 10/sqrt(2). */

    if (zabs < cutoff) {
	*p = expntl * ((((((p6 * zabs + p5) * zabs + p4) * zabs + p3) * zabs 
		+ p2) * zabs + p1) * zabs + p0) / (((((((q7 * zabs + q6) * 
		zabs + q5) * zabs + q4) * zabs + q3) * zabs + q2) * zabs + q1)
		 * zabs + q0);

/* 	|Z| >= CUTOFF. */

    } else {
	*p = *pdf / (zabs + 1. / (zabs + 2. / (zabs + 3. / (zabs + 4. / (zabs 
		+ .65)))));
    }

    if (*z < 0.) {
	*q = 1. - *p;
    } else {
	*q = *p;
	*p = 1. - *q;
    }
    return 0;
} /* normp_ */




/* Subroutine */ int nprob_(z, p, q, pdf)
doublereal *z, *p, *q, *pdf;
{
    /* Initialized data */

    static doublereal a0 = .5;
    static doublereal a1 = .398942280444;
    static doublereal a2 = .399903438504;
    static doublereal a3 = 5.75885480458;
    static doublereal a4 = 29.8213557808;
    static doublereal a5 = 2.62433121679;
    static doublereal a6 = 48.6959930692;
    static doublereal a7 = 5.92885724438;
    static doublereal b0 = .398942280385;
    static doublereal b1 = 3.8052e-8;
    static doublereal b2 = 1.00000615302;
    static doublereal b3 = 3.98064794e-4;
    static doublereal b4 = 1.98615381364;
    static doublereal b5 = .151679116635;
    static doublereal b6 = 5.29330324926;
    static doublereal b7 = 4.8385912808;
    static doublereal b8 = 15.1508972451;
    static doublereal b9 = .742380924027;
    static doublereal b10 = 30.789933034;
    static doublereal b11 = 3.99019417011;

    /* Builtin functions */
    double exp();

    /* Local variables */
    static doublereal zabs, y;


/*       P, Q = PROBABILITIES TO THE LEFT AND RIGHT OF Z */
/*       FOR THE STANDARD NORMAL DISTRIBUTION. */
/*       PDF  = THE PROBABILITY DENSITY FUNCTION */

/*       REFERENCE: ADAMS,A.G. AREAS UNDER THE NORMAL CURVE, */
/*       ALGORITHM 39, COMPUTER J., VOL. 12, 197-8, 1969. */

/*       LATEST REVISION - 23 JANUARY 1981 */

/* ******************************************************************** */


    zabs = abs(*z);
    if (zabs > 12.7) {
	goto L20;
    }
    y = a0 * *z * *z;
    *pdf = exp(-y) * b0;
    if (zabs > 1.28) {
	goto L10;
    }

/*       Z BETWEEN -1.28 AND +1.28 */

    *q = a0 - zabs * (a1 - a2 * y / (y + a3 - a4 / (y + a5 + a6 / (y + a7))));
    if (*z < 0.) {
	goto L30;
    }
    *p = 1. - *q;
    return 0;

/*       ZABS BETWEEN 1.28 AND 12.7 */

L10:
    *q = *pdf / (zabs - b1 + b2 / (zabs + b3 + b4 / (zabs - b5 + b6 / (zabs + 
	    b7 - b8 / (zabs + b9 + b10 / (zabs + b11))))));
    if (*z < 0.) {
	goto L30;
    }
    *p = 1. - *q;
    return 0;

/*       Z FAR OUT IN TAIL */

L20:
    *q = 0.;
    *pdf = 0.;
    if (*z < 0.) {
	goto L30;
    }
    *p = 1.;
    return 0;

/*       NEGATIVE Z, INTERCHANGE P AND Q */

L30:
    *p = *q;
    *q = 1. - *p;
    return 0;
} /* nprob_ */

/* 91.f -- translated by f2c (version 19950110).
   You must link the resulting object file with the libraries:
	-lf2c -lm   (in that order)
*/

doublereal ppchi2_(p, v, g, ifault)
doublereal *p, *v, *g;
integer *ifault;
{
    /* Initialized data */

    static doublereal aa = .6931471806;
    static doublereal six = 6.;
    static doublereal c1 = .01;
    static doublereal c2 = .222222;
    static doublereal c3 = .32;
    static doublereal c4 = .4;
    static doublereal c5 = 1.24;
    static doublereal c6 = 2.2;
    static doublereal c7 = 4.67;
    static doublereal c8 = 6.66;
    static doublereal c9 = 6.73;
    static doublereal e = 5e-7;
    static doublereal c10 = 13.32;
    static doublereal c11 = 60.;
    static doublereal c12 = 70.;
    static doublereal c13 = 84.;
    static doublereal c14 = 105.;
    static doublereal c15 = 120.;
    static doublereal c16 = 127.;
    static doublereal c17 = 140.;
    static doublereal c18 = 175.;
    static doublereal c19 = 210.;
    static doublereal pmin = 2e-6;
    static doublereal c20 = 252.;
    static doublereal c21 = 264.;
    static doublereal c22 = 294.;
    static doublereal c23 = 346.;
    static doublereal c24 = 420.;
    static doublereal c25 = 462.;
    static doublereal c26 = 606.;
    static doublereal c27 = 672.;
    static doublereal c28 = 707.;
    static doublereal c29 = 735.;
    static doublereal pmax = .999998;
    static doublereal c30 = 889.;
    static doublereal c31 = 932.;
    static doublereal c32 = 966.;
    static doublereal c33 = 1141.;
    static doublereal c34 = 1182.;
    static doublereal c35 = 1278.;
    static doublereal c36 = 1740.;
    static doublereal c37 = 2520.;
    static doublereal c38 = 5040.;
    static doublereal zero = 0.;
    static doublereal half = .5;
    static doublereal one = 1.;
    static doublereal two = 2.;
    static doublereal three = 3.;

    /* System generated locals */
    doublereal ret_val, d__1, d__2;

    /* Builtin functions */
//    double log(), exp(), pow_dd(), sqrt();

    /* Local variables */
    extern doublereal ppnd_();
    static doublereal a, b, c;
    static integer i;
    static doublereal q, t, x, p1, p2, s1, s2, s3, s4, s5, s6;
    extern doublereal gammad_();
    static doublereal ch, xx;
    static integer if1;


/*        Algorithm AS 91   Appl. Statist. (1975) Vol.24, P.35 */

/*        To evaluate the percentage points of the chi-squared */
/*        probability distribution function. */

/*        p must lie in the range 0.000002 to 0.999998, */
/*        v must be positive, */
/*        g must be supplied and should be equal to */
/*          ln(gamma(v/2.0)) */

/*     Incorporates the suggested changes in AS R85 (vol.40(1), */
/*     pp.233-5, 1991) which should eliminate the need for the limited */
/*     range for p above, though these limits have not been removed */
/*     from the routine. */
/*     If IFAULT = 4 is returned, the result is probably as accurate as */
/*     the machine will allow. */

/*     Auxiliary routines required: PPND = AS 111 (or AS 241) and */
/*     GAMMAD = AS 239. */



/*        test arguments and initialise */

    ret_val = -one;
    *ifault = 1;
    if (*p < pmin || *p > pmax) {
	return ret_val;
    }
    *ifault = 2;
    if (*v <= zero) {
	return ret_val;
    }
    *ifault = 0;
    xx = half * *v;
    c = xx - one;

/*        starting approximation for small chi-squared */

    if (*v >= -c5 * log(*p)) {
	goto L1;
    }
    d__1 = *p * xx * exp(*g + xx * aa);
    d__2 = one / xx;
    ch = pow_dd(&d__1, &d__2);
    if (ch < e) {
	goto L6;
    }
    goto L4;

/*        starting approximation for v less than or equal to 0.32 */

L1:
    if (*v > c3) {
	goto L3;
    }
    ch = c4;
    a = log(one - *p);
L2:
    q = ch;
    p1 = one + ch * (c7 + ch);
    p2 = ch * (c9 + ch * (c8 + ch));
    t = -half + (c7 + two * ch) / p1 - (c9 + ch * (c10 + three * ch)) / p2;
    ch -= (one - exp(a + *g + half * ch + c * aa) * p2 / p1) / t;
    if ((d__1 = q / ch - one, abs(d__1)) > c1) {
	goto L2;
    }
    goto L4;

/*        call to algorithm AS 111 - note that p has been tested above. */
/* 	 AS 241 could be used as an alternative. */

L3:
    x = ppnd16_(p, &if1);

/*        starting approximation using Wilson and Hilferty estimate */

    p1 = c2 / *v;
/* Computing 3rd power */
    d__1 = x * sqrt(p1) + one - p1, d__2 = d__1;
    ch = *v * (d__2 * (d__1 * d__1));

/*        starting approximation for p tending to 1 */

    if (ch > c6 * *v + six) {
	ch = -two * (log(one - *p) - c * log(half * ch) + *g);
    }


/*        call to algorithm AS 239 and calculation of seven term */
/*        Taylor series */

L4:
    for (i = 1; i <= 20; ++i) {
	q = ch;
	p1 = half * ch;
	p2 = *p - gammad_(&p1, &xx, &if1);
	if (if1 == 0) {
	    goto L5;
	}

	*ifault = 3;
	return ret_val;
L5:
	t = p2 * exp(xx * aa + *g + p1 - c * log(ch));
	b = t / ch;
	a = half * t - b * c;
	s1 = (c19 + a * (c17 + a * (c14 + a * (c13 + a * (c12 + c11 * a))))) /
		 c24;
	s2 = (c24 + a * (c29 + a * (c32 + a * (c33 + c35 * a)))) / c37;
	s3 = (c19 + a * (c25 + a * (c28 + c31 * a))) / c37;
	s4 = (c20 + a * (c27 + c34 * a) + c * (c22 + a * (c30 + c36 * a))) / 
		c38;
	s5 = (c13 + c21 * a + c * (c18 + c26 * a)) / c37;
	s6 = (c15 + c * (c23 + c16 * c)) / c38;
	ch += t * (one + half * t * s1 - b * c * (s1 - b * (s2 - b * (s3 - b *
		 (s4 - b * (s5 - b * s6))))));
	if ((d__1 = q / ch - one, abs(d__1)) > e) {
	    goto L6;
	}
/* L7: */
    }
    *ifault = 4;

L6:
    ret_val = ch;
    return ret_val;
} /* ppchi2_ */




