/* $Header: /home/xplore/cvs/xpl4gr/distinv.C,v 1.3 2002/08/16 12:31:40 sigbert Exp $ */

#include <math.h>
#include "distinv.h"
#include "x4math2.h"

#define min(a,b)   ((a>b) ? b : a)
#define max(a,b)   ((a>b) ? a : b)
#define abs(a)     (sqrt(a*a))

/* Table of constant values */

#define dmax(a,b) (double)max(a,b)

/* ====================================================================== */
/* NIST Guide to Available Math Software. */
/* Fullsource for module 724 from package TOMS. */
/* Retrieved from NETLIB on Mon Apr  6 08:14:28 1998. */
/* ====================================================================== */
/*      ALGORITHM 724, COLLECTED ALGORITHMS FROM ACM. */
/*      THIS WORK PUBLISHED IN TRANSACTIONS ON MATHEMATICAL SOFTWARE, */
/*      VOL. 19, NO. 4, DECEMBER, 1993, PP. 481-483. */

/* ======================================================================= */
/* Software package FINV. */


double zinv_(double p)
{
    /* Initialized data */

     double c0 = 2.515517;
     double c1 = .802853;
     double c2 = .010328;
     double d1 = 1.432788;
     double d2 = .189269;
     double d3 = .001308;

    /* System generated locals */
    double ret_val;

    /* Builtin functions 
    double x4log(), sqrt(); */

    /* Local variables */
    static double z, denom, numer, ptemp;

/*      .. Scalar Arguments .. */
/*      .. Local Scalars .. */
/*      .. Intrinsic Functions .. */
/*      .. Executable Statements .. */
/* ***********************************************************************
 */

/* PURPOSE: This function returns an approximation for the pth percentile 
*/
/*          of the standard normal distribution function. */
/*          This approximation is due to Hastings (1955). */

/* ARGUMENTS: */
/*     P   - The given percentile. */
/*        P is a real number between 0.0 and 1.0. */

/* ***********************************************************************
 */
    ptemp = p;
    if (p > .5) {
	ptemp = 1. - p;
    }
    z = sqrt(x4log(ptemp) * -2.);
    numer = c0 + z * (c1 + z * c2);
    denom = z * (d1 + z * (d2 + z * d3)) + 1.;
    z -= numer / denom;
    if (p <= .5) {
	z = -z;
    }
    ret_val = z;
    return ret_val;
} /* zinv_ */

double tinit_(long n, double p)
{
    /* System generated locals */
    double ret_val;

    /* Local variables */
/*    extern double zinv_();*/
    static double x, xsquar;

/*      .. Scalar Arguments .. */
/*      .. Local Scalars .. */
/*      .. External Functions .. */
/*      .. Intrinsic Functions .. */
/*      .. Executable Statements .. */
/* ***********************************************************************
 */

/* PURPOSE: This function returns an approximation to the pth percentile 
*/
/*          for a students-t distribution with n degrees of freedom. */
/*          This approximation is due to Goldberg and Levine (1945). */

/* ARGUMENTS: */
/*     N    -  The degree of freedom of the students-t distribution */
/*          N is a positive integer. */

/* EXTERNAL FUNCTION CALLED: */
/*     ZINV -   ZINV returns X, the inverse probability for a Standard */
/*          Normal distribution function --- P{Z <= X} = P. */

/* ***********************************************************************
 */
/*         (          USE ZINV TO FIND X WHEN  P = PROB{N(0,1) <= X}. ) */
    x = zinv_(p);
    xsquar = x * x;
    ret_val = x + x * (xsquar + 1.) / ((double) (n) * 4.) + x * (xsquar *
	     (xsquar * 5. + 16.) + 3.) / ((double) (n) * 96. * (
	    double) (n));
    return ret_val;
} /* tinit_ */

double norppf_(double p)
{
    /* Initialized data */

    double p0 = (double)-.322232431088;
    double p1 = (double)-1.;
    double p2 = (double)-.342242088547;
    double p3 = (double)-.0204231210245;
    double p4 = (double)-4.53642210148e-5;
    double q0 = (double).099348462606;
    double q1 = (double).588581570495;
    double q2 = (double).531103462366;
    double q3 = (double).10353775285;
    double q4 = (double).0038560700634;


    /* Local variables */
     double aden, anum, r, t;


/*     PURPOSE--THIS SUBROUTINE COMPUTES THE PERCENT POINT */
/*              FUNCTION VALUE FOR THE NORMAL (GAUSSIAN) */
/*              DISTRIBUTION WITH MEAN = 0 AND STANDARD DEVIATION = 1. */
/*              THIS DISTRIBUTION IS DEFINED FOR ALL X AND HAS */
/*              THE PROBABILITY DENSITY FUNCTION */
/*              F(X) = (1/SQRT(2*PI))*EXP(-X*X/2). */
/*              NOTE THAT THE PERCENT POINT FUNCTION OF A DISTRIBUTION */
/*              IS IDENTICALLY THE SAME AS THE INVERSE CUMULATIVE */
/*              DISTRIBUTION FUNCTION OF THE DISTRIBUTION. */
/*     INPUT  ARGUMENTS--P      = THE SINGLE PRECISION VALUE */
/*                                (BETWEEN 0.0 AND 1.0) */
/*                                AT WHICH THE PERCENT POINT */
/*                                FUNCTION IS TO BE EVALUATED. */
/*     OUTPUT ARGUMENTS--PPF    = THE SINGLE PRECISION PERCENT */
/*                                POINT FUNCTION VALUE. */
/*     OUTPUT--THE SINGLE PRECISION PERCENT POINT */
/*             FUNCTION VALUE PPF. */
/*     PRINTING--NONE UNLESS AN INPUT ARGUMENT ERROR CONDITION EXISTS. */
/*     RESTRICTIONS--P SHOULD BE BETWEEN 0.0 AND 1.0, EXCLUSIVELY. */
/*     OTHER DATAPAC   SUBROUTINES NEEDED--NONE. */
/*     FORTRAN LIBRARY SUBROUTINES NEEDED--SQRT, ALOG. */
/*     MODE OF INTERNAL OPERATIONS--SINGLE PRECISION. */
/*     LANGUAGE--ANSI FORTRAN. */
/*     REFERENCES--ODEH AND EVANS, THE PERCENTAGE POINTS */
/*                 OF THE NORMAL DISTRIBUTION, ALGORTIHM 70, */
/*                 APPLIED STATISTICS, 1974, PAGES 96-97. */
/*               --EVANS, ALGORITHMS FOR MINIMAL DEGREE */
/*                 POLYNOMIAL AND RATIONAL APPROXIMATION, */
/*                 M. SC. THESIS, 1972, UNIVERSITY */
/*                 OF VICTORIA, B. C., CANADA. */
/*               --HASTINGS, APPROXIMATIONS FOR DIGITAL */
/*                 COMPUTERS, 1955, PAGES 113, 191, 192. */
/*               --NATIONAL BUREAU OF STANDARDS APPLIED MATHEMATICS */
/*                 SERIES 55, 1964, PAGE 933, FORMULA 26.2.23. */
/*               --FILLIBEN, SIMPLE AND ROBUST LINEAR ESTIMATION */
/*                 OF THE LOCATION PARAMETER OF A SYMMETRIC */
/*                 DISTRIBUTION (UNPUBLISHED PH.D. DISSERTATION, */
/*                 PRINCETON UNIVERSITY), 1969, PAGES 21-44, 229-231. */
/*               --FILLIBEN, 'THE PERCENT POINT FUNCTION', */
/*                 (UNPUBLISHED MANUSCRIPT), 1970, PAGES 28-31. */
/*               --JOHNSON AND KOTZ, CONTINUOUS UNIVARIATE */
/*                 DISTRIBUTIONS--1, 1970, PAGES 40-111. */
/*               --THE KELLEY STATISTICAL TABLES, 1948. */
/*               --OWEN, HANDBOOK OF STATISTICAL TABLES, */
/*                 1962, PAGES 3-16. */
/*               --PEARSON AND HARTLEY, BIOMETRIKA TABLES */
/*                 FOR STATISTICIANS, VOLUME 1, 1954, */
/*                 PAGES 104-113. */
/*     COMMENTS--THE CODING AS PRESENTED BELOW */
/*               IS ESSENTIALLY IDENTICAL TO THAT */
/*               PRESENTED BY ODEH AND EVANS */
/*               AS ALGORTIHM 70 OF APPLIED STATISTICS. */
/*               THE PRESENT AUTHOR HAS MODIFIED THE */
/*               ORIGINAL ODEH AND EVANS CODE WITH ONLY */
/*               MINOR STYLISTIC CHANGES. */
/*             --AS POINTED OUT BY ODEH AND EVANS */
/*               IN APPLIED STATISTICS, */
/*               THEIR ALGORITHM REPRESENTES A */
/*               SUBSTANTIAL IMPROVEMENT OVER THE */
/*               PREVIOUSLY EMPLOYED */
/*               HASTINGS APPROXIMATION FOR THE */
/*               NORMAL PERCENT POINT FUNCTION-- */
/*               THE ACCURACY OF APPROXIMATION */
/*               BEING IMPROVED FROM 4.5*(10**-4) */
/*               TO 1.5*(10**-8). */
/*     WRITTEN BY--JAMES J. FILLIBEN */
/*                 STATISTICAL ENGINEERING LABORATORY (205.03) */
/*                 NATIONAL BUREAU OF STANDARDS */
/*                 WASHINGTON, D. C. 20234 */
/*                 PHONE:  301-921-2315 */
/*     ORIGINAL VERSION--JUNE      1972. */
/*     UPDATED         --SEPTEMBER 1975. */
/*     UPDATED         --NOVEMBER  1975. */
/*     UPDATED         --OCTOBER   1976. */

/* --------------------------------------------------------------------- 
*/


/* -----START POINT----------------------------------------------------- 
*/
double ppf;

    if (p != (double).5) {
	goto L150;
    }
    ppf = (double)0.;
    return ppf;

L150:
    r = p;
    if (p > (double).5) {
	r = (double)1. - r;
    }
    t = sqrt(x4log(r) * (double)-2.);
    anum = (((t * p4 + p3) * t + p2) * t + p1) * t + p0;
    aden = (((t * q4 + q3) * t + q2) * t + q1) * t + q0;
    ppf = t + anum / aden;
    if (p < (double).5) {
	ppf = -(double)(ppf);
    }
    return ppf;

} /* norppf_ */


double dlbeta_(double a, double b)
{
    /* System generated locals */
    double ret_val, d__1;

    

    /* Local variables */
     double hold, dspi, temp, at, bt;


/* ********************************************************************** 
*/

/* PURPOSE: The function returns the logarithm of the ratio: */
/*                      Gamma(A)*Gamma(B)/Gamma(A+B) */
/*          This routine was written by the authors to complement the */
/*          IMSL routine DBETAI which calls this function. */

/* ARGUMENTS: */
/*     A, B  - These are the parameters of the Beta distribution. */
/*          A and B are real numbers of the form (integer)/2. */

/* ********************************************************************** 
*/
    dspi = sqrt(3.14159265359);
    at = a;
    bt = b;
    if (at < bt) {
	hold = at;
	at = bt;
	bt = hold;
    }
    if (at > 60.) {
	if (bt > 60.) {
	    temp = at * x4log((at + bt) / at) + bt * x4log((at + bt) / bt);
	    temp = x4exp(temp) * sqrt(at * bt / (6.2831853071800001 * (at + bt))
		    );
	} else {
	    temp = 1.;
L10:
	    if (bt <= 1.0001) {
		goto L20;
	    }
	    bt += -1.;
	    temp = temp * (at + bt) / bt;
	    goto L10;
L20:
	    if ((d__1 = bt - .5, abs(d__1)) < 1e-4) {
		temp /= dspi;
	    }
	    temp = x4log(temp) + at * x4log((at + bt) / at) - bt;
	    temp = x4exp(temp) * sqrt(at);
	    if ((d__1 = bt - 1., abs(d__1)) < 1e-4) {
		temp *= sqrt(at + bt);
	    }
	}
    } else {
	temp = 1.;
L30:
	if (at <= 1.0001) {
	    goto L40;
	}
	at += -1.;
	temp = temp * (at + bt) / at;
	goto L30;
L40:
	if (bt <= 1.0001) {
	    goto L50;
	}
	bt += -1.;
	temp = temp * (at + bt) / bt;
	goto L40;
L50:
	if ((d__1 = at + bt - 1.5, abs(d__1)) < 1e-4) {
	    temp = temp * .5 * dspi;
	}
	if ((d__1 = at - .5, abs(d__1)) < 1e-4) {
	    temp /= dspi;
	}
	if ((d__1 = bt - .5, abs(d__1)) < 1e-4) {
	    temp /= dspi;
	}
    }
    ret_val = -x4log(temp);
    return ret_val;
} /* dlbeta_ */


double dbetai_(double x, double pin, double qin)
{
    /* Initialized data */

    double alneps = 0.;
     double alnsml = 0.;
    double eps = 0.;
    double sml = 0.;

    /* System generated locals */
    long i__1;
    double r__1;
    double ret_val, d__1;

    /* Builtin functions 
    double x4log(), d_int(), exp();*/

    /* Local variables */
    double term, c;
    long i, n;
    double p, q, y, p1;
     long ib;
     double xb;
/*    extern double dlbeta_();*/
     double ps, finsum;

/*      .. Scalar Arguments .. */
/*      .. Local Scalars .. */
/*      .. External Functions .. */
/*      .. Intrinsic Functions .. */
/*      .. Executable Statements .. */
/* ********************************************************************** 
*/

/* PURPOSE: Evaluate the incomplete beta function ratio. */

/* ARGUMENTS: */
/*     X   -  Upper Limit of integration. */
/*         X must be in the interval (0.0,1.0) inclusive. */
/*     PIN -  First beta distribution parameter */
/*         PIN must be a positive real number. */
/*     QIN -  Second beta distribution parameter */
/*         QIN must be a positive real number. */

/* EXTERNAL FUNCTIONS CALLED: */
/*     DLBETA-DLBETA(A,B) returns the value of the logarithm of the Beta 
*/
/*         function having parameters A and B. */

/*     Note:  DBETAI is an IMSL library routine.  The authors have been */
/*         granted special permission to include this source code from */
/*         the IMSL library. */
/*         However, anyone wishing to use this code must first */
/*         purchased the library routines from IMSL. */

/* ********************************************************************** 
*/

    if (eps == 0.) {
	eps = 1.19237e-7;
	alneps = x4log(eps);
	sml = 2.9394099999999999e-37;
	alnsml = x4log(sml);
    }
    y = x;
    p = pin;
    q = qin;
    if (q <= p && x < .8) {
	goto L10;
    }
    if (x < .2) {
	goto L10;
    }
    y = 1. - y;
    p = qin;
    q = pin;
L10:
    if ((p + q) * y / (p + 1.) < eps) {
	goto L70;
    }
    ps = q - (long) (q);
    if (ps == 0.) {
	ps = 1.;
    }
    xb = p * x4log(y) - dlbeta_(ps, p) - x4log(p);
    ret_val = 0.;
    if (xb < alnsml) {
	goto L30;
    }
    ret_val = x4exp(xb);
    term = ret_val * p;
    if (ps == 1.) {
	goto L30;
    }
/* Computing MAX */
    r__1 = alneps / x4log(y);
    n = (long) dmax(r__1,(double)4.);
    i__1 = n;
    for (i = 1; i <= i__1; ++i) {
	term = term * ((double) i - ps) * y / (double) i;
	ret_val += term / (p + (double) i);
/* L20: */
    }
L30:
    if (q <= 1.) {
	goto L60;
    }
    xb = p * x4log(y) + q * x4log(1. - y) - dlbeta_(p, q) - x4log(q);
/* Computing MAX */
    r__1 = xb / alnsml;
    ib = (long) dmax(r__1,0.0);
    term = x4exp(xb - (double) ib * alnsml);
    c = 1. / (1. - y);
    p1 = q * c / (p + q - 1.);
    finsum = 0.;
    n = (long) q;
    if (q == (double) n) {
	--n;
    }
    i__1 = n;
    for (i = 1; i <= i__1; ++i) {
	if (p1 <= 1. && term / eps <= finsum) {
	    goto L50;
	}
	term = (q - (double) i + 1.) * c * term / (p + q - (double) i)
		;
	if (term > 1.) {
	    --ib;
	    term *= sml;
	}
	if (ib == 0) {
	    finsum += term;
	}
/* L40: */
    }
L50:
    ret_val += finsum;
L60:
    if (y != x || p != pin) {
	ret_val = 1. - ret_val;
    }
/* Computing MAX */
    d__1 = min(ret_val,1.);
    ret_val = max(d__1,0.);
    goto L9000;
L70:
    ret_val = 0.;
    xb = p * x4log((max(y,sml))) - x4log(p) - dlbeta_(p, q);
    if (xb > alnsml && y != 0.) {
	ret_val = x4exp(xb);
    }
    if (y != x || p != pin) {
	ret_val = (double)1. - ret_val;
    }
L9000:
    return ret_val;
} /* dbetai_ */

double beta_(double x, double a, double b)
{
    /* System generated locals */
    double ret_val, d__1;

    /* Builtin functions 
    double log(), exp(), sqrt(); */

    /* Local variables */
    double hold, temp, at, bt, xt, yt;

/*      .. Parameters .. */
/*      .. Scalar Arguments .. */
/*      .. Local Scalars .. */
/*      .. Intrinsic Functions .. */
/*      .. Executable Statements .. */
/* ***********************************************************************
 */

/* PURPOSE: This function returns the value of the Beta density function: 
*/
/*         BETA(X|A,B), where A and B are the parameters of the Beta. */

/* ARGUMENTS: */
/*     X    - The argument of the Beta density function. */
/*         X is a real number. */
/*     A,B  - The parameters of the Beta density function. */
/*         A and B are real numbers of the form (integer)/2. */

/* ***********************************************************************
 */
    at = a;
    bt = b;
    xt = x;
    yt = 1. - xt;
    if (at < bt) {
	hold = at;
	at = bt;
	bt = hold;
	yt = xt;
	xt = 1. - xt;
    }
    if (at > 60.) {
	if (bt > 60.) {
	    temp = (at - .5) * x4log(xt * (at + bt - 1.) / (at - 1.)) + (bt - 
		    .5) * x4log(yt * (at + bt - 1.) / (bt - 1.));
	    temp = x4exp(temp - 1.) * sqrt((at + bt - 1.) / (6.2831853071800001
		    * xt * yt));
	} else {
	    temp = 1.;
L10:
	    if (bt <= 1.0001) {
		goto L20;
	    }
	    bt += -1.;
	    temp = temp * yt * (at + bt) / bt;
	    goto L10;
L20:
	    if ((d__1 = bt - .5, abs(d__1)) < 1e-4) {
		temp /= sqrt(yt * 3.14159265359);
	    }
	    temp = x4log(temp) - bt + (at - .5) * x4log(xt * (at + bt - 1.) / (at 
		    - 1.)) + bt * x4log(at + bt - 1.);
	    temp = x4exp(temp) / sqrt(xt);
	}
    } else {
	temp = 1.;
L30:
	if (at <= 1.0001) {
	    goto L40;
	}
	at += -1.;
	temp = temp * xt * (at + bt) / at;
	goto L30;
L40:
	if (bt <= 1.0001) {
	    goto L50;
	}
	bt += -1.;
	temp = temp * yt * (at + bt) / bt;
	goto L40;
L50:
	if ((d__1 = at + bt - 1.5, abs(d__1)) < 1e-4) {
	    temp = temp * .5 * sqrt(3.14159265359);
	}
	if ((d__1 = at - .5, abs(d__1)) < 1e-4) {
	    temp /= sqrt(xt * 3.14159265359);
	}
	if ((d__1 = bt - .5, abs(d__1)) < 1e-4) {
	    temp /= sqrt(yt * 3.14159265359);
	}
    }
    ret_val = temp;
    return ret_val;
} /* beta_ */





double binv_(double a, double b, double vapp, double p)
{
    /* System generated locals */
    long i__1, i__2;
    double ret_val, d__1;

    /* Builtin functions 
    long s_wsle(), do_lio(), e_wsle();*/

    /* Local variables */
/*    extern double beta_();*/
     double tail, d[361]	/* was [19][19] */;
     long i, j, k;
     double q, t, v, vhold, s1, s2, bcoeff;
/*    extern double dbetai_(); */
     long loopct;
     double sum;

  
/*      .. Parameters .. */
/*      .. Scalar Arguments .. */
/*      .. Local Scalars .. */
/*      .. Local Array .. */
/*      .. External Functions .. */
/*      .. Intrinsic Functions .. */
/*      .. Executable Statements .. */
/* ***********************************************************************
 */

/* PURPOSE: This function uses a series expansion method to compute */
/*          percentage points for the Beta distribution. */

/* ARGUMENTS: */
/*     A,B  -  The parameters of the Beta distribution. */
/*          A and B must both be positive real numbers of the form */
/*          (integer)/2. */
/*     VAPP -  The initial approximation to the Beta percentile. */
/*          VAPP is a real number between 0.0 and 1.0. */
/*     P    -  The probability for which the inverse Beta percentile is */
/*          to be evaluated. */
/*          P is a real number between 0.0 and 1.0. */

/* EXTERNAL FUNCTION CALLED: */
/*     BETA  -  BETA(A,B,X) returns the value of the density function for 
*/
/*          a Beta(A,B) random variable. */
/*          A and B are elements in  {N/2 | N is a positive integer}. */
/*          0 <= X <= 1. */

/*     DBETAI - DBETAI(X,A,B) returns the probability that a random */
/*          variable from a Beta distribution having parameters (A,B) */
/*          will be less than or equal to X --- P{B(A,B) <= X} = DBETAI. 
*/
/*          A and B are positive real numbers. */

/* *********************************************************************** */
    v = vapp;
    vhold = 0.;
    loopct = 2;
L10:
    if ((d__1 = (v - vhold) / v, abs(d__1)) >= .001 && loopct != 0) {
	vhold = v;
	--loopct;
/*        (          USE DBETAI TO FIND  F(V) = PROB{ BETA(A,B) <= V }. ) */
/*        (          AND THEN COMPUTE    Q = (P - F(V))/f(V).         ) */
	q = (p - dbetai_(v, a, b)) / beta_(v, a, b);
/*        (                       LET D(N,K) = C(N,K)*Q**(N+K-1)/(N-1)! ) */
	t = 1. - v;
	s1 = q * (b - 1.) / t;
	s2 = q * (1. - a) / v;
	d[0] = s1 + s2;
	tail = d[0] * q / 2.;
	v = v + q + tail;
	k = 3;
L20:
	if ((d__1 = tail / v, abs(d__1)) > 1e-8 && k <= 20) {
/*           (                                   FIRST FIND  D(2,K-2).  ) */
	    s1 = q * ((double) k - 2.) * s1 / t;
	    s2 = q * (2. - (double) k) * s2 / v;
	    d[(k - 2) * 19] = s1 + s2;
/*           (  NOW FIND  D(3,K-3), D(4,K-4), D(5,K-5), ... , D(K-1,1). ) */
	    i__1 = k - 1;
	    for (i = 3; i <= i__1; ++i) {
		sum = d[0] * d[i - 1 + (k - i) * 19 - 2];
		bcoeff = 1.;
		i__2 = k - i;
		for (j = 1; j <= i__2; ++j) {
		    bcoeff = bcoeff * (double) (k - i - j + 1) / (
			    double) j;
		    sum += bcoeff * d[j * 19] * d[i - 1 + (k - i - j) * 19 - 
			    2];
/* L30: */
		}
		d[i + (k - i) * 19 - 2] = sum + d[i - 1 + (k - i + 1) * 19 - 
			2] / (double) (i - 1);
/* L40: */
	    }
/*           ( AND THEN COMPUTE D(K,0) AND USE IT TO EXPAND THE SERIES. ) */
	    d[k - 2] = d[0] * d[k - 3] + d[k + 16] / (double) (k - 1);
	    tail = d[k - 2] * q / (double) k;
	    v += tail;
/*           (                           CHECK FOR A DIVERGENT SERIES.  ) */
	    if (v <= 0. || v >= 1.) {
     	v = -1.;
		goto L50;
	    }
	    ++k;
	    goto L20;
	}
	goto L10;
    }
L50:
    ret_val = v;
    return ret_val;
} /* binv_ */

double finit_(long m, long n, double p)
{
    /* System generated locals */
    double ret_val;

    /* Builtin functions 
    double sqrt(), exp();*/

    /* Local variables */
       double a, b, c, d, x;
/*    extern  Subroutine  int norppf_();*/

/*      .. Scalar Arguments .. */
/*      .. Local Scalars .. */
/*      .. External Functions .. */
/*       SUBROUTINE                        NORPPF */
/*       EXTERNAL                          NORPPF */
/*      .. Intrinsic Functions .. */
/*      .. Executable Statements .. */
/* ***********************************************************************
 */

/* PURPOSE: This function returns an approximation to the pth percentile 
*/
/*          for an F distribution with (m,n) degrees of freedom. */
/*          This approximation is due to Carter (1947). */

/* ARGUMENTS: */
/*     M,N  - The degrees of freedom of the F distribution. */
/*         M and N are positive integers, M >=3 and N >= 3. */

/* EXTERNAL FUNCTION CALLED: */
/*     ZINV -   ZINV returns X, the inverse probability for a Standard */
/*           Normal distribution function --- P{Z <= X} = P. */

/* ***********************************************************************
 */
/*           (        USE ZINV TO FIND X WHEN  P = PROB{ N(0,1) <= X }. ) 
*/
    x=norppf_(p);
/*      X = ZINV(P) */
    a = 1. / ((double) (m) - 1.) + 1. / ((double) (n) - 1.);
    b = 1. / ((double) (m) - 1.) - 1. / ((double) (n) - 1.);
    c = (x * x - 3.) / 6.;
    d = x * a * sqrt(2. / a + c) - b * 2. * (c + .83333333333333337 - a / 3.);
    ret_val = x4exp(d);
    return ret_val;
} /* finit_ */


double finv_(long m, long n, double p)
{
    /* System generated locals */
    double ret_val, d__1, d__2, d__3;

    /* Builtin functions */
/*    long s_wsle(), do_lio(), e_wsle();
    double tan(), pow_dd();*/

    /* Local variables */
/*    extern double binv_(); */
      double a, b, f, t;
/*    extern double finit_(), tinit_(); */
      double power;



/* ***********************************************************************
 */

/* PURPOSE: This function returns percentage points for an F-distribution 
*/
/*          having (M,N) degrees of freedom --- P{F(M,N) <= FINV} = P. */

/* ARGUMENTS: */
/*    M,N  -  The degrees of freedom of an F-distribution. */
/*          M and N must both be positive integers. */
/*    P    -  The probability for which the inverse of the F-distribution 
*/
/*          is to be evaluated. 0 <= P < 1. */
/*    FINV -  Function value  (output) */
/*          The probability that a random variable with F-distribution */
/*          takes a value less than or equal to FINV is equal to P. */
/*          If M or N is negative, or if P >=1 or < 0, then FINV */
/*          returns with a negative value and an appropriate error */
/*          message is sent to standard output. */

/* EXTERNAL FUNCTIONS CALLED: */

/*    BINV  -  BINV(A,B,VAPP,P) returns the percentage points for a Beta 
*/
/*          random variable, B(A,B) --- P{B(A,B) <= BINV} = P. */
/*          0 <= VAPP <= 1 is an initial approximation to BINV. */
/*          A and B are elements in {N/2 | N is a positive integer}. */
/*          If the series used in BINV does not converge, the */
/*          expansion is terminated, a negative value is returned and */
/*          a message noting this divergence is sent to standard */
/*          output. */

/*    FINIT -  FINIT(M,N,P) returns an initial approximation for the */
/*          point x such that a random variable, F, with F-distribution */
/*          having (M,N) degrees of freedom will satisfy  P{F <= x} = P. 
*/
/*          0 < P < 1  and  M,N are positive integers >= 3. */
/*          This is the approximation due to Carter (1947). */

/*    TINIT -  TINIT(N,P) returns an initial approximation for the point 
*/
/*          x such that a random variable, T, with t-distribution having 
*/
/*          N degrees of freedom will satisfy  P{T >= x} = P. */
/*          0 < P < 1  and  N is a positive integer. */
/*          This is the approximation due to Goldberg and Levine (1945). 
*/

/* ***********************************************************************
 */
/*                                 ( FIRST, ELIMINATE BAD VALUES FOR P. ) */
    if (p <= 0. || p >= 1.) {
	if (p == 0.) {
	    ret_val = 0.;
	} else if (p < 0. || p >= 1.) {
	    ret_val = -1.; }
    } else if (m <= 0 || n <= 0) {
	ret_val = -1.;

/*                             (          THEN, CONSIDER SPECIAL CASES. ) */
/*  CASE:  M>2 AND N>2:        (       F-DISTRIBUTION--CONVERT TO BETA. ) */
    } else if (m > 2 && n > 2) {
	f = finit_(m, n, p);
/*          (          TRANSFORM F APPROXIMATION TO BETA APPROXIMATION. ) */
	b = (double) (n) / ((double) (n) + (double) (m) * f);
/*          (                           FIND THE BETA PERCENTAGE POINT. ) */
	d__1 = (double) (n) / 2.;
	d__2 = (double) (m) / 2.;
	d__3 = 1. - p;
	a = binv_(d__1, d__2, b, d__3);
/*          (                     CONVERT RESULT TO F PERCENTAGE POINT. ) */
	ret_val = (1. - a) * (double) (n) / (a * (double) (m));
/*  CASE:  M=1, N=1:   (     CAUCHY DISTRIBUTION--DO DIRECT CALCULATION.) */
    } else if (m == 1 && n == 1) {
	a = tan(p * 3.14159265359 / 2);
	ret_val = a * a;
/*  CASE:  M=1 OR N=1, BUT NOT BOTH: ( T DISTRIBUTION--CONVERT TO BETA. ) */
    } else if (m == 1 || n == 1) {
/*          (            GET INITIAL APPROXIMATION FOR T, CONVERT TO F. ) */
	if (m == 1) {
	    d__1 = (1. - p) / 2.;
	    t = tinit_(n, d__1);
	    f = t * t;
	} else {
	    d__1 = p / 2.;
	    t = tinit_(m, d__1);
	    f = 1. / (t * t);
	}
/*          (          TRANSFORM F APPROXIMATION TO BETA APPROXIMATION. ) */
	b = (double) (n) / ((double) (n) + (double) (m) * f);
/*          (                           FIND THE BETA PERCENTAGE POINT. ) */
	d__1 = (double) (n) / 2.;
	d__2 = (double) (m) / 2.;
	d__3 = 1. - p;
	a = binv_(d__1, d__2, b, d__3);
/*          (                     CONVERT RESULT TO F PERCENTAGE POINT. ) */
	ret_val = (1. - a) * (double) (n) / (a * (double) (m));
/*  CASE:  M=2 OR N=2:  (    POLYNOMIAL DENSITY--DO DIRECT CALCULATION. ) */
    } else {
	if (m == 2) {
	    d__1 = 1. - p;
	    d__2 = 2. / (double) (n);
	    power = pow(d__1, d__2);
	    ret_val = (double) (n) * (1. - power) / (power * 2.);
	} else {
	    d__1 = 2. / (double) (m);
/*   check for pow() */
	    power = pow(p, d__1);

	    ret_val = power * 2. / ((1. - power) * (double) (m));
	}
    }
    return ret_val;
} /* finv_ */


