/* $Header: /home/xplore/cvs/xpl4gr/dllfun.C,v 1.5 2003/01/24 13:52:48 cizek Exp $ */

// XploRe dynamic load functions
// see also xplfun.C and xplbasefun.C 
// contains dynamic load interface

//  SK, 30.12.98

#include "machine.h"

#ifdef HAS_DLL

#if HAS_DLL==WINDLL
#include <windows.h>
#endif

#if HAS_DLL==UNIXDLL
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <unistd.h>
#include <math.h>
#include <string.h>
#ifndef __hpux
#include <dlfcn.h>
#endif
#include <ctype.h>
#endif

#include <stdarg.h>
#include <stdio.h>

#include "object.h"
//#include "xplfun.h"
//#include "init.h"
#include "dllfun.h"
#include "directory.h"

typedef int (*myprocex)(int, int*, voidp *);

typedef int (*myproc0)(void);
typedef int (*myproc1)(double*);
typedef int (*myproc2)(double*, double*);
typedef int (*myproc3)(double*, double*, double*);
typedef int (*myproc4)(double*, double*, double*, double*);
typedef int (*myproc5)(double*, double*, double*, double*, double*);
typedef int (*myproc6)(double*, double*, double*, double*, double*, double*);
typedef int (*myproc7)(double*, double*, double*, double*, double*, double*, double*);
typedef int (*myproc8)(double*, double*, double*, double*, double*, double*, double*, double*);
typedef int (*myproc9)(double*, double*, double*, double*, double*, double*, double*, double*, double*);
typedef int (*myproca)(double*, double*, double*, double*, double*, double*, double*, double*, double*, double*);
typedef int (*myprocb)(double*, double*, double*, double*, double*, double*, double*, double*, double*, double*, double*);
typedef int (*myprocc)(double*, double*, double*, double*, double*, double*, double*, double*, double*, double*, double*, double*);
typedef int (*myprocd)(double*, double*, double*, double*, double*, double*, double*, double*, double*, double*, double*, double*, double*);
typedef int (*myproce)(double*, double*, double*, double*, double*, double*, double*, double*, double*, double*, double*, double*, double*, double*);
typedef int (*myprocf)(double*, double*, double*, double*, double*, double*, double*, double*, double*, double*, double*, double*, double*, double*, double*);

#if HAS_DLL==UNIXDLL
#ifdef __hpux

char errmsg[1024];

char *dlerror()
{ switch (errno)
  { case ENOEXEC:
      sprintf (errmsg, "The specified file is not a shared library, or a format error was detected.");
      break;
    case ENOSYM:
      sprintf (errmsg, "Some symbol required by the shared library could not be found.");
      break;
    case EINVAL:
      sprintf (errmsg, "The specified handle or index is not valid or an attempt was made to load a library at an invalid address.");
      break;
    case ENOMEM:
      sprintf (errmsg, "There is insufficient room in the address space to load the library.");
      break;
    case ENOENT:
      sprintf (errmsg, "The specified library does not exist.");
      break;
    case EACCES:       
      sprintf (errmsg, "Read or execute permission is denied for the specified library.");
      break;
    default:
      sprintf (errmsg, "Unknown DLL error."); 
  }
  return (errmsg);
}

#endif
#endif


Dll::Dll()
{ name = location = callformat = NULL;
  count = 0;
  Dllhandle = 0;
}

Dll::~Dll()
{ if (name) delete[] name;
  if (location) delete[] location;
  count = 0;

#if HAS_DLL==WINDLL
  if (Dllhandle) FreeLibrary (Dllhandle);
#endif

}

extern void logt (const char *txt);
extern void logl (long len);

int Dll::open (const char *newname)
{ name  = strdup (newname);
  count = 0;
  return (open());
}

int Dll::open ()
{ char *fdir;

// logt("Opening DLL\n");

  fdir = openfile(DLL, name, "r");
  // logt (fdir);

  if (fdir==NULL) return 1;

  if (location) delete[] location;
  location = strdup(fdir);

/* Maschinenabhaengig */

#if HAS_DLL==WINDLL
  HINSTANCE hinst = LoadLibrary (location);
  if (hinst == NULL) return 2;
  Dllhandle = GetModuleHandle (location);
#endif

#if HAS_DLL==UNIXDLL
#ifdef __hpux
  Dllhandle = ::shl_load(location, BIND_IMMEDIATE, 0L); 
#else
  Dllhandle = ::dlopen (location, RTLD_LAZY);
#endif
  if (Dllhandle == NULL) return 4;
#endif

  callformat = strdup("%s");
  count++;
  return 0;
}

int Dll::close()
{ count--;
  if (count>0) return (1);
  if (count<0)
  { count = 0;
    return (0);
  }
/* Maschinenabhaengig */

#if HAS_DLL==WINDLL
  if (FreeLibrary (Dllhandle)==0) return (-1);
#endif

#if HAS_DLL==UNIXDLL
#ifdef __hpux
  if (::shl_unload(Dllhandle) != 0) return(-1);
#else
  if (::dlclose (Dllhandle) != 0) return(-1);
#endif
#endif

  if (callformat) delete[] callformat;
  if (location) delete[] location;

  callformat = location = NULL;
  count = 0;
  return (0);
}

int Dll::call (const char *func, int param, doublep *c)
{ char fname[256], localcallformat[16]="%s";
  int i, ret, retry = 0;
#if HAS_DLL==WINDLL
  FARPROC Dllsym;
#endif
#if HAS_DLL==UNIXDLL
  void *Dllsym = NULL;
#endif
 
  do
  {
    if (retry > 0)
   	sprintf (fname, localcallformat, func);
    else
	   sprintf (fname, callformat, func);

/* Maschinenabhaengig */

#if HAS_DLL==WINDLL
//  FARPROC Dllsym;
  Dllsym = GetProcAddress (Dllhandle, fname);
#endif

#if HAS_DLL==UNIXDLL
//  void *Dllsym = NULL;
#ifdef __hpux
  if (::shl_findsym (&Dllhandle, fname, TYPE_PROCEDURE, &Dllsym)!=0)
    Dllsym = NULL;
#else
  Dllsym = ::dlsym (Dllhandle, fname);
#endif
#endif

  if (Dllsym!=NULL) break;

    retry++;
    switch (retry)
    {
      case 1: strcpy(localcallformat, "_%s"); break;
      case 2: strcpy(localcallformat, "%s_"); break;
      case 3: strcpy(localcallformat, "_%s_"); break;
      case 4: strcpy(localcallformat, "__%s"); break;
      case 5: strcpy(localcallformat, "%s__"); break;
      default: strcpy(localcallformat, "%s"); break;
    };
  }
  while (retry < 6);
  if (Dllsym==NULL) return -1;

#if __GNUC__
#ifdef _SHARED_MEM_
  if ((i = fork()) == 0)   // the son
  {
#endif
#endif

  switch (param)
  { case  0 :
      ret = ((myproc0) Dllsym) ();
      break;
    case  1 :
      ret = ((myproc1) Dllsym) (c[0]);
      break;
    case  2 :
      ret = ((myproc2) Dllsym) (c[0], c[1]);
      break;
    case  3 :
      ret = ((myproc3) Dllsym) (c[0], c[1], c[2]);
      break;
    case  4 :
      ret = ((myproc4) Dllsym) (c[0], c[1], c[2], c[3]);
      break;
    case  5 :
      ret = ((myproc5) Dllsym) (c[0], c[1], c[2], c[3], c[4]);
      break;
    case  6 :
      ret = ((myproc6) Dllsym) (c[0], c[1], c[2], c[3], c[4], c[5]);
      break;
    case  7 :
      ret = ((myproc7) Dllsym) (c[0], c[1], c[2], c[3], c[4], c[5], c[6]);
      break;
    case  8 :
      ret = ((myproc8) Dllsym) (c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]);
      break;
    case  9 :
      ret = ((myproc9) Dllsym) (c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7], c[8]);
      break;
    case 10 :
      ret = ((myproca) Dllsym) (c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7], c[8], c[9]);
      break;
    case 11 :
      ret = ((myprocb) Dllsym) (c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7], c[8], c[9], c[10]);
      break;
    case 12 :
      ret = ((myprocc) Dllsym) (c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7], c[8], c[9], c[10], c[11]);
      break;
    case 13 :
      ret = ((myprocd) Dllsym) (c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7], c[8], c[9], c[10], c[11], c[12]);
      break;
    case 14 :
      ret = ((myproce) Dllsym) (c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7], c[8], c[9], c[10], c[11], c[12], c[13]);
      break;
    case 15 :
      ret = ((myprocf) Dllsym) (c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7], c[8], c[9], c[10], c[11], c[12], c[13], c[14]);
      break;
    default :
      ret = -2;
  }

#if __GNUC__
#ifdef _SHARED_MEM_
    exit (0);
  }
  else
    wait (NULL);   // the father
#endif
#endif

  return ret;
}

int Dll::callex (const char *func, int param, int *type, voidp *c)
{ char fname[256], localcallformat[16]="%s";
  int i, ret, retry = 0;
#if HAS_DLL==WINDLL
  FARPROC Dllsym;
#endif
#if HAS_DLL==UNIXDLL
  void *Dllsym = NULL;
#endif

  do
  {
    if (retry > 0)
   	sprintf (fname, localcallformat, func);
    else
	   sprintf (fname, callformat, func);

/* Maschinenabahaengig */

#if HAS_DLL==WINDLL
//  FARPROC Dllsym;
  Dllsym = GetProcAddress (Dllhandle, fname);
#endif

#if HAS_DLL==UNIXDLL
//  void *Dllsym = NULL;
#ifdef __hpux
  if (::shl_findsym (&Dllhandle, fname, TYPE_PROCEDURE, &Dllsym)!=0)
    Dllsym = NULL;
#else
  Dllsym = ::dlsym (Dllhandle, fname);
#endif
#endif

    retry++;
    switch (retry)
    {
      case 1: strcpy(localcallformat, "_%s"); break;
      case 2: strcpy(localcallformat, "%s_"); break;
      case 3: strcpy(localcallformat, "_%s_"); break;
      case 4: strcpy(localcallformat, "__%s"); break;
      case 5: strcpy(localcallformat, "%s__"); break;
      default: strcpy(localcallformat, "%s"); break;
    };
  }
  while ((Dllsym==NULL) && (retry < 6));
  if (Dllsym==NULL) return -1;

  ret = ((myprocex) Dllsym) (param, type, c);

  return ret;
}


inline int Dll::is (const char *qname)
{ return (strcmp(qname, name)==0); }

inline char* Dll::getlocation()
{ return (location==NULL ? strdup("") : location); }

inline char* Dll::getname()
{ return (name); }

inline int Dll::getcount()
{ return (count); }

/*

   Dlllist

*/

Dlllist::Dlllist (int length = 16)
{ int i;
  maxlist = length;
  dlllist = new dllp[maxlist];
  for (i=0; i<maxlist; i++)
    dlllist[i] = NULL;
  lasthandle = -1;
}

Dlllist::~Dlllist ()
{ if (dlllist)
  { int i;
    for (i=0; i<maxlist; i++)
      if (dlllist[i]) delete dlllist[i];
    delete[] dlllist;
  }
  dlllist = NULL;
}

int Dlllist::gethandle (const char *hname)
{ int i;

  for (i=0; i<maxlist; i++)
    if ((dlllist[i]!=NULL) && (dlllist[i]->is(hname))) return i;
  return -1;
}

inline int Dlllist::gethandle (int handle)
{ if ((handle<0) || (handle>maxlist)) return -1;
  return handle;
}


int Dlllist::open (const char *hname)
{ int i, ret;
  for (i=0; i<maxlist; i++)
    if (dlllist[i]==NULL) break;
  if (i==maxlist)
  { dllp *tmp = new dllp[2*maxlist];
    if (tmp==NULL) return -3;
    memcpy (tmp, dlllist, sizeof(dllp)*maxlist);
    for (i=maxlist; i<2*maxlist; i++)
      tmp[i] = NULL;
    delete[] dlllist;
    dlllist = tmp;
    maxlist *= 2;
  }
  dlllist[i] = new Dll();
  ret = dlllist[i]->open(hname);
  if (ret>0)
    ret = -ret;
  else
    ret = i;
  return (ret);
}

int Dlllist::open (int handle)
{ if ((handle<0) || (handle>=maxlist)) return -1;
  return (dlllist[handle]->open());
}

int Dlllist::close (int handle)
{ if ((handle<0) || (handle>=maxlist)) return -1;
  if (dlllist[handle]->close()==0)
  {
    dlllist[handle] = NULL;
  }

/*  if (dlllist[handle]->close()==0)
  { delete dlllist[handle];
    dlllist[handle] = NULL;
  }
*/
  return 0;
}

int Dlllist::close ()
{ int i;
  for (i=0; i<maxlist; i++)
  { if (dlllist[i])
    { while (dlllist[i]->close());
/*      delete dlllist[i];	*/
      dlllist[i] = NULL;
    }
  }
  return 0;
}

inline char* Dlllist::getlocation(int handle)
{ if ((handle<0) || (handle>=maxlist)) return ("invalid handle");
  if (dlllist[handle]) return (dlllist[handle]->getlocation());
  return ("unused");
}

inline int Dlllist::getsize()
{ return (maxlist); }

int Dlllist::getname(charp *p)
{ int i;
  for (i=0; i<maxlist; i++)
  { if (dlllist[i])
      p[i] = dlllist[i]->getname();
    else
      p[i] = "";
  }
  return 0;
}

int Dlllist::getlocation(charp *p)
{ int i;
  for (i=0; i<maxlist; i++)
  { if (dlllist[i])
      p[i] = dlllist[i]->getlocation();
    else
      p[i] = "";
  }
  return 0;
}


int Dlllist::getcount(int *p)
{ int i;
  for (i=0; i<maxlist; i++)
  { if (dlllist[i])
      p[i] = dlllist[i]->getcount();
    else
      p[i] = -1;
  }
  return 0;
}

int Dlllist::call (int handle, const char *func, int nop, doublep *param)
{ int i, r;
  if (handle<0)
  { if (lasthandle>=0)
    { if (dlllist[lasthandle])
      { r = dlllist[lasthandle]->call(func, nop, param);
        if (r>-1) return r;
      }
      else
        lasthandle = -1;
    }

    for (i=0; i<maxlist; i++)
    { if (dlllist[i])
      { r = dlllist[i]->call(func, nop, param);
        if (r>-1)
        { lasthandle = i;
          return r;
        }
      }
    }
  }
  else if (dlllist[handle])
  { lasthandle = handle;
    r = dlllist[handle]->call(func, nop, param);
    return (r);
  }
  return (-3);
}

int Dlllist::callex (int handle, const char *func, int nop, int *type, voidp *param)
{ int i, r;
  if (handle<0)
  { if (lasthandle>=0)
    { if (dlllist[lasthandle])
      { r = dlllist[lasthandle]->callex(func, nop, type, param);
        if (r>-1) return r;
      }
      else
        lasthandle = -1;
    }

    for (i=0; i<maxlist; i++)
    { if (dlllist[i])
      { r = dlllist[i]->callex(func, nop, type, param);
        if (r>-1)
        { lasthandle = i;
          return r;
        }
      }
    }
  }
  else if (dlllist[handle])
  { lasthandle = handle;
    r = dlllist[handle]->callex(func, nop, type, param);
    return (r);
  }
  return (-3);
}

/*

	XploRe functions

*/

Dlllist x4dlllist;

XplObject* DLLopen (Liste<XplObject*> *argl)
// expects a file name
{
  char errtxt[1024];
  const char *t;
  int ret;
  XplDouble  *res;
  XplObject  *o;

  if (argl->laenge () != 1)
  { globalError.Set(ERR_ARGS, "dlopen:  Must be 1");
    return (NULL);
  }

  o = (*argl)[0];
  if (o->DefaultType() != C_TEXT)
  { globalError.Set(ERR_TEXT, "dlopen:  first argument");
    return (NULL);
  }

#ifdef _XPLCDEF_
  t = (*o->DefaultComponent()->Content().xplchar)(0).tocstr();
#else
  t = o->DefaultComponent()->Content().text;
#endif

  // logt (t);
 
  ret = x4dlllist.gethandle(t);
  //  logl (ret);
  if (ret==-1)
  { // logt ("In IF");
    ret = x4dlllist.open(t);

    if (ret==-1)
    { sprintf (errtxt, "dlopen: Can not find \"%s\".\r\nCheck XPL4DLL!", t);
      globalError.Set (ERR_DLLOPEN, errtxt);
      return (NULL);
    }
    if (ret==-2)
    { sprintf (errtxt, "dlopen: Can not open\r\n%s.\r\nCheck if it is dynamic link library!", x4dlllist.getlocation(ret));
      globalError.Set (ERR_DLLOPEN, errtxt);
      return (NULL);
    }
    if (ret==-3)
    { globalError.Set (ERR_DLLOPEN, "dlopen: too many libraries opened.");
      return (NULL);
    }
#if HAS_DLL==UNIXDLL
    if (ret==-4)
    { globalError.Set (ERR_DLLOPEN, ::dlerror());
      return (NULL);
    }
#endif

  }
  else
  { // logt ("In ELSE");
    x4dlllist.open(ret);
  }

  res = new XplDouble (1);
  (*res)(0) = ret;

  return (new XplObject (res, "dlopen"));
}

XplObject* DLLclose (Liste<XplObject*> *argl)
{
  XplObject  *o;
  const char *t;
  long       length = argl->laenge (), handle = -1;

  if (length > 1)
  { globalError.Set(ERR_ARGS, "dlclose:  expect no or one argument");
    return (NULL);
  }

  if (length == 1)
  { o = (*argl)[0];
    if (o->DefaultType () == C_TEXT)
    {
#ifdef _XPLCDEF_
      t = (*o->DefaultComponent()->Content().xplchar)(0).tocstr();
#else
      t = o->DefaultComponent()->Content().text;
#endif
      handle = x4dlllist.gethandle (t);
    }

    if (o->DefaultType () == C_XPLDOUBLE)
    { XplDouble *x = o->DefaultXplDouble();
      handle = x4dlllist.gethandle((int) ((*x)(0)));
    }

    if (handle == -1)         // handle not changed
    { globalError.Set(ERR_TYPE, "dlclose:  first argument");
      return (NULL);
    }

    x4dlllist.close(handle);

  }
  else
    x4dlllist.close();

  return (NULL);
}

XplObject* DLLquery (Liste<XplObject*> *argl)
{
  int i, countDLL, *cp;
  charp *cn, *cl;
  XplObject  *o;
#ifdef _XPLCDEF_
  XplChar    *xc;
#endif
  XplDouble  *xr;

  if (argl->laenge () != 0)
  { globalError.Set(ERR_ARGS, "dlquery:  No argument expected");
    return (NULL);
  }

  countDLL = x4dlllist.getsize();
  cn = new charp[countDLL];
  cl = new charp[countDLL];
  cp = new int[countDLL];
  x4dlllist.getname(cn);
  x4dlllist.getlocation(cl);
  x4dlllist.getcount(cp);

  o = new XplObject ("dlquery");

  xc = new XplChar(countDLL);
  for (i=0; i<countDLL; i++)
    (*xc)(i) = xplstring (cn[i]);
  o->AppendComponent (new Component (new XplObject (xc, "name")));
  delete[] cn;

  xc = new XplChar(countDLL);
  for (i=0; i<countDLL; i++)
    (*xc)(i) = xplstring (cl[i]);
  o->AppendComponent (new Component (new XplObject (xc, "location")));
  delete[] cl;

  xr = new XplDouble(countDLL);
  for (i=0; i<countDLL; i++)
    (*xr)(i) = (double) cp[i];
  o->AppendComponent (new Component (new XplObject (xr, "count")));
  delete[] cp;

  return (o);
}

int DLLfunc (XplObject *o, XplObject *p, int *handle, int *i, char *f)
{ const char *t;
  XplDouble *x;

  if (p==NULL)
  { if (o->DefaultType() != C_TEXT) return (2);
    *i = 1;
#ifdef _XPLCDEF_
    strcpy(f, (*o->DefaultComponent()->Content().xplchar)(0).tocstr());
#else
    strcpy(f, o->DefaultComponent()->Content().text);
#endif
    return 0;
  }
  else
  { if (o->DefaultType() == C_TEXT)
    { if (p->DefaultType() == C_TEXT)
      {
#ifdef _XPLCDEF_
        t = (*o->DefaultComponent()->Content().xplchar)(0).tocstr();
#else
        t = o->DefaultComponent()->Content().text;
#endif
        *handle = x4dlllist.gethandle(t);
        if (*handle<0) *handle = x4dlllist.open(t);
        if (*handle<0) return 1;
        *i = 2;
#ifdef _XPLCDEF_
        strcpy (f, (*p->DefaultComponent()->Content().xplchar)(0).tocstr());
#else
        strcpy (f, p->DefaultComponent()->Content().text);
#endif
        return 0;
      }
      else
      { *i = 1;
#ifdef _XPLCDEF_
        strcpy (f, (*o->DefaultComponent()->Content().xplchar)(0).tocstr());
#else
        strcpy (f, o->DefaultComponent()->Content().text);
#endif
        return 0;
      }
    }
    else
    { x = o->DefaultXplDouble();
      if (x==NULL) return 1;
      *handle = x4dlllist.gethandle((int) ((*x)(0)));
      if (*handle==-1) return (1);
      if (p->DefaultType() == C_TEXT)
      { *i = 2;
#ifdef _XPLCDEF_
        strcpy (f, (*p->DefaultComponent()->Content().xplchar)(0).tocstr());
#else
        strcpy (f, p->DefaultComponent()->Content().text);
#endif
        return 0;
      }
    }
  }
  return 1;
}

XplObject* DLLcall (Liste<XplObject*> *argl)
{ XplObject  *o, *p;
  XplDouble  *res, *x;
  int handle = -1, i = 0, j, length = argl->laenge (), err, nop;
  char t[256];
  char errtxt[256];
  doublep *param;

  if (length < 1)
  { globalError.Set(ERR_ARGS, "dlcall:  Must be > 0");
    return (NULL);
  }

  o = (*argl)[0];
  p = (length==1 ? NULL : p = (*argl)[1]);
  err = DLLfunc (o, p, &handle, &i, t);
  if (err)
  { globalError.Set(ERR_DLLCALL, "dlcall: no handle");
    return (NULL);
  }

#if __GNUC__
  int shmid, k;
  double *r;
#endif

  nop   = length-i;
  param = new doublep[nop];
  for (j=0; i+j<length; j++)
  { x = (*argl)[i+j]->DefaultXplDouble();
    if (x==NULL)
    { char text[TEXTLENGTH];
      sprintf (text, "dlcall: %ld argument", j);
      globalError.Set (ERR_NUM, text);
      return (NULL);
    }

#ifdef __BORLANDC__
    param[j] = x->getptr();
#endif

#ifdef __GNUC__
#ifdef _SHARED_MEM_
    shmid = shmget (IPC_PRIVATE, size*sizeof(double), 0xffff);
    if (shmid == -1)     // ok, we have no shared memory
    { cerr << "Can not get shared memory" << endl;
      cerr << "leaving dlcall" << endl;
      cerr << "i = " << i << endl;
      return (NULL);
    }
    param[i] = (double *) shmat (shmid, NULL, 0x666);  // r/w permission user
    if (param[i] == NULL  || (long) param[i] == -1)
    { cerr << "dlcall error: " << i << '\t' << size;
      exit (1);
    }
#else
    param[j] = new double[x->size()];
#endif
    r = param[j];
    for (k=0; k<x->size(); k++)          // copy input arguments
      *r++ = (*x)[k];
#endif

  }

  err = x4dlllist.call (handle, t, nop,param);

#ifdef __GNUC__
  for (j=0; i+j<length; j++)
  { x = (*argl)[i+j]->DefaultXplDouble();
    r = param[j];
    for (k=0; k<x->size(); k++)          // copy back results
      (*x)[k] = *r++;
#ifdef _SHARED_MEM_
    shmdt ((void *) param[i]);
#else
    delete param[j];
#endif
  }
#endif

  if (err==-1)
  { sprintf (errtxt, "dlcall: %s function not found", t);
    globalError.Set(ERR_DLLCALL, errtxt);
    return (NULL);
  }
  if (err==-2)
  { sprintf (errtxt, "dlcall: too many parameters for %s (<=12)", t);
    globalError.Set(ERR_DLLCALL, errtxt);
    return (NULL);
  }
  if (err==-3)
  { globalError.Set(ERR_DLLCALL, "dlcall: invalid handle");
    return (NULL);
  }

  res = new XplDouble(1);
  (*res)(0) = err;

  return (new XplObject (res, "dlcall"));
}

#ifdef __BORLANDC__
#define INT1  __int8
#define INT2  __int16
#define INT4  __int32
#define INT8  __int64
#define DBL4  float
#define DBL8  double
#endif

#ifdef __GNUC__
#define INT1  signed char
#define INT2  short
#define INT4  int
#define INT8  long
#define DBL4  float
#define DBL8  double
#endif


void* comp2ptr (Component *c, int type, int byrow)
{ INT1 *i1, *s1; INT2 *i2, *s2; INT4 *i4, *s4; INT8 *i8, *s8;
  DBL4 *d4, *t4; DBL8 *d8, *t8;
  void *ret = NULL;
  long n, p, np, sz, i, j, k, sn, ss, cl;
  double *xp;
  char *cp;
  const char *ct;
  xplstring *sp;
  XplDouble *x;
  XplChar *s;

  if (c->Type()==C_XPLDOUBLE)
  { x = c->Content().xpldouble;
    if (x==NULL) return NULL;
    sz = x->size();

    n  = x->dimension(0);
    p  = x->dimension(1);
    xp = x->getptr();
    np = n*p;
    sn = sz/np;
    if (type==8)
    { t8 = d8 = new DBL8[sz];
      if (byrow)
      { for (k=0; k<sn; k++)
        { for (i=0; i<n; i++)
          { for (j=0; j<p; j++, t8++)
              *t8 = *(xp+n*j+i);
          }
        }
      }
      else
        memcpy (d8, xp, 8*sz);
      ret = d8;
    }
    else if (type==4)
    { t4 = d4 = new DBL4[sz];
      if (byrow)
      { for (k=0; k<sn; k++)
        { for (i=0; i<n; i++)
          { for (j=0; j<p; j++, t4++)
              *t4 = (DBL4) *(xp+n*j+i);
          }
        }
      }
      else
      { for (k=0; k<sz; k++, t4++, xp++)
          *t4 = (DBL4) *xp;
      }
      ret = d4;
    }
    else if (type==-1)
    { s1 = i1 = new INT1[sz];
      if (byrow)
      { for (k=0; k<sn; k++)
        { for (i=0; i<n; i++)
          { for (j=0; j<p; j++, s1++)
              *s1 = (INT1) *(xp+n*j+i);
          }
        }
      }
      else
      { for (k=0; k<sz; k++, s1++, xp++)
          *s1 = (INT1) *xp;
      }
      ret = i1;
    }
    else if (type==-2)
    { s2 = i2 = new INT2[sz];
      if (byrow)
      { for (k=0; k<sn; k++)
        { for (i=0; i<n; i++)
          { for (j=0; j<p; j++, s2++)
              *s2 = (INT2) *(xp+n*j+i);
          }
        }
      }
      else
      { for (k=0; k<sz; k++, s2++, xp++)
          *s2 = (INT2) *xp;
      }
      ret = i2;
    }
    else if (type==-4)
    { s4 = i4 = new INT4[sz];
      if (byrow)
      { for (k=0; k<sn; k++)
        { for (i=0; i<n; i++)
          { for (j=0; j<p; j++, s4++)
              *s4 = (INT4) *(xp+n*j+i);
          }
        }
      }
      else
      { for (k=0; k<sz; k++, s4++, xp++)
          *s4 = (INT4) *xp;
      }
      ret = i4;
    }
    else if (type==-8)
    { s8 = i8 = new INT8[sz];
      if (byrow)
      { for (k=0; k<sn; k++)
        { for (i=0; i<n; i++)
          { for (j=0; j<p; j++, s8++)
              *s8 = (INT8) *(xp+n*j+i);
          }
        }
      }
      else
      { for (k=0; k<sz; k++, s8++, xp++)
          *s8 = (INT8) *xp;
      }
      ret = i8;
    }
  }
  else if (c->Type()==C_TEXT)
  { s = c->Content().xplchar;
    if (s==NULL) return NULL;
    sz = s->size();
    n  = s->dimension(0);
    p  = s->dimension(1);
    sp = s->getptr();
    np = n*p;
    sn = sz/np;
    for (ss=i=0; i<sz; i++)
      ss += (1+strlen((*s)[i].tocstr()));
    if (type==-1)
    { cp = new char[ss];
      if (byrow)
      { for (sz=k=0; k<sn; k++)
        { for (i=0; i<n; i++)
          { for (j=0; j<p; j++, sz+=cl)
            { ct = (sp+n*j+i)->tocstr();
              cl = 1+strlen(ct);
              memcpy (cp+sz, ct, cl*sizeof(char));
            }
          }
        }
      }
      else
      { for (sn=k=0; k<sz; k++, sn+=cl)
        { ct = (sp+k)->tocstr();
          cl = 1+strlen(ct);
          memcpy (cp+sn, ct, cl*sizeof(char));
        }
      }
      ret = cp;
    }
  }
  return (ret);
}

void ptr2comp (Component *c, int type, int byrow, void *ptr)
{ INT1 *s1; INT2 *s2; INT4 *s4; INT8 *s8;
  DBL4 *t4; DBL8 *t8;
  long n, p, np, sz, i, j, k, sn;
  double *xp;
  char *cp;
  xplstring *sp;
  XplDouble *x;
  XplChar *s;

  if (c->Type()==C_XPLDOUBLE)
  { x = c->Content().xpldouble;
    sz = x->size();
    n  = x->dimension(0);
    p  = x->dimension(1);
    xp = x->getptr();
    np = n*p;
    sn = sz/np;
    if (type==8)
    { t8 = (DBL8 *) ptr;
      if (byrow)
      { for (k=0; k<sn; k++)
        { for (i=0; i<n; i++)
          { for (j=0; j<p; j++, t8++)
              *(xp+n*j+i) = *t8;
          }
        }
      }
      else
        memcpy (xp, t8, 8*sz);
    }
    else if (type==4)
    { t4 = (DBL4 *) ptr;
      if (byrow)
      { for (k=0; k<sn; k++)
        { for (i=0; i<n; i++)
          { for (j=0; j<p; j++, t4++)
              *(xp+n*j+i) = (double) *t4;
          }
        }
      }
      else
      { for (k=0; k<sz; k++, t4++, xp++)
          *xp = (double) *t4;
      }
    }
    else if (type==-1)
    { s1 = (INT1 *) ptr;
      if (byrow)
      { for (k=0; k<sn; k++)
        { for (i=0; i<n; i++)
          { for (j=0; j<p; j++, s1++)
              *(xp+n*j+i) = (double) *s1;
          }
        }
      }
      else
      { for (k=0; k<sz; k++, s1++, xp++)
          *xp = (INT1) *s1;
      }
    }
    else if (type==-2)
    { s2 = (INT2 *) ptr;
      if (byrow)
      { for (k=0; k<sn; k++)
        { for (i=0; i<n; i++)
          { for (j=0; j<p; j++, s2++)
              *(xp+n*j+i) = (double) *s2;
          }
        }
      }
      else
      { for (k=0; k<sz; k++, s2++, xp++)
          *xp = (double) *s2;
      }
    }
    else if (type==-4)
    { s4 = (INT4 *) ptr;
      if (byrow)
      { for (k=0; k<sn; k++)
        { for (i=0; i<n; i++)
          { for (j=0; j<p; j++, s4++)
              *(xp+n*j+i) = (double) *s4;
          }
        }
      }
      else
      { for (k=0; k<sz; k++, s4++, xp++)
          *xp = (double) *s4;
      }
    }
    else if (type==-8)
    { s8 = (INT8 *) ptr;
      if (byrow)
      { for (k=0; k<sn; k++)
        { for (i=0; i<n; i++)
          { for (j=0; j<p; j++, s8++)
              *(xp+n*j+i) = (double) *s8;;
          }
        }
      }
      else
      { for (k=0; k<sz; k++, s8++, xp++)
          *xp = (double) *s8;
      }
    }
  }
  else if (c->Type()==C_TEXT)
  { s = c->Content().xplchar;
    sz = s->size();
    n  = s->dimension(0);
    p  = s->dimension(1);
    sp = s->getptr();
    np = n*p;
    sn = sz/np;
    if (type==-1)
    { cp = (char *) ptr;
      if (byrow)
      { for (k=0; k<sn; k++)
        { for (i=0; i<n; i++)
          { for (j=0; j<p; j++)
            { *(sp+n*j+i) = xplstring (cp);
              cp += (1+strlen(cp));
            }
          }
        }
      }
      else
      { for (k=0; k<sz; k++)
        { *(sp+k) = xplstring (cp);
          cp += (1+strlen(cp));
        }
      }
    }
  }
}

XplObject* DLLcallEx (Liste<XplObject*> *argl)
{ XplObject *o, *p;
  XplDouble *res, *x;
  int handle = -1, i = 0, length = argl->laenge (), err, nop,
      *byrow, *type;
  char t[256];
  long k;
  char errtxt[256];
  Component *c;
  voidp *param;

  if (length < 1)
  { globalError.Set(ERR_ARGS, "dlcallex:  Must be > 0");
    return (NULL);
  }

  o = (*argl)[0];
  p = (length==1 ? NULL : (*argl)[1]);
  err = DLLfunc (o, p, &handle, &i, t);
  if (err)
  { globalError.Set(ERR_DLLCALL, "dlcallex: no handle");
    return (NULL);
  }

  o = (*argl)[i];
  nop = (o->Type()==O_COMPOSED ? o->ComponentNumber() : 1);
  byrow = new int[nop];
  for (k=0; k<nop; k++)
    byrow[k] = 0;
  type  = new int[nop];
  for (k=0; k<nop; k++)
    type[k]  = 8;
  param = new voidp[nop];
  for (k=0; k<nop; k++)
    param[k] = NULL;

  err = 0;
  if (i<length-1)
  { p = (*argl)[i+1];
    if (p->Type()==O_COMPOSED)
    { c = p->GetComponent("type");
      if (c)
      { x = c->Content().xplobject->DefaultXplDouble();
        if (x)
        { if (x->size()==nop)
          { for (k=0; k<nop; k++)
              type[k] = (int) (*x)[k];
          }
          else
            err = 5;
        }
        else
          err = 3;
      }

      c = p->GetComponent("byrow");
      if (c)
      { x = c->Content().xplobject->DefaultXplDouble();
        if (x)
        { if (x->size()==nop)
          { for (k=0; k<nop; k++)
              byrow[k] = (int) (*x)[k];
          }
          else
            err = 5;
        }
        else
          err = 4;
      }
    }
    else
      err = 2;
  }

  if (err==0)
  { switch (o->Type())
    { case O_XPLDOUBLE:
      case O_XPLCHAR:
        param[0] = comp2ptr (o->DefaultComponent(), type[0], byrow[0]);
        if (param[0]==NULL) err = 6;
        break;
      case O_COMPOSED:
        for (k=0; k<nop; k++)
        { param[k] = comp2ptr ((*o)[k]->Content().xplobject->DefaultComponent(),
                               type[k], byrow[k]);
          if (param[k]==NULL) err = 6;
        }
        break;
      default:
        err=1;
    }
  }

  if (err)
  { for (i=0; i<nop; i++)
      if (param[i]) delete[] param[i];
    delete[] param;
    delete[] type;
    delete[] byrow;
    switch (err)
    { case 1 :
        globalError.Set(ERR_DLLCALL, "dlcallex: empty or unsupported parameter");
        break;
      case 2 :
        globalError.Set(ERR_DLLCALL, "dlcallex: uncomposed optionlist");
        break;
      case 3 :
        globalError.Set(ERR_DLLCALL, "dlcallex: opt.type has to be numeric");
        break;
      case 4 :
        globalError.Set(ERR_DLLCALL, "dlcallex: opt.byrow has to be numeric");
        break;
      case 5 :
        globalError.Set(ERR_DLLCALL, "dlcallex: opt.xxx length <> size(list)");
        break;
      case 6 :
        globalError.Set(ERR_DLLCALL, "dlcallex: conversion failed before call");
        break;
    }

    return (NULL);
  }

  err = x4dlllist.callex (handle, t, nop, type, param);

  if (err==-1)
  { sprintf (errtxt, "dlcallex: %s function not found", t);
    globalError.Set(ERR_DLLCALL, errtxt);
    return (NULL);
  }
  if (err==-2)
  { sprintf (errtxt, "dlcallex: too many parameters for %s (<=12)", t);
    globalError.Set(ERR_DLLCALL, errtxt);
    return (NULL);
  }
  if (err==-3)
  { globalError.Set(ERR_DLLCALL, "dlcallex: invalid handle");
    return (NULL);
  }

  switch (o->Type())
  { case O_XPLDOUBLE:
    case O_XPLCHAR:
      ptr2comp (o->DefaultComponent(), type[0], byrow[0], param[0]);
      break;
    case O_COMPOSED:
      for (k=0; k<nop; k++)
        ptr2comp ((*o)[k]->Content().xplobject->DefaultComponent(),
                  type[k], byrow[k], param[k]);
      break;
  }

  for (i=0; i<nop; i++)
    if (param[i]) delete[] param[i];
  delete[] param;
  delete[] type;
  delete[] byrow;

  res = new XplDouble(1);
  (*res)(0) = err;

  return (new XplObject (res, "dlcallex"));
}

#endif

