/* $Header: /home/xplore/cvs/xpl4gr/foreign.C,v 1.4 2002/11/27 09:16:16 cizek Exp $ */

#include <stdlib.h>
#include <math.h>
#include <string.h>


#include "object.h"
#include "function.h"
/*
#include "xplfun.h"
#include "xplerror.h"
#include "init.h"   //                      ????
#include "ttkio.h"
#include "xplexec.h"
#include "matrix.h"
*/

int chol(double *B, double *len, double *bb);
long tie(double * x, double* y, double* weigh, double* w, long length);
void splinec(double* x, double* y, double* D2y, long length);
void splsm_(double* x, double* y, double* D2y, double *w, long length, double lambda);
void splad_(double* x, double* y, double* D2y, double *w, long length);
void splint_(double* x, double* y, double* D2y,  long length, 
            double* x_new, double* y_new, long length_new);
void sqrdc_(double *x, 
            long ldx, 
            long n, 
            long p,
            double  *qraux,
            long    *jpvt,  
            int job);
void qrR(double *QR, long nrow, long ncol, double *R);
void qrQ(double *QR, double *raux, long nrow,long ncol, double *Q,
           double eps);

// Interface parser SWS funcdef
 
XplObject* cholchol(Liste<XplObject*> *argl)
// SWS 8.2.98  Interface zu Golubev cholesky decompos.
{
  long      length = argl->laenge();
  XplDouble *x, *y, *res;

  if (length != 2)
    {
    globalError.Set(ERR_ARGS, "xaLocPol: Must = 2");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "chol: first argument");
    return (NULL);
    }
  x = (*argl)[0]->DefaultXplDouble ();

  if ((*argl)[1]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "chol: second argument");
    return (NULL);
    }
  y = (*argl)[1]->DefaultXplDouble ();

  int xrows, xcols;
  double * content;
  int yrows, ycols, b;
  
  xrows=x->dimension(0);
  yrows=y->dimension(0);
  xcols=x->dimension(1);
  ycols=y->dimension(1);
   
  if (xcols!=xrows)
    {
    globalError.Set (ERR_NUM, "chol: first argument has to be quadratic");
    return (NULL);
    }
  
  if (ycols!=1 || yrows!=1)
    {
    globalError.Set (ERR_NUM, "chol: second argument has to be a scalar");
    return (NULL);
    }
   b=((int)((*y)[0]));   
  content= new double [b*xrows];
  int r, c;
  for (c=0; c<xrows*b; c++)
        content[c]=0;
  
  for (c=0; c<b; c++)
        for (r=c; r<xrows; r++)
                content[c+r*b]=(*x)(r-c, r);
                 
/*
  for (r=0; r<xrows; r++)
  {       for (c=0; c<b; c++)
                cout<<content[c+r*b]<<"\t";
        cout<<endl;
  }
*/
  double bb=b;
  double xxr=xrows;
  chol(content, &xxr, &bb);
    
  
/*
        cout<<"RESULT:"<<endl;
        
  for (r=0; r<xrows; r++)
  {       for (c=0; c<b; c++)
                cout<<content[c+r*b]<<"\t";
        cout<<endl;
  }
*/  

  res = new XplDouble(xrows, xrows);
  
  for (c=0; c<xrows*xrows; c++)
        (*res)[c]=0.0;
  for (c=0; c<b; c++)
        for (r=c; r<xrows; r++)
                (*res)(r-c, r)=content[c+r*b];

//      res->mdump("%f");

   delete [] content;
   if (globalError.Get())
    {
    delete res;
    return (NULL);
    }

  return (new XplObject (res, "chol"));
}


// Interface parser YG / StSp funcdef
 

XplObject* spline(Liste<XplObject*> *argl)
// YG / StSp 16.4.98  Interface zu Golubev spline proc.
{
  long      numb = argl->laenge(), xrows, xcols, yrows, ycols,
       lambdacols, lambdarows, i, j, n;
  XplDouble *x, *y, *ynew,  *xnew, *lambda, *weight;
  double *dx, *dy, *dynew, *dxnew, dlambda, *dweight;
  

  if ((numb<2) || (numb>5))
    {
    globalError.Set(ERR_ARGS, "spline: Must = 2,3,4 or 5");
    return (NULL);
    }

  x = y = ynew = xnew = lambda = weight = NULL;

  switch (numb)
    { case 5:
       if ((*argl)[4]->DefaultType () != C_XPLDOUBLE)
       {
         globalError.Set (ERR_NUM, "spline: fifth argument");
         return (NULL);
       }
       weight = (*argl)[4]->DefaultXplDouble ();
      case 4:
       if ((*argl)[3]->DefaultType () != C_XPLDOUBLE)
       {
         globalError.Set (ERR_NUM, "spline: fourth argument");
         return (NULL);
       }
       lambda = (*argl)[3]->DefaultXplDouble ();
      case 3:
        if ((*argl)[2]->DefaultType () != C_XPLDOUBLE)
        {
          globalError.Set (ERR_NUM, "spline: third argument");
          return (NULL);
        }
        xnew = (*argl)[2]->DefaultXplDouble ();
      default:
        if ((*argl)[1]->DefaultType () != C_XPLDOUBLE)
        {
          globalError.Set (ERR_NUM, "spline: second argument");
          return (NULL);
        }
        y = (*argl)[1]->DefaultXplDouble ();
        if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
        {
          globalError.Set (ERR_NUM, "spline: first argument");
          return (NULL);
        }
        x = (*argl)[0]->DefaultXplDouble ();
     }
   
   dx = x->getptr();  
   dy = y->getptr();  
   int  xnewcols,xnewrows;
   xrows=x->dimension(0);
   yrows=y->dimension(0);
   xcols=x->dimension(1);
   ycols=y->dimension(1);

   if (xnew!=NULL)
   { dxnew = xnew->getptr();
     xnewrows = xnew->dimension(0);
     xnewcols = xnew->dimension(1);
   }
   else
   { dxnew = x->getptr();
     xnewrows = x->dimension(0);
     xnewcols = x->dimension(1);
   }

   if (lambda!=NULL)
   { dlambda = (*lambda)(0);
   }
   else
     dlambda = -1;
   
   if (weight!=NULL)
   { dweight = weight->getptr();
   }
   else
   { dweight = new double[x->dimension(0)];
     for (i=0; i<x->dimension(0); i++)
       dweight[i] = 1.0;
   }

/*  if (xcols!=1)
    {
    globalError.Set (ERR_NUM, "spline: first argument has to be a vector");
    return (NULL);
    }
  
  if (ycols!=1)
    {
    globalError.Set (ERR_NUM, "spline: second argument has to be a vector");
    return (NULL);
    }

 if (xrows!=yrows)
    {
    globalError.Set (ERR_NUM, "spline: first and second argument
         must be of the same dimension");
    return (NULL);
    }
    
 if (xnewcols!=1)
    {
    globalError.Set (ERR_NUM, "spline: third argument has to be a vector");
    return (NULL);
    }

   if ((lambdacols!=1) && (lambdarows!=1))
    {
    globalError.Set (ERR_NUM, "spline: fourth argument must be a scalar");
    return (NULL);
    }         
 */              
  double *w = new double[xrows];
  long templength=tie(dx,dy,dweight, w, xrows);
  delete[] w;
  if (templength<4)
  { globalError.Set (ERR_NUM, "spline: number of distinct points of x must be > 3");
    return (NULL);
  }

  ynew = new XplDouble(xnewrows, 1);
  dynew = new double[long(xnewrows)];
  double* D2y;
  D2y = new double[templength];
  if(dlambda==0)
     splinec(dx,dy,D2y,templength);
  if(dlambda>0)
     splsm_(dx,dy,D2y, dweight, templength,dlambda);
  if(dlambda<0)
    splad_(dx,dy,D2y, dweight, templength); 

  splint_(dx,dy,D2y,templength,dxnew,dynew,long(xnewrows));     
  
  delete D2y;
  if (globalError.Get())
  {
    delete ynew;
    return (NULL);
  }

  ynew->setptr(dynew);

  return (new XplObject (ynew, "spline"));
}

XplObject* xplqr(Liste<XplObject*> *argl)
{
  long      numb = argl->laenge(), n, i, *jpvt, job;
  XplDouble *x, *y, *mx,  *mqraux, *mjpvt;
  double *qrx, *djpvt,  *qraux;
  XplIndex ind;
  XplObject *o;

  if ((numb!=1) && (numb!=2))
    {
    globalError.Set(ERR_ARGS, "qr: Must = 1 or 2");
    return (NULL);
    }

/*    if (numb!=1)
    {
    globalError.Set(ERR_ARGS, "qr: Must = 1");
    return (NULL);
    }
*/
    if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "qr: first argument");
    return (NULL);
    }
    x = (*argl)[0]->DefaultXplDouble ();

   jpvt  = new long[x->dimension(1)];
   for (i=0; i<x->dimension(1); i++)
     jpvt[i] = i;

   if (numb>1)
   { if ((*argl)[1]->DefaultType () != C_XPLDOUBLE)
      {
      globalError.Set (ERR_NUM, "qr: second argument");
      return (NULL);
      }
      y = (*argl)[1]->DefaultXplDouble ();
      for (i=0; i<x->dimension(1); i++)
        jpvt[i] = (long) ((*y)(i));
   }
   else
     y = NULL;

   ind = x->getdim();
   n   = x->dimension(0)*x->dimension(1);
   qrx = new double[n];
   memcpy(qrx, x->getptr(), n*sizeof(double));

   qraux = new double[x->dimension(1)];
   job = (y!=NULL);
//   job = 0;
   sqrdc_(qrx, x->dimension(0), x->dimension(0), x->dimension(1),
          qraux, jpvt, job);

   mx = new XplDouble(1);
   mx->setdim (ind);
   mx->setptr (qrx);

   ind.setdim(0, ind(1));
   ind.setdim(1, 1);
   mqraux = new XplDouble(1);
   mqraux->setdim (ind);
   mqraux->setptr (qraux);

   djpvt = new double[x->dimension(1)];
   for (i=0; i<x->dimension(1); i++)
     djpvt[jpvt[i]] = i+1;
   delete[] jpvt;
   mjpvt = new XplDouble(1);
   mjpvt->setdim (ind);
   mjpvt->setptr (djpvt);

  o = new XplObject ("qrres");
  o->AppendComponent (new Component (new XplObject (mx, "qrx")));
  o->AppendComponent (new Component (new XplObject (mqraux, "qraux")));
  o->AppendComponent (new Component (new XplObject (mjpvt, "jpvt")));

//  o->SetDefaultXplDouble ();    // only for sure

  return (o);

// return (NULL);
}

XplObject* xplqrqr(Liste<XplObject*> *argl)
{
  long      numb = argl->laenge(), n, i;
  XplDouble *x, *y,  *mq, *mr;
  double *qrx, *q, *r,  *qraux;
  XplIndex ind;
  XplObject *o;
  Component *c=NULL;

  if ((numb!=1) && (numb!=2))
    {
    globalError.Set(ERR_ARGS, "qrmat: Must 1 or 2");
    return (NULL);
    }

  if (numb==2)
  {
    if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "qrmat: first argument");
    return (NULL);
    }
    x = (*argl)[0]->DefaultXplDouble ();
   
    if ((*argl)[1]->DefaultType () != C_XPLDOUBLE)
      {
      globalError.Set (ERR_NUM, "qrmat: second argument");
      return (NULL);
      }

    y = (*argl)[1]->DefaultXplDouble ();
  }
  else
  { c   = (*argl)[0]->GetComponent("qrx");
    if (c)
    { x = c->Content().xplobject->DefaultXplDouble ();
      if (x==NULL) 
      { globalError.Set (ERR_NUM, "qrmat: numeric argument for l.x expected");      
        return NULL;
      }
    }  
    c   = (*argl)[0]->GetComponent("qraux");
    if (c)
    { y = c->Content().xplobject->DefaultXplDouble ();
      if (y==NULL) 
      { globalError.Set (ERR_NUM, "qrmat: numeric argument for l.qraux expected");
        return NULL;
      }
    }  
  }

   ind = x->getdim();
   n   = x->dimension(0)*x->dimension(1);
//   qrx = new double[n];
//   memcpy(qrx, x->getptr(), n*sizeof(double));

   qraux = new double[x->dimension(1)];
   memcpy(qraux, y->getptr(), x->dimension(1)*sizeof(double));
 
   q = new double[x->dimension(0)*x->dimension(0)];
   r = new double[x->dimension(0)*x->dimension(1)];

   qrQ (x->getptr(), qraux, x->dimension(0), x->dimension(1), q, 1.0e-12);
   qrR (x->getptr(), x->dimension(0), x->dimension(1), r);

   delete[] qraux;
  
   mr = new XplDouble(1);
   mr->setdim (ind);
   mr->setptr (r);

   ind.setdim(1, ind(0));
   mq = new XplDouble(1);
   mq->setdim (ind);
   mq->setptr (q);

  o = new XplObject ("qrmatres");
  o->AppendComponent (new Component (new XplObject (mq, "q")));
  o->AppendComponent (new Component (new XplObject (mr, "r")));
  o->SetDefaultXplDouble ();    // only for sure

  return (o);

}

extern Vector<Function*> *funList;

int foreigncmd()
{
  funList->haengeAn (new Function ("spline",  spline));    // StSp YG 20.04.98
  funList->haengeAn (new Function ("chold",  cholchol));   // SK 09.02.98
  funList->haengeAn (new Function ("qr",  xplqr));         // Yg + SK 09.02.98
  funList->haengeAn (new Function ("qrmat",  xplqrqr));    // Yg + SK 09.02.98
  return 0;
}

/*  o1 = new XplObject (xb, "xb");
  o2 = new XplObject (yb, "yb");

  o  = new XplObject ("bindata");

  o->AppendComponent (new Component (o1));
  o->AppendComponent (new Component (o2));

  o->SetDefaultXplDouble ();    // only for sure

  return (o);

*/
