/* $Header: /home/xplore/cvs/xpl4gr/genar.C,v 1.4 2002/11/18 15:15:34 cizek Exp $ */

// Generation of autoregressive sequences

// PC 20021118

#include<stdio.h>
#include<math.h>

#include "object.h"   
#include "xplexec.h"
#include "xplerror.h"

#ifndef MATRIX
#define rows() dimension(0)
#define cols() dimension(1)
#define out()  dump("%lg")
#endif


XplDouble* genar(XplDouble x, XplDouble startval, XplDouble alpha)

/*
input:
x        : n x d2 x ... x d8 array of exogenous variables plus noise
startval : p x d2 x ... x d8 array of starting values
alpha    : p x d2 x ... x d8 array of autoregressive parameters

output:
genar    : (n x 1) vector generated as 
	   y_t = x_t + alpha[1]*y_(t-1)+...+alpha[p]*y_(t-p), t=p+1,...,n
	   y_t = startval, t=1,...,p
*/

{ 

  long   i, j, k, ks, ka, t, p, n, ns, nsa, nss;
  double s;

  n = x.rows ();
  ns = x.dim.prod() / n;
  p = alpha.rows ();
  nsa = alpha.dim.prod() / p;
  nss = startval.dim.prod() / p;

// error control

  if ( startval.rows () != p )
  {
    globalError.Set(ERR_GENAR, "the number of rows of alpha and startval must be equal");
	 return(NULL);
  }


  XplDouble y(x);


/*
    for (t=0; t<p; t++)
	{
	y(t) = startval(t);
	}
*/

  double *yptr = y.getptr(), *ypom;
  double *aptr = alpha.getptr(), *apom;
  double *sptr = startval.getptr(), *spom;

  for (k = ka = ks = 0; k < ns; k++, ka++, ks++)
  {
    if (ka >= nsa)
    {
      aptr = alpha.getptr();
      ka = 0;
    }
    if (ks >= nss)
    {
      sptr = startval.getptr();
      ks = 0;
    }

    for(i = 0; i < p; i++, yptr++, sptr++)
      *yptr = *sptr;
      
    for(i = p; i < n; i++, yptr++)
	 {
	   s = 0;
      apom = aptr;
      ypom = yptr - 1;
	   for (j = 0; j < p; j++, apom++, ypom--)
		{
		   s += *apom * *ypom;
		}
      *yptr += s;
	 }
    aptr += p;
  }

  return (new XplDouble (y));
}

