/* $Header: /home/xplore/cvs/xpl4gr/initxa.C,v 1.3 2002/08/16 12:31:44 sigbert Exp $ */

#include <math.h>
#include <time.h>
#include <string.h>
#ifdef __BORLANDC__
#include <io.h>
#else
#include <unistd.h>
#endif
#include "values.h"
#include "liste.h"
#include "xplerror.h"
#include "xarray.hpp"
#include "global.h"
#include "xnan.h"
#include "directory.h"
#include "xplore_strings.h"

#ifdef HAS_STRCMPI
#define strcasecmp strcmpi
#endif

char *Smonth[12] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun",
                   "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
char *Sday[7] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };

double randomval, nrand, iszero = 0.0;
int nset;
char filename[2048], xpl4datatok[4096];
// *xpl4data=NULL,
//     *outheadline, *outlineno, *outlayerline;
// long outmaxdata;
XMemory xmem;
XplStringDatabase xplorebase;
//Info recinfo;

#ifdef SKDEBUG
char *xpl4home = NULL;
#else
#include "xnewf.h"

extern char *xpl4home;
#endif

void InitializeXplArray (void)
{ static int initnec = 1;

  if (initnec)
  { globalWarning.append (new XplError (OPCON,        "Dimensions not conformable"));

    // globalError.append(new XplError (,));

    globalError.append(new XplError (NOMEM,   "No memory"));
    globalError.append(new XplError (DIMBIG,  "Dimension too big"));
    globalError.append(new XplError (DIMCON,  "Dimensions are not conformable"));
    globalError.append(new XplError (NONSORT, "No sorting dimension"));
    globalError.append(new XplError (DIMELEM, "Entry number larger than dimension"));
    globalError.append(new XplError (DIMNEG,  "Dimension negative"));
    globalError.append(new XplError (STRBAD,  "Stream bad"));
    globalError.append(new XplError (NOCAST,  "Casting forbidden"));
    globalError.append(new XplError (NOCONV,  "No convergence"));
    globalError.append(new XplError (RANGE,   "Range too large"));

    globalError.append(new XplError (FUNCCALL,"Quantlet call failed"));
    globalError.append(new XplError (FORMAT,  "Format wrong"));
    globalError.append(new XplError (PARAMFALSE,  "Parameter wrong"));
    globalError.append(new XplError (STRINGBAD,  "String bad"));
    globalError.append(new XplError (ARRAYBAD,  "Array bad"));
    globalError.append(new XplError (TIMEGRAN, "Granulation not supported"));
    globalError.append(new XplError (NOUSE,    "This routine should never be used"));
    globalError.append(new XplError (NN, "nn..."));
    globalError.append(new XplError (UNEXVAL,  "Unexpected value"));
    globalError.append(new XplError (TIES,     "Ties found"));
    globalError.append(new XplError (NNR,      "nnr..."));
    globalError.append(new XplError (OBSOLETE, "Obsolete button/command"));
    // PC 200111127 - number of observations from constant STRSMALLVERSIONOBSERVATIONS
    globalError.append(new XplError (NOSUPPORT, "This version does not support more than " STRSMALLVERSIONOBSERVATIONS " rows per matrix"));
    globalError.append(new XplError (INTERN,    "Internal error - contact mdtech@mdtech.de")); 
    globalError.append(new XplError (LIST,     "List error"));
    globalError.append(new XplError (TIME,     "Time error"));
    globalError.append(new XplError (NULMEM,    XMSG_NULMEM_STR));
/*    outheadline  = strdup ("Contents of %s\r\n\r\n");
    outlayerline = strdup ("[,,%li,%li,%li,%li,%li,%li]\r\n");
    outlineno    = strdup ("[%*li,] ");    
    outmaxdata   = 2048;
*/
    initnec = 0;
  }
}

#ifdef __BORLANDC__           //SWS 16.09
#include <math.h>
int _matherr (struct exception *a)
{
	 return 1;
}
#endif

void SetNaN (double    *X) { *X = ::sqrt(-1); }
void SetNaN (float     *X) { *X = ::sqrt(-1); }
void SetNaN (xplmask   *X) { *X = -1L; }
void SetNaN (xplstring *X) { X->setnull(); }
void SetNaN (long      *X) { *X = -1L; }
void SetNaN (int       *X) { *X = -1L; }

void x4sprintf (char *buf, const char*fmt, double X)  
{ sprintf (buf, fmt, X); }
void x4sprintf (char *buf, const char*fmt, float  X)  
{ sprintf (buf, fmt, X); }
void x4sprintf (char *buf, const char*fmt, long   X)        
{ sprintf (buf, fmt, X); }
void x4sprintf (char *buf, const char*fmt, int    X)        
{ sprintf (buf, fmt, X); }
void x4sprintf (char *buf, const char*fmt, short  X)        
{ sprintf (buf, fmt, X); }
void x4sprintf (char *buf, const char*fmt, unsigned long  X)
{ sprintf (buf, fmt, X); }
void x4sprintf (char *buf, const char*fmt, unsigned int   X)
{ sprintf (buf, fmt, X); }
void x4sprintf (char *buf, const char*fmt, unsigned short X)
{ sprintf (buf, fmt, X); }
void x4sprintf (char *buf, const char*fmt, xplmask    X)    
{ sprintf (buf, fmt, X); }
void x4sprintf (char *buf, const char*fmt, xplstring  X)    
{ char *str = X.tostr();
  sprintf (buf, fmt, str);
}

void x4fprintf (FILE *buf, const char*fmt, double X)  
{ fprintf (buf, fmt, X); }
void x4fprintf (FILE *buf, const char*fmt, float  X)  
{ fprintf (buf, fmt, X); }
void x4fprintf (FILE *buf, const char*fmt, long   X)        
{ fprintf (buf, fmt, X); }
void x4fprintf (FILE *buf, const char*fmt, int    X)        
{ fprintf (buf, fmt, X); }
void x4fprintf (FILE *buf, const char*fmt, short  X)        
{ fprintf (buf, fmt, X); }
void x4fprintf (FILE *buf, const char*fmt, unsigned long  X)
{ fprintf (buf, fmt, X); }
void x4fprintf (FILE *buf, const char*fmt, unsigned int   X)
{ fprintf (buf, fmt, X); }
void x4fprintf (FILE *buf, const char*fmt, unsigned short X)
{ fprintf (buf, fmt, X); }
void x4fprintf (FILE *buf, const char*fmt, xplmask    X)    
{ }
void x4fprintf (FILE *buf, const char*fmt, xplstring  X)    
{ fprintf (buf, fmt, X.tocstr()); }

int  isNaN  (double X)  
{ return (XisNaN(X)); }

int  isNaN  (float  X)  
{ long *l = (long *) &X;
  return (l[0] == 0x80000000);
}

int  isNaN  (long   X)          { return (0); }
int  isNaN  (int    X)          { return (0); }
int  isNaN  (short  X)          { return (0); }
int  isNaN  (unsigned long  X)  { return (0); }
int  isNaN  (unsigned int   X)  { return (0); }
int  isNaN  (unsigned short X)  { return (0); }
int  isNaN  (xplmask    X)      { return (0); }
int  isNaN  (xplstring  X)      { return (X.xplisnull()); }

void SetMin (double *X) { *X = -MAXDOUBLE; }
void SetMin (float  *X) { *X = -MAXFLOAT; }
void SetMin (long   *X) { *X = -MAXLONG; }
void SetMin (int    *X) { *X = -MAXINT; }
void SetMin (short  *X) { *X = -MAXSHORT; }
void SetMin (unsigned long   *X) { *X = 0; }
void SetMin (unsigned int    *X) { *X = 0; }
void SetMin (unsigned short  *X) { *X = 0; }

void SetMax (double *X) { *X = MAXDOUBLE; }
void SetMax (float  *X) { *X = MAXFLOAT; }
void SetMax (long   *X) { *X = MAXLONG; }
void SetMax (int    *X) { *X = MAXINT; }
void SetMax (short  *X) { *X = MAXSHORT; }
void SetMax (unsigned long   *X) { *X = MAXLONG<<1; }
void SetMax (unsigned int    *X) { *X = MAXINT<<1; }
void SetMax (unsigned short  *X) { *X = MAXSHORT<<1; }

int  isZero (double X)  
{ if (iszero==0.0) return (X==0.0);
  return (::fabs(X)<iszero);
}

int  isZero (float  X)
{ if (iszero==0.0) return (X==0.0);
  return (::fabs(X)<iszero);
}

void SetZero (double X) 
{ if (X<0.0)
  { int i;
    for (i=0; i<100; i++)
    { 
#ifdef __BORLANDC__
      iszero = ::pow10(-i);
#else
      iszero = ::pow(10.0, -i);
#endif
      if ((1.0+iszero)==1.0) break;
    }
  }
  else
    iszero = ::fabs (X); 
}

double GetZero (void)   
{ return (iszero); }

long memcopy (xplstring* left, xplstring* right, long elems)
{ long i;
#if SKDEBUG
  printf ("memcopy %li xplstring elements\n", elems);
#endif
  for (i=0; i<elems; i++)
    *(left+i) = *(right+i);
  return(i);
}

long memcopy (double* left, double* right, long elems)
{ long i;
#if SKDEBUG
  printf ("memcopy %li double elements\n", elems);
#endif
  for (i=0; i<elems; i++)
    *(left+i) = *(right+i);
  return(i);
}

long memcopy (long* left, long* right, long elems)
{ long i;
#if SKDEBUG
  printf ("memcopy %li long elements\n", elems);
#endif
  for (i=0; i<elems; i++)
    *(left+i) = *(right+i);
  return(i);
}

long memcopy (int* left, int* right, long elems)
{ long i;
#if SKDEBUG
  printf ("memcopy %li int elements\n", elems);
#endif
  for (i=0; i<elems; i++)
    *(left+i) = *(right+i);
  return(i);
}

long memcopy (char* left, char* right, long elems)
{ long i;
#if SKDEBUG
  printf ("memcopy %li char elements\n", elems);
#endif
  for (i=0; i<elems; i++)
    *(left+i) = *(right+i);
  return(i);
}

long memcopy (xplmask* left, xplmask* right, long elems)
{ long i;
#if SKDEBUG
  printf ("memcopy %li xplmask elements\n", elems);
#endif
  for (i=0; i<elems; i++)
    *(left+i) = *(right+i);
  return(i);
}

extern void logt (const char *txt);

FILE *opendatafile (const char *datei, const char *how)
{ FILE *f = NULL; 
  char *fn;

  fn = openfile (DATA, datei, how);
  //  logt(fn);
  if (fn) f = fopen(fn, how);
  return (f);
}

int removedatafile (const char *datei)
{ char *fn;

  fn = openfile (DATA, datei, "r");
  if (fn==NULL) return (-1);
  return (unlink(fn));
}


// template <class T> long xpltypeid (T* x)
// { return (TYPE_UNKNOWN); }

long xpltypeid (double *X)
{ return (TYPE_DOUBLE); }

long xpltypeid (float  *X)
{ return (TYPE_FLOAT); }

long xpltypeid (long   *X)
{ return (TYPE_LONG); }

long xpltypeid (int    *X)
{ return (TYPE_INT); }

long xpltypeid (short  *X)
{ return (TYPE_SHORT); }

long xpltypeid (unsigned long   *X)
{ return (TYPE_ULONG); }

long xpltypeid (unsigned int    *X)
{ return (TYPE_UINT); }

long xpltypeid (unsigned short  *X)
{ return (TYPE_USHORT); }

long xpltypeid (xplmask    *X)
{ return (TYPE_MASK); }

long xpltypeid (xplstring  *X)
{ return (TYPE_STRING); }

double x4atof (char *tok)
{ char *atofstr = strdup(tok);
  double no;
  int i, mode;
  if ((strcasecmp(tok, ".")==0) || (strcasecmp(tok, "nan")==0) || (strcasecmp(tok, "na")==0))
  { delete[] atofstr;
    return (sqrt(-1));
  }
  for (mode=i=0; i<strlen(tok); i++)
  { // printf ("%i %c\n", mode, tok[i]);
    switch (mode)
    { case  0 :
        mode = -1; 
        if (strchr("+-", tok[i]))
          mode = 1;
        else if (strchr("0123456789", tok[i]))
          mode = 2;
        else if (strchr(".", tok[i]))
          mode = 3;
        else if (strchr(",", tok[i]))
        { mode = 3;
          atofstr[i] = '.';
        }
        break;
      case  1 :
        mode = -1; 
        if (strchr("0123456789", tok[i]))
          mode = 2;
        else if (strchr(".", tok[i]))
          mode = 3;
        else if (strchr(",", tok[i]))
        { mode = 3;
          atofstr[i] = '.';
        }
        break;
      case  2 :
        mode = -1; 
        if (strchr("0123456789", tok[i]))
           mode = 2;
        else if (strchr(".", tok[i]))
          mode = 3;
        else if (strchr(",", tok[i]))
        { mode = 3;
          atofstr[i] = '.';
        }
        else if (strchr("e", tok[i]))
          mode = 5;
        else if (strchr("EdD", tok[i]))
        { mode = 5;
          atofstr[i] = 'e';
        }
        break;
      case  3 :
        mode = -1; 
        if (strchr("0123456789", tok[i]))
          mode = 4;
        else if (strchr("e", tok[i]))
          mode = 5;
        else if (strchr("EdD", tok[i]))
        { mode = 5;
          atofstr[i] = 'e';
        }
        break;
      case  4 :
        mode = -1;
        if (strchr("0123456789", tok[i]))
          mode = 4;
        else if (strchr("e", tok[i]))
          mode = 5;
        else if (strchr("EdD", tok[i]))
        { mode = 5;
          atofstr[i] = 'e';
        }
        break;
      case  5 :
        mode = -1; 
        if (strchr("0123456789", tok[i]))
          mode = 7;
        else if (strchr("-+", tok[i]))
          mode = 7;
        break;
      case  7 :
        mode = -1; 
        if (strchr("0123456789", tok[i]))
          mode = 7;
        break;
      default :
        i = strlen(tok);
    }  
  }
  no = atof(atofstr);
  delete[] atofstr;
  if (mode<0)
    return (sqrt(-1));
  return (no);
}

  
char* format (double *X)         { return (strdup("%g")); }
char* format (float  *X)         { return (strdup("%g")); }  
char* format (long   *X)         { return (strdup("%li")); }        
char* format (int    *X)         { return (strdup("%i")); }        
char* format (short  *X)         { return (strdup("%i")); }        
char* format (unsigned long  *X) { return (strdup("%li")); }
char* format (unsigned int   *X) { return (strdup("%i")); }
char* format (unsigned short *X) { return (strdup("%i")); }
char* format (xplmask    *X)     { return (strdup("R=%i G=%i B=%i SYM=%i S=%i "));  }  
char* format (xplstring  *X)     { return (strdup("%s")); }    

char sktmpnam[512];

void makefmtstring (char *usesep)
{ char *separator;
  int i, j;

  if (usesep==NULL) return;
  separator = strdup(usesep);

  for (j=i=0; i<=strlen(separator); i++, j++)
  { if (separator[i]=='\\')
    { i++;
      if (separator[i]=='t')
        usesep[j] = '\t';
      else if (separator[i]=='n')
        usesep[j] = '\n';
      else if (separator[i]=='r')
        usesep[j] = '\r';
      else if (separator[i]=='\\')
        usesep[j] = '\\';
      else
      { i--;
        usesep[j] = separator[i];
      }
    }
    else
      usesep[j] = separator[i];
  }

  delete[] separator; 
 
}


char *maketxtstring (char *separator)
{ char *usesep;
  int i, j;

  if (separator==NULL) return(NULL);

  usesep = new char[2*strlen(separator)+1];

  for (j=i=0; i<=strlen(separator); i++, j++)
  { if (separator[i]=='\t')
    { usesep[j]   = '\\';
      usesep[j+1] = 't';
      j++;
    }
    else if (separator[i]=='\n')
    { usesep[j]   = '\\';
      usesep[j+1] = 'n';
      j++;
    }
    else if (separator[i]=='\r')
    { usesep[j]   = '\\';
      usesep[j+1] = 'r';
      j++;
    }
    else if (separator[i]=='\\')
    { usesep[j]   = '\\';
      usesep[j+1] = '\\';
      j++;
    }
    else
      usesep[j] = separator[i];
  }

  return (usesep);
}


