/* $Header: /home/xplore/cvs/xpl4gr/liste.h,v 1.3 2002/08/16 12:31:45 sigbert Exp $ */

#ifndef	_LISTE_H_
#define _LISTE_H_

#include <stdlib.h>
#include <stdio.h>

//  Klasse Liste                                                         
//  Der Typ der Listenelemente ist beliebig

//  Elementfunktionen:                                                   

//  int fuegeEin(const T& x, int i)
//           fuegt x als (i + 1)-te Komponente in die Liste ein
//           liefert 1, falls erfolgreich, 0 sonst                       

//  int haengeAn (const T& x)
//           haengt x als letzte Komponente in die Liste an
//           liefert 1, falls erfolgreich, 0 sonst                       

//  int loesche(int i)
//           loescht (i + 1)-te Komponente aus der Liste
//           liefert i, falls erfolgreich, 0 sonst

//  int laenge()
//           liefert die aktuelle Anzahl der Listenelemente              

//  der Indexoperator [] ist fuer const und nicht-const
//           Listenobjekte definiert
//           Ein Zugriff mit falschem Index ruft exit(1) auf             

//  Iteratoren:                                                          

//  ListElem<T>* anfang()
//          liefert einen Zeiger auf den Listenanfang

//  ListElem<T>* iter()
//          positioniert den Zeiger auf das folgende Element             
//          liefert am Ende der Liste NULL

template<class T> struct ListElem 
{
    ListElem(const T& e, ListElem* n = NULL) :
        element(e), nachf(n) { }
    T element;
    ListElem* nachf;
};


                                                                         
template<class T> class Liste                                            
{
public:                                                                  
    Liste () : start(NULL), anzahl(0) { anfang(); }
    virtual ~Liste() { dest(); }
    Liste (const Liste& m) : start(NULL), anzahl(m.anzahl) { copy(m); }
    Liste& operator=(const Liste&);
    int fuegeEin(const T&, int = 0);
    int haengeAn(const T& x)	{ return (fuegeEin (x, anzahl)); }
    int loesche(int);
    T& operator[] (int) ;
    const T& operator[](int) const;
    int laenge() const { return anzahl; }
    void remove ();
protected:
    ListElem<T>* anfang() { aktInd = 0; return (aktZgr = start); }
    ListElem<T>* iter() { return (aktZgr != NULL) ?
        aktInd++, aktZgr = aktZgr->nachf : (ListElem<T>*) NULL; }
private:
    void dest();
    void copy(const Liste&);
    ListElem<T>* start;
    int anzahl;
    ListElem<T>* aktZgr;
    int aktInd;
};


// void error(char*  str =  "")
// {
//     cout << str << endl;
//     exit ( 1 ) ;
// }

void error(char*  str =  "");


template<class T> void Liste<T>::dest()
{
    while  (start  != NULL)  {
        ListElem<T>* tmp = start->nachf;
        delete start;
        start = tmp ;
        }
 }



template<class T> void Liste<T>::copy(const Liste<T>& m)
{
    if ((anzahl = m.anzahl) == 0) {
       anfang();
       return;
       }

    aktInd = m.aktInd;
    ListElem<T>* mZgr = m.start;
    ListElem<T>* z = aktZgr = start = new ListElem<T>(mZgr->element);
    while((mZgr = mZgr->nachf) != NULL)
        {
        z = z->nachf = new ListElem<T>(mZgr->element);
        if (mZgr == m.aktZgr)
           aktZgr = z;
        }
 }

template<class T> Liste<T>& Liste<T>::operator=(const Liste<T>& m) 
{
    if (this == &m)                                                      
        return *this;                                                    
    dest();                                                              
    copy(m);
    return *this;
}
                                                                         
                                                                         
                                                                         
template<class T> int Liste<T>::fuegeEin(const T& x, int i)              
{                                                                         
    if (i < 0 || i > anzahl)
       return 0;
    ListElem<T>* neu = new ListElem<T>(x);                                
    if (neu == NULL)
        return 0;                                                        
    if  (i == 0)  {                                                      
        neu->nachf = start;                                              
        start = neu;                                                     
    } else {
       ListElem<T>* zgr = start;
       for  (int k = 0;  k < i - 1;  k++)
           zgr = zgr->nachf;
       neu->nachf = zgr->nachf;
       zgr->nachf = neu;
    }                                                                    
    anzahl++;
    anfang();
    return 1;
}

template<class T> int Liste<T>::loesche(int i)
{
    if (i < 0 || i >= anzahl)
       return 0;
    ListElem<T>* zgr2 = start;
    ListElem<T>* zgr1 = NULL;
    for (int k = 0; k < i; k++) {
        zgr1 = zgr2;
        zgr2 = zgr2->nachf;
    }

    ((zgr2 == start) ? start : zgr1->nachf) = zgr2->nachf;
    delete zgr2;
    anzahl--;
    anfang();
    return 1;
}

template<class T> T& Liste<T>::operator[](int i)
{
    if (i < 0 || i >= anzahl)
        error( "Liste [] :   Falscher Index" ) ;
    if (i < aktInd)
        anfang();
    else
        i -= aktInd;
    for(int k = 0; k < i; k++)
        iter() ;
    return aktZgr->element;
}

template<class T> const T& Liste<T>::operator[](int i) const
{
    if (i < 0 || i >= anzahl)
        error("Liste[] :   Falscher Index") ;
    ListElem<T>* zgr = start;
    for(int k = 0; k < i; k++)
        zgr = zgr->nachf;
    return zgr->element;
}


template<class T> void Liste<T>::remove (void)
{
//   while (tl.loesche(0) != 0) ;   geht wohl schneller
  int i, length;
  
  if (anzahl == 0)
     return;
     
  length = anzahl;
  
  for (i=0; i<length; i++)
    loesche(0);
}


// template class vector
// TTK, 18.2.97

template<class T> class Vector                                            
{
public:                                                                  
    Vector (long ilen=64L, long off=256L) : length(0), maxLength(ilen), offset(off), feld(new T[ilen]) {
#ifdef TTKDEBUG
if (ttkdebug)
cerr << "Tokl:\tconstruct\t" << this << '\t' << length << endl;
#endif
}
    virtual ~Vector() {
#ifdef TTKDEBUG
if (ttkdebug)
cerr << "Tokl:\tdestruct\t" << this << '\t' << length << endl;
#endif
 delete[] feld; }

//    Vector (const Vector& m) : length(m.length), maxLength(m.maxLength), offset(m.offset), feld(new T[m.maxLength]) { memcpy (feld, m.feld, sizeof(T)*m.maxLength); }
    Vector (const Vector& m) : length(m.length), maxLength(m.length), offset(m.offset), feld(new T[m.length]) {
#ifdef TTKDEBUG
if (ttkdebug)
cerr << "Tokl:\tconstruct\t" << this << '\t' << m.length << endl;
#endif
 memcpy (feld, m.feld, sizeof(T)*m.length); }
    Vector& operator=(const Vector&);
    void haengeAn(const T& x);
//    T& operator[] (const long i)  { return (feld[i]); }
    T& operator[] (const long i);
//    const T& operator[] (const long i) const  { return (feld[i]); }
    const T& operator[] (const long i) const ;
    long laenge() const { return (length); }
    void remove ()  { length = 0; }

protected:
    long length, maxLength, offset;
    T    *feld;

};

template<class T> Vector<T>& Vector<T>::operator= (const Vector& v)
{
  if (this == &v)
    return (*this);

  if (maxLength != v.length)
    {
    delete[] feld;
    feld = new T[v.length];
    }

  length    = v.length;
  maxLength = v.length;
  offset    = v.offset;

  memcpy (feld, v.feld, sizeof(T)*length);
  return (*this);
}

template<class T> void Vector<T>::haengeAn (const T& x)
{
  if (length < maxLength)
    {
    feld[length++] = x;
    return;
    }

  maxLength += offset;

  T *feld2 = new T[maxLength];

  memcpy (feld2, feld, sizeof(T)*length);
  delete[] feld;
  feld = feld2;
  feld[length++] = x;
}

template<class T> const T& Vector<T>::operator[] (const long i) const
{ 
   if (i>=length)
     {
     cerr << "FatalError [] Vector" << endl;
     exit (1);
     }
     
   return (feld[i]);
}

template<class T>  T& Vector<T>::operator[] (const long i)
{ 
   if (i>=length)
     {
     cerr << "FatalError [] Vector" << endl;
     exit (1);
     }
     
   return (feld[i]);
}

#endif
