/* $Header: /home/xplore/cvs/xpl4gr/locpol.C,v 1.3 2002/08/16 12:31:45 sigbert Exp $ */

#ifndef MATRIX
#include <math.h>
#include <string.h>
#include "x4math2.h"

#include "machine.h"

#include "object.h"
#include "function.h"
#include "xplfun.h"
#include "xplerror.h"
#include "init.h"   //                      ????
#include "ttkio.h"
#include "xplexec.h"
#include "matrix.h"


extern "C" int loclin1 
	(double *x_in, double *xg_in, double *y_in,
         double *h_in, double *par_in,         
         double *ma_out);


XplDouble stefan1 (XplDouble *x, XplDouble *xg, XplDouble *y, 
                   XplDouble *h, XplDouble *logi_in )

{ 
	int dim=x->dimension(1);
	int rows=x->dimension(0);
	int dimg=xg->dimension(1);
	int rowg=xg->dimension(0);
	int dimy=y->dimension(1);
	int rowy=y->dimension(0);

        XplDouble rr(1);
	
	if (rowy != rows)
	  {
	  globalError.Set (ERR_LOCPOL, "x and y are not conformable!");
	  return (rr);
	  }
	
	if (h->dimension(0) != dim)
	  {
	  globalError.Set (ERR_LOCPOL, "x and h are not conformable!");
	  return (rr);
	  }
	
	if (h->dimension(0) != dimg)
	  {	
	  globalError.Set (ERR_LOCPOL, "xg and h are not conformable!");
	  return (rr);
	  }

	int logi=(int)( (*logi_in)[0]);
	if (! ( logi==0 || logi==1)) 
	  {	
	  globalError.Set (ERR_LOCPOL, "local linear=0, quadratic=1");
	  return (rr);
	  }	
	double *x_in, *xg_in, *y_in, *h_in, *par_in;
	
	int i, j, l, dd;
	
	x_in= new double [dim*rows];
	for (i=0; i<rows; i++)
	for (j=0; j<dim; j++)
    		x_in[i*dim+j]=(*x)(i, j);	

	xg_in= new double [dimg*rowg];
	for (i=0; i<rowg; i++)
	for (j=0; j<dimg; j++)
		xg_in[i*dimg+j]=(*xg)(i, j);	
			
	y_in= new double [dimy*rows]; 
	for (i=0; i<rows; i++)
	for (j=0; j<dimy; j++)
		y_in[i*dimy+j]=(*y)(i, j);
		
	par_in= new double [5];
	 par_in[0]= rows ;
	 par_in[1]= rowg ;
	 par_in[2]= dim  ;
	 par_in[3]= dimy ;
	 par_in[4]= logi ;

	h_in= new double [dim];
	for (i=0; i<dim; i++)
		h_in[i]=(*h)(i);

        dd  = (int)(1+dim+logi*dim*0.5*(dim+1)) ;
	
	double *ma_out = new double[rowg*dd*dimy];
	
	loclin1(x_in, xg_in, y_in, h_in, par_in, ma_out);

	XplDouble result(rowg, dd, dimy);
	
	for (i=0; i<rowg; i++)
	for (j=0; j<dd; j++)
        for (l=0; l<dimy; l++)
	  {     result(i, j, l)=ma_out[l*rowg*dd+j*rowg+i];
	  }

	XplObject *o =new XplObject( new XplDouble(result), "locpol");
	
	delete[] x_in;
	delete[] xg_in;
	delete[] y_in;
	delete[] h_in;
	delete[] par_in;

	delete[] ma_out;
	
	return(result);
	
}


#ifdef HAS_TIMEH 
#include <time.h>
#else
#include <sys/time.h>
#include <sys/resource.h>
#endif

#define MAX_CORE  (1L << 20L)    /* 1 MB */

#include "matrix.h"
#include <stdio.h>
#include <math.h>
#include <stdlib.h>

/*   **************************************
      C ++ Program       loclin1.C
      
      calculates local polynomials

     **************************************    */
/* 
*  Hints for compilation and necessary c or C files:
*
   g++ -G -O2 -c -DMATRIX matrix.c    	this compilies matrix.c library
   gcc -G -O2 -c dl*.c		        compiles dlubksb.c, dludcmp.c
   gcc -G -O2 -c dja*.c			compiles djacobi.c
   gcc -G -O2 -c nrutil.c		compiles routines from 
					Numerical Recepies
  gcc -G -O2 loclin1.C matrix.o dl*.o -o loclin1.so -lm
 gcc -G -O2 loclin1.C matrix.o dl*.o dja*.o nrutil.o -o loclin1.so -lm
					compiles loclin.C and links it with
					all the other routines
*  INPUTS :
*   xorig = original designmatrix
*   xg    = grid, the points where we want to estimate
*   yorig = original dependent observations
*   h     = bandwidth vector
*   par   = vektor containing
*     n     = number of observations
*     ng    = number of gridpoints
*     d     = number of dimensions with continuous vars.
*     q     = dimension of yorig < 4  !!
*     lorq  = control variable whether linear or quadratic
*  OUTPUT :
*   ma_out= (ng*dd*q) matrix of results
*
*/

extern "C"
{

double p_gauss(long p, double *k)
{ long i;
  double q = 1.0, zpi=6.28318, r;
  
  for (i=0; i<p; i++)
  { r = k[i]*k[i];
    if (r>7.0) return 0.0;
    r = r*(-0.5);
    r = x4exp(r)/sqrt(zpi) ;
    q *= r;     /* q = q * r  */ 
   }
   
  return q;
}

double p_quartic(long p, double *k)
{ long i;
  double q = 1.0, r;
  
  for (i=0; i<p; i++)
  { r = k[i]*k[i];
    if (r>1.0) return 0.0;
    r = 1.0-r;
    r = 0.9375*r*r;
    q *= r;     /* q = q * r  */ 
   }
     return q;
}


int loclin1 (double *x_in, double *xg_in, double *y_in, 
            double *h_in, double *par_in, 
            double *ma_out)

{  long n, ng, d, dd, q, i, j, k, l, ll, lorq, start;
   double  *kern, *w, dd1;

#ifdef HAS_CORE
   struct rlimit   rlp;

   rlp.rlim_cur = MAX_CORE;
   rlp.rlim_max = MAX_CORE;

   setrlimit (RLIMIT_CORE, &rlp);  /* sets core limit  */
#endif

   n	= (long) *(par_in+0);	    /* number of samples */
   d	= (long) *(par_in+2);	    /* number of regressors */
   ng   = (long) *(par_in+1);       /* number of gridpoints */
   q    = (long) *(par_in+3);       
   lorq = (long) *(par_in+4);     /* parameter that controls whether 
			              local linear (lorq = 0) or
			          quadratic (= 1) estimation is used */
   dd1  = 1+d+lorq*d*0.5*(d+1) ;
   dd   = (long)(1+d+lorq*d*0.5*(d+1)) ; 

   kern = new double[sizeof(double)*d];  	// initialize kern
   w	= new double[sizeof(double)*n];		// initialize W
       
//   Matrix Z(n,dd), ZWZ(dd,dd), ZWY(q,dd) ;  // initialize all matrices   
   XplDouble Z(n,dd), ZWZ(dd,dd), ZWY(q,dd) ;  // initialize all matrices   

   // do loop over gridpoints
   for (start=k=0; k<ng; k++)
   {
      for (j=0; j<dd; j++)     
      {
        for (l=0; l<q; l++)
	  ZWY(l,j) = 0 ; 
        for (l=0; l<j+1; l++)
	  ZWZ(l,j) = 0 ; 

      }  
  
      // do loop over observations
      for (i=start; i<n; i++)
      {           
 	 Z(i,0) = 1;   	                      // constant term
         Z(i,1) = *(x_in+i*d) - *(xg_in+k*d);    // first linear term 
 	 *(kern+0) = Z(i,1) / *(h_in) ;
         
         if (*(kern+0)<-1.0)
            start++ ;         
         else
         { if (*(kern+0)>1.0) break;
           for (j=1; j<d; j++)         
           {                       
               Z(i,j+1) = *(x_in+j+i*d) - *(xg_in+j+k*d);    // linear terms 
               *(kern+j) = Z(i,j+1) / *(h_in+j) ;        // prepare vector for kernel 
	   }	

            *(w+i) = p_quartic(d,kern) ;

	    if (lorq==1)
            {  
               ll = 1;
	       for (j=0; j<d; j++)
               { 
                 for (l=j; l<d; l++)           // quadratic and mixed terms
                 {           
                     Z(i,d+ll) = Z(i,j+1) *  Z(i,l+1) ;
                              ll++ ;
                 }
               }
            } 
            for (j=0; j<dd; j++)
            {
               ZWZ(0,j) += Z(i,j) * *(w+i) ;
               for (l=1; l< j+1; l++)
                  ZWZ(l,j) += Z(i,j)*Z(i,l)* *(w+i) ; 
               for (l=0; l<q; l++)
                  ZWY(l,j) += Z(i,j)* *(w+i) * *(y_in+i*q+l) ;    
            }
         }
      } 
      for (l=0; l<dd; l++)
      {
         for (j=l+1; j<dd; j++)
         {
          ZWZ(j,l) = ZWZ(l,j) ;
         }
      }  
           
//        ZWZ.inverse() ;          	// compute inverse of ZT*W*Z
//	ZWY.mult(ZWZ) ;	        	// obtain inv(ZWZ) ZT*W*Y

        ZWZ = ZWZ.inv(ZWZ) ;          	// compute inverse of ZT*W*Z
	ZWY = ZWY.matmult(ZWZ) ;	// obtain inv(ZWZ) ZT*W*Y

      for (j=0; j<dd; j++)      	// read data out of matrices
      {
        for (l=0; l<q; l++)
             *(ma_out+l*ng*dd+j*ng+k) = ZWY(l,j);
       }  
   }
    delete[] kern;			// free workspace
    delete[] w;				// free workspace
 /*
    delete[] Z;  			// free workspace
    delete[] ZWZ;			// free workspace
    delete[] ZWY;			// free workspace
 */  
   return 0;

}

}

extern "C" int loclin2 
	(double *x_in, double *xd_in, double *xg_in, double *y_in,
         double *h_in, double *par_in,         
         double *ma_out) ;


XplDouble stefan2 (XplDouble *x, XplDouble *xd, XplDouble *xg,
                    XplDouble *y, XplDouble *h, XplDouble *logi_in )

{ 	
	int dim=x->dimension(1);
	int rows=x->dimension(0);

	int dimd = xd == NULL ? 0 : xd->dimension(1);
	int rowd = xd == NULL ? 0 : xd->dimension(0);

	int dimg=xg->dimension(1);
	int rowg=xg->dimension(0);
	int dimy=y->dimension(1);
	int rowy=y->dimension(0);

        XplDouble rr(1);	

	if (rowy != rows)
	  {
	  globalError.Set (ERR_LOCPOL, "x and y are not conformable!");
	  return (rr);
	  }
	
	if (h->dimension(0) != dim)
	  {
	  globalError.Set (ERR_LOCPOL, "x and h are not conformable!");
	  return (rr);
	  }
	
	if (xd != NULL && rowd != rows)
	  {
	  globalError.Set (ERR_LOCPOL, "xd and x are not conformable!");
	  return (rr);
	  }

	if (dimg!=dim)
	  {
	  globalError.Set (ERR_LOCPOL, "xg and x are not conformable!");
	  return (rr);
	  }

	int logi=(int)( (*logi_in)[0]);
	if (! ( logi==0 || logi==1)) 
	  {
	  globalError.Set (ERR_LOCPOL, "local linear=0, quadratic=1");
	  return (rr);
	  }
	
	int i, j, l, dd;
	
//	double x_in [dim*rows];
	double *x_in = new double[dim*rows];

	for (i=0; i<rows; i++)
	for (j=0; j<dim; j++)
		x_in[i*dim+j]=(*x)(i, j);	

	
//	double xd_in [dimd*rowd + 1];   // dimd or rowd could be zero !
	double *xd_in = new double[dimd*rowd + 1];

	for (i=0; i<rowd; i++)
	for (j=0; j<dimd; j++)
		xd_in[i*dimd+j]=(*xd)(i, j);	
			
//	double xg_in [dimg*rowg];
	double *xg_in = new double[dimg*rowg];

	for (i=0; i<rowg; i++)
	for (j=0; j<dimg; j++)
		xg_in[i*dimg+j]=(*xg)(i, j);	
	
//	double y_in [dimy*rows]; 
	double *y_in = new double[dimy*rows]; 

	for (i=0; i<rows; i++)
	for (j=0; j<dimy; j++)
		y_in[i*dimy+j]=(*y)(i, j);
		
//	par_in= new double [5];
	double par_in [6];
	 par_in[0]= rows ;
	 par_in[1]= rowg ;
	 par_in[2]= dim  ;
	 par_in[3]= dimd ;
	 par_in[4]= dimy ;
	 par_in[5]= logi ;

//	h_in= new double [dim];
//	double h_in [dim];
	double *h_in = new double[dim];

	for (i=0; i<dim; i++)
		h_in[i]=(*h)(i);



        dd  = (int)(1+dim+logi*dim) ;

//	double *ma_out = new double[rowg*dd*dimy];
//	double ma_out[rowg*dd*dimy];
	double *ma_out = new double[rowg*dd*dimy];


	loclin2(x_in, xd_in, xg_in, y_in, h_in, par_in, ma_out);


	XplDouble result(rowg, dd, dimy);
	
	for (i=0; i<rowg; i++)
	for (j=0; j<dd; j++)
        for (l=0; l<dimy; l++)
	  {     result(i, j, l)=ma_out[l*rowg*dd+j*rowg+i];
	  }

	XplObject *o =new XplObject( new XplDouble(result), "loch");
	


// wieder alle loeschen dank Borland :-(

	delete[] x_in;
	delete[] xd_in;
	delete[] xg_in;
	delete[] y_in;
	delete[] h_in;

	delete[] ma_out;

	return(result);
	
}

#ifdef HAS_TIMEH
#include <time.h>
#else
#include <sys/time.h>
#include <sys/resource.h>
#endif

#define MAX_CORE  (1L << 20L)    /* 1 MB */

#include "matrix.h"
#include <stdio.h>
#include <math.h>
#include <stdlib.h>

/*   **************************************
      C ++ Program       loclin2.C
      
      calculates local polynomials 
      - without mixed terms
      - with partial linear 

     **************************************    */
/* 
*  Hints for compilation and necessary c or C files:
*
   g++ -G -O2 -c -DMATRIX matrix.c    	this compilies matrix.c library
   gcc -G -O2 -c dl*.c		        compiles dlubksb.c, dludcmp.c
   gcc -G -O2 -c dja*.c			compiles djacobi.c
   gcc -G -O2 -c nrutil.c		compiles routines from 
					Numerical Recepies
   gcc -G -O2 loclin2.C matrix.o dl*.o -o loclin2.so -lm 
   gcc -G -O2 loclin2.C matrix.o dl*.o dja*.o nrutil.o -o loclin2.so -lm 
					compiles loclin.C and links it with
					all the other routines
*  INPUTS :
*   xorig = original designmatrix of continuous variables
*   xd    = original designmatrix of discrete variables
*   yorig = original dependent observations
*   h     = bandwidth vector
*   par   = vektor containing
*     n     = number of observations
*     ng    = number of grid points
*     d     = number of dimensions with continuous vars.
*     dis   = number of dimensions with discrete vars.
*     q     = dimension of yorig < 4  !!
*     lorq  = control variable whether linear or quadratic
*  OUTPUT :
*   ma_out= (ng*dd) matrix of results 
*
*/

extern "C"
{

double p_gauss(long p, double *k);
double p_quartic(long p, double *k);


int loclin2 (double *x_in, double *xd_in, double *xg_in, 
             double *y_in, double *h_in, double *par_in, 
             double *ma_out)

{  long n, ng, d, dis, dd, dd2, q, i, j, k, j1, l, lorq, start;
   double  *kern, *w ;

#ifdef HAS_CORE
   struct rlimit   rlp;

   rlp.rlim_cur = MAX_CORE;
   rlp.rlim_max = MAX_CORE;

   setrlimit (RLIMIT_CORE, &rlp);  /* sets core limit  */
#endif
     
   n	= (long) *(par_in+0);    /* number of samples */
   ng   = (long) *(par_in+1);   
   d	= (long) *(par_in+2);    /* number of regressors */
   dis	= (long) *(par_in+3);    /* number of regressors */
   q    = (long) *(par_in+4);       
   lorq = (long) *(par_in+5);  /* parameter that controls whether 
			       local linear (lorq = 0) or
			       quadratic (= 1) estimation is used */
   dd  = (long)(1+d+lorq*d) ;
   dd2 = (long)(1+d+lorq*d+dis) ;
  
   kern = new double[sizeof(double)*d];  	// initialize kern
   w	= new double[sizeof(double)*n];		// initialize W
       
//   Matrix Z(n,dd2), ZWZ(dd2,dd2), ZWY(q,dd2) ;  // initialize all matrices   
   XplDouble Z(n,dd2), ZWZ(dd2,dd2), ZWY(q,dd2) ;  // initialize all matrices   

   // do loop over gridpoints
   for (start=k=0; k<ng; k++)
   {
      for (j=0; j<dd2; j++)     
      {
        for (l=0; l<q; l++)
	  ZWY(l,j) = 0 ; 
        for (l=0; l< j+1; l++)
	  ZWZ(l,j) = 0 ; 

      }  

      // do loop over observations
      for (i=start; i<n; i++)
      {           
 	 Z(i,0) = 1;   	                      // constant term
         Z(i,1) = *(x_in+i*d) - *(xg_in+k*d);    // first linear term 
 	 *(kern+0) = Z(i,1) / *(h_in) ;
         
         if (*(kern+0)<-1.0)
            start++ ;         
         else
         { if (*(kern+0)>1.0) break;
           for (j=1; j<d; j++)         
           {                       
               Z(i,j+1) = *(x_in+j+i*d) - *(xg_in+j+k*d);    // linear terms 
               *(kern+j) = Z(i,j+1) / *(h_in+j) ;        // prepare vector for kernel 
	   }	

            *(w+i) = p_quartic(d,kern) ;

	    if (lorq==1)
            {  
	       for (j=0; j<d; j++)                     // quadratic terms 
                    Z(i,d+1+j) = Z(i,j+1) * Z(i,j+1) ;  
            } 
            for (j1=0; j1<dis; j1++)
                 Z(i,d+d*lorq+1+j1) = *(xd_in+j1+i*dis) ;

            for (j=0; j<dd2; j++)
            {
               ZWZ(0,j) += Z(i,j) * *(w+i) ;
               for (l=1; l< j+1; l++)
                  ZWZ(l,j) += Z(i,j)*Z(i,l)* *(w+i) ; 
               for (l=0; l<q; l++)
                  ZWY(l,j) += Z(i,j)* *(w+i) * *(y_in+i*q+l) ;    
            }
         }
      } 
      for (l=0; l<dd2; l++)
      {
         for (j=l+1; j<dd2; j++)
             ZWZ(j,l) = ZWZ(l,j) ;
      }  
           
//        ZWZ.inverse() ;          	// compute inverse of ZT*W*Z
//	ZWY.mult(ZWZ) ;	        	// obtain inv(ZWZ) ZT*W*Y

        ZWZ = ZWZ.inv(ZWZ) ;          	// compute inverse of ZT*W*Z
	ZWY = ZWY.matmult(ZWZ) ;	      	// obtain inv(ZWZ) ZT*W*Y

      for (j=0; j<dd; j++)      	// read data out of matrices
      {
        for (l=0; l<q; l++)
             *(ma_out+l*ng*dd+j*ng+k) = ZWY(l,j);
      }  
   }
    delete[] kern;			// free workspace
    delete[] w;				// free workspace
  
   return 0;

}

}
#endif // !MATRIX
