/* $Header: /home/xplore/cvs/xpl4gr/machine.C,v 1.3 2002/08/16 12:31:45 sigbert Exp $ */

/*  This file contains system specfic routines
	for XploRe 4. You need to define : 

char *x4findfirst (char * directory, char *filter);
	- Under GNUC it limited to *.xpl and *.hlp2

char *x4findnext ();

void x4date (int *day, int *month, int *year);

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "machine.h"
#include "directory.h"

char tmpfilename[1024];
extern char tmppath[256];


#ifdef __GNUC__
class starexp
{ char *start, *end;
  int exact;

public:

  starexp (char *pat)
  { int i;
    exact = 0;
    for (i=0; i<strlen(pat); i++)
      if (pat[i]=='*') break;
    if (i==0)
    { start = NULL;
      end   = strdup(&pat[1]);
    }
    else if (i==strlen(pat))
    { start = strdup(pat);
      end   = NULL;
      exact = 1;
    }
    else
    { start = strdup(pat);
      start[i] = 0;
      end   = strdup(&start[i+1]);
    }
  }

  ~starexp ()
  { if (start) delete[] start;
    if (end) delete[] end;
  }
  
  int match (char *str)
  { int i;
    if (start)
    { for (i=0; i<strlen(start); i++)
        if (start[i]!=str[i]) return 0;
      if (exact) return (str[i]==0);
    }
    if (end)
    { for (i=0; i<strlen(end); i++)
        if (end[strlen(end)-i-1]!=str[strlen(str)-i-1]) return 0;
    }
    return 1;
  }

};

starexp *se;
DIR *dir;
struct dirent *entry;

char *x4findfirst (char *directory, char *filter)
{ int i;
  se  = new starexp(filter);
  dir = opendir (directory);
  if (dir==NULL) return(NULL);
  while ((entry = readdir(dir))!=NULL)
  { if (se->match(entry->d_name))
    { strcpy (tmpfilename, entry->d_name);
      return (tmpfilename);
    }
  }
  closedir (dir);
  delete se;
  return (NULL);
}

char *x4findnext()
{ while ((entry = readdir(dir))!=NULL)
  { if (se->match(entry->d_name))
    { strcpy (tmpfilename, entry->d_name);
      return (tmpfilename);
    }
  }
  closedir (dir);
  delete se;
  return (NULL);
}

void x4date (int *day, int *month, int *year)
{ struct tm *tmptr;
  time_t tptr;
  ::time(&tptr);
  tmptr = gmtime(&tptr);
  *day   = tmptr->tm_mday;
  *month = tmptr->tm_mon+1;
  *year  = 1900+tmptr->tm_year;
}

char *x4gethome (char *subdir = NULL)
{ char *homepath;

  homepath   = getenv ("HOME");
  if (homepath) 
  { strcpy (tmppath, homepath);
    if (subdir) 
    { strcat (tmppath, SEPCHAR);
      strcat (tmppath, subdir);
    }
    return (tmppath);
  }
  return (NULL);
}

#ifdef __MINGW32__
void x4mkdir (char *dir)
{ mkdir (dir); }
#else
void x4mkdir (char *dir)
{ mkdir (dir, S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH); }
#endif

#endif

// Borland C

#ifdef __BORLANDC__

struct ffblk ffblk;
int done;

char *x4findfirst (char *directory, char *filter)
{ char where[1024];
  sprintf (where, "%s%s%s", directory, SEPCHAR, filter); 
  done = findfirst (where, &ffblk, 0);
  if (done) return (NULL); 
  strcpy (tmpfilename, ffblk.ff_name); 
  return (tmpfilename);
}

char *x4findnext()
{ done = findnext(&ffblk);
  if (done!=0) return (NULL);
  strcpy (tmpfilename, ffblk.ff_name); 
  return (tmpfilename);  
}

void x4date (int *day, int *month, int *year)
{ struct tm *tmptr;
  time_t tptr;
  ::time(&tptr);
  tmptr  = gmtime(&tptr);
  *day   = tmptr->tm_mday;
  *month = tmptr->tm_mon+1;
  *year  = 1900+tmptr->tm_year;
}

char *x4gethome (char *subdir)
{ char *homedrive, *homepath, *userprofile;

  homedrive   = getenv ("HOMEDRIVE");
  if (homedrive) homedrive = strdup(homedrive);
  homepath    = getenv ("HOMEPATH");
  if (homepath) homepath = strdup(homepath);
  userprofile = getenv ("USERPROFILE");
  if (userprofile) userprofile = strdup(userprofile);

  // printf ("home: %s %s %s\n", homedrive, homepath, userprofile);
  if (homedrive && homepath)
  { strcpy (tmppath, homedrive);
    strcat (tmppath, homepath);
    if (subdir)
    { strcat (tmppath, SEPCHAR);
      strcat (tmppath, subdir);
    }
    //printf ("gethome:home %s\n", tmppath);
    if (homedrive) delete[] homedrive;
    if (homepath) delete[] homepath;
    if (userprofile) delete[] userprofile;
    return (tmppath);
  }
  else if (userprofile)
  { strcpy (tmppath, userprofile);
    if (subdir)
    { strcat (tmppath, SEPCHAR);
      strcat (tmppath, subdir);
    }
    //printf ("gethome:user %s\n", tmppath);
    if (homedrive) delete[] homedrive;
    if (homepath) delete[] homepath;
    if (userprofile) delete[] userprofile;
    return (tmppath);
  }
  //printf ("gethome:work %s\n", tmppath);
  if (homedrive) delete[] homedrive;
  if (homepath) delete[] homepath;
  if (userprofile) delete[] userprofile;
  return (getx4cwd(subdir));
}

void x4mkdir (char *dir)
{ mkdir (dir); }
  
#endif

#ifdef __SC__

struct FIND* ffblk;

char *x4findfirst (char *directory, char *filter)
{ char where[1024];
  sprintf (where, "%s%s%s", directory, SEPCHAR, filter); 
  ffblk = findfirst (where, 0);
  if (ffblk==NULL) return (NULL); 
  strcpy (tmpfilename, ffblk->name); 
  return (tmpfilename);
}

char *x4findnext()
{ ffblk = findnext();
  if (ffblk==NULL) return (NULL);
  strcpy (tmpfilename, ffblk->name); 
  return (tmpfilename);  
}

void x4date (int *day, int *month, int *year)
{ struct dos_date_t d;
  dos_getdate (&d);
  *day   = d.day;
  *month = d.month;
  *year  = d.year;
}

char *x4gethome (char *subdir)
{ char *homedrive, *homepath, *userprofile;

  homedrive   = getenv ("HOMEDRIVE");
  if (homedrive) homedrive = strdup(homedrive);
  homepath    = getenv ("HOMEPATH");
  if (homepath) homepath = strdup(homepath);
  userprofile = getenv ("USERPROFILE");
  if (userprofile) userprofile = strdup(userprofile);

  // printf ("home: %s %s %s\n", homedrive, homepath, userprofile);
  if (homedrive && homepath)
  { strcpy (tmppath, homedrive);
    strcat (tmppath, homepath);
    if (subdir)
    { strcat (tmppath, SEPCHAR);
      strcat (tmppath, subdir);
    }
    //printf ("gethome:home %s\n", tmppath);
    if (homedrive) delete[] homedrive;
    if (homepath) delete[] homepath;
    if (userprofile) delete[] userprofile;
    return (tmppath);
  }
  else if (userprofile)
  { strcpy (tmppath, userprofile);
    if (subdir)
    { strcat (tmppath, SEPCHAR);
      strcat (tmppath, subdir);
    }
    //printf ("gethome:user %s\n", tmppath);
    if (homedrive) delete[] homedrive;
    if (homepath) delete[] homepath;
    if (userprofile) delete[] userprofile;
    return (tmppath);
  }
  //printf ("gethome:work %s\n", tmppath);
  if (homedrive) delete[] homedrive;
  if (homepath) delete[] homepath;
  if (userprofile) delete[] userprofile;
  return (getx4cwd(subdir));
}

void x4mkdir (char *dir)
{ mkdir (dir); }

#endif

#ifdef _MSC_VER

struct _finddata_t fileinfo;
long handle;

char *x4findfirst (char *directory, char *filter)
{ char where[1024];
  sprintf (where, "%s%s%s", directory, SEPCHAR, filter); 
  handle = _findfirst (where, &fileinfo);
  if (handle==-1)
  { _findclose (handle);
    return (NULL);
  }
  strcpy (tmpfilename, fileinfo.name);
  return (tmpfilename);
}

char *x4findnext()
{ if (_findnext(handle, &fileinfo))
  { _findclose (handle);
    return (NULL);
  }
  strcpy (tmpfilename, fileinfo.name); 
  return (tmpfilename);  
}

void x4date (int *day, int *month, int *year)
{ struct tm *tmptr;
  time_t tptr;
  ::time(&tptr);
  tmptr = gmtime(&tptr);
  *day   = tmptr->tm_mday;
  *month = tmptr->tm_mon+1;
  *year  = 1900 + tmptr->tm_year;
}

char *x4gethome (char *subdir)
{ char *homedrive, *homepath, *userprofile;

  homedrive   = getenv ("HOMEDRIVE");
  if (homedrive) homedrive = strdup(homedrive);
  homepath    = getenv ("HOMEPATH");
  if (homepath) homepath = strdup(homepath);
  userprofile = getenv ("USERPROFILE");
  if (userprofile) userprofile = strdup(userprofile);

  // printf ("home: %s %s %s\n", homedrive, homepath, userprofile);
  if (homedrive && homepath)
  { strcpy (tmppath, homedrive);
    strcat (tmppath, homepath);
    if (subdir)
    { strcat (tmppath, SEPCHAR);
      strcat (tmppath, subdir);
    }
    //printf ("gethome:home %s\n", tmppath);
    if (homedrive) delete[] homedrive;
    if (homepath) delete[] homepath;
    if (userprofile) delete[] userprofile;
    return (tmppath);
  }
  else if (userprofile)
  { strcpy (tmppath, userprofile);
    if (subdir)
    { strcat (tmppath, SEPCHAR);
      strcat (tmppath, subdir);
    }
    //printf ("gethome:user %s\n", tmppath);
    if (homedrive) delete[] homedrive;
    if (homepath) delete[] homepath;
    if (userprofile) delete[] userprofile;
    return (tmppath);
  }
  //printf ("gethome:work %s\n", tmppath);
  if (homedrive) delete[] homedrive;
  if (homepath) delete[] homepath;
  if (userprofile) delete[] userprofile;
  return (getx4cwd(subdir));
}

void x4mkdir (char *dir)
{ mkdir (dir); }

#endif
