/* $Header: /home/xplore/cvs/xpl4gr/machine.h,v 1.4 2003/02/23 16:54:44 cizek Exp $ */

/* machine.h contains compiler- or platform dependend seetings
   and definition:

   Part I   - Version-dependend common definitions
   -----------------------------------------------
  
   Flags:
     _RPCXGOBI_         compile with Remote Procedure calls (only Unix)
     SMALLVERSION       support only for SMALLVERSIONOBSERVATIONS observations (1000 by default) // PC
                        (STRSMALLVERSIONOBSERVATIONS must be defined as a string containing the number)
    X4WIN_90DAYS        version limited to work only several days after its installation (currently, 
                        the limit is 30 days); just for Windows

   Part II  - Machine-dependend common definitions
   ----------------------------------------------- 

   Flags:
     SEPCHAR            constant with sets the separator between
                        directories
     _EXPIRE_           if defined the X4 version expires lastest
                        2 month after compiling

     HAS_STDSORT        function pointer in template classes seem 
                        to confuse Intel
     HAS_SMPMATH        implements some extended math functions
     HAS_GAMMA          gamma is somewhere defined
     HAS_STRCMPI        string.h contains strcmpi otherwise strcasecmp 
                        is used
     HAS_TMP            has a directory /tmp
     HAS_IOSTREAMH      uses iostream.h rather than stream.h
     HAS_TIMEH          uses time.h rather than sys/time.h and sys/resource.h
     HAS_CORE           supports core file generation 
     HAS_MSWIN          uses Microsoft Windows
     HAS_NOPTRPTR       has difficulties with ** 
     HAS_DLL            supports 'dynamic link libraries' or 
                        'shared libraries'
                        =1 Unix type DLL, =2 Windows type DLL

   To add a new machine figure out which flags your compiler
   defines:

   - GNU C++
     With 'gcc -v' gcc will tell you from where it reads its specs. 
     You should find a section like

     *predefines:
      -Dsun -Dsparc -Dunix -D__svr4__ -D__SVR4   -Asystem(unix) 
         -Asystem(svr4) -Acpu(sparc) -Amachine(sparc)  -D__GCC_NEW_VARARGS__
 
     in that file. The first two flags (sun, sparc) are used to identify 
     the correct machine (here SUN solaris).
   
   - Borland C
     Borland defines __BORLANDC__

   - Visual C++
     Microsoft defines _MSC_VER

   Part III - Compiler-dependend common definitions
   ------------------------------------------------
 
   Part IV  - Definitions from machine.c
   -------------------------------------
     Do not change them. For new compiler it might be necessary
     to reimplement these routines.

*/

/* ----------------------------------------------
   Part I - Some common definitions
   ----------------------------------------------
*/

#include <string.h>
#include <math.h>

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#if !defined (M_E)
#define M_E  2.7182818284590452354
#endif

#define WINDLL    1
#define UNIXDLL   2


/* ----------------------------------------------
   Part I - Version-dependend common definitions
   ----------------------------------------------
*/

#ifdef X4standalone
#endif
#ifdef X4server
#endif
#ifdef X4batch
#endif

// PC 20011127
#ifdef SMALLVERSION
#ifndef SMALLVERSIONOBSERVATIONS
#define SMALLVERSIONOBSERVATIONS 1000
#define STRSMALLVERSIONOBSERVATIONS "1000"
#endif
#else
#define STRSMALLVERSIONOBSERVATIONS "1000"
#endif

/* ----------------------------------------------
   Part II - Machine-dependend common definitions
   ----------------------------------------------
*/

/* SUN Solaris */
#if (defined(sun) && defined(sparc))

#ifdef X4standalone
  #undef  _EXPIRE_
#endif
#ifdef X4server
  #define  _EXPIRE_
#endif
#ifdef X4batch
  #undef  _EXPIRE_
#endif

  #define SEPCHAR "/"
  #undef  _EXPIRE_

  #define HAS_STDSORT
  #undef  HAS_STRCMPI
  #undef  HAS_SMPMATH
  #undef  HAS_GAMMA
  #define HAS_TMP
  #define HAS_IOSTREAMH
  #undef  HAS_TIMEH
  #define HAS_CORE
  #undef  HAS_MSWIN
  #undef  HAS_NOPTPTR
  #define HAS_DLL         UNIXDLL

#endif

/* HP Unix */
#if (defined(hpux) && defined(hppa))

#ifdef X4standalone
  #undef  _EXPIRE_
#endif
#ifdef X4server
  #define  _EXPIRE_
#endif
#ifdef X4batch
  #undef  _EXPIRE_
#endif

  #define SEPCHAR "/"

  #define HAS_STDSORT
  #undef  HAS_STRCMPI
  #undef  HAS_SMPMATH
  #undef  HAS_GAMMA
  #define HAS_TMP
  #undef  HAS_IOSTREAMH
  #undef  HAS_TIMEH
  #define HAS_CORE
  #undef  HAS_MSWIN
  #undef  HAS_NOPTPTR
  #define HAS_DLL        UNIXDLL

#endif

/* IBM AIX */
#if (defined(_AIX) && defined(_IBMR2))

#ifdef X4standalone
  #undef  _EXPIRE_
#endif
#ifdef X4server
  #define  _EXPIRE_
#endif
#ifdef X4batch
  #undef  _EXPIRE_
#endif

  #define SEPCHAR "/"

  #define HAS_STDSORT
  #undef  HAS_STRCMPI
  #undef  HAS_SMPMATH
  #define HAS_GAMMA
  #define HAS_TMP
  #undef  HAS_IOSTREAMH
  #undef  HAS_TIMEH
  #define HAS_CORE
  #undef  HAS_MSWIN
  #undef  HAS_NOPTPTR
  #undef  HAS_DLL			  UNIXDLL

#endif

/* SGI
#if (defined(_AIX) && defined(_IBMR2))

  #define SEPCHAR "/"

  #define HAS_STDSORT
  #undef  HAS_STRCMPI
  #undef  HAS_SMPMATH
  #define HAS_GAMMA
  #define HAS_TMP
  #undef  HAS_IOSTREAMH
  #undef  HAS_TIMEH
  #define HAS_CORE
  #undef  HAS_MSWIN
  #undef  HAS_NOPTPTR
  #undef  HAS_DLL

#
*/

// Linux
#if (defined(i386) && defined(linux))

#ifdef X4standalone
  #undef  _EXPIRE_
#endif
#ifdef X4server
  #define  _EXPIRE_
#endif
#ifdef X4batch
  #undef  _EXPIRE_
#endif

  #define SEPCHAR "/"

  #define HAS_STDSORT
  #undef  HAS_STRCMPI
  #define HAS_SMPMATH
  #undef  HAS_GAMMA
  #define HAS_TMP
  #undef  HAS_IOSTREAMH
  #undef  HAS_TIMEH
  #define HAS_CORE
  #undef  HAS_MSWIN
  #undef  HAS_NOPTPTR
  #define HAS_DLL				UNIXDLL

#endif

// Apple
#ifdef __APPLE__

#ifdef X4standalone
  #undef  _EXPIRE_
#endif
#ifdef X4server
  #define  _EXPIRE_
#endif
#ifdef X4batch
  #undef  _EXPIRE_
#endif

  #define SEPCHAR "/"

  #define HAS_STDSORT
  #undef  HAS_STRCMPI
  #define HAS_SMPMATH
  #define HAS_GAMMA
  #define HAS_TMP
  #define HAS_IOSTREAMH
  #undef  HAS_TIMEH
  #define HAS_CORE
  #undef  HAS_MSWIN
  #undef  HAS_NOPTPTR
  #define HAS_DLL UNIXDLL

#endif

/* DJGPP DOS/Windows
#if (defined(i386) && defined(DJGPP))

  #define SEPCHAR "/"

  #define HAS_STDSORT
  #undef  HAS_STRCMPI
  #define HAS_SMPMATH
  #undef  HAS_GAMMA
  #define HAS_TMP
  #undef  HAS_IOSTREAMH
  #undef  HAS_TIMEH
  #define HAS_CORE
  #undef  HAS_MSWIN
  #undef  HAS_NOPTPTR
  #undef  HAS_DLL

#endif
*/

#ifdef __MINGW32__

  #undef  _EXPIRE_

  #define SEPCHAR "\\"

  #undef  HAS_STDSORT
  #define HAS_STRCMPI
  #define HAS_SMPMATH
  #undef  HAS_GAMMA
  #undef  HAS_TMP
  #define HAS_IOSTREAMH
  #define HAS_TIMEH
  #undef  HAS_CORE
  #define HAS_MSWIN
  #define HAS_NOPTRPTR
  #define HAS_DLL          WINDLL
  #undef  WIN32

#endif
// Windows 95/NT 4.0
#if defined(_MSC_VER)

  #define SEPCHAR "\\"
  #define _EXPIRE_

  #undef  HAS_STDSORT
  #define HAS_STRCMPI
  #define HAS_SMPMATH
  #undef  HAS_GAMMA
  #undef  HAS_TMP
  #define HAS_IOSTREAMH
  #define HAS_TIMEH
  #undef  HAS_CORE
  #define HAS_MSWIN
  #define HAS_NOPTRPTR
  #undef  HAS_DLL          WINDLL

#endif

#if defined(__BORLANDC__)

#ifdef SMALLVERSION
  #undef  _EXPIRE_
#else
  #define _EXPIRE_
#endif

#ifdef NOEXPIRE
  #undef _EXPIRE_
#endif

  #define SEPCHAR "\\"

  #undef  HAS_STDSORT
  #define HAS_STRCMPI
  #define HAS_SMPMATH
  #undef  HAS_GAMMA
  #undef  HAS_TMP
  #define HAS_IOSTREAMH
  #define HAS_TIMEH
  #undef  HAS_CORE
  #define HAS_MSWIN
  #define HAS_NOPTRPTR
  #define HAS_DLL				WINDLL

#endif

#if (defined(mips) && defined(MIPSEL))

  #define SEPCHAR "/"
  #undef  _EXPIRE_

  #define HAS_STDSORT
  #undef  HAS_STRCMPI
  #undef  HAS_SMPMATH
  #undef  HAS_GAMMA
  #define HAS_TMP
  #define HAS_IOSTREAMH
  #undef  HAS_TIMEH
  #define HAS_CORE
  #undef  HAS_MSWIN
  #undef  HAS_NOPTPTR
  #define HAS_DLL         UNIXDLL

#endif

/* ------------------------------------------------
   Part III - Compiler-dependend common definitions
   ------------------------------------------------
*/

#ifdef __GNUC__
  #include <sys/types.h>
  #include <sys/stat.h>
  #include <dirent.h>
  #include <time.h>
#endif

#ifdef __BORLANDC__
  #include <dir.h>
  #include <conio.h>
  #include <time.h>
//  #include "xplwin.h"
#endif

#ifdef _MSC_VER
  #include <conio.h>
  #include <direct.h>
  #include <time.h>
  #include <io.h>
  #define getch _getch
#endif

// SK Test Suite for independent programs

#ifdef __SC__
  #include <dos.h>
  #include <direct.h>

  #define SEPCHAR "\\"

  #define HAS_STRCMPI

#endif


/* ------------------------------------------------
   Part IV - Definitions from machine.c
   ------------------------------------------------
*/

char *x4findfirst (char *directory, char *filter);
char *x4findnext ();
void x4date (int *day, int *month, int *year);
char *x4gethome (char *subdir);
void x4mkdir (char *dir);
int isabspath (char *dir);
