/* $Header: /home/xplore/cvs/xpl4gr/macro.C,v 1.3 2002/08/16 12:31:46 sigbert Exp $ */

// Macro Klasse  10.10.94  TTK

// last modification:  10.10.94

// #pragma implementation

#include <string.h>
#include "liste.h"
#include "object.h"
#include "xplcalc.h"
#include "xplpp.h"
#include "globlist.h"
#include "macro.h"
#include "ttkio.h"
#include "liste.h"
#include "xplerror.h"

Macro::Macro (const char* n, const char *t, 
              Liste<const char*> ip, Liste<const char*> op,
              class GlobalList *gl)
{
  name = strdup (n);
  text = strdup (t);

  if ((text == NULL) || (name == NULL))
    {
    ttkerror ("Fatal Error in Macro:  Out of memory");
    exit (0);
    }

  inpar      = ip;
  outpar     = op;
  globallist = gl;
  objList    = new Liste<class XplObject*>;
  allChecked = FALSE;
  isRunning  = FALSE;

  ip.remove ();
  op.remove ();  
}

Macro::~Macro ()
{
  long i, length;

  delete[] name;
  delete[] text;

  inpar.remove();               // Liste von const char*
  outpar.remove();              // Liste von const char*

  globallist->DeleteElements();
  delete globallist;

  length = objList->laenge ();
  for (i=0; i<length; i++)
    delete (*objList)[i];
  objList->remove ();
  delete objList;
}

void Macro::SetNewDefinition  (const char *t,
   Liste<const char*> ip,  Liste<const char*> op,
   class GlobalList* newgl)
{
  long i, length = objList->laenge ();  

  if (isRunning)
    {
    globalError.Set (ERR_INTERN, "tried to change while running ");
    globalError.AppendText (ERR_INTERN, name);
    return;
    }

  if (newgl == NULL)
    {
    globalError.Set (ERR_INTERN, "Error in SetGlobalList:  macro not modified!");
    return;
    }

  globallist->DeleteElements();

  delete[] text;
  delete   globallist;

  for (i=0; i<length; i++)
    delete (*objList)[i];

  objList->remove ();
  inpar.remove ();
  outpar.remove ();  

  text = strdup (t);

  if (text == NULL)
    {
    ttkerror ("Fatal Error in MacroSetNewDefinition:  Out of memory");
    exit (0);
    }

  globallist = newgl;
  inpar      = ip;
  outpar     = op;
  allChecked = FALSE;

  ip.remove ();
  op.remove ();  
//  PrepareToRun ();   is done in CheckMacro now, TTK, 27.6.97
}

void BuildLocalXplObjectList (Liste<XplObject*> *objList, 
                      Liste<const char*> &inp, Liste<const char*> &outp)
{
  long i, length;

  length = inp.laenge ();

  for (i=0; i<length; i++)
    objList->haengeAn (new XplObject (inp[i]));

  length = outp.laenge ();

  for (i=0; i<length; i++)
    objList->haengeAn (new XplObject (outp[i]));
}

MacroText::MacroText (long n)
{
#ifdef TTKDEBUG
if (ttkdebug)
cerr << "MacText:\tconstruct\t" << this  << endl;
#endif
  if (n <= 0L)
    {
    ttkerror ("Fatal Error in MacroText: Negative Textlength");
    exit(0);
    }

  text = new char[n];

  if (text == NULL)
    {
    ttkerror ("Fatal Error in MacroText:  Out of memory");
    exit (0);
    }

  *text      = '\0';
  length     = 0L;
  maxLength  = n-1;             // for terminal \0
  recordMode = FALSE;
}


void MacroText::AppendText (const char *t)
{
  long tl;
  char *text2 = text;

  if (recordMode == FALSE)
    return;

  tl = strlen (t);

  if (length + tl > maxLength)
    {
    maxLength = length + tl + OFFSET;
    text = new char[maxLength+2];
    if (text == NULL)
      {
      ttkerror("Fatal Error in MacroText:  Out of memory");
      exit (0);
      }

    //    strncpy (text, text2, length+1);   // terminal '\0'
    strcpy (text, text2);
    delete[] text2;
    }

  //  strncat (text, t, tl+1);            // terminal '\0'    new TTK, 18.1.98
  strcat (text, t);
  length += tl;
}

void MacroText::SetNewText ()
{
/*  text = new char[TEXTLENGTH];

  if (text == NULL)
    {
    ttkerror ("Fatal Error in MacroText:  Out of memory");
    exit (0);
    }
*/
  *text      = '\0';
  length     = 0L;
//  maxLength  = TEXTLENGTH;
}

void Macro::PrepareToRun ()
{
  allChecked = FALSE;
}

void Macro::Prepare2Run ()
{
  const char *oldMacroName;
  
  BuildLocalXplObjectList (objList, inpar, outpar);
  ScanFunsAndMacs (globallist);
  ScanMacroVariables (objList, globallist);
  ScanConstants (globallist);
  InputBranch (globallist, 0L);
#ifdef TTKDEBUG
if (ttkdebug)
  ShowTokenList(globallist);
#endif
  globallist->SetPos (0L);


  long i, j, 
  length = inpar.laenge (), length2 = outpar.laenge ();

  // check for double defined formal parameters

  for (i=0; i<inpar.laenge(); i++)
    for (j=i+1; j<inpar.laenge(); j++)
      if (strcmp (inpar[i], inpar[j]) == 0)
	globalError.Set (ERR_INP, inpar[i]);

  for (i=0; i<outpar.laenge(); i++)
    for (j=i+1; j<outpar.laenge(); j++)
      if (strcmp (outpar[i], outpar[j]) == 0)
	globalError.Set (ERR_OUTP, outpar[i]);

  for (i=0; i<length; i++)
    for (j=0; j<length2; j++)
      if (strcmp (inpar[i], outpar[j]) == 0)
	globalError.Set (ERR_PARAM, inpar[i]);

  extern int *varlist;            // flag, whether VAR has already a value 
  length = objList->laenge ();

  varlist = new int[length+2];    // even if length==0, varlist is an array
  for (i=0; i<length; i++)
    varlist[i] = FALSE;

  length = inpar.laenge ();
  for (i=0; i<length; i++)        // input VARs have a value
    varlist[i] = TRUE;

extern int checkMFcand;
extern const char *CurrentMacro;
  checkMFcand = FALSE;

  oldMacroName = CurrentMacro;
  CurrentMacro = name;

#ifdef TTKDEBUG
extern int docheck;
  if (docheck)
    CheckTokenList (*globallist);
#else
  CheckTokenList (*globallist);
#endif
  delete[] varlist;
  varlist = NULL;

  CurrentMacro = oldMacroName;

  if (globalError.Get ())
    globalError.ShowError();    // now show errors, if there are any
}

const int Macro::CheckMacro ()
{
  if (Checked () == TRUE)
    return (TRUE);

  Prepare2Run ();

  const char *oldMacroName;

  ScanVariables (objList, globallist);     // release VARs if possible
  ScanFunsAndMacs (globallist);            // check for new NEWSTRINGs

  if (ScanVariables (objList, globallist) == FALSE)
    return (FALSE);

  extern int *varlist;
  long i, length = objList->laenge ();

  varlist = new int[length+2];
  for (i=0; i<length; i++)
    varlist[i] = FALSE;

  length = inpar.laenge ();
  for (i=0; i<length; i++)
    varlist[i] = TRUE;

extern int checkMFcand;
extern const char *CurrentMacro;
  checkMFcand = TRUE;

  oldMacroName = CurrentMacro;
  CurrentMacro = name;

#ifdef TTKDEBUG
extern int docheck;
  if (docheck)
    CheckTokenList (*globallist);
#else
    CheckTokenList (*globallist);
#endif
  delete[] varlist;
  varlist = NULL;

  CurrentMacro = oldMacroName;

  if (globalError.Get ())
    {
// don't show errors here
//    globalError.ShowError();    // now show errors, if there are any
    return (FALSE);
    }

  SetChecked ();
  return (TRUE);
}
