/* $Header: /home/xplore/cvs/xpl4gr/macro.h,v 1.3 2002/08/16 12:31:46 sigbert Exp $ */

// Macro Klasse fuer XploRe 4

// by TTK   10.10.94

/* the macro class helds following:

   Name:             name of macro
   XplObjectList:    List of local objects
   inpar:            names of input parameters
   outpar:           names of output parameters
   Text:             macro definition text
   allChecked:       Flag, whether TokenList is completely analyzed.
                     If true, TokenList can be processed by the parser,
                     if false, some scanning remains to be done.

*/

#ifndef _MACRO_H_
#define _MACRO_H_

#include <string.h>

#define FALSE 0
#define TRUE  1

#ifdef TTKDEBUG
extern int ttkdebug;
#endif

#define TEXTLENGTH  1024L
#define OFFSET      1024L

class Macro {
 public:  Macro (const char*, const char*, Liste<const char*>, Liste<const char*>, class GlobalList*);
          ~Macro ();
	  void                SetNewDefinition 
     (const char*,  Liste<const char*>,  Liste<const char*>, class GlobalList* );
          void                PrepareToRun  ();    // dummy 
          void                Prepare2Run  ();     // real procedure
          Liste<XplObject*> * XplObjectList ()     { return (objList);    }
	  class GlobalList *  globalList    ()     { return (globallist); }
          const char*	      Name    ()    const  { return (name);       }
	  Liste<const char*>  inPar   ()    const  { return (inpar);      }
	  Liste<const char*>  outPar  ()    const  { return (outpar);     }
	  const char *        Text    ()    const  { return (text);       }
	  const int           Checked ()    const  { return (allChecked); }
	  void                SetChecked ()        { allChecked = TRUE;   }
	  const int           IsRunning  () const  { return (isRunning);  }
	  void                SetRun     (int r)   { isRunning = r;       }
	  const int           CheckMacro ();
          const int   operator< (const Macro &r)
	    {
	      return (strcmp (name, r.name) < 0);
	    }
          const int   operator> (const Macro &r)
	    {
	      return (strcmp (name, r.name) > 0);
	    }

 private: char		      *name, *text;
	  Liste<const char*>  inpar, outpar;
          class GlobalList    *globallist;
          Liste<XplObject*>   *objList;
          int                 isRunning, allChecked;    // flag for TokenList
};


class MacroText {
 public:  MacroText   (long n = TEXTLENGTH);
          ~MacroText  ()                { delete[] text;
#ifdef TTKDEBUG
if (ttkdebug)
cerr << "MacText:\tdestruct\t" << this  << endl;
#endif
 }
          void AppendText  (const char *);
          void SetNewText  ();
          void Record      (int mode)        { recordMode = mode; }
          void Rewind      (void)            { text[0] = '\0'; length = 0L; }
          const char* Text (void) const      { return (text); } 
          void DeleteText  (void)            { delete[] text; 
cerr << "## ~MacroText: " << length << endl;
					       length = maxLength = 0L;
					       recordMode = FALSE;
					     }

 private: char *text;
          long length, maxLength;
          int  recordMode, isRunning;
};

void BuildLocalXplObjectList (Liste<XplObject*> *objList, 
                              Liste<char*> &inp, Liste<char*> &outp);

#endif
