/* $Header: /home/xplore/cvs/xpl4gr/matrix.C,v 1.3 2002/08/16 12:31:46 sigbert Exp $ */

// Matrix Klasse  6.6.94  Thomas Koetter

// last modification:  14.7.94

// Surrogate for XplDouble
// slighty fitted:  dump, operator[]
// 13.2.95  TTK

// #define TTKDEBUG

// #pragma implementation

#ifdef MATRIX

#ifdef SOCKETSERVER
#include "xjprotocol.h"
#include "xsockio.h"        // nur wg xplmessage -> sockout
#define xplmessage sockout
#endif


extern "C" {                // for numerical recipes
/* #include "nr.h" */
#include "nrutil.h"
}


#ifdef TTKDEBUG
int ttkdebug;
#endif

#if !defined (__BORLANDC__) && !defined (WIN32) //LOOK!
#include <stream.h>
#else
#include <iostream.h>
#endif

#include <assert.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "liste.h"
#include "matrix.h"

/*
#include "ttkio.h"
#include "xplio.h"
*/

#ifdef __BORLANDC__
#include "xplwin.h"
#endif

#define MAX_LINE_LENGTH 2560	//Sw
#define MAX_WORD_LENGTH 256	//Sw
#define MAX_WORD_COUNT		1000//Sw

Matrix::~Matrix ()
{
#ifdef TTKDEBUG
extern int ttkdebug;
if (ttkdebug)
  cerr<<"Mat: destruct\t\t"<<this<<endl;
#endif
//cout << "free: " << row*col << endl;

  delete[] elem;
}

Matrix::Matrix (int row)
{
  assert (row > 0);

  this->row = row;
  col = 1;
  ScalarFlag = (row == 1) && (col == 1);
  elem = MatrixAlloc (row, col);
#ifdef TTKDEBUG
if (ttkdebug)
cerr << "Mat:\tconstruct\t" << this << endl;
#endif
}

Matrix::Matrix (int row, int col)
{
  assert (row > 0 && col > 0);

  this->row = row;
  this->col = col;
  ScalarFlag = (row == 1) && (col == 1);
  elem = MatrixAlloc (row, col);
#ifdef TTKDEBUG
if (ttkdebug)
cerr << "Mat:\tconstruct\t" << this << endl;
#endif
}

Matrix::Matrix (double start, double stop, double step)
{
  int		i, length;
  double	h = 0.0;
  
  step = (start > stop) ? -fabs(step) : fabs(step);
  
  length = 1 + (int) floor((stop - start) / step);

  elem = MatrixAlloc (length, 1);
  
  for (i=0; i<length; i++, h += step)
    elem[i] = start + h;
  
  row = length;
  col = 1;
  ScalarFlag = (row == 1) && (col == 1);
  
#ifdef TTKDEBUG
if (ttkdebug)
cerr << "Mat:\tconstruct\t" << this << endl;
#endif
}

/* is it need in Motif version ?  20.12.98
Matrix::Matrix (int row, int col, int i)
{
  int n, m;
  
  assert (row > 0 && col > 0);
  assert (i == 1 && row == col);
  if (! (i==1 && row==col) )
    {
    ttkerror("Error: Constructor Matrix: i must be 1");
    exit (0);
    }
  this->row = row;
  this->col = col;
  ScalarFlag = (row == 1) && (col == 1);
  elem = MatrixAlloc (row, col);
  for (n=0; n<row; n++)
  for (m=0; m<col; m++)
  if (m==n) 
  	elem[m*row+n]=1;
  else elem[m*row+n]=0;
#ifdef TTKDEBUG
if (ttkdebug)
cerr << "Mat:\tconstruct\t" << this << endl;
#endif
}
*/


Matrix::Matrix (const Matrix& v)
{
  int	length;
  
  row = v.row;
  col = v.col;
  ScalarFlag = (row == 1) && (col == 1);
  length = row*col;
  elem = MatrixAlloc (row, col);
  
  //  for (i=0; i<length; i++)
  //    elem[i] = v.elem[i];

  memcpy (elem, v.elem, length*sizeof (double));
  
#ifdef TTKDEBUG
if (ttkdebug)
cerr << "Mat:\tconstruct\t" << this << endl;
#endif
  return;
}  
   
Matrix::Matrix (Matrix *v)
{
  int	length;
  
  if (v == NULL)
    {
    elem = NULL;
    row = col = 0;
    return;
    }
  
  row = v->row;
  col = v->col;
  ScalarFlag = (row == 1) && (col == 1);
  length = row*col;
  elem = MatrixAlloc (row, col);
    
  //  for (i=0; i<length; i++)
  //    elem[i] = v->elem[i];
  
  memcpy (elem, v->elem, length*sizeof (double));

#ifdef TTKDEBUG
if (ttkdebug)
cerr << "Mat:\tconstruct\t" << this << endl;
#endif
  return;
}  

#ifdef NDEBUG
double& Matrix::operator() (int i)
// i : 0 .. row-1, j=0
{
  assert (i < row-1 && col == 1);

  return (elem[i]);
}

double Matrix::operator() (int i) const
// i : 0 .. row-1, j=0
{
  assert (i < row-1 && col == 1);

  return (elem[i]);
}

double& Matrix::operator() (int i, int j)
// i : 0 .. row-1, j : 0 .. col-1
{
  assert (i < row-1 && j < col-1);

  return (elem[i+j*row]);
}

double Matrix::operator() (int i, int j) const
// i : 0 .. row-1, j : 0 .. col-1
{
  assert (i < row-1 && j < col-1);

  return (elem[i+j*row]);
}
#endif   // OPTIMZE

Matrix& Matrix::operator= (const Matrix& v)
{
  int 	length;

  if (this == &v)  return (*this);
  
  row = v.row;
  col = v.col;
  
  if (elem != NULL)
    delete[] elem;
  elem = MatrixAlloc (row, col);

  length = row*col;
  
  memcpy (elem, v.elem, length*sizeof (double));

  return (*this);
}


Matrix Matrix::operator+ (const Matrix& r)
{
  int		i, len;
  double	t;
  
  if (r.isScalar())
    {
    Matrix	res(row, col);
    len = row*col;
    t = r[0];
    for (i=0; i<len; i++)
      res.elem[i] = elem[i] + t;
    
    return (res);
    }
  
  if (isScalar())
    {
    Matrix	res(r.row, r.col);
    len = r.row*r.col;
    t = elem[0];
    for (i=0; i<len; i++)
      res.elem[i] = r.elem[i] + t;
    
    return (res);
    }
  
  if (match(r))
    {
    Matrix	res(row, col);
    len = row*col;
    
    for (i=0; i<len; i++)
      res.elem[i] = elem[i] + r.elem[i];
    
    return (res);
    }

  if (row == r.row)
    {
    int j, p = 0, rp = 0;
    int c = (col > r.col) ? col : r.col;
    Matrix res (row, c);

    for (i=0; i<row; i++)
      for (j=0; j<c; j++)
      {
      res(i, j) = (*this)(i, p++) + r(i, rp++);
      if (p >= col)
	p = 0;
      if (rp >= r.col)
	rp = 0;
      }
    return (res);
    }

  if (col == r.col)
    {
    int j, p, rp;
    int rr = (row > r.row) ? row : r.row;
    Matrix res (rr, col);

    p = rp = 0;
    for (i=0; i<rr; i++)
      {
      for (j=0; j<col; j++)
        res(i, j) = (*this)(p, j) + r(rp, j);

      p++;  rp++;
      if (p >= row)
	p = 0;
      if (rp >= r.row)
	rp = 0;
      }
    return (res);
    }

  assert (0);     // error message

  return (Matrix(1));
}

Matrix Matrix::operator- (const Matrix& r)
{
  int		i, len;
  double	t;
  
  if (r.isScalar())
    {
    Matrix	res(row, col);
    len = row*col;
    t = r[0];
    for (i=0; i<len; i++)
      res.elem[i] = elem[i] - t;
    
    return (res);
    }
  
  if (isScalar())
    {
    Matrix	res(r.row, r.col);
    len = r.row*r.col;
    t = elem[0];
    for (i=0; i<len; i++)
      res.elem[i] = t - r.elem[i];
 
    return (res);
    }

  if (match(r))
    {
    Matrix	res(row, col);
    len = row*col;
    
    for (i=0; i<len; i++)
      res.elem[i] = elem[i] - r.elem[i];
    
    return (res);
    }

  if (row == r.row)
    {
    int j, p = 0, rp = 0;
    int c = (col > r.col) ? col : r.col;
    Matrix res (row, c);

    for (i=0; i<row; i++)
      for (j=0; j<c; j++)
      {
      res(i, j) = (*this)(i, p++) - r(i, rp++);
      if (p >= col)
	p = 0;
      if (rp >= r.col)
	rp = 0;
      }
    return (res);
    }

  if (col == r.col)
    {
    int j, p, rp;
    int rr = (row > r.row) ? row : r.row;
    Matrix res (rr, col);

    p = rp = 0;
    for (i=0; i<rr; i++)
      {
      for (j=0; j<col; j++)
        res(i, j) = (*this)(p, j) - r(rp, j);

      p++;  rp++;
      if (p >= row)
	p = 0;
      if (rp >= r.row)
	rp = 0;
      }
    return (res);
    }

  assert (0);     // error message

  return (Matrix(1));
}

Matrix Matrix::operator* (const Matrix& r)
{
  int		i, len;
  double	t;
  
  if (r.isScalar())
    {
    Matrix	res(row, col);
    len = row*col;
    t = r[0];
    for (i=0; i<len; i++)
      res.elem[i] = elem[i] * t;
    
    return (res);
    }
  
  if (isScalar())
    {
    Matrix	res(r.row, r.col);
    len = r.row*r.col;
    t = elem[0];
    for (i=0; i<len; i++)
      res.elem[i] = t * r.elem[i];
 
    return (res);
    }
  
  if (match(r))
    {
    Matrix	res(row, col);
    len = row*col;
    
    for (i=0; i<len; i++)
      res.elem[i] = elem[i] * r.elem[i];
    
    return (res);
    }

  if (row == r.row)
    {
    int j, p = 0, rp = 0;
    int c = (col > r.col) ? col : r.col;
    Matrix res (row, c);

    for (i=0; i<row; i++)
      for (j=0; j<c; j++)
      {
      res(i, j) = (*this)(i, p++) * r(i, rp++);
      if (p >= col)
	p = 0;
      if (rp >= r.col)
	rp = 0;
      }
    return (res);
    }

  if (col == r.col)
    {
    int j, p, rp;
    int rr = (row > r.row) ? row : r.row;
    Matrix res (rr, col);

    p = rp = 0;
    for (i=0; i<rr; i++)
      {
      for (j=0; j<col; j++)
        res(i, j) = (*this)(p, j) * r(rp, j);

      p++;  rp++;
      if (p >= row)
	p = 0;
      if (rp >= r.row)
	rp = 0;
      }
    return (res);
    }

  assert (0);          // error message
  return (Matrix(1));  // dummy for pedantic compilers
}

Matrix Matrix::operator/ (const Matrix& r)
{
  int		i, len;
  double	t;
  
  if (r.isScalar())
    {
    Matrix	res(row, col);
    len = row*col;
    t = r[0];
    for (i=0; i<len; i++)
      res.elem[i] = elem[i] / t;
    
    return (res);
    }
  
  if (isScalar())
    {
    Matrix	res(r.row, r.col);
    len = r.row*r.col;
    t = elem[0];
    for (i=0; i<len; i++)
      res.elem[i] = t / r.elem[i];
 
    return (res);
    }
  
  if (match(r))
    {
    Matrix	res(row, col);
    len = row*col;
    
    for (i=0; i<len; i++)
      res.elem[i] = elem[i] / r.elem[i];
    
    return (res);
    }

  if (row == r.row)
    {
    int j, p = 0, rp = 0;
    int c = (col > r.col) ? col : r.col;
    Matrix res (row, c);

    for (i=0; i<row; i++)
      for (j=0; j<c; j++)
      {
      res(i, j) = (*this)(i, p++) / r(i, rp++);
      if (p >= col)
	p = 0;
      if (rp >= r.col)
	rp = 0;
      }
    return (res);
    }

  if (col == r.col)
    {
    int j, p, rp;
    int rr = (row > r.row) ? row : r.row;
    Matrix res (rr, col);

    p = rp = 0;
    for (i=0; i<rr; i++)
      {
      for (j=0; j<col; j++)
        res(i, j) = (*this)(p, j) / r(rp, j);

      p++;  rp++;
      if (p >= row)
	p = 0;
      if (rp >= r.row)
	rp = 0;
      }
    return (res);
    }

  assert (0);          // error message
  return (Matrix(1));  // dummy for pedantic compilers
}

Matrix Matrix::operator% (const Matrix& r)
{
  int		i, len;
  double	t;
  
  if (r.isScalar())
    {
    Matrix	res(row, col);
    len = row*col;
    t = r[0];
    for (i=0; i<len; i++)
      res.elem[i] = fmod (elem[i], t);
    
    return (res);
    }
  
  if (isScalar())
    {
    Matrix	res(r.row, r.col);
    len = r.row*r.col;
    t = elem[0];
    for (i=0; i<len; i++)
      res.elem[i] = fmod(t, r.elem[i]);
 
    return (res);
    }
  
  if (match(r))
    {
    Matrix	res(row, col);
    len = row*col;
    
    for (i=0; i<len; i++)
      res.elem[i] = elem[i] * r.elem[i];
    
    return (res);
    }

  if (row == r.row)
    {
    int j, p = 0, rp = 0;
    int c = (col > r.col) ? col : r.col;
    Matrix res (row, c);

    for (i=0; i<row; i++)
      for (j=0; j<c; j++)
      {
      res(i, j) = fmod ((*this)(i, p++), r(i, rp++));
      if (p >= col)
	p = 0;
      if (rp >= r.col)
	rp = 0;
      }
    return (res);
    }

  if (col == r.col)
    {
    int j, p, rp;
    int rr = (row > r.row) ? row : r.row;
    Matrix res (rr, col);

    p = rp = 0;
    for (i=0; i<rr; i++)
      {
      for (j=0; j<col; j++)
        res(i, j) = fmod ((*this)(p, j), r(rp, j));

      p++;  rp++;
      if (p >= row)
	p = 0;
      if (rp >= r.row)
	rp = 0;
      }
    return (res);
    }

  assert (0);          // error message
  return (Matrix(1));  // dummy for pedantic compilers
}

Matrix Matrix::operator== (const Matrix& r)
{
  int		i, len;
  double	t;
  
  if (r.isScalar())
    {
    Matrix	res(row, col);
    len = row*col;
    t = r[0];
    for (i=0; i<len; i++)
      res.elem[i] = elem[i] == t;
    
    return (res);
    }
  
  if (isScalar())
    {
    Matrix	res(r.row, r.col);
    len = r.row*r.col;
    t = elem[0];
    for (i=0; i<len; i++)
      res.elem[i] = r.elem[i] == t;
    
    return (res);
    }
  
  assert (match (r));          // error message
  
  Matrix	res(row, col);
  len = row*col;

  for (i=0; i<len; i++)
    res.elem[i] = elem[i] == r.elem[i];
  
  return (res);
}

Matrix Matrix::operator!= (const Matrix& r)
{
  int		i, len;
  double	t;
  
  if (r.isScalar())
    {
    Matrix	res(row, col);
    len = row*col;
    t = r[0];
    for (i=0; i<len; i++)
      res.elem[i] = elem[i] != t;
    
    return (res);
    }
  
  if (isScalar())
    {
    Matrix	res(r.row, r.col);
    len = r.row*r.col;
    t = elem[0];
    for (i=0; i<len; i++)
      res.elem[i] = r.elem[i] != t;
    
    return (res);
    }
  
  assert (match (r));          // error message
  
  Matrix	res(row, col);
  len = row*col;

  for (i=0; i<len; i++)
    res.elem[i] = elem[i] != r.elem[i];
  
  return (res);
}

Matrix Matrix::operator< (const Matrix& r)
{
  int		i, len;
  double	t;
  
  if (r.isScalar())
    {
    Matrix	res(row, col);
    len = row*col;
    t = r[0];
    for (i=0; i<len; i++)
      res.elem[i] = elem[i] < t;
    
    return (res);
    }
  
  if (isScalar())
    {
    Matrix	res(r.row, r.col);
    len = r.row*r.col;
    t = elem[0];
    for (i=0; i<len; i++)
      res.elem[i] = r.elem[i] < t;
    
    return (res);
    }
  
  assert (match (r));          // error message
  
  Matrix	res(row, col);
  len = row*col;

  for (i=0; i<len; i++)
    res.elem[i] = elem[i] < r.elem[i];
  
  return (res);
}

Matrix Matrix::operator<= (const Matrix& r)
{
  int		i, len;
  double	t;
  
  if (r.isScalar())
    {
    Matrix	res(row, col);
    len = row*col;
    t = r[0];
    for (i=0; i<len; i++)
      res.elem[i] = elem[i] <= t;
    
    return (res);
    }
  
  if (isScalar())
    {
    Matrix	res(r.row, r.col);
    len = r.row*r.col;
    t = elem[0];
    for (i=0; i<len; i++)
      res.elem[i] = r.elem[i] <= t;
    
    return (res);
    }
  
  assert (match (r));          // error message
  
  Matrix	res(row, col);
  len = row*col;

  for (i=0; i<len; i++)
    res.elem[i] = elem[i] <= r.elem[i];
  
  return (res);
}

Matrix Matrix::operator> (const Matrix& r)
{
  int		i, len;
  double	t;
  
  if (r.isScalar())
    {
    Matrix	res(row, col);
    len = row*col;
    t = r[0];
    for (i=0; i<len; i++)
      res.elem[i] = elem[i] > t;
    
    return (res);
    }
  
  if (isScalar())
    {
    Matrix	res(r.row, r.col);
    len = r.row*r.col;
    t = elem[0];
    for (i=0; i<len; i++)
      res.elem[i] = r.elem[i] > t;
    
    return (res);
    }
  
  assert (match (r));          // error message
  
  Matrix	res(row, col);
  len = row*col;

  for (i=0; i<len; i++)
    res.elem[i] = elem[i] > r.elem[i];
  
  return (res);
}

Matrix Matrix::operator>= (const Matrix& r)
{
  int		i, len;
  double	t;
  
  if (r.isScalar())
    {
    Matrix	res(row, col);
    len = row*col;
    t = r[0];
    for (i=0; i<len; i++)
      res.elem[i] = elem[i] >= t;
    
    return (res);
    }
  
  if (isScalar())
    {
    Matrix	res(r.row, r.col);
    len = r.row*r.col;
    t = elem[0];
    for (i=0; i<len; i++)
      res.elem[i] = r.elem[i] >= t;
    
    return (res);
    }
  
  assert (match (r));          // error message
  
  Matrix	res(row, col);
  len = row*col;

  for (i=0; i<len; i++)
    res.elem[i] = elem[i] >= r.elem[i];
  
  return (res);
}

Matrix Matrix::operator&& (const Matrix& r)
{
  int		i, len;
  double	t;
  
  if (r.isScalar())
    {
    Matrix	res(row, col);
    len = row*col;
    t = r[0];
    for (i=0; i<len; i++)
      res.elem[i] = elem[i] && t;
    
    return (res);
    }
  
  if (isScalar())
    {
    Matrix	res(r.row, r.col);
    len = r.row*r.col;
    t = elem[0];
    for (i=0; i<len; i++)
      res.elem[i] = r.elem[i] && t;
    
    return (res);
    }
  
  assert (match (r));          // error message
  
  Matrix	res(row, col);
  len = row*col;

  for (i=0; i<len; i++)
    res.elem[i] = elem[i] && r.elem[i];
  
  return (res);
}

Matrix Matrix::operator|| (const Matrix& r)
{
  int		i, len;
  double	t;
  
  if (r.isScalar())
    {
    Matrix	res(row, col);
    len = row*col;
    t = r[0];
    for (i=0; i<len; i++)
      res.elem[i] = elem[i] || t;
    
    return (res);
    }
  
  if (isScalar())
    {
    Matrix	res(r.row, r.col);
    len = r.row*r.col;
    t = elem[0];
    for (i=0; i<len; i++)
      res.elem[i] = r.elem[i] || t;
    
    return (res);
    }
  
  assert (match (r));          // error message
  
  Matrix	res(row, col);
  len = row*col;

  for (i=0; i<len; i++)
    res.elem[i] = elem[i] || r.elem[i];
  
  return (res);
}

double 	Matrix::min_col(int c) const
{
  assert (0 < c && c < col-1);
  //	if ((c>col-1)||(c<0))
  //	{
  //		ttkerror("ERROR: min_col Matrice has no such column");
  //		exit(0);
  //	}
	double min;
	int i;
  	min=elem[c*row];
	for (i=1; i<row; i++)
  		if (elem[i+c*row]<min) 
  		 	min=elem[i+c*row]; 
	return min;
}

double 	Matrix::max_col(int c) const
{
  assert (0 < c && c < col-1);
  //	if ((c>col-1)||(c<0))
  //	{
  //		ttkerror("ERROR: max_col Matrice has no such column");
  //		exit(0);
  //	}
	double max;
	int i;
  	max=elem[c*row];
	for (i=1; i<row; i++)
  		if (elem[i+c*row]>max) 
  		 	max=elem[i+c*row]; 
	return max;
}

Matrix* Matrix::exec (singlefun fun)
{
  int i, length;
  
  length = row*col;
  
  for (i=0; i<length; i++)    
    elem[i] = fun(elem[i]);
  
  return (this);
}

Matrix*	Matrix::selection (Liste<Matrix*> argl)
{
  int		length, i, j;
  Matrix	rs, cs, *res;
  
  length = argl.laenge();
  
  assert (length < 3);
  //  if (length > 2)
  //    {
  //    ttkerror ("Matrix::selection: Error:  List of args greater than 2");
  //    return (NULL);
  //    }
  
  if (argl[0] == NULL)
    rs = Matrix (1.0, (double) row, 1.0);	// Indexvektor
  else
    rs = argl[0];

  if (length == 1)
    cs = Matrix (1.0, 1.0, 1.0);	// Indexvektor
  else  
    if (argl[1] == NULL)
      cs = Matrix (1.0, (double) col, 1.0);	// Indexvektor
    else
      cs = argl[1];
  
  if (rs.cols()*cs.cols() != 1)
    {
    error ("Matrix::selection: Error:  Args are not vectors");
    return (NULL);
    }
  
  res = new Matrix (rs.rows(), cs.rows());
  
  for (i=0; i<res->rows(); i++)
    for (j=0; j<res->cols(); j++)
      (*res)(i,j) = (*this)((int)rs[i]-1, (int)cs[j]-1);
  
  return (res);
}

int	Matrix::selectassign (Liste<Matrix*> argl, Matrix *m)
{
  int		length, i, j;
  Matrix	rs, cs;
  
  length = argl.laenge();
  
  if (length > 2)
    {
    error ("Matrix::selectassign: Error:  List of args greater than 2");
    return (1);
    }
  
  if (argl[0] == NULL)
    rs = Matrix (1.0, 1.0, 1.0);	// Indexvektor
  else
    rs = argl[0];
  
  if (length == 1)
    cs = Matrix (1.0, (double) col, 1.0);	// Indexvektor
  else  
    if (argl[1] == NULL)
      cs = Matrix (1.0, (double) col, 1.0);	// Indexvektor
    else
      cs = argl[1];
  
  if (rs.cols()*cs.cols() != 1)
    {
    error ("Matrix::selectassign: Error:  Args are not vectors");
    return (1);
    }
  
  for (i=0; i<rs.rows(); i++)
    for (j=0; j<cs.rows(); j++)
      (*this)((int)rs[i]-1,(int)cs[j]-1) = (*m)(i,j);

  
  return (1);
}


void Matrix::mult (Matrix &r)
{
  int i, j, k, n, m, p;
  double sum = 0.0;
  
  if (isScalar() || r.isScalar())
    {
    *this = *this * r;
    return;
    }

  Matrix tmp(row, r.col);
  
  if (col != r.row)
    {
    error("Matrix::mult: Error:  Matrices don't match");
    return;
    }

  n = row;
  m = col;
  p = r.col;
  
  for (i=0; i<n; i++)
    for (j=0; j<p; j++)
      {
      for (k=0; k<m; k++)
        sum += (*this)(i,k) * r(k,j);
      tmp(i,j) = sum;
      sum = 0.0;
      }


/*
  for (i=0; i<n; i++)
    {
    s = r.elem + i*p;
    for (j=0; j<p; j++)
      {
      t = elem + i;
      for (k=0; k<m; k++)
	{
        sum += *t * *s++;
	t += p;
        }
      tmp(i,j) = sum;
      sum = 0.0;
      }
    }
*/
  *this = tmp;
}

void Matrix::tmult (Matrix &r)
// like mult only this is regarded as transposed
{
  int i, j, k, n, m, p;
  double sum = 0.0;
  
  if (isScalar() || r.isScalar())
    {
    *this = *this * r;
    return;
    }

  Matrix tmp(col, r.col);
  
  if (row != r.rows())
    {
    error("Matrix::tmult: Error:  Matrices don't match");
    return;
    }

  n = row;
  m = col;
  p = r.col;
  
  for (i=0; i<m; i++)
    for (j=0; j<p; j++)
      {
      for (k=0; k<n; k++)
        sum += (*this)(k,i) * r(k,j);
      tmp(i,j) = sum;
      sum = 0.0;
      }


/*
  for (i=0; i<n; i++)
    {
    s = r.elem + i*p;
    for (j=0; j<p; j++)
      {
      t = elem + i;
      for (k=0; k<m; k++)
	{
        sum += *t * *s++;
	t += p;
        }
      tmp(i,j) = sum;
      sum = 0.0;
      }
    }
*/
  *this = tmp;
}
  

Matrix* Matrix::transpose ()
{
  Matrix tmp(col, row);
  int	i, j;
  
  for (i=0; i<row; i++)
    for (j=0; j<col; j++)
      tmp(j,i) = (*this)(i,j);
  
  *this = tmp;
  
  return (this);
}


void Matrix::out (void)
{
  int i, j, length, dim;
  
  length = row*col;
  
  if (col == 1)
    dim = 1;
  else
    dim = 2;

#ifndef INETD
#if !defined(_GRAPHICS_) && !defined(SOCKETSERVER)
  cout << "Dim: " << dim << '\n';  
#else
  xplmessage ("Dim: ");
  xplmessage ((long) dim);
  xplmessage ('\n');
  
#endif
#endif

  switch (dim) {
  case 1:	for (i=0; i<length; i++)
                  {
#if !defined(_GRAPHICS_) && !defined(SOCKETSERVER)
  		  cout << elem[i] << '\t'; 
#else
		  xplmessage (elem[i]);
		  xplmessage ('\t');
#endif
		  }
		  break;
  case 2:	for (i=0; i<row; i++)
  		  {
  		  for (j=0; j<col; j++)
		    {
#if !defined(_GRAPHICS_) && !defined(SOCKETSERVER)
  		    cout << (*this)(i,j) << '\t';
#else
		    xplmessage ((*this)(i,j));
		    xplmessage ('\t');
#endif
		    }
#if !defined(_GRAPHICS_) && !defined(SOCKETSERVER)
  		  cout << '\n';
#else
		  xplmessage ('\n');
#endif
  		  }
  		break;
  default:	assert (0);
    //          ttkerror("Error: out Matrix");
    //  	exit (0);
  		break;
  }
    
#if !defined(_GRAPHICS_) && !defined(SOCKETSERVER)
  		  cout << '\n';
#else
		  xplmessage ('\n');
#endif

//  cout << endl;
}

#ifdef SOCKETSERVER
void Matrix::out (SocketIO *sock)
{
  int i, j, length, dim;
  char line[2048];

  
  length = row*col;
  
  if (col == 1)
    dim = 1;
  else
    dim = 2;

  sprintf (line, "Dim: %d\n", dim);
  protocol->sendString (line);

  line[0] = '\0';

  switch (dim) {
  case 1:	for (i=0; i<length; i++)
                  {
		  sprintf (line, "%s%lg\t", line, elem[i]);   
		  protocol->sendString (line);
		  line[0] = '\0';
		  }
		  break;
  case 2:	for (i=0; i<row; i++)
  		  {
  		  for (j=0; j<col; j++)
		    {
		    sprintf (line, "%s%lg\t", line, (*this)(i,j));   
		    if (strlen (line) > 1900)
		      {
		      protocol->sendString (line);
		      line[0] = '\0';
		      }
		    }
		  sprintf (line, "%s\n", line);
		  if (strlen (line) > 1900)
		    {
		    protocol->sendString (line);
		    line[0] = '\0';
		    }
  		  }
                protocol->sendString (line);
  		break;
  default:	assert (0);
    //                ttkerror("Error: out Matrix");
    //  		exit (0);
  		break;
  }
}
#endif  // SOCKETSERVER


void Matrix::fill (const double v)
{
  long i, n = row*col;
  double *p = elem;

  for (i=0; i<n; i++)
    *p++ = v;
}

void Matrix::fill (const Matrix &v)
{
  long i, j, z, nv, n = row*col;
  double *p = elem, *pv = v.elem;

  if ((row == v.row) && (col == v.col))          // matrices match
    {
    for (i=0; i<n; i++)
      *p++ = *pv++;
    return;
    }

  if (row == v.row)                      // number of rows are equal
    {
    for (i=0; i<col; i++)
      {
      if (i > v.col-1)                   // if v has less cols, reset pv
        pv = v.elem;
      for (j=0; j<row; j++)
        *p++ = *pv++;
      }
    return;
    }

  if (col == v.col)                     // number of columns is equal
    {
    for (i=0; i<row; i++)
      {
      p  = elem + i;
      pv = v.elem + i;
      if (i > v.row-1)                  // if v has less rows, reset pv
        pv = v.elem;
      for (j=0; j<col; j++)
        {
        *p = *pv;
        p  += row;                      // increment by cols !!
        pv += v.row;
        }
      }
    return;
    }

  nv = v.col * v.row;                   // neither columns nor rows are equal
  z = 0;

  for (i=0; i<n; i++, z++)
    {
    if (z >= nv)                        // repeat v by rows 
      {
      pv = v.elem;
      z = 0;
      }
    *p++ = *pv++;
    }
}

double * MatrixAlloc (int row, int col)
{
  double *res, *p;
  
  res = new double[row*col];

  p = res;

  //  for (i=0; i<n; i++)
  //    *p++ = 0.0;

  return (res);
}

int ReadMatrix(const char *datei, Matrix &m )
// 11.11 Swetlana Schmelzer
{	long col_current, words_da, words_needed, i, j;
	char line[MAX_LINE_LENGTH], *word;
	float zahl;
	Matrix empty;
	double *ptr;
	
	fstream f(datei, ios::in);
	if ( f.bad() )
	  {	cerr << "fstream is bad" << endl;
	//		ttkerror("fstream is bad");
		return(0);
	}
	m.row=m.col=words_da=0;
	words_needed=MAX_WORD_COUNT;
	delete [] m.elem;
	m.elem = new double [words_needed];
/*
	if (m.elem==NULL)
	{  	ttkerror("Error: No memory enough for ");
		ttkerror((long) words_needed);
	        ttkerror(" double");
		m=empty;
    		return(0);
    	}
    	ttkmessage("Bitte warten!");
*/
	while ( f.getline(line, MAX_LINE_LENGTH-1, '\n'))
	{	col_current=0;
		word=strtok(line, " \t\n");
		while ( word )
		{	if (! (sscanf(word, "%g", &zahl)==1) )
		  //			{	ttkerror("readerror");
		        {	cerr << "readerror" << endl;
				m=empty;
				return(0);
			}
			m.elem[words_da]=(double)zahl;
			col_current++;
			words_da++;
			if (words_da==words_needed)
			{	words_needed+=MAX_WORD_COUNT;	
				m.elem=(double *)
				realloc(m.elem, sizeof(double)*words_needed);
			} 
			word=strtok(NULL, " \t\n");
		}
		if (m.col==0)	//erste zeile
			m.col=col_current;
		else
		if (! ( m.col==col_current) )
		  {	m=empty;
			return(0);
		}
		m.row++;
	}
  	
  	ptr=new double [words_da];
 	for (i=0; i<m.row; i++)
    		for (j=0; j<m.col; j++)
    			ptr[j*m.row+i]=m.elem[i*m.col+j];
  
	delete [] m.elem;
	m.elem=ptr;
	return(1);
}

int Matrix::inverse ()
// Numerical Recipes in C, p. 45
{
  double **a, **ai, *column, d;
  int i, j;
  int *index;

  int n = row;

  if (col != row)
    {
    cerr << "inverse:  Error:  Matrix not symmetric" << endl;
    exit (1);
    }


  a   = dmatrix(1,n,1,n);
  ai  = dmatrix(1,n,1,n);
  column = dvector(1,n);


  for (i=1; i<=n; i++)
    for (j=1; j<=n; j++)
      a[i][j] = (*this)(i-1,j-1);

  index  = new int[n+1];

  ludcmp (a, n, index, &d);

  for (j=1; j<=n; j++)
    {
    for (i=1; i<=n; i++)
      column[i] = 0.0;
    column[j] = 1.0;
    lubksb (a, n, index, column);
    for (i=1; i<=n; i++)
      ai[i][j] = column[i];
    }

  for (i=1; i<=n; i++)
    for (j=1; j<=n; j++)
      (*this)(i-1,j-1) = ai[i][j];

  free_dmatrix (a, 1, n, 1, n);
  free_dmatrix (ai, 1, n, 1, n);

  free_dvector (column, 1, n);

  delete index;

 return (OK);

}

/*
int Matrix::solve (const Matrix &b)
// Numerical Recipes in C, p. 45
{
  double **a, **ai, *column, d;
  int i, j;
  int *index;

  int n = row;

  if (col != b.row || b.col != 1)
    {
    cerr << "solve:  Error:  Vector does not fit" << endl;
    exit (1);
    }


  a   = dmatrix(1,row,1,col);
  ai  = dmatrix(1,row,1,col);
  column = dvector(1,col);


  for (i=1; i<=n; i++)
    for (j=1; j<=n; j++)
      a[i][j] = (*this)(i-1,j-1);

  index  = new int[n+1];

  ludcmp (a, n, index, &d);

  for (j=1; j<=n; j++)
    {
    for (i=1; i<=n; i++)
      column[i] = 0.0;
    column[j] = 1.0;
    lubksb (a, n, index, column);
    for (i=1; i<=n; i++)
      ai[i][j] = column[i];
    }

  for (i=1; i<=n; i++)
    for (j=1; j<=n; j++)
      (*this)(i-1,j-1) = ai[i][j];

  free_dmatrix (a, 1, n, 1, n);
  free_dmatrix (ai, 1, n, 1, n);

  free_dvector (column, 1, n);

  delete index;

 return (OK);

}
*/

int Matrix::LUsubstitution ( Matrix &index, Matrix &b)
// Numerical Recipes in C, p. 44
{
  long   i, ii = 0, ip, j, n = row;
  double sum;

  if (col != n)
    return (NONSYMMETRIC);

  for (i=1; i<=n; i++)
    {
    ip = (long) index(i-1);
    sum = b(ip-1);
    b(ip-1) = b(i-1) ;
    if (ii)
      for (j=ii; j<=i-1; j++)
	sum -= (*this)(i-1, j-1) * b(j-1);
    else 
      if (sum)
	ii=i;
    b(i-1) = sum;
    }

    for (i=n; i>=1; i--)
      {
      sum = b(i-1);
      for (j=i+1; j<=n; j++)
	sum -= (*this)(i-1, j-1) * b(j-1);
      b(i-1) = sum / (*this)(i-1, i-1);
      }
  return (OK);
}


#define TINY 0.0;

int Matrix::LUdecomposition ( Matrix &index, double &d)
// Numerical Recipes in C, p. 43
{
  long    i, j, k, imax, n = row;
  double  big, dum, sum, temp;

  if (col != n)
    return (NONSYMMETRIC);

  Matrix vv(1.0,(double) n,1.0);

  d = 1.0;
  for (i=1; i<=n; i++)
    {
    big = 0.0;
    for (j=1; j<=n; j++)
      if ((temp = fabs ((*this)(i-1, j-1)) > big))
        big = temp;
    if (big == 0.0)
      return (SINGULAR);

    vv(i-1) = 1.0 / big;
    }

  for (j=1; j<=n; j++)
    {
    for (i=1; i<j; i++)
      {
      sum = (*this)(i-1, j-1);
      for (k=1; k<i; k++)
	sum -= (*this)(i-1, k-1) * (*this)(k-1, j-1);
      (*this)(i-1, j-1) = sum;
      }

    big = 0.0;
    for (i=j; i<=n; i++)
      {
      sum = (*this)(i-1, j-1);
      for (k=1; k<j; k++)
	sum -= (*this)(i-1, k-1) * (*this)(k-1, j-1);
      (*this)(i-1, j-1) = sum;
      if ((dum = vv(i-1) * fabs(sum)) >= big)
	{
        big = dum;
	imax = i;
        }

      if (j != imax)
	{
        for (k=1; k<=n; k++)
	  {
          dum = (*this)(imax-1, k-1);
          (*this)(imax-1, k-1) = (*this)(j-1, k-1);
          (*this)(j-1, k-1) = dum;
	  }

          d = -d;
          vv(imax-1) = vv(j-1);
        }
      index(j-1) = imax;

      if ((*this)(j-1, j-1) == 0.0)
	(*this)(j-1, j-1) = TINY;

      if  (j != n)
	{
        dum = 1.0/(*this)(j-1, j-1);
        for (i=j+1; i<=n; i++)
	  (*this)(i-1, j-1) *= dum;
        }
      }
    }
  return (OK);
}

Matrix * Matrix::jacobi (Matrix *evalues)
{
  double **a, **v, *d;
  int i, j;
  int n, nrot;

  n = row;

  if (col != row)
    {
    cerr << "jacobi:  Error:  Matrix not symmetric" << endl;
    exit (1);
    }

  a = dmatrix (1, n, 1, n);
  v = dmatrix (1, n, 1, n);
  d = new double[n+1];

  for (i=1; i<=n; i++)
    for (j=1; j<=n; j++)
      a[i][j] = (*this)(i-1,j-1);

  ::jacobi(a,n,d,v,&nrot);

  Matrix *evecs = new Matrix (n,n);
  Matrix evals(n);

  for (i=1; i<=n; i++)
    {
    evals(i-1) = d[i];
    for (j=1; j<=n; j++)
      (*evecs)(i-1,j-1) = v[i][j];
    }

  (*evalues) = evals;

  free_dmatrix (a, 1, n, 1, n);
  free_dmatrix (v, 1, n, 1, n);

  delete d;

  return (evecs);
}
  
  

void  Matrix::sort (Matrix &index)
/*  sort first column of a matrix, index is the vector of the order of the
    original values
*/
{
  Matrix order(0.0, (double) row-1.0, 1.0);

  SortMatrix (*this, order, 0, row-1);

  index = order;
}


void SortMatrix (Matrix &feld, Matrix &ind, long l, long r)
{
  long   links, rechts, mitte;
  double hilf, x;

  links = l; rechts = r;
  mitte = (l+r)/2;
  x = feld(mitte, 0);
  do {
     while (feld(links, 0)  < x)  links++;
     while (feld(rechts, 0) > x)  rechts--;
     if (links <= rechts)
       {
       hilf = feld(rechts, 0);
       feld(rechts, 0) = feld(links, 0);
       feld(links, 0) = hilf;
       hilf = ind[rechts];
       ind[rechts] = ind[links];
       ind[links] = hilf;
       links++;  rechts--;
       }
     } while (links < rechts);

  if (l < rechts)
    SortMatrix (feld, ind, l, rechts);
  if (links < r)
    SortMatrix (feld, ind, links, r);
}

Matrix* ColumnBind (Matrix &l, Matrix &r)
{
  if (l.row != r.row)
    {
    cerr << "Error:  ColumnBind:  Different number of rows";
    return (NULL);
    }

  Matrix *erg = new Matrix (l.row, l.col+r.col);

  long i, j;

  for (i=0; i<l.row; i++) 
    for (j=0; j<l.col; j++)
      (*erg)(i, j) = l(i,j);

  for (i=0; i<r.row; i++) 
    for (j=0; j<r.col; j++)
      (*erg)(i, j+l.col) = r(i,j);

  return (erg);
}

Matrix* RowBind (Matrix &l, Matrix &r)
{
  if (l.col != r.col)
    {
    cerr << "Error:  ColumnBind:  Different number of rows";
    return (NULL);
    }

  Matrix *erg = new Matrix (l.row+r.row, l.col);

  long i, j;

  for (i=0; i<l.row; i++) 
    for (j=0; j<l.col; j++)
      (*erg)(i, j) = l(i,j);

  for (i=0; i<r.row; i++) 
    for (j=0; j<r.col; j++)
      (*erg)(i+l.row, j) = r(i,j);

  return (erg);
}

Matrix Matrix::operator+ (double d)
{
  Matrix res (*this);
  double *p = res.elem;
  long size = row*col;

  for (long i=0; i<size; i++)
    *p++ += d;

  return (res);
}

Matrix Matrix::operator- (double d)
{
  Matrix res (*this);
  double *p = res.elem;
  long   size = row*col;

  for (long i=0; i<size; i++)
    *p++ -= d;

  return (res);
}

Matrix Matrix::operator* (double d)
{
  long size = row*col;

  for (long i=0; i<size; i++)
    *elem *= d;
}

Matrix Matrix::operator/ (double d)
{
  long size = row*col;

  for (long i=0; i<size; i++)
    *elem /= d;
}

#endif    // of MATRIX
