/* $Header: /home/xplore/cvs/xpl4gr/matrix.h,v 1.3 2002/08/16 12:31:46 sigbert Exp $ */

// Matrix Klasse   TTK  6.6.94
// aus vector.C


// access over [] is now unsecure, no bounds checking is done
// access over () is secure
// last modification:  10.8.95

/*
Folgende Klasse Matrix wird definiert:

Konstruktoren:

-	Matrix ()
	erzeugt leere Matrix. row=col=0;

-	Matrix (int)
	erzeugt Vektor der vorgegebenen Laenge

-	Matrix (int, int);
	erzeugt Matrix mit vorgegebenen Zeilen und Spalten

-	Matrix (double start, double stop, double step)
	erzeugt Vektor mit start beginnend, sich um step veraendernd,
	letzter Wert <= stop

-	Matrix (int, int, int i);
	erzeugt EinheitsMatrix mit vorgegebenen Zeilen und Spalten,
	wenn i==1

Kopierkonstruktoren:

-	Matrix (const Matrix&)
	erzeugt tiefe Kopie der Referenzmatrix

-	Matrix (Matrix*);
	erzeugt tiefe Kopie der Matrix, auf der der Pointer zeigt


Destruktor:

-	~Matrix ()
	gibt des reservierten Speicher fuer die Elemente frei

Operatoren:

-	double& operator[] (int)
	Referenzzugriff auf Matrix mit einer Spalte

-	double  operator[] (int) const;
	Element aus const Matrix lesen

-	double& operator() (int, int);
	Referenzzugriff auf Matrix ueber Zeilen/Spalten
	
-	double  operator() (int, int) const;
	Element aus Matrix (Zeilen/Spalten) lesen
	
-	Matrix  operator+ (const Matrix&);
-	Matrix  operator- (const Matrix&);
-	Matrix  operator* (const Matrix&);
-	Matrix  operator/ (const Matrix&);
-	Matrix  operator< (const Matrix&);
-	Matrix  operator<= (const Matrix&);
-	Matrix  operator> (const Matrix&);
-	Matrix  operator>= (const Matrix&);
-	Matrix  operator== (const Matrix&);
-	Matrix  operator!= (const Matrix&);
-	Matrix  operator|| (const Matrix&);
-	Matrix  operator&& (const Matrix&);
-	Matrizen elementweise verknuepfen. (1x1) Matrix = Skalar
	wird expandiert.

-	Matrix& operator= (const Matrix&);
	Weist Matrix zu

Attributfunktionen:

-	int	isScalar (void) const;
	True, wenn row == col == 1

-	int	rows (void) const;
	liefert Anzahl der Zeilen

-	int	cols (void) const;
	liefert Anzahl der Spalten

Rechnen:

-	Matrix*	exec (singlefun);
	fuehrt die angegebene (double) singlefun(double) auf alle
	Elemente aus

-	void	mult (Matrix&);
	Multiplitziert uebergebene Matrix

-	void	tmult (Matrix&);
	Multiplitziert transponiertes this mit uebergebene Matrix

-	void	transpose (void);
	Transponieren

Spezial:

-	Matrix*	selection (Liste<Matrix*>)

-	int	match (const Matrix&);
	gibt True, wenn uebergeben Matrix gleich gross ist

Ausgabe:

-	void	out (void);
	gibt Matrix nach cout aus

Interna:

-	int row, col;
	Zeilen- und Spaltenzahl

-	double *elem;
	Array der Elemente

-	int ReadMatrix(char *, Matrix &)
	versucht Datei zu lesen und eine Matrix erzeugen
	dibt True, falls OK


*/



#ifndef _MATRIX_H
#define _MATRIX_H

// #pragma interface

#include "liste.h"
#include <fstream.h>
#include <stdlib.h>		//Sw

#ifdef SOCKETSERVER
#include "socketio.h"
#endif


typedef	double (*singlefun) (double);

#define OK           0
#define NONSYMMETRIC 1
#define SINGULAR     2



class Matrix {
  public:	Matrix ()		: row (0), col (0), elem (NULL) {
#ifdef TTKDEBUG
extern int ttkdebug;
if (ttkdebug)
  cerr<<"Mat:\tconstruct\t"<<this<<endl;
#endif
}
  		Matrix (int);
  		Matrix (int, int);
  		Matrix (double, double, double);
  		Matrix (int, int, int);
  		Matrix (const Matrix&);
  		Matrix (Matrix*);
		~Matrix ();
		double& operator[] (int i)        { return (elem[i]); }
		double  operator[] (int i) const  { return (elem[i]); }
#ifdef OPTIMIZE
		double& operator() (int i)           {return (elem[i]);      }
		double  operator() (int i) const     {return (elem[i]);      }
		double& operator() (int i, int j)    {return (elem[i+j*row]);}
		double  operator() (int i, int j) const {return(elem[i+j*row]);}
#else
		double& operator() (int i);
		double  operator() (int i) const;
		double& operator() (int, int);
		double  operator() (int, int) const;
#endif
		Matrix  operator+ (const Matrix&);
		Matrix  operator- (const Matrix&);
		Matrix  operator* (const Matrix&);
		Matrix  operator/ (const Matrix&);
		Matrix  operator% (const Matrix&);
		Matrix  operator== (const Matrix&);
		Matrix  operator!= (const Matrix&);
		Matrix  operator< (const Matrix&);
		Matrix  operator<= (const Matrix&);
		Matrix  operator> (const Matrix&);
		Matrix  operator>= (const Matrix&);
		Matrix  operator&& (const Matrix&);
		Matrix  operator|| (const Matrix&);
		Matrix& operator= (const Matrix&);

		Matrix  operator+ (double d);
		Matrix  operator- (double d);
		Matrix  operator* (double d);
		Matrix  operator/ (double d);
		int     operator> (double d)    { return (*elem > d); }
		int     operator< (double d)    { return (*elem < d); }


		int	isScalar (void) const	{ return (ScalarFlag); }
//		int	isScalar (void) const	{ return (row*col == 1); }
//		int	isScalar (void) const	{ return (row == 1 && row == col); }

		int	rows (void) const	{ return (row); }
		int	cols (void) const	{ return (col); }
		long    size (void) const	{ return (row*col); }
		double 	min_col(int) const;
		double 	max_col(int) const;
		Matrix*	exec (singlefun);
		void	mult  (Matrix&);
		void	tmult (Matrix&);
		Matrix*	transpose (void);
                int     selectassign (Liste<Matrix*>, Matrix*);
		Matrix*	selection (Liste<Matrix*>);
		void	out (void);
#ifdef SOCKETSERVER
		void	out (SocketIO *);
#endif
		void	fill (const double);
		void	fill (const Matrix &);
		void	dump (char *)  { this->out(); }
		void	mdump (char *, ostream& ) { this->out(); }

                void  sort          (Matrix &index);
                int inverse         (void);
                int LUsubstitution  (Matrix &index, Matrix &b);
                int LUdecomposition (Matrix &index, double &d);
                Matrix * jacobi (Matrix *evalues);
	friend	int  ReadMatrix (const char *, Matrix&);
	friend	void SortMatrix (Matrix &in, Matrix &order, long l, long r);
	friend	Matrix* ColumnBind (Matrix &, Matrix &);
	friend	Matrix* RowBind    (Matrix &, Matrix &);

  private:	int 	row, col, ScalarFlag;
		double	*elem;
		int	match (const Matrix&);
  };


double  *MatrixAlloc (int, int);
  
/* for debugging no inline
inline Matrix::~Matrix ()
{
#ifdef TTKDEBUG
extern int ttkdebug;
if (ttkdebug)
  cerr<<"Mat: destruct\t\t"<<this<<endl;
#endif
  delete[] elem;
}  */


#endif
