/* $Header: /home/xplore/cvs/xpl4gr/nnet.C,v 1.3 2002/08/16 12:31:47 sigbert Exp $ */

#include <stdio.h>
#include <math.h>
#include "x4math2.h"

#include "global.h"
#include "xplerror.h"

#include "nnet.hpp"

static long	Epoch;
static double	Decay;
static double	TotalError;

static long	Nunits;	
static long	Ninputs;
static long	FirstHidden;
static long	FirstOutput;
static long	Noutputs;
static long 	NSunits; /* number of sigmoid units */
static long	Nweights;
static long	Entropy;
static long	Linout;
static long      Softmax;

static double	Outputs[N];
static double	ErrorSums[N];
static double	Errors[N];
static long	Nconn[N];
static long	Conn[NWTS];
static double	wts[NWTS];
static double	Slopes[NWTS];
static double   Probs[N];

static long	NTrain;
static long	NTest;
static double	**TrainIn;
static double	**TrainOut;
static double	**TestIn;
static double	**TestOut;
static double	*Weights;

static double toutputs[N];

void set_net(long *n, long *nconn, long *conn, long *nsunits, long *entropy, 
             long *softmax, double *decay)
{  long i;
   Build_Net(n[0], n[1], n[2]);
   for (i = 0; i <= Nunits; i++) Nconn[i] = nconn[i];
   Nweights = Nconn[Nunits];
   for (i = 0; i < Nweights; i++) Conn[i] = conn[i];
   Epoch = 0;
   Decay = *decay;
   TotalError = 0.0;
   NSunits = *nsunits;
   Entropy = *entropy;
   Linout = (NSunits < Nunits);
   Softmax = *softmax;
}

void set_train(long *ntr, double *train, double *weights)
{
   long i,j;
   NTrain = *ntr;
   TrainIn = matrix(NTrain,Ninputs);
   TrainOut = matrix(NTrain,Noutputs);
   Weights = vect(NTrain);
   for(j=0; j < Ninputs; j++)
      for(i=0; i < NTrain; i++) TrainIn[i][j] = *train++; 
/*   printf ("0,0 %g\n", TrainIn[0][0]);
   printf ("0,1 %g\n", TrainIn[0][1]);
   printf ("1,0 %g\n", TrainIn[1][0]);
   printf ("1,1 %g\n", TrainIn[1][1]);
   fflush(stdout);
*/   for(j=0; j < Noutputs; j++)
      for(i=0; i< NTrain; i++) 
      TrainOut[i][j] = *train++;
   for(i=0; i < NTrain; i++) Weights[i] = *weights++;
}

void nntest(long *ntest, double *test, double *result, double *inwts)
{
   long i,j;
    for (i=0; i< Nweights; i++) wts[i] = inwts[i];
    NTest = *ntest;
    if(Nweights==0) {//printf("No model set\n");
                     return;}
    TestIn = matrix(NTest,Ninputs);
    TestOut = matrix(NTest,Noutputs);
    for (i = 0; i < Ninputs; i++)
       for(j = 0; j < NTest; j++)
          TestIn[j][i] = *test++;

    for (i=0; i<Noutputs; i++) toutputs[i] = 0.5;
    for(j=0; j<NTest; j++){
      fpass(TestIn[j], toutputs, 1.0);
      if (Softmax)
        for (i=0; i<Noutputs; i++) TestOut[j][i] = Probs[FirstOutput+i];
      else
        for (i=0; i<Noutputs; i++) TestOut[j][i] = Outputs[FirstOutput+i];
      }
    for (i=0; i<Noutputs; i++)
       for(j=0; j<NTest; j++)
          *result++ = TestOut[j][i];
    free_matrix(TestIn,NTest,Ninputs);
    free_matrix(TestOut,NTest,Noutputs);
}


void Build_Net (long ninputs, long nhidden, long noutputs)
{
  long i;

  Nunits = 1 + ninputs + nhidden + noutputs;
  if (Nunits > N)
  {  errmsg("Too many units");
     return;
  }
  Ninputs  = ninputs;
  FirstHidden = 1 + ninputs;
  FirstOutput = 1 + ninputs + nhidden;
  Noutputs = noutputs;

  for (i=0; i<=Nunits; i++)    {
     Outputs[i] = 0.0;
     Nconn[i] = 0;
     }
  Nconn[Nunits+1] = 0;
  Outputs[0] = 1.0;
}



static double sigmoid(double sum)
{
    if (sum < -15.0) 
      return(0.0);
    else if (sum > 15.0) 
      return(1.0);
    else 
      return (1.0 /(1.0 + x4exp(-sum)));
}

static double E(double y, double t)
{
  double dif, sum = 0;
  if(Entropy) {
  if (t > 0) sum -= t * x4log((y+EPS)/t);
  if (t < 1) sum -= (1-t) * x4log((1-y+EPS)/(1-t));
  } else {
     dif = y - t;
     sum = dif*dif;
  }
  return(sum);
}


static void fpass (double *input, double *goal, double wx)
{
  long i, j;
  double sum, t;
  
  for (i = 0; i < Ninputs; i++) Outputs[i+1] = input[i];

  for (j = FirstHidden; j < Nunits; j++)
    {
      sum = 0.0;
      for (i = Nconn[j]; i < Nconn[j+1]; i++)
	sum += Outputs[Conn[i]] * wts[i];
	if(j < NSunits) sum = sigmoid(sum);
        Outputs[j] = sum;
    }
  if(Softmax) {
    sum = 0.0;
    /* avoid overflows by re-normalizing */
    t = Outputs[FirstOutput];
    for (i = FirstOutput+1; i < Nunits; i++) 
       if(Outputs[i] > t)  t = Outputs[i];
    for (i = FirstOutput; i < Nunits; i++){
      Probs[i] = x4exp(Outputs[i]-t);
      sum += Probs[i];
      }
    for (i = FirstOutput; i < Nunits; i++){
      Probs[i] = Probs[i]/sum;
      t = goal[i-FirstOutput];
      if(t > 0) TotalError += t * x4log(t/Probs[i]);
      }
  } else
    for (i = FirstOutput; i < Nunits; i++)
      TotalError += wx * E(Outputs[i], goal[i-FirstOutput]);
}

static double sigmoid_prime(double value)
{
    return (value * (1.0 - value));
}

static double sigmoid_prime_prime(double  value)
{
	return (value * (1.0 - value) * (1.0-2.0*value));
}

static void bpass (double *goal, double wx)
{
  long i, j, cix; 
  double sum;

  if (Softmax) {
    sum = 0.0;
    for (i = FirstOutput; i < Nunits; i++)
      sum += goal[i-FirstOutput];
    for (i = FirstOutput; i < Nunits; i++) 
      ErrorSums[i] = sum * Probs[i] - goal[i-FirstOutput];
  } else if (Entropy)
    for (i = FirstOutput; i < Nunits; i++) 
      ErrorSums[i] = Outputs[i] - goal[i-FirstOutput];
  else
    for (i = FirstOutput; i < Nunits; i++) { 
      ErrorSums[i] = 2*(Outputs[i] - goal[i-FirstOutput]);
      if (i < NSunits) ErrorSums[i] *= sigmoid_prime(Outputs[i]);
      }
  for (i = FirstHidden; i < FirstOutput; i++)
    ErrorSums[i] = 0.0;

  for (j = Nunits-1; j >= FirstHidden; j--)
    {
      Errors[j] = ErrorSums[j];
      if(j < FirstOutput) Errors[j] *= sigmoid_prime(Outputs[j]); 
      for (i = Nconn[j]; i < Nconn[j+1]; i++)
	{
	  cix = Conn[i];
	  ErrorSums[cix] += Errors[j] * wts[i];
	  Slopes[i] += wx * Errors[j] * Outputs[cix];
	}
    }
}

void dfunc(double *p, double *df, double *fp)
{
  long i,j;
  double sum1;
  for(i = 0; i < Nweights; i++)
      wts[i] = p[i];
  for (j = 0; j < Nweights; j++) Slopes[j] =  2 * Decay * wts[j] ;
  TotalError = 0.0;
  for (i = 0; i < NTrain; i++)
    {
      fpass (TrainIn[i], TrainOut[i], Weights[i]);
      bpass (TrainOut[i], Weights[i]);
    }
  sum1 = 0.0;
  for (i = 1; i <= Nweights; i++) 
     sum1 += p[i]*p[i];
  *fp = TotalError + Decay * sum1;
  for (j = 0; j < Nweights; j++) df[j] = Slopes[j];
  Epoch++;
}

static double fminfn(double  *p)
{
	long  i;
	double  sum1;
	for (i = 0; i < Nweights; i++)
		wts[i] = p[i];
	TotalError = 0.0;
	for (i = 0; i < NTrain; i++)
	{ fpass(TrainIn[i], TrainOut[i], Weights[i]);
/*          printf ("%i %f\n", i , TotalError); */
        }
	sum1 = 0.0;
	for (i = 0; i < Nweights; i++)
		sum1 += p[i] * p[i];
	Epoch++;
	return (TotalError + Decay * sum1);
}

static void fmingr(double *p, double *df)
{
	long  i, j;
	for (i = 0; i < Nweights; i++)
		wts[i] = p[i];
	for (j = 0; j < Nweights; j++)
		Slopes[j] = 2 * Decay * wts[j];
	TotalError = 0.0;
	for (i = 0; i < NTrain; i++) {
		fpass(TrainIn[i], TrainOut[i], Weights[i]);
		bpass(TrainOut[i], Weights[i]);
	}
	for (j = 0; j < Nweights; j++)
		df[j] = Slopes[j];
	Epoch++;
}


static double *vect(long n)
{
	double *v;
	v = (double *) Calloc(n, double);
        long i;
        for (i=0; i<n; i++)
          v[i] = 0.0;
/*	if (!v) 
        { errmsg("allocation failure in vect()");
          return NULL;
*/        
	return v;
}
 
static void free_vect(double *v)
{
	Free(v);
}

static double **matrix(long nrh, long nch)
{
        long i, j;
        double **m;
 
        m=(double **) Calloc((nrh+1), double*);
        for (i=0; i<nrh+1; i++)
          m[i] = NULL;
//        if (!m) errmsg("allocation failure 1 in matrix()");
 
        for(i=0;i<=nrh;i++) {
                m[i]=(double *) Calloc((nch+1), double);
                for (j=0; j<nch+1; j++) m[i][j] = 0.0;
//                if (!m[i]) errmsg("allocation failure 2 in matrix()");
        }
        return m;
}
static void free_matrix(double **m, long nrh, long nch)
{
        long i;
 
        for(i=nrh;i>=0;i--) Free(m[i]);
        Free(m);
}
static double **Lmatrix(long n)
{
	long i, j;
	double **m;

	m=(double **) Calloc(n, double*);
        for (i=0; i<n; i++) m[i] = NULL;
//	if (!m) errmsg("fail1 in Lmatrix()");
 
	for(i=0;i<n;i++) {
		m[i]=(double *) Calloc((i+1), double);
                for (j=0; j<i+1; j++) m[i][j] = 0.0;
//		if (!m[i]) errmsg("fail2 in Lmatrix()");
	}
	return m;
}

static void free_Lmatrix(double **m, long n)
{
        long i;
 
        for(i=n-1;i>=0;i--) Free(m[i]);
        Free(m);
}


void dovm(long *Nweights, double *wts, double *Fmin, long *maxit, long *trace)
{
	vmmin(*Nweights, wts, Fmin, *maxit, *trace);
}

/*  BFGS variable-metric method, based on Pascal code
in J.C. Nash, `Compact Numerical Methods for Computers', 2nd edition,
converted by p2c then re-crafted by B.D. Ripley */

static void vmmin(long n, double *b, double *Fmin, long maxit, long trace) 
/* n      dimension */
/* *b     input vect */
/* *Fmin  minimum value */
/* maxit  maximum number of BFGS updates */
/* trace  trace switch */
{
  booleanr accpoint, enough;
  double  *g, *t, *X, *c, **B;
  long count, funcount, gradcount;
  double f, gradproj;
  long i, j, ilast, iter=0;
  double s, steplength;
  double D1, D2;

  g = vect(n); t = vect(n); X = vect(n); c = vect(n);
  B = Lmatrix(n);
  f = fminfn(b);
  if(trace) {// printf("initial  value %f \n",f); 
                fflush(stdout);}
  {
    *Fmin = f;
    funcount = gradcount = 1;
    fmingr(b, g); iter++;
    ilast = gradcount;

    do {
      if (ilast == gradcount) {
	for (i = 0; i < n; i++) {
	  for (j = 0; j < i; j++)
	    B[i][j] = 0.0;
	  B[i][i] = 1.0;
	}
     }

      for (i = 0; i < n; i++) {
	X[i] = b[i];
	c[i] = g[i];
      }
      gradproj = 0.0;
      for (i = 0; i < n; i++) {
	s = 0.0;
	for (j = 0; j <= i; j++)
	  s -= B[i][j] * g[j];
	for (j = i+1; j < n; j++)
	  s -= B[j][i] * g[j];
        t[i] = s;
        gradproj += s * g[i];
        }

      if (gradproj < 0.0) { /* search direction is downhill */
	steplength = 1.0;
	accpoint = false;
	do {
	  count = 0;
	  for (i = 0; i < n; i++) {
	    b[i] = X[i] + steplength * t[i];
	    if (reltest + X[i] == reltest + b[i]) /* no change */
	      count++;
	  }
	  if (count < n) {
	    f = fminfn(b);
	    funcount++;
	    accpoint = (f <= *Fmin + gradproj * steplength * acctol);

	    if (!accpoint) {
	      steplength *= stepredn;
	    }
	  }
	} while (!(count == n || accpoint));
	enough = (f > abstol) && (f < (1.0-reltol)*(*Fmin));
	/* stop if value if small or if relative change is low */
        if(!enough) count = n;
	if (count < n) { /* making progress */
	  *Fmin = f;
	  fmingr(b, g);
	  gradcount++; iter++;
	  D1 = 0.0;
	  for (i = 0; i < n; i++) {
	    t[i] = steplength * t[i];
	    c[i] = g[i] - c[i];
	    D1 += t[i] * c[i];
	  }
	  if (D1 > 0) {
	    D2 = 0.0;
	    for (i = 0; i < n; i++) {
	      s = 0.0;
	      for (j = 0; j <= i; j++)
		s += B[i][j] * c[j];
	      for (j = i+1; j < n; j++)
		s += B[j][i] * c[j];
	      X[i] = s;
	      D2 += s * c[i];
	    }
	    D2 = 1.0 + D2 / D1;
	    for (i = 0; i < n; i++) {
	      for (j = 0; j <= i; j++)
		B[i][j] += (D2 * t[i] * t[j] - X[i] * t[j] - t[i] * X[j]) / D1;
	    }
	  } else { /* D1 < 0 */
	    ilast = gradcount;
	  }
	} else {  /* no progress */
	  if (ilast < gradcount) {
	    count = 0;
	    ilast = gradcount;
	  }
	}
      } else { /* uphill search */
	count = 0;
	if (ilast == gradcount)
	  count = n;
	else
	  ilast = gradcount;
	/* Resets unless has just been reset */
      }
      if (iter%REPORT == 0 && trace) {
      	// printf("iter%4d value %f\n",iter,f);
      	// fflush(stdout);
      }
      if (iter >= maxit) break;
      if (gradcount - ilast > 2*n) ilast = gradcount; /* periodic restart */
    } while (count != n || ilast != gradcount);
  }
  if(trace) {
    /* printf("final  value %f \n", *Fmin);
     if(iter < maxit) 
  	printf("converged\n");
     else 
  	printf("stopped after %i iterations\n");
	*/
  }
  free_vect(g); free_vect(t); free_vect(X); free_vect(c);
  free_Lmatrix(B,n);
}


double **H, *h, **w;

static void pHessian(double *input, double *goal, double wx)
{
	long i, in1, in2, out1,out2, j, j1,j2, first1, first2;
	double out, s, sum1, sum2, t, tmp;

	fpass(input, goal, 1.0); bpass(goal, 1.0);
   if(Softmax) {
        for (i = 0; i < Nunits; i++){
           sum1 = 0.0;
           for (j = FirstOutput; j < Nunits; j++) sum1 += w[i][j] * Probs[j];
           h[i] = sum1;
        }
	for (in1 = 0; in1 < Nunits; in1++)
	   for (j1 = Nconn[in1]; j1 < Nconn[in1+1]; j1++){
	      out1 = Conn[j1]; first1 = (in1 < FirstOutput);
	      for (in2 = 0; in2 < Nunits; in2++)
	         for (j2 = Nconn[in2]; j2 < Nconn[in2+1]; j2++)
	            if (j2 <= j1) {
	            out2 = Conn[j2]; first2 = (in2 < FirstOutput);
	            if ((!first1) && (!first2)){
	            /* both hidden -> output */
		       tmp = - Probs[in1] * Probs[in2];
	               if(in1 == in2) tmp += Probs[in1];
	                  H[j1][j2] += (tmp * Outputs[out1] * Outputs[out2]);
	            } else if (first1 && first2){
	                sum1 = sum2 = 0.0;
	                for(i = FirstOutput; i < Nunits; i++){
	                   sum1 += Errors[i] * w[in1][i];
			   sum2 += w[in1][i] * w[in2][i] * Probs[i];
	                }
                        sum2 -= h[in1] * h[in2];
	                s = sigmoid_prime(Outputs[in1]) * sigmoid_prime(Outputs[in2])
	                    * sum2;
	                if(in1 == in2)
	                    s += sigmoid_prime_prime(Outputs[in1]) * sum1;
	                H[j1][j2] += (s * Outputs[out1] * Outputs[out2]);
	            } else {
	            /* one input -> hidden, one hidden -> output */
	               if(in1 < in2){
                          sum1 = w[in1][in2] - h[in1];
	                  H[j1][j2] += (Outputs[out1] * sigmoid_prime(Outputs[in1])
	                  * (Outputs[out2] * Probs[in2] * sum1
	                  +((in1==out2)?Errors[in2]:0)));
	               } else {
                          sum1 = w[in2][in1] - h[in2];
	                  H[j1][j2] += (Outputs[out2] * sigmoid_prime(Outputs[in2])
	                  * (Outputs[out1] * Probs[in1] * sum1
	                  +((in2==out1)?Errors[in1]:0)));
	               }
	            }
	        }
	   }
   } else {
	for (i = FirstOutput; i < Nunits; i++) {
	   out = Outputs[i];
	   s = sigmoid_prime(out);
	   t = goal[i - FirstOutput];
	if (Linout) h[i] = 2;
	else if (Entropy)
	   h[i] = out*(1 - out);
	else
	   h[i] = sigmoid_prime_prime(out) * 2 * (out - t) + 2 * s * s;
	}
	for (in1 = 0; in1 < Nunits; in1++)
	   for (j1 = Nconn[in1]; j1 < Nconn[in1+1]; j1++){
	      out1 = Conn[j1]; first1 = (in1 < FirstOutput);
	      for (in2 = 0; in2 < Nunits; in2++)
	         for (j2 = Nconn[in2]; j2 < Nconn[in2+1]; j2++)
	            if (j2 <= j1) {
	            out2 = Conn[j2]; first2 = (in2 < FirstOutput);
	            if ((!first1) && (!first2)){
	            /* both hidden -> output */
	               if(in1 == in2) 
	                  H[j1][j2] += wx*(h[in1] * Outputs[out1] * Outputs[out2]);
	            } else if (first1 && first2){
	                sum1 = sum2 = 0.0;
	                for(i = FirstOutput; i < Nunits; i++){
	                   sum1 += Errors[i] * w[in1][i];
	                   sum2 += w[in1][i] * w[in2][i] * h[i];
	                }
	                s = sigmoid_prime(Outputs[in1]) * sigmoid_prime(Outputs[in2])
	                    * sum2;
	                if(in1 == in2)
	                    s += sigmoid_prime_prime(Outputs[in1]) * sum1;
	                H[j1][j2] += wx*(s * Outputs[out1] * Outputs[out2]);
	            } else {
	            /* one input -> hidden, one hidden -> output */
	               if(in1 < in2){
	                  H[j1][j2] += wx*(Outputs[out1] * sigmoid_prime(Outputs[in1])
	                  *(Outputs[out2]*w[in1][in2]*h[in2]
	                  +((in1==out2)?Errors[in2]:0)));
	               } else {
	                  H[j1][j2] += wx*(Outputs[out2] * sigmoid_prime(Outputs[in2])
	                  *(Outputs[out1]*w[in2][in1]*h[in1]
	                  +((in2==out1)?Errors[in1]:0)));
	               }
	            }
	        }
	   }
   }
}

void Hessian(double *inwts, double *Hess)
{
	long i, j;
	for (i = 0; i< Nweights; i++) wts[i] = inwts[i];
	H = Lmatrix(Nweights);
	h = vect(Nunits);
	w = matrix(Nunits, Nunits);
	for (i = 0; i < Nweights; i++)
	   for ( j = 0; j <= i; j++) 
	      H[i][j] = 0.0;
	for (j = FirstOutput; j < Nunits; j++)
	   for (i = FirstHidden; i < FirstOutput; i++) 
		w[i][j] = 0.0;
	for (j = FirstOutput; j < Nunits; j++) 
	   for (i = Nconn[j]; i < Nconn[j+1]; i++) 
	   	w[Conn[i]][j] = wts[i];
	for (i = 0; i < NTrain; i++){
	   pHessian (TrainIn[i], TrainOut[i], Weights[i]);
	}
	for (i = 0; i < Nweights; i++)
	    H[i][i] += 2 * Decay;

	for (i = 0; i < Nweights; i++)
	   for (j = 0; j < Nweights; j++) *Hess++ = H[max(i,j)][min(i,j)];
	free_Lmatrix(H, Nweights); free_vect(h); free_matrix(w, Nunits, Nunits);

}

