/* $Header: /home/xplore/cvs/xpl4gr/object.h,v 1.4 2003/02/23 17:01:21 cizek Exp $ */

/*
  Header fuer XplObject-Definition fuer XploRe 4

  TTK  2.2.95

  last modification:  21.5.96  TTK  xpltime

  XplObjects are organized as list of pointer to components. Each
  component can be itself a pointer to an XplObject.

  Example:

       XplObject    x.data
                 x.eig.vecs
                 x.eig.vals


  The XplObject x consists of two components: data and eig.  data has the
  type DATA (e.g., double). eig is another XplObject consisiting of further
  two XplObject, vecs and val.

*/

#ifndef _XplObject_H_
#define _XplObject_H_

// #pragma interface

// last modification:  23.2.95  TTK

#define TRUE  1
#define FALSE 0


// Use at the moment Matrix instead of XplReal

#ifdef MATRIX
#include "matrix.h"
typedef Matrix  XplDouble;
#else
#include "xreal.hpp"
#include "xint.hpp"
typedef XplReal<double> XplDouble;
#endif

#if !defined(__APPLE_CC__) && !defined (__BORLANDC__) && !defined (WIN32) 
#include <stream.h>
#else
#include <iostream.h>      // needed by xreal.hpp
#endif

#if defined (_XPLCDEF_) || defined (__BORLANDC__)
#include "xchar.hpp"    //XPLCHAR
#endif

#ifdef _XPLTIME_
#include "xtime.hpp"
#endif

#ifdef _XPLMASK_
#include "xmask.hpp"    //XPLMASK
#endif

#include "liste.h"
#include "xplpp.h"



// Substance holds the different pointers of a component

typedef  union { XplDouble      *xpldouble;
#ifdef _XPLCDEF_
		 XplChar        *xplchar;
#endif
#ifdef _XPLTIME_
		 XplTime        *xpltime;
#endif
#if defined (_GRAPHICS_) || defined (PSGEN) || defined (__BORLANDC__)
		 class XplDisplay     *display;
#endif
#ifdef _XPLMASK_
		 XplMask        *xplmask;
#endif
		 class Macro          *macro;
		 class XplObject      *xplobject;
		 char           *text;
	       } Substance;


// The types of a component

#define  C_EMPTY        0
#define  C_XPLDOUBLE    1
#define  C_XPLLONG      2
#define  C_XplObject    3
#define  C_PLOT         4
#define  C_DISPLAY      5
#define  C_MACRO        6
#define  C_XPLCHAR      7
#define  C_XPLMASK      8
#define  C_XPLTIME      9

#define  C_TEXT         C_XPLCHAR

// The types of an XplObject

#define  O_EMPTY        0x00000000    // no component
#define  O_XPLDOUBLE    1 << 0    // consists only of an XplDouble
#define  O_XPLLONG      1 << 1    // consists only of an XplLong
#define  O_XPLCHAR      1 << 2    // consists only of an XplChar
#define  O_PLOT         1 << 3    // consists only of a plot
#define  O_DISPLAY      1 << 4    // consists only of a display
#define  O_MACRO        1 << 5    // consists only of a macro
#define  O_XPLMASK      1 << 6    // consists only of an XplMask
#define  O_XPLTIME      1 << 7    // consists only of an XplTime

#define  O_TEXT         O_XPLCHAR

#define  O_SIMPLE    (O_XPLDOUBLE|O_XPLLONG|O_XPLCHAR|O_PLOT|O_MACRO|O_XPLMASK|O_XPLTIME)
#define  O_COMPOSED  ~O_SIMPLE   // consists of more than a single component

/* An XploRe variable can consist of several components, which could
   be themselves constructed of variables. A component holds a single
   data structure (XplObject, plot, macro, xpldouble,
   ...). Essentially an XplObject owns a list of components.  Since an
   XploRe variable can be a tree with arbitrary branches and leaves.


   Class Component:

   - Component (const char *name, int typ, Substance content);
     creates a component with the given content.
     allocates memory for
      name

     NOT for
      content

   - Component (XplDouble* x, const char *n = "tmpc");
     creates a component with the given content.xpldouble=x. Type is C_XPLDOUBLE.
     allocates memory for
      name

     NOT for
      content.xpldouble

   - Component (XplObject*);
     creates a component with the given content.XplObject=x. Type is C_XplObject.
     allocates memory for
      name

     NOT for
      content.XplObject

   - Component (Component* c);
     creates a component that is a DEEP copy of c.
     allocates memory for
      name, content

     NOT for
      -

   - ~Component ();
     destroyes component
     frees memory
      name
      content.<type>

   - XplObject * TransferXplObject (XplObject *o);
     if content != C_XplObject  return NULL
     otherwise if content.XplObject == o, o is returned
     otherwise content.XplObject->TransferXplObject is invoked

   - const char* Name (void);
     gives name of component

   - long Type (void);
     gives type of component

   - Substance Content ();
     gives content of component

    - void Print (const char *);
      prints component with delivered name (of object)



    Class XplObject
    each XplObject owns an unique identification number, which is generated
    by construction.

    - XplObject (Component* c, const char *n = "tmp");
      creates XplObject with the single component c.
      allocates memory for
      name

    - XplObject (Liste<Component*> el, const char *n = "tmp");
      creates XplObject with component list el. el is copied, but not the conent!
      allocates memory for
      name, elements (implicit ->liste.h)

    - XplObject (XplDouble *x, const char *n = "tmp");
      creates XplObject with a NEW single component c consisting of x.
      allocates memory for
      name, elements[0]

    - XplObject (XplObject *o);
      creates an XplObject that is a DEEP copy of *o.
      allocates memory for
      name, elements

    - XplObject (XplObject *o, const char *n);
      creates an XplObject that is a DEEP copy of *o with name n.
      allocates memory for
      name, elements

    - ~XplObject ();
      destroyes an XplObject, frees all pointers in elements, invokes elements.remove().
      frees memory
       name, elements[i], elements

    - Component* operator[] (long i);
      delivers component on position i, and NULL, if not found.

    - XplObject& operator= (const XplObject& o);
      generates a DEEP copy of o
      allocates memory for
       name, elements

    - const char* Name (void);
    - void SetName (const char *);
      gets and sets name of component

    - const long Type (void);     // necessary ??
    - void SetType (const long);
      gets and sets type of component

    - SetDefaultXplDouble ();
      derives the default XplDouble for the XplObject

    - const long Id (void);
      delivers id number of XplObject

    - XplDouble* DefaultXplDouble ();
      delivers default XplDouble, and NULL, if not set.

    - Component* DefaultComponent ();
      delivers default XplDouble, and NULL, if not set.

    - XplObject * TransferXplObject (XplObject *o);
      gets XplObject o out of elements of XplObject (this!, recursive)
      delivers NULL is o is not contained in this
      XplObject o is removed from list of elements, NULL is inserted

    - long ComponentNumber (void) const;
      delivers number of component

    - Component* Component (const char *n);
      delivers component with name n, and NULL, if not found.

    - DeleteComponents (void);
      deletes all components
      frees all elements

    - DeleteComponent (long i);
      deletes component on position i, ignores i<0 or i>=elements.laenge().
      frees elements[i]

    - InsertComponent (long i, Component *c);
      inserts component c on position i, if i>=0 and i<=elements.laenge().
 
    - AppendComponent (Component *c);
      appends a component on the list using InsertComponent
 
    - CopyComponent (Component *c, int New = TRUE, long pos = 0);
      copies component c. If New is TRUE, a deep copy is generated, otherwise
      c itself is copied. If pos=0 the new component is appended.
 
    - void Print (void);
     prints XplObject to cout



      Class Variable
      A variable contains two XplObjects. The root XplObject and the use XplObject:

        x.y   x is origObj, y is useObj

      The default XplDouble component of x is hold in copiedXpl (DEEP copy)
      The flag simpleVar is TRUE, if the origObj is simple, i.e., not compounded.
      parentXplObjectId holds the id of the parent XplObject, i.e., the source of
      origObj, that is a DEEP copy of this XplObject. Usually this source is not
      a temporary obejct. Therefore call by references are possible.


      - Variable (XplObject *o);
      creates a variable based on o. Usually needed for CONST
      allocates memory for
        origObj, copiedXpl

      - Variable (Variable *v);
      creates a DEEP copy of *v
      allocates memory for
        origObj, copiedXpl


      - ~Variable ()
      frees memory of origObj, if != NULL
      and copiedXpl

      - void SetSimpleVar (int sv)
      sets simpleVar flag to sv

      - int    SimpleVar() const
      delivers simpleVar flag

      - int    Constant () const
      delivers TRUE if Variable is constant (position < 0)

      - int    Name () const
      delivers name of


*      - Component * Component()
      delivers component

*      - void SetComponent (Component *c)
      sets component c for variable

*/

#ifdef REFERENCE
class Reference {
public:   Reference (Liste<TokenElement> *tl, Liste<long> *idl);
          ~Reference ()  {} ;
          Liste<TokenElement> * TokenList (void)  { return (tlist);  }
          Liste<long>         * IdList    (void)  { return (idlist); }
private:  Liste<TokenElement> *tlist;
          Liste<long>         *idlist;
};

class Dependence {
public:   Dependence (Liste<long> *idl);
          ~Dependence ()  {} ;
          Liste<long>         * IdList    (void)  { return (idlist); }
private:  Liste<long>         *idlist;
};

#endif

class Component {
 public:  
#if defined (_GRAPHICS_) || defined (PSGEN) || defined (__BORLANDC__)
          Component   (XplDisplay *, const char *n = "_tmpdisplay");
#endif
#ifdef _XPLTIME_
          Component   (XplTime    *, const char *n = "_tmpchar");
#endif
#ifdef _XPLCDEF_
          Component   (XplChar    *, const char *n = "_tmpchar");
#endif
#ifdef _XPLMASK_
          Component   (XplMask    *, const char *n = "_tmpmask");
#endif
          Component   (char       *, const char *n = "_tmpc");
          Component   (XplDouble  *, const char *n = "_tmpc");
          Component   (XplObject  *);
          Component   (Component  *);
          ~Component  ();
          XplObject *       TransferXplObject (XplObject *);
          void              Print   (const char *) const;
          const char *      Name    () const  { return (name == NULL ? "tmp" : name);    }
          const long        Type    () const  { return (type);    }
          const Substance   Content () const  { return (content); }
 private: long        type;
          char        *name;
          Substance   content;
};

class XplObject {
 public:  XplObject (Component  *, const char *n = NULL);  //"_tmp");
#if defined (_GRAPHICS_) || defined (PSGEN) || defined (__BORLANDC__)
          XplObject (XplDisplay *, const char *n = NULL);  //"_tmpdisplay");
#endif
#ifdef _XPLTIME_
          XplObject (XplTime    *, const char *n = NULL);  //"_tmp");
#endif
#ifdef _XPLCDEF_
          XplObject (XplChar    *, const char *n = NULL);  //"_tmp");
#endif
#ifdef _XPLMASK_
          XplObject (XplMask    *, const char *n = NULL);  // "_tmp");
#endif
          XplObject (const char *);
          XplObject (XplDouble  *, const char *n = NULL);
          XplObject (const XplObject *);
          XplObject (const XplObject  *, const char *);
          ~XplObject ();
          const char* Name (void)           const   { return (name == NULL ? "_tmp" : name);    }
          const long  Type (void)           const   { return (type);     }
          const long  Id   (void)           const   { return (id);       }
          const long  MotherId (void)       const   { return (motherid); }
          void        SetMotherId (long id)         { motherid = id;     }
          void        SetName  (const char *);
          void        SetType  (void);
          const int   WriteProtection ()     { return (writeprotection); }
          const int   isEmpty ()        const       { return (empty);    }
          const long  ComponentNumber (void) const
                        { return (elements.laenge()); }
#ifdef REFERENCE
          Reference * Reference ()      const       { return (ref);      }
          Dependence* Dependence ()     const       { return (dep);      }
#endif
          void        Print (const char *objname = NULL) const;
          XplDouble * DefaultXplDouble () const     { return (defx);     }
          Component * DefaultComponent () const     { return (defc);     }
          const long  DefaultType (void)  const     
                      { return (defc != NULL ? defc->Type() : C_EMPTY);  }
          Component * GetComponent (const char *n) const;
          Component * GetComponent (const char *n, long &) const;
          Component * operator[]   (long i);
          XplObject * ContainsXplObject (const long id);

          void        SetWriteProtection  ();
          void        DelWriteProtection  ();
          void        SetDefaultXplDouble ();
          void        SetDefaults ()           { SetDefaultXplDouble (); }
          XplObject * TransferXplObject  (XplObject *o);
          void        TransferComponents (XplObject *o);
          XplObject&  operator= (const XplObject&);
          void        DeleteElements (void)         { elements.remove ();
					              type  = O_EMPTY;
						      empty = TRUE;
					              defc  = NULL;
						      defx  = NULL; }
          void        CopyElements   (XplObject *o) { elements = o->elements;
					              SetType();
					              SetDefaultXplDouble ();}
          void        DeleteComponents (void);
          void        DeleteComponent  (long i);
          void        InsertComponent  (long i, Component*);
          void        AppendComponent  (Component *c)
                        { InsertComponent (elements.laenge(),c); }
          void        CopyComponent (Component *c, int New = TRUE, long pos = 0);

 private: char       *name;
          long       type, id, motherid;    // motherid is set if obj is copied
          int        empty, writeprotection;
          XplDouble  *defx;
          Component  *defc;
          Liste<Component*> elements;

          void init (void);

#ifdef REFERENCE
          Reference  *ref;
          Dependence *dep;
#endif
};


class Variable  {
 public:  Variable  (char *, long pos);
          Variable  (XplObject *);
          ~Variable ();
          XplObject *  GetXplObject ();
          void         SetSimpleVar (int sv)          { simpleVar = sv;     }
          void         SetUseXplObject (XplObject *o) { useObj = o;         }
          const long   Position ()        const       { return (position);  }
          const int    SimpleVar (void)   const       { return (simpleVar); }
          const char * Name ()            const       { return (name);      }
          const int    Constant ()        const   { return (position < 0L); }

 private: XplObject    *useObj;
          int          simpleVar;
          char         *name;
          long         position;
};


inline Component* XplObject::operator[] (long i)
{
  return ((0<=i) && (i<elements.laenge()) ? elements[i] : (Component *)NULL);
}

inline XplObject * Component::TransferXplObject (XplObject *o)
{
  if (type == C_XplObject)
    {
    if (o != content.xplobject)
      return (content.xplobject->TransferXplObject (o));
    else
      {
      content.xplobject = NULL;
      return (o);
      }
    }
  
  return (NULL);
}

int IsEqual (XplObject *, XplObject *);
// delivers TRUE if XplObjects have equal values

#endif  // of _XplObject_H_



