/* $Header: /home/xplore/cvs/xpl4gr/printing.cpp,v 1.3 2002/08/16 12:31:48 sigbert Exp $ */



//----------------------------------------------------------------------------

// ObjectWindows

// Copyright (c) 1991, 1995 by Borland International, All Rights Reserved

//  This application displays and prints a Ruler using OWL printer classes.

//----------------------------------------------------------------------------

#include "printing.h"

#include <stdio.h>



// tprintdialog:tdata contain all setting for the printing

// for graphic printing, page number (pnn) has to be always supplied!

// for text printing, number of lines can be used instead

TWindowPrintout::TWindowPrintout(const char* title, TWindow* window,

                                 int pnn, unsigned long lines,

                                 TRect *margins, TPoint *paper,

                                 unsigned tab, char *fontname, unsigned fontsize)

:

TPrintout(title)

{

     Window = window;

     Scale = true;



     // page count / number of lines

     pn = pnn;

     numlines = lines;



     // tab - we do not allow for extreme values

     if ((tab <= 0) || (32 <= tab))

          tabsize = 8;

     else tabsize = tab;



     // margins and paper size

     if (margins)

     	  printmargins = *margins;

     else printmargins.SetEmpty();



     if (paper)

          papersize = *paper;

     else

     {

          papersize.x = 0;

          papersize.y = 0;

          printmargins.SetEmpty();

     }



     // default setting if page setup was not yet done

     if (printmargins.IsNull() || (papersize.X() <= 0) || (papersize.Y() <= 0))

     {

          printmargins.Set(2540, 2540, 2540, 2540);

          papersize.x = 21000;

          papersize.y = 29700;

     }

     devicemargins.SetEmpty();



     // set up print font

     if (fontsize > 1)

          PrtFontSize = fontsize;

     else PrtFontSize = DEFAULTPRTSIZE;



     if (fontname)

     {

          if ((0 < strlen(fontname)) && (strlen(fontname) < PRTFONTNAMELEN))

               strcpy(PrtFontName, fontname);

          else 

          {

               strcpy(PrtFontName, DEFAULTPRTFONT);

               PrtFontSize = DEFAULTPRTSIZE;

          }

     }

     else 

     {

          strcpy(PrtFontName, DEFAULTPRTFONT);

          PrtFontSize = DEFAULTPRTSIZE;

     }

}



// external printing variables

extern bool PrintAsPicture;

extern char *printheader;

extern int *tabs;

// PC: replaced by pointer

// extern char printtext[32550];

extern char *printtext;

char *currentprinttext;

extern long textlen;

long currenttextlen;



bool TWindowPrintout::HasPage(int pageNumber)

{

// PC:     return (pageNumber < (pn+1));

     if (pn > 0)	// number of pages is known

          return (pageNumber < (pn + 1));

     else 		// number of pages not known

          return (numlines > 0);

}



#define LINE_MAL 1.25



void TWindowPrintout::PrintPage(int pnum, TRect& rect, unsigned)

{

  // Conditionally scale the DC to the window so the printout will

  // resemble the window

  //

     int    prevMode;

     TSize  oldVExt, oldWExt;

     TRect PageRect(0, 0, PageSize.X(), PageSize.Y());



  //if (Scale)

     // set mapping mode

     if (PrintAsPicture)

     {

          prevMode = DC->SetMapMode(MM_ISOTROPIC);

          TRect windowSize = Window->GetClientRect();

          DC->SetViewportExt(PageSize, &oldVExt);

          DC->SetWindowExt(windowSize.Size(), &oldWExt);

          DC->IntersectClipRect(windowSize);

          DC->DPtoLP(rect, 2);

     }

    else

     {

          prevMode = DC->SetMapMode(MM_TWIPS);

//          prevMode = DC->SetMapMode(MM_ISOTROPIC);

//          DC->SetViewportExt(PageSize, &oldVExt);

//          DC->SetWindowExt(PageSize, &oldWExt);

// done above         TRect PageRect(0, 0, PageSize.X(), PageSize.Y());

          DC->DPtoLP(PageRect, 2);

          DC->IntersectClipRect(PageRect);

          PageRect.Set(abs(PageRect.Left()), abs(PageRect.Top()), abs(PageRect.Right()), abs(PageRect.Bottom()));

// not needed          DC->DPtoLP(rect, 2);

     }



     // WAS WAR DAS FUER EIN??? - graph printing



     if (PrintAsPicture)

     {

          // Call the window to paint itself

          Window->Paint(*DC, false, rect);

     }

     else

     {

          int gc;

   	  char *curbeg;

          TFont *prtfont = NULL;



          if  (ActPage == 1)

          {

               currenttextlen = textlen;

               currentprinttext = printtext;

          }



          gc = currenttextlen;

   	  curbeg = currentprinttext;



   	  DC->SaveDC();



  	  if (gc == 0)

   	       numlines = 0; // cout<<"no text to get.."<<endl;

   	  else

          {

               TEXTMETRIC fontmetrics;



               // select font

               DC->SelectStockObject(ANSI_FIXED_FONT);



               prtfont = new TFont(PrtFontName, -20 * PrtFontSize);

               if (prtfont)

               {

                    prtfont->GetTextMetrics(fontmetrics);

                    if (fontmetrics.tmPitchAndFamily & TMPF_FIXED_PITCH)

                    {

                         delete prtfont;

                         prtfont = new TFont(DEFAULTPRTFONT, -20 * DEFAULTPRTSIZE);

                    }

               }



               if (prtfont)

                    DC->SelectObject(*prtfont);

               else DC->SelectStockObject(ANSI_FIXED_FONT);



               // get device margins

               devicemargins.Set(

                    (int) ( ((double) printmargins.Left()   / (double) papersize.X()) * (double) PageRect.Width() ),

                    (int) ( ((double) printmargins.Top()    / (double) papersize.Y()) * (double) PageRect.Height() ),

                    (int) ( ((double) printmargins.Right()  / (double) papersize.X()) * (double) PageRect.Width() ),

                    (int) ( ((double) printmargins.Bottom() / (double) papersize.Y()) * (double) PageRect.Height() )

                                );



               // step and offset for the text

               TSize ts = DC->GetTextExtent("M", 1);

               TPoint begp(devicemargins.Left(), -1 * devicemargins.Top());// initial offset

               TPoint offset(0, -1 * ts.Y() * LINE_MAL);                   // line offset



               // line len and page len

               long linelen = (PageRect.Width() - devicemargins.Left() - devicemargins.Right()) / abs(ts.X());

               if (linelen < 1)

                    linelen = 1;

               long pagelen = (PageRect.Height() - devicemargins.Top() - devicemargins.Bottom()) / abs(offset.Y());

               if (pagelen < 1)

                    pagelen = 1;



// PC: not needed

//               // init tabs if they are not supplied

//               int i;

//               if (tabs == NULL)

//   	       {	

//                    tabs = new int [10];

//                    if (tabs)

//		         for (i = 0; i < 10; i++)

//  		              tabs[i] = 8 * i;

//   	       }



               // print header

               if (printheader != NULL)

               {

#define HEADSTRLEN  (256L + 80L)     // filename + --- ... ---

                    char headi[HEADSTRLEN + 4];

                    

           	    sprintf(headi, " %s Page %d ", printheader, ActPage);

                    if (strlen(headi) + 1 < min(linelen,HEADSTRLEN))

                    {

                         long add = (min(linelen,HEADSTRLEN) - strlen(headi)) / 2;

                         char pom[HEADSTRLEN + 4];



                         memset(pom, 0, HEADSTRLEN);

                         memset(pom, '-', add);

                         strcat(headi, pom);

                         strcat(pom, headi);

                         strcpy(headi, pom);

                    }

                    

                    DC->TextOut(begp, headi, strlen(headi));

            	    begp += offset;

            	    begp += offset;

                    

                    pagelen -= 2;  // two lines are occupied by the header

                    if (pagelen < 1)

                         pagelen = 1;

               }

               

               int gelesen = 0;

               int linescount = 0, lineends = 0;

               unsigned linesize, bufsize;

#define PRINTBUFLEN 512

               char printbuf[PRINTBUFLEN + 8];



               while ((gelesen < gc) && (linescount < pagelen))

               {

                    // create the output line

                    linesize = 0;       bufsize = 0;

         	    while ( (gelesen + linesize < gc) && (bufsize < PRINTBUFLEN) &&

                            (curbeg[linesize] != '\r') && (curbeg[linesize] != '\n') &&

                            (bufsize < linelen)

                          )

                    {

                         if (curbeg[linesize] == '\t')

                         {

                              printbuf[bufsize++] = ' ';

                              while ( ((bufsize % tabsize) != 0) && (bufsize < PRINTBUFLEN) && (bufsize < linelen))

                                   printbuf[bufsize++] = ' ';

                         }

                         else if (curbeg[linesize] < ' ')

                              printbuf[bufsize++] = ' ';

                         else printbuf[bufsize++] = curbeg[linesize];

                         linesize++;

                    }

                    printbuf[bufsize] = '\0';



                    // print text

                    if (bufsize > 0)

            	         DC->TextOut(begp, printbuf);

//            	         DC->TabbedTextOut(begp, curbeg, linesize, 10, tabs, 0);



                    // skip the line end

                    while ((curbeg[linesize] == '\r') && (gelesen + linesize < gc))

                         linesize++;

                    if (curbeg[linesize] == '\n')

                    {

                         lineends++;

                         linesize++;

                    }

                    while ((curbeg[linesize] == '\r') && (gelesen + linesize < gc))

                         linesize++;

                    

                    // update cycle variables

                    begp += offset;

                    linescount++;

                    

                    curbeg += linesize;  

                    gelesen += linesize;

               }

               

               // update global vars

               currenttextlen -= gelesen;

               if (currenttextlen < 0)

                   currenttextlen = 0;

               currentprinttext += gelesen;



               numlines = (numlines > lineends) ? numlines - lineends : 0;



               ActPage++;

          }



          DC->RestoreDC();



          // free font

          if (prtfont)

               delete prtfont;

     }    // end of print as text



  // Restore changes made to the DC

     // if (Scale)

     if (PrintAsPicture)

     {

          DC->SetWindowExt(oldWExt);

          DC->SetViewportExt(oldVExt);

          DC->SetMapMode(prevMode);

     }

}



// Do not enable page range in the print dialog since only one page is

// available to be printed

//

void

TWindowPrintout::GetDialogInfo(int& minPage, int& maxPage,

                               int& selFromPage, int& selToPage)

{

     minPage = 0;

     maxPage = 0;

     selFromPage = selToPage = 0;

}



//----------------------------------------------------------------------------

// TRulerWin



class TRulerWin : public TWindow {

  public:

    TRulerWin(TWindow* parent = 0);

   ~TRulerWin();



    void Paint(TDC&, bool, TRect&);



    void CmFilePageSetup();

    void CmFilePrint();



  private:

    TPrinter* Printer;



  DECLARE_RESPONSE_TABLE(TRulerWin);

};



DEFINE_RESPONSE_TABLE1(TRulerWin, TWindow)

  EV_COMMAND(CM_FILEPAGESETUP, CmFilePageSetup),

  EV_COMMAND(CM_FILEPRINT, CmFilePrint),

END_RESPONSE_TABLE;



TRulerWin::TRulerWin(TWindow* parent)

:

  TWindow(parent)

{

  Printer = new TPrinter;

}



TRulerWin::~TRulerWin()

{

  delete Printer;

}



void

TRulerWin::CmFilePageSetup()    // Execute File:Page-setup command

{

  if (Printer)

    Printer->Setup(this);

}



void

TRulerWin::CmFilePrint()          // Execute File:Print command

{

  if (Printer) {

    TWindowPrintout printout("Document", this);

    // TWindowPrintout printout("Ruler Test", this);

    //

    // Using OWL's banding support under Win95 _when_ the EXE's marked 4.0

    // (under Options|Project|Linker|General) will cause the print to fail.

    //

    if(!TSystem::IsWin95())

      printout.SetBanding(true);

    Printer->Print(this, printout, true);

  }

}



// Paint window's contents on any dc, screen or printer or whatever...

//

void

TRulerWin::Paint(TDC& dc, bool , TRect&)

{

  const int UnitsPerInch = 100;                 // Display scale units per inch

  const int NumInches = 8;                      // Size of ruler in inches

  const int MarkFraction = 4;                   // Power of 2 to use for marks

  const int MarksPerInch = (1 << MarkFraction); // Number of markers for each inch

  const int LargeMarkerSize = UnitsPerInch/3;   // Size of large, labeled markers

  const int SmallMarkerSize = UnitsPerInch/8;   // Size of smallest markers

  const int MarkStep = (LargeMarkerSize-SmallMarkerSize) / MarkFraction;



  dc.SaveDC();

  dc.SetMapMode(MM_LOENGLISH);

  int x1 = 0; //0.50 * UnitsPerInch;

  int y1 = x1;

  int x2 = x1 + NumInches * UnitsPerInch;

  int y2 = y1 + 1*UnitsPerInch;

  dc.Rectangle(x1, -y1, x2, -y2);



  // Draw marks

  //

  y2 = y1 + SmallMarkerSize;

  for (int marksPerInch = MarksPerInch; marksPerInch > 0; marksPerInch /= 2) {

    for (int i = 0; i <= NumInches*marksPerInch - 1; i++) {

      int x = x1 + (i * UnitsPerInch) / marksPerInch;

      dc.MoveTo(x, -y1);

      dc.LineTo(x, -y2);

    }

    y2 += MarkStep;

  }



  // Label the inch marks

  //

  y2 = y1 + LargeMarkerSize; // - text height

  for (int i = 1;  i <= NumInches - 1; i++) {

    int x = x1 + i * UnitsPerInch;  // - ( text width + 20%)

    char s[3];

    itoa(i, s, 10);

    dc.TextOut(TPoint(x, -y2), s, strlen(s));

  }



  dc.RestoreDC();

}



/* PC:



//

// class TRulerApp

// ~~~~~ ~~~~~~~~~

class TRulerApp : public TApplication {

  public:

    TRulerApp()

    :

      TApplication()

    {

    }



    void InitMainWindow() {

      EnableCtl3d();

      TFrameWindow* frame = new TFrameWindow(0, "Ruler Printing Demonstration",

        new TRulerWin);

      frame->Attr.X = GetSystemMetrics(SM_CXSCREEN) / 8;

      frame->Attr.Y = GetSystemMetrics(SM_CYSCREEN) / 8;

      frame->Attr.H = frame->Attr.Y * 6;

      frame->Attr.W = frame->Attr.X * 6;

      frame->AssignMenu(IDM_DISPLAY);

      SetMainWindow(frame);

    }

};





int

OwlMain(int , char* [])

{

  return TRulerApp().Run();

}

*/



