/* $Header: /home/xplore/cvs/xpl4gr/qp.C,v 1.3 2002/08/16 12:31:48 sigbert Exp $ */

/*

Copyright (C) 1998 Dieter Baron <dillo@giga.or.at>
and Computer Supported Didactics Working Group,
University of Vienna


This file contains the Whinston-van der Panne algorithm 
for quadratic optimization with linear constraints.

This is a special version of this file for internal use in the 
XploRe
Project only.  The normal version, distributed under the GNU 
General
Public License can be obtained from the author.

License is hereby granted to use this program or a modified 
version of
this program in versions of XploRe or XploRe Java Server, 
provided these
versions are clearly marked as containing this program 
in the places where information about the program is given.


The code is not to be distributed beyond the XploRe Development 
team 
and a few beta testers. 
 
License to use this program expires on October 1, 1998.
After that date, if the license agreement is not renewed, all 
existing
copies of XploRe or XploRe Java Server containing this program 
or
a modified version of this program must be deleted; all source 
code
containing this program or a modified version of this program must
either be deleted or released under the GNU General Public License
version 2 or later, or the GNU Library General Public License, version
2 or later.

Source code containing this program or a modified version of this
program may also be released under the conditions above at any time
prior to expiry of this license.

This file is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.




#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <cfloat>
#include <cmath>

#include <string>

#include "lo-ieee.h"
#include "str-vec.h"

#include "defun.h"
#include "error.h"
#include "gripes.h"
#include "help.h"
#include "oct-map.h"
#include "ov.h"
#include "variables.h"
#include "oct-obj.h"
#include "utils.h"

*/

#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif

#ifndef ABS
#define ABS(x) (((x) < 0) ? (-x) : (x))
#endif



int pivot(XplDouble *A, int r, int c);


/*
DEFUN (pivot, args, ,
       "pivot (A, r, c): pivot A on (r, c)")
{
    octave_value retval;
    Matrix B;
    int nargin, err;

    nargin = args.length ();

    if (nargin != 3 || !args(0).is_defined ()) { // check types, handle 
complex
        print_usage("pivot");
        return retval;
    }

    // XXX: indices are zero origin (one origin in octave) 
    B = args(0).matrix_value();
    
    if ((err=pivot(B,
                   (int)args(1).double_value(),
                   (int)args(2).double_value())) < 0)
        retval = err;
    else {
        // XXX: does this copy? 
        retval = B;
    }
    
    return retval;
}

*/

/* pivot A in place with pivot A(r, c)
   return values:
         0: no error
        -1: pivot coordinates illegal
        -2: pivot is zero
*/

int
pivot(XplDouble *A, int r, int c)
{
    int nr, nc, i, j;
    double pivot;

    nr = (*A).dimension(0);
    nc = (*A).dimension(1);
    
    if (r < 0 || r >= nr || c < 0 || c >= nc)
        return -1;

    pivot = (*A)(r, c);
    if (pivot == 0)
        return -2;

    /* column -> column - (el on pivotrow/pivot) * pivot column */
    for (i=0; i<nr; i++)
        if (i != r)
            for (j=0; j<nc; j++)
                if (j != c)
                    (*A)(i, j) = (*A)(i, j) - ((*A)(r, j)/pivot)*(*A)(i, c);
    
    /* pivot row -> pivot row*pivot */
    for (i=0; i<nc; i++)
        (*A)(r, i) = (*A)(r, i)/pivot;

    /* pivot column -> -pivot column*pivot */
    for (i=0; i<nr; i++)
        (*A)(i, c) = -(*A)(i, c)/pivot;

    /* pivot -> 1/pivot */
    (*A)(r, c) = 1/pivot;

    return 0;
}




//DEFUN (qp, args, nargout,
//       "qp(C, p, A, b) Whinston's quadratic optimizer")

int qp (XplDouble *C, XplDouble *p, XplDouble *A, XplDouble *b, 
         XplDouble *x, XplDouble *t)
{
/*    octave_value_list retval;
    Matrix C, A;
    ColumnVector p, b, x;
    ColumnVector xpos;   //  tracking positions of x_i;
                         //  negative means in head column
*/
    XplDouble *T;
    long *xpos;

    int nargin, n, m, i, j, pos, k, l;
    double min, d;
/*    Matrix T;

    nargin = args.length();
      
    if (nargin != 4) {
        print_usage("qp");
        return retval;
    }
    // XXX: check argument types. 

    C = args(0).matrix_value();
    p = args(1).vector_value();
    A = args(2).matrix_value();
    b = args(3).vector_value();
*/
    /* check dimensions */
    n = p->dimension(0);
    m = b->dimension(0);
/*
    if ((C.dimension(0) != n || C.dimension(1) != n
         || A.dimension(0) != m || A.dimension(1) != n)) {
        print_usage("qp");
        return retval; 
    }
*/
    /* XXX: check that C is positive definite */

    pos = 1;
    for (i=0; i<n; i++) {
        if ((*p)(i) < 0) {
            pos = 0;
            break;
        }
    }

    /* initialize */
    xpos = new long[n];
    for (i=0; i<n; i++)
      xpos[i] = 0;
    T    = new XplDouble(n+m+1, n+m+1);
    for (i=0; i<n+m+1; i++)
      for (j=0; j<n+m+1; j++)
        (*T)(i,j) = 0;


    for (i=0; i<n; i++) {
        (*T)(0, 1+i) = (*p)(i);
        (*T)(1+i, 0) = (*p)(i);
        for (j=0; j<n; j++)
            (*T)(1+i,1+j) = -(*C)(i, j);
    }

    for (i=0; i<m; i++) {
        (*T)(0, 1+n+i) = -(*b)(i);
        (*T)(1+n+i, 0) = (*b)(i);
        for (j=0; j<n; j++) {
            (*T)(1+j, 1+n+i) = -(*A)(i, j);
            (*T)(1+n+i, 1+j) = (*A)(i, j);
        }
    }

    k = -1;

    while (!pos) {
        /* find pivot column */
        if (k == -1) {
            min = 0;
            for (i=1; i<=n; i++) {
                if ((*T)(i, 0) < min) {
                    k = i;
                    min = (*T)(i, 0);
                }
            }
        }
        if (k == -1) {
            printf("can't happen, no negative q_i\n");
            break;
        }
        
        /* find pivot row */
        min = 1.0e138; // DBL_MAX;
        l = -1;
        if ((*T)(k, k) != 0) {
            min = (*T)(k, 0)/(*T)(k, k);
            l = k;
        }
        for (i=1; i<=m; i++) {
            /* XXX: is this really p_{ik}, not r_{ik} */
            if ((*T)(i+n, k) > 0) {
                d = (*T)(n+i, 0)/(*T)(n+i, k);
                if (d < min) {
                    min = d;
                    l = n+i;
                }
            }
        }
        if (l == -1) {
            /* no minimum */
            return -1;
        }

        if (k == l) { /* 2. q_{kk} */
            pivot(T, k, k);
            xpos[k-1] = !xpos[k-1];
            k = -1;
        }
        else if ((*T)(l, l) < 0) { /* 3a. -r_{ll} < 0 */
            pivot(T , l, l);
        }
        else if ((*T)(l, l) == 0 && (*T)(k, k) < 0) { /* 3b. -r_{ll} = 0,
                                                       -q_{kk} < 0 */
            pivot(T, k, k);
            pivot(T, l, l);
            xpos[k-1] = !xpos[k-1];
            k = -1;
        }
        else if ((*T)(l, l) == 0 && (*T)(k, k) == 0) { /* 3c. -r_{ll} = 0,
                                                        -q_{kk} = 0 */
            pivot(T, l, k);
            pivot(T, k, l);
            for (i=0; i<n+m+1; i++) {
                d = (*T)(k, i);
                (*T)(k, i) = (*T)(l, i);
                (*T)(l, i) = d;
            }
            for (i=0; i<n+m+1; i++) {
                d = (*T)(i, k);
                (*T)(i, k) = (*T)(i, l);
                (*T)(i, l) = d;
            }
            xpos[k-1] = !xpos[k-1];
            k = -1;
        }           

        /* check if we found minimum */
        pos = 1;
        for (i=1; i<=n; i++) {
            if ((*T)(i, 0) < 0) {
                pos = 0;
                break;
            }
        }
    }

    for (i=0; i<n; i++) {
        if (xpos[i])
            (*x)(i) = (*T)(1+i, 0);
        else
            (*x)(i) = 0;
    }

    (*t)(0) = (*T)(0, 0)*-.5;

    return 0;
}

