/* $Header: /home/xplore/cvs/xpl4gr/socketio.C,v 1.3 2002/08/16 12:31:50 sigbert Exp $ */

// Socket basic I/O class implementation  TTK, 19.7.98
// last modified 8.8.98
// transparent data exchange between big/little endian
// SK 020107 insert readBytes, modify readSize, readData 

#ifdef __MINGW32__
#include "windows.h"
#endif

#include <iostream.h>
#include <string>
#ifdef X4singleserver
#include <unistd.h>
#include <stdio.h>
#endif
#include "socketio.h"

#define CLIENT_VERSION  "XCversion 2.0"
#define SERVER_VERSION  "XSversion 2.0"

SocketIO::SocketIO()
{ socket = 0;
  status = OK;
  server = 0;

  // connection is always established since middleware starts process
  SizeOfDouble = sizeof (double);
  SizeOfFloat  = sizeof (float);
  SizeOfLong   = sizeof (long);
  SizeOfInt    = sizeof (int);
  SizeOfShort  = sizeof (short);

  ownEndian    = getEndian ();
  remoteEndian = BIG_ENDIAN;
}

void logt(const char *txt);
void logc(long cmd);
void logl(long len);
void logp (char *ptr, long size);

// SK 020107

long SocketIO::readBytes (void *buf, long len) {
  long i, j, rlen;
  int last200;
  unsigned char *vbuf = new (unsigned char)[len];
  unsigned char *ibuf = (unsigned char *) buf;

  if (ibuf==NULL) return (-1L);

  logt ("Readbytes");  
  j = last200 = 0;
  do {
#ifdef X4singleserver
     logt ("start reading");
     logl (len-j);
     rlen = read (fileno(stdin), vbuf, len-j);
     logc (rlen);
     logp ((char *) vbuf, rlen);
     //     cerr << rlen << endl;
#else
     rlen = recv (socket, vbuf, len-j, 0);
#endif
     i = 0;
     if (last200) {
       switch (vbuf[i]) {
       case 200: ibuf[j++] = 200; break;
       case 210: ibuf[j++] = 10; break;
       case 213: ibuf[j++] = 13; break;
       case 226: ibuf[j++] = 26; break;
       }
       i++;
     }
     for (; i<rlen-1; i++) {
       if (vbuf[i]==200) {
	 i++;
	 switch (vbuf[i]) {
	 case 200: ibuf[j++] = 200; break;
	 case 210: ibuf[j++] = 10; break;
	 case 213: ibuf[j++] = 13; break;
	 case 226: ibuf[j++] = 26; break;
	 }
       }
       else
	 ibuf[j++] = vbuf[i];
     }
     if (i==rlen-1) {
       last200 = (vbuf[rlen-1]==200);
       if (!last200) ibuf[j++] = vbuf[rlen-1];
     }
#ifndef __MINGW32__
     if (rlen == -1) {
       delete[] vbuf;
       return (-1L);
     }
#else
     if (rlen == -1) 
     { rlen=0;
     // logt("sleep");
       Sleep(1);
     }
#endif
  } while (j<len);
  logp ((char *) ibuf, len);
  delete[] vbuf;
  return (j);
}

void SocketIO::writeToken (long l)
{
  writeSize (l);
}

long SocketIO::readToken ()
{
  return (readSize());
}

int SocketIO::writeSize (long l)
// writes size as long (4 Bytes)
{
  if (ownEndian != remoteEndian)
    swapEndian (&l, 1);

  return (writeData (&l, SizeOfLong));
}

// SK 020107 
long SocketIO::readSize ()
// reads size as long (4 Bytes)
{
  long  l;
  long  rlen, tlen = SizeOfLong;

  //#ifdef WIN32
  char *buf;

  buf = (char *) &l;
  /*
#else
  void  *buf;

  buf = (void *) &l;
#endif
  */
  logt("SocketIO::readSize");

  rlen = readBytes (buf, tlen);

  if (rlen == -1) {
    //       logt("TRANSMISSION_ERROR");
    status = TRANSMISSION_ERROR;
    return (0L);
  }

  if (ownEndian != remoteEndian)
    swapEndian (&l, 1L);
  
  logp ((char *) &l, SizeOfLong);

  return (l);
}

#define MAXBUFLEN 8192
char *cbuf = NULL;

int SocketIO::writeData (const void *v, long size)
// writes byte array to socket, size is amount of bytes
{
  if (cbuf==NULL)
  { cbuf = new char[MAXBUFLEN];
    if (cbuf==NULL) 
    { status = NOT_ENOUGH_MEMORY;
      logt ("NOT_ENOUGH_MEMORY");
      return (status);
    }
  }

  if (size <= 0)
    return (status);

  if (status != OK)
    return (status);

  char *sbuf;
  const char *vbuf = (const char *) v;
  int i, j;
  long wlen, tlen;
  
  for (j=i=0; i<size; i++)
  { 
#ifdef X4singleserver
    if (vbuf[i]==200)
    { //logt ("200!");
      cbuf[j++] = 200;
      cbuf[j]   = 200;
    }
    else if (vbuf[i]==-56)
    { //logt ("-56!");
      cbuf[j++] = 200;
      cbuf[j]   = 200;
    }
    else if (vbuf[i]==26)
    { //logt(" 26!");
      cbuf[j++] = 200;
      cbuf[j]   = 226;
    }
    else if (vbuf[i]==10)
    { //logt(" 26!");
      cbuf[j++] = 200;
      cbuf[j]   = 210;
    }
    else if (vbuf[i]==13)
    { //logt(" 26!");
      cbuf[j++] = 200;
      cbuf[j]   = 213;
    }
    else
      cbuf[j]   = vbuf[i];
#endif
    j++;
    
    if (j>MAXBUFLEN-2)  
    { tlen = j;
      sbuf = cbuf;
      logt ("IN LOOP");
      logp (sbuf, tlen);
      do {
#ifdef X4singleserver
        logt ("Send bytes"); logl(tlen);
        wlen = write(fileno(stdout), sbuf, tlen);
        logt ("Sent bytes"); logl(wlen);
        fflush(stdout);
        logt ("Flush buffer");
#else
        wlen = send (socket, sbuf, tlen, 0);
#endif
        if (wlen == -1)
        { status = TRANSMISSION_ERROR;
          logt ("TRANSMISSION_ERROR");
          return (status);
        }
        tlen -= wlen;
        sbuf += wlen;
      } 
      while (tlen > 0);
      j = 0;
    }
  }

  if (j)
  { tlen = j;
    sbuf = cbuf;
    logt ("AFTER LOOP");
    logp (cbuf, tlen);
    do {
#ifdef X4singleserver
      logt ("Send bytes"); logl(tlen);
      wlen = write(fileno(stdout), sbuf, tlen);
      logt ("Sent bytes"); logl(wlen);
      fflush(stdout);
      logt ("Flush buffer");
#else
      wlen = send (socket, sbuf, tlen, 0);
#endif
      if (wlen == -1)
      { status = TRANSMISSION_ERROR;
        logt ("TRANSMISSION_ERROR");
        return (status);
      }
      tlen -= wlen;
      sbuf += wlen;
    } while (tlen > 0);
  } 
  return (status);
}

void * SocketIO::readData (long size)
// reads byte array of length size from socket
{
  char *vec;
  long rlen, tlen = size;

  if (size <= 0)
    status = NEGATIVE_LENGTH;

  vec = new char[size];
  if (vec == NULL)
    status = NOT_ENOUGH_MEMORY;

  if (status != OK)
    return (NULL);

  rlen = readBytes (vec, tlen);

  if (rlen == -1) {
    status = TRANSMISSION_ERROR;
    return (NULL);
  }
  
  logt("readData");
  logp(vec, size);

  return ((void *) vec);
}

void SocketIO::swapEndian (void *x, long len, int size)
// swaps the byte order for data type of size size
{
  long i;
  int  j, n;
  char *db, *dl, h;

  db = (char *) x;           // Startposition
  n  = size / 2;

  for (i=0; i<len; i++)
    {
    dl = db + size - 1;      // Endposition
    for (j=0; j<n; j++, dl--, db++)
      {
      h   = *dl;
      *dl = *db;
      *db = h;
      }
    db += n;                 // naechste Position
    }
}

void SocketIO::swapEndian (float *x, long len)
{
  swapEndian ((void *) x, len, SizeOfFloat);
}

void SocketIO::swapEndian (short *x, long len)
{
  swapEndian ((void *) x, len, SizeOfShort);
}

void SocketIO::swapEndian (int *x, long len)
{
  swapEndian ((void *) x, len, SizeOfInt);
}


void SocketIO::swapEndian (double *x, long len)
// swaps the byte order for type double
{
  swapEndian ((void *) x, len, SizeOfDouble);
}

void SocketIO::swapEndian (long *x, long len)
// swaps the byte order for type long
{
  swapEndian ((void *) x, len, SizeOfLong);
}

int SocketIO::getEndian ()
// delivers endian mode for local machine
{
  long l;
  char *c = (char *) &l;

  //  n = sizeof (long) - 4;      // because of 8-Byte longs

  l = (1L << 24) + (2L << 16) + (3L << 8) + 4L;

  if (c[0] == 1 && c[1] == 2 && c[2] == 3 && c[3] == 4)
    return (BIG_ENDIAN);

  if (c[0] == 4 && c[1] == 3 && c[2] == 2 && c[3] == 1)
    return (LITTLE_ENDIAN);

  return (0);
}


int SocketIO::establish ()
// determines endian mode on socket. if sender is true, data will
// be sent to the socket, otherwise data is expected from socket
{
  char *str;

  if (socket <= 0)
    return (-1);

  SizeOfDouble = sizeof (double);
  SizeOfFloat  = sizeof (float);
  SizeOfLong   = sizeof (long);
  SizeOfInt    = sizeof (int);
  SizeOfShort  = sizeof (short);

  ownEndian    = getEndian ();
  remoteEndian = BIG_ENDIAN;

  if (server == TRUE)
    {
    str = (char*) readData (strlen ("request2serve"));
    //                               XCversion 2.0

    // tell old XploRe client to update
    if (strncmp (str, "request2serve", strlen ("request2serve")) == 0)
      {
      str = strdup ("accepted\n1\n1\n");
      writeData (str, strlen (str));

      str = strdup ("Your client is no longer supported\n");
      writeData (str, strlen (str));
      str = strdup ("Please update.\n");
      writeData (str, strlen (str));
      str = strdup ("Server is closing connection ...\n");
      writeData (str, strlen (str));
      str = strdup ("_END_OF_XPLORE_\n");
      writeData (str, strlen (str));

      //      cerr << "unknown client protocol" << endl;
      close (socket);
      exit (1);
      }

    // do we have the right client?
    if (strncmp (str, CLIENT_VERSION, strlen (CLIENT_VERSION)) != 0)
      {
      //  cerr << "unknown client protocol" << endl;
      close (socket);
      exit (1);
      }

    // go on
    str = strdup (SERVER_VERSION);
    writeData (str, strlen (str));
    delete[] str;

    }
  else
    {
    str = strdup (CLIENT_VERSION);
    writeData (str, strlen (str));
    delete[] str;

    str = (char*) readData (strlen (SERVER_VERSION));
    if (strncmp (str, SERVER_VERSION, strlen (SERVER_VERSION)) != 0)
      {
	// cerr << "unknown server protocol" << endl;
      close (socket);
      exit (1);
      }

    delete[] str;
    }

  return (status);
}

int SocketIO::writeString (const char *x)
{
  long len = strlen (x);

  if (len > 0)
    return (writeChar (x, len));

  return (OK);
}

int SocketIO::writeChar (const char *x, long len)
// writes char array to socket

{
  //  printf ("WriteChar %li %s\n", len, x);

  writeSize (len);
  logl(len);
  logt(x);
  return (writeData (x, len));
}

char * SocketIO::readChar (long &len)
// reads char array from socket
{
  char *x, *y;

  len = readSize ();
  logt("SocketIO::readChar");
  logl(len);
  x = (char *) readData (len);
  logp(x, len);
  if (x == NULL)              // transmission error
    return (NULL);

  y = new char[len+1];
  strncpy (y, x, len);
  y[len] = '\0';

  logt(y);
  delete x;

  return (y);
}

int SocketIO::writeShort (short *x, long len)
// writes short array to socket

{
  if (ownEndian != remoteEndian)
    swapEndian (x, len);

  writeSize (len);
  return (writeData (x, len * SizeOfShort));
}

short * SocketIO::readShort (long &len)
// reads short array from socket
{
  short *x;

  len = readSize ();
  x   = (short *) readData (SizeOfShort*len);

  if (x == NULL)
    status = NOT_ENOUGH_MEMORY;

  if (status != OK)
    return (NULL);
  
  if (ownEndian != remoteEndian)
    swapEndian (x, len);

  return (x);
}

int SocketIO::writeInt (int *x, long len)
// writes integer array to socket

{
  if (ownEndian != remoteEndian)
    swapEndian (x, len);

  writeSize (len);
  return (writeData (x, len * SizeOfInt));
}

int * SocketIO::readInt (long &len)
// reads integer array from socket
{
  int *x;

  len = readSize ();
  x   = (int *) readData (SizeOfInt*len);

  if (x == NULL)
    status = NOT_ENOUGH_MEMORY;

  if (status != OK)
    return (NULL);

  if (ownEndian != remoteEndian)
    swapEndian (x, len);

  return (x);
}

int SocketIO::writeLong (long *x, long len)
// writes double array to socket
{
  if (ownEndian != remoteEndian)
    swapEndian (x, len);

  writeSize (len);
  return (writeData (x, len * SizeOfLong));
}

long * SocketIO::readLong (long &len)
// reads double array from socket
{
  long *x;

  len = readSize ();
  logt ("SocketIO::readLong");
  logl (len);
  x   = (long *) readData (SizeOfLong*len);
  logt ("nach readData");
  logp ((char *) x, SizeOfLong*len);

  if (x == NULL)
    status = NOT_ENOUGH_MEMORY;

  if (status != OK)
    return (NULL);
  
  if (ownEndian != remoteEndian)
    swapEndian (x, len);

  return (x);
}

int SocketIO::writeFloat (float *x, long len)
// writes float array to socket

{
  if (ownEndian != remoteEndian)
    swapEndian (x, len);

  writeSize (len);
  return (writeData (x, len * SizeOfFloat));
}

float * SocketIO::readFloat (long &len)
// reads float array from socket
{
  float *x;

  len = readSize ();
  x   = (float *) readData (SizeOfFloat*len);

  if (x == NULL)
    status = NOT_ENOUGH_MEMORY;

  if (status != OK)
    return (NULL);

  if (ownEndian != remoteEndian)
    swapEndian (x, len);

  return (x);
}

int SocketIO::writeDouble (double *x, long len)
// writes double array to socket

{
  if (ownEndian != remoteEndian)
    swapEndian (x, len);

  writeSize (len);
  return (writeData (x, len * SizeOfDouble));
}

double * SocketIO::readDouble (long &len)
// reads double array from socket
{
  double *x;

  len = readSize ();
  x   = (double *) readData (SizeOfDouble*len);
  
  if (x == NULL)
    status = NOT_ENOUGH_MEMORY;

  if (status != OK)
    return (NULL);

  if (ownEndian != remoteEndian)
    swapEndian (x, len);

  return (x);
}



