/* $Header: /home/xplore/cvs/xpl4gr/socketio.h,v 1.3 2002/08/16 12:31:50 sigbert Exp $ */

// Socket basic I/O class

// transparent data exchange between big/little endian

#ifndef _SOCKETIO_
#define _SOCKETIO_

// Microsoft defines different names
#if defined (WIN32)
// following defines break including of windows specific includes
// therefore undefine them
#ifdef N
#undef N
#endif
#ifdef Free
#undef Free
#endif

#ifndef X4singleserver
#include<winsock2.h>
#define close closesocket
#endif

//#define read(a,b,c) recv(a,b,c,0)
//#define write(a,b,c) send(a,b,c,0)
//#define sleep Sleep

#else

#ifndef X4singleserver
#include <sys/types.h>
#include <sys/socket.h>
#endif

#include <unistd.h>

#endif

// #define UNKNOWN       0
#define TRUE  1
#define FALSE 0

#define OK            0
#define NO_CONNECTION 1
#define WRITE_ERROR   2
#define READ_ERROR    3

#ifdef BIG_ENDIAN
#undef BIG_ENDIAN
#endif

#ifdef LITTLE_ENDIAN
#undef LITTLE_ENDIAN
#endif

#define UNKNOWN             0
#define NEGATIVE_LENGTH     1
#define NOT_ENOUGH_MEMORY   2
#define TRANSMISSION_ERROR  3
#define BIG_ENDIAN          4
#define LITTLE_ENDIAN       5
//#define TEXT                6    // neither little nor big endian

class SocketIO {
public:

  SocketIO ();  // this constructor is only called under X4singleserver
  SocketIO (int s, int serv)  : socket (s), status (OK), server (serv) { }
  ~SocketIO ()                         { close (socket); }

  int establish   ();    
// determines endian mode

  const int getSocket () const { return (socket); }

  long readBytes (void *buf, long len);
  int      writeData   (const void *v, long len);
  void *   readData    (long len);

  void   writeToken (long);
  long   readToken  ();

  int    writeSize (long);
  long   readSize  ();

  int  writeString  (const char   *x);          // = writeChar (x, strlen (x));

  int writeChar    (const char *x, long len);      // 1 Byte 
  int writeShort   (short  *x, long len);      // 2 Bytes    (Linux, i386)
  int writeInt     (int    *x, long len);      // 4 Bytes    (Linux, i386)
  int writeLong    (long   *x, long len);      // 4 Bytes    (Linux, i386)
  int writeFloat   (float  *x, long len);      // 4 Bytes    (IEEE 754)
  int writeDouble  (double *x, long len);      // 8 Bytes    (IEEE 754)

  char *   readChar   (long &len);              // 1 Byte 
  short *  readShort  (long &len);              // 2 Bytes    (Linux, i386)
  int  *   readInt    (long &len);              // 4 Bytes    (Linux, i386)
  long *   readLong   (long &len);              // 4 Bytes    (Linux, i386)
  float *  readFloat  (long &len);              // 4 Bytes    (IEEE 754)
  double * readDouble (long &len);              // 8 Bytes    (IEEE 754)



private:
  int  getEndian  ();
  void swapEndian (void   *x, long len, int size);
  void swapEndian (short  *x, long len);
  void swapEndian (int    *x, long len);
  void swapEndian (long   *x, long len);
  void swapEndian (float  *x, long len);
  void swapEndian (double *x, long len);

  int socket, status, server, ownEndian, remoteEndian;

  int SizeOfDouble, SizeOfFloat, SizeOfLong, SizeOfInt, SizeOfShort;

};


#endif

