/* $Header: /home/xplore/cvs/xpl4gr/spline.C,v 1.3 2002/08/16 12:31:50 sigbert Exp $ */


#include <stdlib.h>
#include <math.h>
// v. 22.06.98


void cholesky_5( double** B, double* D, double ** L, long length)
{
long i;
   D[0]=B[0][2];
   if(length >1){ 
     L[1][1]=B[1][1]/D[0]; 
     D[1] = B[1][2]-L[1][1]*L[1][1]*D[0];}
   for(i=3;i<= length; i++){
      L[i-1][0]=B[i-1][0]/D[i-3];
      L[i-1][1] = (B[i-1][1]-L[i-2][1]*L[i-1][0]*D[i-3])/D[i-2];
      D[i-1] = B[i-1][2]-L[i-1][1]*L[i-1][1]*D[i-2]-L[i-1][0]*L[i-1][0]*D[i-3];
     }
}   



double **matrix(long nrow, long ncol)
/* allocate a double matrix with subscript range m[0..nrow-1][0..ncol-1] */
{
	long i;
	double **m;

	/* allocate pointers to rows */
	m = new double*[nrow];
	/* allocate rows and set pointers to them */
	for(i=0;i<nrow;i++)
	{
		m[i]=new double[ncol];
	}
	/* return pointer to array of pointers to rows */
	return m;
}
void  delete_matrix(double **m, long nrow)
{
 long i;
 for(i=0;i<nrow;i++)
 {
	delete[] m[i];
 }
 delete[] m;
}


void eqn(double ** B, double* z, double* x, long length)
{
double* u;
double* D;
double** L;  
long i;
      u=new double[length];
      D= new double[length];
      L= matrix (length,2);
      cholesky_5(B,D,L,length);  
      u[0]=z[0];
      if (length >1)
        u[1]=z[1]-L[1][1]*u[0];
      for(i=3;i<=length;i++)
         u[i-1]=z[i-1]-L[i-1][1]*u[i-2]-L[i-1][0]*u[i-3];
      for(i=0;i<length; i++)
        u[i]=u[i]/D[i];
      x[length-1] = u[length-1];
      if(length >1)
        x[length-2] = u[length-2]-L[length-1][1]*x[length-1];
      for(i=length-2;i>0; i--)
        x[i-1]=u[i-1]- L[i][1]*x[i]-L[i+1][0]*x[i+1];

      delete[] u;
      delete[] D;
      delete_matrix(L,length);  
}

void diagonal(double **B,long length, double** Diag){
  //computes tree main subdiagonals of the matrix B
double* D;
D=new double[length];
double** L;
L = matrix(length,2);
  cholesky_5(B, D, L, length);

int i;
i=length-1;
  Diag[i][2]   = 1.0/D[i];
i=i-1;
  Diag[i+1][1]=-L[i+1][1]*Diag[i+1][2];
  Diag[i][2] = 1.0/D[i]- L[i+1][1]*Diag[i+1][1]; 
   
 
  for(i=length-3; i>=0; i--){

    Diag[i+2][0]=-L[i+2][0]*Diag[i+2][2]-L[i+1][1]*Diag[i+2][1];
 
    Diag[i+1][1]=-L[i+1][1]*Diag[i+1][2]-L[i+2][0]*Diag[i+2][1];

    Diag[i][2] = 1.0/D[i]- L[i+1][1]*Diag[i+1][1]-L[i+2][0]*Diag[i+2][0]; 
  }
 
              
delete_matrix(L,length);   
delete[] D;
}

double trace(double** B, double* h, double* w, long length)
{
long i;
double** Diag;
double tr=0.0;
Diag=matrix(length-2,3);
diagonal(B,length-2, Diag);

i=0;
tr +=   ((1.0/h[i])*Diag[i][2]*(1.0/h[i]))/w[i];
i=i+1;

tr += ((1.0/h[i])*Diag[i][2]*(1.0/h[i])
           + (1.0/h[i]+1.0/h[i-1])*Diag[i-1][2]*(1.0/h[i]+1.0/h[i-1])
          - 2.0*(1.0/h[i])*Diag[i][1]*(1.0/h[i]+1.0/h[i-1]))/w[i];

for(i=2;i<length-2;i++)
  {
    tr +=   ((1.0/h[i])*Diag[i][2]*(1.0/h[i])
             + (1.0/h[i]+1.0/h[i-1])*Diag[i-1][2]*(1.0/h[i]+1.0/h[i-1])
             + (1.0/h[i-1])*Diag[i-2][2]*(1.0/h[i-1])
             - 2.0*(1.0/h[i-1])*Diag[i-1][1]*(1.0/h[i]+1.0/h[i-1])
             + 2.0*(1.0/h[i-1])*Diag[i][0]*(1.0/h[i])
             - 2.0*(1.0/h[i])*Diag[i][1]*(1.0/h[i]+1.0/h[i-1]))/w[i];
  }  
i=length-2;
tr += ((1.0/h[i-1])*Diag[i-2][2]*(1.0/h[i-1])
        + (1.0/h[i]+1.0/h[i-1])*Diag[i-1][2]*(1.0/h[i]+1.0/h[i-1])
        - 2.0*(1.0/h[i-1])*Diag[i-1][1]*(1.0/h[i]+1.0/h[i-1]))/w[i];       
i=i+1;
tr +=((1.0/h[i-1])*Diag[i-2][2]*(1.0/h[i-1]))/w[i];
delete_matrix(Diag,length-2);
return tr;
}
 

void spline_(double* x, double* y, double* D2y, long length)
{
double* right_side;
double** B;
double* temp;
long i;
  temp = new double[length-2];
  right_side = new double[length-2];
    for(i=2;i<length;i++)
      right_side[i-2] = (y[i]-y[i-1])/(x[i]-x[i-1])-
                       (y[i-1]-y[i-2])/(x[i-1]-x[i-2]);


  B = matrix(length-2,3);
  for(i=0; i<length-2; i++)
    B[i][0]=0.0;


  for(i=2; i< length; i++)
    B[i-2][2]=(x[i]-x[i-2])/3;

  for(i=3; i<length; i++)
    B[i-2][1]=(x[i-1]-x[i-2])/6;
    
   B[0][1]=0.0;

  eqn(B,right_side,temp,length-2); 
  D2y[0]=0.0;
  D2y[length-1]=0.0;
  for(i=1;i<length-1;i++)
    D2y[i]=temp[i-1];
  
delete_matrix(B,length-2);
delete[] right_side;
delete[] temp;
}

void splsm_(double* x, double* y, double* D2y, double * w, 
            long length, double lambda)
{
double* h;
long i;
double* temp;

h=new double[length-1];
temp=new double[length];
for(i=0;i<length-1;i++)
    h[i]=x[i+1]-x[i];

for(i=1;i<length-1;i++)
    temp[i-1]=(y[i+1]-y[i])/h[i]-(y[i]-y[i-1])/h[i-1];
 
double ** B;
B = matrix(length-2,3);

for(i=0;i<length-2;i++){
  B[i][2]= (h[i]+h[i+1])/3.0;

}

for(i=1;i<length-2;i++)
  {B[i][1]= h[i]/6.0;}

for(i=0;i<length-2; i++)
  {B[i][2]= B[i][2]+lambda*
     ((1.0/h[i]+1.0/h[i+1])*(1.0/h[i]+1.0/h[i+1])/w[i+1]+1.0/h[i]/h[i]/w[i]
      + 1.0/h[i+1]/h[i+1]/w[i+2]);}

for(i=1;i<length-2; i++)
  B[i][1] = B[i][1]+lambda*
              ((-1.0/h[i-1]-1.0/h[i])/h[i]/w[i]+
               (-1.0/h[i]-1.0/h[i+1])/h[i]/w[i+1]);

for(i=2;i<length-2;i++)
  B[i][0]=lambda/h[i]/h[i-1]/w[i];

eqn(B,temp,D2y,length-2);

for(i=2;i<length-2;i++)
 y[i]=y[i]-lambda*(D2y[i-2]/h[i-1]+D2y[i-1]*(-1.0/h[i-1]-1.0/h[i])+
                   D2y[i]/h[i])/w[i];
y[0]=y[0]-lambda*D2y[0]/h[0]/w[0];
if(length > 3)
y[1]=y[1]-lambda*((-1.0/h[0]-1.0/h[1])*D2y[0]+D2y[1]/h[1])/w[1];

y[length-1]=y[length-1]-lambda*D2y[length-3]/h[length-2]/w[length-1];
i=length-2;

y[i]=y[i]-lambda*(D2y[i-2]/h[i-1]+D2y[i-1]*(-1.0/h[i-1]-1.0/h[i]))/w[i];
  
D2y[length-1]=0.0;
for(i=length-2; i>0; i--)
   D2y[i]=D2y[i-1];
D2y[0]=0.0;

delete_matrix(B,length-2);
delete[] h;
delete[] temp;
}

void splad_(double* x, double* y, double* D2y, double* w, long length)
{
double* h;
long i;
double* temp;
double lambda=1.0;
double** B;
double* data;
double tr;
double risk;
double risk_min;
double* D2y_min;
double* y_min;
long run;
  
   D2y_min=new double[length];
   y_min = new double[length];
   data=new double[length];
   B=matrix(length-2,3);
   h=new double[length-1];
   temp=new double[length];
   
//preliminary estimation of the variance
double var=0.0;
   for(i=1;i<length;i++)
     var+=(y[i-1]-y[i])*(y[i-1]-y[i]);
   risk_min=var/double(length-1);  
   for(i=0;i<length;i++)
      data[i]=y[i];


   for(i=0;i<length-1;i++)
        h[i]=x[i+1]-x[i];
   for(i=1;i<length-1;i++)
         temp[i-1]=(data[i+1]-data[i])/h[i]-(data[i]-data[i-1])/h[i-1];
   for(run=0;run < 32000 ;run++){       
      for(i=0;i<length-2;i++){
         B[i][2]= (h[i]+h[i+1])/3.0;}

      for(i=1;i<length-2;i++)
          {B[i][1]= h[i]/6.0;}

      for(i=0;i<length-2; i++)
         {B[i][2]= B[i][2]+lambda*
                   ((1.0/h[i]+1.0/h[i+1])*(1.0/h[i]+1.0/h[i+1])/w[i+1]+
                    1.0/h[i]/h[i]/w[i]
                   + 1.0/h[i+1]/h[i+1]/w[i+2]);}

      for(i=1;i<length-2; i++)
            B[i][1] = B[i][1]+lambda*
              ((-1.0/h[i-1]-1.0/h[i])/h[i]/w[i]+
               (-1.0/h[i]-1.0/h[i+1])/h[i]/w[i+1]);

      for(i=0;i<length-2;i++)
          B[i][0]=lambda/h[i]/h[i-1]/w[i];

      eqn(B,temp,D2y,length-2);
     
      for(i=2;i<length-2;i++)
         y[i]=data[i]-lambda*(D2y[i-2]/h[i-1]+
                      D2y[i-1]*(-1.0/h[i-1]-1.0/h[i])+
                      D2y[i]/h[i])/w[i];
      y[0] = data[0]-lambda*D2y[0]/h[0]/w[0];
      y[1] = data[1] - lambda*((-1.0/h[0]-1.0/h[1])*D2y[0]
                     + D2y[1]/h[1])/w[1];
      y[length-1]=data[length-1]-lambda*D2y[length-3]/h[length-2]/w[length-1];
      i=length-2;
      y[i]=data[i]-lambda*(D2y[i-2]/h[i-1]+
           D2y[i-1]*(-1.0/h[i-1]-1.0/h[i]))/w[i];
      if(run==1)
         { 
	         for(i=0;i<length;i++)
                  { if (i<length-2)
                     D2y_min[i]=D2y[i]; 
		     y_min[i]=y[i];}
         }
      // computation of the unbiased risk estimator
      //compute the trace of the matrix W^{-1} Q(R+lambda*Q^TW^{-1} Q)^{-1}Q^T
          tr= trace(B, h, w, length);
         // compute the predicted risk
          risk =0.0;
          for(i=0;i<length;i++)
             risk += (data[i]-y[i])*(data[i]-y[i])*w[i];
          risk =risk/(lambda*tr)/(lambda*tr); 
	  if(risk>=risk_min)
               { 
	         for(i=0;i<length;i++){
                   if (i< length-2)
                     D2y_min[i]=D2y[i]; 
                   y_min[i]=y[i];}
                 break; 
               } 
           else{
             risk_min=risk;
             lambda=lambda/1.1;
           }
}  
    
       D2y[length-1]=0.0;
       for(i=0; i<length-1; i++)
         D2y[i+1]=D2y_min[i];

       for(i=0; i<length; i++)
         y[i]=y_min[i];  
       D2y[0]=0.0;
           
     
delete   D2y_min;
delete   y_min;
delete   data;
delete_matrix(B,length-2);
delete   h;
delete   temp;  

}


void splint_(double* x, double* y, double* D2y, long length,
            double* x_new, double* y_new, long length_new)
{
long i,k;

long upper_index=1;
double A,B,C,D;
i=0;

  for(i=0 ;i<length_new;i++){
    if(x_new[i]<x[0]){
       upper_index=1;
       A = (x[upper_index] - x_new[i])/(x[upper_index]-x[upper_index-1]);
       B = 1-A;
       C = (A*A*A-A)*(x[upper_index]-x[upper_index-1])*
           (x[upper_index]-x[upper_index-1])/6.0;
       D = (B*B*B-B)*(x[upper_index]-x[upper_index-1])*
           (x[upper_index]-x[upper_index-1])/6.0; 
       y_new[i] = A*y[upper_index-1]+B*y[upper_index]+
	 C*D2y[upper_index-1]+D*D2y[upper_index];
       goto loop; }

    if(x_new[i]>x[length-1]){
       upper_index=length-1;
       A = (x[upper_index] - x_new[i])/(x[upper_index]-x[upper_index-1]);
       B = 1-A;
       C = (A*A*A-A)*(x[upper_index]-x[upper_index-1])*
           (x[upper_index]-x[upper_index-1])/6.0;
       D = (B*B*B-B)*(x[upper_index]-x[upper_index-1])*
           (x[upper_index]-x[upper_index-1])/6.0; 
       y_new[i] = A*y[upper_index-1]+B*y[upper_index]+
	 C*D2y[upper_index-1]+D*D2y[upper_index];
       goto loop; }

    if(x_new[i]<=x[upper_index]&&x_new[i]>=x[upper_index-1]){
       A = (x[upper_index] - x_new[i])/(x[upper_index]-x[upper_index-1]);
       B = 1-A;
       C = (A*A*A-A)*(x[upper_index]-x[upper_index-1])*
           (x[upper_index]-x[upper_index-1])/6.0;
       D = (B*B*B-B)*(x[upper_index]-x[upper_index-1])*
           (x[upper_index]-x[upper_index-1])/6.0; 
       y_new[i] = A*y[upper_index-1]+B*y[upper_index]+
	          C*D2y[upper_index-1]+D*D2y[upper_index];}
    else{ upper_index=upper_index+1; i=i-1;}
  loop: ;
   }
  
}


long tie(double * x, double* y, double *weigh, double* w, long length)
{
    long i, k, counter;
    double * xtemp;
    double * ytemp;
    xtemp = new double[length];
    ytemp = new double[length];
    double sum;
    counter = 0;
    long flag;
    double temp;
    for(i=0;i<length;i++){
       sum = y[i];
       w[counter]=1.0;
       flag = i;
       for(k=1;k<= length-i;k++){
           if (k+i< length)
              temp=x[k+i];
           else
              temp = x[length-1]+1.0;
           if(temp <= x[i])
             { 
               w[counter]++;
               weigh[counter]+=weigh[k+i]; 
               sum=sum*(1.0-1.0/double(k+1))+y[k+i]/double(k+1);
               flag++;
             }
           else
             {
               xtemp[counter] = x[i];
               ytemp[counter] = sum;
               i=flag;
               counter++;
               break;
             }
	 }
      
    }
    for(i=0;i < counter;i++)
    {
       x[i]=xtemp[i];
       y[i]=ytemp[i];
     }
    delete xtemp; delete ytemp;
    return counter;
} 

long tie(double * x, double* y,  double* w, long length)
{
    long i, k, counter;
    double * xtemp;
    double * ytemp;
    xtemp = new double[length];
    ytemp = new double[length];
    double sum;
    counter = 0;
    long flag;
    double temp;
    for(i=0;i<length;i++){
       sum = y[i];
       w[counter]=1.0;
       flag = i;
       for(k=1;k<= length-i;k++){
           if (k+i< length)
              temp=x[k+i];
           else
              temp = x[length-1]+1.0;
           if(temp <= x[i])
             { 
               w[counter]++; 
               sum=sum*(1.0-1.0/double(k+1))+y[k+i]/double(k+1);
               flag++;
             }
           else
             {
               xtemp[counter] = x[i];
               ytemp[counter] = sum;
               i=flag;
               counter++;
               break;
             }
	 }
      
    }
    for(i=0;i < counter;i++)
    {
       x[i]=xtemp[i];
       y[i]=ytemp[i];
     }
    delete xtemp; delete ytemp;
    return counter;
} 

long untie(double * x,  double *w, long length)
{
    double * xtemp;
    long new_length=0;
    long counter=0;
    long i,k;
    for(i=0;i<length; i++)
      new_length += long(w[i]);
    xtemp = new double[new_length];
    for( i=0;i< length; i++)
        for(k=1;k<= long(w[i]);k++)
        {
           xtemp[counter]=x[i];
           counter++;
        }
    for(i=0;i < new_length;i++)
    {
       x[i]=xtemp[i];
     }
     delete xtemp; 
     return new_length;
} 

void splintc(double* x, double* y, double* D2y, long length,
            double* x_new, double* y_new, long length_new)
{
     double* xtemp;
     double* w;
     xtemp = new double[length];
     w     = new double[length];
     long tie_length,i;
     for(i=0;i<length;i++)
       xtemp[i] = x[i];
     tie_length = tie(x,y,w,length);
     tie_length = tie(xtemp,D2y,w,length);
     if(tie_length < 4)
       {delete w; delete xtemp; return;}
     splint_(x,y,D2y,tie_length,x_new,y_new,length_new);
     delete xtemp; 
     delete w;
}
void splinec(double* x, double* y, double* D2y, long length)
{
   
     double* w;
     long templength;
     w     = new double[length];
     long tie_length,i;
     tie_length = tie(x,y,w,length);
     if(tie_length < 4)
       {delete w; return;}
     spline_(x,y,D2y,tie_length);
     templength = untie(x,w,tie_length);
     templength = untie(D2y,w,tie_length);
     templength = untie(y,w,tie_length);
     delete w;
}

void splsmc(double* x, double* y, double* D2y,  
            long length, double lambda, double *weigh)
{

 
     double* w;
     long templength;
     w     = new double[length];
     long tie_length,i;
     tie_length = tie(x,y,weigh,w,length);
     if(tie_length < 4)
       {delete w; return;}
     splsm_(x,y,D2y,weigh,tie_length,lambda);
     templength = untie(x,w,tie_length);
     templength = untie(D2y,w,tie_length);
     templength = untie(y,w,tie_length);
     delete w;
}

void spladc(double* x, double* y, double* D2y, long length, double *weigh)
{

 
     double* w;
     long templength;
     w     = new double[length];
     long tie_length,i;
     tie_length = tie(x,y,weigh,w,length);
     if(tie_length < 4)
       {delete w; return;}
     splad_(x,y,D2y,weigh,tie_length);
     templength = untie(x,w,tie_length);
     templength = untie(D2y,w,tie_length);
     templength = untie(y,w,tie_length);
     delete w;
}
























