/* $Header: /home/xplore/cvs/xpl4gr/wavelet.C,v 1.3 2002/08/16 12:31:53 sigbert Exp $ */

#include <iostream.h>
#include <stddef.h>
#include <stdlib.h>
#include <math.h>

static double **matrixroot, *matrixelem;

static double *h;
static int RANGE;

inline int circle(int x,
				 int  period){
	if( x>= 0)
	  return x%period;
	else
	  return period-1+(x+1)%period;
 }

 void filter_mirror(double *data,
						  int    length)
 {

 double    *temp, *temp_1;
 double     mult;
 int       i,k,time;
 int       center = length/2;
 temp   = new double[length];
 temp_1 = new double[length];
 for (i=0;i< center;i++) {
	 time = 2*i;
	 temp[i]   = 0.0;
	 temp_1[i] = 0.0;
	 mult = 1.0;
	 for(k=0;k<RANGE;k++){
		mult =-mult;
		temp[i]   +=  h[k]*data[circle(time+k,length)];
		temp_1[i] +=  mult*h[k]*data[circle(time+1-k,length)];
	 }
 }
 for(i=0;i< center;i++){
	data[i]=temp[i];
	data[i+center] =temp_1[i];
 }
 delete[] temp;
 delete[] temp_1;
 }

 void filter_TI(double *data,
					 double *T0_vector,
					 double *T1_vector,
					 int length){
	double *temp;
	double *temp_1;
	int       i;
	int       center = length/2;
	temp   = new double[length];
	temp_1 = new double[length];

				for(i=0;i<length;i++)
				{
					temp[i]=data[i];
					temp_1[i]=data[circle(i-1,length)];
				}
				filter_mirror(temp,length);
				filter_mirror(temp_1,length);
				for(i=0;i<center;i++){
				  T1_vector[i]=temp[i];
				  T1_vector[i+center]=temp_1[i];
				  T0_vector[i]=temp[i+center];
				  T0_vector[i+center]=temp_1[i+center];
				}
	delete [] temp;
	delete [] temp_1;

 }

  void fwt_TI(double *data,
				  double **TI_table, /*TI_table contains all wavelet
										  tramsform of shifted data*/
				  int   length, /*number of rows in TI_table */
				  int   depth)  /*number of colloms in TI_table */
{
		double *T0;
		double *T1;
		double *temp;
		int i,k,blocks;

		T0   = new double[length];
		T1   = new double[length];
		temp = new double[length];
		int loc;
		int size;
		/* initiate TI_table */
		filter_TI(data,T0,T1,length);
		for(i=0;i<length;i++)
		{
			TI_table[depth-1][i] = T0[i];
			TI_table[depth-2][i] = T1[i];
		}
		/* TI_table is ready for iterations*/
		blocks=2;
		for(k=depth-2; k>0; k--)
		{
		  for(loc=0;loc < blocks;loc++)
		  {
			  size=length/blocks;
			  for(i=loc*size;i<(loc+1)*size;i++)
					temp[i-loc*size]=TI_table[k][i];
			  filter_TI(temp,T0,T1,size);
			  for(i=loc*size;i<(loc+1)*size;i++)
				 {
					TI_table[k][i]   = T0[i-loc*size];
					TI_table[k-1][i] = T1[i-loc*size];
				 }
			}
		blocks=2*blocks;
		}
	delete [] T0;
	delete [] T1;
	delete []temp;
 }
/* fast wavelet transform*/
void fwt(double *data,
			int    length, /*length of the input vector; must be a power of 2 */
			int    level)  /*number of father wavelets coeff.
								  must be a power of 2 */

{
int i;
int    local=length;
for(i=1; local > level;i++)
{
	  filter_mirror(data,local);
	  local = local/2;
}
}

 void filter_mother_father(double *x,
									 int   length)
 {
/*
  Recalculates mother wavelet coefficients
  at a certain level in terms of the father
  coefficients at the next level. */

  double *temp;
  double mult = -1.0;
  int     i,k;

  temp = new double[length];
  for(i=0;i< length/2; i++){
	 temp[2*i]=x[i];
	 temp[2*i+1] =0.0;
  }
  for(i= 0;i< length; i++){
	 mult = -mult;
	 x[i]=0.0;
	 for(k=0;k<RANGE;k++)
		x[i] += mult*h[k]*temp[circle(i-1+k,length)];
  }
delete[]temp;
}

 void filter_father_father(double *x,
									  int   length){
/*
  Recalculates father wavelets coefficients
  at a certain level in terms of father wavelets coefficients
  at the next level. */

  double *temp;
  int     i,k;
  temp = new double[length];
  for(i=0;i<length/2; i++)
  {
	 temp[2*i]=x[i];
	 temp[2*i+1]=0.0;
  }
  for(i=0;i< length; i++)
  {
	  x[i] =  0.0;
	  for(k=0;k<RANGE;k++)
		  x[i] += h[k]*temp[circle(i-k,length)];
  }
  delete[]temp;
}
/* fast invert wavelet transform */
void invert_fwt(double *data,
					  int    length,
					  int    level )
{
	 double *temp, *temp_1, *sum;
	 int   mult =level;
	 int   i,k;

	 temp   = new double[length];
	 temp_1 = new double[length];
	 sum    = new double[length];
	 for (i=0;i<length;i++)
	 {
		 sum[i]   = 0.0;
		 temp[i]  = 0.0;
		 temp_1[i]= 0.0;
	 }
	 for(k=0;k< level; k++)
		temp_1[k] = data[k];
	 for (i=0;mult <=length/2; i++ )
	 {
		 for(k=0;k<mult; k++)
			 temp[k]=data[mult+k];
		 filter_mother_father(temp,2*mult);
		 filter_father_father(temp_1,2*mult);
		 for(k=0;k<length;k++)
			 sum[k]=sum[k]+temp[k];
		 mult=2*mult;
		 if(mult <=length/2)
			filter_father_father(sum,2*mult);
	  }
	  for(k=0;k<length;k++)
		 data[k] = sum[k]+temp_1[k];

	  delete[]temp;
	  delete[]temp_1;
	  delete[]sum;
}
void invert_TI(double** TI,
					int length,
					int depth,
					double* vec)
{
int i,k;
double* sum;
double* temp;
double* father;
double* temp_1;
double* mother;
int bl_sz=length;
int block_size;
int left;
//define initial block size
for (i=0;i<depth-1;i++)
{
	  bl_sz=bl_sz/2;
}
temp   = new double[length];
father = new double[length];
temp_1 = new double[length];
mother = new double[length];
sum    = new double[length];
for(i=0;i<length;i++)
{
	sum[i]=0;
}
// calculate the mother part of wavelet transform
block_size=bl_sz;
for(k=1;k<depth;k++)
{
	for(i=0;i<length;i++)
	{
			 mother[i]=TI[k][i];
	}
	for(left=0; left<length;left+=2*block_size)
	{
		for(i=0;i<block_size; i++)
		{
			temp[i]= mother[i+left];
			temp_1[i] = mother[i+left+block_size];
		}
		filter_mother_father(temp,2*block_size);
		filter_mother_father(temp_1,2*block_size);
		for(i=0;i< 2*block_size; i++)
		{
		  father[i+left]=
					(temp[i]+temp_1[circle(i+1,2*block_size)])/2.0;
		}
	}
	for(i=0;i<length;i++)
	{
	  sum[i]+=father[i];
	}
	block_size=2*block_size;
	 if(block_size<length)
	 {
		for(left=0; left<length;left+=2*block_size)
		{
			for(i=0;i<block_size; i++)
			{
				temp[i]= sum[i+left];
				temp_1[i] = sum[i+left+block_size];
			}
			filter_father_father(temp,2*block_size);
			filter_father_father(temp_1,2*block_size);
			for(i=0;i< 2*block_size; i++)
			{
			  sum[i+left]=
				  (temp[i]+temp_1[circle(i+1,2*block_size)])/2.0;
			}
		 }
	 }
}

// calculate the father part of wavelet transform
for(i=0;i<length;i++)
{
		 father[i]=TI[0][i];
}
block_size=bl_sz;
for(k=0;k<depth-1;k++)
{
  for(left=0; left<length;left+=2*block_size)
  {
	 for(i=0;i<block_size; i++)
	 {
		temp[i]=father[i+left];
		temp_1[i] =father[i+left+block_size];
	 }
	 filter_father_father(temp,2*block_size);
	 filter_father_father(temp_1,2*block_size);
	 for(i=0;i< 2*block_size; i++)
	 {
		father[i+left]=(temp[i]+temp_1[circle(i+1,2*block_size)])/2.0;
	 }
  }
  block_size=2*block_size;
}

for(i=0;i< length; i++)
{
  vec[i]=father[i]+sum[i];
}
delete [] temp;
delete [] father;
delete [] temp_1;
delete [] mother;
delete [] sum;
}

void retrieve(double **TI,  //extracts fwt from TI_table
				 int shift,    // shift of the data in time domain
				 int length,
				 int depth,
				 double *vec)
{
  int left=length;
  int i,k;
  int rmd =0;
  int intshift=shift;
  int size=length;
  double* temp;
  temp = new double[length];
  for(k=depth-1; k>0; k--)
  {
		left=left/2;
		size=size/2;
		rmd=rmd+(intshift%2)*size;
		for(i=0; i<size; i++)
		{
		  temp[i]=TI[k][i+rmd];
		}
		intshift=intshift/2;
		for(i=0; i<size; i++)
		{
		  vec[i+left]=temp[circle(i-intshift,size)];
		}
	}
		for(i=0; i<size; i++)
		{
		  temp[i]=TI[0][i+rmd];
		}
		for(i=0; i<size; i++)
		{
		  vec[i]=temp[circle(i-intshift,size)];
		}

  delete [] temp;
}

int get_level(int length, //returns number of father coefficients
				  int  depth)
{
int l=length;
int i;
for(i=1; i<depth;i++)
{
  l=l/2;
}
return l;
}

int get_depth(int length, //returns number of rows in TI_table
				  int  level)
{
int l;
int i=0;
for(l=length; l>=level;l=l/2)
i++;
return i;
}


 double  **wvmatrix(int nrow, int ncol)
/* allocate a double matrix with subscript range m[0..nrow-1][0..ncol-1] */
{
	int i;
	double **m;

	/* allocate pointers to rows */
	m = new double*[nrow];
	/* allocate rows and set pointers to them */
	for(i=0;i<nrow;i++)
	{
		m[i]=new double[ncol];
	}
	/* return pointer to array of pointers to rows */
	return m;
}
void  delete_matrix(double **m, int nrow)
{
 int i;
 for(i=0;i<nrow;i++)
 {
	delete[] m[i];
 }
 delete[] m;
}
void show(double* vector, int length)
{
int i;
 cout.precision(2);
for(i=0; i<length; i=i+4) {
	cout.width(8);
	cout << (vector[i])<<"   ";
	cout.width(8);
	cout << (vector[i+1])<<
  "   ";
  cout.width(8);
  cout << (vector[i+2])<<
  "   ";
  cout.width(8);
  cout <<(vector[i+3])<<"\n";
  }
  cout << "--\n";
}

void setwc (int newrange, double *newh)
{ h = newh;
  RANGE = newrange;
}

/*
void main(){

RANGE = 4;
h=new double[RANGE];
h[0] = 0.4829629131445341;
h[1] = 0.8365163037478077;
h[2] = 0.2241438680420134;
h[3] = -0.1294095225512603;

int len =32;
int dep=4;
 create a signal
double* signal = new double[len];
double* temp   = new double[len];
for(int i=0; i<len;i++)
  signal[i]=exp(double(i)/double(len));
 allocate the TI_table
matrixroot=matrix(dep,len);
calculate TI_table
fwt_TI(signal,matrixroot,len,dep);
 check that fwt_TI and invert_TI work correctly
	 // invert TI_table
	 invert_TI(matrixroot,len,dep,temp);
	 //compare temp[] and signal[]
	 for(i=0;i<len;i++)
		temp[i]=temp[i]-signal[i];
	 show(temp,len);
 check that matrixroot contains the ordinary wavelet
 transform
	 // calculate the ordinary wavelet transform of the shifted signal
	 int shift = 5;
	 double* signal_shifted=new double[len];
	 for(i=0;i<len;i++)
		signal_shifted[i]=signal[circle(i-shift,len)];
	 fwt(signal_shifted,len,get_level(len,dep));
	 // retreive the wavelet coefficient from matrixroot
	 retrieve(matrixroot,shift,len,dep,temp);
	 //compare the results
	 for(i=0;i<len;i++)
		temp[i]=temp[i]-signal_shifted[i];
	 show(temp,len);
}

*/
void filter_mirror2(double** data,
                    long    length,
                    long asm_index)

{
long i,k,ls,as;
double* temp;
temp = new double[length];
  for (i=0;i<length;i++){
    for(k=0; k<length;k++)
      temp[k]=data[k][i];
    ls=length;
    for(as=0;as<=asm_index;as++)
          if(ls > 1)
          {filter_mirror(temp,ls); ls=ls/2;}
    for(k=0; k<length;k++)
         data[k][i]=temp[k];
  }

 for (i=0;i<length;i++){
    for(k=0; k<length;k++)
      temp[k]=data[i][k];
    ls=length;
    for(as=0;as<=asm_index;as++)
          if(ls > 1)
          {filter_mirror(temp,ls); ls=ls/2;}
    for(k=0; k<length;k++)
         data[i][k]=temp[k];
  }
delete [] temp;
}



void invert_filter_mirror(double* data, long block_length){

double* temp;
double* temp1;
long i;
temp  = new double[block_length];
temp1 = new double[block_length];
for(i=0; i<block_length/2; i++) {
  temp[i]= data[i];
  temp1[i] =data[i+block_length/2];
}
filter_father_father(temp, block_length);
filter_mother_father(temp1, block_length);
for(i=0; i< block_length; i++)
  data[i]=temp[i]+temp1[i];
delete [] temp;
delete [] temp1;
}



void invert_filter_mirror2(double** data,
                    long    length,
                    long asm_index)

{
long i,k,ls,as;
double* temp;
temp = new double[length];
long block_length=2*length;
for(as=0;as<=asm_index;as++)
  if(block_length>2)
   block_length=block_length/2;

  for (i=0;i<length;i++){
    for(k=0; k<length;k++)
      temp[k]=data[k][i];
    ls=block_length;
    for(as=0;as<=asm_index;as++)
          if(ls <= length)
          {invert_filter_mirror(temp,ls); ls=ls*2;}
    for(k=0; k<length;k++)
         data[k][i]=temp[k];
  }

 for (i=0;i<length;i++){
    for(k=0; k<length;k++)
      temp[k]=data[i][k];
    ls=block_length;
    for(as=0;as<=asm_index;as++)
          if(ls <= length)
          {invert_filter_mirror(temp,ls); ls=ls*2;}
    for(k=0; k<length;k++)
         data[i][k]=temp[k];
  }
delete [] temp;
}

void invert_fwt2(double** data,
          long   length,
          long   level,
          long   asm_index)
{
 long block_length,min_block_length;
 long  mult=1;
 long as;
 for (as=0;as<=asm_index;as++){
     mult=mult*2;
 }
 for(min_block_length=length; min_block_length>=mult*level;
    min_block_length=min_block_length/mult)
   {}
for(block_length=min_block_length;block_length <= length;
    block_length=mult*block_length)
   {
    invert_filter_mirror2(data,block_length,asm_index);
   }
}

void fwt2(double** data,
          long   length,
          long   level,
          long   asm_index)
{
 long min_block_length,block_length;
 long  mult=1;
 long as;
 for (as=0;as<=asm_index;as++)
     mult=mult*2;
 for(min_block_length=length; min_block_length>=mult*level;
    min_block_length=min_block_length/mult)
   {}
 for(block_length=length;block_length >=min_block_length;
       block_length=block_length/mult)
   {
    filter_mirror2(data,block_length,asm_index);
   }
}







