/*
 * BDT-AsyncReqs.c      modul obsahujici fce pro komumnikaci s uzivatelem
 */

#define BDT_ASYNCHREQS_C

#include "wed_editor.h"
/*** Modify Start : d-tec  ***/
#include <mbctype.h>
/*** Modify End : d-tec  ***/



////////////////////////////////////////////////
// NASLEDUJI RUTINY PRO ASYNCHRONNI REQUESTRY //
////////////////////////////////////////////////

/* pomocna funkce, ktera provede dosazeni daneho Location do kurzoru
 * a popr. take zobrazi vse potrebne ... */
// format = FALSE --> disp udava, zda se ma zobrazit nova pozice
// format = TRUE  --> disp udava parametr zobraz pro FormatLines
// pokud je zadano end --> oznaci se blok mezi pos a end!
// trueend - end ukazuje na konec bloku (posl. znak) --> nutno o znak posunout
void SetCursorPosition(TEditFile *teditfile, Location *pos, Location *end,
                       BOOL format, BOOL disp, BOOL trueend)
//>>
{    
     EditFile *doc = teditfile->GetEditFile();

     Location *cursor = &doc->cursor, *cursorline = &doc->cursorline;
     Location loc, pomend;

     if (doc->block) UnmarkBlock(teditfile, TRUE);

     *cursor = *pos;
     *cursorline = *cursor;
     (void) GetPrevPos(teditfile, cursorline, FALSE, 0);
  // oznaceni blokem
     if (end)
     {    pomend = *end;
          if (trueend) GetNextPos(teditfile, &pomend, TRUE, 1);
          SelectTextBlock(teditfile, pos, &pomend);
     }

//     if (format) FormatLines(doc, GetFirstFormatLine(doc, &loc, FALSE, FALSE), disp, FALSE, TRUE, FALSE, TRUE, 0, FALSE, FALSE, FALSE, FALSE, FALSE);

//     if (disp && NOT format)
//     {    
//          DrawCursor(doc, TRUE, 0, 0, 0, 0, FALSE, TRUE);
//          (void) DrawLine(doc, cursorline, 0, NULL, 0, 0, 0, FALSE, FALSE, FALSE, FALSE, FALSE, TRUE);
//          MakeCursorVisible(doc, ~0L, ~0L, TRUE);
//     }
     if (disp || format)
          MakeCursorVisible(teditfile, ~0L, ~0L, TRUE);
}
//<<

// pomocna funkce, ktera prevede retezec pro vyhledani na binarni podobu
// a ykontroluje, zda tam nejsou nepripustne znaky
// vysledny retezec musi byt pole alespon s MAXSEARCHSTR + MAXWORDLEN znaky
// replace - ma se prevest replace (str2) nebo search (str1)
void TransformSearchReplcString(ReplaceState *state, UBYTE *result, BOOL replace)
//>>
{
     UBYTE *src = (replace) ? state->str2 : state->str1;

     if (*src == '\0') return;

     for(; *src; src++, result++)
     {
          if (*src < ' ')
               *result = ' ';
          else if (*src != '\\')
               *result = *src;
          else // specialni znaky
          {    src++;
               switch (*src)
               {
                    case 't': *result = CH_TAB;   break;
//                    case 'n': *result = CH_NUMBER;break;
//                    case 'N': *result = CH_HIERNUM;    break;
//                    case 'd': *result = CH_DATE;  break;
//                    case 'D': *result = CH_TIME;  break;
//                    case 'p': *result = CH_PAGENUM;    break;
//                    case 'P': *result = CH_PAGECOUNT;  break;
                    case '\\':*result = '\\';     break;
                    default:  src--;   *result = *src; break;
               }
          }
     }

     *result = '\0';
}
//<<

/* pomocna fce pro FindReplace - nalezne dalsi vyskyt slova a oznaci ho
 * jako blok ...; pokud dojde na konec dokumentu vraci FALSE (= nic
 * nenasla), jinak TRUE; hledaci mod a vse dle ReplaceState */
// beg, end - vrati se v nich pozice zacatku a konce ...
// skipchars - udava, kolik znaku mame nejprve preskocit ...
BOOL FindSearchWord(TEditFile *teditfile, ReplaceState *state, Location *beg, Location *end, UWORD skipchars)
//>>
{
     EditFile *doc = teditfile->GetEditFile();

     MemoryBlock *mb = (state->from_beg) ? doc->block_first : doc->cursor.block, *mbpom, *mbpom2;
     ULONG off = (state->from_beg) ? 0 : doc->cursor.block_off, offpom, offpom2;

     Paragraph *para_act = (state->from_beg) ? doc->para_first : doc->cursor.para, *para_pom;
     Paragraph *para_find;

     LONG i, len;
     BOOL nalezeno = FALSE, samechar, first = TRUE;
     UBYTE *pos = mb->memory + off, *str, *pospom;
     UBYTE pomsearchbuf[MAXSEARCHSTR + MAXWORDLEN + 4];

  // uprava retezce
     if (*state->str1 == '\0')
          return(FALSE);

  // klasicke vyhledavani
     TransformSearchReplcString(state, pomsearchbuf, FALSE);
     len = strlen(pomsearchbuf);
     if (len == 0) return(FALSE);

frombeg:
     while (NOT nalezeno)
     {    
          str = pomsearchbuf;
          for( i = (first) ? skipchars : 1; i > 0; i--)
               if ( NULL == (pos = (state->forward) ? IncrementPosMB(&mb, &off) : DecrementPosMB(&mb, &off)) )
                    break;
          first = FALSE;
          if (pos == NULL) break;  // zacatek/konec dokumentu
       // vlastni porovnavani
          samechar = (state->case_sens) ? (*pos == *str) : (ToUpper((ULONG) *pos) == ToUpper((ULONG) *str));
          /*** Modify Start : d-tec ***/
          if (samechar) {
               i = 1;
               if (_ismbslead(mb->memory, mb->memory + off))
               {
                    samechar = (*(pos + 1) == *(str + 1));
                    if (samechar) {
                         str++;
                         i = 2;
                    }
               }
          }
          /*** Modify End : d-tec  ***/
          if (samechar)
          { // test shody; je-li whole_word, musi predchazet SPACE ci TAB
               if (state->whole_word)
               {
                    mbpom = mb;    offpom = off;
                    // ASCII editor: CH_EOL is only at the end of paragraph, nowhere else!
                    pospom = DecrementPosMB(&mbpom, &offpom);
//                    while (ENDLINE( (pospom = DecrementPosMB(&mbpom, &offpom)) ? *pospom : '\0' ));
                    if (NOT ((pospom == NULL) || (*pospom == CH_SPACE) || (*pospom == CH_TAB) || (*pospom == CH_EOL)) )
                         continue;
//                    { 
//                      // musime overit, ze k tomu nedoslo kvuli zacatku odstavce
//                         for(para_pom = para_act; para_pom; para_pom = para_pom->next)
//                              if ((para_pom->block_first == mb) && (para_pom->off_first == off))
//                              {    para_act = para_pom;
//                                   break;
//                              }
//                         if (para_pom == NULL)    // neni tomu tak
//                              continue;
//                    }
               }
               for(i = 1, str++, mbpom = mb, offpom = off; i < len; i++, str++)
               {
                    if (NULL == (pospom = IncrementPosMB(&mbpom, &offpom)) )
                         break;
                    if (ENDLINE(*pospom)) { i--; str--; continue; }
                    samechar = (state->case_sens) ? (*pospom == *str) : (ToUpper((ULONG) *pospom) == ToUpper((ULONG) *str));
                    /*** Modify Start : d-tec ***/
                    if (_ismbslead(mbpom->memory, mbpom->memory + offpom))
                    {
                         if (samechar) samechar = (*(pospom + 1) == *(str + 1));
                         i++; str++;
                    }
                    /*** Modify End : d-tec  ***/
                    if (NOT samechar) break;
               }
               if (i == len)
               if (state->whole_word)
               {
                    mbpom2 = mbpom;          offpom2 = offpom;
                    // ASCII editor: CH_EOL is only at the end of paragraph, nowhere else!
                    pospom = IncrementPosMB(&mbpom2, &offpom2);
//                    while (ENDLINE( (pospom = IncrementPosMB(&mbpom2, &offpom2)) ? *pospom : '\0' ));
                    if ((pospom == NULL) || (*pospom == CH_SPACE) || (*pospom == CH_TAB) || (*pospom == CH_EOL))
                         nalezeno = TRUE;
//                    else
//                    { // musime overit, ze k tomu nedoslo kvuli konci odstavce
//                         for(para_pom = para_act; para_pom; para_pom = para_pom->next)
//                         {
//                              if ((para_pom->block_last == mbpom) && (para_pom->off_last == offpom))
//                              {    para_act = para_pom;
//                                   break;
//                              }
//                              if ((para_pom->block_first == mbpom2) && (para_pom->off_first == offpom2))
//                              {
//                                   break;
//                              }
//                         }
//                         if (para_pom)    // nalezeno
//                              nalezeno = TRUE;
//                    }
               }
               else nalezeno = TRUE;
          }
     }

     if (nalezeno)
     { // nastavime kurzor, hledane slovo - dano mbpom, offpom!
       // mb, off = zacatek; mbpom, offpom = posledni pismeno!
#ifdef CHECK_BUGS
          if (NOT BlockToLocation(teditfile, beg, mb, off, NULL, NULL)) InternalError("FindNextWord - nelze dodelat Location k pozici 1");
          if (NOT BlockToLocation(teditfile, end, mbpom, offpom, NULL, NULL)) InternalError("FindNextWord - nelze dodelat Location k pozici 2");
#else
          (void) BlockToLocation(teditfile, beg, mb, off, NULL, NULL);
          (void) BlockToLocation(teditfile, end, mbpom, offpom, NULL, NULL);
#endif
          if (beg->para != end->para)   // zacatek a konec v ruznych odstavcich
          {    
               nalezeno = FALSE;   
               goto frombeg;    
          }
     }

     return(nalezeno);
}
//<<

/* pomocna fce pro FindReplace - nahradi dany vyskyt slova ... */
/* Provadi kontrolu, zda novy retezec neobsahuje divne znaky ! */
/* format udava, zda se ma formatovat - pak se ale predpoklada,
 * ze je prislusne slovo videt na obrazovce --> zobraz dame TRUE */
// TATO FCE NASTAVUJE KURZOR NA POCATEK NOVEHO SLOVA!
// PO POUZITI JSOU BEG A END NEPLATNE!
// staybefore - ma kurzor zustat pred slovem nebo skoncit za nim ?
void ReplaceThisWord(TEditFile *teditfile, ReplaceState *state, Location *beg, Location *end, BOOL format, BOOL staybefore)
//>>
{
     EditFile *doc = teditfile->GetEditFile();

     Location loc, pomloc, end2;
     BOOL pombool;
//     UBYTE *p;

     if (doc->block) UnmarkBlock(teditfile, TRUE);
  // nastavime kurzor ...
     loc = *beg;
     doc->cursor = loc;
     GetPrevPos(teditfile, &loc, FALSE, 0);  // loc --> na zacatek radku
     doc->cursorline = loc;
//     if (format)
//          pombool = (NULL == GetFirstFormatLine(doc, &loc, FALSE, TRUE)) ? TRUE : FALSE;
  // najdeme pozici pro vlozeni textu...
     end2 = *end;
     GetNextPos(teditfile, &end2, FALSE, 1);
  // pokud ma kurzor byt za slovem --> musime ho tam premistit...
     if (NOT staybefore)
     {    pomloc = end2;
          doc->cursor = pomloc;
          GetPrevPos(teditfile, &pomloc, FALSE, 0);     // pomloc --> na zacatek radku
          doc->cursorline = pomloc;
     }
  // vlozeni str2, pokud v nem neco je
     if (*state->str2)
     {    
          UBYTE pomreplcbuf[MAXSEARCHSTR + MAXWORDLEN + 4];

          TransformSearchReplcString(state, pomreplcbuf, TRUE);
       // vlozeni retezce
          if (NOT InsertInBlock(teditfile, &end2, pomreplcbuf, (ULONG) strlen(pomreplcbuf), FALSE, NUM_DOC_LOCATIONS, NULL, FALSE))
               BeepEditor();
     }
  // smazani puvodniho slova...
     DeleteDataLoc(teditfile, beg, end, FALSE, FALSE, FALSE);
  // pripadne formatovani
//     if (format)
//          FormatLines(doc, GetFirstFormatLine(doc, &loc, pombool, FALSE), TRUE, FALSE, TRUE, FALSE, TRUE, 0, FALSE, FALSE, FALSE, FALSE, FALSE);
  // nastaveni new & changes
     doc->newfile = FALSE;        doc->changes = TRUE;
}
//<<

