/*
 * BDT-Docs-Clipboard.c  modul obsahujici fce pro zapis/cteni clipboardu
 */

#define BDT_DOCS_CLIPBOARD_C

#include "wed_editor.h"

// auxiliary functions - data sizes //

// computes the size = number of bytes for the current line
// or (if max == TRUE) for the longest line
ULONG GetLineSize(TEditFile *teditfile, BOOL max)
{
     EditFile *doc = teditfile->GetEditFile();

     Paragraph *para;
     ULONG size = 0;

     if (max)
     {
          for(para = doc->para_first; para; para = para->next)
               if (para->length >= size)
                    size = para->length;
     }
     else size = doc->cursor.para->length;

     size += 4;	// 2 bytes for EOL, one for '\0'

     return size;
}

// computes the size = number of bytes for selected block (if block is given)
// or (if block == FALSE) the size of the whole document
ULONG GetBlockSize(TEditFile *teditfile, BOOL block)
{
     EditFile *doc = teditfile->GetEditFile();

     Paragraph *para, *paraend;
     MemoryBlock *mb, *mbend;
     ULONG size = 0;

     if (block)
     {
          if (NOT doc->block)
               return 0;

          mb = doc->block_begin.block;
          mbend = doc->block_end.block;
          if (mb == mbend)
               size = doc->block_end.block_off - doc->block_begin.block_off;
          else
          {
               size = mb->size - mb->free - doc->block_begin.block_off;
               for(mb = mb->next; mb && (mb != mbend); mb = mb->next)
                    size += mb->size - mb->free;
               size += doc->block_end.block_off;
#ifdef CHECK_BUGS
               if (mb == NULL) InternalError("GetBlockSize - Wrong block selection!");
#endif
          }

          para = doc->block_begin.para;
          paraend = doc->block_end.para;
          for( ; para && (para != paraend); para = para->next)
               size++;
          size++;
     }
     else
     {
          for(mb = doc->block_first; mb; mb = mb->next)
               size += mb->size - mb->free;

          for(para = doc->para_first; para; para = para->next)
               size++;
     }

     return size;
}

// reads the data from handle
// clip is currently pointer to a string
// return NULL or error message and the number of paragraphs in num_paras
char *LoadClipASCII(TEditFile *teditfile, CLIPHANDLE clip, ULONG *num_paras)
//ERRORCODE LoadClipASCII(EditFile *doc, struct IFFHandle *iff, ULONG *num_paras)
//>>
{
     EditFile *doc = teditfile->GetEditFile();

     Location misto = doc->cursor;
//     UBYTE blok_pom[BLOKPOM_VELIKOST + 16];   // buffer pro nacitani
     CLIPHANDLE blok_pom = clip;
//     LONG error = 0L;
     ULONG kolik_nacist;
     ULONG zbyva_nacist, prave_nacteno; // pro evidenci pri postupnem nacitani textu
     LONG i, pozice, size;
     UBYTE *lasteol;
     char *errorcode = NULL;
//     ERRORCODE errorcode = 0;
//     BOOL byl_CHRS = TRUE;

// eol = FALSE; ... to uvazujeme
     *num_paras = 1;

//          zbyva_nacist = cn->cn_Size;   // velikost nacitaneho ASCII textu
          zbyva_nacist = strlen(clip);   // velikost nacitaneho ASCII textu

          // nacitani textu
          while(zbyva_nacist > 0)
          {
//               if (NOT CheckFreeMemory(0))
//               {
//                    errorcode = GS(MSG_NO_MEMORY_AVAILABLE);
//                    goto end;
//               }
//
//               kolik_nacist = min_ul(BLOKPOM_VELIKOST, zbyva_nacist);
               // everything is in memory - no file/clip reading
               prave_nacteno = kolik_nacist = zbyva_nacist;
               blok_pom = clip;
//               if (NOT (kolik_nacist == (prave_nacteno = ReadChunkBytes(iff, blok_pom, kolik_nacist))))
//               { errorcode = MSG_ERROR_PARSING_CLIP; goto end; }
               // has to be done here - prave_nacteno can change in PrevodMeziKody!
               zbyva_nacist -= prave_nacteno;
               // nacten text o velikosti kolik_nacist == prave_nacteno;

               PrevodMeziKody(blok_pom, &prave_nacteno, strlen(blok_pom) + 1);
//               PrevodMeziKody(clip blok_pom, prave_nacteno, code, TRUE, 2);
               // find the last CH_EOL character - the formating in SplitParagraph has to be done
               // and it is sufficient to do it just at the end!
               // we assume here the clip = blok_pom is C string
               lasteol = (UBYTE *) strrchr((char *) blok_pom, CH_EOL);

               for(i = 0, pozice = 0; (ULONG) i < prave_nacteno; i++)
               {
                    if ((blok_pom[i] == CH_EOL) || ((ULONG) i + 1 == prave_nacteno))
                    {
                         misto.next = NULL;
                         size = i - pozice + ((blok_pom[i] == CH_EOL) ? 0 : 1);
#ifdef CHECK_BUGS
                         if (size < 0) InternalError("LodaCLipASCII - size < 0");
#endif
                         if (size > 0)
                              if (NOT InsertInBlock(teditfile, &misto, blok_pom + pozice, (ULONG) size, FALSE, NUM_DOC_LOCATIONS, &misto, FALSE))
                              {
                                   errorcode = GS(MSG_NO_MEMORY_AVAILABLE);
                                   goto end;
                              }
                    }
                    if (blok_pom[i] == CH_EOL)
                    {
                         BOOL lastpara = (blok_pom + i == lasteol) ? TRUE : FALSE;

                         pozice = i + 1;
     // zbytecne --> SplitPara misto nepozmeni, ale opravi pri pripadnych presunech mezi bloky
                         SplitParagraph(teditfile, &misto, /* doc->cursor.page, */ FALSE, lastpara, lastpara);
//                         SplitParagraph(teditfile, &misto, /* doc->cursor.page, */ FALSE, FALSE, FALSE);
     // zbytecne                    misto = pomloc;
                         GetNextPos(teditfile, &misto, TRUE, 0);
//                    // musime provest kontrolu, protoze ve SplitPara se nedela pri cursor FALSE
//                         CheckTextStyleList(doc, misto.para, misto.para->ts_first, FALSE, NULL);
                         *num_paras += 1;
                    }
               }
          }    // konec cyklu, v nemz se nacita obsah souboru
//     }    // konec nekonecneho cyklu

end:
     return (errorcode);
} /* end LoadClipASCII() */
//<<

// save the data to buffer - clip
// clip is currently pointer to a string
// return NULL or error message and the number of paragraphs in num_paras
// do_crlf = EOLs saved as CR+LF
char *SaveClipASCII(TEditFile *teditfile, CLIPHANDLE clip, BOOL do_crlf)
//>>
{
     EditFile *doc = teditfile->GetEditFile();
     
     char *errorcode = NULL;

     Location *beg = &doc->block_begin, *end = &doc->block_end;
     Paragraph *para_act;
     MemoryBlock *mb, *mbpom;
     UBYTE *pozice;
     ULONG char_offset;
     ULONG block_offset, offpom, size;
     UBYTE *line, *pom, *lineend;
     UBYTE znak;
//     UBYTE blok_pom[BLOKPOM_VELIKOST + 4];
     CLIPHANDLE blok_pom = clip;
     BOOL /* uspech = TRUE, */ eopar, /* special, */ setpom, blockend = FALSE;
     BOOL konecradku, /* nenimisto, */ block_change;

// EOL - zde zapisujeme na konci odstavcu !  eol = FALSE; !

     // here, there is just one chunk of memory, pozice is not reset for every paragraph!
     pozice = blok_pom;
     for(para_act = beg->para; para_act && NOT blockend; para_act = para_act->next)
     { 
          // pocatecni inicializace       
	  eopar = FALSE;
//	  pozice = blok_pom;
	  if (para_act == beg->para)    // 1. odstavec - nebereme od pocatku
	  {
	       char_offset = beg->chars_off;
	       mb = beg->block;              block_offset = beg->block_off;
	       lineend = line = mb->memory + block_offset;
	  }
	  else
	  {
	       char_offset = 0;
	       mb = para_act->block_first;   block_offset = para_act->off_first;
	       lineend = line = mb->memory + block_offset;
	  }

	  // vykonny cyklus - prochazeni odstavce
	  block_change = FALSE;

	  for(pom = line; TRUE; char_offset++)
	  {    
               size = lineend - pom;
               
	       znak = *line;
	       setpom = FALSE;
	       konecradku = (ENDLINE(znak)) ? TRUE : FALSE;
               
	       if ((mb == para_act->block_last) && (block_offset == para_act->off_last))
	       {    
                    eopar = TRUE;    // dosli jsme na konec odstavce
		    if ((end->para == para_act) || (end->para == para_act->next) && (end->chars_off == 0))    // dosli jsme na konec bloku
			 blockend = TRUE;
	       }

//	       nenimisto = (pozice - blok_pom) + size + REZERVA;
	       if ( konecradku || block_change || eopar || blockend )
//	       if ( konecradku || block_change || eopar || special || blockend ||
//		    (nenimisto >= BLOKPOM_VELIKOST) )
	       // s rezervou na vepsani konce radku, data ci cisla stranky...
	       // musime provest zapis do pomocneho bloku pameti
	       {    
                    if (size) 
                         memcpy(pozice, pom, (size_t) size);

		    pozice += size;
		    if (eopar)    // je konec radku a ma se to zapsat
                         // EOL write at the end of paragraph, but only if the block continues to the next paras!
                         if (end->para != para_act)
                         {
                              if (do_crlf)
                                   *pozice++ = CH_EOLD;
                              *pozice++ = CH_EOL;
                         }
//			 switch (znak)
//			 {
//                              case CH_EOLD:
//#ifdef CHECK_BUGS
//				   InternalError("BlockToclip - CH_EOLD na konci para");
//#endif
//				   *pozice++ = '-';
//			      case CH_EOL:
//			      case CH_EOP:
//				   *pozice++ = CH_EOL;
//				   break;
//			 }

//		    if (nenimisto || eopar || blockend)  // zde prekonvertovat a zapsat do souboru:
//		    {    PrevodMeziKody(blok_pom, pozice - blok_pom, code, FALSE, 2);
//			 if (pozice - blok_pom != WriteChunkBytes(iff, blok_pom, pozice - blok_pom))
//			 {    errorcode = MSG_ERROR_SAVING_FILE; goto end;    }
//			 pozice = blok_pom;
//		    }
		    pom = line;
	       } // end if (preruseni ...)

	       if (eopar || blockend) break;     // je konec odstavce ?
//#ifdef CHECK_BUGS
//	       else if (znak == CH_EOP) InternalError("BlockToCLip - ma byt konec odstavce");
//#endif
	       // znak konce radku ci specialni znaky preskocime - viz vyse
	       if (konecradku) setpom = TRUE;
//	       if (konecradku || special) setpom = TRUE;

	       // ziskani dalsiho znaku
	       mbpom = mb;
	       offpom = block_offset;
	       lineend = line + 1;
	       line = IncrementPosMB(&mb, &block_offset);
	       if (mb == mbpom)         // zmena pametoveho bloku ?
	       { 
                    block_change = FALSE; 
                    lineend = line; 
               }
	       else block_change = TRUE;

               // EOL - jump over, already written
	       if (setpom) pom = line;
               
	    // pokud jsme na konci bloku --> prerusime
	       if ((end->block == mb) && (end->block_off == block_offset))
	       {    
                    blockend = TRUE;    
               }
	  } // konec prochazeni odstavce
     } // konec cyklu pro dokument

     return errorcode;
}

/* hlavni fce modulu */
/* vlastni fce pro ulozeni bloku oznaceneho do clipboardu */
BOOL BlockToClip(TEditFile *teditfile, BOOL undo)
//>>
{    
     EditFile *doc = teditfile->GetEditFile();

     TClipboard clipboard(teditfile->GetHandle());
     HANDLE cliphandle;
     CLIPHANDLE buffer;
//     Paragraph *para_pom;
//     ParaStyle *parastyle_pom, ps_kopie;
//     TextStyle *ts_pom;
//     ERRORCODE errorcode = 0;
     char *errorcode = 0;
// nepouzito     ULONG nula = 0L;
     ULONG blocklen;
     ULONG i;
     BOOL uspech = TRUE;
     BOOL savelastparaend = FALSE;

// musi byt oznacen blok !
#ifdef CHECK_BUGS
     if (NOT doc->block) { InternalError("BlockToClip - neni blok"); return(TRUE); }
#else
     if (NOT doc->block) return(TRUE);
#endif
     
// calculate the size of block to be cliped
     blocklen = GetBlockSize(teditfile, TRUE);
     if (blocklen == 0)  // nothing to be done
          return(TRUE);
     
// open clipboard
     if (NOT clipboard)  // clipboard opened by somebody else
          return(FALSE);
// not necessary - constructor does it
//     if (NOT clipboard.OpenClipboard())
//          return(FALSE);
     if (NOT clipboard.EmptyClipboard())
     {
          errorcode = GS(MSG_ERROR_CLIPEMPTY);   goto end;
     }
     // GHND includes ZEROINIT option, therefore, we do not have to write \0
     if (NULL == (cliphandle = ::GlobalAlloc(GHND, blocklen + 1)))    // one extra byte for '\0'
     {
          errorcode = GS(MSG_NOT_ENOUGH_MEMORY);   goto end;
     }
     if (NULL == (buffer = (CLIPHANDLE) ::GlobalLock(cliphandle)))
     {
          errorcode = GS(MSG_NOT_ENOUGH_MEMORY);   goto end;
     }

// wait pointer
     if (NOT undo)
          ChangePointer(teditfile, GS(MSG_BLOCK_TO_CLIP), TRUE, TRUE);

// write the clipboard data
     if (errorcode = SaveClipASCII(teditfile, buffer, GETGLOBALPREFS->crlf_eols))
          goto end;

// set the clipboard data
     ::GlobalUnlock(cliphandle);
     clipboard.SetClipboardData(CF_TEXT, cliphandle);

end:
// not necessary - constructor / destructor does it automatically
//   clipboard.CloseClipboard();

     if (errorcode)
     {
	  uspech = FALSE;
          if (cliphandle) ::GlobalFree(cliphandle);
	  if (NOT undo) Information2(teditfile, errorcode);
     }

     if (NOT undo)
          ChangePointer(teditfile, doc->name, FALSE, FALSE);

     return (uspech);
}    /* end BlockToClip */
//<<

/* fce pro vlozeni bloku v clipboardu do textu */
BOOL ClipToText(TEditFile *teditfile, ULONG *num_paras, BOOL undo, char *textsrc)
// if textsrc is non-null, the text is inserted from this string rather than clipboard
//>>
{
     EditFile *doc = teditfile->GetEditFile();

     TClipboard clipboard(teditfile->GetHandle());
     HANDLE cliphandle;
     CLIPHANDLE text;
     LONG error = 0L;
     char *errorcode = NULL;
//     ERRORCODE errorcode = 0;
     BOOL uspech = TRUE;

     *num_paras = 0;

#ifdef CHECK_BUGS
     if (doc == NULL) { InternalError("ClipToDoc"); return(FALSE); }
#else
     if (doc == NULL) return(FALSE);
#endif

     if (textsrc)
          text = textsrc;
     else
     {
          if (NOT clipboard)  // clipboard opened by somebody else
               return(FALSE);
          if (NOT clipboard.IsClipboardFormatAvailable(CF_TEXT))
               return(FALSE);
          if (NULL == (cliphandle = clipboard.GetClipboardData(CF_TEXT)))
               return(FALSE);
          if (NULL == (text = (CLIPHANDLE) ::GlobalLock(cliphandle)))
               return(FALSE);
     }

// wait pointer
     ChangePointer(teditfile, GS(MSG_CLIP_TO_TEXT), TRUE, TRUE);

// load data from clipboard
     if (errorcode = LoadClipASCII(teditfile, text, num_paras))
          goto end;

end: if (errorcode)
     {    
          uspech = FALSE;
	  Information2(teditfile, errorcode);
     }
     
     ChangePointer(teditfile, doc->name, FALSE, FALSE);
     if (textsrc == NULL)
          ::GlobalUnlock(cliphandle);

     return (uspech);
}
//<<

