//----------------------------------------------------------------------------
// BDEditor 2.2 - simplified ASCII only version
// Copyright (c) 2000, Pavel Cizek
// Compatibility file - defines analog of AmigaOS functions
//----------------------------------------------------------------------------

#define COMPAT_C

#include "wed_editor.h"

/////////////////////////
// general C functions //
/////////////////////////

////////////////////////////
// System Amiga functions //
////////////////////////////

// memory
APTR LibAllocPooled( APTR poolHeader, unsigned long memSize )
{
     APTR mem = malloc(memSize);
     
// it must clear the allocated memory! - MEMF_CLEAR
     if (mem != NULL)
          clrmem(mem, memSize);
     
     return mem;	// should return memory address
}

APTR LibCreatePool( unsigned long memFlags, unsigned long puddleSize, unsigned long threshSize )
{
     return (APTR) 1;	// return something nonzero - just in case sb. wants to test it, no pools under Windows
}

void LibDeletePool( APTR poolHeader )
{
     // it does nothing - no pools under Windows
}

void LibFreePooled( APTR poolHeader, APTR memory, unsigned long memSize )
{
     free(memory);
}

// files
int ExamineFileMode(const char *filename)
// returns type of the file (see EX_FM_#? constants) or EX_ERROR if error occurs
{ 
     struct stat statbuf;
     
     if (stat(filename, &statbuf))
          return EX_ERROR;
     if (statbuf.st_mode & S_IFREG)
          return EX_FM_REGULARFILE;
     else return EX_FM_DEVICE;
}

long ExamineFileSize(const char *filename)
// returns size of the file or EX_ERROR if error occurs
{
     struct stat statbuf;
     
     if (stat(filename, &statbuf))
          return EX_ERROR;
     if (statbuf.st_mode & S_IFREG)
          return statbuf.st_size;
     else return EX_ERROR;
}
     
long GetOpenFileSize(BPTR handle)
// returns size of the file or EX_ERROR if error occurs
{
     long curpos = ftell(handle);
     long size = EX_ERROR;

     if (curpos == -1L)
          return size;
     
     if (0 == fseek(handle, 0L, SEEK_END))
     {
          size = ftell(handle);
          if (size == -1L) size = EX_ERROR;
          if (fseek(handle, curpos, SEEK_SET))
               size = EX_ERROR;
     }     
     
     return size;
          
//     fstat works with a file handle!
//     struct stat statbuf;
//     
//     if (fstat(handle, &statbuf))
//          return EX_ERROR;
//     if (statbuf.st_mode & S_IFREG)
//          return statbuf.st_size;
//     else return EX_ERROR;
}

BPTR Open(const char *filename, ULONG mode)
// opens (binary) file and return its handle or NULL if not successfull
{
     if (mode == MODE_OLDFILE)
          return fopen(filename, "rb");

     if (mode == MODE_NEWFILE)
          return fopen(filename, "w+b");
     
     return NULL;
}

void Close(BPTR file)
// closes file
{ 
     fclose(file);
}

ULONG Read(BPTR file, UBYTE *buffer, ULONG bytes)
// reads from file bytes bytes and stores them to buffer (which has to be big enough)
{
    if (1 == fread(buffer, bytes, 1, file))
         return bytes;
    else return 0;
}

ULONG Write(BPTR file, UBYTE *buffer, ULONG bytes)
// writes to file bytes bytes from buffer
{
    if (1 == fwrite(buffer, bytes, 1, file))
         return bytes;
    else return 0;
}
    
int BDTDeleteFile(const char *filename)
// deletes file, return EX_OK or EX_ERROR 
{
     if (remove(filename))
          return EX_ERROR;
     else return EX_OK;
}

int Rename(const char *oldname, const char *newname)
// rename file oldname to newname
{
     if (rename(oldname, newname))
          return EX_ERROR;
     else return EX_OK;
}

// requesters
BOOL BDTFileRequester(TEditFile *teditfile, STRPTR text, STRPTR oktext, STRPTR mask, BOOL save, char *filename)
// opens a file requester using teditfile and parameters and stores the chosen file to editfile path and name
// if filename is supplied, then it is placed instead of requester in doc
// return TRUE (file selected) or FALSE (nothing selected)
{
     EditFile *doc = teditfile->GetEditFile();
     GlobalPrefs *prefs = GETGLOBALPREFS;

     char drive[MAXDRIVE], dir[MAXDIR], name[MAXFILE], ext[MAXEXT], nameext[MAXFILE+MAXEXT+4];
     char *newfile, *dfn = new char [MAXFILE];

     BOOL selectok = FALSE;

     // if open path was set and save path not yet
     if (strlen(prefs->def_savereq_path) == 0)
          if (strlen(prefs->def_openreq_path) > 0)
               strcpy(prefs->def_savereq_path, prefs->def_openreq_path);

     if (save) // save
     {
          (void) fnsplit(doc->path, drive, dir, name, ext);

          TFileSaveDialog::TData data(OFN_HIDEREADONLY | OFN_PATHMUSTEXIST | OFN_OVERWRITEPROMPT | OFN_EXPLORER,
                                      (*prefs->def_filter) ? prefs->def_filter : BDT_FILEMASK, 0,
                                      (doc->last_saved) ? dir : prefs->def_savereq_path,
                                      BDT_DEFAULT_EXT);
          strcpy(nameext, name);
          strcat(nameext, ext);
          if (dfn)
          {
            strcpy(dfn, nameext);
            data.FileName = dfn;
          }

          newfile = NULL;
          if (filename)
               newfile = filename;
          else if (TFileSaveDialog(teditfile, data).Execute() == IDOK)
               newfile = data.FileName;

          if (newfile)
          {
               int flags = fnsplit(newfile, drive, dir, name, ext);

               // set path and name of document
               strncpy(doc->path, newfile, PATH_LEN + FILENAME);
               doc->path[PATH_LEN + FILENAME] = '\0';  // strncpy does not have to place zero at the string end

               strcpy(doc->name, name);
               strcat(doc->name, ext);

               // update prefs - requester path
               if (filename == NULL)
                    strcpy(prefs->def_savereq_path, dir);

               selectok = TRUE;
          }
     }
     else // open
     {
          TFileOpenDialog::TData data(OFN_FILEMUSTEXIST | OFN_PATHMUSTEXIST | OFN_EXPLORER,
                                      (*prefs->def_filter) ? prefs->def_filter : BDT_FILEMASK, 0,
                                      prefs->def_openreq_path,
                                      BDT_DEFAULT_EXT);

          newfile = NULL;
          if (filename)
               newfile = filename;
          else if (TFileOpenDialog(teditfile, data).Execute() == IDOK)
               newfile = data.FileName;

          if (newfile)
          {
               int flags = fnsplit(newfile, drive, dir, name, ext);

               // set path and name of document
               strncpy(doc->path, newfile, PATH_LEN + FILENAME);
               doc->path[PATH_LEN + FILENAME] = '\0';  // strncpy does not have to place zero at the string end

               strcpy(doc->name, name);
               strcat(doc->name, ext);

               // update prefs - requester path
               if (filename == NULL)
                    strcpy(prefs->def_openreq_path, dir);

               selectok = TRUE;
          }
     }

     return selectok;
}

// misc
void CurrentTime(ULONG *secs, ULONG *mics)
{
     (void) time((long *) secs);
}

// nasledujici funkce slouzi pro konverzi na velka / mala pismena podle
// nastaveni editoru a locale
ULONG LetterUpperCase(ULONG znak)
//>>
{
     return toupper(znak);
}
//<<

ULONG LetterLowerCase(ULONG znak)
//>>
{
     return tolower(znak);
}
//<<

ULONG LetterChangeCase(ULONG znak)
//>>
{
     if (isalpha(znak)) 
	  znak = (znak & 0x20) ? (znak & ~0x20) : (znak | 0x20);

     return(znak);
}
//<<

ULONG ToLower(ULONG znak)
{
     return tolower(znak);
}

ULONG ToUpper(ULONG znak)
{
     return toupper(znak);
}

///////////////////////
// Windows functions //
///////////////////////

// ALL INVALIDATE FUNCTIONS INVALIDATE WHOLE WINDOW IF THEY DO NOT HAVE INFORMATION (W, H, FONT)

void InvalidateLines(TEditFile *teditfile, long lines, unsigned offset, BOOL erase = TRUE)
// lines = 0 ... invalidate whole window
//	      < 0 ... invalidate previous abs(lines)
//	      > 0 ... invalidate next abs(lines)
// offset > 0 ... invalidate from position = line given by offset (including the offset line itself)
//                (line == 1, offset == 1 <=> invalidate second line in window
//                (line == -1, offset == 1 <=> invalidate first line in window
{
     EditFile *doc = teditfile->GetEditFile();
     TRect rect;
     int wh = doc->winheight, ww = doc->winwidth;
     ULONG fh = doc->fontheight, fw = doc->fontwidth;
     ULONG wch = fh * doc->wincheight, wcw = fw * doc->wincwidth;

     if ( (ww == 0) || (wh == 0) || (wcw == 0) || (wch == 0) || (fw == 0) || (fh == 0) ||
          (lines == 0) || (abs_l(lines) >= doc->wincheight) && (offset == 0) )
     {
          teditfile->Invalidate();
          return;
     }

     if (lines > 0)
     	  rect.Set(0, offset * fh, ww, min_l((offset + lines) * fh, wh));
     else // if (lines < 0)
      	  rect.Set(0, max_l(0, (long) offset + lines) * fh, ww, offset * fh);

     teditfile->InvalidateRect(rect, FALSE);
}

void InvalidateLines(TEditFile *teditfile, long lines = 0)
// lines = 0 ... invalidate whole window
//   	 < 0 ... invalidate first abs(lines)
//       > 0 ... invalidate last abs(lines)
{
     EditFile *doc = teditfile->GetEditFile();
     TRect rect;
     int wh = doc->winheight, ww = doc->winwidth;
     ULONG fh = doc->fontheight, fw = doc->fontwidth;
     ULONG wch = fh * doc->wincheight, wcw = fw * doc->wincwidth;

     if ( (ww == 0) || (wh == 0) || (wcw == 0) || (wch == 0) || (fw == 0) || (fh == 0) ||
          (lines == 0) || (abs_l(lines) >= doc->wincheight))
     {
          teditfile->Invalidate();
          return;
     }

     if (lines > 0) // last lines
          rect.Set(0, wch - lines * fh /* + 1 */, ww, wh);
     else // if (lines < 0)
          rect.Set(0, 0, ww, -1 *  lines * fh /* - 1*/ );

     teditfile->InvalidateRect(rect, FALSE);
}

void InvalidateColumns(TEditFile *teditfile, long columns = 0)
// columns = 0 ... invalidate whole window
//			< 0 ... invalidate first abs(columns)
//	      > 0 ... invalidate last abs(columns)
{
     EditFile *doc = teditfile->GetEditFile();
     TRect rect;
     int wh = doc->winheight, ww = doc->winwidth;
     ULONG fh = doc->fontheight, fw = doc->fontwidth;
     ULONG wch = fh * doc->wincheight, wcw = fw * doc->wincwidth;

     if ( (ww == 0) || (wh == 0) || (wcw == 0) || (wch == 0) || (fw == 0) || (fh == 0) ||
          (columns == 0) || (abs_l(columns) > doc->wincwidth) )
     {
          teditfile->Invalidate();
          return;
     }

     if (columns > 0)
          rect.Set(wcw - columns * fw /* + 1 */, 0, ww, wh);
     else // if (columns < 0)
          rect.Set(0, 0, -1 * columns * fw /* - 1 */, wh);

     teditfile->InvalidateRect(rect, FALSE);
}

// ALL SCROLL FUNCTIONS INVALIDATE WHOLE WINDOW IF THEY DO NOT HAVE INFORMATION (W, H, FONT)
// (INVALIDATE = REDRAW OF THE WHOLE WINDOW ... RESULT IS OK)

void ScrollLines(TEditFile *teditfile, long lines)
// lines: negative - cursor up (to the BOF), positive - cursor down (to the EOF)
// should invalidate the empty area
{
     EditFile *doc = teditfile->GetEditFile();
     TRect rect;
     int wh = doc->winheight, ww = doc->winwidth;
     ULONG fh = doc->fontheight, fw = doc->fontwidth;
     ULONG wch = fh * doc->wincheight, wcw = fw * doc->wincwidth;

     if (lines == 0) return;

     if ( (ww == 0) || (wh == 0) || (wcw == 0) || (wch == 0) || (fw == 0) || (fh == 0) ||
          (abs_l(lines) >= doc->wincheight) )
     {
          teditfile->Invalidate();
          return;
     }

     teditfile->ScrollWindow(0, -1 * lines * fh);
     if (lines > 0)
          rect.Set(0, wch - lines * fh /* + 1 */, ww, wh);
     else // if (lines < 0)
          rect.Set(0, 0, ww, -1 * lines * fh /* - 1*/);

     teditfile->InvalidateRect(rect, FALSE);
}


void ScrollColumns(TEditFile *teditfile, long columns)
// columns: negative - left (to the BOL), positive - right (to the EOL)
// should invalidate the empty area
{
     EditFile *doc = teditfile->GetEditFile();
     TRect rect;
     int wh = doc->winheight, ww = doc->winwidth;
     ULONG fh = doc->fontheight, fw = doc->fontwidth;
     ULONG wch = fh * doc->wincheight, wcw = fw * doc->wincwidth;

     if (columns == 0) return;

     if ( (ww == 0) || (wh == 0) || (wcw == 0) || (wch == 0) || (fw == 0) || (fh == 0) ||
          (abs_l(columns) >= doc->wincwidth) )
     {
          teditfile->Invalidate();
          return;
     }

     teditfile->ScrollWindow(-1 * columns * fw, 0);
     if (columns > 0)
          rect.Set(wcw - columns * fw + 1, 0, ww, wh);
     else // if (columns < 0)
          rect.Set(0, 0, columns * fw - 1, wh);

     teditfile->InvalidateRect(rect, FALSE);
//     teditfile->InvalidateRect(rect);
//     teditfile->Invalidate(FALSE);
}

///////////////////////////
// misc Editor functions //
///////////////////////////

#define MESSBOX_FLAGS	(MB_APPLMODAL | MB_TOPMOST)

#ifdef CHECK_BUGS
void InternalError(const char *)
{
     (void) MessageBeep(MB_ICONEXCLAMATION);
}
#endif

void AlertMessage(TEditFile *teditfile, STRPTR txt1, STRPTR txt2)
{
     teditfile->MessageBox((char *) txt1);
     teditfile->MessageBox((char *) txt2);
}

void BeepEditor(void)
{
//     (void) MessageBeep(0xFFFFFFFFL);
}

void Information2(TEditFile *teditfile, STRPTR body)
{
     EditFile *doc = teditfile->GetEditFile();

     teditfile->MessageBox((char *) body, doc->name, MB_ICONINFORMATION | MESSBOX_FLAGS);
}

//LONG Question(TEditFile *teditfile, STRPTR title, STRPTR body, STRPTR answer)
// returns the number of pressed button: 1, .., N, 0
//{
//     // windows does not have a general requester function!
//}

LONG QuestionOkCancel(TEditFile *teditfile, STRPTR title, STRPTR body)
// returns the number of pressed button: 1 (OK), 0 (CANCEL)
{
     EditFile *doc = teditfile->GetEditFile();

     int resp = teditfile->MessageBox((char *) body, (title) ? title : doc->name, MB_OKCANCEL | MB_ICONQUESTION | MESSBOX_FLAGS);

     if (resp == IDOK)
          return 1;
     else return 0;
}

LONG QuestionYesNo(TEditFile *teditfile, STRPTR title, STRPTR body)
// returns the number of pressed button: 1 (YES), 0 (NO)
{
     EditFile *doc = teditfile->GetEditFile();

     int resp = teditfile->MessageBox((char *) body, (title) ? title : doc->name, MB_YESNO | MB_ICONQUESTION | MESSBOX_FLAGS);

     if (resp == IDYES)
          return 1;
     else return 0;
}

LONG QuestionYesNoCancel(TEditFile *teditfile, STRPTR title, STRPTR body)
// returns the number of pressed button: 1 (YES), 2 (NO), 0 (CANCEL)
{
     EditFile *doc = teditfile->GetEditFile();

     int resp = teditfile->MessageBox((char *) body, (title) ? title : doc->name, MB_YESNOCANCEL | MB_ICONQUESTION | MESSBOX_FLAGS);

     if (resp == IDYES)
          return 1;
     else if (resp == IDNO)
          return 2;
     else return 0;
}

void SetWindowTitle(TEditFile *teditfile, char *title)
{
     TWindow *tw = teditfile->GetParentO();
     EditFile *doc = teditfile->GetEditFile();
     char newname[PATH_LEN + FILENAME + 64];

     if (title == NULL) title = doc->name;
     if (title == doc->name)
     {
          strcpy(newname, XEDITORTITLE);
          strcat(newname, doc->path);
          // strcat(newname, doc->name);
          if (doc->changes)
               strcat(newname, XEDITORCHANGES);
          title = newname;
     }

     if (tw)
     {
//          tw->GetWindowTextTitle();
          if (strlen(title) != tw->GetWindowTextLength())
               tw->SetCaption(title);
          else
          {
               char actname[PATH_LEN + FILENAME + 64];

               (void) tw->GetWindowText(actname, FILENAME);
               if (strcmp(actname, title))
                    tw->SetCaption(title);
          }
     }
}

void ChangePointer(TEditFile *teditfile, char *title, BOOL wait, BOOL blockinput)
{
     if (title)
          SetWindowTitle(teditfile, (char *) title);

     if (wait)
     {
          if (teditfile->IncBusyPointer() <= 1)
               teditfile->SetCursor(NULL, IDC_WAIT);
     }
     else
     {
          if (teditfile->DecBusyPointer() == 0)
               teditfile->SetCursor(NULL, IDC_IBEAM);
     }
}

