//----------------------------------------------------------------------------
// BDEditor 2.2 - simplified ASCII only version
// Copyright (c) 2000, Pavel Cizek
// Compatibility header file - defines analog of AmigaOS functions
//----------------------------------------------------------------------------

// C constants
#define	NOT 		!
#define FOREVER         for(;;)
#define GS(strid)	(strid##_STR)	// Windows do not have localization support

// constants for return values file functions
// general
#define EX_ERROR              -1
#define EX_OK                 0

// open
#define MODE_OLDFILE          0
#define MODE_NEWFILE          1

// ExamineFileMode
#define EX_FM_REGULARFILE     0
#define EX_FM_DIRECTORY       1
#define EX_FM_DEVICE          2
#define EX_FM_DOESNOTEXIST    3

// file related constant
#define BDT_BACKUP_EXT        ".bak"
#define BDT_AUTOSAVE_PREP     "Xplore_Autosave " 
#define BDT_AUTOSAVE_EXT      "~~~"	// ~ is used by Vim and other Unixish editors
#ifdef JAPANESE
/* Modified by D-TEC */
#define BDT_FILEMASK          "گ (*.xpl)|*.xpl|ײ (*.lib)|*.lib|÷ ̧ (*.txt)|*.txt|ׂĂ̧|*.*|"
#else
#define BDT_FILEMASK          "Quantlets (*.xpl)|*.xpl|Libraries (*.lib)|*.lib|Text files (*.txt)|*.txt|All files|*.*|"
#endif
#define BDT_DEFAULT_EXT       "xpl"
#ifdef JAPANESE
/* Modified by D-TEC */
#define XEDITORTITLE          "ҏW޳ - "
#else
#define XEDITORTITLE          "Editor - "
#endif
#define XEDITORCHANGES        " *"

// typedefs
typedef FILE *BPTR;      // BPTR should be defined as the handle of file (i.e., the type that is returned when a file is opened)

// cursor blinking speed in miliseconds
#define IDT_TIMERINVERVAL     100
#define IDT_CURSORTICKS       7
#define IDT_TICKSPERMINUTE    600L



/////////////////////////
// general C functions //
/////////////////////////

inline void clrmem(void *buffer, unsigned bytes)
{
     setmem(buffer, bytes, 0);
}

inline size_t strlen(const UBYTE *s)
{
     return strlen((const char *) s);
}

inline UBYTE *strcpy(UBYTE *s1, UBYTE *s2)
{
     return (UBYTE *) strcpy((char *) s1, (char *) s2);
}

////////////////////////////
// System Amiga functions //
////////////////////////////

#ifndef COMPAT_C

// memory
extern APTR LibAllocPooled( APTR poolHeader, unsigned long memSize );
extern APTR LibCreatePool( unsigned long memFlags, unsigned long puddleSize, unsigned long threshSize );
extern void LibDeletePool( APTR poolHeader );
extern void LibFreePooled( APTR poolHeader, APTR memory, unsigned long memSize );

// files
extern int ExamineFileMode(const char *filename);
extern long ExamineFileSize(const char *filename); 
extern long GetOpenFileSize(BPTR handle);   

extern BPTR Open(const char *filename, ULONG mode);
extern void Close(BPTR file);
extern ULONG Read(BPTR file, UBYTE *buffer, ULONG bytes);
extern ULONG Write(BPTR file, UBYTE *buffer, ULONG bytes);

// DeleteFile function is a part of Windows API - we have to fool it around
#define DeleteFile BDTDeleteFile
extern int BDTDeleteFile(const char *filename); 
extern int Rename(const char *oldname, const char *newname); 

// requesters
extern BOOL BDTFileRequester(TEditFile *teditfile, STRPTR text, STRPTR oktext, STRPTR mask, BOOL save, char *filename);

// misc
extern void CurrentTime(ULONG *secs, ULONG *mics);
extern ULONG LetterUpperCase(ULONG znak);
extern ULONG LetterLowerCase(ULONG znak);
extern ULONG LetterChangeCase(ULONG znak);
extern ULONG ToLower(ULONG znak);
extern ULONG ToUpper(ULONG znak);

///////////////////////
// Windows functions //
///////////////////////

extern void InvalidateLines(TEditFile *teditfile, long lines, unsigned offset, BOOL erase = TRUE);
extern void InvalidateLines(TEditFile *teditfile, long lines = 0);
extern void InvalidateColumns(TEditFile *teditfile, long columns = 0);
extern void ScrollLines(TEditFile *teditfile, long lines);
extern void ScrollColumns(TEditFile *teditfile, long columns);

///////////////////////////
// misc Editor functions //
///////////////////////////

#ifdef CHECK_BUGS
extern void InternalError(const char *);
#endif
extern void AlertMessage(TEditFile *teditfile, STRPTR txt1, STRPTR txt2);
extern void BeepEditor(void);
// do we need to block input?
inline void BlockAllWindows(BOOL block) { }
extern void ChangePointer(TEditFile *teditfile, char *title, BOOL wait, BOOL blockinput);
extern void Information2(TEditFile *teditfile, STRPTR body);
inline void Information(TEditFile *teditfile, STRPTR body)
{
     Information2(teditfile, body);
}
//extern LONG Question(TEditFile *teditfile, STRPTR title, STRPTR body, STRPTR answer);
extern LONG QuestionOkCancel(TEditFile *teditfile, STRPTR title, STRPTR body);
extern LONG QuestionYesNo(TEditFile *teditfile, STRPTR title, STRPTR body);
extern LONG QuestionYesNoCancel(TEditFile *teditfile, STRPTR title, STRPTR body);
//extern STRPTR TextTitleDocWindow(TEditFile *teditfile);
extern void SetWindowTitle(TEditFile *teditfile, char *title);

#endif

