/*
 * BDT-Draw.c       modul s fcemi pro vykreslovani vsech veci u dokumentu
 */

#define BDT_DRAW_C

#include "wed_editor.h"
/*** Modify Start : d-tec  ***/
#include <mbctype.h>
/*** Modify End : d-tec  ***/

// #include "BDT-Defs-Protos.h"



// deklarace fci tohoto modulu
void RedrawVisibleClip(TEditFile *teditfile, Location *start_loc, Location *end_loc, LONG from, LONG to, BOOL text, BOOL cleararound);


// FUNCTIONS FOR HIGHLIGHTING //

// identification constants
#define KEYWORDS    3
#define FUNCTIONS   2
#define OBJECTS     1
#define MACROS      0

#ifndef X4standalone
int GetNames(int type, char ***names, bool only_counts)
{
     return 0;
}
#endif

// string sort function
int wordcmp(const char **s1, const char **s2)
{
     if (*s1 && *s2)
          return strcmp( *s1, *s2 );
     else if (*s1)
          return -1;
     else if (*s2)
          return 1;
     else return 0;
}

int wordcmpsort(const void *elem1, const void *elem2)
{
     return wordcmp((const char **) elem1, (const char **) elem2);
}

int wordcmp(const char *s1, const char *s2)
{
     return wordcmp(&s1, &s2);
}

// initialization of highlighting arrays
// return TRUE, if changes are done
BOOL InitHighlightInfo(void)
{
     extern int GetNames(int type, char ***names, bool only_counts);
     
#ifdef X4standalone
#define NUM_KEYWORDS     16
     static char *keywords[NUM_KEYWORDS] = { "proc", "endp", "if", "else", "endif", "switch", "case", "endsw", "default", "while", "endo", "do", "warning", "error", "break", "until" };
#else
#define NUM_KEYWORDS     (8+5+9+3+3+3)
     static char *keywords[NUM_KEYWORDS] = { 
                                             "char", "int", "float", "double", "short", "long", "unsigned", "void",
                                             "extern", "static", "auto", "register", "const"
                                             "if", "else", "switch", "case", "default", "while", "do", "until", "for", 
                                             "continue", "break", "return",
                                             "sizeof", "typedef", "struct", 
                                             "class", "new", "delete"
                                           };
#endif
          
     char **names, **oldnames;
     ULONG i, numnames, oldnumnames;
     
     BOOL changes = FALSE;

     // keywords
     if (TEditFile::hi_keywords == NULL)
     {
          qsort(keywords, NUM_KEYWORDS, sizeof(char *), wordcmpsort);
          TEditFile::hi_keywords = keywords;
          TEditFile::hi_num_keywords = NUM_KEYWORDS;
          
          changes = TRUE;
     }

     // functions
     if (TEditFile::hi_num_functions != GetNames(FUNCTIONS, &names, TRUE))
     {
          oldnames = TEditFile::hi_functions;
          TEditFile::hi_functions = NULL;

          oldnumnames = TEditFile::hi_num_functions;
          TEditFile::hi_num_functions = 0;

          // get words & sort them
          numnames = GetNames(FUNCTIONS, &names, FALSE);
          if (names)
               qsort(names, numnames, sizeof(char *), wordcmpsort);

          // delete old words
          if (oldnames)
          {
               for(i = 0; i < oldnumnames; i++)
                    if (oldnames[i])
                         delete[] oldnames[i];
               delete[] oldnames;
          }

          TEditFile::hi_num_functions = numnames;
          TEditFile::hi_functions = names;
          
          changes = TRUE;
     }
     
     // libraries
     if (TEditFile::hi_num_libraries != GetNames(MACROS, &names, TRUE))
     {
          oldnames = TEditFile::hi_libraries;
          TEditFile::hi_libraries = NULL;

          oldnumnames = TEditFile::hi_num_libraries;
          TEditFile::hi_num_libraries = 0;
          
          // get words & sort them
          numnames = GetNames(MACROS, &names, FALSE);
          if (names)
               qsort(names, numnames, sizeof(char *), wordcmpsort);

          // delete old words
          if (oldnames)
          {
               for(i = 0; i < oldnumnames; i++)
                    if (oldnames[i])
                         delete[] oldnames[i];
               delete[] oldnames;
          }

          TEditFile::hi_num_libraries = numnames;
          TEditFile::hi_libraries = names;

          changes = TRUE;          
     }

     return changes;
}

// search function - finds out whether the given string is the given string array (see constants above)
BOOL stringstrarray(int strarray, const char *str, char **pi1 = NULL, char **pi2 = NULL)
// if pi1, pi2 are supplied, the final strings corresponding to i1, i2 are returned
{
     long i, i1, i2, numnames;
     int cmp;
     char **names;

     switch (strarray)
     {
          case KEYWORDS:
               numnames = TEditFile::hi_num_keywords;
               names = TEditFile::hi_keywords;
               break;
          case FUNCTIONS:
               numnames = TEditFile::hi_num_functions;
               names = TEditFile::hi_functions;
               break;
          case MACROS:
               numnames = TEditFile::hi_num_libraries;
               names = TEditFile::hi_libraries;
               break;
          default:
               return FALSE;
     }
     if ((names == NULL) || (numnames <= 0))
     {
          if (pi1) *pi1 = NULL;
          if (pi2) *pi2 = NULL;
          return FALSE;
     }

     for(i1 = 0, i2 = numnames - 1; i1 <= i2; )
     {
          i = (i1 + i2) / 2;
          cmp = wordcmp(names[i], str);
          if (cmp == 0)
          {
               if (pi1) *pi1 = names[i];
               if (pi2) *pi2 = names[i];
               return TRUE;
          }
          if (cmp < 0)
               i1 = (i1 < i) ? i : i1 + 1;
          if (cmp > 0)
               i2 = (i2 > i) ? i : i2 - 1;
     }
     if (pi1) *pi1 = names[max_l(0,min_l(i, numnames-1))];
     if (pi2) *pi2 = names[max_l(0,min_l(i, numnames-1))];

     return FALSE;
}

// function that sets check the word and sends back the highlighting regime
ULONG GetHighlightMode(UBYTE *word, BOOL init, ULONG *highlight, BOOL *comment, BOOL *string, 
                       UBYTE *prevchar, UBYTE nextchar = CH_SPACE)
// if init == TRUE, highlighting is initialized (beg of line)
{
     // initialization
     if (init)
     {
          *highlight = EC_HI_DEFAULT;
          *comment = FALSE;
          *string = FALSE;
          *prevchar = CH_EOL;

          return *highlight;
     }

     if ((word == NULL) || (strlen(word) == 0))
          return *highlight;

     // check various alternatives - the string is either alfanumeric or just one character
     
     if (*comment)       // comment
     {
          *highlight = EC_HI_COMMENT;
          *prevchar = word[strlen(word) - 1];
          return *highlight;
     }
     if (*string)        // string
     {
          *highlight = EC_HI_STRING;
          *prevchar = word[strlen(word) - 1];
          if (*word != '"')
               return *highlight;
     }
     
     /*** Modify Start : d-tec  ***/
     if (_ismbslead((char *)word, (char *)word))
          return *highlight;
     /*** Modify End : d-tec  ***/

     // other possibilities
     if (isalpha(*word))           // ? keyword
     {
          if (stringstrarray(KEYWORDS, word))
               *highlight = EC_HI_KEYWORDS;
          else if (stringstrarray(FUNCTIONS, word))
               *highlight = EC_HI_FUNCTIONS;
          else if (stringstrarray(MACROS, word))
               *highlight = EC_HI_LIBRARIES;
          else *highlight = EC_HI_DEFAULT;
     }
     else if (isdigit(*word))      // ? number
     {
          while (*word && isdigit(*word))
               word++;
          
          if (*word)
          {
               if ((*word == 'e') || (*word == 'E'))
                    word++;
               if (*word)
                    if ((*word == '-') || (*word == '+'))
                         word++;
          }
          
          while (*word && isdigit(*word))
               word++;
               
          *highlight = (*word) ? EC_HI_DEFAULT : EC_HI_NUMBER;
     }
     else if (*word == '/')        // ? comment
     {
          if ((*prevchar == '/') || (nextchar == '/'))
          {
               *comment = TRUE;
               *highlight = EC_HI_COMMENT;
          }
     }
     else if (*word == ';')        // comment
     {
          *comment = TRUE;
          *highlight = EC_HI_COMMENT;
     }
     else if (*word == '"')        // string
     {
          *string = NOT *string;
          *highlight = EC_HI_STRING;
     }
     else *highlight = EC_HI_DEFAULT;
     
     // set prev char for next round
     *prevchar = word[strlen(word) - 1];
          
     return *highlight;
}

// version for one character
ULONG GetHighlightMode(UBYTE onechar, BOOL init, ULONG *highlight, BOOL *comment, BOOL *string, 
                       UBYTE *prevchar, UBYTE nextchar = CH_SPACE)
{
     UBYTE buf[2];

     buf[0] = onechar;
     buf[1] = '\0';

     return GetHighlightMode(buf, init, highlight, comment, string, prevchar, nextchar);
}

// NEW FUNCTION FOR WINDOWS DRAWING //

void UpdateWindowInfo(TEditFile *teditfile)
// sets the new dimensions of the window
{
     EditFile *doc = teditfile->GetEditFile();
     TRect rect;

     rect = teditfile->GetClientRect();
     doc->winheight = rect.Height();
     if (doc->winheight < 0) doc->winheight = 0;
     doc->winwidth = rect.Width();
     if (doc->winwidth < 0) doc->winwidth = 0;
}

void UpdateFontInfo(TEditFile *teditfile)
// sets window character dimensions
// font dims has to be known
{
     EditFile *doc = teditfile->GetEditFile();

     // if sizes are not known - nothing to do
     if ((doc->fontwidth == 0) || (doc->fontheight == 0))
          return;

     if ((doc->winwidth > 0) && (doc->fontwidth > 0))
          doc->wincwidth = (ULONG) doc->winwidth / (ULONG) doc->fontwidth;
     else doc->wincwidth = 0;
     if ((doc->winheight > 0) && (doc->fontheight > 0))
          doc->wincheight = (ULONG) doc->winheight / (ULONG) doc->fontheight;
     else doc->wincheight = 0;
}

void UpdateFontInfo(TEditFile *teditfile, TDC& dc)
// sets the dimensions of the font used - done only if it is not already set
// the same for window character dimensions
// it assumes that the font is already selected to DC
{
     EditFile *doc = teditfile->GetEditFile();

     // get font metrics if sizes are not known
     if ((doc->fontwidth == 0) || (doc->fontheight == 0))
     {
	  TEXTMETRIC fontmetrics;

	  dc.GetTextMetrics(fontmetrics);
	  doc->fontheight = fontmetrics.tmHeight;
          if (doc->fontheight < 0) doc->fontheight = 0;
	  (void) dc.GetCharWidth(' ', ' ', &doc->fontwidth);
          if (doc->fontwidth < 0) doc->fontwidth = 0;
     }

     if ((doc->winwidth > 0) && (doc->fontwidth > 0))
          doc->wincwidth = (ULONG) doc->winwidth / (ULONG) doc->fontwidth;
     else doc->wincwidth = 0;
     if ((doc->winheight > 0) && (doc->fontheight > 0))
          doc->wincheight = (ULONG) doc->winheight / (ULONG) doc->fontheight;
     else doc->wincheight = 0;
}

void DrawCursor(TEditFile *teditfile, TDC& dc, ULONG curpos)
// draw cursor; curpos gives the column in which is the cursor
{
     EditFile *doc = teditfile->GetEditFile();
     Location *cursor = &doc->cursor;
     ULONG fh = doc->fontheight, fw = doc->fontwidth;
     int x, y;
     BOOL insert = GETGLOBALPREFS->insert;

     doc->last_cursor_pos = curpos;
     doc->ticks = 0;

     if ((fw == 0) || (fh == 0)) return;
     if (cursor->line_num < doc->display_begin.line_num) return;
     if (cursor->line_num > doc->display_end.line_num) return;
     if (curpos < doc->disp_offset) return;

     x = fw * (curpos - doc->disp_offset);
     y = fh * (cursor->line_num - doc->display_begin.line_num);

/*** Modify Start : d-tec ***/
     if (_ismbslead(cursor->block->memory, cursor->block->memory + cursor->block_off)) fw=fw*2;
/*** Modify End : d-tec  ***/

     dc.InvertRect(x, y, x + ((insert) ? 3 : fw), y + fh /*- 1*/);
}

// block modus
#define BLOCKMOD_NONE    0
#define BLOCKMOD_BEGIN   1
#define BLOCKMOD_END     2
#define BLOCKMOD_ALL     4

// TextOuts - prints the given text, variables correspond to those in DrawLine
void SimpleTextOut(TDC &dc, int x, int y, UBYTE *text, ULONG highlight, BOOL block, int count = -1)
{
     GlobalPrefs *prefs = GETGLOBALPREFS;
     
     // select colors
     dc.SetTextColor(TColor::SysWindowText);
     dc.SetBkColor(TColor::SysWindow);

     if (block)
     {
          dc.SetTextColor(TColor::SysHighlightText);
          dc.SetBkColor(TColor::SysHighlight);
     }
     else if (highlight != EC_HI_DEFAULT)
     {
          if (prefs->highlight_foreground[highlight] != COLOR_NOT_DEFINED)
               dc.SetTextColor(TColor(prefs->highlight_foreground[highlight]));
          if (prefs->highlight_background[highlight] != COLOR_NOT_DEFINED)
               dc.SetBkColor(TColor(prefs->highlight_background[highlight]));
     }

     // print text
     dc.TextOut(x, y, (char *) text, count);

     // select default colors
     dc.SetTextColor(TColor::SysWindowText);
     dc.SetBkColor(TColor::SysWindow);
}

void ComplexTextOut(TDC &dc, long x, long y, UBYTE *text, ULONG *chartext, ULONG highlight, 
                    BOOL block, BOOL blockmod, ULONG blockstart, ULONG blockend,
                    ULONG charsoff, ULONG bufchars, ULONG fw)
{
     int count, charcount;

     // determine what is in block
     if (blockmod == BLOCKMOD_NONE)
     {
          SimpleTextOut(dc, x, y, text, highlight, FALSE);   return;
     }
     else if (blockmod & BLOCKMOD_ALL)
     {
          SimpleTextOut(dc, x, y, text, highlight, TRUE);    return;
     }
     else if (blockmod & BLOCKMOD_BEGIN)
     {
          if (blockmod & BLOCKMOD_END)
          {
               if ((charsoff >= blockend) || (charsoff + bufchars < blockstart))
               {
                    SimpleTextOut(dc, x, y, text, highlight, FALSE);  return;
               }
               if (charsoff < blockstart)
               {
                    charcount = (charsoff + bufchars <= blockstart) ? -1 : blockstart - charsoff;
                    count = (charcount >= 0) ? chartext[charcount] - *chartext : -1;
                    SimpleTextOut(dc, x, y, text, highlight, FALSE, count);
                    if (count < 0)
                         return;
                    text += count;	chartext += charcount;
                    x += fw * count;
               }
               // now, charsoff + bufchars > blockstart and charsoff < blockend
               charcount = (charsoff + bufchars <= blockend) ? -1 : blockend - max_ul(charsoff, blockstart);
               count = (charcount >= 0) ? chartext[charcount] - *chartext : -1;
               SimpleTextOut(dc, x, y, text, highlight, TRUE, count);
               if (count < 0)
                    return;
               text += count;	   chartext += charcount;
               x += fw * count;
               // now, charsoff + bufchars > blockend
               SimpleTextOut(dc, x, y, text, highlight, FALSE);
               return;
          }
          else
          {
               if (charsoff < blockstart)
               {
                    charcount = (charsoff + bufchars <= blockstart) ? -1 : blockstart - charsoff;
                    count = (charcount >= 0) ? chartext[charcount] - *chartext : -1;
                    SimpleTextOut(dc, x, y, text, highlight, FALSE, count);
                    if (count < 0)
                         return;
                    text += count;	chartext += charcount;
                    x += fw * count;
               }
               // now, charsoff + bufchars > blockstart
               SimpleTextOut(dc, x, y, text, highlight, TRUE);
               return;
          }
     }
     else
     {
          if (blockmod & BLOCKMOD_END)
          {
               if (charsoff < blockend)
               {
                    charcount = (charsoff + bufchars <= blockend) ? -1 : blockend - charsoff;
                    count = (charcount >= 0) ? chartext[charcount] - *chartext : -1;
                    SimpleTextOut(dc, x, y, text, highlight, TRUE, count);
                    if (count < 0)
                         return;
                    text += count;	chartext += charcount;
                    x += fw * count;
               }
               // now, charsoff + bufchars > blockend
               SimpleTextOut(dc, x, y, text, highlight, FALSE);
               return;
          }
          else
          {
               SimpleTextOut(dc, x, y, text, highlight, FALSE);
               return;
          }
     }

     SimpleTextOut(dc, x, y, text, highlight, FALSE);
}

void DrawLine(TEditFile *teditfile, TDC& dc, Location *loc, LONG drawfrom, LONG drawto, BOOL block, BOOL text, BOOL cursor)
// draw given line
// font has to be already selected!
// block - block is drawn
// text - text is drawn
// cursor - cursor is drawn
{
     EditFile *doc = teditfile->GetEditFile();

     MemoryBlock *mb = loc->block, *mbcur = doc->cursor.block, *mbpom;
     ULONG mboff = loc->block_off, mbcuroff = doc->cursor.block_off, mboffpom;

     int wh = doc->winheight, ww = doc->winwidth;
     ULONG fh = doc->fontheight, fw = doc->fontwidth;
     ULONG wch = doc->wincheight, wcw = doc->wincwidth;
     ULONG lh = (loc->line_num - doc->display_begin.line_num) * fh;
     ULONG dispoff = doc->disp_offset;
     ULONG len;
     long x, y;

     BOOL blockmod;
     ULONG blockstart, blockend;

     ULONG scrpos, bufpos, curpos = 0xFFFFFFFFL, charsoff, bufchars, pomoff;
     UWORD wordcolor, tabsize = GETGLOBALPREFS->tab_size;
     UBYTE wordbuf[MAXWORDLEN + 4], *line, *prtstr, *pomptr;
     ULONG wordcharbuf[MAXWORDLEN + 4], *charstr; // for every character contains its position in wordbuf

     ULONG startcol, endcol;

     // variables for highlighting
     ULONG highlight, oldhighlight;
     BOOL comment, string, oldcomment, oldstring;
     UBYTE prevchar, oldprevchar;

     // zero dimension - nothing done
     if ( (ww == 0) || (wh == 0) || (wcw == 0) || (wch == 0) || (fw == 0) || (fh == 0) )
          return;
     // if line is not between disp_beg / end - nothing done
#ifdef CHECK_BUGS
     if (doc->display_begin.line_num > doc->display_end.line_num) InternalError("MakeCursorVisible - line_nums");
#endif
     if ((loc->line_num < doc->display_begin.line_num) || (loc->line_num > doc->display_end.line_num))
          return; 

// determine drawing bounds - from which to which column is refresh done
     if (drawfrom > drawto) drawfrom = drawto = 0;     // security settings
     if (drawfrom < 0) drawfrom = 0;
     if (drawto > ww) drawto = ww;
     if ( (drawfrom > ww) || (drawto < 0) ) return;    // nothing to do
     
     startcol = dispoff;
     if (drawfrom > 0)
          startcol = min_ul(drawfrom / fw, dispoff + wcw);

     endcol = dispoff + wcw;
     if (drawto < ww)
          endcol = max_ul(drawto / fw + 1, dispoff + wcw);
          
// do initialization of block variables
     if (block && doc->block)
     {
          blockmod = BLOCKMOD_NONE;
          switch (CompareParaLinePos(doc->block_begin.para, loc->para))
          {
               case -1:
                    switch (CompareParaLinePos(doc->block_end.para, loc->para))
                    {
                         case -1:
                              blockmod = BLOCKMOD_NONE;
                              break;
                         case 0:
                              blockmod |= BLOCKMOD_END;
                              blockend = doc->block_end.chars_off;
                              break;
                         case 1:
                              blockmod |= BLOCKMOD_ALL;
                              break;
                    }
                    break;
               case 0:
                    blockmod |= BLOCKMOD_BEGIN;   blockstart = doc->block_begin.chars_off;
                    if (0 == CompareParaLinePos(doc->block_end.para, loc->para))
                    {
                         blockmod |= BLOCKMOD_END;     blockend = doc->block_end.chars_off;
                    }
                    break; 
               case 1:
                    blockmod = BLOCKMOD_NONE;     
                    break;
          }
          
     }
     else blockmod = BLOCKMOD_NONE;
     
// notes:
// 	if not text --> highlighting and text output is not done!

// init highlighting function
     (void) GetHighlightMode((UBYTE *) NULL, TRUE, &highlight, &comment, &string, &prevchar);

     for( charsoff = 0, scrpos = 0, bufpos = 0, bufchars = 0, line = GetPosMB(mb, mboff); 
          !ENDLINE(*line) && (scrpos < dispoff + wcw); 
          bufpos = 0, bufchars = 0, line = GetPosMB(mb, mboff)
        )
     // always: skip white spaces and read word
     // check highlighting and print the word
     // check and print other separators
     // finish at the end of paragraph = once we reach CH_EOL
     {
          // skip whitespaces
          bufpos = 0;    bufchars = 0;
          for(line = GetPosMB(mb, mboff); WHITESPACE(*line) && (bufpos + tabsize < MAXWORDLEN); line = IncrementPosMB(&mb, &mboff))
          {
               if (cursor)
                    if ((mb == mbcur) && (mboff == mbcuroff))
                         curpos = scrpos + bufpos;
               wordcharbuf[bufchars++] = bufpos;
               if (*line == CH_TAB)
                    bufpos += tabsize - (scrpos + bufpos) % tabsize;
               else bufpos++;
          }
          if (bufpos > 0)
               setmem(wordbuf, bufpos, CH_SPACE);
          wordcharbuf[bufchars] = bufpos;
          wordbuf[bufpos] = '\0';

          if (bufpos > 0)
          {
               // output
               len = bufpos;
               if ((scrpos < endcol) && (scrpos + len > startcol))
//               if ((scrpos < dispoff + wcw) && (scrpos + len > dispoff))
               {
//                    x = pomoff = max_ul(scrpos, dispoff);
//                    prtstr = (x > scrpos) ? wordbuf + (x - scrpos) : wordbuf;
//                    charstr = (x > scrpos) ? wordcharbuf + (x - scrpos) : wordcharbuf;
//                    prtstr = wordbuf;
//                    charstr = wordcharbuf;
                    x = ((long) scrpos) - ((long) dispoff);
                    if (text) ComplexTextOut(dc, x * (long) fw, (long) lh, wordbuf, wordcharbuf, 
                                             EC_HI_DEFAULT /*highlight*/, // white spaces are not highlighted
                                             block, blockmod, blockstart, blockend, charsoff, bufchars, fw);
               }
               scrpos += len;      charsoff += bufchars;
          }

          // read next word and print it
          bufpos = 0;    bufchars = 0;
          for(line = GetPosMB(mb, mboff); isalnum(*line) && (bufpos < MAXWORDLEN); line = IncrementPosMB(&mb, &mboff))
          {
               if (cursor)
                    if ((mb == mbcur) && (mboff == mbcuroff))
                         curpos = scrpos + bufpos;
               wordcharbuf[bufchars++] = bufpos;
               wordbuf[bufpos++] = *line;
/*** Modify Start : d-tec  ***/
               if (_ismbslead(mb->memory, mb->memory + mboff)) {
                    wordcharbuf[bufchars++] = bufpos;
                    wordbuf[bufpos++] = *(line + 1);
               }
/*** Modify End : d-tec  ***/
          }
          wordcharbuf[bufchars] = bufpos;
          wordbuf[bufpos] = '\0';

          if (bufpos > 0)
          {

               // highlighting should be done here - send the word

               // output
               len = bufpos;  // strlen(wordbuf);
               if ((scrpos < endcol) && (scrpos + len > startcol))
//               if ((scrpos < dispoff + wcw) && (scrpos + len > dispoff))
               {
//                    x = max_ul(scrpos, dispoff);
//                    prtstr = (x > scrpos) ? wordbuf + (x - scrpos) : wordbuf;
//                    charstr = (x > scrpos) ? wordcharbuf + (x - scrpos) : wordcharbuf;
//                    prtstr = wordbuf;
//                    charstr = wordcharbuf;
                    x = ((long) scrpos) - ((long) dispoff);
                    if (text) ComplexTextOut(dc, x * (long) fw, (long) lh, wordbuf, wordcharbuf, 
                                             GetHighlightMode(wordbuf, FALSE, &highlight, &comment, &string, &prevchar),
                                             block, blockmod, blockstart, blockend, charsoff, bufchars, fw);
               }
               scrpos += len; charsoff += bufchars;
          }

          // read non-white non-letter characters - currently has to be printed one by one
          bufpos = 0;    bufchars = 0;
          for( 
               line = GetPosMB(mb, mboff), 
               oldhighlight = highlight, oldcomment = comment, oldstring = string, oldprevchar = prevchar,
               mbpom = mb, mboffpom = mboff;
               
               !isalnum(*line) && !WHITESPACE(*line) && !ENDLINE(*line) && (bufpos < MAXWORDLEN) &&
               ( (oldhighlight == GetHighlightMode(*line, FALSE, &highlight, &comment, &string, &prevchar, (pomptr = IncrementPosMB(&mbpom, &mboffpom)) ? *pomptr : CH_SPACE))
                 || (bufpos == 0) );
               
               line = IncrementPosMB(&mb, &mboff),
               oldhighlight = highlight, oldcomment = comment, oldstring = string, oldprevchar = prevchar,
               mbpom = mb, mboffpom = mboff
             )
          {
               if (cursor)
                    if ((mb == mbcur) && (mboff == mbcuroff))
                         curpos = scrpos + bufpos;
               wordcharbuf[bufchars++] = bufpos;
               wordbuf[bufpos++] = *line;
/*** Modify Start : d-tec  ***/
               if (_ismbslead(mb->memory, mb->memory + mboff)) {
                    wordcharbuf[bufchars++] = bufpos;
                    wordbuf[bufpos++] = *(line + 1);
               }
/*** Modify End : d-tec  ***/
          }
          wordcharbuf[bufchars] = bufpos;
          wordbuf[bufpos] = '\0';
          
          // restore correct highlighting information
          highlight = oldhighlight;
          comment = oldcomment;
          string = oldstring;
          prevchar = oldprevchar;

          if (bufpos > 0)
          {

               // highlighting should be done here - send the word

               // output
               len = bufpos; // strlen(wordbuf);
               if ((scrpos < endcol) && (scrpos + len > startcol))
//               if ((scrpos < dispoff + wcw) && (scrpos + len > dispoff))
               {
//                    x = max_ul(scrpos, dispoff);
//                    prtstr = (x > scrpos) ? wordbuf + (x - scrpos) : wordbuf;
//                    charstr = (x > scrpos) ? wordcharbuf + (x - scrpos) : wordcharbuf;
//                    prtstr = wordbuf;
//                    charstr = wordcharbuf;
                    x = ((long) scrpos) - ((long) dispoff);
                    if (text) ComplexTextOut(dc, x * (long) fw, (long) lh, wordbuf, wordcharbuf, highlight,
                                             block, blockmod, blockstart, blockend, charsoff, bufchars, fw);
               }
               scrpos += len;      charsoff += bufchars;
          }
     }
     // check cursor - it can be at the end of line
     if (ENDLINE(*line))
     {
          if (cursor)
               if ((mb == mbcur) && (mboff == mbcuroff))
                    curpos = scrpos;

          // draw end of paragraph symbol (currently space) + clear to the end of line
          if (scrpos < dispoff + wcw)
          {
               TBrush brush( ((blockmod & BLOCKMOD_ALL) || (blockmod & BLOCKMOD_BEGIN) && NOT (blockmod & BLOCKMOD_END)) ? TColor::SysHighlight : TColor::SysWindow);

               x = (long) (scrpos >= dispoff) ? (scrpos - dispoff) * fw : 0;
               dc.FillRect(x, lh, ww, lh + fh, brush);
          }
     }
     // PC 20020420: save the expanded length of line
     loc->para->explength = scrpos + 1;

     // draw cursor if it was found
     if (cursor && (curpos != 0xFFFFFFFFL))
          DrawCursor(teditfile, dc, curpos);
}

// PC 20020420 - returns whether scrollers needs to be updated
BOOL RedrawEditorRect(TEditFile *teditfile, TDC& dc, TRect& rect, BOOL cursor)
// redraw all lines that are within (touch) the given rectangle
// if window dimensions are zero, nothing happens
// teditfile - editor file
// dc - drawing context
// rect - area to be refreshed
// cursor - should cursor be drawn
// font has to be already selected!
{
     EditFile *doc = teditfile->GetEditFile();
     struct Location line = doc->display_begin;

     int wh = doc->winheight, ww = doc->winwidth;
     ULONG fh = doc->fontheight, fw = doc->fontwidth;
     ULONG wch = fh * doc->wincheight, wcw = fw * doc->wincwidth;

     ULONG num = 0, pomlen;
     BOOL konec = FALSE, drawcursor = FALSE, doupdate = FALSE;

     // zero dimension - nothing done
     if ( (ww == 0) || (wh == 0) || (wcw == 0) || (wch == 0) || (fw == 0) || (fh == 0) )
          return(doupdate);

     while (NOT konec)
     {
       /* vlastni vykresleni prislusneho radku */
          if ((rect.Top() /* + 1 */ < (long) ((num + 1) * fh)) && ((long) (num * fh) < rect.Bottom()))
          {
            pomlen = line.para->explength;
	         DrawLine(teditfile, dc, &line, rect.Left(), rect.Right(), doc->block, TRUE, cursor);
            // PC 20020420 - check whether the length has changed
            if ((line.para->explength > line.para->length) && (line.para->explength != pomlen))
              doupdate = TRUE;
          }
       /* test, zda je vsechno prekresleno */
          konec = (line.para == doc->display_end.para) ? TRUE : FALSE;
       /* test, zda jsme prejeli kurzor */
//       	  if (cursor)
//                if (line.para == doc->cursor.para) drawcursor = TRUE;
       /* nalezeni dalsiho radku */
          if (NOT konec) GetNextPos(teditfile, &line, TRUE, 0);
            num++;
     } // end while (NOT konec)

     // if we did all lines in rect and the bottom one is above rect, we have to fill the rest
     if ((long) (num * fh) < rect.Bottom())
     {
          TBrush brush(TColor::SysWindow);

          dc.FillRect(rect.Left(), num * fh, rect.Right(), rect.Bottom(), brush);
     }

     return(doupdate);
}



/* pomocna fce pro spravne nastaveni display_#? odpovidajici akt.
 * zobrazovanemu vyrezu dokumetu; check - argument pro GetLineOnHPos */
/* predpoklada se, ze cursor je spravne nastaven */
/* smer == 0 - odvodime od pocatku stranky, na niz se nachazime,
 *         1 - vychazime od aktualniho display_begin
 *        -1 - vychazime od aktualniho display_begin, ale smerem vzad ! */
void SetDisplayBegEnd(TEditFile *teditfile, BOOL check, long smer)
//void SetDisplayBegEnd(EditFile *doc, BOOL check, WORD smer, Page *page1)
//>>
{    
     EditFile *doc = teditfile->GetEditFile();

// right now, there is just a simple treatment
//   if smer is 0 ... just disp_end is set, disp_beg is untouched
//              -1/1 ... disp_beg = cursor / disp_end = cursor and the other disp is set correspondingly

     ULONG i;
     
     switch (smer)
     {
          case 0:
               doc->display_end = doc->display_begin;
               smer = -1;
               break;

          case -1:
               doc->display_begin = doc->cursor;
               (void) GetPrevPos(teditfile, &doc->display_begin, FALSE, 0);
               break;

          case 1:
               doc->display_end = doc->cursor;
               (void) GetPrevPos(teditfile, &doc->display_end, FALSE, 0);
               break;
     }

     // set the other display
     if (smer < 0)  // set disp_end
     {
          doc->display_end = doc->display_begin;
          for (i = 1; i < doc->wincheight; i++)
               if (NOT GetNextPos(teditfile, &doc->display_end, TRUE, 0)) break;
     }
     if (smer > 0)  // set disp_beg
     {
          doc->display_begin = doc->display_end;
          for (i = 1; i < doc->wincheight; i++)
          {
               if (NOT GetPrevPos(teditfile, &doc->display_begin, TRUE, 0)) break;
	       if (NOT GetPrevPos(teditfile, &doc->display_begin, FALSE, 0)) break;
          }
     }
}
//<<



// FCE PRO PREKRESLENI OBRAZOVKY //

/* nasledujici funkce zkontroluje, zda je kurzor ve viditelnem vyrezu;
 * pokud ne, tak zajisti takovy pohled na dokument, aby byl videt */
// refresh_rooler - obnoveni pravitka se zarazkami
// menu_refresh - kody pro RedrawMenu - co se ma obnovit
// gadget_refresh - kody pro RedrawGadgets - co se ma obnovit
BOOL MakeCursorVisible(TEditFile *teditfile, ULONG menu_refresh, ULONG gadget_refresh, BOOL refresh_rooler)
//>>
{    
     EditFile *doc = teditfile->GetEditFile();

//     class TEditorFile *twin = doc->teditorfile;
//     struct RastPort *rp = doc->rp;
     Location beg = doc->display_begin, end = doc->display_end, cur = doc->cursor;
//     Paragraph *begdisp_para = doc->display_begin.para, *enddisp_para = doc->display_end.para, *cursor_para = doc->cursor.para;
     Paragraph *para;
//     PageSize *pgs = &doc->pagesize;
//     Page *page = doc->display_begin.page;
  // beg.page je vzdy na strance, kde je soupatko!!
     BOOL docmove = FALSE;
//     BOOL page_ok = (beg.page == cur.page) ? TRUE : FALSE, docmove = FALSE;
     ULONG disp_dist = 0, curpos;
//     LONG xmin, ymin, xmax, ymax, dx, dy;
//     MUNIT oldte = doc->te_dclip, oldle = doc->le_dclip, oldw = doc->w_dclip, oldh = doc->h_dclip;
//     MUNIT max_te = pgs->height - doc->h_dclip, max_le = pgs->width - doc->w_dclip;
//     MUNIT le_cursor, le_cur, pom_cur1, pom_cur2, pom_ruler;
//     WORD lte = UnitToPointNoCheck(doc, cur.page_off, FALSE);
//     WORD lbe = UnitToPointNoCheck(doc, cur.page_off + cur.line->height, FALSE);
//     WORD lle, lre;
     long smer = 1;

// compute disp_dist
#ifdef CHECK_BUGS
     if (beg.line_num > end.line_num) InternalError("MakeCursorVisible - line_nums");
#endif
     smer = 0;
     disp_dist = 0;
     if (cur.line_num < beg.line_num)
     {
          smer = -1;     disp_dist = beg.line_num - cur.line_num;
     }
     else if (cur.line_num > end.line_num)
     {
          smer =  1;     disp_dist = cur.line_num - end.line_num;
     }
     else
     {
          smer =  0;     disp_dist = 0;
     }

/* original version without num_line
     // is it between display_begin and display_end?
     for(para = begdisp_para; para && (para != enddisp_para); para = para->next)
          if (para == cursor_para)
          {    smer = 0; // no move
               disp_dist = 0;
               break;
          }     
     
     // compute disp_dist
     if (smer)
     {
          for(para = cursor_para, disp_dist = 0; para && (para != begdisp_para); para = para->next, disp_dist++);
          if (para == NULL)
          {
               for(para = enddisp_para, disp_dist = 0; para && (para != cursor_para); para = para->next, disp_dist++);
               smer = 1;      // down
          }
          else smer = -1;     // up
     }
#ifdef CHECK_BUGS
     if (para == NULL) InternalError("MakeCursorVisible - disp_dist")
#endif
*/

// vertical cursor check
     
     if ( doc->wincheight > 0 )
     if ( smer || disp_dist )
//     if ( NOT page_ok || (disp_dist != 0))
     if ( 2*disp_dist >= doc->wincheight )
//     if ( NOT page_ok || (5L*disp_dist >= 4L*oldh) )
     {
          docmove = TRUE;
//	  NumMouseClicks = 0;
       // smazat se musi ve stare poloze ... (aby to vyslo dobre)
//	  DrawCursor(teditfile, TRUE, 0, 0, 0, 0, FALSE, TRUE);
//       // zde staci posunout te_dclip + nastavit display_begin a _end
//	  doc->te_dclip = min_ul(cur.page_off, max_te);
//	  pom_cur1 = (5L*cur.line->height < doc->h_dclip) ? cur.line->height : 0;
//	  if (doc->te_dclip < max_te)   // jsme nejnize jak to jde
//	       doc->te_dclip -= (doc->te_dclip > pom_cur1)  ? pom_cur1 : doc->te_dclip;
//	  doc->te_dclip = TEDCLIP_ROUND(doc->te_dclip);
//	  doc->te_clip = doc->te_dclip;
//       // nove display_#?
	  SetDisplayBegEnd(teditfile, TRUE, smer);
       // prezobrazeni vyrezu dokumentu ...
	  doc->cursor_checked = TRUE;
//	  RedrawNewSizeWindow(doc, TRUE, FALSE, TRUE);
          InvalidateLines(teditfile);
     }
     else // pripad, kdy budeme odrolovavat o kousek
     {
          docmove = TRUE;
//	  NumMouseClicks = 0;
//	  DrawCursor(teditfile, TRUE, 0, 0, 0, 0, FALSE, TRUE);
//	  smer = (oldte >= cur.page_off) ? -1 : 1;
//       // urceni noveho vyrezu
//	  if (smer == -1)
//	  {    doc->te_dclip = min_ul(cur.page_off, max_te);
//	       if (doc->te_dclip < max_te)
//		    doc->te_dclip -= (doc->te_dclip > INCH_4_MUNIT) ? INCH_4_MUNIT : doc->te_dclip;
//	  }
//	  else
//	  {    MUNIT h = cur.line->height;
//	       doc->te_dclip = cur.page_off + h - doc->h_dclip;
//	       if ((5L*h < doc->h_dclip) && (pgs->height - 2*h > doc->te_dclip))
//		    doc->te_dclip += 2*h;
//	       if (doc->te_dclip > max_te) doc->te_dclip = max_te;
//	  }
//	  doc->te_dclip = TEDCLIP_ROUND(doc->te_dclip);
//
//	  if (0 == UPLengthZoom(doc, (MUNIT) ((oldte >= doc->te_dclip) ? oldte - doc->te_dclip : doc->te_dclip - oldte), FALSE))
//	  {    MUNIT pom = PULengthZoom(doc, 2, FALSE);
//	       if (smer == -1)
//		    doc->te_dclip -= (doc->te_dclip > pom) ? pom : doc->te_dclip;
//	       else doc->te_dclip += pom;
//	  }
////          doc->te_clip = CorrectNumeric1(doc, &doc->te_dclip);
//	  doc->te_dclip = TEDCLIP_ROUND(doc->te_dclip);
//	  doc->te_clip = doc->te_dclip;
//       // nove display_#?
	  SetDisplayBegEnd(teditfile, TRUE, smer);
       // spravne vyobrazeni vyrezu dokumentu ...
	       // odrolovani stavajiciho obsahu
//	  dx = 0;
//          dy = UnitToPointNoCheck(doc, doc->te_dclip, FALSE);
//          dy -= UnitToPointNoCheck(doc, oldte, FALSE);
//	  dy = smer * UPLengthZoom(doc, (MUNIT) ((oldte >= doc->te_dclip) ? oldte - doc->te_dclip : doc->te_dclip - oldte), FALSE);
//	  xmin = xmax = doc->le_dwinclip;    xmax += doc->w_dwinclip;
//	  if (doc->ruler_vert) xmin -= VERT_RULER_SIZE;
//	  ymin = ymax = doc->te_dwinclip;    ymax += doc->h_dwinclip;
////          if (smer == -1) ymax += dy; else ymin += dy;
//	  SetBPen(rp, (ULONG) doc->display.paper_color);
//	  MyScrollRaster(doc, dx, dy, xmin, ymin, xmax, ymax, (smer == -1) ? &beg : &doc->display_begin, (smer == -1) ? &doc->display_end : &end);
          ScrollLines(teditfile, smer * disp_dist);
	       // vykresleni pravitka a noveho pozadi ...
//	  if (doc->ruler_vert) xmin += VERT_RULER_SIZE;
//	  if (smer == -1)
//	  {    BGFillDocWin(doc, (WORD) xmin, (WORD) ymin, (WORD) (xmax - xmin), (WORD) (-1 * dy), TRUE, TRUE);
//	       DrawRulers1(doc, 0, 0, doc->te_dclip, oldte, FALSE);
//	  }
//	  else
//	  {    BGFillDocWin(doc, (WORD) xmin, (WORD) (ymax - dy), (WORD) (xmax - xmin), (WORD) dy, TRUE, TRUE);
//	    // takto se to musi udelat, nebot jinak pri posunu umazeme konce cislic u predchozich jednotek
//	       pom_ruler = ((oldte + oldh) / doc->main_coef) * doc->main_coef;
//	       DrawRulers1(doc, 0, 0, max_ul(pom_ruler, doc->te_dclip), doc->te_dclip + doc->h_dclip, FALSE);
//	  }
	       // vykresleni novych radku 
	  doc->cursor_checked = TRUE;
////          if (NOT doc->block) doc->clear_before_writing = FALSE;
//	  if (smer == -1)
//	       RedrawVisibleClip(doc, &doc->display_begin, &beg, 0, 0, TRUE, FALSE);
//	  else RedrawVisibleClip(doc, &end, &doc->display_end, 0, 0, TRUE, FALSE);
////          if (NOT doc->block) doc->clear_before_writing = TRUE;
//	       // pokud se nahodou nevykreslil kurzor - nutno napravit
//	  if (NOT doc->old_ok)
//	       DrawLine(doc, &doc->cursorline, 0, NULL, 0, 0, 0, doc->block, TRUE, FALSE, TRUE, FALSE, TRUE);
     }
//     if (NOT page_ok) doc->act_page = doc->display_begin.page->num;

// horizontal cursor check

// compute disp_dist
     smer = 0;
     disp_dist = 0;
     curpos = GetCursorColumn(teditfile, &cur);
     if (curpos < doc->disp_offset)
     {
          smer = -1;
          disp_dist = doc->disp_offset - curpos;
     }
     if (curpos >= doc->disp_offset + doc->wincwidth)
     {
          smer =  1;
          disp_dist = curpos - (doc->disp_offset + doc->wincwidth) + 1;
     }

     if ( doc->wincwidth > 0 )
     if (smer || disp_dist)
     if (3*disp_dist >= doc->wincwidth)
     {
          docmove = TRUE;
//	  NumMouseClicks = 0;
//	  DrawCursor(teditfile, TRUE, 0, 0, 0, 0, FALSE, TRUE);
//       // zde staci posunout le_dclip
//	  doc->le_dclip = (le_cur > oldw/2) ? le_cur - oldw/2 : 0;
//	  if (doc->le_dclip > max_le) doc->le_dclip = max_le;
//	  doc->le_clip = doc->le_dclip;
       // change the disp_offset
          doc->disp_offset += smer * disp_dist;
       // prezobrazeni vyrezu dokumentu ...
	  doc->cursor_checked = TRUE;
//	  RedrawNewSizeWindow(doc, TRUE, FALSE, TRUE);
          InvalidateColumns(teditfile);
     }
     else
     {
          docmove = TRUE;
//	  NumMouseClicks = 0;
//	  DrawCursor(teditfile, TRUE, 0, 0, 0, 0, FALSE, TRUE);
//	  disp_dist *= 5;     disp_dist /= 4;     disp_dist += INCH_4_MUNIT;
//          if (disp_dist < 3) disp_dist = 3;  // should we do it? to minimize redrawing
//#ifdef CHECK_BUGS
//	  if (MAX_DOC_SIZE - oldle < oldw) InternalError("MakeCursorVisible 2");
//#endif
//       // vypocet polohy
//	  if (smer == -1)
//	       doc->le_dclip = (disp_dist > oldle) ? 0 : oldle - disp_dist;
//	  else doc->le_dclip = (disp_dist < MAX_DOC_SIZE - oldle) ? oldle + (MUNIT) disp_dist : max_le;
//	  if (doc->le_dclip > max_le) doc->le_dclip = max_le;
//	  doc->le_clip = doc->le_dclip;
//       // spravne vyobrazeni vyrezu dokumentu ...
//	       // odrolovani stavajiciho obsahu
//	  dx = smer * UPLengthZoom(doc, (MUNIT) ((doc->le_dclip > oldle) ? doc->le_dclip - oldle : oldle - doc->le_dclip), TRUE);
//	  dy = 0;
//	  xmin = xmax = doc->le_dwinclip;    xmax += doc->w_dwinclip - 1;
////          if (smer == -1) xmax += dx; else xmin += dx;
//	  ymin = ymax = doc->te_dwinclip;    ymax += doc->h_dwinclip - 1;
//	  if (doc->ruler_horiz) ymin -= HORIZ_RULER_SIZE + MARGIN_TAB_SIZE;
//	  SetBPen(rp, (ULONG) doc->display.paper_color);
       // change the disp_offset
          doc->disp_offset += smer * disp_dist;
          // scroll window
          ScrollColumns(teditfile, smer * disp_dist);
//	  MyScrollRaster(doc, dx, dy, xmin, ymin, xmax, ymax, NULL, NULL);
//	  if (doc->ruler_horiz) ymin += HORIZ_RULER_SIZE + MARGIN_TAB_SIZE;
//	       // vykresleni pravitka a pozadi ...
//	  if (smer == -1)
//	  {    BGFillDocWin(doc, (WORD) xmin, (WORD) ymin, (WORD) (-1 * dx), (WORD) (ymax - ymin), TRUE, TRUE);
//	       DrawRulers1(doc, doc->le_dclip, oldle, 0, 0, FALSE);
//	       disp_dist = UPLengthZoom(doc, oldle - doc->le_dclip, TRUE);
//	  }
//	  else // zde je mozna chyba
//	  {    BGFillDocWin(doc, (WORD) (xmax - dx), (WORD) ymin, (WORD) dx, (WORD) (ymax - ymin), TRUE, TRUE);
//	    // takto se to musi udelat, nebot jinak pri posunu umazeme konce cislic u predchozich jednotek
//	       pom_ruler = ((oldle + oldw) / doc->main_coef) * doc->main_coef;
//	       DrawRulers1(doc, max_ul(pom_ruler, doc->le_dclip), doc->le_dclip + doc->w_dclip, 0, 0, FALSE);
//	       disp_dist = UPLengthZoom(doc, doc->le_dclip - oldle, TRUE);
//	  }
//	  DrawSymbols(doc, ~0L, 0, FALSE, TRUE);   refresh_rooler = FALSE;
//	       // vykresleni novych radku
	  doc->cursor_checked = TRUE;
//	  if (smer == -1)
//	       RedrawVisibleClip(doc, &doc->display_begin, &doc->display_end, 0, 0, TRUE, FALSE);
//	  else RedrawVisibleClip(doc, &doc->display_begin, &doc->display_end, 0, 0, TRUE, FALSE);
///* dobre, ale blbnou prevody souradnic          if (smer == -1)
//	       RedrawVisibleClip(doc, &doc->display_begin, &doc->display_end, xmin, xmin - dx, TRUE);
//	  else RedrawVisibleClip(doc, &doc->display_begin, &doc->display_end, xmax - dx, xmax, TRUE);
//*/
//          if (smer == -1)
//               RedrawVisibleClip(doc, &doc->display_begin, &doc->display_end, doc->le_dwinclip, doc->le_dwinclip + disp_dist, TRUE);
//          else RedrawVisibleClip(doc, &doc->display_begin, &doc->display_end, doc->le_dwinclip + doc->w_dwinclip - disp_dist, doc->le_dwinclip + doc->w_dwinclip, TRUE);
     }

/* spravne nastaveni soupatek ... */
     if (docmove) gadget_refresh |= GG_SCROLLERS;

     doc->cursor_checked = TRUE;
     return(docmove);
} /* end MakeCursorVisible */
//<<

// NASLEDUJI FCE PRO POSUN VIDITELNEHO VYREZU TEXTU //

/* tato fce zobrazi novy vyrez textu - ten je urcen parametrem
 * newpos - vzdalenost horniho okraje od pocatku dokumentu */
/* pokud je zadan delta - chape se jako zmena a newpos se polozi
 * rovno actpos + delta; vsechny vstupy jsou v MUNIT */
BOOL SetNewVertPos(TEditFile *teditfile, ULONG newpos, LONG delta, BOOL draw_cursor)
// new version:
//   newpos - new position in lines from the BOF
//   delta - change in lines
//void SetNewVertPos(EditFile *doc, MUNIT newpos, LONG delta, BOOL draw_cursor)
//>>
{
     EditFile *doc = teditfile->GetEditFile();

//     struct RastPort *rp = doc->rp;
//
//     Location beg = doc->display_begin, end = doc->display_end, *cur = &doc->cursor;
//     PageSize *pgs = &doc->pagesize;
//     Page *pompage;
//     Page *page = doc->display_begin.page;
//
//     ULONG basepos= (beg.page->num - 1) * pgs->height;
     ULONG actline = doc->display_begin.line_num;
//     ULONG actpos = doc->te_dclip + basepos;
     ULONG numlines = max_ul(doc->num_lines, doc->wincheight) - doc->wincheight; 
//     ULONG maxpos = (pgs->height < doc->h_dclip) ? (doc->num_pages - 1) * pgs->height : doc->num_pages * pgs->height - doc->h_dclip;
//     ULONG maxpagepos = pgs->height - doc->h_dclip;
//     MUNIT oldte = doc->te_dclip, oldh = doc->h_dclip, pom_ruler;
//     BOOL page_ok;
//
     ULONG disp_dist = 0, i;
//     LONG xmin, ymin, xmax, ymax, dx, dy, i;
     long smer = 0;

// urceni newpos, je-li delta != 0
     if (delta != 0)
     {    if ((delta < 0) && (abs_l(delta) > actline))
	       newpos = 0;
	  else newpos = actline + delta;
     }
     if (newpos > numlines) newpos = numlines;
//     switch (SIGN(delta))
//     {
//	  case -1:  case 0:
//	       if (newpos % pgs->height > maxpagepos)
//		    newpos -= newpos % pgs->height - maxpagepos;
//	       break;
//	  case 1:
//	       if (newpos % pgs->height > maxpagepos)
//		    newpos += pgs->height - newpos % pgs->height;
//	       break;
//     }
//     (void) CorrectNumeric2(doc, &newpos);

     if (actline == newpos) return(FALSE);
//     {    doc->act_page = newpos / pgs->height + 1;    return;    }
//     page_ok = (actpos / pgs->height == newpos / pgs->height) ? TRUE : FALSE;
     disp_dist = (actline > newpos) ? actline - newpos : newpos - actline;
     smer = (actline > newpos) ? -1 : 1;
// pokud se nevykresluje kurzor --> neni kontrolovan
     if (NOT draw_cursor)
     {    doc->cursor_checked = FALSE;  /* usemcursor = TRUE; */   }

// we have to set the display_begin and display_end
     for(i = 0; i < disp_dist; i++)
          if (smer > 0)
          {
               if (NOT GetNextPos(teditfile, &doc->display_begin, TRUE, 0)) break;
          }
          else
          {
               if (NOT GetPrevPos(teditfile, &doc->display_begin, TRUE, 0)) break;
	       if (NOT GetPrevPos(teditfile, &doc->display_begin, FALSE, 0)) break;
          }
     SetDisplayBegEnd(teditfile, FALSE, 0);
          
// jsou 2 moznosti --> bud o kousek odrolujeme, nebo je nutno vse prekreslit
     if ( doc->wincheight > 0 )
     if ( 2*disp_dist > doc->wincheight )
//     if ( NOT page_ok || (5L*dist >= 4L*oldh) )
     {
//	  NumMouseClicks = 0;
//       // nove display_#?
//	  pompage = cur->page;
//	  for( i = newpos / pgs->height, cur->page = doc->page_first;
//	       (i > 0) && cur->page; i--, cur->page = cur->page->next);
//	  SetDisplayBegEnd(teditfile, FALSE, 0);
//	  cur->page = pompage;
//       // prezobrazeni vyrezu dokumentu ...
//	  RedrawNewSizeWindow(doc, TRUE, FALSE, TRUE);
          InvalidateLines(teditfile);
     }
     else // pripad, kdy budeme odrolovavat o kousek
     {
//	  NumMouseClicks = 0;
       // nove display_#?
//	  SetDisplayBegEnd(teditfile, FALSE, 0);
//       // spravne vyobrazeni vyrezu dokumentu ...
//	       // odrolovani stavajiciho obsahu
//	  dx = 0;
//	  dy = smer * UPLengthZoom(doc, (MUNIT) ((oldte >= doc->te_dclip) ? oldte - doc->te_dclip : doc->te_dclip - oldte), FALSE);
//	  xmin = xmax = doc->le_dwinclip;    xmax += doc->w_dwinclip;
//	  if (doc->ruler_vert) xmin -= VERT_RULER_SIZE;
//	  ymin = ymax = doc->te_dwinclip;    ymax += doc->h_dwinclip;
////          if (smer == -1) ymax += dy; else ymin += dy;
          ScrollLines(teditfile, smer * disp_dist);
//	  MyScrollRaster(doc, dx, dy, xmin, ymin, xmax, ymax, (smer == -1) ? &beg : &doc->display_begin, (smer == -1) ? &doc->display_end : &end);
//	       // vykresleni pravitka a noveho pozadi ...
//	  if (doc->ruler_vert) xmin += VERT_RULER_SIZE;
//	  if (smer == -1)
//	  {    BGFillDocWin(doc, (WORD) xmin, (WORD) ymin, (WORD) (xmax - xmin), (WORD) (-1 * dy), TRUE, TRUE);
//	       DrawRulers1(doc, 0, 0, doc->te_dclip, oldte, FALSE);
//	  }
//	  else
//	  {    BGFillDocWin(doc, (WORD) xmin, (WORD) (ymax - dy), (WORD) (xmax - xmin), (WORD) dy, TRUE, TRUE);
//	    // takto se to musi udelat, nebot jinak pri posunu umazeme konce cislic u predchozich jednotek
//	       pom_ruler = ((oldte + oldh) / doc->main_coef) * doc->main_coef;
//	       DrawRulers1(doc, 0, 0, max_ul(pom_ruler, doc->te_dclip), doc->te_dclip + doc->h_dclip, FALSE);
//	  }
//	       // vykresleni novych radku
////          if (NOT doc->block) doc->clear_before_writing = FALSE;
//	  if (smer == -1)
//	       RedrawVisibleClip(doc, &doc->display_begin, &beg, 0, 0, TRUE, FALSE);
//	  else RedrawVisibleClip(doc, &end, &doc->display_end, 0, 0, TRUE, FALSE);
////          if (NOT doc->block) doc->clear_before_writing = TRUE;
//	       // pokud se nahodou nevykreslil kurzor - nutno napravit
//	  if (NOT doc->old_ok && draw_cursor)
//	       DrawLine(doc, &doc->cursorline, 0, NULL, 0, 0, 0, doc->block, TRUE, FALSE, TRUE, FALSE, TRUE);
     }
     
     return(TRUE);
}
//<<

/* tato fce zobrazi novy vyrez textu - ten je urcen parametrem
 * left - vzdalenost od leveho okraje; delta - viz vyse */
BOOL SetNewHorizPos(TEditFile *teditfile, ULONG newpos, long delta, BOOL draw_cursor)
// new version:
//   newpos - new position in lines from the BOF
//   delta - change in lines
//void SetNewHorizPos(EditFile *doc, MUNIT newpos, WORD delta, BOOL draw_cursor)
//>>
{
     EditFile *doc = teditfile->GetEditFile();
     ULONG maxlinelen = teditfile->GetVisibleTextWidth();

//     struct RastPort *rp = doc->rp;
//
//// nepouzito     Location beg = doc->display_begin, end = doc->display_end;
//     PageSize *pgs = &doc->pagesize;
//     Page *page = doc->display_begin.page;

     ULONG actpos = doc->disp_offset;
     ULONG numcols = max_ul(maxlinelen, doc->wincwidth) - doc->wincwidth; 
//     MUNIT actpos = doc->le_dclip;
//     MUNIT maxpos = pgs->width - doc->w_dclip;
//     MUNIT oldle = doc->le_dclip, oldw = doc->w_dclip, pom_ruler;

     ULONG disp_dist = 0;
//     MUNIT dist = 0;
//     LONG xmin, ymin, xmax, ymax, dx, dy;
     long smer = 0;

// urceni newpos, je-li delta != 0
     if (delta != 0)
     {    if ((delta < 0) && (abs_l(delta) > actpos))
               newpos = 0;
          else newpos = actpos + delta;
     }
     if (newpos > numcols) newpos = numcols;

     if (actpos == newpos) return(FALSE);
     disp_dist = (actpos > newpos) ? actpos - newpos : newpos - actpos;
     smer = (actpos > newpos) ? -1 : 1;
// pokud se nevykresluje kurzor --> neni kontrolovan
     if (NOT draw_cursor)
     {    doc->cursor_checked = FALSE;  /* usemcursor = TRUE; */   }

     doc->disp_offset = newpos;
// jsou 2 moznosti --> bud o kousek odrolujeme, nebo je nutno vse prekreslit
     if ( doc->wincwidth > 0 )
     if (3*disp_dist >= doc->wincwidth)
//     if (3L*dist >= 2*doc->oldw)
     { // prezobrazeni vyrezu dokumentu ...
//	  NumMouseClicks = 0;
//	  RedrawNewSizeWindow(doc, TRUE, FALSE, TRUE);
          InvalidateColumns(teditfile);
     }
     else
     { // spravne vyobrazeni vyrezu dokumentu ...
	       // odrolovani stavajiciho obsahu
//	  NumMouseClicks = 0;
//	  dx = smer * UPLengthZoom(doc, (MUNIT) ((doc->le_dclip > oldle) ? doc->le_dclip - oldle : oldle - doc->le_dclip), TRUE);
//	  dy = 0;
//	  xmin = xmax = doc->le_dwinclip;    xmax += doc->w_dwinclip - 1;
////          if (smer == -1) xmax += dx; else xmin += dx;
//	  ymin = ymax = doc->te_dwinclip;    ymax += doc->h_dwinclip - 1;
//	  if (doc->ruler_horiz) ymin -= HORIZ_RULER_SIZE + MARGIN_TAB_SIZE;
//	  MyScrollRaster(doc, dx, dy, xmin, ymin, xmax, ymax, NULL, NULL);
          ScrollColumns(teditfile, smer * disp_dist);
//	  if (doc->ruler_horiz) ymin += HORIZ_RULER_SIZE + MARGIN_TAB_SIZE;
//	       // vykresleni pravitka a pozadi ...
//	  if (smer == -1)
//	  {    BGFillDocWin(doc, (WORD) xmin, (WORD) ymin, (WORD) (-1 * dx), (WORD) (ymax - ymin), TRUE, TRUE);
//	       DrawRulers1(doc, doc->le_dclip, oldle, 0, 0, FALSE);
//	       dist = UPLengthZoom(doc, oldle - doc->le_dclip, TRUE);
//	  }
//	  else // zde je mozna chyba
//	  {    BGFillDocWin(doc, (WORD) (xmax - dx), (WORD) ymin, (WORD) dx, (WORD) (ymax - ymin), TRUE, TRUE);
//	    // takto se to musi udelat, nebot jinak pri posunu umazeme konce cislic u predchozich jednotek
//	       pom_ruler = ((oldle + oldw) / doc->main_coef) * doc->main_coef;
//	       DrawRulers1(doc, max_ul(pom_ruler, doc->le_dclip), doc->le_dclip + doc->w_dclip, 0, 0, FALSE);
//	       dist = UPLengthZoom(doc, doc->le_dclip - oldle, TRUE);
//	  }
//	  DrawSymbols(doc, ~0L, 0, FALSE, TRUE);
//	       // vykresleni novych radku
//	  if (dx > 0) dx += 2; else dx -= 2;
//	  if (smer == -1)
//	       RedrawVisibleClip(doc, &doc->display_begin, &doc->display_end, 0, 0, TRUE, FALSE);
//	  else RedrawVisibleClip(doc, &doc->display_begin, &doc->display_end, 0, 0, TRUE, FALSE);
///* takto je to dobre, ale blbnou nam prevody souradnic         if (smer == -1)
//	       RedrawVisibleClip(doc, &doc->display_begin, &doc->display_end, xmin, xmin - dx, TRUE);
//	  else RedrawVisibleClip(doc, &doc->display_begin, &doc->display_end, xmax - dx, xmax, TRUE);
//*/
//          if (smer == -1)
//               RedrawVisibleClip(doc, &doc->display_begin, &doc->display_end, doc->le_dwinclip, doc->le_dwinclip + dist, TRUE);
//          else RedrawVisibleClip(doc, &doc->display_begin, &doc->display_end, doc->le_dwinclip + doc->w_dwinclip - dist, doc->le_dwinclip + doc->w_dwinclip, TRUE);
     }

     return(TRUE);
}
//<<

