//----------------------------------------------------------------------------
// BDEditor 2.2 - simplified ASCII only version
// Copyright (c) 2000, Pavel Cizek
// Main file
//----------------------------------------------------------------------------

#include "wed_editor.h"

#include "wed_editor.rh"
#include "xplore.rh"

#include "directory.h"

#define IDT_TIMER	1


////////////////////////
// TEditFile Elements //
////////////////////////

char **TEditFile::hi_keywords = NULL;
long TEditFile::hi_num_keywords = 0L;
char **TEditFile::hi_functions = NULL;
long TEditFile::hi_num_functions = 0L;
char **TEditFile::hi_libraries = NULL;
long TEditFile::hi_num_libraries = 0L;

// Global preferences
// the names of ini file keywords are in the second comment
GlobalPrefs TEditFile::Prefs =
     {
          TRUE,     // EOL = CR+LF		// PCEOLS
          TRUE,     // insert mode
          8,        // tab size			// TabSize
          TRUE,     // backup			// BackUp
          TRUE,     // autosave
          10,       // autosave interval
          16,       // undo_steps
          "",       // dialog preferences: open, save, filter
          "",
          "",
          {    // foreground
               COLOR_NOT_DEFINED,       // EC_HI_DEFAULT
               COLOR_NOT_DEFINED,       // EC_HI_TEXT
               MAKERGB(0, 0, 255),      // EC_HI_STRING
               MAKERGB(0, 0, 255),      // EC_HI_NUMBER
               MAKERGB(0, 0, 128),      // EC_HI_COMMENT
               COLOR_NOT_DEFINED,       // EC_HI_OPERATORS
               COLOR_NOT_DEFINED,       // EC_HI_BRACKETS
               MAKERGB(255, 0, 0),      // EC_HI_KEYWORDS
               MAKERGB(200, 0, 200),    // EC_HI_FUNCTIONS
               MAKERGB(0, 160, 0),      // EC_HI_LIBRARIES
               MAKERGB(255, 255, 255),  // EC_HI_ILLEGALCHAR
          },
          {    // background
               COLOR_NOT_DEFINED,       // EC_HI_DEFAULT
               COLOR_NOT_DEFINED,       // EC_HI_TEXT
               COLOR_NOT_DEFINED,       // EC_HI_STRING
               COLOR_NOT_DEFINED,       // EC_HI_NUMBER
               COLOR_NOT_DEFINED,       // EC_HI_COMMENT
               COLOR_NOT_DEFINED,       // EC_HI_OPERATORS
               COLOR_NOT_DEFINED,       // EC_HI_BRACKETS
               COLOR_NOT_DEFINED,       // EC_HI_KEYWORDS
               COLOR_NOT_DEFINED,       // EC_HI_FUNCTIONS
               COLOR_NOT_DEFINED,       // EC_HI_LIBRARIES
               MAKERGB(255, 0, 0),      // EC_HI_ILLEGALCHAR
          },
          TRUE,      // do highlighting  // SuntaxHiglighting
          MSG_DEFAULT_FONT_STR,   // font name  // FontName
          10               // font size  // FontSize
     };

APTR TEditFile::PoolHeader = NULL;
ULONG TEditFile::NumEditedFiles = 0L;
#ifdef X4standalone
BOOL TEditFile::inifile_read = FALSE;
#endif
TFont *TEditFile::editorfont = NULL;
BOOL TEditFile::openfont = TRUE;

#ifdef X4standalone
Properties editorproperties;
#endif

///////////////////////
// TEditFile Methods //
///////////////////////

// private functions //
ULONG TEditFile::GetVisibleTextWidth(void)
{
     Paragraph *para = editfile.display_begin.para, *paraend = editfile.display_end.para;
     ULONG maxlinelen, uselen;

     // PC 20020420
     maxlinelen = (para->explength >= para->length) ? para->explength : para->length;

     if (para != paraend)
     {
          do {
               para = para->next;
               // PC 20020420 - uselen - to use tab-expanded length
               uselen = (para->explength >= para->length) ? para->explength : para->length;
               if (maxlinelen < uselen)
                    maxlinelen = uselen;
          } while (para && (para != paraend));
     }

     return maxlinelen;
}

// PC 20020420 - added parameter for paint
void TEditFile::updatedisplay(BOOL frompaint)
{
     ULONG rangev = max_ul(editfile.num_lines, editfile.wincheight) - editfile.wincheight;
     ULONG rangeh, maxlinelen;

     // obtain the maximum line length
     if (editfile.wincheight > 0)
     {
          maxlinelen = GetVisibleTextWidth();
          rangeh = max_ul(maxlinelen, editfile.wincwidth) - editfile.wincwidth;
     }
     else
     {
          maxlinelen = 0;
          rangeh = 0;
     }

     // enable / disable scrollbar
     ShowScrollBar(SB_BOTH);
     EnableScrollBar(SB_VERT, (rangev > 0) ? ESB_ENABLE_BOTH : ESB_DISABLE_BOTH);
     EnableScrollBar(SB_HORZ, ((rangeh > 0) || (editfile.disp_offset > 0)) ? ESB_ENABLE_BOTH : ESB_DISABLE_BOTH);

     // update scrollbar
     SCROLLINFO siv = {
     		                sizeof(SCROLLINFO),                //    UINT cbSize;
                         SIF_ALL,                           //    UINT fMask;
                         0,                                 //    int  nMin;
                         editfile.num_lines - 1,            //    int  nMax;
                         editfile.wincheight,               //    UINT nPage;
                         editfile.display_begin.line_num,   //    int  nPos;
                         0                                  //    int  nTrackPos;
                     };
     SCROLLINFO sih = {
                         sizeof(SCROLLINFO),                //    UINT cbSize;
                         SIF_ALL,                           //    UINT fMask;
                         0,                                 //    int  nMin;
                         maxlinelen - 1,                    //    int  nMax;
                         editfile.wincwidth,                //    UINT nPage;
                         editfile.disp_offset,              //    int  nPos;
                         0                                  //    int  nTrackPos;
                     };
     // PC 20020415 - code for the case of big window and still shifted text
     SCROLLINFO sihalt = {
                         sizeof(SCROLLINFO),                //    UINT cbSize;
                         SIF_ALL,                           //    UINT fMask;
                         0,                                 //    int  nMin;
                         editfile.disp_offset + editfile.wincwidth - 1, //    int  nMax;
                         editfile.wincwidth,                //    UINT nPage;
                         editfile.disp_offset,              //    int  nPos;
                         0                                  //    int  nTrackPos;
                     };

     // scrollbar
     if (rangev > 0)
          (void) SetScrollInfo(SB_VERT, &siv, TRUE);
     if (rangeh > 0)
          (void) SetScrollInfo(SB_HORZ, &sih, TRUE);
     else if (editfile.disp_offset > 0)
          (void) SetScrollInfo(SB_HORZ, &sihalt, TRUE);

     if (NOT frompaint)
     {
          // correct title of the window
          SetWindowTitle(this, NULL);

          // set status line
#ifdef X4standalone
          setfilepos();
#endif

          SendMessage(WM_PAINT);
     }
}

void TEditFile::DoFindReplace(BOOL fromdialog)
{
     ULONG i;
     BOOL draw = FALSE, found;
     char *message = NULL, buf[100];
     
     // initialize replacedata
     strcpy((char *) replacedata.str1, searchdata.FindWhat);
     strcpy((char *) replacedata.str2, searchdata.ReplaceWith);
     replacedata.case_sens = (searchdata.Flags & FR_MATCHCASE) ? TRUE : FALSE;
     replacedata.whole_word = (searchdata.Flags & FR_WHOLEWORD) ? TRUE : FALSE;
     replacedata.forward = TRUE;
     replacedata.from_beg = FALSE;

     if (searchdata.Flags & FR_REPLACEALL)
     {
          i = 0;

          ChangePointer(this, GS(MSG_REPLACING_TITLE), TRUE, TRUE);
          do
          {
               if (DispatchMessages(this, MN_ReplaceNext, 1, NULL, FALSE))
                    draw = TRUE;
               if (found = replacedata.success)
                    i++;
          } while (found);
          ChangePointer(this, editfile.name, FALSE, FALSE);

          if (i == 0)
               message = GS(MSG_SEARCH_STRING_NOT_FOUND);
          else if (i == 1)
               message = GS(MSG_NUMBER_OF_REPLACED_STRING);
          else
          {
               message = buf;
               sprintf(buf, GS(MSG_NUMBER_OF_REPLACED_STRINGS), i);
          }
          
          noreplaceyet = FALSE;
     }
     else if (searchdata.Flags & FR_REPLACE)
     {
          if (DispatchMessages(this, MN_ReplaceNext, 0, NULL, FALSE))
               draw = TRUE;

          if (NOT replacedata.success)
               message = GS(MSG_SEARCH_STRING_NOT_FOUND);
          else if (fromdialog)
               if (DispatchMessages(this, MN_SearchNext, 2, NULL, FALSE))
                    draw = TRUE;

          noreplaceyet = FALSE;
     }
     else if (searchdata.Flags & FR_FINDNEXT)
     {
          if (DispatchMessages(this, MN_SearchNext, 0, NULL, FALSE))
               draw = TRUE;
          if (NOT replacedata.success)
               message = GS(MSG_SEARCH_STRING_NOT_FOUND);
     }

     // display update & message
     if (draw)
          updatedisplay();

     if (message)
          Information(this, message);
}

BOOL TEditFile::ClearEditFile(BOOL ask)
// if ask, then it returns whether users canceled the action (FALSE) or cleaning was done (TRUE)
{
     ULONG fh = editfile.fontheight;
     ULONG fw = editfile.fontwidth;
     LONG i = 2;

     if (ask)
          if (editfile.changes)
               i = QuestionYesNoCancel(this, editfile.name, GS(MSG_TEXT_NOT_SAVED));

     // action canceled
     if (i == 0)
	  return(FALSE);

     // save document
     if (i == 1)
     { 
          if (DispatchMessages(this, MN_SaveDoc, 0, NULL, FALSE))
               updatedisplay();

          if (editfile.last_saved == 0)
               return(FALSE);
     }
     
     // clear document
     if (NOT editfile.newfile)
     {
          CloseDoc(this, FALSE);
          initfinished = CreateDoc(this);
     }
//     CloseDoc(this, FALSE);
//     initfinished = CreateDoc(this);

     editfile.fontheight = fh;
     editfile.fontwidth = fw;

     UpdateWindowInfo(this);
     UpdateFontInfo(this);

     Invalidate();
     updatedisplay();

     return (TRUE);
}

#ifdef X4standalone
void TEditFile::InitFromIniFile(void)
{
     extern char *getx4cwd (char *subdir);
     extern void logt (char *txt);

     char *colorfkeys[] = {
                              "SyntaxForeString",
                              "SyntaxForeNumber",
                              "SyntaxForeComment",
                              "SyntaxForeKeyword",
                              "SyntaxForeFunction",
                              "SyntaxForeQuantlet",
                              NULL
                          };
     char *colorbkeys[] = {
                              "SyntaxBackString",
                              "SyntaxBackNumber",
                              "SyntaxBackComment",
                              "SyntaxBackKeyword",
                              "SyntaxBackFunction",
                              "SyntaxBackQuantlet",
                              NULL
                          };
     int colorinds[] = {
                         EC_HI_STRING,
                         EC_HI_NUMBER,
                         EC_HI_COMMENT,
                         EC_HI_KEYWORDS,
                         EC_HI_FUNCTIONS,
                         EC_HI_LIBRARIES,
                         0
                       };
     char *boolkeys[] = {
                         "PCEOLS",
                         "BackUp",
                         "SyntaxHighlighting",
                         NULL
                      };
     BOOL *boolkeysvalues[] = {
                              &Prefs.crlf_eols,
                              &Prefs.backup,
                              &Prefs.do_highlighting,
                              NULL
                            };
     char *uwordkeys[] = {
                              "TabSize",
                              "FontSize",
                              NULL
                         };
     UWORD *uwordkeysvalues[] = {
                                   &Prefs.tab_size,
                                   &Prefs.fontheight,
                                   NULL
                                };

     int i;
     UWORD uw, r, g, b;
     long l;
     char path[PATH_LEN + FILENAME + 4], *res;

/*** Modify Start : d-tec ***/
     extern char* xpl4homewin;
//     sprintf (path, "%s%sXploRe_Editor.ini", getx4cwd(NULL), SEPCHAR);
     sprintf (path, "%s%sXploRe_Editor.ini", xpl4homewin, SEPCHAR);
/*** Modify End : d-tec  ***/
     editorproperties.load(path);

     // check keys and set global prefs appropriately
     // BOOLs
     for(i = 0; boolkeys[i] && boolkeysvalues[i]; i++)
     {
          res = editorproperties.getProperty(boolkeys[i]);
          if ((res == NULL) || (strlen(res) == 0))
               continue;
          
          if (sscanf(res, "%ld", &l) >= 1)
               *(boolkeysvalues[i]) = (l) ? TRUE : FALSE;
     }

     // UWORDs
     for(i = 0; uwordkeys[i] && uwordkeysvalues[i]; i++)
     {
          res = editorproperties.getProperty(uwordkeys[i]);
          if ((res == NULL) || (strlen(res) == 0))
               continue;
          
          if (sscanf(res, "%hu", &uw) >= 1)
               *(uwordkeysvalues[i]) = uw;
     }

     // colors
     for(i = 0; colorfkeys[i] && colorinds[i]; i++)
     {
          res = editorproperties.getProperty(colorfkeys[i]);
          if ((res == NULL) || (strlen(res) == 0))
               continue;
          
          if (sscanf(res, "%hu %hu %hu", &r, &g, &b) >= 3)
               Prefs.highlight_foreground[colorinds[i]] = MAKERGB(r, g, b);
     }
     for(i = 0; colorbkeys[i] && colorinds[i]; i++)
     {
          res = editorproperties.getProperty(colorbkeys[i]);
          if ((res == NULL) || (strlen(res) == 0))
               continue;

          if (sscanf(res, "%hu %hu %hu", &r, &g, &b) >= 3)
               Prefs.highlight_background[colorinds[i]] = MAKERGB(r, g, b);
     }

     // fontname
     res = editorproperties.getProperty("FontName");
     if (res && strlen(res))
     {
     	  clrmem(Prefs.fontname, FILENAME);
          strncpy(Prefs.fontname, res, FILENAME - 1);
     }
}
#endif

// this function is not called from Dispatch (ineffective), it rather does
// the job by calling different DispatchMessages functions
BOOL TEditFile::DoUndoOperation(void)
//>>
{
     EditFile *doc = &editfile;

     Undo *pundo;
     UWORD undo_command;
     ULONG i, pom;
     BOOL draw = FALSE;
     UBYTE *pchar;

     // return if no undo is available
     if (doc->undo_steps == 0) return draw;

     pundo = &doc->undo_info[doc->undo_steps - 1];
     undo_command = pundo->undo_command;
     if ((undo_command == UNDO_Empty) || (undo_command == UNDO_Special))
          return draw;

     // wait pointer
     undoing_active = TRUE;
     ChangePointer(this, NULL, TRUE, TRUE);

     // set cursor position - position after the action
     pom = pundo->para_after;
     if (DispatchMessages(this, MN_GoToLine, 0, &pom, FALSE))
          draw = TRUE;
     pom = pundo->charoff_after + 1;
     if (DispatchMessages(this, MN_GoToColumn, 0, &pom, FALSE))
          draw = TRUE;
     
     // undo
     switch (undo_command)
     {
          case UNDO_DelBlock:
          case UNDO_DelText:  // inserted the deleted text
               if ((pundo->bufsize == 0) || (pundo->buf == NULL))
                    break;
               
               if (DispatchMessages(this, MN_Paste, 0, pundo->buf, FALSE))
                    draw = TRUE;
               break;

          case UNDO_FontFormat:
               if ((pundo->bufsize == 0) || (pundo->buf == NULL))
                    break;

               if (pundo->block)
               {
                    // select the original block
                    pom = pundo->para_beg;
                    if (DispatchMessages(this, MN_GoToLine, 0, &pom, FALSE))
                         draw = TRUE;
                    pom = pundo->charoff_beg + 1;
                    if (DispatchMessages(this, MN_GoToColumn, 0, &pom, FALSE))
                         draw = TRUE;

                    pom = pundo->para_end;
                    if (DispatchMessages(this, MN_GoToLineBlock, 0, &pom, TRUE))
                         draw = TRUE;
                    pom = pundo->charoff_end + 1;
                    if (DispatchMessages(this, MN_GoToColumnBlock, 0, &pom, TRUE))
                         draw = TRUE;

                    // paste over the saved version
                    if (DispatchMessages(this, MN_Paste, 0, pundo->buf, FALSE))
                         draw = TRUE;
               }
               else
               {
                    // set cursor to original position
                    pom = pundo->para_before;
                    if (DispatchMessages(this, MN_GoToLineBlock, 0, &pom, TRUE))
                         draw = TRUE;
                    pom = pundo->charoff_before + 1;
                    if (DispatchMessages(this, MN_GoToColumnBlock, 0, &pom, TRUE))
                         draw = TRUE;

                    // put the original character on its place
                    pchar = GetPosMB(doc->cursor.block, doc->cursor.block_off);
                    if (isalpha(*pchar))
                    {
                         if (DispatchMessages(this, BO_OverwriteChar, (UWORD) (*pundo->buf), NULL, 0))
                              draw = TRUE;
                         if (DispatchMessages(this, BO_CursorLeft, 0, NULL, FALSE))
                              draw = TRUE;
                    }
               }
               break;
               
          case UNDO_InsText:
               // select the inserted text
               pom = pundo->para_before;
               if (DispatchMessages(this, MN_GoToLineBlock, 0, &pom, TRUE))
                    draw = TRUE;
               pom = pundo->charoff_before + 1;
               if (DispatchMessages(this, MN_GoToColumnBlock, 0, &pom, TRUE))
                    draw = TRUE;

               // deleted
               if (DispatchMessages(this, MN_Erase, 0, NULL, FALSE))
                    draw = TRUE;
               break;
               
          case UNDO_InsClip:
               // select the inserted text
               pom = pundo->para_beg;
               if (DispatchMessages(this, MN_GoToLine, 0, &pom, FALSE))
                    draw = TRUE;
               pom = pundo->charoff_beg + 1;
               if (DispatchMessages(this, MN_GoToColumn, 0, &pom, FALSE))
                    draw = TRUE;
               
               pom = pundo->para_end;
               if (DispatchMessages(this, MN_GoToLineBlock, 0, &pom, TRUE))
                    draw = TRUE;
               pom = pundo->charoff_end + 1;
               if (DispatchMessages(this, MN_GoToColumnBlock, 0, &pom, TRUE))
                    draw = TRUE;

               // deleted
               if (DispatchMessages(this, MN_Erase, 0, NULL, FALSE))
                    draw = TRUE;
               break;

          case UNDO_OverwriteText:
               if ((pundo->bufsize == 0) || (pundo->buf == NULL))
                    break;

               // set cursor to original position
               pom = pundo->para_before;
               if (DispatchMessages(this, MN_GoToLineBlock, 0, &pom, TRUE))
                    draw = TRUE;
               pom = pundo->charoff_before + 1;
               if (DispatchMessages(this, MN_GoToColumnBlock, 0, &pom, TRUE))
                    draw = TRUE;

               // put the original character on its place
               if (DispatchMessages(this, BO_OverwriteChar, (UWORD) (*pundo->buf), NULL, 0))
                    draw = TRUE;
               if (DispatchMessages(this, BO_CursorLeft, 0, NULL, FALSE))
                    draw = TRUE;
               break;

          case UNDO_ReplaceText:
               // select the inserted text
               pom = pundo->para_before;
               if (DispatchMessages(this, MN_GoToLineBlock, 0, &pom, TRUE))
                    draw = TRUE;
               pom = pundo->charoff_before + 1;
               if (DispatchMessages(this, MN_GoToColumnBlock, 0, &pom, TRUE))
                    draw = TRUE;

               // deleted
               if (DispatchMessages(this, MN_Erase, 0, NULL, FALSE))
                    draw = TRUE;

               // insert replaced text
               if ((pundo->bufsize == 0) || (pundo->buf == NULL))
                    break;
               
               if (DispatchMessages(this, MN_Paste, 0, pundo->buf, FALSE))
                    draw = TRUE;
               break;
     }

     // set cursor position - position before the action, so undo is completed
     pom = pundo->para_before;
     if (DispatchMessages(this, MN_GoToLine, 0, &pom, FALSE))
          draw = TRUE;
     pom = pundo->charoff_before + 1;
     if (DispatchMessages(this, MN_GoToColumn, 0, &pom, FALSE))
          draw = TRUE;

     // zruseni ukazatele
     undoing_active = FALSE;
     ChangePointer(this, NULL, FALSE, FALSE);

     // remove the undo information
     (void) FreeUndoData(this);

     return draw;
}

// constructor //

TEditFile::TEditFile(TWindow *parent) : TWindow(parent)
{
// read ini file
#ifdef X4standalone
     if (NOT inifile_read)
     {
          InitFromIniFile();
          inifile_read = TRUE;
     }
#endif

// setup find/replace
     searchdata.Flags = FR_DOWN | FR_HIDEUPDOWN;
     searchdialog = NULL;
     searchreplace = FALSE;
     noreplaceyet = TRUE;

// create document
     initfinished = CreateDoc(this);

// rest of initialization
     existtimer = FALSE;
     mousecaptured = FALSE;
     lbuttondown = FALSE;

     busypointer = 0;

     undoing_active = FALSE;

     NumEditedFiles++;
}

TEditFile::~TEditFile()
{
     char buf[2*FILENAME + 64];
     
     // remove autosave file
     (void) DeleteFile(CreateAutosaveName(buf, editfile.name));
     editfile.saveticks = 0;
     
     // close document
     CloseDoc(this, FALSE);

     // release system resources
     if (searchdialog)
     {
          delete searchdialog;
          searchdialog = NULL;
     }

     if (existtimer)
     {
          KillTimer(IDT_TIMER);
          existtimer = FALSE;
     }

     if (mousecaptured)
     {    
          ReleaseCapture();
          mousecaptured = FALSE;
     }

     if (editorfont)
     {
          delete editorfont;
          editorfont = NULL;
     }

     // static members of TEditFile
     if (NumEditedFiles > 0)
          NumEditedFiles--;

     if (NumEditedFiles == 0)
     {    // free highlighting information
          long i, numnames;
          char **names;
          
          TEditFile::hi_keywords = NULL;
          TEditFile::hi_num_keywords = 0L;
          
          names = TEditFile::hi_functions;
          numnames = TEditFile::hi_num_functions;
          TEditFile::hi_functions = NULL;
          TEditFile::hi_num_functions = 0L;

          if (names)
          {
               for(i = 0; i < numnames; i++)
                    if (names[i]) 
                         delete[] names[i];
               delete[] names;
          }
          
          names = TEditFile::hi_libraries;
          numnames = TEditFile::hi_num_libraries;
          TEditFile::hi_libraries = NULL;
          TEditFile::hi_num_libraries = 0L;
          
          if (names)
          {
               for(i = 0; i < numnames; i++)
                    if (names[i])
                         delete[] names[i];
               delete[] names;
          }
     }
}

// Xplore interface //
void TEditFile::SetFilter(char *filter)
{
     clrmem(Prefs.def_filter, FILENAME);
     strncpy(Prefs.def_filter, filter, FILENAME - 2);
}

const char *TEditFile::GetFileName(BOOL full)
// if full is true, complete path is returned; otherwise just the filename
{
     return (full) ? editfile.path : editfile.name;
}

bool TEditFile::IsModified(void)
{
     return((editfile.changes) ? TRUE : FALSE);
}
     
ULONG TEditFile::GetNumLines(void)
{
     return editfile.num_lines;
}

ULONG TEditFile::GetCursorLine(void)
{
     return( 1 + editfile.cursor.line_num );
}

ULONG TEditFile::GetCursorCol(void)
{
     return( 1 + GetCursorColumn(this, &editfile.cursor) );
}

ULONG TEditFile::GetCurrentLineLength(void)
{
     return editfile.cursor.para->length;
}

ULONG TEditFile::GetTextSize(BOOL all, BOOL line, BOOL max)
// returns the size of the selected or whole text (depends whether user has selected a block)
// if all == TRUE, the size for whole text is returned
// if line == TRUE, the length of the current (max = FALSE) or the longest (max = TRUE)
// line is returned
{
     if (line)
          return (GetLineSize(this, max) + 4); // 4 - reserve for one CH_EOL and final '\0'
     else return (GetBlockSize(this, (all) ? FALSE : editfile.block) + 4); // 4 - reserve for one CH_EOL and final '\0'
}

void TEditFile::GetEditFileText(char *buffer, ULONG buflen, BOOL all, BOOL line)
// this function assumes that buffer is big enough!
// returns the selected or whole text (depends whether user has selected a block)
// if all == TRUE, the whole text is returned
// if line == TRUE, the content of the current line is returned
{
     Location lb, le;
     
     lb = editfile.block_begin;
     le = editfile.block_end;

     if (line)
     {
          editfile.block_begin = editfile.cursorline;
          editfile.block_end   = editfile.cursorline;
          (void) GetNextPos(this, &editfile.block_end, TRUE, 0);
     }
     else if (all || NOT editfile.block)
     {
          (void) BlockToLocation(this, &editfile.block_begin, editfile.block_first, 0, NULL, 0);
          (void) BlockToLocation(this, &editfile.block_end, editfile.block_last, editfile.block_last->size - editfile.block_last->free - 1, editfile.para_last, editfile.num_lines - 1);
     }

     clrmem(buffer, buflen);

     (void) SaveClipASCII(this, buffer, FALSE);

     // last EOL is not saved!
     if ( all || (line && (editfile.cursor.para == editfile.para_last)) )
          strcat(buffer, "\n");

// is it needed - probably not     strcat(buffer, "\n");

     editfile.block_begin = lb;
     editfile.block_end = le;
}

void TEditFile::InsertEditFileText(char *text, BOOL overblock)
// text is assumed to be a C string
// return TRUE (OK) or FALSE (not enough memory)
{
     ULONG dummy = 0;
     BOOL draw = FALSE;
     char *error;

     if (editfile.block && NOT overblock)
          if (DispatchMessages(this, BO_CursorLeft, 0, 0, FALSE))
               draw = TRUE;

     if (DispatchMessages(this, MN_Paste, 0, text, FALSE))
          draw = TRUE;

     if (draw)
          updatedisplay();
}

#ifdef X4standalone
void TEditFile::setfilepos(void)
{
     extern void statusmessage(const char *msg);

     ULONG lno, pos;
     char linemsg[PATH_LEN + FILENAME + 64];

     pos = GetCursorCol();
     lno = GetCursorLine();
     sprintf (linemsg, "%s (%lu,%lu)", GetFileName(TRUE), lno, pos);

     statusmessage (linemsg);
}
#endif

void TEditFile::RedrawWindow(void)
{
     Invalidate();
     updatedisplay();
}

char *TEditFile::CreateAutosaveName(char *buf, char *name)
// create name for autosave file given name (file name)
// buf has to be at least FILENAME + strlen(name) + 64 long
// returns address of buf
{
     char *tempname;

     tempname = getenv ("TEMP");
     if (tempname)
     {
          strcpy(buf, tempname);
          strcat(buf, "\\");
     }
     else *buf = '\0';

     strcat(buf, BDT_AUTOSAVE_PREP);
     strcat(buf, name);

     return buf;
}

int TEditFile::GetCursorChar(void)
{
     return (int) *GetPosMB(editfile.cursor.block, editfile.cursor.block_off);
}
     
char *TEditFile::GetCursorWord(void)
// return the word under the cursor
// string is created by strdup - if there is not enough memory or string is empty,
// it returns NULL
{
     char word[MAXWORDLEN + 4];

     (void) DispatchMessages(this, SPEC_GetCursorWord, MAXWORDLEN, word, FALSE);

     if (strlen(word) == 0)
          return NULL;
     else return strdup(word);
}

BOOL TEditFile::SelectCursorWord(void)
// select (as block) the word under cursor - the same as returned by the above function
// return TRUE (selected) or FALSE (nothing selected)
{
     BOOL draw = FALSE, selected = FALSE;
     char word[MAXWORDLEN + 4];

     if (DispatchMessages(this, SPEC_GetCursorWord, MAXWORDLEN, word, TRUE))
          draw = TRUE;

     if (strlen(word) > 0)
     {
          selected = TRUE;
          if (DispatchMessages(this, BO_NextWordEnd, 0, NULL, TRUE))
               draw = TRUE;
     }

     if (draw)
          updatedisplay();

     return selected;
}

BOOL TEditFile::UndoingActive(void)
{
     return undoing_active;
}

// TEditFile response table //
DEFINE_RESPONSE_TABLE1(TEditFile, TWindow)
     EV_COMMAND(IDM_MENU_QUIT, CmQuit),
     EV_COMMAND_ENABLE(IDM_MENU_QUIT, CeTrue),
     EV_COMMAND(IDM_MENU_NEW, CmNew),
     EV_COMMAND_ENABLE(IDM_MENU_NEW, CeTrue),
     EV_COMMAND(IDM_MENU_OPEN, CmOpen),
     EV_COMMAND_ENABLE(IDM_MENU_OPEN, CeTrue),
     EV_COMMAND(IDM_MENU_CLOSE, CmClose),
     EV_COMMAND_ENABLE(IDM_MENU_CLOSE, CeTrue),
     EV_COMMAND(IDM_MENU_SAVE, CmSave),
//     EV_COMMAND_ENABLE(IDM_MENU_SAVE, CeTrue),
     EV_COMMAND_ENABLE(IDM_MENU_SAVE, CeSave),
     EV_COMMAND(IDM_MENU_SAVEAS, CmSaveas),
     EV_COMMAND_ENABLE(IDM_MENU_SAVEAS, CeTrue),
     EV_COMMAND(IDM_MENU_RELOAD, CmReload),
     EV_COMMAND_ENABLE(IDM_MENU_RELOAD, CeReload),

     EV_COMMAND(IDM_MENU_UNDO, CmUndo),
     EV_COMMAND_ENABLE(IDM_MENU_UNDO, CeUndo),
     EV_COMMAND(IDM_MENU_CUT, CmCut),
     EV_COMMAND_ENABLE(IDM_MENU_CUT, CeBlock),
     EV_COMMAND(IDM_MENU_COPY, CmCopy),
     EV_COMMAND_ENABLE(IDM_MENU_COPY, CeBlock),
     EV_COMMAND(IDM_MENU_PASTE, CmPaste),
     EV_COMMAND_ENABLE(IDM_MENU_PASTE, CePaste),
     EV_COMMAND(IDM_MENU_DELETE, CmDelete),
     EV_COMMAND_ENABLE(IDM_MENU_DELETE, CeBlock),
     EV_COMMAND(IDM_MENU_SELECTALL, CmSelectAll),
     EV_COMMAND_ENABLE(IDM_MENU_SELECTALL, CeTrue),
     EV_COMMAND(IDM_MENU_CASETOGGLE, CmCaseToggle),
     EV_COMMAND_ENABLE(IDM_MENU_CASETOGGLE, CeTrue),
     EV_COMMAND(IDM_MENU_CASEUPPER, CmCaseUpper),
     EV_COMMAND_ENABLE(IDM_MENU_CASEUPPER, CeTrue),
     EV_COMMAND(IDM_MENU_CASELOWER, CmCaseLower),
     EV_COMMAND_ENABLE(IDM_MENU_CASELOWER, CeTrue),
     EV_COMMAND(IDM_MENU_LINEDELETE, CmDeleteLine),
     EV_COMMAND_ENABLE(IDM_MENU_LINEDELETE, CeNotBlock),
     EV_COMMAND(IDM_MENU_LINEDUPLICATE, CmDuplicateLine),
     EV_COMMAND_ENABLE(IDM_MENU_LINEDUPLICATE, CeNotBlock),
     EV_COMMAND(IDM_MENU_LINESWAP, CmSwapLines),
     EV_COMMAND_ENABLE(IDM_MENU_LINESWAP, CeSwapLine),
     EV_COMMAND(IDM_MENU_LIBCOMPLETE, CmCompleteLibrary),
     EV_COMMAND_ENABLE(IDM_MENU_LIBCOMPLETE, CeTrue),
     EV_COMMAND(IDM_MENU_TEMPLATECOMPLETE, CmCompleteTemplate),
     EV_COMMAND_ENABLE(IDM_MENU_TEMPLATECOMPLETE, CeTrue),
     EV_COMMAND(IDM_MENU_INSERTPATH, CmInsertPath),
     EV_COMMAND_ENABLE(IDM_MENU_INSERTPATH, CeTrue),

     EV_COMMAND(IDM_MENU_FIND, CmFind),
     EV_COMMAND_ENABLE(IDM_MENU_FIND, CeFind),
     EV_COMMAND(IDM_MENU_REPLACE, CmReplace),
     EV_COMMAND_ENABLE(IDM_MENU_REPLACE, CeFind),
     EV_COMMAND(IDM_MENU_FINDNEXT, CmFindNext),
     EV_COMMAND_ENABLE(IDM_MENU_FINDNEXT, CeTrue),
     EV_COMMAND(IDM_MENU_REPLACENEXT, CmReplaceNext),
     EV_COMMAND_ENABLE(IDM_MENU_REPLACENEXT, CeReplaceNext),
     EV_REGISTERED(FINDMSGSTRING, EvFindMsg),

     EV_COMMAND(IDM_MENU_GOTOLINE, CmGotoLine),
     EV_COMMAND_ENABLE(IDM_MENU_GOTOLINE, CeTrue),
     EV_COMMAND(IDM_MENU_GOTOCOLUMN, CmGotoColumn),
     EV_COMMAND_ENABLE(IDM_MENU_GOTOCOLUMN, CeTrue),
     EV_COMMAND(IDM_MENU_SET1, CmSet1),
     EV_COMMAND_ENABLE(IDM_MENU_SET1, CeTrue),
     EV_COMMAND(IDM_MENU_SET2, CmSet2),
     EV_COMMAND_ENABLE(IDM_MENU_SET2, CeTrue),
     EV_COMMAND(IDM_MENU_SET3, CmSet3),
     EV_COMMAND_ENABLE(IDM_MENU_SET3, CeTrue),
     EV_COMMAND(IDM_MENU_GOTO1, CmGoto1),
     EV_COMMAND_ENABLE(IDM_MENU_GOTO1, CeGoto1),
     EV_COMMAND(IDM_MENU_GOTO2, CmGoto2),
     EV_COMMAND_ENABLE(IDM_MENU_GOTO2, CeGoto2),
     EV_COMMAND(IDM_MENU_GOTO3, CmGoto3),
     EV_COMMAND_ENABLE(IDM_MENU_GOTO3, CeGoto3),
     
     EV_COMMAND(IDM_MENU_CONTEXTHELP, CmContextHelp),
     EV_COMMAND_ENABLE(IDM_MENU_CONTEXTHELP, CeTrue),

     EV_COMMAND_AND_ID(IDK_KEYB_LEFT, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_RIGHT, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_UP, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_DOWN, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_BEGVIEW, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_ENDVIEW, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_BEGFILE, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_ENDFILE, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_PAGEUP, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_PAGEDOWN, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_HOME, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_END, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_NEXTPARA, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_PREVPARA, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_NEXTWORD, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_NEXTWHITEWORD, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_PREVWORD, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_PREVWHITEWORD, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_DELEOL, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_DELBOL, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_DELLINE, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_DELPREVWORD, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_DELNEXTWORD, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_BACKSPACE, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_DELETE, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_SLEFT, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_SRIGHT, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_SUP, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_SDOWN, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_SBEGVIEW, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_SENDVIEW, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_SBEGFILE, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_SENDFILE, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_SPAGEUP, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_SPAGEDOWN, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_SHOME, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_SEND, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_SNEXTPARA, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_SPREVPARA, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_SNEXTWORD, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_SNEXTWHITEWORD, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_SPREVWORD, CmKeyboard),
     EV_COMMAND_AND_ID(IDK_KEYB_SPREVWHITEWORD, CmKeyboard),

     EV_COMMAND(IDK_KEYB_INSERTKEY, CmKeyInsert),

     EV_WM_ACTIVATE,
     EV_WM_KILLFOCUS,
     EV_WM_SETFOCUS,
     
     EV_WM_CAPTURECHANGED,
     EV_WM_CHAR,

/*** Modify Start : d-tec  ***/
     EV_MESSAGE(WM_IME_CHAR, EvImeChar),
     EV_MESSAGE(WM_IME_STARTCOMPOSITION, EvImeStartcomposition),
/*** Modify End : d-tec  ***/

     EV_WM_LBUTTONDBLCLK,
     EV_WM_LBUTTONDOWN,
     EV_WM_LBUTTONUP,
     EV_WM_MBUTTONDOWN,
     EV_WM_RBUTTONDOWN,
     EV_WM_MOUSEMOVE,
     EV_WM_SIZE,
     EV_WM_TIMER,
     EV_WM_VSCROLL,
     EV_WM_HSCROLL,
END_RESPONSE_TABLE;

// command handlers
void TEditFile::CmQuit(void)
{
     SendMessage(WM_CLOSE);
}

void TEditFile::CmNew(void)
{
     (void) ClearEditFile(TRUE);
}

void TEditFile::CmOpenFilename(char *filename)
{
     ULONG fh = editfile.fontheight;
     ULONG fw = editfile.fontwidth;

     if (NOT ClearEditFile((filename == NULL) ? TRUE : FALSE))
          return;

     if (DispatchMessages(this, MN_OpenDoc, 0, filename, FALSE))
          updatedisplay();
}

void TEditFile::CmOpen(void)
{
     CmOpenFilename(NULL);
}

void TEditFile::CmClose()
{
     SendMessage(WM_CLOSE);
}

void TEditFile::CmSave()
{
     if (DispatchMessages(this, MN_SaveDoc, 0, NULL, FALSE))
          updatedisplay();
}

void TEditFile::CmSaveas()
{
     char buf[2*FILENAME + 64], origname[FILENAME + 2];
     strcpy(origname, editfile.name);

     // save as operation
     if (DispatchMessages(this, MN_SaveAsDoc, 0, NULL, FALSE))
          updatedisplay();

     // remove autosave file
     (void) DeleteFile(CreateAutosaveName(buf, origname));
     editfile.saveticks = 0;
}

BOOL TEditFile::CmSaveTempFilename(char *filename, BOOL ext)
// if ext is true, then filename contains just an extension that has to be added
// to the current filename (including the ".")
{
     ULONG lastsave;
     BOOL draw, changes, saved = TRUE;
     char savefile[PATH_LEN + FILENAME + 16];
     char path[PATH_LEN + FILENAME + 16];
     char name[FILENAME + 2];

     if (filename == NULL)
          return(FALSE);
     if (strlen(filename) == 0)
          return(FALSE);
     if (strlen(filename) + ((ext) ? strlen(editfile.path) : 0) > PATH_LEN + FILENAME)
          return(FALSE);

     // backup doc info
     strcpy(path, editfile.path);
     strcpy(name, editfile.name);
     changes = editfile.changes;
     lastsave = editfile.last_saved;

     // create save filename + save
     if (ext)
     {
          strcpy(savefile, editfile.path);
          strcat(savefile, filename);
     }
     else strcpy(savefile, filename);

     // as no backups are required for the temporary files,
     // we have to delete the file if it exists
     (void) DeleteFile(savefile);

     // save file
     draw = DispatchMessages(this, MN_SaveAsDoc, 0, savefile, FALSE);
     if (editfile.last_saved == 0)
          saved = FALSE;

     // reset doc info
     strcpy(editfile.path, path);
     strcpy(editfile.name, name);
     editfile.changes = changes;
     editfile.last_saved = lastsave;

// has to be done always:      if (draw)
// to update title bar:
     updatedisplay();

     return saved;
}

void TEditFile::CmReload()
{
     char path[PATH_LEN + FILENAME + 16];
     char name[FILENAME + 2];

     if (NOT editfile.changes || NOT editfile.last_saved)
          return;

     if (0 == QuestionYesNo(this, editfile.name, GS(MSG_TEXT_RELOAD)))
          return;

     strcpy(path, editfile.path);
     strcpy(name, editfile.name);

     if (NOT ClearEditFile(FALSE))
          return;

     strcpy(editfile.path, path);
     strcpy(editfile.name, name);

     if (DispatchMessages(this, MN_OpenDoc, 0, path, FALSE))
          updatedisplay();
}

void TEditFile::CmUndo()
{
     if (DoUndoOperation())
          updatedisplay();
}

void TEditFile::CmCut()
{
     if (DispatchMessages(this, MN_Cut, 0, NULL, FALSE))
          updatedisplay();
}

void TEditFile::CmCopy()
{
     if (DispatchMessages(this, MN_Copy, 0, NULL, FALSE))
          updatedisplay();
}

void TEditFile::CmPaste()
{
     if (DispatchMessages(this, MN_Paste, 0, NULL, FALSE))
          updatedisplay();
}

void TEditFile::CmDelete()
{
     if (DispatchMessages(this, MN_Erase, 0, NULL, FALSE))
          updatedisplay();
}

void TEditFile::CmSelectAll()
{
     if (DispatchMessages(this, MN_SelectAll, 0, NULL, FALSE))
          updatedisplay();
}

void TEditFile::CmCaseToggle()
{
     BOOL draw = FALSE;

     if (DispatchMessages(this, MN_CaseChange, 0, NULL, FALSE))
          draw = TRUE;
     if (NOT editfile.block)
          if (DispatchMessages(this, BO_CursorRight, 0, NULL, FALSE))
               draw = TRUE;

     if (draw)
          updatedisplay();
}

void TEditFile::CmCaseUpper()
{
     BOOL draw = FALSE;

     if (DispatchMessages(this, MN_CaseUpper, 0, NULL, FALSE))
          draw = TRUE;
     if (NOT editfile.block)
          if (DispatchMessages(this, BO_CursorRight, 0, NULL, FALSE))
               draw = TRUE;

     if (draw)
          updatedisplay();
}

void TEditFile::CmCaseLower()
{
     BOOL draw = FALSE;

     if (DispatchMessages(this, MN_CaseLower, 0, NULL, FALSE))
          draw = TRUE;
     if (NOT editfile.block)
          if (DispatchMessages(this, BO_CursorRight, 0, NULL, FALSE))
               draw = TRUE;

     if (draw)
          updatedisplay();
}

void TEditFile::CmDeleteLine()
{
     CmKeyboard(IDK_KEYB_DELLINE);
}

void TEditFile::CmDuplicateLine()
{
     ULONG linesize = GetTextSize(FALSE, TRUE, FALSE) + 2, len;
     char *buf = new char[linesize];
     BOOL draw = FALSE;

     if (buf == NULL)
          return;
     if (editfile.block) // not possible if block is selected
          return;

     // goto beg of line
     if (DispatchMessages(this, BO_BegLine, 0, NULL, FALSE))
          draw = TRUE;
     
     // copy the line
     GetEditFileText(buf, linesize, FALSE, TRUE);
     len = strlen(buf);
     if (len > 0)   // there should be exactly one EOL
     {
          if (buf[len - 1] != CH_EOL)
               strcat(buf, "\n");
     }
     else
     {
          buf[0] = '\n';
          buf[1] = '\0';
     }
     
     // insert the line
     if (DispatchMessages(this, MN_Paste, 0, buf, FALSE))
          draw = TRUE;

     if (draw)
          updatedisplay();
     
     // free buffer
     delete buf;
}

void TEditFile::CmSwapLines()
{
     ULONG linesize = GetTextSize(FALSE, TRUE, FALSE) + 2, len;
     char *buf = new char[linesize];
     BOOL draw = FALSE, pom;

     if (buf == NULL)
          return;
     if (editfile.block) // not possible if block is selected
          return;
     if (editfile.cursor.para == editfile.para_last)   // nothing to swap
          return;

     // goto beg of line
     if (DispatchMessages(this, BO_BegLine, 0, NULL, FALSE))
          draw = TRUE;
     
     // copy the line
     GetEditFileText(buf, linesize, FALSE, TRUE);
     len = strlen(buf);
     if (len > 0)   // there should be exactly one EOL
     {
          if (buf[len - 1] != CH_EOL)
               strcat(buf, "\n");
     }
     else
     {
          buf[0] = '\n';
          buf[1] = '\0';
     }
     
     // delete the line
     if (DispatchMessages(this, BO_DelLine, 0, NULL, FALSE))
          draw = TRUE;

     // move one line down
     pom = (editfile.cursor.para == editfile.para_last) ? TRUE : FALSE;
     if (DispatchMessages(this, BO_CursorDown, 0, NULL, FALSE))
          draw = TRUE;
     if (pom)
          if (DispatchMessages(this, BO_NewPara, 0, NULL, FALSE))
               draw = TRUE;
     
     // insert the line
     if (DispatchMessages(this, MN_Paste, 0, buf, FALSE))
          draw = TRUE;

     if (draw)
          updatedisplay();
     
     // free buffer
     delete buf;
}

// get common length - the number of characters that are the same at the beginning of the string
int getcommonlen(const char *s1, const char *s2)
{
     int num;

     if ((s1 == NULL) || (s2 == NULL))
          return 0;

     for(num = 0; (*s1 == *s2) && *s1 && *s2; s1++, s2++)
          num++;

     return num;
}

char *TEditFile::searchstringlists(char *word, char *statusbuf, ULONG statuslen, int *foundstat)
// find the replacement of the word in lists of keywords
// foundstat 0 (work is already a keyword), -1 (no expansions found), 1 (more expansions found)
{
#define KEYWORDS    3
#define FUNCTIONS   2
#define OBJECTS     1
#define MACROS      0
     extern BOOL stringstrarray(int strarray, const char *str, char **pi1 = NULL, char **pi2 = NULL);

     long numnames;
     long i, j, best, chars, numlongest, foundindex;
     char *foundwords[2*MAXWORDLEN + 4];
     char *bestword, **names, exp[MAXWORDLEN + 4];

     clrmem(exp, MAXWORDLEN + 4);
     clrmem(statusbuf, statuslen);
     if (foundstat)
          *foundstat = -1;

     // go through the list and find the best word around
     // we have three list
     best = 0;      bestword = NULL;    // best - len in chars, bestword - pointer to the best keyword
     chars = 0;     numlongest = 0;     // chars - for common len, numlongest - number of best words
     foundindex = 0;                    // index for foundwords array
     for(i = 0; i < 3; i++)
     {
          switch (i)
          {
               case 0: 
                    numnames = TEditFile::hi_num_keywords;
                    names = TEditFile::hi_keywords;    break;
               case 1:
                    numnames = TEditFile::hi_num_functions;
                    names = TEditFile::hi_functions;   break;
               case 2:
                    numnames = TEditFile::hi_num_libraries;
                    names = TEditFile::hi_libraries;   break;
          }
          if ((names == NULL) || (numnames == 0))
               break;
          
          // search through the list
          for(j = 0; j < numnames; j++)
          {
               if (names[j] == NULL) 
                    continue;
               
               chars = getcommonlen(word, names[j]);
               // update best result
               if (chars > best)
               {
                    best = chars;
                    bestword = names[j];
                    numlongest = 0;
                    *statusbuf = '\0';
                    foundindex = 0;
               }
               // update strbuf array
               if (chars == best)
               {
                    numlongest++;
                    foundwords[foundindex++] = names[j];
                    
                    if (statuslen && statusbuf)
                         if (statuslen > strlen(statusbuf) + strlen(names[j]) + 4)
                         {
                              strcat(statusbuf, names[j]);
                              strcat(statusbuf, "; ");
                         }
               }
          }
     }
     if (best == 0)
     {
          if (foundstat)
               *foundstat = -1;
          *statusbuf = '\0';
          return NULL;
     }

     // create replacement string
     if (numlongest <= 1)
     {
          if (foundstat)
               *foundstat = 0;
          *statusbuf = '\0';
          strncpy(exp, bestword, MAXWORDLEN);
     }
     else
     {
          if (foundstat)
               *foundstat = 1;

          // if best == strlen(word) --- check whether we can get a longer word
          if ((best == strlen(word)) && (foundindex > 0))
          {
               // try, whether we can increase best
               for(i = 1, chars = strlen(foundwords[0]); i < foundindex; i++)
               {
                    j = getcommonlen(foundwords[0], foundwords[i]);
                    if (j < chars)
                         chars = j;
               }

               if (chars > best)
                    best = chars;
          }
          
          if (best > MAXWORDLEN)
               best = MAXWORDLEN;
          strncpy(exp, bestword, best);
     }

     // return replacement string (if it is different from word)
     if (0 == stricmp(exp, word))
          return NULL;
     else return strdup(exp);
}

void TEditFile::CmCompleteLibrary()
{
#define STATUSLINELEN	512
     char *word = GetCursorWord(), *found, statusbuf[STATUSLINELEN];

     if ((word == NULL) || (strlen(word) == 0))
          return;

     // try to find a word
     found = searchstringlists(word, statusbuf, STATUSLINELEN);
     if (found == NULL)  // not found  or cannot expand - quit                 
     {
#ifdef X4standalone
     extern void statusmessage(const char *msg);
     
     if (strlen(statusbuf) > 0)
          statusmessage(statusbuf);
#endif

          delete[] word;
          return;
     }
     
     // select the word to be overwritten and insert the text
     SelectCursorWord();
     InsertEditFileText(found, TRUE);

     // set cursor to the end of the word
     if (DispatchMessages(this, BO_NextWordEnd, 0, NULL, FALSE))
          updatedisplay();

     // set status line
#ifdef X4standalone
     extern void statusmessage(const char *msg);
     
     if (strlen(statusbuf) > 0)
          statusmessage(statusbuf);
#endif

     delete[] word;
     delete[] found;
}

void TEditFile::CmCompleteTemplate()
{
     // arrays have to contain as the last element NULL pointer
     static char *templates[] = { 
                                   "proc", "while", "if", "ifelse", "switch", NULL 
                                };
     static char *expansions[] = {// last endline is inserted automatically
                                   "proc() = \nendp", "while ()\nendo",
                                   "if ()\nendif", "if ()\nelse\nendif",
                                   "switch\nendsw", "do\nuntil",
                                   NULL 
                                 };
     
     int i, rows;
     char *word = GetCursorWord(), *exp, *pom, znak;

     if (word == NULL)
          return;

     // search whether there is a template
     for(i = 0; templates[i]; i++)
          if (0 == stricmp(word, templates[i]))
               break;

     if (templates[i] == NULL)     // nothing found?
     {
          delete[] word;
          return;
     }

     // found? select the template and insert the expansion
     SelectCursorWord();

     // insert text by rows
     rows = 0;
     exp = expansions[i];
     while (*exp)
     {
          pom = strchr(exp, CH_EOL);
          if (pom)
          {
               znak = pom[(rows == 0) ? 1 : 0];
               pom[(rows == 0) ? 1 : 0] = '\0';
          }

          InsertEditFileText(exp, TRUE);

          // insert new para
          if (pom)
          {
               pom[(rows == 0) ? 1 : 0] = znak;

               CmKeyboard(IDK_KEYB_END);
               if (DispatchMessages(this, BO_NewPara, 0, NULL, FALSE))
                    updatedisplay();
               
               exp = pom + 1;
               rows++; 
          }
          else break;    // no other line to be inserted
     }

     // get the specified number of rows back
     while (rows > 0)
     {
          CmKeyboard(IDK_KEYB_UP);
          rows--;
     }

     delete[] word;
}

void TEditFile::CmInsertPath()
{
     TFileOpenDialog::TData data(OFN_EXPLORER,
                                 (Prefs.def_filter) ? Prefs.def_filter : BDT_FILEMASK, 0,
                                 Prefs.def_openreq_path,
                                 BDT_DEFAULT_EXT);

     if (TFileOpenDialog(this, data).Execute() == IDOK)
          InsertEditFileText(data.FileName, TRUE);
}

void TEditFile::CmFind()
{
     if (searchdialog == NULL)
     {
          searchreplace = FALSE;
          searchdialog = new TFindDialog(this, searchdata);
          searchdialog->Create();
     }
}

void TEditFile::CmReplace()
{
     if (searchdialog == NULL)
     {
          searchreplace = TRUE;
          searchdialog = new TReplaceDialog(this, searchdata);
          searchdialog->Create();
     }
}

void TEditFile::CmFindNext()
{

     if ( (searchdialog == NULL) && (searchdata.FindWhat == NULL) )
     {
          CmFind(); 	return;
     }
     if ( (searchdialog == NULL) && (*searchdata.FindWhat == 0) )
     {
          CmFind(); 	return;
     }

     if (searchdialog)
          searchdialog->UpdateData();
     searchdata.Flags &= ~FR_REPLACE;
     searchdata.Flags &= ~FR_REPLACEALL;
     searchdata.Flags |= FR_FINDNEXT;

     if (*searchdata.FindWhat == 0)
     {
          Information(this, GS(MSG_NO_SEARCH_STRING));	return;
     }

     DoFindReplace(FALSE);
}

void TEditFile::CmReplaceNext()
{
     if ( (searchdialog == NULL) &&
          ((searchdata.FindWhat == NULL) || (*searchdata.FindWhat == 0) || noreplaceyet)
        )
     {
          CmReplace(); 	return;
     }
     if (searchdialog /* && NOT searchreplace */)
          return;

     if (searchdialog)
          searchdialog->UpdateData();
     searchdata.Flags &= ~FR_FINDNEXT;
     searchdata.Flags &= ~FR_REPLACEALL;
     searchdata.Flags |= FR_REPLACE;

     if (*searchdata.FindWhat == 0)
     {
          Information(this, GS(MSG_NO_SEARCH_STRING));	return;
     }

     DoFindReplace(FALSE);
}

TResult TEditFile::EvFindMsg(TParam1 param1, TParam2 param2)
{
     if (searchdialog == NULL)
          return 0;

     searchdialog->UpdateData(param2);
     if (searchdata.Flags & FR_DIALOGTERM)
     {
          delete searchdialog;
          searchdialog = NULL;
     }
     else DoFindReplace(TRUE);

     return 0;
}

void TEditFile::CmGotoLine()
{
     if (DispatchMessages(this, MN_GoToLine, 0, NULL, FALSE))
          updatedisplay();
}

void TEditFile::CmGotoGivenLine(ULONG line)
{
     if (DispatchMessages(this, MN_GoToLine, 0, &line, FALSE))
          updatedisplay();
}

void TEditFile::CmGotoColumn()
{
     if (DispatchMessages(this, MN_GoToColumn, 0, NULL, FALSE))
          updatedisplay();
}

void TEditFile::CmGotoGivenColumn(ULONG col)
{
     if (DispatchMessages(this, MN_GoToColumn, 0, &col, FALSE))
          updatedisplay();
}

void TEditFile::CmSet1()
{
     if (DispatchMessages(this, MN_SetBookmark, 0, NULL, FALSE))
          updatedisplay();
}

void TEditFile::CmSet2()
{
     if (DispatchMessages(this, MN_SetBookmark2, 0, NULL, FALSE))
          updatedisplay();
}

void TEditFile::CmSet3()
{
     if (DispatchMessages(this, MN_SetBookmark3, 0, NULL, FALSE))
          updatedisplay();
}

void TEditFile::CmGoto1()
{
     if (DispatchMessages(this, MN_GoToBookmark, 0, NULL, FALSE))
          updatedisplay();
}

void TEditFile::CmGoto2()
{
     if (DispatchMessages(this, MN_GoToBookmark2, 0, NULL, FALSE))
          updatedisplay();
}

void TEditFile::CmGoto3()
{
     if (DispatchMessages(this, MN_GoToBookmark3, 0, NULL, FALSE))
          updatedisplay();
}

// PC20011101 - context help
void TEditFile::CmContextHelp()
{
#define STATUSLINELEN	256
     char path[300 + STATUSLINELEN + 4], *tmp;
     char *word = GetCursorWord(), *found, statusbuf[STATUSLINELEN];
     int notfound;

     if ((word == NULL) || (strlen(word) == 0))
          return;

     // try to find a word
     found = searchstringlists(word, statusbuf, STATUSLINELEN, &notfound);
     if (found || notfound)  // more possibilities - open general help
     {
          tmp = openfile (HELP, "00.html", "r");
          if (tmp)
          {
               strcpy(path, tmp);
               ShellExecute(NULL,"open",path,"","",SW_SHOWNORMAL);
          }

          delete[] word;
          return;
     }
     
     // select the word and show help --- the word is in found
     memset(statusbuf, 0, STATUSLINELEN);
     strncpy(statusbuf, word, STATUSLINELEN - 10);
     strcat(statusbuf, ".html");

     tmp = openfile (HELP, statusbuf, "r");
     if (tmp)
     { 
          strcpy(path, tmp);
          ShellExecute(NULL,"open",path,"","",SW_SHOWNORMAL);
     }
     else
     {
          tmp = openfile (HELP, "_Xpl_Start.html", "r");
          if (tmp)
          { 
               strcpy(path, tmp);
               ShellExecute(NULL,"open",path,"","",SW_SHOWNORMAL);
          }
     }
     
     delete[] word;
     delete[] found;
}

// command enablers
void TEditFile::CeTrue(TCommandEnabler& ce)
{
     ce.Enable(TRUE);
}

void TEditFile::CeSave(TCommandEnabler& ce)
{
     ce.Enable((editfile.changes) ? TRUE : FALSE);
}

void TEditFile::CeReload(TCommandEnabler& ce)
{
     ce.Enable((editfile.changes && editfile.last_saved) ? TRUE : FALSE);
}

void TEditFile::CeBlock(TCommandEnabler& ce)
{
     ce.Enable((editfile.block) ? TRUE : FALSE);
}

void TEditFile::CeNotBlock(TCommandEnabler& ce)
{
     ce.Enable((editfile.block) ? FALSE : TRUE);
}

void TEditFile::CeUndo(TCommandEnabler& ce)
{
     ce.Enable(((editfile.undo_steps > 0) && NOT undoing_active) ? TRUE : FALSE);
}

void TEditFile::CePaste(TCommandEnabler& ce)
{
     TClipboard& clipboard = OpenClipboard();
     ce.Enable(clipboard && clipboard.IsClipboardFormatAvailable(CF_TEXT));
     clipboard.CloseClipboard();
}

void TEditFile::CeSwapLine(TCommandEnabler& ce)
{
     ce.Enable((editfile.block || (editfile.cursor.para == editfile.para_last)) ? FALSE : TRUE);
}

void TEditFile::CeFind(TCommandEnabler& ce)
{
  ce.Enable((searchdialog) ? FALSE : TRUE);
}

void TEditFile::CeReplaceNext(TCommandEnabler& ce)
{
  ce.Enable((searchdialog) ? (  /* (searchreplace) ? TRUE : */ FALSE ) : TRUE);
}

void TEditFile::CeGoto1(TCommandEnabler& ce)
{
     ce.Enable((editfile.bookmark.block) ? TRUE : FALSE);
}

void TEditFile::CeGoto2(TCommandEnabler& ce)
{
     ce.Enable((editfile.bookmark2.block) ? TRUE : FALSE);
}

void TEditFile::CeGoto3(TCommandEnabler& ce)
{
     ce.Enable((editfile.bookmark3.block) ? TRUE : FALSE);
}

// keyboard messages handler
void TEditFile::CmKeyboard(WPARAM id)
{
     Command comm = COMM_Dummy;
     APTR addr = NULL;
     UWORD key = 0;
     BOOL draw = FALSE, block = FALSE;

     // find command
     switch (id)
     {
          case IDK_KEYB_SBEGVIEW:
                    block = TRUE;
          case IDK_KEYB_BEGVIEW:
                    comm = BO_BegView;        break;
          case IDK_KEYB_SENDVIEW:
                         block = TRUE;
          case IDK_KEYB_ENDVIEW:
                    comm = BO_EndView;        break;
          case IDK_KEYB_SBEGFILE:
                         block = TRUE;
          case IDK_KEYB_BEGFILE:
                    comm = BO_BegFile;        break;
          case IDK_KEYB_SENDFILE:
                         block = TRUE;
          case IDK_KEYB_ENDFILE:
                    comm = BO_EndFile;        break;
          case IDK_KEYB_SPAGEUP:
                         block = TRUE;
          case IDK_KEYB_PAGEUP:
                    comm = BO_PageUp;         break;
          case IDK_KEYB_SPAGEDOWN:
                         block = TRUE;
          case IDK_KEYB_PAGEDOWN:
                    comm = BO_PageDown;       break;
          case IDK_KEYB_SHOME:
                         block = TRUE;
          case IDK_KEYB_HOME:
                    comm = BO_BegLine;        break;
          case IDK_KEYB_SEND:
                         block = TRUE;
          case IDK_KEYB_END:
                    comm = BO_EndLine;        break;
          case IDK_KEYB_SNEXTPARA:
                         block = TRUE;
          case IDK_KEYB_NEXTPARA:
                    comm = BO_ParaDown;       break;
          case IDK_KEYB_SPREVPARA:
                         block = TRUE;
          case IDK_KEYB_PREVPARA:
                    comm = BO_ParaUp;         break;
          case IDK_KEYB_SNEXTWORD:
                         block = TRUE;
          case IDK_KEYB_NEXTWORD:
                    comm = BO_NextWord;       break;
          case IDK_KEYB_SNEXTWHITEWORD:
                         block = TRUE;
          case IDK_KEYB_NEXTWHITEWORD:
                    comm = BO_NextWhiteWord;  break;
          case IDK_KEYB_SPREVWORD:
                         block = TRUE;
          case IDK_KEYB_PREVWORD:
                    comm = BO_PrevWord;       break;
          case IDK_KEYB_SPREVWHITEWORD:
                         block = TRUE;
          case IDK_KEYB_PREVWHITEWORD:
                    comm = BO_PrevWhiteWord;  break;
     	  case IDK_KEYB_SLEFT:
                         block = TRUE;
     	  case IDK_KEYB_LEFT:
                    comm = BO_CursorLeft;     break;
     	  case IDK_KEYB_SRIGHT:
                         block = TRUE;
     	  case IDK_KEYB_RIGHT:
                    comm = BO_CursorRight;    break;
     	  case IDK_KEYB_SUP:
                         block = TRUE;
     	  case IDK_KEYB_UP:
                    comm = BO_CursorUp;       break;
     	  case IDK_KEYB_SDOWN:
                         block = TRUE;
     	  case IDK_KEYB_DOWN:
                    comm = BO_CursorDown;     break;

          case IDK_KEYB_DELEOL:
                    comm = BO_DelEOL;         break;
          case IDK_KEYB_DELBOL:
                    comm = BO_DelBOL;         break;
          case IDK_KEYB_DELLINE:
                    comm = BO_DelLine;        break;
          case IDK_KEYB_DELPREVWORD:
                    comm = BO_DelPrevWord;    break;
          case IDK_KEYB_DELNEXTWORD:
                    comm = BO_DelWord;        break;
          case IDK_KEYB_BACKSPACE:
                    comm = BO_DelPrevChar;    break;
          case IDK_KEYB_DELETE:
                    comm = BO_DelChar;        break;

          default:  comm = COMM_Dummy;   break;
     }

     // send command
     if (comm != COMM_Dummy)
          draw = DispatchMessages(this, comm, key, addr, block);

     if (draw)
          updatedisplay();
}

void TEditFile::CmKeyInsert()
{
     Prefs.insert = NOT Prefs.insert;

     // redraw because of cursor change
     if ( (editfile.display_begin.line_num <= editfile.cursor.line_num) &&
          (editfile.display_end.line_num   >= editfile.cursor.line_num) )
     {
          InvalidateLines(this, 1, editfile.cursor.line_num - editfile.display_begin.line_num, FALSE);
          updatedisplay();
     }
}

// event handlers
void TEditFile::EvActivate(uint active, bool minimized, HWND hWndOther)
{
     TWindow::EvActivate(active, minimized, hWndOther);

     UpdateWindowInfo(this);
     UpdateFontInfo(this);

     if (active == WA_INACTIVE)
          return;

     GetParentO()->SetActiveWindow();
     GetParentO()->SetCaption(editfile.name);
     SetFocus();
     ChangePointer(this, NULL, FALSE, FALSE);

     // draw whatever is needed:
     updatedisplay();
}

void TEditFile::EvKillFocus(HWND hWndGetFocus)
{
     editfile.active = FALSE;
}

void TEditFile::EvSetFocus(HWND hWndLostFocus)
{
     editfile.active = TRUE;

     // draw whatever is needed:
     updatedisplay();
}

void TEditFile::EvCaptureChanged(THandle hwnd)
{
     if (mousecaptured)
     {
          ReleaseCapture();
          mousecaptured = FALSE;
     }
}

void TEditFile::EvChar(uint key, uint repeatCount, uint flags)
{
     TWindow::EvChar(key, repeatCount, flags);

     BOOL draw = FALSE;

     // send command
     if ((key >= CH_SPACE) || (key == VK_TAB))
#ifndef JAPANESE
          draw = DispatchMessages(this, (Prefs.insert) ? BO_InsertChar : BO_OverwriteChar, (UWORD) key, NULL, 0);
#else
          /*** Modify Start : d-tec  ***/
          if (Prefs.insert) {
               //Insert
               draw = DispatchMessages(this, BO_InsertChar, (UWORD) key, NULL, 0);
          } else {
               //Overwrite
               draw = DispatchMessages(this, BO_DelChar, 0, NULL, 0);
               draw = DispatchMessages(this, BO_InsertChar, (UWORD) key, NULL, 0);
          }
          /*** Modify End : d-tec  ***/
#endif
/*          draw = DispatchMessages(this, (Prefs.insert) ? BO_InsertChar : BO_OverwriteChar, (UWORD) key, NULL, 0);     */
     if ((key == VK_RETURN) || (key == CH_EOL) || (key == VK_CLEAR))	// FF = page down?
          draw = DispatchMessages(this, (key == VK_CLEAR) ? BO_NewPara/*BP*/ : BO_NewPara, (UWORD) key, NULL, 0);
//     if (key == VK_BACK)
//          draw = DispatchMessages(this, BO_DelPrevChar, (UWORD) key, NULL, 0);

     if (draw)
          updatedisplay();
}

/*** Modify Start : d-tec  ***/
TResult TEditFile::EvImeChar(WPARAM wchar, LPARAM param2)
{
	BOOL draw = FALSE;
	char w_str[3];
	
	w_str[0] = HIBYTE(wchar);
	if (w_str[0] == '\0') {
		w_str[0] = LOBYTE(wchar);
		w_str[1] = '\0';
	} else {
		w_str[1] = LOBYTE(wchar);
	}
	w_str[2] = '\0';

/*** Debug Start : d-tec ***/
//		sprintf(debug_buffer, "EvImeChar:wchar=%s",w_str);
//		write_to_log(debug_buffer);
/*** Debug End : d-tec  ***/

	if (Prefs.insert) {
		//Insert
		draw = DispatchMessages(this, BO_InsertWord, NULL, w_str, 0);
	} else {
		//Overwrite
		draw = DispatchMessages(this, BO_DelChar, 0, NULL, FALSE);
		draw = DispatchMessages(this, BO_InsertWord, NULL, w_str, 0);
	}
	if (draw)
		updatedisplay();
}

TResult TEditFile::EvImeStartcomposition(WPARAM param1, LPARAM param2)
{
	HWND hWnd = GetHandle();
	HIMC hIMC = ImmGetContext(hWnd);
	LOGFONT lf;
	COMPOSITIONFORM cf;
	ULONG fh = editfile.fontheight, fw = editfile.fontwidth;
	int x, y;
	
	// IME FONT
	if (editorfont) {
		editorfont->GetObject(lf);
	} else {
		HFONT hFont=(HFONT) GetStockObject(SYSTEM_FIXED_FONT);
		GetObject(hFont, sizeof(LOGFONT), &lf);
	}
	ImmSetCompositionFont(hIMC, &lf);
	
	// IME POSITION
	x = fw * (editfile.last_cursor_pos - editfile.disp_offset);
	y = fh * (editfile.cursor.line_num - editfile.display_begin.line_num);
/*** Debug Start : d-tec ***/
//		sprintf(debug_buffer, "EvImeStartcomposition:x=%d,y=%d",x,y);
//		write_to_log(debug_buffer);
/*** Debug End : d-tec  ***/
	cf.dwStyle = CFS_POINT;
	cf.ptCurrentPos.x = x;
	cf.ptCurrentPos.y = y;
	ImmSetCompositionWindow(hIMC, &cf);
	
	ImmReleaseContext(hWnd, hIMC);
	
	return DefWindowProc(GetCurrentEvent().Message, param1, param2);
}
/*** Modify End : d-tec  ***/

void TEditFile::EvLButtonDblClk(uint modKeys, TPoint& point)
{
     EvLButtonDown(modKeys, point);
//     lbuttondown = TRUE;
     lbuttondown = FALSE;	// no further block selection allowed

     if (DispatchMessages(this, MN_SelectWord, 0, 0, FALSE))
          updatedisplay();
}

void TEditFile::EvLButtonDown(uint modKeys, TPoint& point)
{
     int coords[2] = { point.X(), point.Y() };

     // with right button = middle button
     if (modKeys & MK_RBUTTON)
     {
          if (editfile.block)
               CmCopy();
     }

     lbuttondown = TRUE;

     if (DispatchMessages(this, MN_GoToCursor, 0, coords, FALSE))
          updatedisplay();

     // with right button = middle button
     if (modKeys & MK_RBUTTON)
     {
          lbuttondown = FALSE;

          CmPaste();
     }
}

void TEditFile::EvLButtonUp(uint modKeys, TPoint& point)
{
     if (mousecaptured)
     {
          ReleaseCapture();
          mousecaptured = FALSE;
     }
     lbuttondown = FALSE;
}

void TEditFile::EvMButtonDown(uint modKeys, TPoint& point)
{
     if (editfile.block)
          CmCopy();

     EvLButtonDown(modKeys, point);
     lbuttondown = FALSE;

     CmPaste();
}

// PC 20010712 - context menu added
void TEditFile::EvRButtonDown(uint modKeys, TPoint& point)
{
/* not enabled for now
     if (modKeys & MK_LBUTTON)
     {
          EvLButtonDown(modKeys, point);
          lbuttondown = FALSE;
     }
*/
  TPopupMenu cmenu;

  // construct menu
  cmenu.AppendMenu(MF_ENABLED, IDM_MENU_UNDO, MSG_EDCMENU_UNDO_STR);
  cmenu.AppendMenu(MF_ENABLED | MF_SEPARATOR);
  cmenu.AppendMenu(MF_ENABLED, IDM_MENU_CUT, MSG_EDCMENU_CUT_STR);
  cmenu.AppendMenu(MF_ENABLED, IDM_MENU_COPY, MSG_EDCMENU_COPY_STR);
  cmenu.AppendMenu(MF_ENABLED, IDM_MENU_PASTE, MSG_EDCMENU_PASTE_STR);
  cmenu.AppendMenu(MF_ENABLED, IDM_MENU_DELETE, MSG_EDCMENU_DELETE_STR);
  cmenu.AppendMenu(MF_ENABLED | MF_SEPARATOR);
  cmenu.AppendMenu(MF_ENABLED, IDM_MENU_SELECTALL, MSG_EDCMENU_SELECTALL_STR);
  cmenu.AppendMenu(MF_ENABLED | MF_SEPARATOR);
  cmenu.AppendMenu(MF_ENABLED, CM_EXECFILE, MSG_EDCMENU_EXECUTE_STR);
  cmenu.AppendMenu(MF_ENABLED | MF_SEPARATOR);
  cmenu.AppendMenu(MF_ENABLED, CM_FORMAT, MSG_EDCMENU_FORMAT_STR);
  cmenu.AppendMenu(MF_ENABLED, CM_INSERTHELP, MSG_EDCMENU_HELP_STR);
  cmenu.AppendMenu(MF_ENABLED, IDM_MENU_CONTEXTHELP, MSG_EDCMENU_CONTEXTHELP_STR);

  // display menu  
  ClientToScreen(point);
  cmenu.TrackPopupMenu(TPM_LEFTBUTTON | TPM_RIGHTBUTTON, point, 0, *this);
}

void TEditFile::EvMouseMove(uint modKeys, TPoint& point)
{
     int coords[2] = { point.X(), point.Y() };

     if ((modKeys & MK_LBUTTON) == 0)
     {
          lbuttondown = FALSE;
          if (mousecaptured)
          {
               ReleaseCapture();
               mousecaptured = FALSE;
          }
          return;
     }
     if (NOT lbuttondown) return;
     if (NOT mousecaptured && GetCapture())
          return;

     if (NOT mousecaptured)   // CaptureMouse for block selection
     {
          SetCapture();
          mousecaptured = TRUE;
     }

     if (DispatchMessages(this, MN_GoToCursor, 0, coords, TRUE))
          updatedisplay();
}

void TEditFile::EvSize(uint sizeType, TSize& size)
{
     ULONG wc = editfile.wincwidth;
     ULONG hc = editfile.wincheight;
     TSize oldsize(editfile.winwidth, editfile.winheight);

     TWindow::EvSize(sizeType, size);

     // get window size, update dimensions and view
     UpdateWindowInfo(this);
     UpdateFontInfo(this);
     SetDisplayBegEnd(this, TRUE, 0);

     // invalidate & redraw
     if ( (size.X() > oldsize.X()) && (size.Y() > 0) )
          if (editfile.wincwidth > wc)
               InvalidateColumns(this, editfile.wincwidth - wc);
     if ( (size.Y() > oldsize.Y()) && (size.X() > 0) )
          if (editfile.wincheight > hc)
               InvalidateLines(this, editfile.wincheight - hc);

     updatedisplay();
}

void TEditFile::EvHScroll(uint scrollCode, uint thumbPos, HWND hWndCtl)
{
     ULONG maxlinelen = GetVisibleTextWidth();
     ULONG newpos = editfile.disp_offset;
     LONG delta = 0;
     BOOL draw = FALSE;

     switch (scrollCode)
     {
     	    case SB_BOTTOM:
               newpos = maxlinelen;
               break;
          case SB_ENDSCROLL:
               break;
          case SB_LINEDOWN:
               delta = 1;     break;
          case SB_LINEUP:
               delta = -1;    break;
          case SB_PAGEDOWN:
               delta = editfile.wincwidth;       break;
          case SB_PAGEUP:
               delta = -1 * editfile.wincwidth;  break;
          case SB_THUMBPOSITION:
          case SB_THUMBTRACK:
               newpos = thumbPos;  break;
          case SB_TOP:
               newpos = 0;
               break;
     }

     if (delta)
          draw = DispatchMessages(this, SPEC_ScrollLeft, 0, (APTR) &delta, FALSE);
     else draw = DispatchMessages(this, SPEC_NewHorizPos, 0, (APTR) &newpos, FALSE);

     if (draw)
          updatedisplay();
}

void TEditFile::EvVScroll(uint scrollCode, uint thumbPos, HWND hWndCtl)
{
     ULONG newpos = editfile.display_begin.line_num;
     LONG delta = 0;
     BOOL draw = FALSE;

     switch (scrollCode)
     {
     	  case SB_BOTTOM:
               newpos = editfile.num_lines;
               break;
	  case SB_ENDSCROLL:
               break;
	  case SB_LINEDOWN:
               delta = 1;     break;
          case SB_LINEUP:
               delta = -1;    break;
          case SB_PAGEDOWN:
               delta = editfile.wincheight;       break;
          case SB_PAGEUP:
               delta = -1 * editfile.wincheight;  break;
          case SB_THUMBPOSITION:
          case SB_THUMBTRACK:
               newpos = thumbPos;  break;
          case SB_TOP:
               newpos = 0;
               break;
     }

     if (delta)
          draw = DispatchMessages(this, SPEC_ScrollUp, 0, (APTR) &delta, FALSE);
     else draw = DispatchMessages(this, SPEC_NewVertPos, 0, (APTR) &newpos, FALSE);

     if (draw)
          updatedisplay();
}

void TEditFile::EvTimer(uint timerId)
{
     // block movement, cursor blinking
     if (editfile.active)
     {
          TClientDC dc(GetHandle());

          // block selection
          if (mousecaptured)
          {
               TPoint mouse;

               GetCursorPos(mouse);
               ScreenToClient(mouse);
               if ( (mouse.X() < 0) || (mouse.X() > editfile.winwidth) ||
                    (mouse.Y() < 0) || (mouse.Y() > editfile.winheight)
                  )
                    EvMouseMove(MK_LBUTTON, mouse);
          }

          // cursor blinking
          editfile.ticks++;
          if (editfile.ticks >= IDT_CURSORTICKS)
          {
               DrawCursor(this, dc, editfile.last_cursor_pos);
               editfile.ticks = 0;
          }
     }

     // autosave function
     if (GETGLOBALPREFS->autosave)
     {
          editfile.saveticks++;
          if (editfile.saveticks > IDT_TICKSPERMINUTE * (ULONG) GETGLOBALPREFS->autosave_interval)
          {
               char buf[2*FILENAME + 64];

               (void) CmSaveTempFilename(CreateAutosaveName(buf, editfile.name), FALSE);
                    
               editfile.saveticks = 0;
          }
     }
}

// virtual TWindow functions
bool TEditFile::CanClose()
{
     EditFile *doc = &editfile;

     if (NOT doc->changes) return TRUE;

     return ClearEditFile(TRUE);
}

void TEditFile::Paint(TDC& dc, bool erase, TRect& rect)
{
     TWindow::Paint(dc, erase, rect);

     // get the font, if it is not available
     if ( (editorfont == NULL) && openfont &&
          Prefs.fontname && strlen(Prefs.fontname) && (Prefs.fontheight > 0))
     {
          TEXTMETRIC fontmetrics;

          editorfont = new TFont(Prefs.fontname, -1 * MulDiv((int) Prefs.fontheight, dc.GetDeviceCaps(LOGPIXELSY), 72));
          if (editorfont)
          {
          	editorfont->GetTextMetrics(fontmetrics);
                if (fontmetrics.tmPitchAndFamily & TMPF_FIXED_PITCH)
                {
                     delete editorfont;
                     editorfont = NULL;
                     openfont = FALSE;
                }
          }
     }

     if (editorfont)
          dc.SelectObject(*editorfont);
     else dc.SelectStockObject(SYSTEM_FIXED_FONT);
     UpdateFontInfo(this, dc);

     if (rect.IsNull()) return;

     // PC 20020420 - update scrollers if necessary
     if (RedrawEditorRect(this, dc, rect, editfile.active))
          updatedisplay(TRUE);
}

void TEditFile::SetupWindow(void)
{
     TWindow::SetupWindow();

// set window parameters
     SetStyle(Attr.Style | WS_VSCROLL);

     SetWindowFont((HFONT) GetStockObject(SYSTEM_FIXED_FONT), TRUE);

// get timer object
     existtimer = (SetTimer(IDT_TIMER, IDT_TIMERINVERVAL)) ? TRUE : FALSE;
}



//////////////////////
// class TEditorApp //
//////////////////////

#ifndef X4standalone

class TEditorApp : public TApplication {
     public:
          TEditorApp() : TApplication("BD Programmers' Editor") {}
          void InitMainWindow();
};

void TEditorApp::InitMainWindow()
{
     MainWindow = new TFrameWindow(0, "Editor", new TEditFile);
     MainWindow->AssignMenu(IDM_EDITOR_MENU);
     MainWindow->Attr.AccelTable = TResId(IDK_EDITOR_KEYBOARD);
}

int OwlMain(int /*argc*/, char* /*argv*/ [])
{
  return TEditorApp().Run();
}

#endif

