//----------------------------------------------------------------------------
// BDEditor 2.2 - simplified ASCII only version
// Copyright (c) 2000, Pavel Cizek
// Main header file
//----------------------------------------------------------------------------

// Includes //

// Windows
#include <owl/applicat.h>
#include <owl/findrepl.h>
#include <owl/framewin.h>
#include <owl/inputdia.h>
#include <owl/opensave.h>
#include <owl/scroller.h>
#include <owl/validate.h>
#include <owl/window.h>
#include <winsys/color.h>

// C
#include <ctype.h>
#include <dir.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <sys/stat.h>

// Editor types
#include "wed_types.h"

// Editor constants
#include "wed_strings.h"
#include "wed_tags.h"

// Editor structures
#include "wed_structures.h"

#ifdef X4standalone
#include "machine.h"
#include "property.h"
#endif



/////////////////////
// TEditFile Class //
/////////////////////

class TEditFile : public TWindow
{
// public methods
     public:
// not good - use separate function for it	  TEditFile(char *filename, TWindow *parent = 0);
	  TEditFile(TWindow *parent = 0);
          ~TEditFile();

          // Xplore interface
          void SetFilter(char *filter);
          const char *GetFileName(BOOL full);
          bool IsModified(void);
          ULONG GetNumLines(void);
          ULONG GetCursorLine(void);
          ULONG GetCursorCol(void);
          ULONG GetCurrentLineLength(void);
          ULONG GetTextSize(BOOL all, BOOL line, BOOL max);
          void GetEditFileText(char *buffer, ULONG buflen, BOOL all, BOOL line);
          void InsertEditFileText(char *text, BOOL overblock);
#ifdef X4standalone
          void TEditFile::setfilepos(void);
#endif
          void RedrawWindow(void);
          char *CreateAutosaveName(char *buf, char *name);
          int GetCursorChar(void);
          char *GetCursorWord(void);
          BOOL SelectCursorWord(void);
          BOOL UndoingActive(void);

          // references to private data members
          inline EditFile *GetEditFile(void)
          {
          	return &editfile;
          }
          inline ReplaceState *GetReplaceData(void)
          {
          	return &replacedata;
          }
#define GETGLOBALPREFS	TEditFile::GetGlobalPrefs()
          inline static GlobalPrefs *GetGlobalPrefs(void)
          {
          	return &Prefs;
          }
          inline static void SetGlobalPrefs(GlobalPrefs &prefs)
          {
          	Prefs = prefs;
          }
#define GETPOOLHEADER	TEditFile::GetPoolHeader()
          inline static APTR GetPoolHeader(void)
          {
          	return PoolHeader;
          }
          inline static void SetPoolHeader(APTR header)
          {
          	PoolHeader = header;
          }


          // command handlers
          void CmQuit();
          void CmNew();
          void CmOpen();
          void CmOpenFilename(char *filename);
          void CmClose();
          void CmSave();
          void CmSaveas();
          BOOL CmSaveTempFilename(char *filename, BOOL ext);
          void CmReload();

          void CmUndo();
          void CmCut();
          void CmCopy();
          void CmPaste();
          void CmDelete();
          void CmSelectAll();
          void CmCaseToggle();
          void CmCaseUpper();
          void CmCaseLower();
          void CmDeleteLine();
          void CmDuplicateLine();
          void CmSwapLines();
          void CmCompleteLibrary();
          void CmCompleteTemplate();
          void CmInsertPath();

          void CmFind();
          void CmReplace();
          void CmFindNext();
          void CmReplaceNext();
          TResult EvFindMsg(TParam1 param1, TParam2 param2);

          void CmGotoLine();
          void CmGotoGivenLine(ULONG line);
          void CmGotoColumn();
          void CmGotoGivenColumn(ULONG col);
          void CmSet1();
          void CmSet2();
          void CmSet3();
          void CmGoto1();
          void CmGoto2();
          void CmGoto3();
          
          void CmContextHelp();

          void CmKeyboard(WPARAM id);
 	  void CmKeyInsert();
          
          // command enablers
          void CeTrue(TCommandEnabler& ce);
          void CeSave(TCommandEnabler& ce);
          void CeReload(TCommandEnabler& ce);
          void CeUndo(TCommandEnabler& ce);
          void CeBlock(TCommandEnabler& ce);
          void CeNotBlock(TCommandEnabler& ce);
          void CePaste(TCommandEnabler& ce);
          void CeSwapLine(TCommandEnabler& ce);
          void CeFind(TCommandEnabler& ce);
          void CeReplaceNext(TCommandEnabler& ce);
          void CeGoto1(TCommandEnabler& ce);
          void CeGoto2(TCommandEnabler& ce);
          void CeGoto3(TCommandEnabler& ce);

          // event handlers
          void EvActivate(uint active, bool minimized, HWND hWndOther);
          void EvKillFocus(HWND hWndGetFocus);
          void EvSetFocus(HWND hWndLostFocus);

          void EvCaptureChanged(THandle hwnd);
          void EvChar(uint key, uint repeatCount, uint flags);
          
/*** Modify Start : d-tec  ***/
          TResult EvImeChar(WPARAM wchar, LPARAM param2);
          TResult EvImeStartcomposition(WPARAM param1, LPARAM param2);
/*** Modify End : d-tec  ***/
          
          void EvLButtonDblClk(uint modKeys, TPoint& point);
          void EvLButtonDown(uint modKeys, TPoint& point);
          void EvLButtonUp(uint modKeys, TPoint& point);
          void EvMButtonDown(uint modKeys, TPoint& point);
          void EvMouseMove(uint modKeys, TPoint& point);
          void EvRButtonDown(uint modKeys, TPoint& point);
          void EvSize(uint sizeType, TSize& size);
          void EvTimer(uint timerId);
          void EvHScroll(uint scrollCode, uint thumbPos, HWND hWndCtl);
          void EvVScroll(uint scrollCode, uint thumbPos, HWND hWndCtl);

          // virtual TWindow functions
          bool CanClose();
          void SetupWindow();
          void Paint(TDC& dc, bool erase, TRect& rect);

          // response table
          DECLARE_RESPONSE_TABLE(TEditFile);

          // miscellaneous
          inline ULONG IncBusyPointer(void)
          {
               return (++busypointer);
          }
          inline ULONG DecBusyPointer(void)
          {
               if (busypointer > 0)
                    busypointer--;
               return busypointer;
          }
          ULONG GetVisibleTextWidth(void);

// public data elements
     public:
          BOOL lbuttondown;	// set if a open/save or other blocking requester is opened

          static char **hi_keywords, **hi_functions, **hi_libraries;
          static long hi_num_keywords, hi_num_functions, hi_num_libraries;

// private data elements
     private:
          EditFile editfile;
          ReplaceState replacedata;

          TFindReplaceDialog::TData searchdata; // (FR_DOWN | FR_HIDEUPDOWN, MAXWORDLEN - 10);
          TFindReplaceDialog* searchdialog;
          BOOL searchreplace;	// is dialog opened for replace?
          BOOL noreplaceyet;	// did user opened already replace?

          BOOL initfinished;
          BOOL existtimer;
          BOOL mousecaptured;

          ULONG busypointer;	// indicates how many times was pointer set to busy

          BOOL undoing_active;  // is undo in progress?

// private functions
          void updatedisplay(BOOL frompaint = FALSE);	// PC 20020420 - added parameter for Paint
          void DoFindReplace(BOOL fromdialog);
          BOOL TEditFile::ClearEditFile(BOOL ask);
          char *searchstringlists(char *word, char *statusbuf, ULONG statuslen, int *foundstat = NULL);
#ifdef X4standalone
          void InitFromIniFile(void);
#endif
          BOOL DoUndoOperation(void);

// protected data
     protected:
	  static GlobalPrefs Prefs;
          static APTR PoolHeader;
          static ULONG NumEditedFiles;
#ifdef X4standalone
          static BOOL inifile_read;
#endif
          static TFont *editorfont;
          static BOOL openfont;
};

//////////////////////
// END of TEditFile //
//////////////////////


// Editor functions
#include "wed_compat.h"
#include "wed_protos.h"

