/*
 * BDT-Lists.c      modul s fcemi pro praci se seznamy tykajicimi se BDT
 */

#define BDT_LISTS_C

#include "wed_editor.h"



/* globalni promenne tohoto modulu */
/* nasledujici pole obsahuje velikosti vsech pouzivanych struktur */
// unsigned int = size_t type!
static ULONG structure_size[] =
//>>
{ 0,
  4,                     4,                      sizeof(MemoryBlock),
  4,                     4,                      4,                     
  4,                     4,                      sizeof(Location),
  4,                     sizeof(Paragraph),      4,                     
  4,                     4,                      4,                     
  4,                     4,                      4,                     
  4,                     4,                      4,                     
  4,                     4,                      4,                     
  4,                     4,                      4,                     
  sizeof(GlobalPrefs),   sizeof(EditFile),       4,                     
  4,                     sizeof(ReplaceState),   4,                     
  4,                     4,                      4,                     
  4,                     4,                     
  0
};
//<<



// FCE PRO INICIALIZACI A UVOLNOVANI STRUKTUR //

/* nasledujici fce alokuje pamet pro zadanou strukturu */
APTR InitStructure(ULONG id, APTR init_val)
//>>
{
     APTR object = LibAllocPooled(GETPOOLHEADER, structure_size[id]);

     if (object && init_val) 
          memcpy(object, init_val, (size_t) structure_size[id]);

     return(object);
} /* end InitStructure */
//<<

/* nasledujici fce uvolni zadanou strukturu z pameti; pokud ma struktura
 * prvni polozku next (uk. na dalsi strukturu), pak tato funkce ji umi
 * vratit v promenne next */
APTR FreeStructure(ULONG id, APTR structure)
//>>
{
     APTR pom = *((APTR *) structure);
     
     if (structure) LibFreePooled(GETPOOLHEADER, structure, structure_size[id]);

     return(pom);
} /* end FreeStructure */
//<<

// DALSI POMOCNE FCE PRO VKLADANI A MAZANI ZE SEZNAMU STRUKTUR //

/* nasleduji fce pro vytvareni/vypousteni MemoryBlock i s pameti */
MemoryBlock *AddNewMemoryBlock(TEditFile *teditfile, MemoryBlock *kam, ULONG size)
//>>
{
     EditFile *doc = teditfile->GetEditFile();

     MemoryBlock *newmb = (MemoryBlock *) InitStructure(MEMORYBLOCK, NULL);
     MemoryBlock *next = (kam) ? kam->next : doc->block_first;

     if (newmb == NULL) return(NULL);
     if ( NULL == (newmb->memory = (UBYTE *) LibAllocPooled(GETPOOLHEADER, size)) )
     {
          FreeStructure(MEMORYBLOCK, newmb);
          return(NULL);
     }

     newmb->prev = kam;            newmb->next = next;
     if (kam)  kam->next  = newmb; else doc->block_first = newmb;
     if (next) next->prev = newmb; else doc->block_last  = newmb;
     newmb->free = size;           newmb->size = size;

     return(newmb);
}
//<<

/* opravy struktur ukazujicich na dany MemoryBlock musi delat volajici fce */
MemoryBlock *RemoveMemoryBlock(TEditFile *teditfile, MemoryBlock *mb)
//>>
{
     EditFile *doc = teditfile->GetEditFile();

     MemoryBlock *prev = mb->prev, *next = mb->next;
#ifdef CHECK_BUGS
     if ((mb == NULL) || (mb->memory == NULL)) InternalError("Modul Lists - RemoveMemoryBlock, mb || mb->memory = NULL");
     if ((mb->prev == NULL) && (mb->next == NULL)) InternalError("Modul Lists - RemoveMemoryBlock, next = prev = NULL");
#endif
     if (mb == NULL) return(NULL);

     if (prev) prev->next = next; else doc->block_first = next;
     if (next) next->prev = prev; else doc->block_last  = prev;
     if (mb->memory)
     	  LibFreePooled(GETPOOLHEADER, mb->memory, mb->size);
     FreeStructure(MEMORYBLOCK, mb);

     return(next);
}
//<<

/* nasleduji fce pro vytvoreni/vypusteni struktury noveho odstavce;
 * musi uz nejaky odstavec existovat ! */
/* length, lines, topedge - inicializace na 0;
 * LineNode a MemoryBlock (first i last) budou hodnoty nasledujici za
 * prislusnymi polozkami prev odstavce - tj. MemoryBlock i LineNode nasledujici musi existovat !
 * Pokud ovsem nastavime set_mb_line = FALSE - pak se nic nedosazuje */
// nps - jaky se ma priradit styl
// updates num_lines in EditFile!
Paragraph *AddNewParagraph(TEditFile *teditfile, Paragraph *prev, /* ParaStyle *ps, */ BOOL set_mb_line /*, NamedParaStyle *nps*/)
//>>
{
     EditFile *doc = teditfile->GetEditFile();

     Paragraph *newpp = (Paragraph *) InitStructure(PARAGRAPH, NULL);
     Paragraph *next = prev->next;
//     TextStyle *ts = InitStructure(TEXTSTYLE, &ps->ts);

//#ifdef CHECK_BUGS
//     if ((ps == NULL) || (prev == NULL)) InternalError("Modul Lists - AddNewParagraph, ps == NULL || prev == NULL");
//#endif
     if ((newpp == NULL) || (prev == NULL))
//     if ((newpp == NULL) || (ts == NULL) || (ps == NULL) || (prev == NULL))
     {    // FreeStructure(TEXTSTYLE, ts); 
          FreeStructure(PARAGRAPH, newpp);
	  return(NULL);
     }
// zapojeni do seznamu
     newpp->prev = prev;           newpp->next = next;
     prev->next = newpp;
     if (next) next->prev = newpp; else doc->para_last = newpp;
     doc->num_lines++;
// nastaveni ukazatele do pameti (first = last)
     newpp->block_first = prev->block_last;
     newpp->off_first   = prev->off_last;
     if (set_mb_line)
	  (void) IncrementPosMB(&newpp->block_first, &newpp->off_first);
     newpp->block_last  = newpp->block_first;
     newpp->off_last    = newpp->off_first;
// length, lines, topedge = 0
// nastaveni parastyle, textstyle - POZOR ! nepouzili jsme fci na vytvoreni
// noveho textoveho stylu !
//     newpp->style = ps;       ps->used++;    // musi se zvetsit used !
//     newpp->nps_para = nps;
//     newpp->ts_first = newpp->ts_last = ts;
//     ts->offset = 0;     ts->prev = NULL;    ts->next = NULL;
//     ts->font->used++;        // musi se zvetsit used i pro font !
// LineNode
//     if (prev->line_last)
//     {    newpp->line_first = prev->line_last;
//// neexistuje           newpp->ld_first   = prev->ld_last;
//	  if (set_mb_line)
//	  {
//#ifdef CHECK_BUGS
//	       if (NOT NextPosLN(&newpp->line_first)) InternalError("Modul Lists - AddNewParagraph, no next ln");
//#else
//	       (void) NextPosLN(&newpp->line_first);
//#endif
//	  }
//	  newpp->line_last  = newpp->line_first;
//// neexistuje           newpp->ld_last    = newpp->ld_first;
//     }
     return(newpp);
}
//<<

/* uvolneni odstavce ze seznamu - free_ts umoznuje zachovat vsechny jeho TextStyles */
/* pokud je procedeno uvolneni seznamu stylu -> volajici fce musi zaridit
 * opravy struktur apod. ! */
// updates num_lines in EditFile!
Paragraph *RemoveParagraph(TEditFile *teditfile, Paragraph *para, BOOL free_ts)
//>>
{
     EditFile *doc = teditfile->GetEditFile();

     Paragraph *prev = para->prev, *next = para->next;
//     TextStyle *ts;
#ifdef CHECK_BUGS
     if (para == NULL) InternalError("Modul Lists - RemoveParagraph, para = NULL");
     if (doc->num_lines <= 1) InternalError("Modul Lists - RemoveParagraph, num_lines <= 1");
#endif
     if (para == NULL) return(NULL);

//// uvolneni text style - musime odzadu
//     if (free_ts)
//	  for(ts = para->ts_last; ts; ) // neopravujeme nic
//	       ts = RemoveTextStyleP(doc, para, ts, NULL, FALSE);
// oprava seznamu
     if (prev) prev->next = next; else doc->para_first = next;
     if (next) next->prev = prev; else doc->para_last  = prev;
     doc->num_lines--;
// uvolneni
//     para->style->used--;          // uz nebude timto odstavcem pouzivan
//     if (para->style->used == 0) RemoveParaStyle(doc, para->style);
     FreeStructure(PARAGRAPH, para);

     return(next);
}
//<<

