/*
 *   Structures.h    header file with definitions of basic editor structures
 */

#ifndef BDT_DEFS_STRUCTURES_H
#define BDT_DEFS_STRUCTURES_H

/* definice typu chyby */
typedef ULONG ERRORCODE;


////////////////////////////////////////////////
// DEFINICE ZAKLADNICH VYCTOVYCH TYPU A POJMU //
////////////////////////////////////////////////

/* HLAVNI VELICINY URCUJICI MOZNOSTI V DANYCH OBLASTECH */

#define PATH_LEN         MAXDIR // delka cesty, kterou jeste akceptujeme - dir.h
//#define PATH_LEN         640  // delka cesty, kterou jeste akceptujeme
//#define MACROFILE_LEN    80   // delka cesty k makrum
//#define PATTERN_LEN      64   // delka masky ...
#define MAXSEARCHSTR     256  // 40   // nejdelsi slovo, ktere lze vyhledavat/nahrazovat
#define MAXWORDLEN       256  // 64   // maximalni delka slova, ktere budeme akceptovat
                              // musi byt vetsi nez FILENAME!
                              // must be bigger than FindReplace string (81 by default)

///* POMOCNE KONSTANTY OBECNEHO VYZNAMU */

#define FILENAME         MAXFILE  // see dir.h
//#define FILENAME         256  // 30   // velikost jmena souboru
#define COMMENTSIZE      80   // velikost komentare

// SPECIALNI ZNAKY A SYMBOLY V NASI REPREZENTACI TEXTU //
#define ESC              "\33"          // Escape - pro tisk
#define LF               "\12"          // LineFeed - pro tisk, jako retezec
#define FF               "\14"          // FormFeed - pro tisk, jako retezec

#define CH_TAB           '\t'           // tabelator = 9
#define CH_EOL           '\12'          // konec radky = 10
#define CH_EOP           '\14'          // konec stranky = 12 (= page break)
#define CH_EOLD          '\15'          // konec radky s delenim = 13

#define CH_SPACE         ' '            // mezera
#define CH_PUNCT         '.'            // tecka
#define CH_COMMA         ','            // carka
#define CH_HARDSPACE     ''            // tvrda mezera 160

// obecne pomocna makra
// v clib/macros.h - #define ABS(x) ( ((x) >= 0) ? (x) : -(x) )
#define SGN(x) ( ((x) > 0) ? 1 : ((x) < 0) ? -1 : 0 )



////////////////////////////////////
// DEFINICE STRUKTURY COMMANDITEM //
////////////////////////////////////

typedef ULONG Command;



//////////////////////////////////////////
// STRUKTURY KE SPRAVE PAMETI PRO TEXTY //
//////////////////////////////////////////

#define MAXBLOCKSIZE     30000     // omezime maximalni delku bloku

#define PUDDLESIZE       64000
#define TRESHSIZE        32000

#define SIZE_TEXT        4096      // mozne hodnoty mezi 5K - 10K - otestovat nejrychlejsi
//#define SIZE_HEADER      1024      // hlavicky 
#define BLOKPOM_VELIKOST (SIZE_TEXT)// pro vyrovnavaci buffery pri IO

#define FREEMEM_MIN      4    // mozna hodnota i nula
#define FREEMEM_MAX      512  // souvisi take s velikosti bloku !
#define FREEMEM_INITIALSTATE  258       // std hodnota pro volnou pamet
#define FREEMEM_CORRECT       254       // pocet bytu pro preneseni pri preteceni
                                        // daneho pametoveho bloku

// MemoryBlock - popisuje blok pameti pro text //
struct MemoryBlock
{
     struct MemoryBlock *next, *prev;
     UBYTE *memory; // ukazatel na pamet pro data
     ULONG size;    // velikost pameti pro data - viz vyse
     ULONG free;    // volne misto v bloku - posledni blok - jasne
                    // normalni blok - mezi hodnotami FREEMEM_MIN a FREEMEM_MAX
                    // presahne-li se jeden limit -> zarovna se to tak, aby
                    // (jde-li to) bylo free prumer FREEMEM_#?
     ULONG user;    // misto pro nejaka pripojena data
     ULONG user2;   // dalsi data...
     BOOL changed;  // urcuje, zda se s nim manipulovalo ...
//     BOOL last;     // udava, zda je to posledni blok, s kterym se manipulovalo ...
};
typedef struct MemoryBlock MemoryBlock;



//////////////////////////////////////////////////
// STRUKTURY POPISUJICI DOKUMENT - PRO EDITFILE //
//////////////////////////////////////////////////

// Location - popisuje pozici neceho v souboru //
struct Location
{
     struct Location *next;        // nekde se to bude hodit
     struct MemoryBlock *block;    // pametovy blok
     struct Paragraph *para;       // odstavec
//     struct Page *page;            // strana
//     struct TextStyle *style;      // styl pisma v danem miste
//     struct LineNode *line;        // rozmery radky
     ULONG block_off;              // offset od pocatku pametoveho bloku
//     MUNIT page_off;               // vzdalenost od pocatku stranky v MUNIT
//     MUNIT para_off;               // vzdalenost od pocatku odstavce v MUNIT
     ULONG chars_off;              // vzdalenost od pocatku odstavce ve znacich
     ULONG line_num;		   // number of line (0,...) = number of the paragraph
};
typedef struct Location Location;

#define SC_PAGEBREAK     12        // FORM FEED
#define SC_TAB           9         // TABELATOR

// Paragraph - popisuje odstavec //
// POZOR - konce radku budou znaceny '\n', v pripade, ze se v tom miste deli
// slovo, tak znakem '\r' = CH_EOLD
struct Paragraph
{
     struct Paragraph *next, *prev;     // nasledujici a predchozi odstavec
          // pametove bloky, v nichz je odstavec ulozen
     struct MemoryBlock *block_first, *block_last;
          // offsety v blocich pameti - prvni a posledni ukazuji na prvni a
          // _na_ posledni pismeno odstavce; bude-li omezena delka
          // MemoryBlock -> staci definovat typ UWORD, jinak ULONG
     ULONG off_first, off_last;
     ULONG length;       // delka odstavce ve znacich
     ULONG explength;    // PC 20020420: expandovana delka odstavce ve znacich (tabelatory se expanduji)
//     ULONG lines;        // pocet radku odstavce - should be always 1 now!

//     struct ParaStyle *style;      // "layout" odstavce, parametry
//     struct TextStyle *ts_first, *ts_last;

//     MUNIT topedge;      // poloha v MUNIT od zacatku dokumentu
//        // ukazatele na uzly pro prvni a posledni radek odstavce
//     struct LineNode *line_first, *line_last;

     BOOL changed;  // urcuje, ze probehly nejake jinak nezjistitelne zmeny

//     struct NamedParaStyle *nps_para;   // pojmenovany styl
};
typedef struct Paragraph Paragraph;



//////////////////////////////////////////////////////
// DEFINICE STRUKTUR POPISUJICICH NASTAVENI EDITORU //
//////////////////////////////////////////////////////

#define MAXAUTOWORDS     256
#define MAXAUTOPAIRLEN   50
#define MAXSAVEAUTOPAIRLEN   54
// Hlavni struktura obsahujici vsechny nastaveni a volby uzivatele //
struct GlobalPrefs
{
     BOOL crlf_eols;	// are EOLs saved as crlf?
     
  /* toto jsou default hodnoty, ktere se nastavi po startu; vsechny
   * (s vyjimkou keymap) jsou vsak specificke pro dany soubor */
//     BOOL show_invisibles, show_guides, show_pictures; // viz Settings/Show
//     BOOL show_invisibles; 
//     BOOL show_headers, show_footers;   // Document/Header/show ...
//     BOOL ruler_horiz, ruler_vert;      // viz Settings/Ruler
     BOOL insert;                       // viz Settings/Insert mode
//  /* urcuje klavesnici - ta bude na rozdil od predesleho globalni */
//     enum CurrentKeyMap keymap;         // viz Settings/KeyMap
//  // standardni zoom v dokumentu, standardne je to 100 - zatim nevyuzito
////     UWORD doc_zoom;     // zoom dokumentu, je to v procentech (25% - 400%)

//  /* nasledujici polozky standardni hodnoty pro layout, stranku a odstavec;
//   * kazdy dokument vsak obsahuje tyto informace sam o sobe */
//     struct TFont def_font;
//     struct ParaStyle def_parastyle;
//     struct Layout def_layout;
//     struct PageSize def_pagesize;

  /* nasledujici polozky (struktury) budou obsahovat aktualni nastaveni ruznych
     parametru a rozdelenim budou odpovidat polozkam v menu SETTINGS ... */
//  /* prvni tri polozky jsou charakteristicke pro kazdy dokument */
//     struct DisplayPrefs display;
//     struct ItemFormats def_items;
//     struct TabPrefs def_tabs;
     UWORD tab_size;
//  /* nasledujici polozky jsou jiz globalni nastaveni editoru */
//     struct GeneralPrefs general;
//     struct LocalePrefs locale;
//     struct IOPrefs ioprefs;
     BOOL backup;
     BOOL autosave;
     UWORD autosave_interval;

     // how many steps is remembered by undo
     UWORD undo_steps;

   /////
//     struct ScreenPrefs screen;
//     struct PalettePrefs palette;
   /////
//     struct SpellerPrefs speller;
//  /* a jeste makra prirazena do menu MACROS */
//     char macros[NUMMACROS][MACROFILE_LEN + 2];
//  /* nasledujici polozka charakterizuje nastaveni pro tisk; tato nastaveni
//     jsou platna globalne */
//     struct PrintPrefs print;
//  /* nasledujici struktura (bude-li vyuzita) bude obsahovat polohy requestru,
//     ktere si uzivatel pri manipulaci s nimi nastavil */
//     struct ReqPrefs file, font, screen_mode;
//     char def_req_path[PATH_LEN + 2], def_req_pattern[PATTERN_LEN + 2];
     char def_openreq_path[PATH_LEN + 2];
     char def_savereq_path[PATH_LEN + 2];
     char def_filter[FILENAME + 2];
  /* slova pro automaticke opravy - v kazdem retezci je dvojice slov,
     ktera jsou oddelena znakem 187  */
//     UBYTE autocorr_words[MAXAUTOWORDS + 1][MAXSAVEAUTOPAIRLEN];
//     BOOL autocorrect;

//  /* nova charakteristika dokumentu - styl indexu */
//     struct IndexPrefs def_index;
//  /* preference nastrojove listy */
//     struct ToolbarPrefs toolbar;
//  /* preference pro cislovani */
//     struct NumberFormats def_numformat;
//  /* barva titulku pro dripens */
//     UWORD title_color;       // barva titulku oken apod. - WHITE or BLACK

     // colors, elements correspond to highlighting constant for foreground and background
// highlight constants
#define EC_HI_DEFAULT         0
#define EC_HI_TEXT            1    // normal text
#define EC_HI_STRING          2    // "ahoj", 'hello'
#define EC_HI_NUMBER          3    // 100, 1e-10
#define EC_HI_COMMENT         4    // ;, // 
#define EC_HI_OPERATORS       5
#define EC_HI_BRACKETS        6
#define EC_HI_KEYWORDS        7    // proc, endp, if, else, endif, switch, case, endws, default, case, while, endo 
#define EC_HI_FUNCTIONS       8    // built-in functions: atof, max, ...
#define EC_HI_LIBRARIES       9    // external procedures: quantlets lts, ...
#define EC_HI_ILLEGALCHAR     10  
#define EC_HI_NUMITEMS        11

#define COLOR_NOT_DEFINED     0xFFFFFFFF
     
// macro for creating palette entries
#define MAKERGB(r, g, b)      ( (r & 0xFFL) | ((g & 0xFFL) << 8L) | ((b & 0xFFL) << 16L) )

     long highlight_foreground[EC_HI_NUMITEMS];
     long highlight_background[EC_HI_NUMITEMS];
     BOOL do_highlighting;         // if TRUE, highlighting is active

// fonts
     char fontname[FILENAME + 4];  // font name for the edit window
     UWORD fontheight;		   // height of the requested font
};
typedef struct GlobalPrefs GlobalPrefs;



////////////////////////
// STRUKTURA EDITFILE //
////////////////////////

#define NUM_UNDO_STEPS	64
#define UNDO_INSPARA     ~0UL  // byl vlozen konec odstavce, ne text

// struktura Undo - popis vsech informaci pro porvadeni teto operace
struct Undo
{
     UWORD undo_command;    // prikaz, ktery se ma odvolat
          // je-li COMM_Dummy - zadne undo neni k dispozici
     UWORD undo_clip;    // identifikator clip jednotky pro tento dokument
     STRPTR buf;	 // pointer to the data for undo, they always form a string
     ULONG bufsize;      // size of the data

// position of cursor before and after the command
     ULONG para_before, para_after;	  // number of line = paragraph
     ULONG charoff_before, charoff_after; // character offset from the BOL

// position of the begin and the end of inserted or modified text
     ULONG para_beg, para_end;	        // number of line = paragraph
     ULONG charoff_beg, charoff_end;    // character offset from the BOL
     BOOL block;                        // byl blok oznacen?
};
typedef struct Undo Undo;

// EditFile - hlavni struktura popisujici editovany soubor       //
//            budeme mit pole o MAXFILES - ukazatele na EditFile //
struct EditFile
{
  /* polozky tykajici se AmigaDOSu - ulozeni apod. */
     char path[PATH_LEN + FILENAME + 16];     // cesta k souboru  (including the name of file?)
     char name[FILENAME + 2];   // jmeno souboru
//     char comment[COMMENTSIZE];    // zatim asi nevyuzito
     ULONG opened;            // cas otevreni (v sekundach)
     ULONG last_saved;        // cas posledniho ulozeni (v sekundach);
          // je-li nula --> jeste se neukladal

  /* dalsi polozky se tykaji primo editovaneho souboru - ulozeni a struktury */
          // specifikace ulozeni souboru v pameti; vzdy alespon 1 blok pritomen
     struct MemoryBlock *block_first, *block_last;
          // jednotlive odstavce souboru
     struct Paragraph *para_first, *para_last;

// this constant is equal to the number of Location structures below
#define NUM_DOC_LOCATIONS     	9	// originally: 7 & one bookmark
#define DOC_LOCATION_CURSOR		0
#define DOC_LOCATION_CURSORLINE	1
  /* nasledujici struktury Location popisuji polohy dulezitych bodu v textu */
     struct Location cursor;       // kurzoru
     struct Location cursorline;   // zacatek radky s kurzorem
     struct Location bookmark;     // zalozka
     struct Location bookmark2;     // zalozka
     struct Location bookmark3;     // zalozka
     struct Location block_begin, block_end;      // blok - zacatek a konec
          // ZACATEK prvni a posledni zobrazovane radky;
          // pokud je v display_#? nastaveno vse na 0, neni nic videt z textu
     struct Location display_begin, display_end;

//  /* ciselne udaje o editovanem souboru */
     ULONG num_lines;
     ULONG num_columns;
     
/* a polozky urcujici stav toho, co se edituje */
     BOOL active;        // soubor je aktivni, je prave editovan !
                         // spravne nastaveni je dulezite pro ostatni funkce
     BOOL block;         // je aktivni blok (platny)
     BOOL newfile;         // soubor je novy a neposkozeny (lze do nej nahrat neco)
     BOOL changes;       // v souboru byly provedeny zmeny
     BOOL undo;          // je platne undo

  /* dale jsou uvedeny ruzne souradnice a rozmery viditelnych casti textu */
  /* a zobrazovanych objektu, jako kurzor, aktualni radek atd.            */
     int fontwidth, fontheight;  // dimensions of SYSTEM_FIXED_FONT
     int winwidth, winheight;    // width and height of the window in pixels
     ULONG wincwidth, wincheight; 	// and in chars (SYSTEM_FIXED_FONT)
     ULONG disp_offset;          // offset of the display = which column is displayed
     										// as the leftmost one; 0 = first column displayed
     // pozice kurzoru - mela by odpovidat zacatku pismene, na nejz
     // ukazuje --> z toho je pak jasne, co a kam se posune a zobrazi
     // pri vkladani/mazani ...; kurzor ma konstantni tloustku
     // rozmery jsou absolutni od okraje stranky, bez layoutu!
     BOOL cursor_checked;               // je hotova kontrola viditelnosti kurzoru ?
     BOOL cursor_visible;               // je kurzor nakreslen ?
     ULONG ticks;   // pocet tiku -> z toho se urci, kdy ma kurzor bliknout
//     UBYTE refresh_ticks;// slouzi k urceni, kdy uz se ma obnovit obrazovka
     ULONG saveticks;    // slouzi k testovani pro funkci autosave
     ULONG last_cursor_pos;	// DrawCursor save here the column in which was cursor drawn

/* polozky z Prefs, ktere jsou charakteristicke pro kazdy soubor - viz Prefs */

     // nastaveni zobrazovani a psani; musi nasledovat za sebou !
     BOOL show_invisibles;
     BOOL insert;
     UWORD tab_size;

// pomocne promenne pro znaceni bloku
/* pomocna promenna urcujici, zda se prave voli blok */
     BOOL block_select;  // std je FALSE - inicializuje se spravne
     BOOL zamena_bb_be;  // std je FALSE - inicializuje se spravne
     Location blockpoint;     // urcuje "pevny bod" pri znaceni bloku; nastavuje volajici funkce

     // informace pro Undo
     struct Undo undo_info[NUM_UNDO_STEPS];  // contain the undo steps definition
     ULONG undo_steps;	// contain the number of undo steps in the buffer

     // some previously global variables
     BOOL no_separator_separator;
};
typedef struct EditFile EditFile;



/////////////////////////////////////////////////
// STRUKTURY PRO POPIS GLOBALNICH OKEN EDITORU //
/////////////////////////////////////////////////

// ReplaceState - popisuje stav okna pro nahrazovani //
// str ma delku MAXSEARCHSTR + MAXWORDLEN, aby se dal pouzit
// i pri kontrole pravopisu
struct ReplaceState
{
     UBYTE str1[MAXSEARCHSTR + MAXWORDLEN];      // hledany retezec
     UBYTE str2[MAXSEARCHSTR + MAXWORDLEN];      // novy retezec
     BOOL case_sens;     // case sensitive ?
     BOOL whole_word;    // jenom cela slova vyhledat ?
     BOOL forward;       // vyhledavat vpred ci vzad
     BOOL from_beg;      // vyhledavat od pocatku ci od kurzoru
//     NamedParaStyle *nps;// styl, ve kterem se ma hledat; je-li NULL
//                       // hleda se vsude
     BOOL success;	 // can contain a return value:
                         // was string found / replaced (TRUE) or not (FALSE)
};
typedef struct ReplaceState ReplaceState;



///////////////////////////////////////////////////////////////////////
// DEKLARACE IDENTIFIKATORU PRO JEDNOTLIVE STRUKTURY - INITSTRUCTURE //
///////////////////////////////////////////////////////////////////////

/* identifikatory jednotlivych struktur */
// byvalo #define COMMANDITEM           1
//#define MESSAGEITEM           2
#define MEMORYBLOCK           3
// byvalo #define EXTLINE               4
//#define TAB                   5
//#define TFONT                 6
//#define TEXTSTYLE             7
//#define PARASTYLE             8
#define LOCATION              9
//#define LINENODE              10
#define PARAGRAPH             11
//#define PAGE                  12
//#define HEADER                13
//#define PICTURE               14
//#define PAGESIZE              15
//#define LAYOUT                16
//#define PRINTPREFS            17
//#define ITEMFORMATS           18
//#define TABPREFS              19
//#define GENERALPREFS          20
//#define LOCALEPREFS           21
//#define IOPREFS               22
//#define DISPLAYPREFS          23
//#define SCREENPREFS           24
//#define PALETTEPREFS          25
//#define SPELLERPREFS          26
//#define REQPREFS              27
#define GLOBALPREFS           28
#define EDITFILE              29
//#define INSERTLETTER          30
//#define SEARCHSTATE           31
#define REPLACESTATE          32
//#define SPELLERSTATE          33
//#define FRAME                 34
//#define VECTOROBJECT          35
//#define NAMEDPARASTYLE        36
//#define INDEXPREFS            37
//#define TEXTINFO              38

#define NUM_STRUCTS           38

#endif /* BDT-Defs-Structures.h */

