/* $Header: /home/xplore/cvs/xpl4gr/xarray.C,v 1.3 2002/08/16 12:31:54 sigbert Exp $ */

#include "xarray.hpp"
#include "x4math2.h"

#ifdef HAS_STRCMPI
#define strcasecmp strcmpi
#endif

template<class T> T *vecsort (T *lptr, long rlen, long rdim, long ndir, long *dir)
{ long *sortorder, *iptr, *irange, *nrange, i, j, k, maxsort, 
       iszero, ranges, m, l;
  T *destmat, *subptr;
  XplSort<T> xs;

#ifndef MEMDEBUG
  sortorder = new long[rdim];
  iptr      = new long[rlen];
  irange    = new long[rlen+2];
  nrange    = new long[rlen+2];
  destmat   = new T[rlen*rdim];
#else
  sortorder = (long *) xmem.newptr (sizeof(long), rdim, "XplArray", "vecsort", "sortorder");
  iptr      = (long *) xmem.newptr (sizeof(long), rlen, "XplArray", "vecsort", "iptr");
  irange    = (long *) xmem.newptr (sizeof(long), rlen+2, "XplArray", "vecsort", "irange");
  nrange    = (long *) xmem.newptr (sizeof(long), rlen+2, "XplArray", "vecsort", "nrange");
  destmat   = (T *) xmem.newptr (sizeof(T), rlen*rdim, "XplArray", "vecsort", "destmat");
#endif
  
  for (i=0; i<rlen*rdim; i++)
    SetNaN (destmat+i);

  maxsort = 0;
  iszero  = 0;
  irange[0] = 0;
  irange[1] = rlen;
  ranges    = 1;
  for (i=0; i<rlen; i++)
    iptr[i]   = i;

  for (j=0; j<ndir; j++)
  { if (maxsort>=rdim)
    { XAerrort (DIMCON, "More order columns than matrix columns");
      return (destmat);
    }
    sortorder[maxsort] = *(dir+j);
    iszero = iszero || (sortorder[maxsort]==0);
    if (labs(sortorder[maxsort])>rdim)
    { XAerrort (DIMCON, "Sorting after not existing column");
      return (destmat);
    }
    maxsort++;
  }
  if (maxsort==0) 
  { for (i=0; i<rdim; i++)
      sortorder[i] = i+1;
    maxsort = rdim;
  }
  if (!iszero)
  { for (i=0; (i<maxsort) && (ranges<=rlen); i++)
    { subptr = (lptr+rlen* (::labs(sortorder[i])-1) );
      xs.sortvec(subptr);
      for (k=0; k<ranges; k++)
      { if ((irange[k+1]-irange[k])>1)
          xs.quicksort(&iptr[irange[k]], irange[k+1]-irange[k], sortorder[i]>0);
      }
      m = 0;
      for (j=0; j<ranges; j++)
      { nrange[m] = irange[j];
        for (k=irange[j]; k<irange[j+1]; k++)
        { if (k<irange[j+1]-1)
          { if (subptr[iptr[k]]!=subptr[iptr[k+1]])
            { m++;
              nrange[m] = k+1;
            }
          }
          else
          { m++;
            nrange[m] = k+1;
          }
        }
      }
      m++;
      for (j=0; j<m; j++)
        irange[j] = nrange[j];
      irange[m] = rlen;
      ranges = m;
    }
    for (k=0; k<rdim; k++)
    { for (l=0; l<rlen; l++)
        *(destmat+k*rlen+l) = *(lptr+k*rlen+iptr[l]);
    }
  }
  else /* 0 found, sort all seperatly !!!! */
  { for (k=0; k<rdim; k++)
    { for (i=0; i<rlen; i++)
        iptr[i] = i;
      subptr = (lptr+rlen*k);
      xs.sortvec(subptr);
      xs.quicksort(iptr, rlen, 1);
      for (l=0; l<rlen; l++)
        *(destmat+k*rlen+l) = *(lptr+k*rlen+iptr[l]);
    }
  }

#ifndef MEMDEBUG
  delete[] sortorder;
  delete[] iptr;
  delete[] irange;
  delete[] nrange;
#else
  xmem.delptr (sortorder);
  xmem.delptr (iptr);
  xmem.delptr (irange);
  xmem.delptr (nrange);
  sortorder = NULL;
  iptr = NULL;
  irange = NULL;
  nrange = NULL;
#endif
  return (destmat);
}

template int *vecsort (int *lptr, long rlen, long rdim, long ndir, long *dir);
template long *vecsort (long *lptr, long rlen, long rdim, long ndir, long *dir);
template double *vecsort (double *lptr, long rlen, long rdim, long ndir, long *dir);
template xplmask *vecsort (xplmask *lptr, long rlen, long rdim, long ndir, long *dir);
template xplstring *vecsort (xplstring *lptr, long rlen, long rdim, long ndir, long *dir);

template<class T> void mattrn (T *tptr, long n, long p)
{ long i, j;
  T *tmp; 
#ifndef MEMDEBUG
  tmp = new T[n*p];
#else
  tmp =(T *) xmem.newptr (sizeof(T), n*p, "XplArray", "mattrn", "tmp");
#endif  

//  memcpy (tmp, tptr, n*p*sizeof(T));
  ::memcopy (tmp, tptr, n*p);
  for (i=0; i<p; i++)
  { for (j=0; j<n; j++)
      *(tptr+j*p+i) = *(tmp+i*n+j);    
  }
#ifndef MEMDEBUG
  delete[] tmp;
#else
  xmem.delptr (tmp);
  tmp = NULL;
#endif
}

template void mattrn (int *tptr, long n, long p);
template void mattrn (long *tptr, long n, long p);
template void mattrn (double *tptr, long n, long p);
template void mattrn (xplmask *tptr, long n, long p);
template void mattrn (xplstring *tptr, long n, long p);

  template<class T> void XplArray<T>::construct(XplArray<T> *x)
  { char out[256];
    long i;
    long entries = dim.prod();
    if (entries==0)
    { XAerror(NOMEM);
      return;
    }
#ifdef SMALLVERSION
    if (dim(0)>SMALLVERSIONOBSERVATIONS) XAerror(NOSUPPORT);     // PC 20011127 changed to SMALLVERSIONOBSERVATIONS
#endif
#ifndef MEMDEBUG
    x->ptr = new T[entries];
#else
    x->ptr = (T *) xmem.newptr (sizeof(T), entries, "XplArray", "construct", "x->ptr");
#endif
    if (x->ptr==NULL)
    { XAerror(NOMEM);
      return;
    }
  }

  template<class T> void XplArray<T>::indexx (XplIndex sortact, int sortdim, long *indx, int nlen)
  { long n, l, j, ir, indxt, i;
    T q, cl, cr;
    n = dim(sortdim);
    if (nlen == 1) return;
    l=(nlen >> 1)+1;
    ir=nlen;
    for (;;)
    { if (l > 1)
      { indxt=indx[--l];
        sortact.setdim(sortdim, indxt-1);
        q=*(ptr+dim.value(sortact));
      }
      else 
      { indxt=indx[ir];
        sortact.setdim(sortdim, indxt-1);
        q=*(ptr+dim.value(sortact));
        indx[ir]=indx[1];
        if (--ir == 1) 
        { indx[1]=indxt;
	  return;
        }
      }
      i=l;
      j=l << 1;
      while (j <= ir) 
      { sortact.setdim(sortdim, indx[j]-1);
        cl= *(ptr+dim.value(sortact));
        sortact.setdim(sortdim, indx[j+1]-1);
        cr= *(ptr+dim.value(sortact));
        if ((j<ir) && (cl<cr)) j++;
        sortact.setdim(sortdim, indx[j]-1);
        cl= *(ptr+dim.value(sortact));
        if (q < cl) 
        { indx[i]=indx[j];
	  j += (i=j);
        }
        else j=ir+1;
      }
      indx[i]=indxt;
    }
  }

  template<class T> T* XplArray<T>::getptr (void) const         
  { return (ptr); }

  template<class T> void XplArray<T>::setptr (T* newptr)   
  { if (ptr) 
#ifndef MEMDEBUG
      delete[] ptr; 
#else
      xmem.delptr (ptr);
#endif
    ptr = newptr; 
  }

  template<class T> XplIndex XplArray<T>::getdim (void) const
  { return (dim); }

  template<class T> void XplArray<T>::setdim (const XplIndex &d) 
  { dim = d; }

  template<class T> XplArray<T>::XplArray ()
  { ptr = NULL;
#ifdef OUTPUT 
    outputfmt = format (ptr);
#endif
  }

  template<class T> XplArray<T>::XplArray (const XplIndex &newdim)
  { unsigned i, c;
    dim = newdim;
    construct(this);
#ifdef OUTPUT 
    outputfmt = format (ptr);
#endif
  }

  template<class T> XplArray<T>::XplArray (long d0, long d1, long d2, long d3, long d4, 
                                           long d5, long d6, long d7)
  { XplIndex tmp (d0, d1, d2, d3, d4, d5, d6, d7);
    dim = tmp;
    construct(this); 
#ifdef OUTPUT 
    outputfmt = format (ptr);
#endif
  }

  template<class T> XplArray<T>::XplArray (const XplArray<T> & x)
  { // assignment (x);
    dim = x.dim;
#ifdef SMALLVERSION 
    if (dim(0)>SMALLVERSIONOBSERVATIONS) XAerror(NOSUPPORT);       // PC 20011127 changed to SMALLVERSIONOBSERVATIONS
#endif
#ifndef MEMDEBUG
    long s  = x.size();
    ptr = new T[s]; 
#else
    ptr = (T *) xmem.newptr (sizeof(T), x.size(), "XplArray", "XplArray", "ptr");
#endif
    ::memcopy (ptr, x.ptr, x.size()); 
  }

  template<class T> XplArray<T>::~XplArray ()
  { if (ptr) 
#ifndef MEMDEBUG
      delete[] ptr; 
#else
    { xmem.delptr (ptr);
      ptr = NULL;
    }
#endif
#ifdef OUTPUT
    delete[] outputfmt;
    outputfmt = NULL;
#endif
  }

  template<class T> unsigned XplArray<T>::dimension (void) const       
  { int i;
    for (i=MAX_XARRAY-1; i>-1; i--)
      if (dim(i)>1) return (i+1);
    return (0);
  }

  template<class T> long XplArray<T>::dimension (unsigned index) const 
  { return (dim(index)); }

  template<class T> T& XplArray<T>::operator[] (long d0)
  { return *(ptr+d0); }

  template<class T> T& XplArray<T>::operator[] (long d0) const
  { return *(ptr+d0); }

  template<class T> T& XplArray<T>::operator() (const XplIndex &tmp) const
  { int i;
#ifndef MEMDEBUG
    T *err = new T[1];
#else
    T *err = (T*) xmem.newptr (sizeof(T), 1, "XplArray", "operator()", "err");
#endif
    for (i=0; i<MAX_XARRAY; i++)
    { if ((tmp(i)<0) || (tmp(i)>=dim(i)))
      { XAerrort (RANGE, "Access refused");
        return *err;
      }
    }
#ifndef MEMDEBUG
    delete[] err;
#else
    xmem.delptr (err);
    err = NULL;
#endif
    return *(ptr+dim.value(tmp)); 
  }

// const methods for operator()

  template<class T> T& XplArray<T>::operator() (void)  const              
  { return *ptr; }

  template<class T> T& XplArray<T>::operator() (long d0) const
  {
#ifndef MEMDEBUG
    T *err = new T[1];
#else
    T *err = (T*) xmem.newptr (sizeof(T), 1, "XplArray", "operator(long)", "err");
#endif
    if ((d0<0) || (d0>=dim(0)))
    { XAerrort (RANGE, "Access refused");
      return *err;
    }
#ifndef MEMDEBUG
    delete[] err;
#else
    xmem.delptr (err);
    err = NULL;
#endif
    return (*(ptr+d0));
  }

  template<class T> T& XplArray<T>::operator() (long d0, long d1) const
  { 
#ifndef MEMDEBUG
    T *err = new T[1];
#else
    T *err = (T*) xmem.newptr (sizeof(T), 1, "XplArray", "operator(long, long)", "err");
#endif
    if ((d0<0) || (d0>=dim(0)) ||
        (d1<0) || (d1>=dim(1)))
    { XAerrort (RANGE, "Access refused");
      return *err;
    }
#ifndef MEMDEBUG
    delete[] err;
#else
    xmem.delptr (err);
    err = NULL;
#endif
    return (*(ptr+d1*dim(0)+d0));
  }

  template<class T> T& XplArray<T>::operator() (long d0, long d1, long d2) const
  { 
#ifndef MEMDEBUG
     T *err = new T[1];
#else
     T *err = (T*) xmem.newptr (sizeof(T), 1, "XplArray", "operator(long,long,long)", "err");
#endif
     if ((d0<0) || (d0>=dim(0)) ||
        (d1<0) || (d1>=dim(1)) ||
        (d2<0) || (d2>=dim(2)))
    { XAerrort (RANGE, "Access refused");
      return *err;
    }
#ifndef MEMDEBUG
    delete[] err;
#else
    xmem.delptr (err);
    err = NULL;
#endif
    return (*(ptr+(d2*dim(1)+d1)*dim(0)+d0));
  }

  template<class T> T& XplArray<T>::operator() (long d0, long d1, long d2, long d3) const 
  {  
#ifndef MEMDEBUG
     T *err = new T[1];
#else
     T *err = (T*) xmem.newptr (sizeof(T), 1, "XplArray", "operator(long,long,long,long)", "err");
#endif
     if ((d0<0) || (d0>=dim(0)) ||
        (d1<0) || (d1>=dim(1)) ||
        (d2<0) || (d2>=dim(2)) ||
        (d3<0) || (d3>=dim(3)))
    { XAerrort (RANGE, "Access refused");
      return *err;
    }
#ifndef MEMDEBUG
    delete[] err;
#else
    xmem.delptr (err);
    err = NULL;
#endif
    return (*(ptr+((d3*dim(2)+d2)*dim(1)+d1)*dim(0)+d0));
  }

// non-const methods for operator()

  template<class T> T& XplArray<T>::operator() (const XplIndex &tmp)
  { int i;
#ifndef MEMDEBUG
    T *err = new T[1];
#else
    T *err = (T*) xmem.newptr (sizeof(T), 1, "XplArray", "operator()", "err");
#endif
    for (i=0; i<MAX_XARRAY; i++)
    { if ((tmp(i)<0) || (tmp(i)>=dim(i)))
      { XAerrort (RANGE, "Access refused");
        return *err;
      }
    }
#ifndef MEMDEBUG
    delete[] err;
#else
    xmem.delptr (err);
    err = NULL;
#endif
    return *(ptr+dim.value(tmp)); 
  }

  template<class T> T& XplArray<T>::operator() (void)             
  { return *ptr; }

  template<class T> T& XplArray<T>::operator() (long d0)
  {
#ifndef MEMDEBUG
    T *err = new T[1];
#else
    T *err = (T*) xmem.newptr (sizeof(T), 1, "XplArray", "operator(long)", "err");
#endif
    if ((d0<0) || (d0>=dim(0)))
    { XAerrort (RANGE, "Access refused");
      return *err;
    }
#ifndef MEMDEBUG
    delete[] err;
#else
    xmem.delptr (err);
    err = NULL;
#endif
    return (*(ptr+d0));
  }

  template<class T> T& XplArray<T>::operator() (long d0, long d1)
  { 
#ifndef MEMDEBUG
    T *err = new T[1];
#else
    T *err = (T*) xmem.newptr (sizeof(T), 1, "XplArray", "operator(long, long)", "err");
#endif
    if ((d0<0) || (d0>=dim(0)) ||
        (d1<0) || (d1>=dim(1)))
    { XAerrort (RANGE, "Access refused");
      return *err;
    }
#ifndef MEMDEBUG
    delete[] err;
#else
    xmem.delptr (err);
    err = NULL;
#endif
    return (*(ptr+d1*dim(0)+d0));
  }

  template<class T> T& XplArray<T>::operator() (long d0, long d1, long d2)
  { 
#ifndef MEMDEBUG
     T *err = new T[1];
#else
     T *err = (T*) xmem.newptr (sizeof(T), 1, "XplArray", "operator(long,long,long)", "err");
#endif
     if ((d0<0) || (d0>=dim(0)) ||
        (d1<0) || (d1>=dim(1)) ||
        (d2<0) || (d2>=dim(2)))
    { XAerrort (RANGE, "Access refused");
      return *err;
    }
#ifndef MEMDEBUG
    delete[] err;
#else
    xmem.delptr (err);
    err = NULL;
#endif
    return (*(ptr+(d2*dim(1)+d1)*dim(0)+d0));
  }

  template<class T> T& XplArray<T>::operator() (long d0, long d1, long d2, long d3)
  {  
#ifndef MEMDEBUG
     T *err = new T[1];
#else
     T *err = (T*) xmem.newptr (sizeof(T), 1, "XplArray", "operator(long,long,long,long)", "err");
#endif
     if ((d0<0) || (d0>=dim(0)) ||
        (d1<0) || (d1>=dim(1)) ||
        (d2<0) || (d2>=dim(2)) ||
        (d3<0) || (d3>=dim(3)))
    { XAerrort (RANGE, "Access refused");
      return *err;
    }
#ifndef MEMDEBUG
    delete[] err;
#else
    xmem.delptr (err);
    err = NULL;
#endif
    return (*(ptr+((d3*dim(2)+d2)*dim(1)+d1)*dim(0)+d0));
  }



  template<class T> XplArray<T> XplArray<T>::bind (const XplArray<T>& x, unsigned dimen) const
  { unsigned i;
    XplArray err;
    long loop, sizeleft, sizeright, entries = 1L, j;
    XplIndex bound;
    if (dimen>=MAX_XARRAY) 
    { XAerrort(DIMBIG, "~, |, or bind : prod(dim) > 2^31");
      return err;
    }
    for (i=0; i<MAX_XARRAY; i++)
    { if (i==dimen)
        bound.setdim(i, x.dim(i)+dim(i));
      else if (x.dim(i)!=dim(i))
        XAerrort(DIMCON, "~, | or bind : layer(x_i) <> layer(y_i)");
      else
        bound.setdim(i, dim(i));
    }
    XplArray<T> ret(bound);
    loop = bound.prod();
    sizeleft  = dim.prod(dimen+1);
    sizeright = x.dim.prod(dimen+1);
    T *rightptr = x.getptr(), *leftptr = ptr,
       *retptr;
    retptr = ret.getptr(); 
    loop /= (sizeleft+sizeright);
    for (j=0; j<loop; j++)
    { ::memcopy (retptr, leftptr, sizeleft); 
      retptr   += sizeleft;
      leftptr  += sizeleft;
      ::memcopy (retptr, rightptr, sizeright);
      retptr   += sizeright;
      rightptr += sizeright;
    }  
    return (ret);
  }

  template<class T> XplArray<T>& XplArray<T>::assignment (const XplArray<T>& x)
  { if (this != &x)
    { if (ptr !=NULL)
#ifndef MEMDEBUG
        delete[] ptr;
#else
        xmem.delptr (ptr);
#endif
      dim = x.dim;
#ifndef MEMDEBUG
      ptr = new T[x.size()];
#else
      ptr = (T *) xmem.newptr (sizeof(T), x.size(), "XplArray", "assignment", "ptr");
#endif
      ::memcopy (ptr, x.ptr, x.size());
    }
    return (*this);
  }

  template<class T> XplArray<T>& XplArray<T>::operator= (const XplArray<T> &x)
  { assignment (x);
    return (*this);
  }  

  template<class T> XplArray<T>& XplArray<T>::operator= (T val)
  { *ptr = val;
    return (*this);
  }  

  template<class T> void XplArray<T>::fill (T val)
  { long i, max = dim.prod();
    for (i=0; i<max; i++)
      *(ptr+i) = val;
  }

  template<class T> void XplArray<T>::fill (const XplArray<T>& x)

  { XplIndex tmp;
    tmp = 0L;
    do
    { *(ptr+dim.value(tmp)) = 
         x(tmp % x.dim); 
    }
    while (tmp.inc(dim));  
  }

  template<class T> void XplArray<T>::reshape (const XplIndex newdim)
  { if (dim.prod()!=newdim.prod()) XAerrort(DIMCON, "reshape : dim(x)<>prod(v)");
    dim = newdim;
    return;
  }

  template<class T> int XplArray<T>::isconformable (const XplArray<T>& x) const 
  { return (dim.isconf(x.dim)); }

  template<class T> void XplArray<T>::dump (const char *fmt) const
  { char out[256];
    XplIndex tmp;
    cout << "DIMENSION:\n";
    dim.dump();
    cout << "PTR: ";
    sprintf (out, "%p\n", ptr);
    cout << out;
    cout << "CONTENT:\n";
    tmp = 0L;
    do
    { x4sprintf (out, fmt, *(ptr+dim.value(tmp)));
      cout << out;
    }
    while (tmp.inc(dim));
  }

  template<class T> void XplArray<T>::mdump (const char *fmt, Xplostream *stream, const char *varname, int pdim) const
  { char out[256], *ufmt, *txt;
    T *subptr;
    long i, j, k, l, s, m, starfmt;
    unsigned long count, maxcount;
    XplIndex tmp;
    Xplostream streammdump(stream);
//    printf ("mdump : fmt     = %s\n", fmt);
//    printf ("mdump : varname = %s\n", varname);
//    printf ("mdump : pdim    = %i\n", pdim);
//    printf ("mdump : this    = %p\n", this);
//    printf ("mdump : ptr     = %p\n", ptr);
//    dim.dump();
//    printf ("\n");
    char 
      *outheadline = property.getProperty("outheadline"), 
      *outlineno = property.getProperty("outlineno"), 
      *outlayerline = property.getProperty("outlayerline"),
      *outputformat = property.getProperty("outputformat");
    long outmaxdata = atol (property.getProperty("outmaxdata"));

    if (size()<=0) 
    { XAerrort (ARRAYBAD, "Arraysize <= 0");
      return;
    }
    l = dim(0)*dim(1);    
    s = dim.prod()/l;
    m = (long) (1.0+x4log10((double) (dim(0))));
#ifdef OUTPUT
    ufmt = (fmt==NULL ? outputfmt : fmt);
#else
    ufmt = (char *) fmt;
#endif
    
    if (varname)
      sprintf (out, outheadline, varname);
    else
      sprintf (out, outheadline, "unnamed object");
    streammdump << out;

    // printf ("mdump : ufmt     = %s\n", ufmt);

    starfmt = (strchr(outlineno, '*') != NULL);
    maxcount = (outmaxdata<1 ? 0xffffffff : outmaxdata); 
    count    = 0L;

    for (tmp=0, k=0; (k<s) && (count<maxcount) ; k++)
    { subptr = ptr+k*l;
      if (pdim)
      { if (s>1)
        { sprintf (out, outlayerline, 
                    tmp(2)+1,tmp(3)+1,tmp(4)+1,tmp(5)+1,tmp(6)+1,tmp(7)+1);
          streammdump << out;
        }
      }

      for (i=0; (i<dim(0)) && (count<maxcount); i++)
      { if (starfmt)
          sprintf (out, outlineno, m, i+1);
        else
          sprintf (out, outlineno, i+1);
        streammdump << out;
        for (j=0; (j<dim(1)) && (count<maxcount); j++, tmp.inc(dim), count++)
        { x4sprintf (out, outputformat, *(subptr+j*dim(0)+i));
          streammdump << out;
          streammdump << " ";
        }
        streammdump << "\n";
      }
    } 
    if (count==maxcount)
      streammdump << "\r\nSorry, this object is too big !!\r\n";
  }

  template<class T> void XplArray<T>::sdump (const char *fmt) const
  { char out[256];
    sprintf (out, "%-20s %p", fmt, ptr);   
    cout << out;
  }  

  template<class T> void XplArray<T>::tfunc (T (*func) (void))
  { unsigned long i, m;
    T *tptr;
    m = dim.prod();
    for (i=0L; i<m; i++)
      *(ptr+i) = func();
  }

  template<class T> XplArray<T> XplArray<T>::tfunct (int nr, T (*func) (T l))
  { long tmp, s = size();
    for (tmp=0; tmp<s; tmp++)
      *(ptr+tmp) = func (*(ptr+tmp));
    return (*this);
  }

  template<class T> XplArray<T> XplArray<T>::tfunctt (const XplArray<T> &right, T (*func) (T l, T r)) const
  { XplArray<T> res (dim.maxindex(right.dim));
    XplIndex tmp;
    T *rptr, *resptr, val;
    long lind, rind, tind, i, lsize, rsize;

    lsize  = size();
    rsize  = right.size();
    resptr = res.getptr();
    rptr   = right.getptr();

    if (lsize==1)
    { val = *ptr;
      for (i=0; i<rsize; i++)
        *(resptr+i) = func (val, *(rptr+i));
    }
    else if (rsize==1)
    { val = *rptr;
      for (i=0; i<lsize; i++)
        *(resptr+i) = func (*(ptr+i), val);
    }
    else if (dim == right.dim)    // PC: special code for matrices of the same dimensions
    {	// now, lsize == rsize!
    	 resptr = res.getptr();
         rptr = right.getptr();

         for(i = 0; i < lsize; i++)
              *(resptr+i) = func(*(ptr+i), *(rptr+i));
    }
    else
    { resptr = res.getptr();
      rptr = right.getptr();
      tmp = 0L;
      do
      { tind = res.dim.value(tmp);
        rind = right.dim.value(tmp % right.dim);
        lind = dim.value(tmp % dim);
        *(resptr+tind) = func(*(ptr+lind), *(rptr+rind));
      }
      while (tmp.inc(res.dim));
    }
    return (res);
  }

  template<class T> XplArray<T> XplArray<T>::tfunct (const XplArray<T> &right, T (*func) (T l, T *r), T *val)
  { long tmp, s = size();
    T run = *val;
    for (tmp=0; tmp<s; tmp++)
      *(ptr+tmp) = func (*(ptr+tmp), &run);
    *val = run;
    return (*this);
  }

  template<class T> XplArray<T> XplArray<T>::layernpnp (void (*func) (T* ptr, long n, long p))
  { long tmp, s = size(), n = dim(0), p = dim(1);
//  printf ("Size: %li\n", s);
//  printf ("N:    %li\n", n);
//  printf ("P:    %li\n", p);
    for (tmp=0; tmp < s; tmp += n*p)
      func (ptr+tmp, n, p);
    return (*this);
  }

  template<class T> XplArray<T> XplArray<T>::layernn (void (*func) (T* ptr, long n))
  { long tmp, s = size(), n = dim(0);
//  printf ("Size: %li\n", s);
//  printf ("N:    %li\n", n);
//  printf ("P:    %li\n", p);
    for (tmp=0; tmp < s; tmp += n)
      func (ptr+tmp, n);
    return (*this);
  }

  template<class T> XplArray<T> XplArray<T>::layernpnpsort (T* (*func) (T* ptr, long n, long p, long ndir, long *dir), const XplArray<long> &dirmat)
  { long tmp, s = size(), n = dim(0), p = dim(1), *pdir, ndir, adir, k;
    T *resptr;
    ndir = dirmat.dim(0);
    pdir = dirmat.getptr();
    adir = 0;
    for (tmp=0; tmp < s; tmp += n*p, adir+=ndir)
    { adir   = adir%dirmat.size();
      resptr = func (ptr+tmp, n, p, ndir, pdir+adir);
      ::memcopy (ptr+tmp, resptr, n*p); 
#ifndef MEMDEBUG
      delete[] resptr;
#else
      xmem.delptr (resptr);
      resptr = NULL;
#endif
    }
    return (*this);
  }


  template<class T> XplArray<T> XplArray<T>::layernpnp1 (void (*func) (T* ptr, long n, long p, T *retval)) const
  { long tmp, s = size(), n = dim(0), p = dim(1), k;
//  printf ("Size: %li\n", s);
//  printf ("N:    %li\n", n);
//  printf ("P:    %li\n", p);
    T resdet;
    k=0;
    XplArray<T> res(s/(n*p));	
    for (tmp=0; tmp < s; tmp += n*p)
    { func (ptr+tmp, n, p, &resdet);
      res(k) = resdet;
      k=k+1;
    }
    return (res);
  }

  template<class T> XplArray<T> XplArray<T>::layernpnpdist (void (*func) (T* ptr, long n, long p, T *retval,
                                           const char *distm), const char *dist) const
  { long tmp, tmp2, s = size(), n = dim(0), p = dim(1);
    XplIndex resdim = dim;
    resdim.setdim (1, dim(0));
    XplArray<T> res(resdim);	
    s = dim.prod()/(n*p);
    for (tmp=tmp2=0; tmp < s; tmp += n*p, tmp2 += n*n)
      func (ptr+tmp, n, p, res.ptr+tmp2, dist);
    return (res);
  }

  template<class T> XplArray<T> XplArray<T>::layernpnp2 (void (*func) (T* ptr, long n, long p, T* ptrl, T* ptru, long *ptrix), const XplArray<T> &l, const XplArray<T> &u, const XplArray<long> &ix)
  { long tmp, s = size(), n = dim(0), p = dim(1);
    T *ptrl, *ptru;
    long *ptrix;
    ptrl=l.getptr();
    ptru=u.getptr();
    ptrix=ix.getptr();
    long k;
//  printf ("Size: %li\n", s);
//  printf ("N:    %li\n", n);
//  printf ("P:    %li\n", p);
    k=0;
    for (tmp=0; tmp < s; tmp += n*p)
      {
      func (ptr+tmp, n, p, ptrl+tmp, ptru+tmp, ptrix+n*k);
       k=k+1;
      }
    return(*this);
  }

  template<class T> XplArray<T> XplArray<T>::layernpnp3 (void (*func) (T *x, long n, long p, 
                                        T *a, long na, long pa,
                                        T *b, long nb, long pb),
                          const XplArray<T> &aa, const XplArray<T> &bb) const
  { XplArray<T> res = *this;

    T *resptr, *aptr, *bptr;
    long nr, pr, naa, paa, nbb, pbb, s, i;
  
    resptr = res.getptr();
    aptr   = aa.getptr();
    bptr   = bb.getptr();
    
    nr = res.dim(0);
    pr = res.dim(1);
    naa = aa.dim(0);
    paa = aa.dim(1);
    nbb = bb.dim(0);
    pbb = bb.dim(1);

    s = (res.dim).prod() / (nr*pr);
    for (i=0; i<s; i++, resptr+=nr*pr, aptr +=naa*paa, bptr +=nbb*pbb)
      func (resptr, nr, pr, aptr, naa, paa, bptr, nbb, pbb);

    return (res);
  }

  template<class T> XplArray<T>& XplArray<T>::selection (const XplArray<T>& x, int nsel, const XplSelection *sel)
  { long i, j, k, *lptr[MAX_XARRAY], s;
    XplIndex tmp, count;
    dim = x.dim;
    for (i=0; i<MAX_XARRAY; i++)
    { lptr[i] = NULL;
      if ((i<nsel) && ((sel+i)->ptr()))
      { dim.setdim(i, (sel+i)->len());
        lptr[i] = (sel+i)->ptr();
      }
    }
    if (ptr) 
#ifndef MEMDEBUG
      delete[] ptr;
#else
      xmem.delptr (ptr);
#endif
    s = size();
#ifndef MEMDEBUG
    ptr = new T[s];
#else
    ptr = (T *) xmem.newptr (sizeof(T), s, "XplArray", "selection", "ptr");
#endif
    for (i=0; i<MAX_XARRAY; i++)
    { tmp.setdim(i, (lptr[i] ? lptr[i][0] : 0));
      count.setdim (i, 0);
    }
    for (i=0; i<s; i++)
    { //printf("Loop %i\n", i);
      //tmp.dump();
      for (k=0; k<MAX_XARRAY; k++)
        if ((tmp(k)<0) || (tmp(k)>=x.dim(k)))	// PC 20010709 : here was && instead of ||
        { XAerrort (RANGE, "Access refused");
          return (*this);
        }
      *(ptr+i) = *(x.ptr+x.dim.value(tmp));
      for (j=0; j<MAX_XARRAY; j++)
      { if (lptr[j])
        { count.setdim (j, count(j)+1);
          if (count(j)<dim(j))
	  { tmp.setdim(j, lptr[j][count(j)]);
            break;
          }
          count.setdim (j, 0);
          tmp.setdim(j, lptr[j][0]);
        }
        else
        { tmp.setdim (j, tmp(j)+1);
          if (tmp(j)<dim(j)) break;
          tmp.setdim(j, 0);
        } 
      }
    } 
    return (*this);
  }

  template<class T> XplArray<T>& XplArray<T>::assignment (int nsel, const XplSelection *sel, const XplArray<T>& x)
  { long i, j, k, *lptr[MAX_XARRAY], s, maxs;
    XplIndex tmp, count, actdim;
    actdim = dim;
    if (&x==NULL)
    { XAerrort (ARRAYBAD, "Can not assign to a numeric matrix a textvalue");
      return (*this);
    }
    for (i=0; i<MAX_XARRAY; i++)
    { lptr[i] = NULL;
      if ((i<nsel) && ((sel+i)->ptr()))
      { actdim.setdim (i, (sel+i)->len());
        lptr[i] = (sel+i)->ptr();
      }
      else
        actdim.setdim (i, 1);
    }
    s = actdim.prod();
    for (i=0; i<MAX_XARRAY; i++)
    { tmp.setdim (i, (lptr[i] ? lptr[i][0] : 0));
      count.setdim (i, 0);
    }
    maxs = (x.dim).prod();
//  printf ("Len: %li\n", s);
//    tmp.dump();
    for (i=0; i<s; i++)
    { for (k=0; k<MAX_XARRAY; k++)
        if ((tmp(k)<0) || (tmp(k)>=dim(k)))
//          XAwarnt (RANGE, "Cycling in []");
	{ XAerrort (RANGE, "Access refused");
          return (*this);
        }
      *(ptr+dim.value(tmp%dim)) = *(x.ptr+(i%maxs));
      for (j=0; j<MAX_XARRAY; j++)
      { if (lptr[j])
        { count.setdim (j, count(j)+1);
          if (count(j)<actdim(j))
	  { tmp.setdim (j, lptr[j][count(j)]);
            break;
          }
          count.setdim (j, 0);
          tmp.setdim (j, lptr[j][0]);
        }
        else
        { tmp.setdim (j, tmp(j)+1);
          if (tmp(j)<actdim(j)) break;
          tmp.setdim(j, 0);
        } 
      }
    } 
    return (*this);
  }

  template<class T> XplArray<T>& XplArray<T>::sort (int n, const XplIndex *vec, int *dir)
  { long *index, i, len, *sortlist, sortlistlen, *newsortlist, newsortlistlen,
         k, l, size, blocksize;
    int sortdim, j, lastcomp;
    XplIndex sortact;
    T ot, nt, *sortptr;
    sortdim = -1;
    for (i=0; i<MAX_XARRAY; i++)
    { if ((vec[0])(i)<0)
      { if (sortdim>-1) XAerror(NONSORT);
          sortdim = i;
      }
      else if ((vec[0])(i)>=dim(i))
      { XAerror (NONSORT);
        return (*this);
      }
    }
    if (sortdim<0) 
    { XAerror(NONSORT);
      return (*this);
    }
    for (j=1; j<n; j++)
    { for (i=0; i<MAX_XARRAY; i++)
      { if (i!=sortdim) 
        { if ((vec[j])(i)<0) 
          { XAerror(DIMNEG);
            return (*this);
          }
          if ((vec[j])(i)>=dim(i)) 
          { XAerror(DIMELEM);
            return (*this);
          }
        }
      }
    }
    len = dim(sortdim);
#ifndef MEMDEBUG
    index = new long[len+1];
    sortlist = new long[2*len];
    newsortlist = new long[2*len];
#else
    index = (long *) xmem.newptr (sizeof(long), len+1, "XplArray", "sort", "index");
    sortlist = (long *) xmem.newptr (sizeof(long), 2*len, "XplArray", "sort", "sortlist");
    newsortlist = (long *) xmem.newptr (sizeof(long), 2*len, "XplArray", "sort", "newsortlist");
#endif
    sortlist[0] = 1;
    sortlist[1] = len;
    sortlistlen = 2;
    newsortlistlen = 0;
    for (i=0; i<len+1; i++) index[i] = i;
    j = 0;
    while ((sortlistlen>0) && (j<n))
    { sortact = vec[j];
      for (k=0; k<sortlistlen; k += 2)
      { indexx (sortact, sortdim, &(index[sortlist[k]-1]), 
                sortlist[k+1]-sortlist[k]+1); 
        sortact.setdim (sortdim, index[sortlist[k]]-1);
        ot = *(ptr+dim.value(sortact));
        lastcomp = 0; 
        for (l=1; l<sortlist[k+1]-sortlist[k]+1; l++)
        { sortact.setdim(sortdim, index[sortlist[k]+l]-1);
          nt = *(ptr+dim.value(sortact)); 
          if (ot==nt)
          { if (lastcomp==0)
	    { newsortlist[newsortlistlen]=sortlist[k]+l-1;
              newsortlistlen++;
              newsortlist[newsortlistlen]=sortlist[k]+l;
              lastcomp = 1;
            }
            else if (lastcomp==1)
              newsortlist[newsortlistlen]=sortlist[k]+l;
          }
          else
	  { if (lastcomp)
	    { newsortlistlen++;
              lastcomp = 0;
            }
          }
          ot = nt;
        }
      } 
      ::memcopy (sortlist, newsortlist, 2*len);
      sortlistlen = newsortlistlen;
      j++;
    }

#ifndef MEMDEBUG  
    delete[] sortlist;
    delete[] newsortlist;
#else
    xmem.delptr (sortlist);
    xmem.delptr (newsortlist);
    sortlist = NULL;
    newsortlist = NULL;
#endif
    size = dim.prod();
    blocksize = dim.prod (sortdim);
#ifndef MEMDEBUG
    sortptr = new T[size];
#else
    sortptr = (T*) xmem.newptr(sizeof(T), size, "XplArray", "sort", "sortlist");
#endif
    j = -len;
    for (i=0; i<size/blocksize; i++)
    { if ((i%len)==0) j += len;
      ::memcopy(sortptr+i*blocksize, ptr+(j+index[(i%len)+1]-1)*blocksize, blocksize); 
    }
    setptr (sortptr);    

    return (*this);
  }

  template<class T> XplArray<T> XplArray<T>::sortvec (const XplArray<long> &dir) const
  { XplArray<T> ret;
    ret = *this;
    if (dir.dim(1)!=1)
    { XAerrort (DIMCON, "Direction matrix should have only one column"); 
      return ret;
    }
    ret.layernpnpsort (::vecsort, dir);
    return (ret);
  }

  template<class T> XplArray<int> XplArray<T>::ifunctt (const XplArray<T> &right, int (*func) (const T l, const T r)) const 
  { XplIndex tmp, retdim=dim.maxindex(right.dim);
    XplArray<int> ret (retdim);
    T *rptr;
    int *retptr;
    long lind, rind, tind, i, s;
    retptr = ret.getptr();
    rptr = right.getptr();
    if (dim==right.dim)
    { s = dim.prod();
      for (i=0; i<s; i++)
        *(retptr+i) = func(*(ptr+i), *(rptr+i));    
    }
    else
    { tmp = 0L;
      do
      { tind = retdim.value(tmp);
        rind = right.dim.value(tmp % right.dim);
        lind = dim.value(tmp % dim);
        *(retptr+tind) = func(*(ptr+lind), *(rptr+rind));
      }
      while (tmp.inc(retdim));
    }
    return (ret);
  }

// PC - auxiliary function for ReadASCIIXpl - checks conditions
// index should range from 1..n
bool checkreadasciicondition(long index, long *list, long listlen,
                             long *cindex, long *conds, long condslen,
                             bool increment)
{
     long i, pom, bottom, top;
     int type;
     bool res1, res2;

     if (index <= 0)
          return false;

     // list of inclusion (pos) / exclusion (neg) intervals
     // we assume that the lower index comes always first!
     if ((list != NULL) && (listlen > 0))
     {
          // check what is default for res1:
          //   if only exclusion restrictions are present, res1 = true by default
          //   if only inclusion restrictions are present, res1 = false by default
          //   mixture -- we return false always, so nothing is read at all
          type = (list[0] > 0) ? 1 : -1;
          for(i = 1; i < 2*listlen; i++)
          {
               if ((list[i] > 0) && (type <= 0))
               {
                    type = 0;
                    break;
               }
               if ((list[i] < 0) && (type >= 0))
               {
                    type = 0;
                    break;
               }
          }
          if (type == 0)
               res1 = true;
          else if (type > 0)
               res1 = false;  // inclusion
          else if (type < 0)
               res1 = true;   // exclusion

          if (type != 0)		// we check it only if one type of conditions is present
          for(i = 0; i < listlen; i++)
          {
               bottom = list[i];
               top = list[listlen + i];

               if ((bottom < 0) && (top < 0)) // exclusion
               {
                    bottom *= -1;
                    top *= -1;
                    if ((bottom <= index) && (index <= top))
                    {
                         res1 = false;
                         break;
                    }
               }
               else if ((bottom > 0) && (top > 0)) // inclusion
               {
                    if ((bottom <= index) && (index <= top))
                    {
                         res1 = true;
                         break;
                    }
               }
               // otherwise ignored
          } // end for
     }
     else res1 = true;   // no list

     // boolean condition vector - evaluate only if res1 == true
     if (res1)
     {
          if (increment)
               *cindex += 1;
          if (*cindex <= 0)
               return false;
          if ((conds != NULL) && (condslen > 0))
               res2 = (conds[(*cindex-1) % condslen] != 0) ? true : false;
          else res2 = true;   // no conditions
     }

     return (res1 && res2) ? true : false;
}

  // PC 2001: read rewritten so that it allows for selection of rows / columns by list or by conditions
  template<class T> int XplArray<T>::ReadASCIIXplArray (char const *datei, XplArray<long> *selrows, XplArray<long> *selcols,
                                                                           XplArray<long> *rowconds, XplArray<long> *colconds)
  { FILE *f = NULL;

    T *newptr;
    char *epl;
    char errmsg[1024];

    T *tmpt;
    long actualsize = MAX_BLOCKSIZE, count, rows = 0, cols = 0, linecol, i, j, k, s;
    long truerows = 0, truelinecols = 0;
    long truecondrows = 0, truecondlinecols = 0;
    char line[MAX_LINE_LENGTH], *tok, l=0, *tmpc;
    char bakline[MAX_LINE_LENGTH], pomchar;
    long readlinebuf;
    bool waseol, acteol;  // did we find EOL last time?

    // PC: condition variables
    long *selrowsptr = NULL, *selcolsptr = NULL, *rowcondsptr = NULL, *colcondsptr = NULL;
    long selrowslen = 0, selcolslen = 0, rowcondslen = 0, colcondslen = 0, pomstrlen;
    char *fgetsres;

    if (ReadXXplArray(datei)==1) return (1);
/*
    FILE *log = fopen ("x.log", "w");
    printf ("try to open: %s\n", datei);
    fclose(log);
*/
    f = opendatafile (datei, "r");
    if (f==NULL)
    { sprintf (errmsg, "Cannot open data file : %s", datei);
      XAerrort (STRBAD, errmsg);
      return(0);
    }

// check parameters regarding conditions
     if (selrows != NULL)
     if ((selrows->dim(1) == 2) && (selrows->dim.prod(2) == selrows->dim.prod(MAX_XARRAY)))
     {
          selrowsptr = selrows->getptr();
          selrowslen = selrows->dim(0);
     }
     if (selcols != NULL)
     if ((selcols->dim(1) == 2) && (selcols->dim.prod(2) == selcols->dim.prod(MAX_XARRAY)))
     {
          selcolsptr = selcols->getptr();
          selcolslen = selcols->dim(0);
     }
     if (rowconds != NULL)
     if (rowconds->dim.prod(MAX_XARRAY) > 1)
     {
          rowcondsptr = rowconds->getptr();
          rowcondslen = rowconds->dim.prod(MAX_XARRAY);
     }
     if (colconds != NULL)
     if (colconds->dim.prod(MAX_XARRAY) > 1)
     {
          colcondsptr = colconds->getptr();
          colcondslen = colconds->dim.prod(MAX_XARRAY);
     }

// memory allocation
#ifndef MEMDEBUG
    newptr = new T[actualsize];
    epl    = new char[actualsize];
#else
    newptr = (T *) xmem.newptr (sizeof(T), actualsize, "XplArray", "ReadASCIIXplArray", "newptr");
    epl    = (char *) xmem.newptr (sizeof(char), actualsize, "XplArray", "ReadASCIIXplArray", "epl");
#endif

     count = 0;
     waseol = true;
     acteol = true;
     readlinebuf = MAX_LINE_LENGTH - 2;
     *line = '\0';
     // PC 20010815 - fix for cases when the last line does not have EOL
     while ((fgetsres = fgets(line + strlen(line), readlinebuf, f)) || !waseol)
     {
          if (!waseol && (fgetsres == NULL))	// last line does not have EOL
          {
				   strcat(line, "\n");
          }

          if (waseol)    // increase rows just at the end of the line
          {
               rows++;
               truerows++;
               linecol = 0;
               truelinecols = 0;
               truecondlinecols = 0;
          }

          strcpy(bakline, line);
          tok = line;

          if (pomstrlen = strlen(line))
               acteol = (line[pomstrlen - 1] == '\n') ? true : false;
      
          for (i=0; i<strlen(tok); i++)
          {    if ((!isprint(tok[i])) && (tok[i]!='\n') && (tok[i]!='\t'))
               {    if (tok[i]!=26)  //Ctrl-Z
                    {    fclose(f);
                         sprintf (errmsg, "Cannot read data file : %s strange character(s)", datei);
                         XAerrort (STRBAD, errmsg);
#ifndef MEMDEBUG
                         delete[] newptr;
                         delete[] epl;
#else
                         xmem.delptr (newptr);
                         xmem.delptr (epl);
#endif
                         return 0;
                    }
                    tok[i] = ' ';  
               }
          }
          
          // remove uncomplete tokens
          for(pomstrlen = strlen(line)-1; pomstrlen > 0; pomstrlen--)
          {
               pomchar = line[pomstrlen];
               if ((pomchar == ' ') || (pomchar == '\n') || (pomchar == '\t'))
               {
                    pomstrlen++;
                    break;
               }
          }
          if (pomstrlen >= strlen(line))
               *bakline = '\0';
          else
          {
               strcpy(bakline, line + pomstrlen);
               line[pomstrlen] = '\0';
          }

          // get tokens
          tok = strtok (line, " \t\n");
          while (tok)
          {
               if ( checkreadasciicondition(truerows, selrowsptr, selrowslen, &truecondrows, rowcondsptr, rowcondslen, truelinecols==0) &&
                    checkreadasciicondition(truelinecols+1, selcolsptr, selcolslen, &truecondlinecols, colcondsptr, colcondslen, true) )
               {
                    if ( (strcmp(".", tok)==0) || (strcasecmp("NaN", tok)==0) ||
                         (strcasecmp("na", tok)==0))
                         *(newptr+count) = (T) sqrt(-1);
                    else
                         *(newptr+count) = (T) x4atof(tok);
                    *(epl+count) = l;

                    count++;
                    linecol++;
                    truelinecols++;
               }
               else truelinecols++;

               if (count==actualsize)
               {    actualsize += MAX_BLOCKSIZE;
#ifndef MEMDEBUG
                    tmpt = new T[actualsize];
                    tmpc = new char[actualsize];
#else
                    tmpt = (T *) xmem.newptr (sizeof(T), actualsize, "XplArray", "ReadASCIIXplArray", "tmpt");
                    tmpc = (char *) xmem.newptr (sizeof(char), actualsize, "XplArray", "ReadASCIIXplArray", "tmpc");
#endif
                    ::memcopy (tmpt, newptr, actualsize-MAX_BLOCKSIZE); //sizeof(T)*actualsize);
                    ::memcopy (tmpc, epl, actualsize-MAX_BLOCKSIZE); //sizeof(char)*actualsize);

#ifndef MEMDEBUG
                    delete[] newptr;
                    delete[] epl;
#else
                    xmem.delptr (newptr);
                    xmem.delptr (epl);
#endif
                    newptr = tmpt;
                    epl    = tmpc;
               }
        
               tok = strtok (NULL, " \t\n"); // get next token
          } // end of while (tok)

          // if we are at the end of line...
          if (acteol)
          {
               if (linecol == 0)
                    rows--;
               else
                    l = 1-l;
               if (linecol > cols) cols = linecol;

               if (truelinecols == 0)
                    truerows--;

               *line = '\0';	// we have to clean the line!
               readlinebuf = MAX_LINE_LENGTH - 2;
          }
          else // move the rest of the buffer
          {
               strcpy(line, bakline);
               readlinebuf = MAX_LINE_LENGTH - strlen(line) - 3;
               if (readlinebuf < 2)
               {
                    *line = '\0';
                    readlinebuf = MAX_LINE_LENGTH - 2;
               }
          }

          waseol = acteol;    // set EOL
     }
     // last row does not have to have EOL at the end!
     if (!waseol)
     {
          if (linecol == 0)
               rows--;
          else
               l = 1-l;
          if (linecol > cols) cols = linecol;

          if (truelinecols == 0)
               truerows--;

          waseol = true;
     }

     // the rest is the same as in the original routine
    fclose(f);
    if ((rows==0) || (cols==0)) // SK 10.08.98
    { sprintf (errmsg, "No data found: %s", datei);
      XAerrort (STRBAD, errmsg);
#ifndef MEMDEBUG
      delete[] newptr;
      delete[] epl;
#else
      xmem.delptr (newptr);
      xmem.delptr (epl);
#endif
      return 0;
    }

//    printf ("read: %li rows and %li cols recognized, %li values read\n", rows, cols, count);
    dim = XplIndex (rows, cols);
    l = 0;
    s = dim.prod();  // SK 02.09.97
#ifndef MEMDEBUG 
    delete[] ptr;
    ptr = new T[s];
#else
    xmem.delptr (ptr);
    ptr = (T *) xmem.newptr (sizeof(T), s, "XplArray", "ReadASCIIXplArray", "ptr");
#endif
//    printf ("Allocted: %li\n", s);
    i = 0;
    j = 0;
    l = 0;
    for (k=0; k<count; k++)
    { // printf ("count=%li, row=%li, col=%li\n", k, i, j); 
      if (*(epl+k)==l)
      { *(ptr+j*rows+i) = *(newptr+k);
        j++;
      }
      else
      { while (j<cols)
        { *(ptr+rows*j+i) = (T) sqrt(-1);
          j++; 
        }
        i++;
        *(ptr+i) = *(newptr+k);
        j = 1;
        l = 1-l;
      }  
    }
//    printf ("End: count=%li, row=%li, col=%li\n", k, i, j);
//    printf ("Rows: %li - Cols: %li\n", rows, cols); 
    for (k=count; k<s; k++, j++)
    { if (j==cols)
      { j = 0;
        i++;
        if (i==rows) break;
      }
//      printf ("count=%li, row=%li, col=%li\n", k, i, j); 
      *(ptr+rows*j+i) = (T) sqrt(-1);
    }
#ifndef MEMDEBUG    
    delete[] newptr;
    delete[] epl;
#else
    xmem.delptr (newptr);
    xmem.delptr (epl);
    newptr = NULL;
    epl = NULL;
#endif
    return (1);
  }

//  PC: rewritten
//  template<class T> int XplArray<T>::ReadASCIIXplArray (char const *datei)
//  { FILE *f = NULL;
//
//    T *newptr;
//    char *epl;
//    char errmsg[1024];
//
//    T *tmpt;
//    long actualsize = MAX_BLOCKSIZE, count, rows = 0, cols = 0, linecol, i, j, k, s;
//    char line[MAX_LINE_LENGTH], *tok, l=0, *tmpc;
//
//    if (ReadXXplArray(datei)==1) return (1);
///*
//    FILE *log = fopen ("x.log", "w");
//    printf ("try to open: %s\n", datei);
//    fclose(log);
//*/
//    f = opendatafile (datei, "r");
//    if (f==NULL)
//    { sprintf (errmsg, "Can not open data file : %s", datei);
//      XAerrort (STRBAD, errmsg);
//      return(0);
//    }
//
//#ifndef MEMDEBUG
//    newptr = new T[actualsize];
//    epl    = new char[actualsize];
//#else
//    newptr = (T *) xmem.newptr (sizeof(T), actualsize, "XplArray", "ReadASCIIXplArray", "newptr");
//    epl    = (char *) xmem.newptr (sizeof(char), actualsize, "XplArray", "ReadASCIIXplArray", "epl");
//#endif
//    count = 0;
//    while (fgets(line, MAX_LINE_LENGTH-1, f))
//    { rows++;
//      tok = line;
//      for (i=0; i<strlen(tok); i++)
//      { if ((!isprint(tok[i])) && (tok[i]!='\n') && (tok[i]!='\t'))
//        { if (tok[i]!=26)  //Ctrl-Z
//          { fclose(f);
//            sprintf (errmsg, "Can not read data file : %s strange character(s)", datei);
//            XAerrort (STRBAD, errmsg);
//#ifndef MEMDEBUG
//            delete[] newptr;
//            delete[] epl;
//#else
//            xmem.delptr (newptr);
//            xmem.delptr (epl);
//#endif
//            return 0;
//          }
//          tok[i] = ' ';
//        }
//      }
//      tok = strtok (line, " \t\n");
//      linecol = 0;
//      while (tok)
//      { if ((strcmp(".", tok)==0) || (strcasecmp("NaN", tok)==0) ||
//            (strcasecmp("na", tok)==0))
//          *(newptr+count) = (T) sqrt(-1);
//        else
//          *(newptr+count) = (T) x4atof(tok);
//        *(epl+count) = l;
//        count++;
//        linecol++;
//        if (count==actualsize)
//        { actualsize += MAX_BLOCKSIZE;
//#ifndef MEMDEBUG
//          tmpt = new T[actualsize];
//          tmpc = new char[actualsize];
//#else
//          tmpt = (T *) xmem.newptr (sizeof(T), actualsize, "XplArray", "ReadASCIIXplArray", "tmpt");
//          tmpc = (char *) xmem.newptr (sizeof(char), actualsize, "XplArray", "ReadASCIIXplArray", "tmpc");
//#endif
//          ::memcopy (tmpt, newptr, actualsize-MAX_BLOCKSIZE); //sizeof(T)*actualsize);
//          ::memcopy (tmpc, epl, actualsize-MAX_BLOCKSIZE); //sizeof(char)*actualsize);
//
//#ifndef MEMDEBUG
//          delete[] newptr;
//          delete[] epl;
//#else
//          xmem.delptr (newptr);
//          xmem.delptr (epl);
//#endif
//          newptr = tmpt;
//          epl    = tmpc;
//        }
//        tok = strtok (NULL, " \t\n");
//      }
//      if (linecol==0)
//        rows--;
//      else
//        l = 1-l;
//      if (linecol>cols) cols = linecol;
//    }
//    fclose(f);
//    if ((rows==0) || (cols==0)) // SK 10.08.98
//    { sprintf (errmsg, "No data found: %s", datei);
//      XAerrort (STRBAD, errmsg);
//#ifndef MEMDEBUG
//      delete[] newptr;
//      delete[] epl;
//#else
//      xmem.delptr (newptr);
//      xmem.delptr (epl);
//      return 0;
//#endif
//    }
//
////    printf ("read: %li rows and %li cols recognized, %li values read\n", rows, cols, count);
//    dim = XplIndex (rows, cols);
//    l = 0;
//    s = dim.prod();  // SK 02.09.97
//#ifndef MEMDEBUG
//    delete[] ptr;
//    ptr = new T[s];
//#else
//    xmem.delptr (ptr);
//    ptr = (T *) xmem.newptr (sizeof(T), s, "XplArray", "ReadASCIIXplArray", "ptr");
//#endif
////    printf ("Allocted: %li\n", s);
//    i = 0;
//    j = 0;
//    l = 0;
//    for (k=0; k<count; k++)
//    { // printf ("count=%li, row=%li, col=%li\n", k, i, j);
//      if (*(epl+k)==l)
//      { *(ptr+j*rows+i) = *(newptr+k);
//        j++;
//      }
//      else
//      { while (j<cols)
//        { *(ptr+rows*j+i) = (T) sqrt(-1);
//          j++;
//        }
//        i++;
//        *(ptr+i) = *(newptr+k);
//        j = 1;
//        l = 1-l;
//      }
//    }
////    printf ("End: count=%li, row=%li, col=%li\n", k, i, j);
////    printf ("Rows: %li - Cols: %li\n", rows, cols);
//    for (k=count; k<s; k++, j++)
//    { if (j==cols)
//      { j = 0;
//        i++;
//        if (i==rows) break;
//      }
////      printf ("count=%li, row=%li, col=%li\n", k, i, j);
//      *(ptr+rows*j+i) = (T) sqrt(-1);
//    }
//#ifndef MEMDEBUG
//    delete[] newptr;
//    delete[] epl;
//#else
//    xmem.delptr (newptr);
//    xmem.delptr (epl);
//    newptr = NULL;
//    epl = NULL;
//#endif
//    return (1);
//  }

  template<class T> int XplArray<T>::ReadXXplArray (char const *datei)
  { FILE *f = NULL;
    int i, rform = 0;
    long maxdim, s, adim;
    char format[8], *tok;
    char errmsg[1024];

    T *newptr;

    f = opendatafile (datei, "rb");
    if (f==NULL)
    { sprintf (errmsg, "Can not open data file : %s", datei);
      XAerrort (STRBAD, errmsg);
      return(0);
    }

    fread (format, sizeof(char), 8, f);
    if ((format[0]==26) && (format[1]==27) && (format[2]==27) && (format[3]==27) &&
        (format[4]==27) && (format[5]==27) && (format[6]==27) && (format[7]==27))
      rform = 401;
    if (rform == 0) 
    { fclose (f);
      return(-1);
    }
    fread (&maxdim, sizeof(long), 1, f);
    if (maxdim > MAX_XARRAY) 
    { sprintf (errmsg, "X Format file with mismatching dimensions : %s", datei);
      XAerrort (DIMCON, errmsg);
      fclose (f);
      return (0);
    }
    dim = 1;
    for (i=0; i<maxdim; i++)
    { fread (&adim, sizeof(long), 1, f);
      dim.setdim(i, adim);
    }
    s = dim.prod();
#ifndef MEMDEBUG
    newptr = new T[s];
#else
    newptr = (T *) xmem.newptr (sizeof(T), s, "XplArray", "ReadASCIIXplArray", "newptr");
#endif
    fread (newptr, sizeof(T), s, f);
    setptr(newptr);
    fclose (f);
    return (1);
  }

  template<class T> int XplArray<T>::WriteXXplArray (char const *datei) const
  { FILE *f = NULL;
    int i;
    long maxdim = MAX_XARRAY, s, adim;
    char format[8], *tok;
    char errmsg[1024];

    f = opendatafile (datei, "wb");
    if (f==NULL)
    { sprintf (errmsg, "Can not open file for writing : %s", datei);
      XAerrort (STRBAD, errmsg);
      return(0);
    }

    for (i=0; i<8; i++)
      format[i] = (i ? 27 : 26);
    fwrite (format, sizeof(char), 8, f);
    fwrite (&maxdim, sizeof(long), 1, f);
    for (i=0; i<MAX_XARRAY; i++)
    { adim = dim(i);
      fwrite (&adim, sizeof(long), 1, f);
    }
    s = dim.prod();
    fwrite (ptr, sizeof(T), s, f);
    fclose(f);
    return 0;
  }

  template<class T> int XplArray<T>::WriteASCIIXplArray (char const *datei, const char *fmt, int mode) const
  { FILE *f = NULL;
    T *subptr;
    long i, j, k, l, s;
    XplIndex tmp;
    char *tok, errmsg[1024];

    if (mode)
      f = opendatafile (datei, "w");
    else
      f = opendatafile (datei, "a");
    if (f==NULL)
    { sprintf (errmsg, "Can not open file for writing : %s", datei);
      XAerrort (STRBAD, errmsg);
      return(0);
    }

    l = dim(0)*dim(1);    
    s = dim.prod()/l;
    for (tmp=0, k=0; k<s; k++)
    { subptr = ptr+k*l;
      for (i=0; i<dim(0); i++)
      { for (j=0; j<dim(1); j++, tmp.inc(dim))
        { x4fprintf (f, fmt, *(subptr+j*dim(0)+i));
          fprintf (f, (j==dim(1)-1) ? "\n" : " ");
        }
      }
    } 
    fclose(f);   
	return 0;
  }

  template<class T> void XplArray<T>::scalar (T val)
  { long i;
    T *newptr;
    dim = 1;
#ifndef MEMDEBUG
    newptr = new T[1];
#else
    newptr = (T *) xmem.newptr (sizeof(T), 1, "XplArray", "scalar", "newptr");
#endif
    *newptr = val;
    setptr (newptr);
  }

  template<class T> void XplArray<T>::vector (long n, T *tptr)
  { long i;
    T *newptr;
    dim = 1;
    dim.setdim (0, n);
#ifndef MEMDEBUG
    newptr = new T[n];
#else
    newptr = (T *) xmem.newptr (sizeof(T), n, "XplArray", "vector", "newptr");
#endif
    ::memcopy (newptr, tptr, n); 
    setptr (newptr);
  }

  template<class T> void XplArray<T>::matrix (long n, long p, T *tptr)
  { long i;
    T *newptr;
    dim = 1;
    dim.setdim(0, n);
    dim.setdim(1, p);
#ifndef MEMDEBUG
    newptr = new T[n*p];
#else
    newptr = (T *) xmem.newptr (sizeof(T), n*p, "XplArray", "matrix", "newptr");
#endif
    ::memcopy (newptr, tptr, n*p); 
    setptr (newptr);
  }

  template<class T> long XplArray<T>::xpltypeof (void) const
  { return (xpltypeid(ptr)); }

  template<class T> int XplArray<T>::WriteXArray (FILE *f, long type) const
  { long i, s = size(), md = MAX_XARRAY, adim;
    XplIndex d = getdim();
    T *t;
    if ((type!=TYPE_LONG) && (type!=TYPE_DOUBLE)) return (1);
    fwrite (&type, sizeof(long), 1, f);
    fwrite (&md, sizeof(long), 1, f);
    for (i=0; i<MAX_XARRAY; i++)
    { adim = d(1);
      fwrite (&adim, sizeof(long), 1, f);
    }
    t = getptr();
    fwrite (t, sizeof(T), s, f);
    return 0;
  }

  template<class T> XplArray<T> XplArray<T>::index (const XplArray<long> &ind, long vdir) const
  { XplArray<T> ret (ind.dim);
    long s, sv, i, j;
    XplIndex tmp, itmp;
    if (vdir>=MAX_XARRAY)
    { XAerrort (RANGE, "index : vdir out of range");
      return (ret);
    }
    
    s  = size();
    sv = s/dim(vdir);
    for (i=0; i<ind.dim(vdir); i++)
    { tmp = 0;
      for (j=0; j<sv; j++, tmp.inc(dim, vdir))
      { itmp       = tmp;
        tmp.setdim  (vdir, i);
        itmp.setdim (vdir, ind(tmp)-1);
        ret(tmp)   = (*this)(itmp);  
      }
    }
    return (ret);
  }

  template<class T> XplArray<T>& XplArray<T>::trn (const XplArray<T> &x, int dim0, int dim1)
  { long tmp;
    if ((dim0<0) || (dim0>=MAX_XARRAY))
    { XAerrort (DIMCON, "Dim out of range");
      return (*this);
    }
    if ((dim1<0) || (dim1>=MAX_XARRAY))
    { XAerrort (DIMCON, "Dim out of range");
      return (*this);
    }
    *this = x; 
    if (dim0!=dim1)
    { 

      layernpnp (mattrn);
      dim.swap (0, 1);
/*
      XplIndex oldind, newind, xdim;
      tmp = dim[dim0];
      dim[dim0] = dim[dim1];
      dim[dim1] = tmp;
      xdim = x.getdim();
      oldind = 0;
      do
      { newind = oldind;
        tmp = newind[dim0];
        newind[dim0] = newind[dim1];
        newind[dim1] = tmp;
        (*this)(newind) = x(oldind);
      }
      while (oldind.inc(xdim));
*/
    }
    return (*this);
  }


#include "cast.hpp"

template class XplArray<int>;
template class XplArray<long>;
template class XplArray<double>;
template class XplArray<xplmask>;
template class XplArray<xplstring>;
