/* $Header: /home/xplore/cvs/xpl4gr/xarray.hpp,v 1.3 2002/08/16 12:31:54 sigbert Exp $ */

#include <string.h>	        
#include <stdlib.h>	        
#include <fstream.h>
#include <iostream.h>
#include <math.h>
#include <ctype.h>

#include "machine.h"
#include "otxtstream.hpp"
#include "global.h"
#include "xindex.hpp"
#include "xselect.hpp"
#include "initxa.hpp"
#include "property.h"
#include "xsort.hpp"

#ifdef __BORLANDC__
#include<alloc.h>

#endif

#define MAX_LINE_LENGTH 2560	
#define MAX_BLOCKSIZE  65536   

template<class T> T *vecsort (T *lptr, long rlen, long rdim, long ndir, long *dir);
template<class T> void mattrn (T *tptr, long n, long p);

#ifndef XPLARRAY
#define XPLARRAY

template<class T> class XplArray {

  T *vecsort (T *lptr, long rlen, long rdim, long ndir, long *dir);

public:
  XplIndex dim;

protected:
  T *ptr;
#ifdef OUTPUT
  char *outputfmt;
#endif

  void construct(XplArray<T> *x);
  void indexx (XplIndex sortact, int sortdim, long *indx, int nlen);

public:

  T* getptr (void) const; 
  void setptr (T* newptr);   
  XplIndex getdim (void) const ;
  void setdim (const XplIndex &d);

  XplArray ();
  XplArray (const XplIndex &newdim);
  XplArray (long d0, long d1=1, long d2=1, long d3=1, long d4=1, 
            long d5=1, long d6=1, long d7=1);
  XplArray (const XplArray<T> & x);
  ~XplArray ();

  unsigned dimension (void) const;      
  long dimension (unsigned index) const;
  inline long size (void) const { return (dim.prod()); }
 
// const
  T& operator[] (long d0) const;
  T& operator() (const XplIndex &tmp) const; 
  T& operator() (void) const;         
  T& operator() (long d0) const;
  T& operator() (long d0, long d1) const;
  T& operator() (long d0, long d1, long d2) const;
  T& operator() (long d0, long d1, long d2, long d3) const;
// non-const
  T& operator[] (long d0);
  T& operator() (const XplIndex &tmp); 
  T& operator() (void);         
  T& operator() (long d0);
  T& operator() (long d0, long d1);
  T& operator() (long d0, long d1, long d2);
  T& operator() (long d0, long d1, long d2, long d3);

  XplArray<T> bind (const XplArray<T>& x, unsigned dimen) const;
  XplArray<T>& assignment (const XplArray<T>& x);
  XplArray<T>& operator= (const XplArray<T> &x);
  XplArray<T>& operator= (T val);
  void fill (T val);
  void fill (const XplArray<T>& x);

  void reshape (XplIndex newdim);
  int isconformable (const XplArray<T>& x) const;

  void dump (const char *fmt) const;
  void mdump (const char *fmt, Xplostream *stream = NULL, const char *varname = NULL, int pdim = 1) const;
  void sdump (const char *fmt) const;

  void tfunc (T (*func) (void));
  XplArray<T> tfunct (int nr, T (*func) (T l));
  XplArray<T> tfunctt (const XplArray<T> &right, T (*func) (T l, T r)) const;
  XplArray<T> tfunct (const XplArray<T> &right, T (*func) (T l, T *r), T *val);
  XplArray<T> layernpnp (void (*func) (T* ptr, long n, long p));
  XplArray<T> layernn (void (*func) (T* ptr, long n));
  XplArray<T> layernpnpsort (T* (*func) (T* ptr, long n, long p, long ndir, long *dir), const XplArray<long> &dirmat);
  XplArray<T> layernpnp1 (void (*func) (T* ptr, long n, long p, T *retval)) const;
  XplArray<T> layernpnpdist (void (*func) (T* ptr, long n, long p, T *retval,
                                           const char *distm), const char *dist) const;
  XplArray<T> layernpnp2 (void (*func) (T* ptr, long n, long p, T* ptrl, T* ptru, long *ptrix), const XplArray<T> &l, 
                          const XplArray<T> &u, const XplArray<long> &ix);
  XplArray<T> layernpnp3 (void (*func) (T *x, long n, long p, 
                                        T *a, long na, long pa,
                                        T *b, long nb, long pb),
                          const XplArray<T> &aa, const XplArray<T> &bb) const;

  XplArray<T>& selection (const XplArray<T>& x, int nsel, const XplSelection *sel);
  XplArray<T>& assignment (int nsel, const XplSelection *sel, const XplArray<T>& x);

  XplArray<T>& sort (int n, const XplIndex *vec, int *dir);
  XplArray<T> sortvec (const XplArray<long> &dir) const;

  XplArray<int> ifunctt (const XplArray<T> &right, int (*func) (const T l, const T r)) const;

  // PC 20010608 - added selection and condition parameters
  int ReadASCIIXplArray (char const *datei, XplArray<long> *selrows = NULL, XplArray<long> *selcols = NULL, XplArray<long> *rowconds = NULL, XplArray<long> *colconds = NULL);
  int ReadXXplArray (char const *datei);
  int WriteXXplArray (char const *datei) const;
  int WriteASCIIXplArray (char const *datei, const char *fmt = "%f", int mode = 1) const;

  void scalar (T val);
  void vector (long n, T *tptr);
  void matrix (long n, long p, T *tptr);

  long xpltypeof (void) const;

  int WriteXArray (FILE *f, long type = TYPE_UNKNOWN) const;
  XplArray<T> index (const XplArray<long> &ind, long vdir = 0) const;
  XplArray<T>& trn (const XplArray<T> &x, int dim0 = 0, int dim1 = 1);

};

#include "cast.hpp"

#endif

