/* $Header: /home/xplore/cvs/xpl4gr/xindex.C,v 1.3 2002/08/16 12:31:55 sigbert Exp $ */

#include "xindex.hpp"

  long XplIndex::update (void)
  { unsigned i; 
    pind = 1L;
    for (i=0; i<MAX_XARRAY; i++)
    { shift[i] = pind;
      pind *= index[i];
    }
    return (pind);
  }

XplIndex::XplIndex (long d0, long d1, long d2, long d3, 
            long d4, long d5, long d6, long d7)
  { unsigned i;
    index[0] = d0;
    index[1] = d1;
    index[2] = d2;
    index[3] = d3;
    index[4] = d4;
    index[5] = d5;
    index[6] = d6;
    index[7] = d7;
    for (i=8; i<MAX_XARRAY; i++)
      index[i] = 1L;
    update();
  }

  XplIndex& XplIndex::operator= (const XplIndex &x)
  { unsigned i;
    if (&x !=this)
    { memcpy (index, x.index, MAX_XARRAY*sizeof(long));
      update();
    }
    return *this;
  }

  XplIndex& XplIndex::operator= (long val)
  { unsigned i;
    for (i=0; i<MAX_XARRAY; i++)
      index[i] = val;
    update();
    return *this;
  }

  int XplIndex::operator< (XplIndex& x)
  { unsigned i;
    for (i=0; i<MAX_XARRAY; i++)
      if (index[i]>=x.index[i]) return 0;
    return 1;
  }

  XplIndex XplIndex::operator<< (int s)
  { XplIndex ret;
    int i;
    for (i=0; i<MAX_XARRAY; i++)
      ret.index[i] = (i< s ? 0 : index[i]);
    ret.update();
    return (ret);
  }

  XplIndex XplIndex::operator>> (int s)
  { XplIndex ret;
    int i;
    for (i=0; i<MAX_XARRAY; i++)
      ret.index[i] = (i< MAX_XARRAY - s ? index[i+s] : 0);
    ret.update();
    return (ret);
  }

  void XplIndex::setdim(unsigned no, long newdim)
  { // if ((no<0) || (no>=MAX_XARRAY)) XAerrort(DIMBIG, "wrong dimension");
    // if (newdim<=0) XAerrort(DIMNEG, "negative dimension");
    index[no] = newdim;
    update();
  }

  long XplIndex::operator()(unsigned no) const
  { //if (no>=MAX_XARRAY) XAerrort(DIMBIG, "XplIndex []");
    if (no>=MAX_XARRAY) return (index[MAX_XARRAY-1]);
    return (index[no]);
  }

  XplIndex XplIndex::operator% (const XplIndex& x) const
  { XplIndex res;
    unsigned i;
    for (i=0; i<MAX_XARRAY; i++)
      res.index[i] = index[i] % x.index[i];
    res.update();
    return (res);
  }

  int XplIndex::inc (const XplIndex& x, long dim)
  { unsigned i;
    for (i=0; i<MAX_XARRAY; i++)
    { if (i!=dim)
      { if (index[i]<x.index[i]-1)
        { (index[i])++;
          break;
        }
        else
          index[i] = 0;
      }
    }
    update();
    for (i=0; i<MAX_XARRAY; i++)
      if ((i!=dim) && (index[i])) return (1);
    return(0);
  }

  int XplIndex::equal (XplIndex&  x, unsigned k) const
  { int i;
    for (i=0; i<MAX_XARRAY; i++)
      if ((i!=k) && (index[i] != x.index[i]))
        return 0;
    return 1;
  }

  int XplIndex::matequal (XplIndex&  x) const
  { int i;
    for (i=2; i<MAX_XARRAY; i++)
      if ((index[i] != x.index[i]) && (index[i]!=1) && (x.index[i]!=1))
        return(0);
    return 1;
  }


  XplIndex XplIndex::sum (XplIndex& x, unsigned k) const
  { int i;
    XplIndex res;
    for (i=0; i<MAX_XARRAY; i++)
    { if (i==k)
        res.index[i] = index[i] + x.index[i];
      else
        res.index[i] = index[i];
    }
    res.update();
    return res;
  }

  int XplIndex::isconf (const XplIndex& x) const
  { unsigned i;
    for (i=0; i<MAX_XARRAY; i++)
      if ((index[i]!=x.index[i]) && (index[i]!=1L) && (x.index[i]!=1L))
        return 0;
    return 1;
  }

  long XplIndex::prod (int k) const 
  { int i;
    long res = 1L;
    for (i=0; i<k; i++)
      res *= index[i];
    return res;
  }

  void XplIndex::swap (long l, long r)
  { long tmp;
    tmp = index[l];
    index[l] = index[r];
    index[r] = tmp;
    tmp = shift[l];
    shift[l] = shift[r];
    shift[r] = tmp;
  }

  long XplIndex::value (XplIndex act) const
  { int i;
#ifndef X45
    long res = 0L, size = 1L;
    for (i=0; i<MAX_XARRAY; i++)
    { res += size*(act.index[i] % index[i]);
      size *= index[i];
    }
#else
    long res = 0L;
    for (i=0; i<MAX_XARRAY; i++)
      res += (act.index[i] % index[i])*shift[i];
#endif
    return (res);
  }

  XplIndex XplIndex::value (long pos) const
  { XplIndex ret;
    long apos = pos, size = 1L, i, msize = 1L;

    for (i=0; i<MAX_XARRAY; i++)
    { ret.index[i] = apos%index[i];
      apos   = apos/index[i];
    }
    ret.update();
    return (ret);
  }

  XplIndex XplIndex::maxindex (const XplIndex& right) const 
  { int i;
    XplIndex res;

    for (i=0; i<MAX_XARRAY; i++)
      res.index[i] = (index[i] < right.index[i] ? right.index[i] : index[i]);
    res.update();
    return res;
  }

  XplIndex XplIndex::maxmatindex (XplIndex& right)
  { int i;
    for (i=2; i<MAX_XARRAY; i++)
      index[i] = (index[i] < right.index[i] ? right.index[i] : index[i]);
    update();
    return *this;
  }

  void XplIndex::dump (void) const
  { int i;
    char out[256];
    for (i=0; i<MAX_XARRAY; i++)
    { sprintf (out, "%li ", index[i]);
      cout << out;
    }
    cout << "\n";
  }


  int XplIndex::operator== (const XplIndex &right) const 
  { int i;
    for (i=0; i<MAX_XARRAY; i++)
      if (index[i]!=right.index[i]) return 0;
    return (1);
  }


