/* $Header: /home/xplore/cvs/xpl4gr/xjprotocol.C,v 1.4 2003/03/21 13:28:06 cizek Exp $ */

// implementation of client/server protocol version 0.1

// TTK, 26.9.98  last modified 1.11.98

#if defined(SOCKETSERVER)||defined(X4singleserver)

#include "object.h"
#include "setgopt.h"
#include "xjprotocol.h"

extern void logt (const char *txt);
//extern void logp (const char *txt, int size);
extern void logptr (void *ptr);
extern char* xtmpnam(void);

int  XJprotocol::checkProtocol ()
{
  char *msg;
  long len, minorVersion, majorVersion,
       answer = XJ_ACCEPTED;

  msg = sockIO->readChar (len);

  if (msg == NULL)
    return (XJ_TRANSMISSION_ERROR);

  majorVersion = getLongParameter ();
  minorVersion = getLongParameter ();

  if (majorVersion == 0L && minorVersion == 0L)
    return (XJ_TRANSMISSION_ERROR);

  // check the version 

  if (majorVersion < 0L)
    answer = XJ_DENIED;

  sendCommand (answer);

  return (answer);
}



long XJprotocol::getLongParameter ()
{
  return (sockIO->readToken ());
}

void XJprotocol::sendLongParameter (long par)
{
  sockIO->writeToken (par);
}

float XJprotocol::getFloatParameter ()
{
  float res, *r;
  long  dummy;

  r = sockIO->readFloat (dummy);

  if (dummy != 1L)
    cerr << "getFloatParameter: got more than a single value" << endl;

  res = *r;
  delete[] r;
  
  return (res);
}

void XJprotocol::sendFloatParameter (float f)
{
  sockIO->writeFloat (&f, 1);
}

double XJprotocol::getDoubleParameter ()
{
  double res, *r;
  long   dummy;

  r = sockIO->readDouble (dummy);

  if (dummy != 1L)
    cerr << "getDoubleParameter: got more than a single value" << endl;

  res = *r;
  delete[] r;
  
  return (res);
}

void XJprotocol::sendDoubleParameter (double d)
{
  sockIO->writeDouble (&d, 1);
}

long XJprotocol::getCommand ()
{
  return (sockIO->readToken ());
}

void XJprotocol::sendCommand (long cmd)
{
  sockIO->writeToken (cmd);
}

void  XJprotocol::sendLong   (long *l, long n)
{
  sockIO->writeLong (l, n);
}

void  XJprotocol::sendFloat   (float *f, long n)
{
  sockIO->writeFloat (f, n);
}

float *XJprotocol::getFloat (long n)
{
  return (sockIO->readFloat (n));
}

long *XJprotocol::getLong (long n)
{
  return (sockIO->readLong (n));
}

char * XJprotocol::readXploReCommands ()
{
  long dummy;

  return (sockIO->readChar (dummy));
}

void XJprotocol::sendString      (const char *s)
{
  logt ("XJprotocol::sendString");
  sendCommand (RESULTS);
  sendCommand (END_STRING);
  sockIO->writeString (s);
}

void XJprotocol::sendStringParameter  (const char *s)
{
  sockIO->writeString (s);
}

char * XJprotocol::getStringParameter  ()
{
  long dummy;

  return (sockIO->readChar (dummy));
}


void XJprotocol::collectString      (const char *s)
{
  long len = strlen (s);

  while (currentLength + len >= XJmaxLength)
    {
         if (currentLength > 0)     // PC 20011214 - do it only if buffer is filled
         {
    sendCommand (CONTINUE_STRING);
    sockIO->writeString (StringBuffer);
    StringBuffer[0] = '\0';        
    currentLength = 0;
         }

    strncpy (StringBuffer, s, XJmaxLength);
    currentLength = strlen (StringBuffer);
    s   += currentLength;
    len -= currentLength;
    }

  strncat (StringBuffer, s, len);
  currentLength += len;
}

void XJprotocol::sendCollectedString ()
{
  sendCommand (END_STRING);
  sockIO->writeString (StringBuffer);
  
  StringBuffer[0] = '\0';
  currentLength   = 0L;
}

void XJprotocol::sendDisplayData (long id, const char *name,
				  long nrow, long ncol, long type, long dim)
{
  long param[5];

  param[0] = id;
  param[1] = nrow;
  param[2] = ncol;
  param[3] = type;
  param[4] = dim;

  sockIO->writeLong (param, 5);
  sockIO->writeString (name);
}

void XJprotocol::sendPlotData (long id, long nrow, long ncol)
{
  long param[3];

  param[0] = id;
  param[1] = nrow;
  param[2] = ncol;

  sockIO->writeLong (param, 3);
}


void XJprotocol::sendContent (XplDouble &x, const char *varstr, char *str)
#ifndef MATRIX
{
// PC 20011213 - completely rewritten
     Xplostream outstream;

     logt ("XJprotocol::sendContent(XplDouble)");
     sendCommand (RESULTS);

     logt ("begin params");
     //     logp (str, strlen(str));
     //     logp (varstr, strlen(varstr));
     logptr (&x);
     logptr (&outstream);
     logt ("end params");
     if (str != NULL)
          x.mdump(str, &outstream, varstr);
     else
          x.mdump("%5.5g", &outstream, varstr);

     logt ("collectString");
    collectString (outstream.ptr());

    logt ("sendCollectedString");
    sendCollectedString ();

// PC 20011213 : original code - writing to disk
/*
  char line[1024], *fname;
  long length;
		
  fname = xtmpnam(); // SK 05.06.2000
  //  logt ("XJprotocol::sendContent #ifndef MATRIX");
  //logt (fname);
  fstream f (fname, ios::out);

  if (f.bad ())
    {
    cerr << "Couldn't open temporary file (out). Exit ..." << endl;
    exit (1);
    }

  Xplostream tmpstream(f);     // tk + sk 20.2.98
  if (str != NULL)
    x.mdump (str, &tmpstream, varstr);
   else
    x.mdump ("%5.5g", &tmpstream, varstr);

  f.close();
  
  fstream fi (fname, ios::in);
  if (fi.bad ())
    {
    cerr << "Couldn't open temporary file (in). Exit ..." << endl;
    exit (1);
    }

  sendCommand (RESULTS);

  while (fi.getline (line, 1023))
    {
    length = strlen (line);
    if (length > 0L)
      line[length-1] = '\n';
    collectString (line);
    }

  sendCollectedString ();
  remove (fname);
*/
}
#else  
// void XJprotocol::sendContent
{
  x.out (sockIO);
}
#endif  // !MATRIX

#ifdef _XPLCDEF_
void XJprotocol::sendContent (XplChar &x, const char *varstr, char *fst )
{
  char line[1030], *fname;
  long length;
		
  fname = xtmpnam();  // SK 05.06.2000
  //  logt ("XJprotocol::sendContent #ifdef _XPLCDEF_");
  //logt (fname);
  fstream f(fname, ios::out);
  if (f.bad())
    {
    cerr << "Couldn't open temporary file (out). Exit ..." << endl;
    exit (1);
    }
  Xplostream tmpstream(f);      // tk + sk 20.2.98
  x.mdump ("%s", &tmpstream, varstr);
  f.close ();
  	
  fstream fi(fname, ios::in);
  if (fi.bad ())
    {
    cerr << "Couldn't open temporary file (in). Exit ..." << endl;
    exit (1);
    }
  logt ("XJprotocol::sendContent((XplChar)");
  sendCommand (RESULTS);

  while (fi.getline (line, 1023))
    {

    length = strlen (line);
    // TK+SK 27.01.99
    if (length>0L)
    { line[length] = '\r';
      line[length+1] = '\n';
      line[length+2] = 0;
    }
    //    if (length > 0L)
    //      line[length-1] = '\n';
    // printf ("sending textoutput : %s", line);
    collectString (line);
    }

  sendCollectedString ();
  remove(fname);

}
#endif   // XPLCDEF

#endif  // SOCKETSERVER
