/* $Header: /home/xplore/cvs/xpl4gr/xjprotocol.h,v 1.4 2003/03/21 13:28:06 cizek Exp $ */

// Protocol header for client/server communication

// TTK, 26.9.98

#if defined(SOCKETSERVER) || defined(X4singleserver)

#ifndef _XJPROTOCOL_
#define _XJPROTOCOL_

#define BEGIN_SEND_DATA     1L
#define END_SEND_DATA       2L
#define XJmaxLength         4095L
//#define XJmaxLength 127L

#define XJ_ACCEPTED           10L
#define XJ_DENIED             20L
#define XJ_TOO_MANY_CLIENTS   30L
#define XJ_TRANSMISSION_ERROR 40L

#include "object.h"             // wg XplDouble <-> Matrix
#include "socketio.h"

// undef close (definition in socketio.h)
#ifdef close
#undef close
#endif

#define REQUEST2PING    10001L
#define REQUEST2BREAK   10002L
#define REQUEST2SERVE   10003L

#define END_OF_SESSION  10004L
#define SEND_COMMANDS   10005L
#define ACCEPTED        10006L

#define RESULTS         10007L

#define CREATE_DISPLAY  10008L
#define SHOW2D          10009L
#define SHOW3D          -9L
#define DATA_PART       10010L
#define END_DATA_PART   10011L
#define POLYGON         10012L

#define CONTINUE_STRING 10013L
#define END_STRING      10014L

#define READ_VALUE      10015L
#define SELECT_ITEM     10016L

#define REQUEST_DATA    10017L
#define TRANSMIT_DATA   10018L

#define END_OF_XPLORE   10019L

#define POINTTEXT       10020L // SK + TK 21.04.99
#define READY2SERVE     10021L // SK 12.06.99
#define STATUSMSG       10022L // SK 06.02.00 

#define REQUEST_DOUBLE  10027L // SK 15.10.01
#define TRANSMIT_DOUBLE 10028L // SK 15.10.01

#define FAIL            10098L // SK 22.10.01
#define SUCCESS         10099L // SK 22.10.01

#define SETGOPT           10100L // SK + TK 29.11.99
#define SETGOPT_BASE      10101L // SK 21.11.2002
#define END_OF_SETGOPT    10199L // SK 21.11.2002

#define SETSIZE          10200L // SK 18.02.2000
#define AXESON           10201L // SK 04.04.2000
#define AXESOFF          10202L // SK 04.04.2000

#define BEGINXPLCMD      11000L // SK 16.01.2002
#define ENDXPLCMD        11001L // SK 16.01.2002
#define BEGINTXTARRAY    11010L // SK 16.01.2002
#define ENDTXTARRAY      11011L // SK 16.01.2002
#define BEGINNUMARRAY    11020L // SK 16.01.2002
#define ENDNUMARRAY      11021L // SK 16.01.2002

/*
  General:

  Client and server exchange data, where a token introduces
  the communication. The allowed tokens are defined as constants.
  The tokens determine the following data. E.g. receiving the
  token 'SHOW_2D' means that the following data describe fully
  which display is involved, which coordinates in the display are
  set, all data belonging to the plot and all display attributes.

  Scheme:

  1. Establish connection
  2. Make sure that client and server match
  3. Server handles request to server. If it denies, the
     connection will be closed.
  4. Client listens for user action and sends data to the
     server, if necessary.
  5. Server processes the data and sends back the results.
  6. If user finishes the session, the client logs off and exits.
  7. The server closes the conncetion and listens for new ones.
  */

class XJprotocol {

public:

XJprotocol (SocketIO *socketIO) : sockIO(socketIO), currentLength(0L) { StringBuffer[0] = '\0'; }

~XJprotocol ()                    { delete sockIO;     }

void append (SocketIO *socketIO)  { sockIO = socketIO; }
int  checkProtocol ();
void close ()                     { delete sockIO;     }

// I/O functions


// Following functions exchange commands between server and client.
long getCommand  ();
void sendCommand (long cmd);


/* Following functions exchange array of length n between server and
   client.

   ATTENTION: Long means exactly 4 bytes signed integer, float means
              exactly 4 Bytes floating point numbers.
   REMEMBER:  In Java the 4 bytes signed integer is 'int' !

   Code, that checks the 'long condition (4 bytes)' has still to be
   implemented
 */

void   sendLong   (long *, long n);
long * getLong    ();

void    sendFloat (float *, long n);
//void    getFloat (float *, long n);
float * getFloat  (long n);
long * getLong (long n);

// sendString sends a character array, which will be printed in the clients
// output window
void sendString           (const char *);


// Following functions exchange parameters of different types between
// server and client.

void sendStringParameter  (const char *);
void sendLongParameter    (long);
void sendFloatParameter   (float);
void sendDoubleParameter  (double);

char * getStringParameter ();
long   getLongParameter   ();
float  getFloatParameter  ();
double getDoubleParameter ();

// Following functions collect a string in a buffer. The whole
// transfer is finished by the final call of sendCollectedString.
// The functions are intended for sending a block of character lines
// at once instead of each lines separated, where the block is built
// during execution of several lines of code.
// The buffer handling (overflow) is handled automatically.
void collectString       (const char *s);
void sendCollectedString ();


// Following functions send the content of XplDouble and XplChar
// objects to the client's output window.
void sendContent (XplDouble &x, const char *varstr, char *str);
#ifdef _XPLCDEF_
void sendContent (XplChar &x, const char *varstr, char *str);
#endif

// gets commands form client and executes them 
char * readXploReCommands ();

// sends output to client
int sendOutput      (const char *oname);


// sendDisplayData sends display related header info to the client
void sendDisplayData (long id, const char *name,
		     long nrow, long ncol, long type, long dim);
// SK + TK, 25.01.99
void sendPlotData (long id, long nrow, long ncol);
// Sk 29.11.99

private:
  char StringBuffer[XJmaxLength+1L];
  long currentLength;

  SocketIO *sockIO;
};
  
#endif

#endif  // SOCKETSERVER

