/* $Header: /home/xplore/cvs/xpl4gr/xmask.C,v 1.3 2002/08/16 12:31:56 sigbert Exp $ */

#include "xmask.hpp"
#include "x4math2.h"

  XplMask::XplMask (const XplIndex newdim) : XplArray<xplmask>(newdim) { };

  XplMask::XplMask (long d0, long d1 , long d2, long d3, long d4, long d5, long d6, long d7) : XplArray<xplmask>(d0, d1, d2, d3, d4, d5, d6, d7) { };

  XplMask::XplMask (const XplArray<xplmask>& x) : XplArray<xplmask>(x) {};

  XplMask& XplMask::operator= (const XplArray<xplmask>& x)
  { if (this != &x)
    { long i,s;
      if (ptr !=NULL)
#ifndef MEMDEBUG
        delete[] ptr;
#else
        xmem.delptr (ptr);
#endif
      dim = x.dim;
#ifndef MEMDEBUG
      ptr = new xplmask[x.size()]; // malloc(x.size()*sizeof(T));
#else
      ptr = (xplmask *) xmem.newptr (sizeof(xplmask), x.size(), "XplMask", "operator=", "ptr");
#endif
      ::memcopy (ptr, &(x[0]), x.size()); //*sizeof(xplmask));
    }
    return (*this);
  }

  long XplMask::xpltypeof (void)
  { return (TYPE_MASK); }

  int XplMask::WriteXArray (FILE *f)
  { long i, s = size(), type = TYPE_MASK, md = MAX_XARRAY, tmp;
    XplIndex d = getdim();
    fwrite (&type, sizeof(long), 1, f);
    fwrite (&md, sizeof(long), 1, f);
    for (i=0; i<MAX_XARRAY; i++)
    { tmp = dim(i);
      fwrite (&tmp, sizeof(long), 1, f);
    }
    fwrite (ptr, sizeof(xplmask), s, f);
    return 0;
  }

  void XplMask::mdump (const char *fmt, Xplostream *stream, const char *varname,
              int pdim)
  { char out[256];
    xplmask *subptr;
    long i, j, k, l, s, m, starfmt;
    unsigned long count, maxcount;
    XplIndex tmp;
    Xplostream streammdump(stream);
    int r,g,b,sy,si;
    char 
      *outheadline = property.getProperty("outheadline"), 
      *outlineno = property.getProperty("outlineno"), 
      *outlayerline = property.getProperty("outlayerline");
    long outmaxdata = atol (property.getProperty("outmaxdata"));
    if (dim.prod()<=0)
    { XAerrort (ARRAYBAD, "Arraysize <= 0");
      return;
    }

    l = dim(0)*dim(1);
    s = dim.prod()/l;
    m = (long) (1.0+x4log10((double) (dim(0))));
    if (varname)
      sprintf (out, outheadline, varname);
    else
      sprintf (out, outheadline, "unnamed object");
    streammdump << out;

    starfmt = (strchr(outlineno, '*') != NULL);
    maxcount = (outmaxdata<1 ? 0xffffffff : outmaxdata); 
    count    = 0L;

    for (tmp=0, k=0; (k<s) && (count<maxcount) ; k++)
    { subptr = ptr+k*l;
      if (pdim)
      { if (s>1)
        { sprintf (out, outlayerline, 
                    tmp(2)+1,tmp(3)+1,tmp(4)+1,tmp(5)+1,tmp(6)+1,tmp(7)+1);
          streammdump << out;
        }
      }

      for (i=0; (i<dim(0)) && (count<maxcount); i++)
      { if (starfmt)
          sprintf (out, outlineno, m, i+1);
        else
          sprintf (out, outlineno, i+1);
        streammdump << out;
        for (j=0; (j<dim(1)) && (count<maxcount); j++, tmp.inc(dim), count++)
        { (subptr+j*dim(0)+i)->GetRGB (&r, &g, &b);
 	  (subptr+j*dim(0)+i)->GetSymbol (&sy);
          (subptr+j*dim(0)+i)->GetSize (&si);
          sprintf (out, "R=%i G=%i B=%i SYM=%i S=%i", r,g,b,sy,si);
          streammdump << out;
          streammdump << " ";
        }
        streammdump << "\n";
      }
    }
    if (count==maxcount)
      streammdump << "\r\nSorry, this object is too big !!\r\n";

  }

  XplMask& XplMask::operator= (const xplmask m)
  { long i, s;
    s = size();
    for (i=0; i<s; i++)
      *(ptr+i) = m;
    return (*this);
  }

  XplMask& XplMask::fill (const xplmask from, const xplmask to)
  { float fh, fl, fs, fsy, fsi, th, tl, ts, tsy, tsi, l, r, g, b, mi, ma;
    long i, s=size();
    int ri, gi, bi, sy, si;

    // From RGB to HLS
    from.GetRGB(&ri, &gi, &bi);
    r  = ri/255.0;
    g  = gi/255.0;
    b  = bi/255.0;
    mi = (r<g  ? r : g);
    mi = (b<mi ? b : mi);
    ma = (r>g  ? r : g);
    ma = (b>ma ? b : ma);
    fl = (mi+ma)/2;
    if (mi==ma)
    { fs = fh = 0.0; }
    else
    { fs = ((l<=0.5) ? (ma-mi)/(ma+mi) : (ma-mi)/(2-ma-mi));
      mi = ma-mi;
      if (r==ma)
        fh = (g-b)/mi;
      else if (g==ma)
        fh = 2+(b-r)/mi;
      else
        fh = 4+(r-g)/mi;
      fh *= 60;
      if (fh<0) fh += 360;
    }
    from.GetSymbol (&ri);
    fsy = ri;
    from.GetSize (&ri);
    fsi = ri;

    to.GetRGB(&ri, &gi, &bi);
    r  = ri/255.0;
    g  = gi/255.0;
    b  = bi/255.0;
    mi = (r<g  ? r : g);
    mi = (b<mi ? b : mi);
    ma = (r>g  ? r : g);
    ma = (b>ma ? b : ma);
    tl = (mi+ma)/2;
    if (mi==ma)
    { ts = fh = 0.0; }
    else
    { ts = ((l<=0.5) ? (ma-mi)/(ma+mi) : (ma-mi)/(2-ma-mi));
      mi = ma-mi;
      if (r==ma)
        th = (g-b)/mi;
      else if (g==ma)
        th = 2+(b-r)/mi;
      else
        th = 4+(r-g)/mi;
      th *= 60;
      if (th<0) th += 360;
    }
    from.GetSymbol (&ri);
    tsy = ri;
    from.GetSize (&ri);
    tsi = ri;

    if (fh>th)
    { if (fh-th>180.0) th += 360; }
    else
    { if (th-fh>180.0) fh += 360; }
    for (i=0; i<s; i++)
    { l = ((float) i)/(s-1);
      r = fmod ((1-l) * fh + l * th, 360.0);
      g = (1-l) * fl + l * tl;
      b = (1-l) * fs + l * ts;
      (*(ptr+i)).SetHLS (r, g, b);
      sy = (int) ((1-l) * fsy + l * tsy);
      (*(ptr+i)).SetSymbol (sy);
      si = (int) ((1-l) * fsi + l * tsi);
      (*(ptr+i)).SetSize (si);
    }
    return (*this);
  }

  void XplMask::setred (const XplArray<long> &x)
  { long i, s, *xptr;
    for (i=0; i<MAX_XARRAY; i++)
    { if (x.dim(i)!=dim(i))
      { XAerrort(DIMCON, "setred : dimension are not conformable");
        return;
      }
    }
    xptr = x.getptr();
    s = size();
    for (i=0; i<s; i++)
      (*(ptr+i)).SetRed(*(xptr+i));
  }

  void XplMask::setgreen (const XplArray<long> &x)
  { long i, s, *xptr;
    for (i=0; i<MAX_XARRAY; i++)
    { if (x.dim(i)!=dim(i))
      { XAerrort(DIMCON, "setgreen : dimension are not conformable");
        return;
      }
    }
    xptr = x.getptr();
    s = size();
    for (i=0; i<s; i++)
      (*(ptr+i)).SetGreen(*(xptr+i));
  }

  void XplMask::setblue (const XplArray<long> &x)
  { long i, s, *xptr;
    for (i=0; i<MAX_XARRAY; i++)
    { if (x.dim(i)!=dim(i))
      { XAerrort(DIMCON, "setblue : dimension are not conformable");
        return;
      }
    }
    s = size();
    xptr = x.getptr();
    for (i=0; i<s; i++)
      (*(ptr+i)).SetBlue (*(xptr+i));
  }

  void XplMask::setsymbol (const XplArray<long> &x)
  { long i, s, *xptr;
    for (i=0; i<MAX_XARRAY; i++)
    { if (x.dim(i)!=dim(i))
      { XAerrort(DIMCON, "setsymbol : dimension are not conformable");
        return;
      }
    }
    xptr = x.getptr();
    s = size();
    for (i=0; i<s; i++)
      (*(ptr+i)).SetSymbol(*(xptr+i));
  }

  void XplMask::setsize (const XplArray<long> &x)
  { long i, s, *xptr;
    for (i=0; i<MAX_XARRAY; i++)
    { if (x.dim(i)!=dim(i))
      { XAerrort(DIMCON, "setsize : dimension are not conformable");
        return;
      }
    }
    xptr = x.getptr();
    s = size();
    for (i=0; i<s; i++)
      (*(ptr+i)).SetSize(*(xptr+i));
  }

  XplArray<long> XplMask::getred (void) const
  { long i, s, *xptr;
    int tmp;
    XplArray<long> ret (dim);

    xptr = ret.getptr();
    s = size();
    for (i=0; i<s; i++)
    { (*(ptr+i)).GetRed (&tmp);
       *(xptr+i) = tmp;
    }
    return (ret);
  }

  XplArray<long> XplMask::getgreen (void) const
  { long i, s, *xptr;
    int tmp;
    XplArray<long> ret (dim);

    xptr = ret.getptr();
    s = size();
    for (i=0; i<s; i++)
    { (*(ptr+i)).GetGreen (&tmp);
       *(xptr+i) = tmp;
    }
    return (ret);
  }

  XplArray<long> XplMask::getblue (void) const
  { long i, s, *xptr;
    int tmp;
    XplArray<long> ret (dim);

    xptr = ret.getptr();
    s = size();
    for (i=0; i<s; i++)
    { (*(ptr+i)).GetBlue (&tmp);
       *(xptr+i) = tmp;
    }
    return (ret);
  }

  XplArray<long> XplMask::getsymbol (void) const
  { long i, s, *xptr;
    int tmp;
    XplArray<long> ret (dim);

    xptr = ret.getptr();
    s = size();
    for (i=0; i<s; i++)
    { (*(ptr+i)).GetSymbol (&tmp);
       *(xptr+i) = tmp;
    }
    return (ret);
  }

  XplArray<long> XplMask::getsize (void)const 
  { long i, s, *xptr;
    int tmp;
    XplArray<long> ret (dim);

    xptr = ret.getptr();
    s = size();
    for (i=0; i<s; i++)
    { (*(ptr+i)).GetSize(&tmp);
       *(xptr+i) = tmp;
    }
    return (ret);
  }

  XplArray<long> XplMask::group (void) const
  { long i, j, s, *xptr;
    XplArray<long> ret (dim);

    xptr = ret.getptr();
    s = size();
    for (i=0; i<s; i++)
    { for (j=0; j<i; j++)
       if (ptr[i]==ptr[j]) break;
      *(xptr+i) = j+1;
    }
    return (ret);
  }
