/* $Header: /home/xplore/cvs/xpl4gr/xnan.C,v 1.3 2002/08/16 12:31:56 sigbert Exp $ */

/* NaN Handling of XploRe

   tested for

     Solaris 2 / Sparc
     Linux 2   / Intel
     HP UX 10  / A 9000
     SGI       / Irix


   by TTK, 23.4.97   28.3.98

*/

/* Copyright (C) 1992 Free Software Foundation, Inc.
This file is part of the GNU C Library.

The GNU C Library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

The GNU C Library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with the GNU C Library; see the file COPYING.LIB.  If
not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA.  */


/* Following setting works for the tested machines. Maybe for other
   machines the endian type must be set explicitly.

   Sparc and PA is big endian.
   Intel is little endian but this flag is set before by the GNU compiler.
*/

#ifdef WIN32
#include <float.h>
#endif

#ifndef __BIG_ENDIAN
#define __BIG_ENDIAN 4321
#endif

#ifndef __LITTLE_ENDIAN
#define __LITTLE_ENDIAN 1234
#endif

#ifdef i386
#define __BYTE_ORDER __LITTLE_ENDIAN
#endif

#ifdef sparc
#define __BYTE_ORDER __BIG_ENDIAN
#endif

#ifdef hppa
#define __BYTE_ORDER __BIG_ENDIAN
#endif

#ifdef __sgi
#define __BYTE_ORDER __BIG_ENDIAN
#endif

// default is little endian
#ifndef __BYTE_ORDER
#define __BYTE_ORDER __LITTLE_ENDIAN
#endif

union ieee754_double
  {
    double d;
    
    /* This is the IEEE 754 double-precision format.  */
    struct
      {
#if	__BYTE_ORDER == __BIG_ENDIAN
	unsigned int negative:1;
	unsigned int exponent:11;
	/* Together these comprise the mantissa.  */
	unsigned int mantissa0:20;
	unsigned int mantissa1:32;
#endif
#if	__BYTE_ORDER == __LITTLE_ENDIAN
	/* Together these comprise the mantissa.  */
	unsigned int mantissa1:32;
	unsigned int mantissa0:20;
	unsigned int exponent:11;
	unsigned int negative:1;
#endif
      } ieee;
    struct
      {
#if	__BYTE_ORDER == __BIG_ENDIAN
	unsigned int negative:1;
	unsigned int exponent:11;
	unsigned int quiet_nan:1;
	/* Together these conprise the mantissa.  */
	unsigned int mantissa0:19;
	unsigned int mantissa1:32;
#else
	/* Together these conprise the mantissa.  */
	unsigned int mantissa1:32;
	unsigned int mantissa0:19;
	unsigned int quiet_nan:1;
	unsigned int exponent:11;
	unsigned int negative:1;
#endif
      } ieee_nan;
  };

#define _IEEE754_DOUBLE_BIAS            0x3ff   /* added to exp of ieee754_double */


#include "xnan.h"

//inline
int  XisInf    (double x)
// Bit pattern for Inf:
// x    11111111111  0000000000000000000000000000000000000000000000000000
// sign  exponent      mantissa
{
#ifdef WIN32
  
  if (_finite (x))
    return (FALSE);
  if (!_isnan (x))
	return (FALSE);
  return (TRUE);

#else
	
  ieee754_double xx;

  xx.d = x;
  if (xx.ieee.exponent == 0x07ffUL)
    if ((xx.ieee.mantissa0 == 0) && (xx.ieee.mantissa1 == 0))
      return (TRUE);

  return (FALSE);

#endif   // WIN32
}

//inline
int  XisPosInf    (double x)
// Bit pattern for Inf:
// x    11111111111  0000000000000000000000000000000000000000000000000000
// sign  exponent      mantissa
{
#ifdef WIN32

  return ((_fpclass (x) == _FPCLASS_PINF) ? TRUE : FALSE);

#else
	
  ieee754_double xx;

  xx.d = x;
  if (XisInf (x))
    return (xx.ieee.negative ? FALSE : TRUE);

  return (FALSE);

#endif   // WIN32
}

//inline
int  XisNegInf    (double x)
// Bit pattern for Inf:
// x    11111111111  0000000000000000000000000000000000000000000000000000
// sign  exponent      mantissa
{
#ifdef WIN32

  return ((_fpclass (x) == _FPCLASS_NINF) ? TRUE : FALSE);

#else

  ieee754_double xx;

  xx.d = x;
  if (XisInf (x))
    return (xx.ieee.negative ? TRUE : FALSE);

  return (FALSE);

#endif   / WIN32
}

int  XisNaN    (double x)
// Bit pattern for Inf:
// x    11111111111  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
// sign  exponent      mantissa != 0
{
#ifdef WIN32

	return (_isnan (x));

#else

  ieee754_double xx;

  xx.d = x;
  if (xx.ieee.exponent == 0x07ffUL)
    if ((xx.ieee.mantissa0 != 0) || (xx.ieee.mantissa1 != 0))
      return (TRUE);

  return (FALSE);

#endif  // WIN32
}

//inline
int  XisNumber (double x)
{
  ieee754_double xx;

  xx.d = x;
  return (XisInf (x) || XisNaN (x) ? FALSE : TRUE);
}
