/* $Header: /home/xplore/cvs/xpl4gr/xnumber.hpp,v 1.3 2002/08/16 12:31:57 sigbert Exp $ */

#include <math.h>
#include "values.h"
#include "xarray.hpp"
#include "xplexec.h"

#ifndef XPLNUMBER
#define XPLNUMBER

template<class T> T plus (T l, T r);
template<class T> T minus (T l, T r);
template<class T> T divi (T l, T r);
template<class T> T modu (T l, T r);

template<class T> T mult (T l, T r);

template<class T> int eq (const T l, const T r);
template<class T> int leq (const T l, const T r);
template<class T> int neq (const T l, const T r);
template<class T> int geq (const T l, const T r);
template<class T> int le (const T l, const T r);
template<class T> int ge (const T l, const T r);

template<class T> void fillgrid (T* ptr, long p, long *n, T*
start, T* inc);

template<class T> class XplNumber : public XplArray<T> {

  public:
  
  XplNumber (const XplIndex newdim);
  XplNumber ();
  XplNumber (long d0, long d1=1, long d2=1, long d3=1, long d4=1, long d5=1, long d6=1, long d7=1);
  ~XplNumber ();
  XplNumber (const XplArray<T> &x);
//  XplNumber (const XplNumber<T> &x); SK 270398

  XplNumber<T>& operator= (const XplArray<T> &x);		
  XplNumber<T>& operator= (const XplNumber<T> &x);		

  XplNumber<T> operator+(const XplNumber<T> &right) const;
  XplNumber<T> operator-(const XplNumber<T> &right) const;
  XplNumber<T> operator*(const XplNumber<T> &right) const;
  XplNumber<T> operator/(const XplNumber<T> &right) const;
  XplNumber<T> operator%(const XplNumber<T> &right) const;

  XplNumber<int> operator==(const XplNumber<T> &right) const;
  XplNumber<int> operator<=(const XplNumber<T> &right) const;
  XplNumber<int> operator!=(const XplNumber<T> &right) const;
  XplNumber<int> operator>=(const XplNumber<T> &right) const;
  XplNumber<int> operator<(const XplNumber<T> &right) const;
  XplNumber<int> operator>(const XplNumber<T> &right) const;

  XplNumber<T> extreme (const XplNumber<T> &right, int dir, int cmp);

  XplNumber<long> mknn (int knn, const XplNumber<T> &xest) const;

  XplNumber<T> matmult (const XplNumber<T> &right) const;      

  void bindata (const XplNumber<T> delta, const XplNumber<T> orig, XplNumber<T> *y,
                XplNumber<long> *xb, XplNumber<T> *yb) const;

  XplNumber<T> tmatmult (const XplNumber<T> &right) const;      

  void conv (const XplNumber<long> &xb,const  XplNumber<long> &wx, const XplNumber<T> &wy, XplNumber<long> *sym, XplNumber<long> *xc, XplNumber<T> *yc, XplNumber<long> *or) const;

  XplNumber<T> grid (const XplNumber<T> &h, const XplNumber<long> &n) const;

  XplNumber<T> sum (long d, const XplNumber<long> *cond = NULL) const;
  XplNumber<T> sumnan (long d) const;
  XplNumber<T> mul (long d, const XplNumber<long> *cond = NULL) const;
  XplNumber<T> mulnan (long d) const;

  XplNumber<T> maximum (long d, const XplNumber<long> *cond=NULL) const;
  XplNumber<T> minimum (long d, const XplNumber<long> *cond=NULL) const;

  XplNumber<T> cumsum (long d, const XplNumber<long> *cond=NULL) const;
  XplNumber<T> cumsumnan (long d) const;
  XplNumber<T> cummul (long d, const XplNumber<long> *cond=NULL) const;
  XplNumber<T> cummulnan (long d) const;

  void fillvec (T start, T step, int op);

  XplNumber<long> maxind (long d, const XplNumber<long> *cond=NULL) const;
  XplNumber<long> minind (long d, const XplNumber<long> *cond=NULL) const;

};

#endif





