/* $Header: /home/xplore/cvs/xpl4gr/xplbasefun.C,v 1.3 2002/08/16 12:31:57 sigbert Exp $ */

// XploRe core functions

// here are all basic mathematical functions implemented, which
// transform reals to reals (e.g. sin)

//  TTK, 9.6.95

#include <math.h>
#include "x4math2.h"

#include "machine.h"

#ifdef HAS_SMPMATH
#include "x4math.h"
#endif

#include "object.h"
#include "function.h"
#include "xplfun.h"
#include "init.h"
#include "ttkio.h"

double x4acos (double x); // cause on SUN acos(x) for abs(x)>1 gives a wrong result

XplObject*	matabs (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "abs: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "abs: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (fabs);
#else
  s->tfunct (0, fabs);
#endif

  return (new XplObject (s, "abs"));
}

XplObject*	matacos (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "acos: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "acos: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (acos);
#else
  s->tfunct (0, x4acos);
#endif

  return (new XplObject (s, "acos"));
}

XplObject*	matacosh (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "acosh: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "acosh: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (acosh);
#else
  s->tfunct (0, acosh);
#endif

  return (new XplObject (s, "acosh"));
}

XplObject*	matasin (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;

  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "asin: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "asin: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (x4asin);
#else
  s->tfunct (0, x4asin);
#endif

  return (new XplObject (s, "asin"));
}

XplObject*	matasinh (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;

  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "asinh: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "asinh: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (asinh);
#else
  s->tfunct (0, asinh);
#endif

  return (new XplObject (s, "asinh"));
}

XplObject*	matatan (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "atan: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "atan: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (atan);
#else
  s->tfunct (0, atan);
#endif

  return (new XplObject (s, "atan"));
}

XplObject*	matatanh (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "atanh: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "atanh: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (atanh);
#else
  s->tfunct (0, atanh);
#endif

  return (new XplObject (s, "atanh"));
}

XplObject*	matceil (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "ceil: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "ceil: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (ceil);
#else
  s->tfunct (0, ceil);
#endif

  return (new XplObject (s, "ceil"));
}

XplObject*	matcos (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "cos: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "cos: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (cos);
#else
  s->tfunct (0, cos);
#endif

  return (new XplObject (s, "cos"));
}

XplObject*	matcosh (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "cosh: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "cosh: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (x4cosh);
#else
  s->tfunct (0, x4cosh);
#endif

  return (new XplObject (s, "cosh"));
}

XplObject*	materf (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "erf: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "erf: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (erf);
#else
  s->tfunct (0, erf);
#endif

  return (new XplObject (s, "erf"));
}

XplObject*	materfc (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "erfc: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "erfc: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (erfc);
#else
  s->tfunct (0, erfc);
#endif

  return (new XplObject (s, "erfc"));
}

XplObject*	matexp (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "exp: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "exp: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (x4exp);
#else
  s->tfunct (0, x4exp);
#endif

  return (new XplObject (s, "exp"));
}

/*
XplObject*	matexp10 (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "exp10: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "exp10: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (exp10);
#else
  s->tfunct (0, exp10);
#endif

  return (new XplObject (s, "exp10"));
}
*/

XplObject*	matexpm1 (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;

  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "expm1: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "expm1: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (x4expm1);
#else
  s->tfunct (0, x4expm1);
#endif

  return (new XplObject (s, "expm1"));
}

XplObject*	matfloor (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;

  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "floor: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "floor: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (floor);
#else
  s->tfunct (0, floor);
#endif

  return (new XplObject (s, "floor"));
}

/*
XplObject*	matgamma (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "gamma: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "gamma: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (gamma);
#else
  s->tfunct (0, gamma);
#endif

  return (new XplObject (s, "gamma"));
}
*/

XplObject*	matlgamma (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "lgamma: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "lgamma: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (lgamma);
#else
  s->tfunct (0, lgamma);
#endif

  return (new XplObject (s, "lgamma"));
}

/*
XplObject*	matgammaf (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "gammaf: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "gammaf: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (gammaf);
#else
  s->tfunct (0, gammaf);
#endif

  return (new XplObject (s, "gammaf"));
}

XplObject*	matln (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "ln: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "ln: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (ln);
#else
  s->tfunct (0, ln);
#endif

  return (new XplObject (s, "ln"));
}
*/

XplObject*	matlog (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "log: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "log: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (x4log);
#else
  s->tfunct (0, x4log);
#endif

  return (new XplObject (s, "log"));
}

XplObject*	matlog10 (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "log10: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "log10: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (x4log10);
#else
  s->tfunct (0, x4log10);
#endif

  return (new XplObject (s, "log10"));
}

XplObject*	matlog1p (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "log1p: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "log1p: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (log1p);
#else
  s->tfunct (0, log1p);
#endif

  return (new XplObject (s, "log1p"));
}

/*
XplObject*	matlog2 (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "log2: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "log2: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (log2);
#else
  s->tfunct (0, log2);
#endif

  return (new XplObject (s, "log2"));
}
*/

XplObject*	matrint (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "rint: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "rint: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (rint);
#else
  s->tfunct (0, rint);
#endif

  return (new XplObject (s, "rint"));
}

XplObject*	matsin (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "sin: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "sin: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (sin);
#else
  s->tfunct (0, sin);
#endif

  return (new XplObject (s, "sin"));
}

XplObject*	matsinh (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "sinh: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "sinh: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (sinh);
#else
  s->tfunct (0, sinh);
#endif

  return (new XplObject (s, "sinh"));
}

XplObject*	matsqrt (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "sqrt: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "sqrt: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (sqrt);
#else
  s->tfunct (0, sqrt);
#endif

  return (new XplObject (s, "sqrt"));
}

XplObject*	mattan (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "tan: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "tan: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (tan);
#else
  s->tfunct (0, tan);
#endif

  return (new XplObject (s, "tan"));
}

XplObject*	mattanh (Liste<XplObject*> *argl)
{
  long length = argl->laenge ();
  XplDouble *s1, *s;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "tanh: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "tanh: first argument");
    return (NULL);
    }

  s = new XplDouble (s1);

#ifdef MATRIX
  s->exec (x4tanh);
#else
  s->tfunct (0, x4tanh);
#endif

  return (new XplObject (s, "tanh"));
}

#define TTKNAN sqrt(-1.0)

XplObject*	qfn (Liste<XplObject*> *argl)
// Formel: Hartung
{
  long i, m, n, length = argl->laenge ();
  XplDouble *s1, *res;

  const double c0 = 2.515517, c1 = 0.802853, c2 = 0.010328,
               d1 = 1.432788, d2 = 0.189269, d3 = 0.001308;

  double t, f;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "qfn: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "qfn: first argument");
    return (NULL);
    }

#ifdef MATRIX
  n = s1->rows() * s1->cols();
#else
  n = 1L;
  m = s1->dimension();
  for (i=0; i<m; i++)
    n *= s1->dimension (i);
#endif

  res = new XplDouble (s1);

  for (i=0; i<n; i++)    
    {
    t = (*s1)[i];
    if ((t <= 0.0) || (t >= 1.0))
      {
      (*res)[i] = TTKNAN;
      continue;
      }
    if (t < 0.5)
      {
      t = 1.0 - t;
      f = -1.0;
      }
    else
      f = 1.0;
    t = sqrt(-2.0*x4log(1-t));
    (*res)[i] = f*(t - ((c2*t+c1)*t+c0) / (((d3*t+d2)*t+d1)*t+1));
    }

  return (new XplObject (res, "qfn"));
}

XplObject*	pdfn (Liste<XplObject*> *argl)
// Formel: Graf, Henning, Stange, Wilrich: Formel und Tabellen ...
{
  long i, m, n, length = argl->laenge ();
  XplDouble *s1, *res;

  const double factor = 1.0 / sqrt(2.0*M_PI);

  double u;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "pdfn: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "pdfn: first argument");
    return (NULL);
    }

#ifdef MATRIX
  n = s1->rows() * s1->cols();
#else
  n = 1L;
  m = s1->dimension();
  for (i=0; i<m; i++)
    n *= s1->dimension (i);
#endif

  res = new XplDouble (s1);

  for (i=0; i<n; i++)    
    {
    u = (*s1)[i];
    (*res)[i] = factor*x4exp(-0.5*u*u);
    }

  return (new XplObject (res, "pdfn"));
}

XplObject*	cdfn (Liste<XplObject*> *argl)
// Formel: Graf, Henning, Stange, Wilrich: Formel und Tabellen ...
{
  long i, m, n, length = argl->laenge ();
  XplDouble *s1, *res;

  const double d1 = 0.0498673470, d2 = 0.0211410061, d3 = 0.0032776263,
               d4 = 0.0000380036, d5 = 0.0000488906, d6 = 0.0000053830;

  double u, x;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "cdfn: Must be 1");
    return (NULL);
    }

  s1 = (*argl)[0]->DefaultXplDouble ();

  if (s1 == NULL)
    {
    globalError.Set (ERR_NUM, "cdfn: first argument");
    return (NULL);
    }

#ifdef MATRIX
  n = s1->rows() * s1->cols();
#else
  n = 1L;
  m = s1->dimension();
  for (i=0; i<m; i++)
    n *= s1->dimension (i);
#endif

  res = new XplDouble (s1);

  for (i=0; i<n; i++)    
    {
    x = (*s1)[i];
    u = fabs (x);
    u = (((((d6*u + d5)*u + d4)*u + d3)*u + d2)*u + d1)*u + 1;
    u = u*u;
    u = u*u;
    u = u*u;
    u = u*u;
    (*res)[i] = (x > 0) ? 1 - 0.5/u : 0.5/u;
    }

  return (new XplObject (res, "cdfn"));
}

