/* $Header: /home/xplore/cvs/xpl4gr/xplcalc.C,v 1.3 2002/08/16 12:31:57 sigbert Exp $ */

#include "machine.h"

#ifdef HAS_IOSTREAMH
#include <iostream.h>
#else
#include <stream.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "liste.h"
#include "xplcalc.h"
#include "xplpp.h"
#include "init.h"
// #include "macro.h"
#include "ttkio.h"

void error(char*  str)
{
    ttkerror(str);
//    exit ( 1 ) ;
}


inline double minus (double x)
{
  return (-x);
}

inline double lognot (double x)
{
  return (x == 0.0);
}

XplObject*	operate (XplObject *lo, XplObject *ro, int op)
{
  if (globalError.Get())
    {
    if (lo != NULL)
      if (! lo->WriteProtection ())
	delete lo;
    if (ro != NULL)
      if (! ro->WriteProtection ())
	delete ro;
    return (NULL);
    }

  if ((lo == NULL) || (ro == NULL))
    {
    cerr << "Operate:  argument is NULL" << endl;
    return (NULL);
    }

  if (lo->DefaultType() == C_EMPTY)
    {
    globalError.Set (ERR_EMPTY, lo->Name());
    return (NULL);
    }

  if (ro->DefaultType() == C_EMPTY)
    {
    globalError.Set (ERR_EMPTY, ro->Name());
    return (NULL);
    }

#ifdef _XPLCDEF_
  if ((lo->DefaultType() == C_XPLCHAR) && (ro->DefaultType() == C_XPLCHAR))
    {
    XplChar *rs, *ls, *tmp = NULL;
    XplDouble *t = NULL;

    ls = lo->DefaultComponent()->Content().xplchar;
    rs = ro->DefaultComponent()->Content().xplchar;

    if ((op == CBIND) || (op == RBIND) || (op == PLUS))
      tmp = new XplChar (1);
    else
      t = new XplDouble(1);

    switch (op) {
      case PLUS:	        *tmp = ls->concat (*rs);   break;
      case CBIND:		*tmp = ls->bind (*rs, 1);  break;
      case RBIND:		*tmp = ls->bind (*rs, 0);  break;
      case EQUAL:		*t   = cast (*ls == *rs);  break;
      case NEQ:                 *t   = cast (*ls != *rs);  break;
      case LESS:		*t   = cast (*ls <  *rs);  break;
      case GREATER:	        *t   = cast (*ls >  *rs);  break;
      case LEQ:		        *t   = cast (*ls <= *rs);  break;
      case GEQ:		        *t   = cast (*ls >= *rs);  break;
      default:	cerr << "Error in operate with XPLCHAR: Unknown Operation " << endl;
    		return (NULL);
      }

    if (! lo->WriteProtection ())       // delete the objects, if they
      delete lo;                        // are temporary
    if (! ro->WriteProtection ())
      delete ro;

    if (globalError.Get())              // any error in operator
      {
      if (tmp != NULL)
	delete tmp;
      else
	delete t;
      return (NULL);
      }

    if (tmp == NULL)
      return (new XplObject (new Component (t)));
    else
      return (new XplObject (new Component (tmp)));
    }
#endif  // of _XPLCDEF_


  XplDouble *l, *r, *t = new XplDouble(1);
  static XplDouble h(1);                 // stays constant

  h(0) = 1.0; 

// it is not necessary to generate copies of write protected objects,
// as no operation will change their operands. Therefore following
// code was replaced:
/*
if (lo->WriteProtection())
    l = new XplDouble (lo->DefaultXplDouble());
  else
    l = lo->DefaultXplDouble();
*/

  r = ro->DefaultXplDouble();
  l = lo->DefaultXplDouble();

  if (l == NULL)
    {
    cerr << "Operate:  left argument type XPLDOUBLE = NULL" << endl;
    globalError.Set (ERR_EMPTY, lo->Name());
    return (NULL);
    }

  if (r == NULL)
    {
    cerr << "Operate:  right argument type XPLDOUBLE = NULL" << endl;
    globalError.Set (ERR_EMPTY, ro->Name());
    return (NULL);
    }

  
  switch (op) {
    case PLUS:		*t = *l + *r;  break;
    case MINUS:		*t = *l - *r;  break;
    case MULT:		*t = *l * *r;  break;
    case DIV:		*t = *l / *r;  break;
    case MOD:		*t = *l % *r;  break;

    case POW:		delete t;  t = power (*l, *r);
			if (t == NULL)
			  return (NULL);
			break;

#ifdef MATRIX
    case MMULT:		*t = *l;   t->mult (*r);    break;
    case TMULT:		if (l->isScalar () || r->isScalar ())
                          *t = *l * *r;
			else
			  {
//			  *t = *(l->transpose());  t->mult (*r);
			  *t = *l;
                          t->tmult (*r);
			  }
                        break;
    case CBIND:		delete t;  t = ColumnBind (*l, *r);  break;
    case RBIND:		delete t;  t = RowBind    (*l, *r);  break;
    case EQUAL:		*t = *l == *r;  break;
    case NEQ:		*t = *l != *r;  break;
    case LESS:		*t = *l < *r;   break;
    case GREATER:	*t = *l > *r;   break;
    case LEQ:		*t = *l <= *r;  break;
    case GEQ:		*t = *l >= *r;  break;
    case AND:		*t = *l && *r;  break;
    case OR:		*t = *l || *r;  break;
    case SEQ:		if (l->isScalar() && r->isScalar())
    			  *t = XplDouble((*l)(0), (*r)(0), 1.0);
			else
			  {
			  delete t;
			  return (NULL);
			  }
			break;
#else
    case EQUAL:		*t = cast (*l == *r);  break;
    case NEQ:		*t = cast (*l != *r);  break;
    case LESS:		*t = cast (*l < *r);   break;
    case GREATER:	*t = cast (*l > *r);   break;
    case LEQ:		*t = cast (*l <= *r);  break;
    case GEQ:		*t = cast (*l >= *r);  break;
//    case AND:		*t = cast ((*l->fabs() * *r->fabs()) >= h);  break;
    case AND:		*t = cast ((*l * *r) >= h);  break;
//    case OR:		*t = cast ((*l->fabs() + *r->fabs()) >= h);  break;
    case OR:		*t = cast ((*l + *r) >= h);  break;
    case MMULT:		if ((l->dimension () == 0) || (r->dimension () == 0))
                          *t = *l * *r;
			else
			  *t = l->matmult (*r);  break;
    case TMULT:		if ((l->dimension () == 0) || (r->dimension () == 0))
    						{ // PC 20010418
                       *t = *l * *r;
                       if (l->dimension () != 0)
                       {
                         *t =  t->trn(*t);
                       }
                     }
                          // original version: *t = *l * *r;
			else
			  *t = l->tmatmult (*r);  break;
    case CBIND:		*t = l->bind (*r, 1);  break;
    case RBIND:		*t = l->bind (*r, 0);  break;
    case SEQ:		long a, b, n, m, i;
			a = l->dimension (0);
                        b = r->dimension (0);
			if ((a == 1) && (b == 1))
			  {
			  a = (long) (*l)(0);
			  b = (long) (*r)(0);
			  n = labs(b-a) + 1;
			  m = b < a ? -1 : 1;
    			  *t = XplDouble (n);
			  for (i=0; i<n; i++)
			    (*t)(i) = a + i*m;
			  }
			else
			  {
			  delete t;  return (NULL);
			  }
			break;
#endif

    default:	cerr << "Error in operate: Unknown Operation " << endl;
//                xplerror(op);
    		return (NULL);
    }

  if (! lo->WriteProtection ())       // delete the objects, if they
    delete lo;                        // are temporary
  if (! ro->WriteProtection ())
    delete ro;

  if (globalError.Get())              // any error in operator
    {
    delete t;
    return (NULL);
    }

  return (new XplObject (t));         // t must not be deleted !!
}

XplObject*	operate (XplObject *lo, int op)
{
  if (lo == NULL)
    {
    globalError.Set (ERR_EMPTY, "(no object)");
    return (NULL);
    }

  if (lo->DefaultType() == C_EMPTY)
    {
    globalError.Set (ERR_EMPTY, lo->Name());
    return (NULL);
    }

#ifdef _XPLCDEF_
  if (lo->DefaultType() == C_XPLCHAR)
    {
    XplChar *ls, *tmp = NULL;

    ls = lo->DefaultComponent()->Content().xplchar;

    switch (op) {
      case TRANS: tmp = new XplChar(1);  // SK 15.10.97
   	          tmp->trn (*ls); 
                  break;
      default:	cerr << "Error in operate with XPLCHAR: Unknown Operation " << endl;
    		return (NULL);
      }
    return (new XplObject (new Component (tmp)));
    }
#endif  // of _XPLCDEF_

  if (lo->DefaultType() != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, lo->Name());
    return (NULL);
    }

  XplDouble *l, *t = new XplDouble(1);

  if (lo->WriteProtection())
    l = new XplDouble (lo->DefaultXplDouble());
  else
    l = lo->DefaultXplDouble();

  switch (op) {
#ifdef MATRIX
    case TRANS:		*t = *l; t->transpose ();   break;
    case MINUS:		*t = *l; t->exec (minus);   break;
    case NOT:		*t = *l; t->exec (lognot);  break;
#else
    case TRANS:		*t =  l->trn (*l);               break;
    case MINUS:		*t = *l; t->tfunct (0, minus);   break;
    case NOT:		*t = *l; t->tfunct (0, lognot);  break;
#endif

    default:	cerr << "Fatal Error in operate: Unknown Operation " << op  << endl;
    		exit (1);
    }

    if (! lo->WriteProtection ())
      delete lo;
    else
      delete l;
 
    return (new XplObject (t));         // t must not be deleted !!
}

void            assign  (Liste<const char*> *as, XplObject *o)
{
  long      i, length = as->laenge ();
  XplObject *obj;

  if (length > o->ComponentNumber())
    {
    globalError.Set (ERR_ASSIGN, "object has less components than requested");
    return;
    }

  if (length == 1)        // list consists only of one element
    {
    obj = GetXplObject (objList, (*as)[0]);
    if (obj == NULL)
      {
      obj = new XplObject ((*as)[0]);
      objList->haengeAn (obj);
      }
      obj->TransferComponents (o);

      if (o->Type() & O_DISPLAY)
	cerr << " *** TTK: display in assign!! CHECK!" << endl;
    }

else
  for (i=0; i<length; i++)
    {
    obj = GetXplObject (objList, (*as)[i]);
    if (obj == NULL)
      {
      obj = new XplObject ((*as)[i]);
      objList->haengeAn (obj);
      }
      obj->TransferComponents ((*o)[i]->Content().xplobject);

      if ((*o)[i]->Content().xplobject->Type() == O_DISPLAY)
	cerr << " *** TTK: display in assign!! CHECK!" << endl;
    }
  obj->SetDefaultXplDouble();            // very necessary!
}

XplDouble * power (XplDouble l, XplDouble r)
{
  long   i, n, n1, m, m1;
  double p;

#ifdef MATRIX          // first case, same size
  if ((r.rows() == l.rows()) && (r.cols() == l.cols()))
#else
  if (r.dimension() == l.dimension())
#endif
    {
/* #ifndef MATRIX
  if (l.dimension() > 2)
    {
    globalError.Set (ERR_INTERN, "power: only for matrices");
    return (NULL);
    }
#endif
*/

#ifdef MATRIX
    n = l.rows() * l.cols();
#else
    n = n1 = 1L;
    m1 = r.dimension();
    m = l.dimension();
    for (i=0; i<m; i++)
      {
      n  *= l.dimension (i);
      n1 *= r.dimension (i);
      }
#endif

    if (n == n1)
      {
      XplDouble *res = new XplDouble(l);

      for (i=0; i<n; i++)    
	(*res)[i] = pow (l[i], r[i]);

      return (res);
      }
    }    

#ifdef MATRIX          // first case, right argument is scalar
  if (r.isScalar())
#else
  if (r.dimension() == 0)
#endif
    {
/* #ifndef MATRIX
  if (l.dimension() > 2)
    {
    globalError.Set (ERR_INTERN, "power: only for matrices");
    return (NULL);
    }
#endif
*/

    p = r(0);

#ifdef MATRIX
    n = l.rows() * l.cols();
#else
    n = 1L;
    m = l.dimension();
    for (i=0; i<m; i++)
      n *= l.dimension (i);
#endif

    XplDouble *res = new XplDouble(l);

    for (i=0; i<n; i++)    
      (*res)[i] = pow (l[i], p);

    return (res);
    }

#ifdef MATRIX          // second case, left argument is scalar
  if (l.isScalar())
#else
  if (l.dimension() == 0)
#endif
    {
/* #ifndef MATRIX
  if (l.dimension() > 2)
    {
    globalError.Set (ERR_INTERN, "power: only for matrices");
    return (NULL);
    }
#endif
*/

    p = l(0);

#ifdef MATRIX
    n = r.rows() * r.cols();
#else
    n = 1L;
    m = r.dimension();
    for (i=0; i<m; i++)
      n *= r.dimension (i);
#endif

    XplDouble *res = new XplDouble(r);

    for (i=0; i<n; i++)    
      (*res)[i] = pow (p, r[i]);

    return (res);
    }

  globalError.Set (ERR_INTERN, "power: at least one arg has to be scalar");
  return (NULL);
}
