/* $Header: /home/xplore/cvs/xpl4gr/xpldllfun.C,v 1.3 2002/08/16 12:31:57 sigbert Exp $ */

// XploRe dynamic load functions
// see also xplfun.C and xplbasefun.C
// contains dynamic load interface (ELF)


//  TTK, 23.10.06

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <unistd.h>
#include <math.h>
#include <string.h>
#include <dlfcn.h>
#include <ctype.h>

#include "object.h"
#include "xpldllfun.h"
#include "xplerror.h"
#include "init.h"
#include "directory.h" 

#define TEXTLENGTH  1024L

#ifdef __BORLANDC__
#include "xplwin.h"
#endif


// #define _SHARED_MEM_  1

int     maxDLL;
DLLctrl *DLLlib;

XplObject* DLLopen (Liste<XplObject*> *argl)
// expects a file name
{
  long       i;
  void       *DLLhandle = NULL;
  int        mode = RTLD_LAZY;
  const char *t;
  XplDouble  *res;
  XplObject  *o;
  char       *fdir, loaddll[TEXTLENGTH];

  static long counter = 1;

  if (argl->laenge () != 1)
    {
    globalError.Set(ERR_ARGS, "DLLopen:  Must be 1");
    return (NULL);
    }

  o = (*argl)[0];
  if (o->DefaultType() != C_TEXT)
    {
    globalError.Set(ERR_TEXT, "DLLopen:  first argument");
    return (NULL);
    }

#ifdef _XPLCDEF_
  t = (*o->DefaultComponent()->Content().xplchar)(0).tocstr();
#else
  t = o->DefaultComponent()->Content().text;
#endif

  for (i=0; i<maxDLL; i++)           // equal names ?
    if (DLLlib[i].DLLid != 0)
      if (strcmp (DLLlib[i].name, t) == 0)
	globalWarning.Set (WARN_DLLOPEN, t);

  for (i=0; i<maxDLL; i++)           // search empty position
    if (DLLlib[i].DLLid == 0)
      break;

  if (i >= maxDLL)
    {
    globalError.Set (ERR_DLLOPEN, "too many libraries opened. Please close some.");
    return (NULL);
    }
/*
  fdir = FindDir (t, xpl4dll);      // search in global directory
  if (fdir == NULL)
    fdir = FindDir (t, ".");        // search in current dir
  if (fdir == NULL)
    fdir = FindDir (t, "");         // search absolut path name
SK 23.06.98 */


  fdir = openfile (DLL, t, "r");
  if (fdir == NULL)
    {
    globalError.Set (ERR_DLLOPEN, t);
    return (NULL);
    }

  sprintf (loaddll, "%s", fdir);

//  sprintf (loaddll, "/home/klinke/xpl4gr/dll/%s", t);
  
  DLLhandle = dlopen (loaddll, mode);
  if (DLLhandle == NULL)
    {
    globalError.Set (ERR_DLLOPEN, dlerror());
    return (NULL);
    }
  DLLlib[i].name      = strdup (t);
  DLLlib[i].DLLid     = counter;
  DLLlib[i].DLLhandle = DLLhandle;
  //  DLLlib[i].DLLhandle = dlopen (t, mode);

  res = new XplDouble (1);
  (*res)(0) = counter++;

  return (new XplObject (res, "dlopen"));
}

XplObject* DLLclose (Liste<XplObject*> *argl)
// expects nothing. DLLlist will be handled in the future
{
  XplObject  *o;
  XplDouble  *x = NULL;
  long       id, i = -1, length = argl->laenge ();
  const char *t;

  if (length > 1)
    {
    globalError.Set(ERR_ARGS, "DLLclose:  expect no or one argument");
    return (NULL);
    }

  if (length == 1)
    {
    o = (*argl)[0];
    if (o->DefaultType () == C_TEXT)
      {
#ifdef _XPLCDEF_
      t = (*o->DefaultComponent()->Content().xplchar)(0).tocstr();
#else
      t = o->DefaultComponent()->Content().text;
#endif

      for (i=0; i<maxDLL; i++)           // equal names ?
	if (DLLlib[i].DLLid != 0)
	  if (strcmp (DLLlib[i].name, t) == 0)
	    {
	    id = DLLlib[i].DLLid;
	    break;
	    }
      }
    if (o->DefaultType () == C_XPLDOUBLE)
      {
      x = o->DefaultXplDouble ();
      id = (long) (*x)(0);
      if (id <= 0)
	{
        globalWarning.Set (WARN_DLLCLOSE, "invalid library handle");
	return (NULL);
        }
      }
    if (id == -1)         // id not changed
      {
      globalError.Set(ERR_TYPE, "DLLclose:  first argument");
      return (NULL);
      }

    for (i=0; i<maxDLL; i++)
      if (DLLlib[i].DLLid == id)
	{
	delete [] DLLlib[i].name;
	DLLlib[i].name  = NULL;
	DLLlib[i].DLLid = 0;
	if (dlclose (DLLlib[i].DLLhandle) != 0)
	  globalError.Set (ERR_DLLCLOSE, dlerror());
	DLLlib[i].DLLhandle = NULL;
	break;
        }
    if (i >= maxDLL)         // lib not found
      {
      globalWarning.Set (WARN_DLLCLOSE);
      return (NULL);
      }

    }
  else                              // close all
    for (i=0; i<maxDLL; i++)
      if (DLLlib[i].DLLid != 0)
	{
	delete [] DLLlib[i].name;
	DLLlib[i].name  = NULL;
	DLLlib[i].DLLid = 0;
	if (dlclose (DLLlib[i].DLLhandle) != 0)
	  globalError.Set (ERR_DLLCLOSE, dlerror());
	DLLlib[i].DLLhandle = NULL;
        }

  return (NULL);
}

XplObject* DLLcall (Liste<XplObject*> *argl)
// expects a file name
{
  long       ioffset, i, j, n, id, size, lres=-4165, length = argl->laenge ();
  const char *t;
  XplObject  *o;
  XplDouble  *res, *x;
  void       *DLLsym = NULL;
  void       *DLLhandle = NULL;

  if (length < 1)
    {
    globalError.Set(ERR_ARGS, "DLLcall:  Must be >= 1");
    return (NULL);
    }

  o = (*argl)[0];
  x = o->DefaultXplDouble ();
  if (x != NULL)
    {
    id = (long) (*x)(0);
    for (i=0; i<maxDLL; i++)
      if (DLLlib[i].DLLid == id)
	if (NULL == (DLLhandle = DLLlib[i].DLLhandle))
	  {
	  globalError.Set(ERR_DLLCALL, "no library with this handler");
	  return (NULL);
	  }
    if (DLLhandle == 0)
      {
      globalError.Set(ERR_DLLCALL, "no library with this handler");
      return (NULL);
      }
    o = (*argl)[1];
    }

  if (o->DefaultType() != C_TEXT)
    {
    globalError.Set(ERR_TEXT, "DLLcall:  first argument");
    return (NULL);
    }

#ifdef _XPLCDEF_
  t = (*o->DefaultComponent()->Content().xplchar)(0).tocstr();
#else
  t = o->DefaultComponent()->Content().text;
#endif

  if (DLLhandle == NULL)
    {
    for (i=0; i<maxDLL; i++)
      if (NULL != (DLLhandle = DLLlib[i].DLLhandle))
	if (NULL != (DLLsym = dlsym (DLLhandle, t)))
	  break;
    }
  else
    DLLsym = dlsym (DLLhandle, t);

  if (DLLsym == NULL)
    {
    globalError.Set (ERR_DLLCALL, t);
    return (NULL);
    }

  if (x == NULL)
    {
    n = length == 1 ? 1 : length-1;
    ioffset = 1;
    }
  else
    {
    n = length == 2 ? 1 : length-2;
    ioffset = 2;
    }

  double    *r, *c[n];
  int       shmid;
  void      *shm;

  for (i=0; i<length-ioffset; i++)       // ioffset depends on first arg
    {
    x = (*argl)[i+ioffset]->DefaultXplDouble();
    if (x == NULL)
      {
      char text[TEXTLENGTH];
      sprintf (text, "DLLcall: %ld argument", i+ioffset);
      globalError.Set (ERR_NUM, text);
      return (NULL);
      }
    size = x->size();
    
#ifdef _SHARED_MEM_
    shmid = shmget (IPC_PRIVATE, size*sizeof(double), 0xffff);
    if (shmid == -1)     // ok, we have no shared memory
      {
      cerr << "Can not get shared memory" << endl;
      cerr << "leaving dlcall" << endl;
      cerr << "i = " << i << endl;
      return (NULL);
      }
/*
    else
      {
      shm = shmat (shmid, NULL, 0xffff);
      if ( (int) shm == -1)
	{
	cerr << "Can not attach shared memory" << endl;
	cerr << "leaving dlcall" << endl;
	cerr << "i = " << i << endl;
	return (NULL);
        }
*/
    c[i] = (double *) shmat (shmid, NULL, 0x666);  // r/w permission user
    if (c[i] == NULL  || (long) c[i] == -1)
      {
      cerr << "dlcall error: " << i << '\t' << size;
      exit (1);
      }
#else
    c[i] = new double[size];
#endif
    r = c[i];
    for (j=0; j<size; j++)          // copy input arguments
      *r++ = (*x)[j];
    }

fun0 fptr0;
fun1 fptr1;
fun2 fptr2;
fun3 fptr3;
fun4 fptr4;
fun5 fptr5;
fun6 fptr6;
fun7 fptr7;
fun8 fptr8;
fun9 fptr9;
fun10 fptr10;
fun11 fptr11;
fun12 fptr12;

#ifdef _SHARED_MEM_
  if ((i = fork()) == 0)   // the son
    {
#endif
   switch (length-ioffset)
    {
    case 0:   fptr0 = (fun0) DLLsym;
	      lres = (*fptr0) ();  break;
    case 1:   fptr1 = (fun1) DLLsym;
	      lres = (*fptr1) (c[0]);  break;
    case 2:   fptr2 = (fun2) DLLsym;
	      lres = (*fptr2) (c[0], c[1]);  break;
    case 3:   fptr3 = (fun3) DLLsym;
	      lres = (*fptr3) (c[0], c[1], c[2]);  break;
    case 4:   fptr4 = (fun4) DLLsym;
	      lres = (*fptr4) (c[0], c[1], c[2], c[3]);  break;
    case 5:   fptr5 = (fun5) DLLsym;
	      lres = (*fptr5) (c[0], c[1], c[2], c[3], c[4]);  break;
    case 6:   fptr6 = (fun6) DLLsym;
	      lres = (*fptr6) (c[0], c[1], c[2], c[3], c[4], c[5]);  break;
    case 7:   fptr7 = (fun7) DLLsym;
	      lres = (*fptr7) (c[0], c[1], c[2], c[3], c[4], c[5], c[6]);  break;
    case 8:   fptr8 = (fun8) DLLsym;
	      lres = (*fptr8) (c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]);  break;
    case 9:   fptr9 = (fun9) DLLsym;
	      lres = (*fptr9) (c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7], c[8]);  break;
    case 10:   fptr10 = (fun10) DLLsym;
	      lres = (*fptr10) (c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7], c[8], c[9]);  break;
    case 11:   fptr11 = (fun11) DLLsym;
	      lres = (*fptr11) (c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7], c[8], c[9], c[10]);  break;
    case 12:   fptr12 = (fun12) DLLsym;
	      lres = (*fptr12) (c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7], c[8], c[9], c[10], c[12]);  break;
    default:  globalError.Set (ERR_INTERN, "too many args for dlcall");
               lres = 0;
    }
#ifdef _SHARED_MEM_
   exit (0);
   }
  else
   wait (NULL);   // the father
#endif
  
  for (i=0; i<length-ioffset; i++)
    {
    x = (*argl)[i+ioffset]->DefaultXplDouble();
    if (x == NULL)
      {
      char text[TEXTLENGTH];
      sprintf (text, "DLLcall: %ld argument", i+ioffset);
      globalError.Set (ERR_NUM, text);
      return (NULL);
      }
    r = c[i];
    size = x->size();
    for (j=0; j<size; j++)          // copy back results
      (*x)[j] = *r++;
#ifdef _SHARED_MEM_
    shmdt ((void *) c[i]);
#else
    delete c[i];
#endif
    }

  res = new XplDouble (1);
  (*res)[0] = (double) lres;

  return (new XplObject (res, "dlcall"));
}


XplObject* DLLquery (Liste<XplObject*> *argl)
// expects a file name
{
  long       i, j = 0, countDLL = 0;
  XplDouble  *r1, *r2;
  XplObject  *o;
#ifdef _XPLCDEF_
  XplChar    *xc;
#endif

  if (argl->laenge () != 0)
    {
    globalError.Set(ERR_ARGS, "DLLquery:  No argument expected");
    return (NULL);
    }

  for (i=0; i<maxDLL; i++)           // count libs
    if (DLLlib[i].DLLid != 0)
      countDLL++;

  r1 = new XplDouble (1);
  (*r1)(0) = (double) countDLL;
  if (countDLL == 0)
    return (new XplObject (r1, "dlquery"));

  r2 = new XplDouble (countDLL);
#ifdef _XPLCDEF_
  xc = new XplChar   (countDLL);
#endif

  for (i=0; i<maxDLL; i++)           // count libs
    if (DLLlib[i].DLLid != 0)
      {
#ifdef _XPLCDEF_
      (*xc)(j) = xplstring (DLLlib[i].name);
#endif
      (*r2)(j++) = (double) DLLlib[i].DLLid;
      }

  o = new XplObject ("dlquery");
  o->AppendComponent (new Component (new XplObject (r1, "n")));
  o->AppendComponent (new Component (new XplObject (r2, "handle")));
#ifdef _XPLCDEF_
  o->AppendComponent (new Component (new XplObject (xc, "name")));
#endif

  return (o);
}

