/* $Header: /home/xplore/cvs/xpl4gr/xplerror.C,v 1.3 2002/08/16 12:31:58 sigbert Exp $ */

// XploRe 4 Error Class Implementation

// TTK, 4.7.95

// #pragma implementation

#ifdef __BORLANDC__
#include <windows.h>
#endif

#include "machine.h"

#ifdef HAS_IOSTREAMH
#include <iostream.h>
#else
#include <stream.h>
#endif

#include <stdlib.h>
#include <string.h>

#include "xplerror.h"
#include "globlist.h"
#include "liste.h"
#include "xjprotocol.h"

#ifndef MATRIX
#include "xplio.h"
#endif

#ifdef __BORLANDC__
#include "xplwin.h"
#endif

// localized strings - PC 20010320
#include "xplore_strings.h"

XplGlobalError    globalError (0);
XplGlobalWarning  globalWarning (0);

static int warnCounter = 0L;
//static int errCounter  = 0L;

const  int maxWarnings = 5;
const  int maxErrors   = 10;

XplError::XplError (long n)
{
  number    = n;
  text      = NULL;
}

XplError::XplError (long n, const char *t)
{
  number = n;
  text   = strdup (t);

  if (text == NULL)
    {
    cerr << "Error:  XplError: Not enough memory ..." << endl;
    exit (1);
    }
}

XplError::~XplError ()
{
  if (text != NULL)
    delete[] text;
}

void XplError::AppendText (const char *t)
{
  long n1, n2;
  char *t1;

//  if (errCounter++ > maxErrors)
//    return;

  n1 = (text == NULL) ? 0 : strlen (text);
  n2 = strlen (t);

  t1  = new char[n1+n2+2];    // for \0 and " "
  *t1 = '\0';

  if (t1 == NULL)
    {
    cerr << "Error:  XplError: Not enough memory ..." << endl;
    exit (1);
    }

  if (text != NULL)
    {
    strcpy (t1, text);
    strcat (t1, "\n");
    delete [] text;
    }

  strcat (t1, t);

  text = t1;
}

void XplError::AppendText (long n, const char *t)
{
  if (n == number)        // accept new text only for the right error number
    AppendText (t);
}

void XplGlobalError::Clear (void)
{
  number   = 0;
  happened = FALSE;
  if (text != NULL)
    delete[] text;
}

void XplGlobalError::Set (long n)
{
  if (happened)
    return;
  number = n;
  happened = TRUE;
}

void XplGlobalError::Set (long n, const char *t)
{
extern const char *CurrentMacro;    // defined in init.C
extern GlobalList parselist;

  char buffer[TEXTLENGTH];

  if (happened)
    return;
  number = n;
//  sprintf (buffer, "%s %s %ld", CurrentMacro, t, parselist.Pos());
  sprintf (buffer, XMSG_1GLOBAL_ERROR_LINE_STR, CurrentMacro, parselist.CurrentLine(), t);
  AppendText (buffer);
  happened = TRUE;
}

void XplGlobalError::Set (long n, const char *t, const char *cm, long line)
{
// sets full information for error

  char buffer[TEXTLENGTH];

  if (happened)
    return;
  number = n;
  sprintf (buffer, XMSG_2GLOBAL_ERROR_LINE_STR, cm, line, t);
  AppendText (buffer);
  happened = TRUE;
}

void XplGlobalError::ShowError ()
{
  if (happened == FALSE)
    return;

#ifdef INETD
  cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif

#if !defined(SOCKETSERVER) && !defined(X4singleserver)
#ifndef __BORLANDC__   // SWS
  cerr << "XplGlobalError\n";
  cerr << text  << endl;
#endif
#endif

  XplError *err = GetXplError ();

  long tx = 0L;

  if (text != NULL)
    tx = strlen (text);
  if (err->GetText () != NULL)
    tx += 1+strlen (err->GetText ());   // for extra " "

  char *outtext = new char[tx+1];
  *outtext = '\0';                  // this is necessary !!

  if (err->GetText () != NULL)
    {
    strcat (outtext, err->GetText());
    strcat (outtext, " ");
    }
  if (text != NULL)
    strcat (outtext, text);

#if defined(SOCKETSERVER) || defined(X4singleserver)
  protocol->sendString (outtext);
  //  sockout (outtext);
  //  sockout ('\n');
#else
#ifndef __BORLANDC__   // SK 14.11.97
#ifndef _UNIX_
  cerr << err->GetText () << endl;    // nothing to stderr in server mode
  cerr << "XplGlobalError\n";
  cerr << text  << endl;
#else
 xplerror(outtext);
#endif
#else
  MessageBox (NULL, outtext, XMSG_XPLORE_ERROR_STR, MB_ICONERROR | MB_OK | MB_TOPMOST | MB_TASKMODAL);
#endif

#endif

  delete [] outtext;
  delete [] text;

  text     = NULL;      // necessary (see AppendText)
  number   = 0;
  happened = FALSE;
}

XplError* XplGlobalError::GetXplError ()  const
{
  int i, n = errList.laenge ();

  for (i=0; i<n; i++)
    if (errList[i]->GetNumber () == number)
      return (errList[i]);

#ifdef INETD
cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif

#if !defined(SOCKETSERVER) && !defined(X4singleserver)
  cerr << "Error:  GetXplError requires a non existing error ..." << endl;
  exit (1);
#endif
  return (NULL);        // necessary for some compilers
}

void XplGlobalError::append      (XplError *x)
  { errList.haengeAn (x); }


//  end of XplGlobalError

void XplGlobalWarning::Set (long n, const char *t)
{
extern const char *CurrentMacro;    // defined in init.C
extern GlobalList parselist;

  if (warnCounter++ > maxWarnings)
    return;

  char buffer[TEXTLENGTH];

  number   = n;
  happened = TRUE;

  XplError *err = GetXplError ();

  if (err->GetText () != NULL)
    AppendText (err->GetText ());

  if (t != NULL)
    sprintf (buffer, XMSG_3GLOBAL_ERROR_LINE_STR, CurrentMacro, parselist.CurrentLine(), t);
else
    sprintf (buffer, XMSG_4GLOBAL_ERROR_LINE_STR, CurrentMacro, parselist.CurrentLine());

  AppendText (buffer);
}

void XplGlobalWarning::Set (long n, const char *t, const char *cm, long line)
{
// sets full information for error

  if (warnCounter++ > maxWarnings)
    return;
  char buffer[TEXTLENGTH];

  number   = n;
  happened = TRUE;

  XplError *err = GetXplError ();

  if (err->GetText () != NULL)
    AppendText (err->GetText ());

  sprintf (buffer, XMSG_5GLOBAL_WARNING_LINE_STR, cm, line, t);
  AppendText (buffer);
}

void XplGlobalWarning::ShowError ()
{
  if (happened == FALSE)
    return;

  warnCounter = 0;

#ifdef INETD
  cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif

#if !defined(SOCKETSERVER) && !defined(X4singleserver)
#ifndef __BORLANDC__  // SK 09.04.99
  cerr << "XplGlobalWarning\n";
  cerr << text  << endl;
#endif
#endif

#ifdef _GRAPHICS_
  ttkwarning (text);
#else
#ifndef __BORLANDC__  // Sk 09.04.99
  cerr << text << endl;
#else
  MessageBox (NULL, text, XMSG_XPLORE_WARNING_STR, MB_ICONERROR | MB_OK | MB_SETFOREGROUND);
#endif
#endif
#if defined(SOCKETSERVER) || defined(SOCKETSERVER)
  protocol->sendString (text);
  //  sockout (text);
  //  sockout ('\n');
#endif

  delete [] text;

  text     = NULL;      // necessary (see AppendText)
  number   = 0;
  happened = FALSE;

}

//  end of XplGlobalWarning

void InitializeErrorList ()
{
  static int alreadyInvoked = FALSE;

  if (alreadyInvoked)
    {
#ifdef INETD
cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif
    cerr << "ERROR:  InitializeErrorList already called. Ignored!" << endl;
    return;
    }

// firstly the warnings

  globalWarning.append (new XplError (WARN_COMMON,        XMSG_WARNING_STR));
  globalWarning.append (new XplError (WARN_NOCONVERGENCE, XMSG_NO_CONVERGENCE_STR));
  globalWarning.append (new XplError (WARN_MACRO,         XMSG_WARNING_IN_QUANTLET_STR));
  globalWarning.append (new XplError (WARN_DLLOPEN,       XMSG_DYNAMIC_LIBRARY_OPENED_AGAIN_STR));
  globalWarning.append (new XplError (WARN_DLLCLOSE,      XMSG_DYNAMIC_LIBRARY_NOT_FOUND_STR));
  globalWarning.append (new XplError (WARN_OUTPUT,        XMSG_NO_OUTPUT_FILE_WAS_OPEN_STR));
  globalWarning.append (new XplError (WARN_SELECT,        XMSG_IGNORE_SELECTION_DIMENSION_BY_STR));
  globalWarning.append (new XplError (WARN_FREE,          XMSG_FREE_NOT_ALLOWED_INSIDE_STR));

// now the errors

  globalError.append (new XplError (ERR_INPUT,   XMSG_INVALID_CHARACTER_STR));
  globalError.append (new XplError (ERR_IF,      XMSG_IF_STR));
  globalError.append (new XplError (ERR_ELSE,    XMSG_ELSE_STR));
  globalError.append (new XplError (ERR_ENDIF,   XMSG_ENDIF_STR));
  globalError.append (new XplError (ERR_WHILE,   XMSG_WHILE_STR));
  globalError.append (new XplError (ERR_ENDO,    XMSG_ENDO_STR));
  globalError.append (new XplError (ERR_REPEAT,  XMSG_REPEAT_STR));
  globalError.append (new XplError (ERR_UNTIL,   XMSG_UNTIL_STR));
  globalError.append (new XplError (ERR_SWITCH,  XMSG_SWITCH_STR));
  globalError.append (new XplError (ERR_CASE,    XMSG_CASE_STR));
  globalError.append (new XplError (ERR_DEFAULT, XMSG_DEFAULT_STR));
  globalError.append (new XplError (ERR_SBREAK,  XMSG_BREAK_STR));
  globalError.append (new XplError (ERR_ENDSW,   XMSG_ENDSW_STR));
  globalError.append (new XplError (ERR_PROC,    XMSG_PROC_STR));
  globalError.append (new XplError (ERR_ENDP,    XMSG_ENDP_STR));
  globalError.append (new XplError (ERR_COMBEGIN,"'/*'"));
  globalError.append (new XplError (ERR_COMEND,  "'*/'"));

  globalError.append (new XplError (ERR_ARGS, XMSG_WRONG_NUM_OF_ARGS_STR));

  globalError.append (new XplError (ERR_DIM,    XMSG_WRONG_DIMENSION_STR));
  globalError.append (new XplError (ERR_FILE,   XMSG_CAN_NOT_OPEN_FILE_STR));
  globalError.append (new XplError (ERR_OBJ,    XMSG_CAN_NOT_FIND_OBJECT_STR));
  globalError.append (new XplError (ERR_MAC,    XMSG_CAN_NOT_FIND_QUANTLET_STR));
  globalError.append (new XplError (ERR_COMMON, XMSG_ERROR_STR));

  globalError.append (new XplError (ERR_VAL,  XMSG_WRONG_VALUE_STR));
  globalError.append (new XplError (ERR_EMPTY,XMSG_EMPTY_OBJECT__STR));
  globalError.append (new XplError (ERR_TYPE, XMSG_WRONG_ARGUMENT_TYPE_STR));
  globalError.append (
	  new XplError (ERR_NUM,  XMSG_WRONG_1ARGUMENT_TYPE_STR));
  globalError.append (
          new XplError (ERR_TEXT, XMSG_WRONG_2ARGUMENT_TYPE_STR));
  globalError.append (
          new XplError (ERR_PLOT, XMSG_WRONG_3ARGUMENT_TYPE_STR));
  globalError.append (
          new XplError (ERR_MASK, XMSG_WRONG_4ARGUMENT_TYPE_STR));
  globalError.append (
          new XplError (ERR_TIME, XMSG_WRONG_5ARGUMENT_TYPE_STR));

  globalError.append (new XplError (ERR_VAR,    XMSG_UNKNOWN_VARIABLE_STR));
  globalError.append (new XplError (ERR_FUN,    XMSG_UNKNOWN_FUNCTION_STR));
  globalError.append (new XplError (ERR_SYNTAX, XMSG_SYNTAX_ERROR_STR));
  globalError.append (new XplError (ERR_ASSIGN, XMSG_ERROR_IN_ASSIGNMENT_STR));
  globalError.append (new XplError (ERR_SELECT, XMSG_ERROR_IN_SELECTION_STR));
  globalError.append (new XplError (ERR_COMP,   XMSG_COMPONENT_NOT_FOUND_STR));
  globalError.append (new XplError (ERR_DOT,    XMSG_VARIABLE_END_WITH__STR));
  globalError.append (new XplError (ERR_REF,    XMSG_UNKNOWN_REFERENCE_STR));
  globalError.append (new XplError (ERR_NUMARG, XMSG_WRONG_NUMBER_OF_ARGUMENTS_STR));
  globalError.append (new XplError (ERR_RETARG, XMSG_MISSING_RETURN_ARGUMENT_STR));

  globalError.append (new XplError (ERR_INTERN, XMSG_INTERNAL_ERROR_STR));

  globalError.append (new XplError (ERR_RUNTIME, XMSG_RUNTIME_ERROR_STR));
  globalError.append (new XplError (ERR_PAR,     " "));    // bracket error
  globalError.append (new XplError (ERR_HELP,    XMSG_NO_HELP_FOR__STR));

  globalError.append (new XplError (ERR_INP,     XMSG_INPUT_PARAMETER_ERROR_STR));
  globalError.append (new XplError (ERR_OUTP,    XMSG_OUTPUT_PARAMETER_ERROR_STR));
  globalError.append (new XplError (ERR_PARAM,   XMSG_EQUAL_NAME_FOR_IO_STR));
  globalError.append (new XplError (ERR_CAST,    XMSG_TYPE_MISMATCH_STR));

  globalError.append (new XplError (ERR_NELMIN,  XMSG_ERROR_IN_NELMIN_STR));
  globalError.append (new XplError (ERR_GENAR,   XMSG_ERROR_IN_GENAR_STR));
  globalError.append (new XplError (ERR_DLLOPEN, XMSG_CANT_LOAD_DLL_STR));
  globalError.append (new XplError (ERR_DLLCLOSE,XMSG_CLOSING_DLL_STR));
  globalError.append (new XplError (ERR_DLLCALL, XMSG_CANT_LOAD_SYMBOL_STR));
  globalError.append (new XplError (ERR_SIMPSON, XMSG_ERROR_IN_SIMPSON_STR));
  globalError.append (new XplError (ERR_LOCPOL,  XMSG_ERROR_IN_LOCPOL_STR));

  globalError.append (new XplError (ERR_INTR,   " ### "));
  globalError.append (new XplError (ERR_BREAK,  XMSG_ERROR_IN_QUANTLET___STR));

  alreadyInvoked = TRUE;
}

