/* $Header: /home/xplore/cvs/xpl4gr/xplerror.h,v 1.3 2002/08/16 12:31:58 sigbert Exp $ */

// XploRe 4 Error Class used also for warnings

//  TTK, 4.7.95

/* Description:

   Class XplError:

   This class defines a data structure consisting of an error number and
   an error text. The text can be extended, if necessary.

   Constructors:

   -  XplError (long n)
      generates XplError without text and number n

   -  XplError (long n, char *t)
      generates XplError with text t and number n

   -  ~XplError
      frees the memory of the text

   -  AppendText (char *t)
      appends text t to the existing text

   -  char * GetText () const
      delivers the pointer to text

   -  int GetNumber () const
      delivers the number of the error



   Class XplGlobalError:

   This class is derived from XplError. In addition it contains
   the full error list of XploRe 4 and access functions to these errors.
   Contrary to XplError only on instance of XplGlobalError will be
   generated. It holds the additional protected variable happened, that
   indicates an occurence of an error.

   Constructors:

   -  same as for XplError, happened is initilized with FALSE

   -  Clear ()
   sets happened to FALSE, number to 0, and frees text

   -  Set (long n)
   sets number to n

   -  Set (long n, char *t)
   sets number to n and appends t

   -  Set (long n, char *t, const char *cm, long line)
   sets number to n and appends t, uses as CurrentMacro cm and takes line
   instead of the information from parselist

   -  Get ()
   delivers value of happened

   - ShowError ()
   if happened is TRUE, the to number regarding XplGlobalError.Text is emitted
   followed by the XplError.Text. Afterwards the text is deleted, number is
   set to 0, and happened to FALSE

   - XplError * GetXplError
   searches XplError in list regarding to number. If no XplError exists
   the program is stopped!

*/

#ifndef _XPLERROR_
#define _XPLERROR_

// #pragma interface

#define TRUE  1
#define FALSE 0

#include "liste.h"

#define  WARN_COMMON           10000
#define  WARN_NOCONVERGENCE    10001
#define  WARN_MACRO            10002
#define  WARN_DLLOPEN          10003
#define  WARN_DLLCLOSE         10004
#define  WARN_OUTPUT           10005
#define  WARN_SELECT           10006
#define  WARN_FREE             10007

#define  ERR_INPUT    800

#define  ERR_IF       900
#define  ERR_ELSE     901
#define  ERR_ENDIF    902
#define  ERR_WHILE    903
#define  ERR_ENDO     904
#define  ERR_REPEAT   905
#define  ERR_UNTIL    906
#define  ERR_SWITCH   907
#define  ERR_CASE     908
#define  ERR_DEFAULT  909
#define  ERR_SBREAK   910
#define  ERR_ENDSW    911
#define  ERR_PROC     912
#define  ERR_ENDP     913
#define  ERR_COMBEGIN 914
#define  ERR_COMEND   915

#define  ERR_ARGS     1001
#define  ERR_DIM      1003
#define  ERR_FILE     1004
#define  ERR_OBJ      1005
#define  ERR_MAC      1006
#define  ERR_COMMON   1007
#define  ERR_TYPE     1010
#define  ERR_NUM      1011
#define  ERR_TEXT     1012
#define  ERR_PLOT     1013
#define  ERR_VAL      1014
#define  ERR_EMPTY    1015
#define  ERR_MASK     1016
#define  ERR_TIME     1017

#define  ERR_VAR      2001
#define  ERR_FUN      2002
#define  ERR_SYNTAX   2003
#define  ERR_ASSIGN   2004
#define  ERR_SELECT   2005
#define  ERR_COMP     2006
#define  ERR_DOT      2007
#define  ERR_REF      2008
#define  ERR_NUMARG   2009
#define  ERR_RETARG   2010
#define  ERR_INTERN   2011
#define  ERR_RUNTIME  2012
#define  ERR_PAR      2013
#define  ERR_HELP     2014
#define  ERR_INP      2015
#define  ERR_OUTP     2016
#define  ERR_PARAM    2017
#define  ERR_CAST     2018

#define  ERR_NELMIN   3001
#define  ERR_GENAR    3002
#define  ERR_DLLOPEN  3003
#define  ERR_DLLCLOSE 3004
#define  ERR_DLLCALL  3005
#define  ERR_SIMPSON  3006
#define  ERR_LOCPOL   3007

#define  ERR_INTR     4000       // break for running macro (Ctrl - C)
#define  ERR_BREAK    4001       // break via error function

void  InitializeErrorList   (void);


class XplError {
public:
  XplError (long);
  XplError (long, const char *);
  ~XplError ();

  void         AppendText (const char *t);
  void         AppendText (long n, const char *t);
  const char * GetText    (void) const    { return (text);   }
  long         GetNumber  (void) const    { return (number); }

protected:  long number;
            char *text;
};


class XplGlobalError : public XplError {
public:
  XplGlobalError (long n)               : XplError (n)    { happened = FALSE; }
  XplGlobalError (long n, const char *t): XplError (n, t) { happened = FALSE; }
  ~XplGlobalError () { };

  void      append      (XplError *x);
  void      Clear       (void);
  void      Set         (long n);
  void      Set         (long n, const char *t);
  void      Set         (long n, const char *t, const char *cm, long line);
  int       Get         (void) const                     { return (happened); }
  int       GetError    (void);
  XplError* GetXplError (void) const;
  void      ShowError   (void);
  

protected:  int              happened;
            Liste<class XplError*> errList;
};

class XplGlobalWarning : public XplGlobalError {
public:
  XplGlobalWarning  (long n)               : XplGlobalError (n)    { }
  XplGlobalWarning  (long n, const char *t): XplGlobalError (n, t) { }
  ~XplGlobalWarning () { };

  void      Set         (long n, const char *t = NULL);
  void      Set         (long n, const char *t, const char *cm, long line);
  void      ShowError   (void);
};

extern XplGlobalError    globalError;
extern XplGlobalWarning  globalWarning;
#endif
