/* $Header: /home/xplore/cvs/xpl4gr/xplexec.C,v 1.3 2002/08/16 12:31:58 sigbert Exp $ */

#include "object.h"
#include "xplpp.h"
#include "xplexec.h"
#include "xplerror.h"
#include "init.h"
#include "liste.h"

static int depth = 0;

#ifdef TTKDEBUG
static long count = 0;
#endif

extern Liste<XplObject *> *objList;


void * StartExecution ()
{
#ifdef TTKDEBUG
  count = 0;
#endif
  depth++;

#ifdef TTKDEBUG
  if (depth > 20)
    {
cerr << "StartExecution: depth = " << depth << endl;
cerr << "Are you sure that nothing is going crazy ?" << endl;
//exit (1);
    }
#endif

  return (new Liste<XplObject *>);
}

void EndExecution (void * id)
{
  if (id == NULL)
    {
    globalError.Set (ERR_INTERN, "EndExecution: id = 0");
    return;
    }
  if (depth <= 0)
    {
cerr << "Fatal Error: "; 
cerr << "EndExecution: depth less than zero " << depth << endl;
exit(1);
    globalError.Set (ERR_INTERN, "EndExecution: depth less than zero");
    return;
    }
#ifdef TTKDEBUG
if (ttkdebug)
  cerr << "EndExecution: count " << count << endl;
#endif

  depth--;
long length = ( (Liste<XplObject *>*) id)->laenge ();
for (long i=0; i<length; i++)
  {
  if ((*( (Liste<XplObject *>*) id))[i]->WriteProtection())
    (*( (Liste<XplObject *>*) id))[i]->DelWriteProtection();
  delete (*( (Liste<XplObject *>*) id))[i];
  }

  delete (Liste<XplObject *>*) id;
}

int     Execute   (const char *cmd, void *id, long &translated)
{
  FILE                *fout, *fin;
  Liste<XplObject*>   *tmpList;

#ifdef TTKDEBUG
    count++;
#endif

  static GlobalList gl;
  static long transId = 1;

  //  if ((transId != translated) || translated == 0L)
  if (1)
    {
    if (++transId == 0L)          // if there is an overflow
      transId++;
    translated = transId;         // confirm TokenList for next call
    if (strlen (cmd) == 0)
      return (0);
  
    if (NULL == (fout = fopen (xpl4parsefile, "w")))
      {
      globalError.Set (ERR_INTERN, "Execute: can't open temporary file");
      return (globalError.GetNumber());
      }
    fputs (cmd, fout);
    fputc ('\n', fout);
    fclose (fout);
  
    fin = fopen (xpl4parsefile, "r");
    yyrestart (fin);

    tmpList = objList;              // swap objList with localObjList
    objList = (Liste<XplObject *> *) id;

    DeleteVariables (&gl);          // last chance to delete variables
    gl.Clear ();                    // delete old information
    //    delete gl.GetTokenList ();      // BuildTokenList adds
    BuildTokenList (&gl, '\0');     // new TokenList to gl
    ScanFunsAndMacs (&gl);
    ScanVariables (objList, &gl);
    InputBranch (&gl, 0L);

    objList = tmpList;              // swap back

    fclose (fin);
    remove (xpl4parsefile);
    }

  Vector<TokenElement> *t1 = gl.GetTokenList ();
  GlobalList gl1 (t1);

  tmpList = objList;              // swap objList with localObjList
  objList = (Liste<XplObject *> *) id;

#ifdef TTKDEBUG
if (ttkdebug)
    ShowTokenList (&gl1);
#endif
    ParseTokenList (gl1);
    SetVariables2Null (&gl1);    // resets the VARs in tl (recursion!)

  objList = tmpList;           // swap back

  yyrestart (stdin);


  if (globalError.Get())
    return (globalError.GetNumber());

  return (FALSE);
}

int GenerateXplObject (const char *name, void *x, void *id, int overwrite)
{
  long type;
  XplObject *xo  = GetXplObject ((Liste<XplObject*> *) id, name);

#ifndef MATRIX
  type = ((XplDouble *) x)->xpltypeof ();

  if (type != TYPE_DOUBLE)
    {
cerr << "GenerateXplDouble:  Type not supported" << endl;
return (FALSE);
}
#endif

  if (xo != NULL)
    if (overwrite)
      if (xo->DefaultXplDouble () == NULL)
	{
	xo->DeleteComponents ();
//	xo->AppendComponent (new Component (new XplDouble (&x)));
	xo->AppendComponent (new Component 
			     (new XplDouble ((XplDouble*) x)));
	return (TRUE);
        }
      else
	{
	*(xo->DefaultXplDouble ()) = *((XplDouble*) x);
	return (TRUE);
        }
    else
      return (FALSE);

  ((Liste<XplObject*> *) id)->haengeAn (new XplObject (new XplDouble ((XplDouble*) x), name));
  return (TRUE);
}


void * GetXplObject (const char *name, void *id)
{
  XplObject *xo;

  xo = GetXplObject ((Liste<XplObject*> *) id, name);
  if (xo == NULL)
    {
    return (NULL);
    }

  return (xo->DefaultXplDouble ());
}


int intervall (const char *fun, double &l, double &r)
{
  int       count = 0;
  long      translated = 0L;
  double    fl, fr, fm, m;
  char      cmd[1024];
  XplDouble *x, y(1), *yy;

  void *id;

  id = StartExecution ();

  GenerateXplObject ("yy", &y, id, TRUE);
  GenerateXplObject ("x", &y, id, TRUE);

  yy  = (XplDouble *) GetXplObject ("yy", id);
//  GetXplObject ("yy", yy);

  m = (l+r)/2.0;
  (*yy)(0) = l;
  sprintf (cmd, "x = %s (yy)", fun);
  translated = 0;
  Execute (cmd, id, translated);            // x = fun (l)
  x  = (XplDouble *) GetXplObject ("x", id);
//  GetXplObject ("x", x);
  fl = (*x)(0);
  (*yy)(0) = r;
  translated = 0;
  Execute (cmd, id, translated);            // x = fun (r)
  x  = (XplDouble *) GetXplObject ("x", id);
//  GetXplObject ("x", x);
  fr = (*x)(0);
  (*yy)(0) = m;
  translated = 0;
  Execute (cmd, id, translated);            // x = fun (m)
  x  = (XplDouble *) GetXplObject ("x", id);
//  GetXplObject ("x", x);
  fm = (*x)(0);

  if (fl*fr >= 0)
    {
    cerr << "no zero" << endl;
    return (0);
    }

  while (r-l > 1e-5)
    {
    if (fl*fm > 0.0)
      l = m;
    else
      r = m; 
    m = (l+r)/2.0;
    (*yy)(0) = m;
    //    translated = 0;
    Execute (cmd, id, translated);            // x = fun (m)
    x  = (XplDouble *) GetXplObject ("x", id);
//    GetXplObject ("x", x);
    fm = (*x)(0);
    count++;
    }

  EndExecution (id);

  cerr << "count: " << count << endl;

  return (count);
}

int intervall2 (const char *fun, double &l, double &r)
{
  int       count = 0;
  long      translated = 0L;
  double    fl, fr, fm, m;
  char      cmd[1024];
  XplDouble *x, y(1), *yy;

  void *id;

  id = StartExecution ();

  GenerateXplObject ("yy", &y, id, TRUE);
  GenerateXplObject ("x", &y, id, TRUE);

  yy  = (XplDouble *) GetXplObject ("yy", id);
//  GetXplObject ("yy", yy);

  m = (l+r)/2.0;
  (*yy)(0) = l;
  sprintf (cmd, "x = %s (yy)", fun);
  Execute (cmd, id, translated);            // x = fun (l)
  x  = (XplDouble *) GetXplObject ("x", id);
//  GetXplObject ("x", x);
  fl = (*x)(0);
  (*yy)(0) = r;
  Execute (cmd, id, translated);            // x = fun (r)
  x  = (XplDouble *) GetXplObject ("x", id);
//  GetXplObject ("x", x);
  fr = (*x)(0);
  (*yy)(0) = m;
  Execute (cmd, id, translated);            // x = fun (m)
  x  = (XplDouble *) GetXplObject ("x", id);
//  GetXplObject ("x", x);
  fm = (*x)(0);

  if (fl*fr >= 0)
    {
    cerr << "no zero" << endl;
    return (0);
    }

  while (r-l > 1e-10)
    {
    if (fl*fm > 0.0)
      l = m;
    else
      r = m; 
    m = (l+r)/2.0;
    (*yy)(0) = m;
    Execute (cmd, id, translated);            // x = fun (m)
    x  = (XplDouble *) GetXplObject ("x", id);
//    GetXplObject ("x", x);
    fm = (*x)(0);
    count++;
    }

  EndExecution (id);

  cerr << "count: " << count << endl;

  return (count);
}

