/* $Header: /home/xplore/cvs/xpl4gr/xplexec.h,v 1.3 2002/08/16 12:31:58 sigbert Exp $ */

// XploRe 4 Execute interface for using the X4 language
// from C++ routines

//  TTK, 2.8.95

//  new version with local object lists

/*  local objects list was added for Execution, Generate and Get. This
    is necessary because of nested exection commands in different
    macros. Furthermore no global variable will be generated.
*/

//  TTK, 12.9.95
//  **************************************************
//  Currently only the type XplDouble is supported !!!
//  **************************************************


/* Description:

   This module contains interface functions to invoke X4 macros from
   C++ functions. To have local X4 objects the invoking C++ function
   has to apply for an X4 session with StartExecution. Afterwards
   local objects can be generated, which are only accessible, if
   the right object list identifier provided by StartExecution is given.
   Generating objects is done by the function GenerateXplObject.
   It is compulsory to Generate each object explicitly, even it would
   be generated by the X4 command.

   Example:

   long      translated = 0L;
   XplDouble x(1);

   id = StartExecution();

   GenerateXplObject ("xx", &x, id);
   Execute ("y=sin(xx)", id, translated);


   This would not work as the object "y" is not known within
   the execution module.

   GenerateXplObject creates a copy of the given argument *x, so it
   is even possible to take local instances to initialize X4 objects.
   With get object the C++ function gets the pointer to the XplArray
   structure of the X4 object. The C++ code can so change the content
   of X4 objects immediately. Finally EndExecution frees all X4 objects,
   which are generated in the object list identified by the result
   of StartExecution.


   There are three interface function for communicating with X4.

   - void * StartExecution (void)
     in order to get for each Execution call a certain object list, it
     is necessary to request for execution. This functions provides
     an identifier for such an object list.

   - void EndExecution (void *)
     deletes all objects in the object list which was provide by 
     StartExecution. ATTENTION: neither it is checked, whether id is
     really a pointer allocated by StartExecution nor whether
     EndExecution was already invoked with this identifier!
     If EndExecution is invoked twice or with another pointer, not
     get via StartExecution, X4 will surely crash.

   - int Execute (const char *cmd, void id, long &translated)
     delivers the command cmd to the X4 interpreter
     The objects are local to the environment specified by StartExecution.
     That means, that all objects, which are created during a command
     belong to an own object list. They will be destroyed by EndExecution.
     The return value is zero, if no error occured. Otherwise it is
     the error number, which was set by a globalError.Set call.

   - int GenerateXplObject 
          (const char *name, void *x, void *id, int overwrite = FALSE);

     generates a new XplObject in the object list, which is actually
     id. If an object with the same name already exists, it is only
     overwritten if overwrite is TRUE.  The return value is TRUE if an
     object containing x could be generated otherwise FALSE.

   - void * GetXplObject (const char *name, void *id);
     delivers the default component of the XplObject name.  If there
     is no object name, NULL is the result. NULL can also be the
     result, if the default XplDouble component of an existing object
     is NULL. The objects are searched in the object list, which is
     determined by id (actually this is (Liste<XplObject*> *) id)

*/

#ifndef  _XPLEXEC_
#define  _XPLEXEC_

#define  TRUE   1
#define  FALSE  0

void *  StartExecution    (void);
void    EndExecution      (void *);

int     Execute           (const char *, void *id, long &translated);

int GenerateXplObject (const char *name, 
                       void *x, void *id, int overwrite = FALSE);

void * GetXplObject (const char *name, void  *id);

int intervall  (const char *, double &, double &);
int intervall2 (const char *, double &, double &);

#endif
