/* $Header: /home/xplore/cvs/xpl4gr/xplfun3.C,v 1.5 2002/12/09 13:25:49 cizek Exp $ */

// XploRe core function for different distributions
// see also xplfun.C, xplfun2.C and xplbasefun.C

//  TTK, 5.12.95

#include <math.h>
#include <string.h>
#include <ctype.h>

#include "object.h"
#include "function.h"
#include "xplfun.h"
#include "xplerror.h"
#include "init.h"
#include "ttkio.h"
#include "xplexec.h"
#include "property.h"
#include "directory.h"

#include "locpol.h"

#ifdef PSGEN
#include "plot.h"
#endif

#if defined (_GRAPHICS_) || defined (__BORLANDC__)
#include "display.h"
#include "plot.h"
#endif

#define TEXTLENGTH  1024L

#ifdef _XPLMASK_
XplObject* GenerateXplMask  (Liste<XplObject*> *argl)
// generates a Maskvektor
// TTK, 17.3.96
{
  long      length = argl->laenge ();
  XplObject *o;

  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "generateXplMask: Must be 1");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_TEXT, "GXM: first argument");
    return (NULL);
    }

  o = new XplObject (new XplMask ((*argl)[0]->DefaultXplDouble()->getdim()), "GXM");
  o->GetComponent("GXM")->Content().xplmask->dump("%d");

  return (o);
}

XplObject* SetXplMask  (Liste<XplObject*> *argl)
// generates a Maskvektor
// TTK, 17.3.96
{
  long      length = argl->laenge ();
  XplDouble *x;
  XplMask   *m;
  XplObject *o;

  if (length != 2)
    {
    globalError.Set (ERR_ARGS, "SetXplMask: Must be 2");
    return (NULL);
    }

  if ((*argl)[0]->DefaultComponent()->Type () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "SXM: first argument");
    return (NULL);
    }
  x = (*argl)[0]->DefaultXplDouble ();

  if ((*argl)[1]->DefaultType () != C_XPLMASK)
    {
    globalError.Set (ERR_MASK, "SXM: second argument");
    return (NULL);
    }
  m = (*argl)[1]->DefaultComponent()->Content().xplmask;

  o = GetXplObject (objList, (*argl)[0]->MotherId());
  o->AppendComponent (new Component (new XplObject (new XplMask (*m),"SXM")));
  
  return (NULL);
}
#endif

XplObject* deleteObj  (Liste<XplObject*> *argl)
// deletes global objects of objList if it is not called in a macro
// TTK, 9.5.96
{
  long       pos, i, length = argl->laenge ();
  char const *s1;
  XplObject  *o, *o2;

  // not allowed in macros
  if (objList != globObjList)
    {
    globalWarning.Set (WARN_FREE);
    return (NULL);
    }

  if (length == 0)    // delete all global objects
    {
    length = objList->laenge();
    for (i=0; i<length; i++)
      {
      o = (*objList)[i];
      o->DelWriteProtection ();
      delete o;
      }
    objList->remove ();
    return (NULL);
    }

  // run over all elements
  for (i=0; i<length; i++)
    {
    o = (*argl)[i];
    switch (o->DefaultType ())
      {
      case C_TEXT:
	if (!o->WriteProtection ())       // the obj with this name is meant
	  {
#ifdef _XPLCDEF_
	  s1 = (*o->DefaultComponent()->Content().xplchar)(0).tocstr();
#else
	  s1 = o->DefaultComponent()->Content().text;
#endif
	  o = GetXplObject (objList, s1, pos);
	  if (o == NULL)
	    {
            globalError.Set (ERR_OBJ, s1);
	    return (NULL);
	    }
	  objList->loesche (pos);
	  o->DelWriteProtection ();
	  delete o;
	  }
	else            // obj itself is meant
	  {
	  o = GetXplObject (objList, o->Id (), pos);
	  if (o == NULL)
	    {
		//exit (2);    // cannot happen
		globalError.Set (ERR_OBJ, "should not happen");  // only for testing 6.1.99
	    return (NULL);
	    }
	  objList->loesche (pos);
	  o->DelWriteProtection ();
	  delete o;
          }
	break;
      case C_DISPLAY:
	o2 = GetXplObject (displayList, o->Id(), pos);
	if (o != o2)
	  {
cerr << "Fatal Error: displayList and objList are not consistent" << endl;
exit (1);
          }
	displayList->loesche (pos);
	o2 = GetXplObject (objList, o->Name(), pos);
	objList->loesche (pos);
	o->DelWriteProtection ();
	o->DeleteComponents ();
	delete o;
	break;
      default:
	if (o->WriteProtection ())       // is obj in objList?
	  {
          o = GetXplObject (objList, o->Id(), pos);
          if (o == NULL)
	    {
            globalError.Set (ERR_OBJ, "should not happen");
	    return (NULL);
	    }
	  objList->loesche (pos);
	  o->DelWriteProtection ();
	  delete o;
	  }
      }
      break;
  }
  return (NULL);
}

XplObject* setenv  (Liste<XplObject*> *argl)
// sets environment variables, which are given as first arg
// TTK, 19.2.96
{
  long       n, length = argl->laenge ();
  const char *ss;
  char       *s1, *s2;
 
  if (length != 2)
    {
    globalError.Set (ERR_ARGS, "setenv: Must be 2");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_TEXT)
    {
    globalError.Set (ERR_TEXT, "setenv: first argument");
    return (NULL);
    }

  if ((*argl)[1]->DefaultType () != C_TEXT)
    {
    globalError.Set (ERR_TEXT, "setenv: second argument");
    return (NULL);
    }

#ifdef _XPLCDEF_
  ss = (*(*argl)[0]->DefaultComponent()->Content().xplchar)(0).tocstr();
  s2 = strdup((*(*argl)[1]->DefaultComponent()->Content().xplchar)(0).tocstr());
#else
  ss = (*argl)[0]->DefaultComponent()->Content().text;
  s2 = (*argl)[1]->DefaultComponent()->Content().text;
#endif

  s1 = strdup (ss);

  n = strlen (s1);
  for (long i=0; i<n; i++)
    s1[i] = tolower (s1[i]);

  makefmtstring(s2);

/* SK preliminary 24.06.98
  if (strcmp ("xpl4home", s1) == 0)
    {
    delete[] xpl4home;
    delete[] s1;
    xpl4home = strdup (s2);
    return (NULL);
    }

  if (strcmp ("xpl4lib", s1) == 0)
    {
    delete[] xpl4lib;
    delete[] s1;
    xpl4lib = strdup (s2);
    return (NULL);
    }

  if (strcmp ("xpl4dll", s1) == 0)
    {
    delete[] xpl4dll;
    delete[] s1;
    xpl4dll = strdup (s2);
    return (NULL);
    }

  if (strcmp ("xpl4data", s1) == 0)
    {
    delete[] xpl4data;
    delete[] s1;
    xpl4data = strdup (s2);
    return (NULL);
    }

  if (strcmp ("xpl4help", s1) == 0)
    {
    delete[] xpl4help;
    delete[] s1;
    xpl4help = strdup (s2);
    return (NULL);
    }

  if (strcmp ("xpl4prog", s1) == 0)
    {
    delete[] xpl4prog;
    delete[] s1;
    xpl4prog = strdup (s2);
    return (NULL);
    }

  if (strcmp ("xpl4backup", s1) == 0)
    {
    delete[] s1;
    if (xpl4backup != NULL)
      delete[] xpl4backup;
    xpl4backup = strdup (s2);
    return (NULL);
    }
*/
/*  if (strcmp ("format", s1) == 0)
    {
    delete[] s1;
    if (outputformat != NULL)
      delete[] outputformat;
    outputformat = strdup (s2);
    return (NULL);
    }

  if (strcmp ("stringformat", s1) == 0)
    {
    delete[] s1;
    if (outputstringformat != NULL)
      delete[] outputstringformat;
    outputstringformat = strdup (s2);
    return (NULL);
    }
*/
#ifdef _XPLCDEF_
// sk + tk 310398 begin, neue env-variablen zur steuerung des outputs
/*  if (strcmp ("outheadline", s1) == 0)
  { delete[] s1;
    if (outheadline != NULL)
      delete[] outheadline;
    outheadline = strdup(s2);
    makefmtstring (outheadline);
    return (NULL);
  }
  if (strcmp ("outlineno", s1) == 0)
  { delete[] s1;
    if (outlineno != NULL)
      delete[] outlineno;
    outlineno = strdup(s2);
    makefmtstring (outlineno);
    return (NULL);
  }
  if (strcmp ("outlayerline", s1) == 0)
  { delete[] s1;
    if (outlayerline != NULL)
      delete[] outlayerline;
    outlayerline = strdup(s2);
    makefmtstring (outlayerline);
    return (NULL);
  }
  if (strcmp ("outmaxdata", s1) == 0)
  { delete[] s1;
    outmaxdata = atol(s2);
    return (NULL);
  }*/
// sk + tk 310398 end
#endif
/* sk 30.06.98
  if (strcmp ("browser", s1) == 0)
    {
    delete[] s1;
    if (WWWBrowser != NULL)
      delete[] WWWBrowser;
    WWWBrowser = strdup (s2);
    return (NULL);
    }
*/
  if (property.setProperty(s1, s2))
    globalError.Set (ERR_INTERN, "setenv: environment protected");
  else
    directoryupdate (s1);

  delete[] s1;
  delete[] s2;
  return (NULL);
}

XplObject* getenv  (Liste<XplObject*> *argl)
// reads environment variables, which are given as first arg
// TTK, 2.5.96
{
  long       length = argl->laenge ();
  char       *s1, *res = NULL;
  const char *ss;
 
  if (length != 1)
    {
/* SK 30.06.98 
    globalError.Set (ERR_ARGS, "getenv: Must be 1");
    return (NULL);
*/
     int i, n = property.length();
#ifdef _XPLCDEF_
     XplChar xplc(n, 2);
     for (i=0; i<n; i++)
     { s1 = property.propertyName(i);
       xplc(i,0) = xplstring(s1);
       xplc(i,1) = xplstring(maketxtstring(property.getProperty(s1)));
     }   
     return (new XplObject (new Component (new XplChar (xplc)), "getenv"));
#endif
    }

#ifdef _XPLCDEF_
  XplChar    xplc(1);
#endif
 
  if ((*argl)[0]->DefaultType () != C_TEXT)
    {
    globalError.Set (ERR_TEXT, "getenv: first argument");
    return (NULL);
    }

#ifdef _XPLCDEF_
  ss = (*(*argl)[0]->DefaultComponent()->Content().xplchar)(0).tocstr();
#else
  ss = (*argl)[0]->DefaultComponent()->Content().text;
#endif

  s1 = strdup (ss);

  for (unsigned int i=0; i<strlen (s1); i++)
    s1[i] = tolower (s1[i]);

/* SK 24.06.98 preliminary 
  if (strcmp ("xpl4home", s1) == 0)
    res = strdup (xpl4home);

  if (strcmp ("xpl4lib", s1) == 0)
    res = strdup (xpl4lib);

  if (strcmp ("xpl4dll", s1) == 0)
    res = strdup (xpl4dll);

  if (strcmp ("xpl4data", s1) == 0)
    res = strdup (xpl4data);

  if (strcmp ("xpl4help", s1) == 0)
    res = strdup (xpl4help);

  if (strcmp ("xpl4prog", s1) == 0)
    res = strdup (xpl4prog);

  // only xpl4backup can be NULL
  if ((xpl4backup == NULL) && (strcmp ("xpl4backup", s1) == 0))
     res = strdup ("");

  if ((xpl4backup != NULL) && (strcmp ("xpl4backup", s1) == 0))
    res = strdup (xpl4backup);
*/
/*  if (strcmp ("format", s1) == 0)
    res = strdup (outputformat);

  if (strcmp ("stringformat", s1) == 0)
    res = strdup (outputstringformat);
*/
/* Sk 30.06.98

  if (strcmp ("browser", s1) == 0)
    res = strdup (WWWBrowser);
*/
#ifdef _XPLCDEF_
// sk + tk 310398 begin, neue env-variablen zur steuerung des outputs
/*  if (strcmp("outheadline", s1)==0) // SK 26.03.98
    res = maketxtstring  (outheadline);

  if (strcmp("outlineno", s1)==0) // SK 26.03.98
    res = maketxtstring  (outlineno);

 if (strcmp("outlayerline", s1)==0) // SK 26.03.98
    res = maketxtstring (outlayerline);

  if (strcmp("outmaxdata", s1)==0) // SK 26.03.98
  { char out[64];
    sprintf (out, "%li", outmaxdata);
    res = strdup (out);
  }
*/                                  // SK 26.03.98 + 26.06.98
  if (res==NULL)  // SK 26.06.98
    res = maketxtstring(property.getProperty(s1));

#endif
// sk + tk 310398 end

  delete[] s1;

  if (res == NULL) res = "";
// SK 27.10.98
/*    {
    globalError.Set (ERR_INTERN, "setenv: unknown environment");
    return (NULL);
    }
*/
#ifdef _XPLCDEF_
  xplc(0) = xplstring (res);
  return (new XplObject (new Component (new XplChar (xplc)), "getenv"));
#else
  return (new XplObject (new Component (res), "getenv"));
#endif

}

#ifndef MATRIX
XplObject* xaAtan2  (Liste<XplObject*> *argl)
// arc tan
{
  long      length = argl->laenge ();
  XplDouble *x, *d, *res;
  XplObject *o;

  if (length != 2)
    {
    globalError.Set (ERR_ARGS, "xaAtan2: Must be 2");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaAtan2: first argument");
    return (NULL);
    }

  if ((*argl)[1]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaAtan2: second argument");
    return (NULL);
    }

  x = (*argl)[0]->DefaultXplDouble ();
  d = (*argl)[1]->DefaultXplDouble ();

  res  = new XplDouble(1);
  *res = x->atan2 (*d);

  if (globalError.Get())
    {
    delete res;
    return (NULL);
    }

  o  = new XplObject (res, "atan2");
  return (o);
}

XplObject* xaPdfc  (Liste<XplObject*> *argl)
// density function of Chi^2
{
  long      length = argl->laenge ();
  XplDouble *x, *d, *res;
  XplObject *o;

  if (length != 2)
    {
    globalError.Set (ERR_ARGS, "xaPdfc: Must be 2");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaPdfc: first argument");
    return (NULL);
    }

  if ((*argl)[1]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaPdfc: second argument");
    return (NULL);
    }

  x = (*argl)[0]->DefaultXplDouble ();
  d = (*argl)[1]->DefaultXplDouble ();

  res  = new XplDouble(1);
  *res = x->pdfc (*d);

  if (globalError.Get())
    {
    delete res;
    return (NULL);
    }

  o  = new XplObject (res, "pdfc");
  return (o);
}

XplObject* xaPdfb  (Liste<XplObject*> *argl)
// density function of the Beta distribution
{
  long      length = argl->laenge ();
  XplDouble *x, *a, *b, *res;
  XplObject *o;

  if (length != 3)
    {
    globalError.Set (ERR_ARGS, "xaPdfb: Must be 3");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaPdfb: first argument");
    return (NULL);
    }

  if ((*argl)[1]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaPdfb: second argument");
    return (NULL);
    }

  if ((*argl)[2]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaPdfb: third argument");
    return (NULL);
    }

  x = (*argl)[0]->DefaultXplDouble ();
  a = (*argl)[1]->DefaultXplDouble ();
  b = (*argl)[2]->DefaultXplDouble ();

  res  = new XplDouble(1);
  *res = x->pdfb (*a, *b);

  if (globalError.Get())
    {
    delete res;
    return (NULL);
    }

  o  = new XplObject (res, "pdfb");
  return (o);
}

XplObject* xaPdft  (Liste<XplObject*> *argl)
// density function of the Beta distribution
{
  long      length = argl->laenge ();
  XplDouble *x, *a, *res;
  XplObject *o;

  if (length != 2)
    {
    globalError.Set (ERR_ARGS, "xaPdft: Must be 2");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaPdft: first argument");
    return (NULL);
    }

  if ((*argl)[1]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaPdft: second argument");
    return (NULL);
    }

  x = (*argl)[0]->DefaultXplDouble ();
  a = (*argl)[1]->DefaultXplDouble ();

  res  = new XplDouble(1);
  *res = x->pdft (*a);

  if (globalError.Get())
    {
    delete res;
    return (NULL);
    }

  o  = new XplObject (res, "pdft");
  return (o);
}

XplObject* xaPdfn  (Liste<XplObject*> *argl)
// density function of the Beta distribution
{
  long      length = argl->laenge ();
  XplDouble *x, *res;
  XplObject *o;

  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "xaPdfn: Must be 1");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaPdfn: first argument");
    return (NULL);
    }

  x = (*argl)[0]->DefaultXplDouble ();

  res  = new XplDouble(1);
  *res = x->pdfn ();

  if (globalError.Get())
    {
    delete res;
    return (NULL);
    }

  o  = new XplObject (res, "pdfn");
  return (o);
}

XplObject* xaPdff  (Liste<XplObject*> *argl)
// density function of the F-distribution
{
  long      length = argl->laenge ();
  XplDouble *x, *a, *b, *res;
  XplObject *o;

  if (length != 3)
    {
    globalError.Set (ERR_ARGS, "xaPdff: Must be 3");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaPdff: first argument");
    return (NULL);
    }

  if ((*argl)[1]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaPdff: second argument");
    return (NULL);
    }

  if ((*argl)[2]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaPdff: third argument");
    return (NULL);
    }

  x = (*argl)[0]->DefaultXplDouble ();
  a = (*argl)[1]->DefaultXplDouble ();
  b = (*argl)[2]->DefaultXplDouble ();

  res  = new XplDouble(1);
  *res = x->pdff (*a, *b);

  if (globalError.Get())
    {
    delete res;
    return (NULL);
    }

  o  = new XplObject (res, "pdff");
  return (o);
}

XplObject* xaCdfn  (Liste<XplObject*> *argl)
// normal distribution function
{
  long      length = argl->laenge ();
  XplDouble *x, *res;
  XplObject *o;

  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "xaCdfn: Must be 1");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaCdfn: first argument");
    return (NULL);
    }

  x = (*argl)[0]->DefaultXplDouble ();

  res  = new XplDouble(1);
  *res = x->cdfn ();

  if (globalError.Get())
    {
    delete res;
    return (NULL);
    }

  o  = new XplObject (res, "cdfn");
  return (o);
}

XplObject* xaCdft  (Liste<XplObject*> *argl)
// distribution function of the t-distribution
{
  long      length = argl->laenge ();
  XplDouble *x, *a, *res;
  XplObject *o;

  if (length != 2)
    {
    globalError.Set (ERR_ARGS, "xaCdft: Must be 2");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaCdft: first argument");
    return (NULL);
    }

  if ((*argl)[1]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaCdft: second argument");
    return (NULL);
    }

  x = (*argl)[0]->DefaultXplDouble ();
  a = (*argl)[1]->DefaultXplDouble ();

  res  = new XplDouble(1);
  *res = x->cdft (*a);

  if (globalError.Get())
    {
    delete res;
    return (NULL);
    }

  o  = new XplObject (res, "cdft");
  return (o);
}

XplObject* xaCdfc  (Liste<XplObject*> *argl)
// distribution function of the Chi^2-distribution
{
  long      length = argl->laenge ();
  XplDouble *x, *a, *res;
  XplObject *o;

  if (length != 2)
    {
    globalError.Set (ERR_ARGS, "xaCdfc: Must be 2");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaCdfc: first argument");
    return (NULL);
    }

  if ((*argl)[1]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaCdfc: second argument");
    return (NULL);
    }

  x = (*argl)[0]->DefaultXplDouble ();
  a = (*argl)[1]->DefaultXplDouble ();

  res  = new XplDouble(1);
  *res = x->cdfc (*a);

  if (globalError.Get())
    {
    delete res;
    return (NULL);
    }

  o  = new XplObject (res, "cdfc");
  return (o);
}

XplObject* xaCdfb  (Liste<XplObject*> *argl)
// distribution function of the Beta distribution
{
  long      length = argl->laenge ();
  XplDouble *x, *a, *b, *res;
  XplObject *o;

  if (length != 3)
    {
    globalError.Set (ERR_ARGS, "xaCdfb: Must be 3");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaCdfb: first argument");
    return (NULL);
    }

  if ((*argl)[1]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaCdfb: second argument");
    return (NULL);
    }

  if ((*argl)[2]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaCdfb: third argument");
    return (NULL);
    }

  x = (*argl)[0]->DefaultXplDouble ();
  a = (*argl)[1]->DefaultXplDouble ();
  b = (*argl)[2]->DefaultXplDouble ();

  res  = new XplDouble(1);
  *res = x->cdfb (*a, *b);

  if (globalError.Get())
    {
    delete res;
    return (NULL);
    }

  o  = new XplObject (res, "cdfb");
  return (o);
}

XplObject* xaCdff  (Liste<XplObject*> *argl)
// distribution function of the F-distribution
{
  long      length = argl->laenge ();
  XplDouble *x, *a, *b, *res;
  XplObject *o;

  if (length != 3)
    {
    globalError.Set (ERR_ARGS, "xaCdff: Must be 3");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaPdff: first argument");
    return (NULL);
    }

  if ((*argl)[1]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaCdff: second argument");
    return (NULL);
    }

  if ((*argl)[2]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaCdff: third argument");
    return (NULL);
    }

  x = (*argl)[0]->DefaultXplDouble ();
  a = (*argl)[1]->DefaultXplDouble ();
  b = (*argl)[2]->DefaultXplDouble ();

  res  = new XplDouble(1);
  *res = x->cdff (*a, *b);

  if (globalError.Get())
    {
    delete res;
    return (NULL);
    }

  o  = new XplObject (res, "cdff");
  return (o);
}

XplObject* xaQfn  (Liste<XplObject*> *argl)
// quantile function of the normal distribution
{
  long      length = argl->laenge ();
  XplDouble *x, *res;
  XplObject *o;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "xaQfn: Must be 1");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaQfn: first argument");
    return (NULL);
    }

  x = (*argl)[0]->DefaultXplDouble ();
  res  = new XplDouble(1);
  *res = x->qfn ();

  if (globalError.Get())
    {
    delete res;
    return (NULL);
    }

  o  = new XplObject (res, "qfn");
  return (o);
}

XplObject* xaQft  (Liste<XplObject*> *argl)
// quantile function of the t-distribution
{
  long      length = argl->laenge ();
  XplDouble *x, *a, *res;
  XplObject *o;

  if (length != 2)
    {
    globalError.Set (ERR_ARGS, "xaQft: Must be 2");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaQft: first argument");
    return (NULL);
    }

  if ((*argl)[1]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaQft: second argument");
    return (NULL);
    }

  x = (*argl)[0]->DefaultXplDouble ();
  a = (*argl)[1]->DefaultXplDouble ();

  res  = new XplDouble(1);
  *res = x->qft (*a);

  if (globalError.Get())
    {
    delete res;
    return (NULL);
    }

  o  = new XplObject (res, "qft");
  return (o);
}

XplObject* xaQfc  (Liste<XplObject*> *argl)
// quantile function of the Chi^2-distribution
{
  long      length = argl->laenge ();
  XplDouble *x, *a, *res;
  XplObject *o;
 
  if (length != 2)
    {
    globalError.Set (ERR_ARGS, "xaQfc: Must be 2");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaQfc: first argument");
    return (NULL);
    }

  if ((*argl)[1]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaQfc: second argument");
    return (NULL);
    }

  x = (*argl)[0]->DefaultXplDouble ();
  a = (*argl)[1]->DefaultXplDouble ();

  res  = new XplDouble(1);
  *res = x->qfc (*a);

  if (globalError.Get())
    {
    delete res;
    return (NULL);
    }

  o  = new XplObject (res, "qfc");
  return (o);
}

XplObject* xaQff  (Liste<XplObject*> *argl)
// quantile function of the F-distribution
{
  long      length = argl->laenge ();
  XplDouble *x, *a, *b, *res;
  XplObject *o;

  if (length != 3)
    {
    globalError.Set (ERR_ARGS, "xaQff: Must be 3");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaQff: first argument");
    return (NULL);
    }

  if ((*argl)[1]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaQff: second argument");
    return (NULL);
    }

  if ((*argl)[2]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaQff: third argument");
    return (NULL);
    }

  x = (*argl)[0]->DefaultXplDouble ();
  a = (*argl)[1]->DefaultXplDouble ();
  b = (*argl)[2]->DefaultXplDouble ();

  res  = new XplDouble(1);
  *res = x->qff (*a, *b);

  if (globalError.Get())
    {
    delete res;
    return (NULL);
    }

  o  = new XplObject (res, "qff");
  return (o);
}

XplObject* xaNNinit   (Liste<XplObject*> *argl)
// TTK 1.11.96
{
  long      length = argl->laenge();
  XplDouble *x, *unit, *sa;
  XplLong   *lsa;

  if (length != 2)
    {
    globalError.Set(ERR_ARGS, "xaNNinit: Must = 2");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaNNinit: first argument");
    return (NULL);
    }
  x = (*argl)[0]->DefaultXplDouble ();

  if ((*argl)[1]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaNNinit: second argument");
    return (NULL);
    }
  unit = (*argl)[1]->DefaultXplDouble ();

  lsa = new XplLong;
  x->nninit (*unit, lsa);

  if (globalError.Get())
    {
    delete lsa;
    return (NULL);
    }

  sa = new XplDouble(1);
  *sa = cast (*lsa);
  delete lsa;

  return (new XplObject (sa, "nninit"));
}

XplObject* xaNNfunc   (Liste<XplObject*> *argl)
// TTK 1.11.96
{
  long      length = argl->laenge();
  XplDouble *x, *unit, *xa, *ya;

  if (length != 3)
    {
    globalError.Set(ERR_ARGS, "xaNNfunc: Must = 3");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaNNfunc: first argument");
    return (NULL);
    }
  x = (*argl)[0]->DefaultXplDouble ();

  if ((*argl)[1]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaNNfunc: second argument");
    return (NULL);
    }
  unit = (*argl)[1]->DefaultXplDouble ();

  if ((*argl)[2]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaNNfunc: third argument");
    return (NULL);
    }
  xa = (*argl)[2]->DefaultXplDouble ();

  ya = new XplDouble(1);
  x->nnfunc (*unit, *xa, ya);

  if (globalError.Get())
    {
    delete ya;
    return (NULL);
    }

  return (new XplObject (ya, "nnfunc"));
}

XplObject* xaNNvisu   (Liste<XplObject*> *argl)
// TTK 1.11.96
{
  long      length = argl->laenge();
  XplDouble *x, *xv, *xvd, *ver;
  XplObject *o;

  if (length != 2)
    {
    globalError.Set(ERR_ARGS, "xaNNvisu: Must = 2");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaNNvisu: first argument");
    return (NULL);
    }
  x = (*argl)[0]->DefaultXplDouble ();

  if ((*argl)[1]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaNNvisu: second argument");
    return (NULL);
    }
  xv = (*argl)[1]->DefaultXplDouble ();

  xvd = new XplDouble(1);
  ver = new XplDouble(1);
  x->nnvisu (*xv, xvd, ver );

  if (globalError.Get())
    {
    delete xvd;
    delete ver;
    return (NULL);
    }

  o = new XplObject ("nnvisu");
  o->AppendComponent (new Component (new XplObject (xvd, "xvd")));
  o->AppendComponent (new Component (new XplObject (ver, "ver")));

  return (o);
}


XplObject* xaIsoReg   (Liste<XplObject*> *argl)
// TTK 26.11.96
{
  long      length = argl->laenge();
  XplDouble *x, *res;

  if (length != 1)
    {
    globalError.Set(ERR_ARGS, "xaIsoReg: Must = 1");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaIsoReg: first argument");
    return (NULL);
    }
  x = (*argl)[0]->DefaultXplDouble ();

  res = new XplDouble(1);
  *res = x->isoreg ();

  if (globalError.Get())
    {
    delete res;
    return (NULL);
    }

  return (new XplObject (res, "isoreg"));
}

XplObject* xaRmed   (Liste<XplObject*> *argl)
// TTK 13.2.97
{
  long      length = argl->laenge();
  XplDouble *x, *kr, *res;
  XplLong   k;

  if (length != 2)
    {
    globalError.Set(ERR_ARGS, "xaRmed: Must = 2");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaRmed: first argument");
    return (NULL);
    }
  x = (*argl)[0]->DefaultXplDouble ();

  if ((*argl)[1]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaRmed: second argument");
    return (NULL);
    }
  kr = (*argl)[1]->DefaultXplDouble ();

  k = cast (*kr);

  res = new XplDouble(1);
  *res = x->rmed (k);

  if (globalError.Get())
    {
    delete res;
    return (NULL);
    }

  return (new XplObject (res, "rmed"));
}

XplObject* xaL1line   (Liste<XplObject*> *argl)
// TTK 13.2.97
{
  long      length = argl->laenge();
  XplDouble *x, *y, *res;

  if (length != 2)
    {
    globalError.Set(ERR_ARGS, "xaL1line: Must = 2");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaL1line: first argument");
    return (NULL);
    }
  x = (*argl)[0]->DefaultXplDouble ();

  if ((*argl)[1]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaL1line: second argument");
    return (NULL);
    }
  y = (*argl)[1]->DefaultXplDouble ();

  res = new XplDouble(1);
  *res = x->l1line (*y);

  if (globalError.Get())
    {
    delete res;
    return (NULL);
    }

  return (new XplObject (res, "l1line"));
}

XplObject* xaLowess   (Liste<XplObject*> *argl)
// TTK 13.2.97
{
  long      length = argl->laenge();
  XplDouble *x, *y, *f = NULL, *res;

  if (length != 2 && length != 3)
    {
    globalError.Set(ERR_ARGS, "xaLowess: Must = 2 or 3");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaLowess: first argument");
    return (NULL);
    }
  x = (*argl)[0]->DefaultXplDouble ();

  if ((*argl)[1]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaLowess: second argument");
    return (NULL);
    }
  y = (*argl)[1]->DefaultXplDouble ();

  if (length == 3)
    {
    if ((*argl)[2]->DefaultType () != C_XPLDOUBLE)
      {
      globalError.Set (ERR_NUM, "xaLowess: third argument");
      return (NULL);
      }
    f = (*argl)[2]->DefaultXplDouble ();
    }

  res = new XplDouble(1);
  *res = x->lowess (*y, f);

  if (globalError.Get())
    {
    delete res;
    return (NULL);
    }

  return (new XplObject (res, "lowess"));
}


XplObject* xaLocPol(Liste<XplObject*> *argl)
// TTK 8.2.97  Interface zu StSp local polynomial
{
  long      length = argl->laenge();
  XplDouble *x, *xg, *y, *h, *l, *res;

  if (length != 5)
    {
    globalError.Set(ERR_ARGS, "xaLocPol: Must = 5");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaLocPol: first argument");
    return (NULL);
    }
  x = (*argl)[0]->DefaultXplDouble ();

  if ((*argl)[1]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaLocPol: second argument");
    return (NULL);
    }
  xg = (*argl)[1]->DefaultXplDouble ();

  if ((*argl)[2]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaLocPol: third argument");
    return (NULL);
    }
  y = (*argl)[2]->DefaultXplDouble ();

  if ((*argl)[3]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaLocPol: forth argument");
    return (NULL);
    }
  h = (*argl)[3]->DefaultXplDouble ();

  if ((*argl)[4]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaLocPol: fifth argument");
    return (NULL);
    }
  l = (*argl)[4]->DefaultXplDouble ();

  res = new XplDouble(1);
  *res = stefan1 (x, xg, y, h, l);

  if (globalError.Get())
    {
    delete res;
    return (NULL);
    }

  return (new XplObject (res, "locpol"));
}

XplObject* xaLocH(Liste<XplObject*> *argl)
// TTK 8.2.97  Interface zu StSp local polynomial
{
  long      length = argl->laenge();
  XplDouble *x, *xg, *xd, *y, *h, *l, *res;

  if (length != 5 && length != 6)
    {
    globalError.Set(ERR_ARGS, "xaLocH: Must be 5 or 6");
    return (NULL);
    }

  if ((*argl)[0]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaLocH: first argument");
    return (NULL);
    }
  x = (*argl)[0]->DefaultXplDouble ();

  if ((*argl)[1]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaLocH: second argument");
    return (NULL);
    }
  xg = (*argl)[1]->DefaultXplDouble ();

  if ((*argl)[2]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaLocH: third argument");
    return (NULL);
    }
  y = (*argl)[2]->DefaultXplDouble ();

  if ((*argl)[3]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaLocH: forth argument");
    return (NULL);
    }
  h = (*argl)[3]->DefaultXplDouble ();

  if ((*argl)[4]->DefaultType () != C_XPLDOUBLE)
    {
    globalError.Set (ERR_NUM, "xaLocPol: fivth argument");
    return (NULL);
    }
  l = (*argl)[4]->DefaultXplDouble ();

  if (length == 6)
    {
    if ((*argl)[5]->DefaultType () != C_XPLDOUBLE)
      {
      globalError.Set (ERR_NUM, "xaLocH: sixth argument");
      return (NULL);
      }
    xd = (*argl)[5]->DefaultXplDouble ();
    }
  else
    xd = NULL;

  res = new XplDouble(1);
  *res = stefan2 (x, xd, xg, y, h, l);

  if (globalError.Get())
    {
    delete res;
    return (NULL);
    }

  return (new XplObject (res, "loch"));
}



#endif  // of ! MATRIX

XplObject* list   (Liste<XplObject*> *argl)
{
  XplObject   *t;
  long     i, length = argl->laenge(), z = 1;
  char   n[128];

  if (length < 1)
    {
    globalError.Set(ERR_ARGS, "list: Must >= 1");
    return (NULL);
    }

  if (strcmp ((*argl)[0]->Name (), "_tmp") == 0)
    {
    sprintf (n, "el%ld\0", z);
    t = new XplObject (new Component (new XplObject ((*argl)[0], n)));
    }
  else
    t = new XplObject (new Component (new XplObject ((*argl)[0])));

  z++;        // z is position in list

  for (i=1; i<length; i++)
    {
    if (strcmp ((*argl)[i]->Name (), "_tmp") == 0)
      {
      sprintf (n, "el%ld\0", z);
      t->AppendComponent (new Component (new XplObject ((*argl)[i], n)));
      }
    else
      t->AppendComponent (new Component (new XplObject ((*argl)[i])));
    z++;
    }
  return (t);
}

XplObject* NumberOfComponents  (Liste<XplObject*> *argl)
// number of elements of an object
// return 0, if object is simple
{
  long      length = argl->laenge ();
  XplDouble *x;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "size: accepts just one parameter");
    return (NULL);
    }

  x = new XplDouble (1);

  // PC 20021122
  if ((*argl)[0]->Type () & O_SIMPLE)
  {
    if ((*argl)[0]->Type() == O_XPLDOUBLE)
    {
      XplDouble *o = (*argl)[0]->DefaultXplDouble();
      (*x)(0) = o->size();
    }
    else
    if ((*argl)[0]->Type() == O_XPLCHAR)
    {
      XplChar *o = (*argl)[0]->DefaultComponent()->Content().xplchar;
      (*x)(0) = o->size();
    }
    else
/*    if ((*argl)[0]->Type() == O_DISPLAY)
    {
      XplDisplay *o = (*argl)[0]->DefaultComponent()->Content().display;
      (*x)(0) = o->GetRows() * o->GetColumns();
    }
    else
*/
    if ((*argl)[0]->Type() == O_XPLTIME)
    {
      XplTime *o = (*argl)[0]->DefaultComponent()->Content().xpltime;
      (*x)(0) = o->size();
    }
    else (*x)(0) = 0.0;          // the object is simple !
  }
  else
    (*x)(0) = (double) (*argl)[0]->ComponentNumber ();

  return (new XplObject (x, "size"));
}

XplObject* appendObject  (Liste<XplObject*> *argl)
// appends an the second object on the first
// if the first is simple, both objects are appended to a new one
{
  long       length = argl->laenge ();
  XplObject  *o, *a, *n;
  char       name[64];
  const char *oname; 
  if (length != 2)
    {
    globalError.Set (ERR_ARGS, "appendObject: must be 1");
    return (NULL);
    }

  o = (*argl)[0];
  if (o == NULL)
    {
    globalError.Set (ERR_COMMON, "appendObject: empty argument objects");
    return (NULL);
    }

  if (!o->WriteProtection ())
    {
    globalError.Set (ERR_COMMON, "appendObject: does not work with temporary objects");
    return (NULL);
    }

  sprintf (name, "el%ld", o->ComponentNumber ()+1);
  a = (*argl)[1];
  if (a->WriteProtection ())
    oname = a->Name ();
  else
    oname = name;

  if (o->Type () & O_SIMPLE)   // wrap the original object
    {
    n = new XplObject (o->Name ());
    n->AppendComponent (new Component (new XplObject (o)));
    n->AppendComponent (new Component (new XplObject (a, oname)));
    o->TransferComponents (n);
    }
  else
    {
    o->AppendComponent (new Component (new XplObject (a, oname)));
    }

  if (o->ComponentNumber () == 2)     // changes the standard view ?
    {
    XplObject *po = GetXplObject (objList, o->Id ());
    if (po == NULL)
      po = GetContainerXplObject (objList, o->Id ());
    if (po == NULL)
      {
cerr << "Fatal Error in appendObject from list object" << endl;
exit (1);
      }
    po->SetDefaults ();
    return (NULL);
    }

  return (NULL);
}

XplObject* deleteObject  (Liste<XplObject*> *argl)
// number of elements of an object
// return 0, if object is simple
{
  long      pos, length = argl->laenge ();
  XplDouble *x;
  XplObject *o;
 
  if (length != 2)
    {
    globalError.Set (ERR_ARGS, "deleteObjects: must be 2");
    return (NULL);
    }

  o = (*argl)[0];
  if (o == NULL)
    {
    globalError.Set (ERR_COMMON, "appendObject: empty argument objects");
    return (NULL);
    }

  if (!o->WriteProtection ())
    {
    globalError.Set (ERR_COMMON, "appendObject: does not work with temporary objects");
    return (NULL);
    }

  if (o->Type () & O_SIMPLE)
    {
    globalError.Set (ERR_COMMON, "deleteObject: first argument expected list");
    return (NULL);
    }

  x = (*argl)[1]->DefaultXplDouble ();
  if (x == NULL)
    {
    globalError.Set (ERR_NUM, "deleteObject: second argument");
    return (NULL);
    }

  pos = -1 + (long) (*x)(0);

  if ((pos < 0) || (pos >= (*argl)[0]->ComponentNumber ()))
    {
    globalError.Set (ERR_COMMON, "deleteObject: no argument at position");
    return (NULL);
    }

  // if component on position pos contains object, delete write protection
  // TTK, 31.5.98
  if ((*o)[pos]->Type () == C_XplObject)
    (*o)[pos]->Content().xplobject->DelWriteProtection ();

  if (pos == 0)     // changes the standard view ?
    {
    XplObject *po = GetXplObject (objList, o->Id ());
    if (po == NULL)
      po = GetContainerXplObject (objList, o->Id ());
    if (po == NULL)
      {
cerr << "Fatal Error in deleteObject from list object" << endl;
exit (1);
      }
    
    o->DeleteComponent (pos);
    po->SetDefaults ();
    return (NULL);
    }
  o->DeleteComponent (pos);
  return (NULL);
}

XplObject* insertObject  (Liste<XplObject*> *argl)
// number of elements of an object
// return 0, if object is simple
{
  long      pos, length = argl->laenge ();
  XplDouble *x;
  XplObject *o, *a;
 
  if (length != 3)
    {
    globalError.Set (ERR_ARGS, "insertObjects: must be 1");
    return (NULL);
    }

  o = (*argl)[0];
  if (o == NULL)
    {
    globalError.Set (ERR_COMMON, "insertObject: empty argument objects");
    return (NULL);
    }

  if (!o->WriteProtection ())
    {
    globalError.Set (ERR_COMMON, "insertObject: does not work with temporary objects");
    return (NULL);
    }

  a = (*argl)[2];
  if (o->Type () & O_SIMPLE)
    {
    globalError.Set (ERR_COMMON, "insertObject: first argument expected list");
    return (NULL);
    }

  x = (*argl)[1]->DefaultXplDouble ();
  if (x == NULL)
    {
    globalError.Set (ERR_NUM, "insertObject: second argument");
    return (NULL);
    }

  pos = -1 + (long) (*x)(0);

  if ((pos < 0) || (pos >= (*argl)[0]->ComponentNumber ()))
    {
    globalError.Set (ERR_COMMON, "insertObject: position not valid");
    return (NULL);
    }

  char name[64];
  sprintf (name, "el%ld", pos+1);
  o->InsertComponent (pos, new Component (new XplObject (a, name)));

  if (pos == 0)     // has the standard view changed ?
    {
    XplObject *po = GetXplObject (objList, o->Id ());
    if (po == NULL)
      po = GetContainerXplObject (objList, o->Id ());
    if (po == NULL)
      {
cerr << "Fatal Error in insertObject from list object" << endl;
exit (1);
      }
    po->SetDefaults ();
    }
  return (NULL);
}

XplObject* containsComponent  (Liste<XplObject*> *argl)
// if object is composed the position of the object with the given 
// name is deliverd
// otherweise 0
{
  long       pos, length = argl->laenge ();
  XplObject  *o;
  XplDouble  *x;
  const char *t1, *t2;
 
  if (length != 2)
    {
    globalError.Set (ERR_ARGS, "containsObjects: must be 1");
    return (NULL);
    }

  if ((*argl)[1]->DefaultType() != C_TEXT)
    {
    globalError.Set(ERR_TEXT, "containsComponent:  second argument");
    return (NULL);
    }
#ifdef _XPLCDEF_
  t2 = (*(*argl)[1]->DefaultComponent()->Content().xplchar)(0).tocstr();
#else
  t2 = (*argl)[1]->DefaultComponent()->Content().text;
#endif

  o = (*argl)[0];
  if (o->Type () & O_SIMPLE)
    {
    if (o->DefaultType () != C_TEXT)
      {
      XplDouble *res = new XplDouble (1);
      (*res)(0) = 0.0;
      return (new XplObject (res, "comp"));
      }
#ifdef _XPLCDEF_
    t1 = (*o->DefaultComponent()->Content().xplchar)(0).tocstr();
#else
    t1 = o->DefaultComponent()->Content().text;
#endif
    o = GetXplObject (objList, t1);
    if (o == NULL)
      {
      XplDouble *res = new XplDouble (1);
      (*res)(0) = -1.0;
      return (new XplObject (new Component (res), "comp"));
      }
    }

  if (o->Type () & O_SIMPLE)
    {
    XplDouble *res = new XplDouble (1);
    (*res)(0) = 0.0;
    return (new XplObject (res, "comp"));
    }

  x = new XplDouble (1);

  if (o->GetComponent (t2, pos) == NULL)
    (*x)(0) = 0.0;             // component doesn't exist
  else
    (*x)(0) = (double) pos+1;

  return (new XplObject (x, "comp"));
}

#ifdef _XPLCDEF_
XplObject* namesOfComponents  (Liste<XplObject*> *argl)
// if object is composed the names of the contained objects are deliverd
// ohterweise NULL
{
  long      length = argl->laenge ();
  XplObject *o;
 
  if (length != 1)
    {
    globalError.Set (ERR_ARGS, "namesOfObjects: must be 1");
    return (NULL);
    }

  o = (*argl)[0];
  if (o->Type () & O_SIMPLE)
    {
    globalError.Set (ERR_COMMON, "namesOfObject: object is simple");
    return (NULL);
    }

  long i, n = o->ComponentNumber ();
  XplChar xplc(n);

  for (i=0; i<n; i++)
    {
    if ((*o)[i]->Type () != C_XplObject)
      {
      cerr << "ERROR: Component should be an object!" << endl;
      }
    xplc(i) = xplstring ((*o)[i]->Name ());
    }

  return (new XplObject (new XplChar (xplc), "names"));
}
#endif  // XPLCDEF
