/* $Header: /home/xplore/cvs/xpl4gr/xplmask.C,v 1.3 2002/08/16 12:31:59 sigbert Exp $ */

#include <math.h>
#include "xplmask.hpp"

  xplmask::xplmask ()
  { red = green = blue = symbol = size = 0; }

  xplmask::xplmask (int r, int g, int b)
  { red   = r;
    if (r>255) red = 255;
    if (r<0) red = 0;
    green = g;
    if (g>255) green = 255;
    if (g<0) green = 0;
    blue  = b;
    if (b>255) blue = 255;
    if (b<0) blue = 0;
    symbol = size = 0;
  }

  xplmask::xplmask (int r, int g, int b, int sy, int si)
  { red    = r;
    if (r>255) red = 255;
    if (r<0) red = 0;
    green  = g;
    if (g>255) green = 255;
    if (g<0) green = 0;
    blue   = b;
    if (b>255) blue = 255;
    if (b<0) blue = 0;
    symbol = sy;
    if ((sy<0) || (sy>15)) symbol = 0;
    size   = si;
    if ((si<0) || (si>15)) size = 0;
  }

  xplmask::xplmask (long mask)
  { red    = (mask & MASK_RED);
    green  = (mask & MASK_GREEN)  >> 8;
    blue   = (mask & MASK_BLUE)   >> 16;
    symbol = (mask & MASK_SYMBOL) >> 20;
    size   = (mask & MASK_SIZE)   >> 24;
  }

  xplmask::xplmask (double dmask)
  { long mask = (long) dmask;
    red    = (mask & MASK_RED);
    green  = (mask & MASK_GREEN)  >> 8;
    blue   = (mask & MASK_BLUE)   >> 16;
    symbol = (mask & MASK_SYMBOL) >> 20;
    size   = (mask & MASK_SIZE)   >> 24;
  }

  void xplmask::SetRGB (int r, int g, int b)
  { red   = r;
    if (r>255) red = 255;
    if (r<0) red = 0;
    green = g;
    if (g>255) green = 255;
    if (g<0) green = 0;
    blue  = b;
    if (b>255) blue = 255;
    if (b<0) blue = 0;
  }

  void xplmask::SetHLS (float h, float l, float s)
  { float m1, m2, hue;
    m2 = (l<0.5 ? l*(1+s) : l+s-l*s);
    m1 = 2*l-m2;
    if (s<1.0e-6)
      red = green = blue = 0;
    else
    { // red 
      hue = fmod (h+120.0, 360.0);
      if (hue<60)
        red = (long) (255*(m1+(m2-m1)*hue/60));
      else if (hue<180)
        red = (long) (255*m2);
      else if (hue<240)
        red = (long) (255*(m1+(m2-m1)*(240-hue)/60));
      else
        red = (long) (255*m1);
      // green 
      hue = fmod (h, 360.0);
      if (hue<60)
        green = (long) (255*(m1+(m2-m1)*hue/60));
      else if (hue<180)
        green = (long) (255*m2);
      else if (hue<240)
        green = (long) (255*(m1+(m2-m1)*(240-hue)/60));
      else
        green = (long) (255*m1);
      // blue 
      hue = fmod (h-120.0, 360.0);
      if (hue<60)
        blue = (long) (255*(m1+(m2-m1)*hue/60));
      else if (hue<180)
        blue = (long) (255*m2);
      else if (hue<240)
        blue = (long) (255*(m1+(m2-m1)*(240-hue)/60));
      else
        blue = (long) (255*m1);
    }
  }

  void xplmask::SetRed (int r)
  { red = r;
    if (r>255) red = 255;
    if (r<0) red = 0;
  } 
  
  void xplmask::SetBlue (int b)
  { blue = b;
    if (b>255) blue = 255;
    if (b<0) blue = 0;
  }

  void xplmask::SetGreen (int g)
  { green = g; 
    if (g>255) green = 255;
    if (g<0) green = 0;
  }

  void xplmask::SetSymbol (int sy)
  { symbol = sy; 
    if ((sy<0) || (sy>15)) symbol = 0;
  }

  void xplmask::SetSize (int si)
  { size = si; 
    if ((si<0) || (si>15)) size = 0;
  }

  void xplmask::GetRGB (int *r, int *g, int *b) const
  { *r = red;
    *g = green;
    *b = blue;
  }

  void xplmask::GetRed (int *r) const
  { *r = red; } 
  
  void xplmask::GetBlue (int *b) const
  { *b = blue; }

  void xplmask::GetGreen (int *g) const
  { *g = green; }

  void xplmask::GetSymbol (int *sy) const
  { *sy = symbol; }

  void xplmask::GetSize (int *si) const
  { *si = size; }

  int xplmask::operator!= (const xplmask r) const
  { return ((red!=r.red) || (blue!=r.blue) || (green!=r.green) ||
            (symbol!=r.symbol) || (size!=r.size)); }


  int xplmask::operator== (const xplmask r) const
  { return ((red==r.red) && (blue==r.blue) && (green==r.green) && 
            (symbol==r.symbol) && (size==r.size)); }

  int xplmask::operator< (const xplmask r) const
  { return 0; }

  int xplmask::operator> (const xplmask r) const
  { return 0; }

