/* $Header: /home/xplore/cvs/xpl4gr/xplmat.h,v 1.3 2002/08/16 12:32:00 sigbert Exp $ */

// XplMatrix Klasse   TTK+SwS  6.6.94
// aus vector.C

// last modification:  24.8.95

/*
Folgende Klasse XplMatrix wird definiert:

Konstruktoren:

-	XplMatrix ()
	erzeugt leere XplMatrix. row=col=0;

-	XplMatrix (const int)
	erzeugt Vektor der vorgegebenen Laenge

-	XplMatrix (const int, const int);
	erzeugt XplMatrix mit vorgegebenen Zeilen und Spalten

-	XplMatrix (const int, const int, const int i);
	erzeugt EinheitsXplMatrix mit vorgegebenen Zeilen und Spalten,
	wenn i==1

Kopierkonstruktoren:

-	XplMatrix (const XplMatrix&)
	erzeugt tiefe Kopie der Referenzmatrix

-	XplMatrix (XplMatrix*);
	erzeugt tiefe Kopie der XplMatrix, auf der der Pointer zeigt


Destruktor:

-	~XplMatrix ()
	gibt des reservierten Speicher fuer die Elemente frei

Operatoren:


-	double& operator() (const int, const int);
	Referenzzugriff auf XplMatrix ueber Zeilen/Spalten
	
-	double  operator() (const int, const int) const;
	Element aus XplMatrix (Zeilen/Spalten) lesen
	
-	XplMatrix  operator+ (const XplMatrix&);	//elementweise
-	XplMatrix  operator- (const XplMatrix&);	//elementweise

-	XplMatrix  operator* (const double); 		
-	XplMatrix  operator/ (const double); 		

-	XplMatrix  operator* (const XplMatrix&); 	//matrixmult

-	Boolean  operator== (const XplMatrix&);
-	Boolean  operator!= (const XplMatrix&);

-	XplMatrix& operator= (const XplMatrix&);
	Weist XplMatrix zu

Attributfunktionen:

-	int	isScalar (void) const;
	True, wenn row == col == 1

-	int	rows (void) const;
	liefert Anzahl der Zeilen

-	int	cols (void) const;
	liefert Anzahl der Spalten

Extremum berechnen:
-		XplMatrix & 	min_col() const;
-		XplMatrix & 	max_col() const;
-		XplMatrix &	extremum() const;

Rechnen:


-	XplMatrix& transpose (void) const;
	Transponieren

-	XplMatrix& inv(void) const;	//nur fuer diagon. natricen!!!

-	void	fill (const double);

Ausgabe:

-	void	out (void) const;
	gibt XplMatrix nach cout aus

Interna:

-	int row, col;
	Zeilen- und Spaltenzahl

-	double *elem;
	Array der Elemente


*/



#ifndef _XPLMATRIX_H
#define _XPLMATRIX_H

#include "liste.h"
#include <fstream.h>
#include <stdlib.h>		//Sw


#if defined (_UNIX_) 
#include "motif.h"
#else
#define Boolean bool
#define False 0
#define True 1
#endif

/*--------------------------------------------------------------------*/

struct _MY_POINT_	{	short x;
			short y;
		};
		
typedef struct _MY_POINT_	MyPoint;
/*	soll mit XPoint uebereinstimmen!!!	*/

/*--------------------------------------------------------------------*/

typedef	double (*singlefun) (double);


class XplMatrix {
  public:	XplMatrix ();
 
#ifndef __BORLANDC__ //LOOK! the same story
   		XplMatrix (const int);
  		XplMatrix (const int, const int);
  		XplMatrix (const int, const int, const int);
#else
		XplMatrix (int);
		XplMatrix (int, int);
		XplMatrix (int, int, int);
#endif
		
		XplMatrix (const XplMatrix&);
  		XplMatrix (XplMatrix* const);
		~XplMatrix ();


#ifndef __BORLANDC__ //LOOK! the same story
		double& operator() (const int, const int);
		double  operator() (const int, const int) const;
#else
		double& operator() (int, int);
		double  operator() (int, int) const;
#endif

		XplMatrix  operator+ (const XplMatrix&);
		XplMatrix  operator- (const XplMatrix&);

		XplMatrix  operator* (const double); 		
		XplMatrix  operator/ (const double); 
	
		XplMatrix  operator* (const XplMatrix&);

		Boolean  operator== (const XplMatrix&);
		Boolean  operator!= (const XplMatrix&);

		XplMatrix& operator= (const XplMatrix&);

		int	isScalar (void) const	{ return (row*col == 1); }
		
		int	rows (void) const	{ return (row); }
		int	cols (void) const	{ return (col); }

		XplMatrix  	min_col(void) const;
		XplMatrix  	max_col(void) const;
		XplMatrix	extremum(void) const;
		
		XplMatrix 	transpose (void) const;
		XplMatrix 	inv(void) const;
 
                int 	inverse         (void);
		void	out (void) const;
		void	fill (const double);

friend void Compute(const XplMatrix & data, const XplMatrix & mult, 
			const XplMatrix & add, MyPoint * wohin);

  private:	int 	row, col;
		double	*elem;
		int	match (const XplMatrix&);
  };  

double * XplMatrixAlloc (int, int);
#endif
