/* $Header: /home/xplore/cvs/xpl4gr/xplore.withoutmac.y,v 1.3 2002/08/16 12:32:01 sigbert Exp $ */

/*
Syntaktische Analyse fuer XploRe (Prototyp fuer XploRe 4 Interpreter):

by TTK 30.6.94  - 6.2.95 (repeat until)

Verarbeitung von Matrizen:

Folgende Operatoren werden unterstuetzt (in dieser Prioritaet):

	Operator	Bedeutung

	~ |		horizontale und vertikale Verkettung
	+ -		Addition, Subtraktion von Matrizen
	.* / *		elementweise Mult. und Division, Matrixmult.
	-		Vorzeichen
	:		Sequenz, nicht assoziativ
	< <= > >=	Relation
	== !=		Vergleich
	|| &&		OR / AND
	
Der Parser versteht folgende Regeln:

	Klammern (), Selektieren[], Funktion auf Matrixelemente,
	Zuweisung =, Vektorliste #(), Argumentliste (Selektor, Funktion)
*/
	
%{
#include <stream.h>
#include "liste.h"
#include "object.h"
#include "function.h"
#include "xplcalc.h"
#include "xplpp.h"
#include "init.h"

YYSTYPE yylval;		// needed for mylex !!

int errorToken = 0;     // for error state

XplDouble *x;
Object    *o;

Liste<Object*>             *objList;          // globals
Liste<Function*>           *funList;
class GlobalList           *newgl;

Liste<Liste<Variable*>*>   argumlist;         // global in xplore.y
Liste<Variable*>           *tmparg;

Liste<double>		tmp;
// Liste<Object*>	        *varlist;
// Liste<Component*>	*argl, *tmp2, *argl1, *tmp1;
// Liste<char*>		cargl, *ctmp2, *ctmp3;

// Liste<Liste<Component*>*>   argumlist;
// Liste<Liste<char*>*>	 cargumlist;

extern "C" {
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <math.h>

int	yywrap (void);		// Prototypes necessary here
void	yyerror (const char *);
int	yyparse (void);
int	mylex (void);

%}


%union {
int		  Index;
double		  Real;
char              *Name;
class Object  	  *Obj;
// class Component	  *Comp;
class Variable    *Var;
class Function    *Fun;
Liste<double>	  *List;
Liste<Variable*>  *VarList;
long		  Branch;
}
  /* for IF, ELSE, ENDIF, REPEAT, UNTIL */
  /*     WHILE, ENDWHILE */


%pure_parser
%start	program

%token	<Var>   VARIABLE
// %token	<Comp>  COMPONENT
%token	<Real>  CONST
%token	<Index> OP_PM OP_MD OP_REL OP_EQ OP_AND OP_OR OP_SEQ
%token  <Index> DOT
%token  <Name>  NEWSTRING 
%token  <Fun>   FUNCTION 

%token	<Index> IF ELSE ENDIF WHILE ENDO REPEAT UNTIL ENDP

%type	<Var>		constvec ausdruck
%type   <List>          list
%type   <VarList>       arglist tmparglist

%left	'~' '|'
%left	OP_OR
%left	OP_AND
%left	OP_EQ
%left	OP_REL
%left	OP_PM
%left	OP_MD
%nonassoc OP_SEQ
%left	SIGN
%left	TRAN


%%

program         :  /* empty */
                | program eingabe
                ;

eingabe		: ausdruck		{ if ($1 != NULL)
					    {
					    $1->Object()->Print(); 
					    if ($1->XplDouble () != NULL)
					      $1->XplDouble().dump("%lf");
					    delete $1;
					    }
					}
                | '\n'
//		| VARIABLE '=' ausdruck   { AssignObject ($1, $3); }
// objList->fuegeEin ( new Object ($3));}
		| NEWSTRING '=' ausdruck  { if ($3 != NULL)
					      {
					      o = $3->TransferObject();
					      o->SetName ($1);
					      o->SetDefaultXplDouble ();
					      objList->fuegeEin (o);
					      delete $3;
					      }
//					    else error
					  }
		;


list		: CONST			{ tmp.fuegeEin($1,0); $$ = &tmp; }
		| OP_PM CONST 		{ if ($1 == MINUS)
					    tmp.fuegeEin(-$2,0);
					  else
					    tmp.fuegeEin($2,0);
					  $$ = &tmp; }
		| list ',' CONST	{ tmp.fuegeEin($3,0); }
		| list ',' OP_PM CONST	{ if ($3 == MINUS)
					    tmp.fuegeEin(-$4,0);
					  else
					    tmp.fuegeEin($4,0);
					  $$ = &tmp; }
		;
		
constvec	: '#' '(' list ')'	{ x = new XplDouble ($3->laenge());
					  for (int j=0, i=$3->laenge(); i>0;)
					    (*x)(--i) = (*$3)[j++];
					  $3->remove();
                           		  $$ = new Variable (x);
                                        }
                ;

ausdruck	: constvec		{ $$ = $1; }
		| VARIABLE		{ $$ = $1; }
//                             { $$ = ($1==NULL) ? NULL : new Variable ($1);}
                                         // generates a DEEP copy
		| CONST			{ x = new XplDouble (1);
					  (*x)(0) = $1;
                           		  $$ = new Variable (x); }
		| '(' ausdruck ')'		{ $$ = $2; }
		| ausdruck OP_MD ausdruck { $$ = operate($1,$3,$2); }
		| ausdruck OP_PM ausdruck { $$ = operate($1,$3,$2); }
                | FUNCTION '('      { argumlist.fuegeEin(tmparg, 0);
						  tmparg = new Liste<Variable*>; }
                               arglist ')'        {
						  $$ = ($1->Fun ()) ($4);
			//			  tmparg->remove();  should be not necessary
			                          long i, length = tmparg.laenge();
						  for (i=0; i<length; i++)
						    delete (*tmparg)[i];
						  delete tmparg;     // entries deleted ? TEST
						  tmparg = argumlist[0];
						  argumlist.loesche(0); }
                ;

tmparglist      : ausdruck	          { if ($1 != NULL)
					      tmparg->haengeAn ($1);   $$ = tmparg; }
		| tmparglist ',' ausdruck { if ($3 != NULL)
					      tmparg->haengeAn ($3);   $$ = tmparg; }
                ;

arglist         : tmparglist   { $$ = $1; }

%%


#include "lex.yy.c"

}  // of extern C

      			
#include "xplpp.h"

#ifndef _GRAPHICS_
int main (int argc, char **argv)
{
  Liste<TokenElement> tl;

#ifdef YYDEBUG
  yydebug  = (argc >= 3);        // debug info if dummy parameter is present
#endif
#ifdef TTKDEBUG
extern int ttkdebug;
  ttkdebug = (argc == 2) || (argc >= 4);
#endif

  InitializeInterpreter ();  

while (1)
  {
PrintObjectList (objList);
  BuildTokenList (&parselist);
//  InputBranch (&parselist, 0L);
ScanObjects (objList, &parselist);
  ShowTokenList (&parselist);
  ParseTokenList (parselist);
  parselist.GetTokenList()->remove();
  }

return (0);
}
#endif     // of _GRAPHICS_
