/* $Header: /home/xplore/cvs/xpl4gr/xplore.y,v 1.6 2003/05/23 17:35:10 cizek Exp $ */
/*
Syntaktische Analyse fuer XploRe (Prototyp fuer XploRe 4 Interpreter):

by TTK 30.6.94  - 6.2.95 (repeat until)

Verarbeitung von Matrizen:

Folgende Operatoren werden unterstuetzt (in dieser Prioritaet):

	Operator	Bedeutung

	~ |		horizontale und vertikale Verkettung
	+ -		Addition, Subtraktion von Matrizen
	.* / *		elementweise Mult. und Division, Matrixmult.
	-		Vorzeichen
	:		Sequenz, nicht assoziativ
	< <= > >=	Relation
	== !=		Vergleich
	|| &&		OR / AND
	
Der Parser versteht folgende Regeln:

	Klammern (), Selektieren[], Funktion auf Matrixelemente,
	Zuweisung =, Vektorliste #(), Argumentliste (Selektor, Funktion)
*/
	
%{
#include <string.h>
#include "machine.h"
//#if !defined (__BORLANDC__) && !defined (WIN32) //LOOK!
#ifndef HAS_IOSTREAMH
#include <stream.h>
#else
#include <iostream.h>
#include <malloc.h>
#endif

// #pragma implementation

#include "liste.h"
#include "object.h"
#include "function.h"
#include "macro.h"
#include "xplcalc.h"
#include "xplpp.h"
#include "init.h"
// ini file 	sk + tk 090798 SK 05.02.99
#include "property.h"
#include "directory.h"
// ini file 	sk + tk 090798 SK 05.02.99

// PC - D-TEC - Japanese
#include "xplore_strings.h"
// end of PC - D-TEC - Japanese

#ifdef TTKDEBUG
extern long XplObjectId;
extern char *programId;

extern long objcounter;

extern int ttkdebug;
extern int tracedebug;
extern int docheck;
#endif

#ifdef __BORLANDC__ //LOOK!
#include "xplwin.h"
void startxplore (void);

#endif

#if defined(__BORLANDC__) || defined(SOCKETSERVER) || defined(X4singleserver)
void statusmessage(const char *msg);
#endif

YYSTYPE yylval;		// needed for mylex !!

int errorToken = 0;     // for error state

int          nsel;
long         i;
XplDouble    *xx;
XplObject    *o, *obj;
Macro        *mac;
Function     *fun;
const char   *text1;
char         *xpltext;

#ifdef _XPLCDEF_
XplChar      xplc(1);
#endif

XplObject * isNotEmpty (Liste<XplObject*> *argl);  // xplfun2.h
XplObject * deleteObj  (Liste<XplObject*> *argl);  // xplfun3.h

void ExitXpl       (int);

Liste<XplObject*>          *displayList, *constList,
                           *objList, *globObjList;   // globals
Vector<Function*>           *funList;
Vector<Macro*>              *macList;
class GlobalList           *newgl;

Liste<Liste<XplObject*>*>   argumlist;         // global in xplore.y
Liste<XplObject*>           *tmparg, *seltmp, *sold[20], *csel;
Liste<const char*>	    *ctmp;
Liste<XplDouble*>	    *xpldoublelist;
static int soldcounter = 0;

#include <math.h>               // necessary for HP UX

extern void logt (const char *txt);  // SK 31.05.2000


extern "C" {
#include <stdio.h>
#include <errno.h>
#include <ctype.h>

int	yywrap  (void);		// Prototypes necessary here
void	yyerror (const char *);
int	yyparse (void);
int	mylex   (void);

%}


%union {
int		    Index;
double		    Real;
char                *Name;
class XplObject     *Obj;
class Variable      *Var;
class Function      *Fun;
class Macro         *Mac;
class MacroText	    *MacText;
Liste<Variable*>    *VarList;
Liste<XplObject*>   *ObjList;
Liste<const char*>  *CharList;
long		    Branch;
}
  /* for IF, ELSE, ENDIF, REPEAT, UNTIL */
  /*     WHILE, ENDWHILE */


%pure_parser
%start	program

%token	<Var>   VARIABLE
%token	<Obj>   CONSTVAR
%token	<Real>  CONSTANT
%token	<Index> OP_PM OP_MD OP_TMD OP_P OP_REL OP_EQ OP_AND OP_OR OP_SEQ OP_CON
%token  <Index> SINGLE_OPN SINGLE_OPV

%token  <Index> DOT COMMENT COMBEGIN COMEND INTERRUPT ASSIGNBRACKET
%token  <Name>  NEWSTRING STRCONST REFMAC MFCAND UNKNOWNINPUT
%token  <Fun>   FUNCTION
%token  <Mac>   MACRO

%token  <MacText>       XPROC

%token	<Index> IF ELSE ENDIF WHILE ENDO REPEAT UNTIL ENDP
%token	<Index> SWITCH CASE BREAK DEFAULT ENDSW NOP

%type   <Branch>        block         // probably not necessary

%type	<Obj>		ausdruck
%type   <ObjList>       arglist tmparglist tsellist sellist compsellist
%type   <CharList>      carglist parlist assignlist

%left	  OP_CON
%left	  OP_OR
%left	  OP_AND
%left	  OP_EQ
%left	  OP_REL
%nonassoc OP_SEQ
%left	  OP_PM
%left	  OP_MD
%left	  OP_TMD         //  y*x'*z    '* must have higher priority
%left	  SINGLE_OPN
%left	  SINGLE_OPV
%left	  SIGN
%right	  OP_P


%%

program         :  /* empty */
                | program block
		| error { if (!errorToken)        // get the FIRST errTok !
			    {
#ifdef INETD
cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif
			    // SK 991211 cerr << "Error: " << yychar << endl; 
			    errorToken = yychar;
			    }
			}
                  '\n'	{ yyerrok;
#ifndef __BORLANDC__	//SWS 19.12            
                 // tk 240398	  PrintToken (errorToken); 
			  parselist.PrintCurrentToken ();
#endif
//globalError.ShowError ();
//			  mt.SetNewText();       // if error with proc def
			  errorToken = 0; }
                ;

block		: eingabe         {}  
		| INTERRUPT       { YYERROR; }
		| block eingabe   {}
		| IF     '(' ausdruck ')'      {  $$ = IfHandler     ($3, $1); }
		| WHILE  '(' ausdruck ')'      {  $$ = WhileHandler  ($3, $1); }
		| UNTIL  '(' ausdruck ')'      {  $$ = UntilHandler  ($3, $1); }
		| SWITCH '(' ausdruck ')'      {  $$ = SwitchHandler ($3, $1); }
		| CASE   '(' ausdruck ')'      {  $$ = CaseHandler   ($3, $1); }
		| XPROC '(' parlist ')' '=' NEWSTRING '(' parlist ')'
		 { newgl = CopyTokenList();
//	6.12.96	   if (GetXplObject (objList, $6) == NULL)
//		     {
		     if (newgl != NULL)
		       {
		       mac = new Macro($6, $1->Text(), *$8,*$3, newgl);
		       macList->haengeAn (mac);
//		       mac->PrepareToRun ();
		       delete $1;           // MacroText must be deleted
		       }
		     else
		       {
#ifdef INETD
cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif
		       cerr << "Error in Definition of " << $6 << endl;
                       }
//		     }
//		   else
//		     {
//#ifdef INETD
//cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
//#endif
//		     cerr << $6 << " is already Variable" << endl;
//                    globalError.Set (ERR_INTERN, "Variable cannot be macro");
//		     if (newgl != NULL)
//		       delete newgl;           // what is with list elements?
//		     }
		   }
		| XPROC '(' parlist ')' '=' MACRO '(' parlist ')'
		 { $6->SetNewDefinition ($1->Text(), *$8, *$3, CopyTokenList());
		   delete $1;             // MacroText must be deleted
		   }
		;
eingabe		: ausdruck  { if ($1 != NULL)
				{
				$1->Print (); 
				if ((! $1->WriteProtection ()) &&
				    ($1->DefaultType() != C_PLOT))
				  delete $1;
			        }
			    }
                | '\n'
		| NOP { }
                | assignlist '=' ausdruck { if ($3 == NULL)
					      {
if (globalError.GetNumber () != ERR_INTR)      // error only if not interrupted
  {
#ifdef INETD
cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif
  cerr << "Error in list assignment: no Variables generated"<<endl;
globalError.Set (ERR_INTERN, "Error in list assignment: no obj generated ");
  }
}
					    else
					      {
                                              assign ($1, $3);
					      if (!$3->WriteProtection ())
						delete $3;
					      }
		                            $1->remove ();
					    delete $1;
                                          }
		| NEWSTRING '=' ausdruck  { if ($3 != NULL)
					      {
					      o = GetXplObject (objList, $1);
					      if (o != NULL)
						{
                      o->DelWriteProtection ();          // allow deletion of all components
                      o->TransferComponents ($3);
                      o->SetWriteProtection ();          // protect XplObject
                      GetXplObject(objList, $1)->SetDefaultXplDouble ();
		    }
					      else
						if ($3->WriteProtection ())
						  {
						  o = new XplObject ($3, $1);
						  o->SetWriteProtection ();
						  objList->haengeAn (o);
						  }
						else
						  {
						  $3->SetName ($1);
						  $3->SetWriteProtection ();
						  objList->haengeAn ($3);
						  }
					      }
					    else 
					      {
#ifdef INETD
cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif
cerr << "Error in assignment: Variable " << $1 << " not generated"<<endl;
globalError.Set (ERR_INTERN, "Error in assignment: obj not generated ");
globalError.AppendText (ERR_INTERN, $1);
                                              }
					  }
		| VARIABLE '=' ausdruck   { if ($3 != NULL)
					      {
                      o = $1->GetXplObject ();
//                      o->DelWriteProtection ();          // allow deletion of all components

if ($3->WriteProtection ())	     // wg  x=x.el1.el2   etc.
  {
  o->DelWriteProtection ();          // allow deletion of all components
  if (!$3->WriteProtection ())       // is $3 element of o?
    {
    $3->SetWriteProtection ();       // yes: protect again
    o->TransferComponents ($3);
    $3->DelWriteProtection ();       // now it's save to remove it
    }
  else
    o->TransferComponents ($3);      // if $3 is not affected by DelWriteP
  }
else
  o->TransferComponents ($3);        // $3 is temporary

                      o->SetWriteProtection ();          // protect XplObject
                      GetXplObject(objList, $1->Name())->SetDefaultXplDouble ();
                      if (! $3->WriteProtection ())
		        delete $3;
					      }
					    else
					      {
#ifdef INETD
cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif
cerr << "Error in assignment: Variable " << $1->Name() << " not changed"<<endl;
globalError.Set (ERR_INTERN, "Error in assignment: ");
globalError.AppendText (ERR_INTERN, $1->Name ());
                                              }
					  }

		| VARIABLE  sellist '=' ausdruck  {
XplDouble           *xs = NULL, *x1;
#ifdef _XPLCDEF_
XplChar             *xd = NULL;
#endif
    obj = $1->GetXplObject();
    switch (obj->DefaultType ())
      {
      case C_XPLDOUBLE:  xs  = obj->DefaultXplDouble();   break;
#ifdef _XPLCDEF_
      case C_XPLCHAR  :  xd  = obj->DefaultComponent()->Content().xplchar;
	                 break;
#endif
      case C_EMPTY    :  globalError.Set (ERR_INTERN, "emtpy object in [] ");
                         break;
      default         :  globalError.Set (ERR_INTERN, "wrong type in [] ");
      }
#ifndef MATRIX
XplSelection *xplsel;     // must be defined inside yyparse
                          // because of recursion   TTK, 9.1.97
    if ($4 == NULL)
      globalError.Set (ERR_INTERN, "Var [] doesn't exist ");

    else if (obj->DefaultType () != $4->DefaultType ())  // SK 19.01.2000
      globalError.Set (ERR_CAST, obj->Name ());

    if (!globalError.Get ())
      {                      
#ifdef _XPLCDEF_
      if (xd != NULL)
	xplsel = GenerateXplSelection (xd->getdim (), nsel, MakeArglist($2, obj));
      else
#endif
	xplsel = GenerateXplSelection (xs->getdim (), nsel, MakeArglist($2, obj));
      if (xplsel == NULL)
	{
#ifdef INETD
cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif
        cerr << "ERROR: Selector for " << $1->Name () << endl;
        }
      else
	{
#ifdef _XPLCDEF_
	if (xd != NULL)
	  xd->assignment (nsel, xplsel, *($4->DefaultComponent()->Content().xplchar));
	else
#endif
	  xs->assignment (nsel, xplsel, *($4->DefaultXplDouble ()));
	delete[] xplsel;
        }
      if (! $4->WriteProtection ())
	delete $4;
      }  // !globalError
#else  // of MATRIX
    xs  = obj->DefaultXplDouble();
    if (xs == NULL || $4 == NULL)
      globalError.Set (ERR_INTERN, "Var [] doesn't exist ");
    else
      {                      
      xpldoublelist = MakeArglist($2, obj);
      xs = $4->DefaultXplDouble ();
      $1->GetXplObject()->DefaultXplDouble()->selectassign(*xpldoublelist, xs);
      if (! $4->WriteProtection ())
        delete $4;
      xpldoublelist->remove();
      delete xpldoublelist;
      }
#endif
    if ($2 != NULL)
      {
      long length = $2->laenge ();  // now delete tmp selection
      for (i=0; i<length; i++)
	{
	if ((*$2)[i] != NULL && !(*$2)[i]->WriteProtection ())
	  delete (*$2)[i];
        }
      $2->remove();
      delete $2;
      }
  }
		| VARIABLE  compsellist '=' ausdruck  {
		    obj = $1->GetXplObject ();
		    if (obj != NULL)
		      {
		      o = SelectComponentList (obj, $2);
		      AssignSelection (o, $4);
		      long i, length = $2->laenge();
		      for (i=0; i<length; i++)
			if (! (*$2)[i]->WriteProtection ())
			  delete (*$2)[i];
		      delete $2;
                      if (! $4->WriteProtection ())
		        delete $4;
		      obj->SetDefaultXplDouble ();
		      }
		    else
                      globalError.Set (ERR_INTERN, "empty obejct in slc");
		    }
		;

		| VARIABLE  compsellist sellist '=' ausdruck  {       // new  TTK, 13.1.98
		    obj = $1->GetXplObject ();
		    if (obj != NULL)
		      {
		      o = SelectComponentList (obj, $2);
// AssignSelection (o, $4);
// copy of  VARIABLE  sellist '=' ausdruck    $2 -> $3,  $4 -> $5

XplDouble           *xs = NULL, *x1;
#ifdef _XPLCDEF_
XplChar             *xd = NULL;
#endif
    obj = o;
    switch (obj->DefaultType ())
      {
      case C_XPLDOUBLE:  xs  = obj->DefaultXplDouble();   break;
#ifdef _XPLCDEF_
      case C_XPLCHAR  :  xd  = obj->DefaultComponent()->Content().xplchar;
	                 break;
#endif
      case C_EMPTY    :  globalError.Set (ERR_INTERN, "emtpy object in [] ");
                         break;
      default         :  globalError.Set (ERR_INTERN, "wrong type in [] ");
      }
#ifndef MATRIX
XplSelection *xplsel;     // must be defined inside yyparse
                          // because of recursion   TTK, 9.1.97
    if ($5 == NULL)
      globalError.Set (ERR_INTERN, "Var [] doesn't exist ");

    if (!globalError.Get ())
      {                      
#ifdef _XPLCDEF_
      if (xd != NULL)
	xplsel = GenerateXplSelection (xd->getdim (), nsel, MakeArglist($3, obj));
      else
#endif
	xplsel = GenerateXplSelection (xs->getdim (), nsel, MakeArglist($3, obj));
      if (xplsel == NULL)
	{
#ifdef INETD
cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif
        cerr << "ERROR: Selector for " << $1->Name () << endl;
        }
      else
	{
#ifdef _XPLCDEF_
	if (xd != NULL)
	  xd->assignment (nsel, xplsel, *($5->DefaultComponent()->Content().xplchar));
	else
#endif
	  xs->assignment (nsel, xplsel, *($5->DefaultXplDouble ()));
	delete[] xplsel;
        }
//      if (! $5->WriteProtection ())   siehe unten
//	delete $5;
      }  // !globalError
#else  // of MATRIX
// following not tested yet  TTK, 13.1.98
    xs  = obj->DefaultXplDouble();
    if (xs == NULL || $5 == NULL)
      globalError.Set (ERR_INTERN, "Var [] doesn't exist ");
    else
      {                      
      xpldoublelist = MakeArglist($3, obj);
      xs = $5->DefaultXplDouble ();
//    $1->GetXplObject()->DefaultXplDouble()->selectassign(*xpldoublelist, xs);
      obj->DefaultXplDouble()->selectassign(*xpldoublelist, xs);
      xpldoublelist->remove();
      delete xpldoublelist;
      }
#endif
// end of   VARIABLE  compsellist '=' ausdruck

		      long i, length = $2->laenge();
		      for (i=0; i<length; i++)
			if (! (*$2)[i]->WriteProtection ())
			  delete (*$2)[i];
                      $2->remove();
		      delete $2;
                      if (! $5->WriteProtection ())
		        delete $5;
		      obj->SetDefaultXplDouble ();
		      }
		    else
                      globalError.Set (ERR_INTERN, "empty obejct in slc");
		    }
		;

ausdruck	: VARIABLE      { o = $1->GetXplObject ();
                                  if (o->isEmpty ())
                                    {
// globalError.Set (ERR_INTERN, "empty Var ");
// globalError.AppendText (ERR_INTERN, o->Name());
//                                   yyerror ("empty Var");
#ifdef TTKDEBUG
if (tracedebug)
  cerr << "empty object: shouldn't matter " << o->Name() << endl;
#endif
// parselist.SetPos (parselist.GetTokenList()->laenge()-1);
//  TTK 2.2.96                                    YYERROR;
                                    }
                                  $$ = o;  }
		| CONSTVAR      { $$ = $1; }
		| VARIABLE sellist
      { XplDouble         *xs = NULL, *xs1;
#ifdef _XPLCDEF_
        XplChar           *xd = NULL, *xd1;
#endif
		  obj = $1->GetXplObject();
        switch (obj->DefaultType ())
        { case C_XPLDOUBLE:
            xs  = obj->DefaultXplDouble();
            break;
#ifdef _XPLCDEF_
          case C_XPLCHAR  :
            xd  = obj->DefaultComponent()->Content().xplchar;
            break;
#endif
          case C_EMPTY    :
            globalError.Set (ERR_INTERN, "emtpy object in [] ");
            $$ = NULL;
            break;
          default         :
            globalError.Set (ERR_INTERN, "wrong type in [] ");
            $$ = NULL;
        }
#ifndef MATRIX
        XplSelection *xplsel;
          // must be defined inside yyparse
          // because of recursion   TTK, 9.1.97

        if (!globalError.Get ())
        {
#ifdef _XPLCDEF_
          if (xd != NULL)
            xplsel = GenerateXplSelection (xd->getdim (), nsel, MakeArglist($2, obj));
          else
#endif
	       xplsel = GenerateXplSelection (xs->getdim (), nsel, MakeArglist($2, obj));
          if (xplsel == NULL)
	       {
#ifdef INETD
cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif
            cerr << "ERROR: Selector for " << $1->Name () << endl;
            $$ = NULL;
          }
          else
	       {
#ifdef _XPLCDEF_
	         if (xd != NULL)
            { xd1 = new XplChar (1);
              xd1->selection (*xd, nsel, xplsel);
	           $$ = new XplObject (xd1);
            }
            else
#endif
	         {
            xs1 = new XplDouble (1);
            xs1->selection (*xs, nsel, xplsel);
            $$ = new XplObject (xs1);
	         }
	         delete[] xplsel;
          }
// werden alle Matrizen geloescht?
        }   // of !globalError
        else
          $$ = NULL;
#else     // of MATRIX
        xpldoublelist = MakeArglist($2, obj);
        $$ = new XplObject ($1->GetXplObject()->DefaultXplDouble()->selection (*xpldoublelist));
// werden alle Matrizen geloescht?
        xpldoublelist->remove();
        delete xpldoublelist;
#endif

        if ($2 != NULL)
        { long length = $2->laenge ();  // now delete tmp selection
          for (i=0; i<length; i++)
	       { if ((*$2)[i] != NULL && !(*$2)[i]->WriteProtection ())
	         delete (*$2)[i];
          }
          $2->remove();
          delete $2;
        }
      }
		| ausdruck sellist    {
XplDouble         *xs = NULL, *xs1;
#ifdef _XPLCDEF_
XplChar           *xd = NULL, *xd1;
#endif
#ifndef MATRIX
XplSelection *xplsel;     // must be defined inside yyparse
                          // because of recursion   TTJ, 9.1.97
		  obj = $1;
		  if (obj != NULL)
		    {
switch (obj->DefaultType ())
  {
  case C_XPLDOUBLE:  xs  = obj->DefaultXplDouble();   break;
#ifdef _XPLCDEF_
  case C_XPLCHAR  :  xd  = obj->DefaultComponent()->Content().xplchar;
                     break;
#endif
  case C_EMPTY    :  globalError.Set (ERR_INTERN, "emtpy object in [] ");
                     $$ = NULL;
                     break;
  default         :  globalError.Set (ERR_INTERN, "wrong type in [] ");
                     $$ = NULL;
  }

#ifdef _XPLCDEF_
		      if (xd != NULL)
			xplsel = GenerateXplSelection (xd->getdim (), nsel, MakeArglist($2, obj));
                      else
#endif
			xplsel = GenerateXplSelection (xs->getdim (), nsel, MakeArglist($2, obj));
		    if (xplsel == NULL)
		      {
#ifdef INETD
cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif
		      cerr << "ERROR: Selector for term" << endl;
		      $$ = NULL;
		      }
		    else
		      {
#ifdef _XPLCDEF_
		      if (xd != NULL)
			{
			xd1 = new XplChar (1);
			xd1->selection (*xd, nsel, xplsel);
			$$ = new XplObject (xd1);
		        }
else
#endif
                        {
		        xs1 = new XplDouble (1);
		        xs1->selection (*xs, nsel, xplsel);
		        $$ = new XplObject (xs1);
		        }
		      }
                      delete[] xplsel;
		      }
		  else
		    {
		    globalError.Set (ERR_INTERN, "selection of empty arg");
		    $$ = NULL;
		    }
#else
		  if ($1 != NULL)
		    {
		    xpldoublelist = MakeArglist($2, $1);
    $$ = new XplObject ($1->DefaultXplDouble()->selection (*xpldoublelist));
// werden alle Matrizen geloescht?
                    xpldoublelist->remove();
		    delete xpldoublelist; 
		    }
		  else
		    {
		    globalError.Set (ERR_INTERN, "selection of empty arg");
		    $$ = NULL;
		    }

#endif
		  if (! $1->WriteProtection ())   // must be deleted
		      delete $1;

		    if ($2 != NULL)
		      {
		      long length = $2->laenge ();  // now delete tmp selection
		      for (i=0; i<length; i++)
			{
			if ((*$2)[i] != NULL && !(*$2)[i]->WriteProtection ())
			  delete (*$2)[i];
                        }
		      $2->remove();
		      delete $2;
		      }
              }
		| ausdruck compsellist
                           { o = SelectComponentList ($1, $2);
			     if (o != NULL)
			       {
			       if (! $1->WriteProtection ())
				 {
				 $$ = $1->TransferXplObject (o);
				 delete $1;
			         }
			       else
				 $$ = new XplObject (o);
			       }
			     else
			       {
			       if (! $1->WriteProtection ())
				 delete $1;
		    globalError.Set (ERR_INTERN, "selection{} failed");
			       $$ = NULL;
			       }
			     long i, length = $2->laenge ();
			     for (i=0; i<length; i++)
			       if (! (*$2)[i]->WriteProtection ())
				 delete (*$2)[i];
			   }
		| VARIABLE compsellist
                           { obj = $1->GetXplObject ();
                             o   = SelectComponentList (obj, $2);
			     if (o != NULL)
			       {
			       if (! obj->WriteProtection ())
				 {
				 $$ = new XplObject("csel");
				 $$->TransferXplObject (o);
				 delete obj;
			         }
			       else
				 $$ = new XplObject (o);
			       }
			     else
			       {
		    globalError.Set (ERR_INTERN, "selection{} failed");
			       $$ = NULL;
			       }
			     long i, length = $2->laenge ();
			     for (i=0; i<length; i++)
			       if (! (*$2)[i]->WriteProtection ())
				 delete (*$2)[i];
			   }
		| VARIABLE compsellist sellist             // new TTK, 15.1.98
                           { obj = $1->GetXplObject ();
                             o   = SelectComponentList (obj, $2);
			     if (o != NULL)
			       {
			       if (! obj->WriteProtection ())
				 {
				 $$ = new XplObject("csel");
				 $$->TransferXplObject (o);
				 delete obj;
			         }
			       else
				 $$ = new XplObject (o);
			       }
			     else
			       {
		    globalError.Set (ERR_INTERN, "selection{} failed");
			       $$ = NULL;
			       }
			     long i, length = $2->laenge ();
			     for (i=0; i<length; i++)
			       if (! (*$2)[i]->WriteProtection ())
				 delete (*$2)[i];
// copy of ausdruck sellist   $2 -> $3
//                | ausdruck sellist    
if (o != NULL)
  {
XplDouble         *xs = NULL, *xs1;
#ifdef _XPLCDEF_
XplChar           *xd = NULL, *xd1;
#endif
#ifndef MATRIX
XplSelection *xplsel;     // must be defined inside yyparse
                          // because of recursion   TTK, 9.1.97
		  obj = o;
		  if (obj != NULL)
		    {
switch (obj->DefaultType ())
  {
  case C_XPLDOUBLE:  xs  = obj->DefaultXplDouble();   break;
#ifdef _XPLCDEF_
  case C_XPLCHAR  :  xd  = obj->DefaultComponent()->Content().xplchar;
                     break;
#endif
  case C_EMPTY    :  globalError.Set (ERR_INTERN, "emtpy object in [] ");
                     $$ = NULL;
                     break;
  default         :  globalError.Set (ERR_INTERN, "wrong type in [] ");
                     $$ = NULL;
  }

#ifdef _XPLCDEF_
		      if (xd != NULL)
			xplsel = GenerateXplSelection (xd->getdim (), nsel, MakeArglist($3, obj));
                      else
#endif
			xplsel = GenerateXplSelection (xs->getdim (), nsel, MakeArglist($3, obj));
		    if (xplsel == NULL)
		      {
#ifdef INETD
cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif
		      cerr << "ERROR: Selector for term" << endl;
		      $$ = NULL;
		      }
		    else
		      {
#ifdef _XPLCDEF_
		      if (xd != NULL)
			{
			xd1 = new XplChar (1);
			xd1->selection (*xd, nsel, xplsel);
			$$ = new XplObject (xd1);
		        }
else
#endif
                        {
		        xs1 = new XplDouble (1);
		        xs1->selection (*xs, nsel, xplsel);
		        $$ = new XplObject (xs1);
		        }
		      }
                      delete[] xplsel;
		      }
		  else
		    {
		    globalError.Set (ERR_INTERN, "selection of empty arg");
		    $$ = NULL;
		    }
#else
		  if ($1 != NULL)
		    {
		    xpldoublelist = MakeArglist($3, o);
    $$ = new XplObject (o->DefaultXplDouble()->selection (*xpldoublelist));
// werden alle Matrizen geloescht?
                    xpldoublelist->remove();
		    delete xpldoublelist; 
		    }
		  else
		    {
		    globalError.Set (ERR_INTERN, "selection of empty arg");
		    $$ = NULL;
		    }
#endif
		    if ($3 != NULL)
		      {
		      long length = $3->laenge ();  // now delete tmp selection
		      for (i=0; i<length; i++)
			{
			if ((*$3)[i] != NULL && !(*$3)[i]->WriteProtection ())
			  delete (*$3)[i];
                        }
		      $3->remove();
		      delete $3;
		      }
// end of copy
			   }
}
		| CONSTANT	   { xx = new XplDouble (1);
				     (*xx)(0) = $1;
				     $$ = new XplObject (xx); }
		| STRCONST {
//           text = new char[strlen($1)+1];
//	   sprintf (text, "%s", $1);
                           text1 = $1;    // strdup ($1);
#ifdef _XPLCDEF_
			   xplc(0) = xplstring (text1);
			   $$ = new XplObject (new Component (new XplChar (xplc)));
#else
                           $$ = new XplObject (new Component (strdup (text1)));
#endif
			   }
		| '(' ausdruck ')'		{ $$ = $2; }
		| OP_PM ausdruck  %prec SIGN	{ if ($1 == MINUS)
						    $$ = operate($2, $1);
						  else
						    $$ = $2;         }
		| SINGLE_OPV ausdruck      { $$ = operate($2,$1);    }
		| ausdruck SINGLE_OPN      { $$ = operate($1,$2);    }
		| ausdruck OP_CON ausdruck { $$ = operate($1,$3,$2); }
		| ausdruck OP_TMD ausdruck { $$ = operate($1,$3,$2); }
		| ausdruck OP_MD ausdruck  { $$ = operate($1,$3,$2); }
		| ausdruck OP_PM ausdruck  { $$ = operate($1,$3,$2); }
		| ausdruck OP_P  ausdruck  { $$ = operate($1,$3,$2); }
		| ausdruck OP_EQ ausdruck  { $$ = operate($1,$3,$2); }
		| ausdruck OP_REL ausdruck { $$ = operate($1,$3,$2); }
		| ausdruck OP_SEQ ausdruck { $$ = operate($1,$3,$2); }
		| ausdruck OP_AND ausdruck { $$ = operate($1,$3,$2); }
		| ausdruck OP_OR  ausdruck { $$ = operate($1,$3,$2); }
                | FUNCTION '('  { argumlist.fuegeEin (tmparg, 0);
			          tmparg = new Liste<XplObject*>; }
                               arglist ')' {
#ifdef TTKDEBUG
if (tracedebug)
  cerr << " Enter fun: " << $1->Name();
#endif
logt("Enter fun: "); logt($1->Name());
// isNotEmpty and deleteObj are the only functions which accepts empty objects
//                                           if (($1->Fun() == isNotEmpty) || ($1->Fun() == deleteObj))
                                           if (!$1->needTemps())
       					     $$ = $1->Fun() ($4);
                                           else
                                             if (GenerateTempXplObjects ($4))
					       $$ = ($1->Fun()) ($4);
                                             else
					       $$ = NULL;

//globalError.ShowError ();
                                           long i, length = $4->laenge();
                                           if ($1->Fun() != deleteObj)
					    for (i=0; i<length; i++)
					     {
// cerr << "Fun: " << i << (*tmparg)[i]->WriteProtection () << endl;
                                             if (!(*$4)[i]->WriteProtection ())
					       delete (*$4)[i];
                                             }
					   delete $4;
#ifdef TTKDEBUG
if (tracedebug)
  cerr << " leaving fun: " << $1->Name() << endl;
#endif
logt(" leaving fun: "); logt($1->Name());

                                           tmparg = argumlist[0];
					   argumlist.loesche(0); }
                | MACRO '('  { argumlist.fuegeEin (tmparg, 0);
			       tmparg = new Liste<XplObject*>; }
                   arglist ')'   {
#ifdef TTKDEBUG
if (tracedebug)
  cerr << " Enter mac: " << $1->Name();
#endif
logt("Enter mac: "); logt($1->Name());
#if defined(__BORLANDC__) || defined(SOCKETSERVER) || defined(X4singleserver) // SK 06.04.00
//#ifdef __BORLANDC__    // SK 05.02.99
// pc 20030131
  char *statmess = property.getProperty("statusmessage");
  if (statmess)
  {
      int i = strlen(statmess);
      for(; i > 0; i--)
        statmess[i] = tolower(statmess[i]);
  }
  if ( (statmess == NULL) || (strcmp(statmess, "off")!=0) && (strcmp(statmess, "no")!=0) )
  {
    char msg[256];
    sprintf (msg, XMSG_PARSER_ENTERQUANTLET_STR, $1->Name());
    statusmessage (msg);
  }
  if (statmess) free(statmess);
//  char msg[256];
//  sprintf (msg, "Entering quantlet: %s", $1->Name());
//  statusmessage (msg);
#endif

		                 $$ = ParseMacro ($1, $4);
//globalError.ShowError ();
		                 long i, length = $4->laenge();
				 for (i=0; i<length; i++)
				   {
// cerr << "Mac: " << i << (*tmparg)[i]->WriteProtection () << endl;
                                   if (!(*$4)[i]->WriteProtection ())
				     {
				     delete (*$4)[i];
                                     }
                                   }
				 delete $4;
				 tmparg = argumlist[0];
				 argumlist.loesche(0);
#ifdef TTKDEBUG
if (tracedebug)
  cerr << " leaving mac: " << $1->Name() << endl;
#endif
logt(" leaving mac: "); logt($1->Name());

#if defined(__BORLANDC__) || defined(SOCKETSERVER) || defined(X4singleserver) // SK 06.04.00
//#ifdef __BORLANDC__  // SK 05.02.99
// pc 20030131
  statmess = property.getProperty("statusmessage");
  if (statmess)
  {
      int i = strlen(statmess);
      for(; i > 0; i--)
        statmess[i] = tolower(statmess[i]);
  }
  if ( (statmess == NULL) || (strcmp(statmess, "off")!=0) && (strcmp(statmess, "no")!=0) )
  {
    char msg[256];
    sprintf (msg, XMSG_PARSER_LEAVEQUANTLET_STR, $1->Name());
    statusmessage (msg);
  }
  if (statmess) free(statmess);

//  sprintf (msg, "Leaving quantlet: %s", $1->Name());
//  statusmessage (msg);
#endif
				 if ($$ != NULL)
				   $$->DelWriteProtection ();  }
//                | NEWSTRING '('  { argumlist.fuegeEin (tmparg, 0);
                | MFCAND '('  { argumlist.fuegeEin (tmparg, 0);
			       tmparg = new Liste<XplObject*>; }
                   arglist ')'   {
		                 if (NULL != (mac = GetMacro (macList, $1)))
				   {
#ifdef TTKDEBUG
if (tracedebug)
  cerr << " Enter mfc: " << $1;
#endif
logt("Enter mfc: "); logt($1);

				   $$ = ParseMacro (mac, $4);
//globalError.ShowError ();
				   long i, length = $4->laenge();
				   for (i=0; i<length; i++)
				     {
// cerr << "MFCAND MACRO\t" << $1 << endl;
// cerr << "Mac: " << i << (*tmparg)[i]->WriteProtection () << endl;
                                     if (!(*$4)[i]->WriteProtection ())
				       delete (*$4)[i];
                                     }
				   delete $4;
				   tmparg = argumlist[0];
				   argumlist.loesche(0);
				   if ($$ != NULL)
				     $$->DelWriteProtection ();
#ifdef TTKDEBUG
if (tracedebug)
  cerr << " leaving mfc: " << $1 << endl;
#endif
logt(" leaving mfc: "); logt($1);

				 }
				 else {
globalError.Set (ERR_MAC, $1);
//globalError.ShowError ();
$$ = NULL;
                                 } }
                | REFMAC '('  { argumlist.fuegeEin (tmparg, 0);
			       tmparg = new Liste<XplObject*>; }
                   arglist ')'   {
o = GetXplObject (objList, $1);
if (o == NULL)
  {
#ifdef INETD
cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif
  cerr << "REFMAC Error: REF = " << $1 << endl;
  globalError.Set (ERR_MAC, $1);
  mac = NULL;
  fun = NULL;     // following if will be skipped now
  }
else
if (o->DefaultComponent()->Type() != C_TEXT)
  {
#ifdef INETD
cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif
  cerr << "REFMAC Error: REF = " << $1 << endl;
  globalError.Set (ERR_MAC, "String variable expected");
  mac = NULL;
  fun = NULL;     // following if will be skipped now
  }
else
  {
#ifdef _XPLCDEF_
  text1 = (*o->DefaultComponent()->Content().xplchar)(0).tocstr();
#else
  text1 = o->DefaultComponent()->Content().text;
#endif
//  cerr << "REFMAC text = " << text << endl;
  mac = NULL;
  fun = NULL;
  mac = GetMacro (macList, text1);
  fun = GetFunction (text1);
  }
		                 if ((fun != NULL) || (mac != NULL))
				   {
#ifdef TTKDEBUG
if (tracedebug)
  cerr << " Enter fun/mac: " << text1;
#endif
logt("Enter fun/mac: "); logt(text1);


				   if (mac != NULL)
				     $$ = ParseMacro (mac, $4);
				   else
				     $$ = (fun->Fun ()) ($4);
//globalError.ShowError ();
				   long i, length = $4->laenge();
				   for (i=0; i<length; i++)
				     {
//cerr << "NEWSTRING MACRO\t" << $1 << endl;
// cerr << "Mac: " << i << (*tmparg)[i]->WriteProtection () << endl;
                                     if (!(*$4)[i]->WriteProtection ())
				       delete (*$4)[i];
                                     }
				   delete $4;
				   tmparg = argumlist[0];
				   argumlist.loesche(0);
				   if ($$ != NULL)
				     $$->DelWriteProtection ();
#ifdef TTKDEBUG
if (tracedebug)
  cerr << " leaving fun/mac: " << text1 << endl;
#endif
logt(" leaving fun/mac: "); logt(text1);
				 }
				 else {
globalError.Set (ERR_MAC, $1);
//globalError.ShowError ();
$$ = NULL;
                                 } }
                ;

tsellist	: ausdruck	      { seltmp->haengeAn ($1);   $$ = seltmp; }
		| ','		      { seltmp->haengeAn ((XplObject *)NULL); $$ = seltmp; }
		| tsellist ausdruck   { seltmp->haengeAn ($2);   $$ = seltmp; }
		| tsellist ','	      { seltmp->haengeAn ((XplObject *)NULL); $$ = seltmp; }
		;
sellist         : '[' { sold[soldcounter++] = seltmp;
		        seltmp = new Liste<XplObject*>; }
		      tsellist ']'    { $$ = $3;  seltmp = sold[--soldcounter]; }
                ;

compsellist     : '{' ausdruck '}'    { csel = new Liste <XplObject*>;
					csel->haengeAn ($2);
					$$ = csel; }
		| compsellist '{' ausdruck '}'  {  csel->haengeAn ($3);
						   $$ = csel; }
		;


tmparglist      : /* empty */       { $$ = tmparg; }
                | ausdruck	    { if ($1 != NULL)
				        tmparg->haengeAn ($1);   $$ = tmparg; }
		| tmparglist ',' ausdruck 
                                    { if ($3 != NULL)
				        tmparg->haengeAn ($3);   $$ = tmparg; }
                ;

arglist         : tmparglist              { $$ = $1; }
                ;

carglist	: /* empty */             { $$ = ctmp; }
                | NEWSTRING               { ctmp->haengeAn ($1);  $$ = ctmp; }
// cerr << $1 << " NEWSTRING " << (long) $1 << endl; }
                | VARIABLE                { ctmp->haengeAn ($1->Name());
					    $$ = ctmp; }
// cerr << $1->Name () << " VARIABLE " << (long) $1->Name() << endl; }
		| carglist ',' NEWSTRING  { ctmp->haengeAn ($3);  $$ = ctmp; }
		| carglist ',' VARIABLE   { ctmp->haengeAn ($3->Name());
					    $$ = ctmp; }
// MACROs in carglist moeglich? w'lich nicht TTK, 10.1.99
                | MACRO                   { ctmp->haengeAn ($1->Name());
					    $$ = ctmp; 
cerr << $1->Name () << " MACRO " << (long) $1->Name() << endl; }
		| carglist ',' MACRO      { ctmp->haengeAn ($3->Name());
					    $$ = ctmp; }
		;

parlist		: carglist	          { $$ = $1; ctmp = new Liste<const char*>; }
		;

assignlist	: ASSIGNBRACKET parlist '}' { $$ = $2; }
		;

%%

#include "lex.yy.c"

}  // of extern C


void ExitXpl (int n)
{
  cerr << "ExitXpl ..." << endl;

  remove (xpl4parsefile);
  exit (n);
}

#if !defined(X4standalone) && !defined(X4batch) && !defined(X4singleserver)
      			
#include "xplpp.h"

#if !defined (WIN32) || defined (_CONSOLE)   // Windows Flag for console  TTK, 20.12.98
#ifndef _GRAPHICS_
#ifndef __BORLANDC__
int main (int argc, char **argv)
{
#if !defined (SOCKETSERVER) && !defined (PSGEN)
#ifdef YYDEBUG
  yydebug  = (argc >= 3);        // debug info if dummy parameter is present
#endif
#ifdef TTKDEBUG
  ttkdebug = (argc == 2) || (argc >= 4);
#endif
#endif

#ifdef INETD
cout << "Content-Type: multipart/mixed; boundary=\"-------=_aaaaaaa\"\n\n";
#endif

#ifdef RPC
  X4Start(argc, argv);
  StartRPCServer (argc, argv);
  cerr << "RPC server: Good bye ..." << endl;
  exit (0);
#endif

#ifdef SOCKETSERVER
  X4Start(argc, argv);
  StartSocketServer (argc, argv);
  cerr << "Socket server: Good bye ..." << endl;
  exit (0);
#endif

#if !defined (RPC) && !defined (SOCKETSERVER)
  X4Start(argc, argv);
  StartXploRe (argc, argv);
  cerr << "XploRe: Good bye ..." << endl;
  exit (0);
#endif

  cerr << "should never be seen ..." << endl;
  return (0);
}

#else   // Windows case 
extern int argc;
extern char** argv;
void startxplore (void)
{

  Vector<TokenElement> tl;
  InitializeProperties(argc, argv);          // SK 05.02.99
  InitializeInterpreter (argc, argv);
}
#endif // end of Windows case
#endif     // of _GRAPHICS_
#endif     // WIN32  _CONSOLE

#if defined (WIN32) && !defined (_CONSOLE)
int WINAPI WinMain (HINSTANCE hInst, HINSTANCE hPrecInst, LPSTR cmdline, int argc)
{
// build argc, argc from Windows commandline

  char **argv, *cmd = strdup (cmdline);
  char seps[]   = " ,\t\n";
  char *token;

// translate cmdline into argc, argc format
  token = strtok (cmd, seps);
  argc = 0;

  while (token != NULL)
    {
    argc++;
    token = strtok (NULL, seps);
    }
  
  argv = new char*[argc+1];
  strcpy (cmd, cmdline);

  token = strtok (cmd, seps);

  int i = 0;
  while (token != NULL)
  {
        argv[i++] = token;
    token = strtok (NULL, seps);
  }

  InitializeProperties(argc, argv);

#ifdef RPC
  StartRPCServer (argc, argv);
  cerr << "RPC server: Good bye ..." << endl;
  exit (0);
#endif

#ifdef SOCKETSERVER
  StartSocketServer (argc, argv);
  cerr << "Socket server: Good bye ..." << endl;
  exit (0);
#endif

#if !defined (RPC) && !defined (SOCKETSERVER)
  StartXploRe (argc, argv);
  cerr << "XploRe: Good bye ..." << endl;
  exit (0);
#endif

  cerr << "should never be seen ..." << endl;
  return (0);
}
#endif   // WIN32 && !_CONSOLE

#endif // !defined(X4standalone) && !defined(X4batch) && !defined(X4singleserver)

