/* $Header: /home/xplore/cvs/xpl4gr/xplore1.y,v 1.3 2002/08/16 12:32:01 sigbert Exp $ */

/*
Syntaktische Analyse fuer XploRe (Prototyp fuer XploRe 4 Interpreter):

by TTK 30.6.94  - 6.2.95 (repeat until)

Verarbeitung von Matrizen:

Folgende Operatoren werden unterstuetzt (in dieser Prioritaet):

	Operator	Bedeutung

	~ |		horizontale und vertikale Verkettung
	+ -		Addition, Subtraktion von Matrizen
	.* / *		elementweise Mult. und Division, Matrixmult.
	-		Vorzeichen
	:		Sequenz, nicht assoziativ
	< <= > >=	Relation
	== !=		Vergleich
	|| &&		OR / AND
	
Der Parser versteht folgende Regeln:

	Klammern (), Selektieren[], Funktion auf Matrixelemente,
	Zuweisung =, Vektorliste #(), Argumentliste (Selektor, Funktion)
*/
	
%{
#include "machine.h"
#ifdef HAS_IOSTREAMH
#include <iostream.h>
#include <malloc.h>
#else
#include <stream.h>
#endif

// #pragma implementation

#include "object.h"
#include "function.h"
#include "macro.h"
#include "xplcalc.h"
#include "init.h"

YYSTYPE          yylval;       // needed for mylex !!
Liste<Variable*> vlist;        // stores VARs, needed for assignlist
int              *varlist;     // flag for assigned VARs
int              checkMFcand;  // flag for checking MFCAND
long             l;
char             sbuffer[1024];

extern Vector<Macro*>              *macList;

extern "C" {

int	checkwrap  (void);		// Prototypes necessary here
void	checkerror (const char *);
int	checkparse (void);

%}

%union {
int		    Index;
double		    Real;
char                *Name;
class XplObject     *Obj;
class Variable      *Var;
class Function      *Fun;
class Macro         *Mac;
class MacroText	    *MacText;
Liste<Variable*>    *VarList;
Liste<XplObject*>   *ObjList;
Liste<const char*>  *CharList;
long		    Branch;
}

%pure_parser
%start	program

%token	<Var> VARIABLE
%token	 CONSTVAR
%token	 CONSTANT
%token	 OP_PM OP_MD OP_TMD OP_P OP_REL OP_EQ OP_AND OP_OR OP_SEQ OP_CON 
%token   SINGLE_OPN SINGLE_OPV

%token   DOT COMMENT COMBEGIN COMEND INTERRUPT ASSIGNBRACKET
%token   NEWSTRING STRCONST REFMAC
%token  <Name> MFCAND UNKNOWNINPUT
%token   FUNCTION 
%token   MACRO 

%token   XPROC

%token	 IF ELSE ENDIF WHILE ENDO REPEAT UNTIL ENDP
%token	 SWITCH CASE BREAK DEFAULT ENDSW NOP

%left	  OP_CON
%left	  OP_OR
%left	  OP_AND
%left	  OP_EQ
%left	  OP_REL
%nonassoc OP_SEQ
%left	  OP_PM
%left	  OP_MD
%left	  OP_TMD         //  y*x'*z    '* must have higher priority
%left	  OP_P
%left	  SINGLE_OPN
%left	  SINGLE_OPV
%left	  SIGN


%%

program         : blocks
		| error 
		| procdef
                ;

block		: eingabe
		| IF     '(' ausdruck ')' blocks ENDIF
		| IF     '(' ausdruck ')' blocks ELSE  blocks ENDIF
		| WHILE  '(' ausdruck ')' blocks ENDO
		| REPEAT blocks UNTIL  '(' ausdruck ')' 
		| SWITCH caseblocks ENDSW
		| SWITCH caseblocks defaultblock ENDSW
		;

blocks		: /* empty */
		| blocks block
		;

caseblock	: CASE '(' ausdruck ')' blocks BREAK
		| CASE '(' ausdruck ')' blocks
		;
caseblocks	: /* empty */
		| caseblocks caseblock
		;

defaultblock	: DEFAULT blocks BREAK
		| DEFAULT blocks
		;

procdef		: XPROC parlist '=' NEWSTRING parlist blocks ENDP
		;

eingabe		: '\n'
		| NOP
		| ausdruck
                | assignlist '=' ausdruck
                              { for (int i=0; i<vlist.laenge(); i++)
                                  varlist[vlist[i]->Position()] = TRUE;
                                vlist.remove ();
                              }

		| varstruct '=' ausdruck
		| VARIABLE '=' ausdruck  { l = $1->Position();
                                           if (l >= 0) 
                                             varlist[l] = TRUE; }
		| VARIABLE  compsellist sellist '=' ausdruck       // new  TTK, 13.1.98
                     { l = $1->Position ();
                       if (l >= 0)
                         if (!varlist[l])
                           {
                           if (globalError.Get ())
                             globalError.AppendText("\tCannot assign selected variable ");
                           else
                             globalError.Set(ERR_SYNTAX, "\tCannot assign selected variable ");
                           globalError.AppendText ($1->Name());
  sprintf (sbuffer, " on position %ld in line %ld\n", parselist.Pos(), parselist.CurrentLine());
  globalError.AppendText (sbuffer);
                           }
                     }
		| VARIABLE compsellist '=' ausdruck
                     { l = $1->Position ();
                       if (l >= 0)
                         if (!varlist[l])
                           {
                           if (globalError.Get ())
                             globalError.AppendText("\tCannot assign selected variable ");
                           else
                             globalError.Set(ERR_SYNTAX, "\tCannot assign selected variable ");
                           globalError.AppendText ($1->Name());
  sprintf (sbuffer, " on position %ld in line %ld\n", parselist.Pos(), parselist.CurrentLine());
  globalError.AppendText (sbuffer);
                           }
                     }
		| varstruct sellist '=' ausdruck      // TTK, 21.8.97
		| VARIABLE sellist '=' ausdruck 
                     { l = $1->Position ();
                       if (l >= 0)
                         if (!varlist[l])
                           {
                           if (globalError.Get ())
                             globalError.AppendText("\tCannot assign selected variable ");
                           else
                             globalError.Set(ERR_SYNTAX, "\tCannot assign selected variable ");
                           globalError.AppendText ($1->Name());
  sprintf (sbuffer, " on position %ld in line %ld\n", parselist.Pos(), parselist.CurrentLine());
  globalError.AppendText (sbuffer);
                           }
                     }
		;

ausdruck	: VARIABLE  { l = $1->Position ();
                              if (l >= 0)
                                if (!varlist[l])
                           {
                           if (globalError.Get ())
                             globalError.AppendText("\tCannot access variable ");
                           else
                             globalError.Set(ERR_SYNTAX, "\tCannot access variable ");
                           globalError.AppendText ($1->Name());
  sprintf (sbuffer, " on position %ld in line %ld\n", parselist.Pos(), parselist.CurrentLine());
  globalError.AppendText (sbuffer);
                           } }
		| VARIABLE sellist { l = $1->Position ();
                                     if (l >= 0)
                                       if (!varlist[l])
                           {
                           if (globalError.Get ())
                             globalError.AppendText("\tCannot access selection of variable ");
                           else
                             globalError.Set(ERR_SYNTAX, "\tCannot access selection of variable ");
                           globalError.AppendText ($1->Name());
  sprintf (sbuffer, " on position %ld in line %ld\n", parselist.Pos(), parselist.CurrentLine());
  globalError.AppendText (sbuffer);
                           } }
		| VARIABLE compsellist sellist { }
		| ausdruck compsellist
		| VARIABLE compsellist {}
		| ausdruck sellist
		| CONSTANT
		| CONSTVAR
		| STRCONST
		| varstruct
		| '(' ausdruck ')'
		| OP_PM ausdruck       %prec SIGN
		| SINGLE_OPV ausdruck
		| ausdruck SINGLE_OPN      
		| ausdruck OP_CON ausdruck
		| ausdruck OP_TMD ausdruck
		| ausdruck OP_MD ausdruck 
		| ausdruck OP_PM ausdruck 
		| ausdruck OP_P  ausdruck 
		| ausdruck OP_EQ ausdruck 
		| ausdruck OP_REL ausdruck
		| ausdruck OP_SEQ ausdruck
		| ausdruck OP_AND ausdruck
		| ausdruck OP_OR  ausdruck
                | REFMAC arglist
                | MACRO arglist
                | FUNCTION arglist
                | MFCAND arglist  {
		    if (checkMFcand)
		     {
		    if (GetMacro (macList, $1) == NULL)
                      {
                      if (globalError.Get ())
                        globalError.AppendText("\tCannot find quantlet ");
                      else
                        globalError.Set(ERR_SYNTAX, "\tCannot find quantlet ");
                      globalError.AppendText ($1);
  sprintf (sbuffer, " on position %ld in line %ld\n", parselist.Pos(), parselist.CurrentLine());
  globalError.AppendText (sbuffer);
                      } } }
                ;

tsellist	: ausdruck	   
		| ','		   
		| tsellist ausdruck
		| tsellist ','	   
		;
sellist         : '[' tsellist ']'
                ;

targlist	: /* empty */
                | ausdruck
		| targlist ',' ausdruck 
                ;

arglist         : '(' targlist ')'
                ;

tparlist	: /* empty */
                | VARIABLE                { vlist.haengeAn ($1); }
		| tparlist ',' VARIABLE   { vlist.haengeAn ($3); }
		;

parlist		: '(' tparlist ')'
		;

assignlist	: ASSIGNBRACKET tparlist '}'
		;

compsellist     : '{' ausdruck '}'
		| compsellist '{' ausdruck '}'
		;

varstruct	: VARIABLE DOT VARIABLE   {}
                | VARIABLE DOT FUNCTION   {}
                | VARIABLE DOT MACRO      {}
		| varstruct DOT VARIABLE
		| varstruct DOT FUNCTION
		| varstruct DOT MACRO
		;


%%

}  // of extern C

      			
void checkerror (const char *s)
/* needed by the yacc parser, prints out an error string */
{
  extern GlobalList parselist;
  char buffer[1024];

  if (globalError.Get ())
    globalError.AppendText("Parse error ");
  else
    globalError.Set(ERR_SYNTAX, "Parse Error ");

  sprintf (buffer, " on position %ld in line %ld\n", parselist.Pos(), parselist.CurrentLine());
  globalError.AppendText (buffer);
  
//  parselist.PrintCurrentToken ();
}
