/* $Header: /home/xplore/cvs/xpl4gr/xplorematnew.y,v 1.3 2002/08/16 12:32:01 sigbert Exp $ */

/*
Syntaktische Analyse fuer XploRe (Prototyp fuer XploRe 4 Interpreter):

by TTK 30.6.94  - 6.2.95 (repeat until)

Verarbeitung von Matrizen:

Folgende Operatoren werden unterstuetzt (in dieser Prioritaet):

	Operator	Bedeutung

	~ |		horizontale und vertikale Verkettung
	+ -		Addition, Subtraktion von Matrizen
	.* / *		elementweise Mult. und Division, Matrixmult.
	-		Vorzeichen
	:		Sequenz, nicht assoziativ
	< <= > >=	Relation
	== !=		Vergleich
	|| &&		OR / AND
	
Der Parser versteht folgende Regeln:

	Klammern (), Selektieren[], Funktion auf Matrixelemente,
	Zuweisung =, Vektorliste #(), Argumentliste (Selektor, Funktion)
*/
	
%{
#include <stream.h>
#include "liste.h"
#include "object.h"
#include "function.h"
#include "macro.h"
#include "xplcalc.h"
#include "xplpp.h"
#include "init.h"

YYSTYPE yylval;		// needed for mylex !!

int errorToken = 0;     // for error state

XplDouble *x;
XplObject    *o;
Macro     *mac;

long ifhandler (XplObject*, int);
long whilehandler (XplObject*, int);
long untilhandler (XplObject*, int);

Liste<XplObject*>             *objList;          // globals
Liste<Function*>           *funList;
Liste<Macro*>              *macList;
class GlobalList           *newgl;

Liste<Liste<XplObject*>*>   argumlist;         // global in xplore.y
Liste<XplObject*>           *tmparg, *seltmp;
Liste<char*>		   *ctmp;
Liste<XplDouble*>	 *xpldoublelist;

Liste<double>		   tmp;


extern "C" {
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <math.h>

int	yywrap (void);		// Prototypes necessary here
void	yyerror (const char *);
int	yyparse (void);
int	mylex (void);

%}


%union {
int		  Index;
double		  Real;
char              *Name;
class XplObject  	  *Obj;
class Variable    *Var;
class Function    *Fun;
class Macro       *Mac;
class MacroText	  *MacText;
Liste<double>	  *List;
Liste<Variable*>  *VarList;
Liste<XplObject*>    *ObjList;
Liste<char*>      *CharList;
long		  Branch;
}
  /* for IF, ELSE, ENDIF, REPEAT, UNTIL */
  /*     WHILE, ENDWHILE */


%pure_parser
%start	program

%token	<Var>   VARIABLE
%token	<Real>  CONST
%token	<Index> OP_PM OP_MD OP_REL OP_EQ OP_AND OP_OR OP_SEQ OP_CON SINGLE_OP
%token  <Index> DOT
%token  <Name>  NEWSTRING STRCONST
%token  <Fun>   FUNCTION 
%token  <Mac>   MACRO 

%token  <MacText>       PROC

%token	<Index> IF ELSE ENDIF WHILE ENDO REPEAT UNTIL ENDP

%type   <Branch>        block         // probably not necessary

%type	<Obj>		constvec expr
%type   <List>          list
%type   <ObjList>       arglist tmparglist tsellist sellist
%type   <CharList>      carglist parlist

%left	OP_CON
%left	OP_OR
%left	OP_AND
%left	OP_EQ
%left	OP_REL
%left	OP_PM
%left	OP_MD
%nonassoc OP_SEQ
%left	SIGN
%left	SINGLE_OP


%%

program         :  /* empty */
                | program block
		| error { if (!errorToken)        // get the FIRST errTok !
			    {
			    cerr << "Error: " << yychar << endl; 
			    errorToken = yychar;
			    }
			}
                  '\n'	{ yyerrok; PrintToken (errorToken); 
			  parselist.PrintCurrentToken ();
			  errorToken = 0; }
                ;

block		: input         {}  
		| block input   {}
		| IF '(' expr ')'            {  $$ = ifhandler($3, $1); }
		| WHILE '(' expr ')'         {  $$ = whilehandler($3, $1); }
		| UNTIL '(' expr ')'         {  $$ = untilhandler($3, $1); }
		| PROC  parlist  '=' NEWSTRING  parlist 
		 { newgl = CopyTokenList();
		   if (newgl != NULL)
		     {
		     mac = new Macro($4, $1->Text(), *$5, *$2, newgl);
		     macList->haengeAn (mac);
		     mac->PrepareToRun ();
		     }
		   else
		     cerr << "Error in Definition of " << endl;
		   }
		| PROC  parlist  '=' MACRO  parlist 
		 { $4->SetNewDefinition ($1->Text(), *$5, *$2, CopyTokenList());
		   }
		;
input		: expr		{ if ($1 != NULL)
				    {
				    $1->Print(); 
				    if ($1->DefaultXplDouble () != NULL)
				      $1->DefaultXplDouble().dump("%lf");
			if ($1->DefaultComponent()->Type() != C_PLOT)
			              if (! $1->WriteProtection ())
					delete $1;
				     }
				}
                | '\n'
		| NEWSTRING '=' expr  { if ($3 != NULL)
					  {
					  if ($3->WriteProtection ())
					    objList->fuegeEin (new XplObject ($3, $1));
					  else
					    {
					    $3->SetName ($1);
					    objList->fuegeEin ($3);
					    }
				      }
//					    else error
					  }
		| VARIABLE '=' expr   { if ($3 != NULL)
					      {
                      o = $1->GetXplObject ();
                      o->DelWriteProtection ();          // allow deletion of all components
                      o->TransferComponents ($3);
                      o->SetWriteProtection ();          // protect XplObject
                      GetXplObject(objList, $1->Name())->SetDefaultXplDouble ();
                      if (! $3->WriteProtection ())
		        delete $3;
					      }
//					    else error
					  }
		| VARIABLE  sellist '=' expr	
		         { xpldoublelist = MakeArglist($2);
			   $2->remove();
			   delete $2;
			   x = $4->DefaultXplDouble ();
         $1->GetXplObject()->DefaultXplDouble()->selectassign(*xpldoublelist, x);
			   delete $4;
			   xpldoublelist->remove();
			   delete xpldoublelist; }
		;


list		: CONST			{ tmp.fuegeEin($1,0); $$ = &tmp; }
		| OP_PM CONST 		{ if ($1 == MINUS)
					    tmp.fuegeEin(-$2,0);
					  else
					    tmp.fuegeEin($2,0);
					  $$ = &tmp; }
		| list ',' CONST	{ tmp.fuegeEin($3,0); }
		| list ',' OP_PM CONST	{ if ($3 == MINUS)
					    tmp.fuegeEin(-$4,0);
					  else
					    tmp.fuegeEin($4,0);
					  $$ = &tmp; }
		;
		
constvec	: '#' '(' list ')'	{ x = new XplDouble ($3->laenge());
					  for (int j=0, i=$3->laenge(); i>0;)
					    (*x)(--i) = (*$3)[j++];
					  $3->remove();
                           		  $$ = new XplObject(x);
                                        }
                ;

expr	        : constvec		{ $$ = $1; }
		| VARIABLE      { o = $1->GetXplObject();
                                  if (o->isEmpty ())
                                    {
                                    yyerror ("empty Var");
cerr << "Couldn't access " << o->Name() << endl;
parselist.SetPos (parselist.GetTokenList()->laenge()-1);
                                    YYERROR;
                                    }
                                  $$ = o;  }
		| VARIABLE sellist    { xpldoublelist = MakeArglist($2);
					       $2->remove();
					       delete $2;
    $$ = new XplObject ($1->GetXplObject()->DefaultXplDouble()->selection (*xpldoublelist));
// werden alle Matrizen geloescht?
						  xpldoublelist->remove();
						  delete xpldoublelist; }
		| expr sellist       { xpldoublelist = MakeArglist($2);
				       $2->remove();
				       delete $2;
    $$ = new XplObject ($1->DefaultXplDouble()->selection (*xpldoublelist));
// werden alle Matrizen geloescht?
                                       xpldoublelist->remove();
				       delete xpldoublelist; }
		| CONST		{ x = new XplDouble (1);
				  (*x)(0) = $1;
				  $$ = new XplObject (x); }
		| STRCONST	{ $$ = new XplObject (new Component ($1)); }
		| '(' expr ')'		{ $$ = $2; }
		| expr SINGLE_OP       { $$ = operate($1,$2); }
		| expr OP_CON expr { $$ = operate($1,$3,$2); }
		| expr OP_MD expr  { $$ = operate($1,$3,$2); }
		| expr OP_PM expr  { $$ = operate($1,$3,$2); }
		| expr OP_EQ expr  { $$ = operate($1,$3,$2); }
		| expr OP_REL expr { $$ = operate($1,$3,$2); }
		| expr OP_SEQ expr { $$ = operate($1,$3,$2); }
		| expr OP_AND expr { $$ = operate($1,$3,$2); }
		| expr OP_OR  expr { $$ = operate($1,$3,$2); }
                | FUNCTION '('      { argumlist.fuegeEin (tmparg, 0);
				      tmparg = new Liste<XplObject*>; }
                               arglist ')' {
GenerateTempXplObjects ($4);
				           $$ = ($1->Fun ()) ($4);
                                           long i, length = tmparg.laenge();
						  for (i=0; i<length; i++)
						    {
// cerr << "Fun: " << i << (*tmparg)[i]->WriteProtection () << endl;
if (!(*tmparg)[i]->WriteProtection ())
						    delete (*tmparg)[i];
}
						  delete tmparg;
                                           tmparg = argumlist[0];
					   argumlist.loesche(0); }
                | MACRO '('      { argumlist.fuegeEin (tmparg, 0);
				   tmparg = new Liste<XplObject*>; }
                               arglist ')'      {
						$$ = ParseMacro ($1, $4);
			                          long i, length = tmparg.laenge();
						  for (i=0; i<length; i++)
                                                    {
// cerr << "Mac: " << i << (*tmparg)[i]->WriteProtection () << endl;
if (!(*tmparg)[i]->WriteProtection ())
						    delete (*tmparg)[i];
                                                    }
						  delete tmparg;
						  tmparg = argumlist[0];
						  argumlist.loesche(0);
					          $$->DelWriteProtection ();  }
                ;

tsellist	: expr		{ seltmp->haengeAn ($1);   $$ = seltmp; }
		| ','			{ seltmp->haengeAn (NULL); $$ = seltmp; }
		| tsellist expr	{ seltmp->haengeAn ($2);   $$ = seltmp; }
		| tsellist ','		{ seltmp->haengeAn (NULL); $$ = seltmp; }
		;
sellist         : '[' tsellist ']'   { $$ = $2;  seltmp = new Liste<XplObject *>;  }
                ;

tmparglist      : expr	          { if ($1 != NULL)
					      tmparg->haengeAn ($1);   $$ = tmparg; }
		| tmparglist ',' expr { if ($3 != NULL)
					      tmparg->haengeAn ($3);   $$ = tmparg; }
                ;

arglist         : tmparglist   { $$ = $1; }
                ;

carglist	: NEWSTRING               { ctmp->haengeAn ($1);  $$ = ctmp; }
                | VARIABLE                { ctmp->haengeAn ($1->Name());
					    $$ = ctmp; }
		| carglist ',' NEWSTRING  { ctmp->haengeAn ($3);  $$ = ctmp; }
		| carglist ',' VARIABLE   { ctmp->haengeAn ($3->Name());
					    $$ = ctmp; }
                | MACRO                   { ctmp->haengeAn ($1->Name());
					    $$ = ctmp; }
		| carglist ',' MACRO      { ctmp->haengeAn ($3->Name());
					    $$ = ctmp; }
		;

parlist		: '(' carglist ')'        { $$ = $2; ctmp = new Liste<char*>; }
		;


%%


#include "lex.yy.c"

}  // of extern C

      			
#include "xplpp.h"

#ifndef _GRAPHICS_
int main (int argc, char **argv)
{
  Liste<TokenElement> tl;

#ifdef YYDEBUG
  yydebug  = (argc >= 3);        // debug info if dummy parameter is present
#endif
#ifdef TTKDEBUG
extern int ttkdebug;
  ttkdebug = (argc == 2) || (argc >= 4);
#endif

  InitializeInterpreter ();  

while (1)
  {

  BuildTokenList (&parselist);
ScanFunsAndMacs (objList, &parselist);
ScanVariables (objList, &parselist);
  InputBranch (&parselist, 0L);
  ShowTokenList (&parselist);
  ParseTokenList (parselist);
  DeleteVariables (&parselist);
  parselist.GetTokenList()->remove();
  }

return (0);
}
#endif     // of _GRAPHICS_

long ifhandler (XplObject *m, int jump)
{
  if (! (*(m->DefaultXplDouble()))(0,0))
    parselist.SetJump (jump + 1);

  delete m;
  return 0L;
}

long whilehandler (XplObject *m, int jump)
{
  if (! (*(m->DefaultXplDouble()))(0,0))
    parselist.SetJump (jump + 1);

  delete m;
  return 0L;
}

long untilhandler (XplObject *m, int jump)
{
  if (! (*(m->DefaultXplDouble()))(0,0))
    parselist.SetJump (jump + 1);

  delete m;
  return 0L;
}

