/* $Header: /home/xplore/cvs/xpl4gr/xplorexa.y,v 1.3 2002/08/16 12:32:02 sigbert Exp $ */

/*
Syntaktische Analyse fuer XploRe (Prototyp fuer XploRe 4 Interpreter):

by TTK 30.6.94  - 6.2.95 (repeat until)

Verarbeitung von Matrizen:

Folgende Operatoren werden unterstuetzt (in dieser Prioritaet):

	Operator	Bedeutung

	~ |		horizontale und vertikale Verkettung
	+ -		Addition, Subtraktion von Matrizen
	.* / *		elementweise Mult. und Division, Matrixmult.
	-		Vorzeichen
	:		Sequenz, nicht assoziativ
	< <= > >=	Relation
	== !=		Vergleich
	|| &&		OR / AND
	
Der Parser versteht folgende Regeln:

	Klammern (), Selektieren[], Funktion auf Matrixelemente,
	Zuweisung =, Vektorliste #(), Argumentliste (Selektor, Funktion)
*/
	
%{
#include <stream.h>
#include "liste.h"
#include "object.h"
#include "function.h"
#include "macro.h"
#include "xplcalc.h"
#include "xplpp.h"
#include "init.h"

#ifdef TTKDEBUG
extern int ttkdebug;
#endif

YYSTYPE yylval;		// needed for mylex !!

int errorToken = 0;     // for error state

int          nsel, switchcount = 0;
XplSelection *xplsel;
XplDouble    *x, *x1;
XplObject    *o, *obj;
Macro        *mac;
Function     *fun;
char         *text;

#ifdef _XPLCDEF_
XplChar      xplc(1);
#endif

long casehandler   (XplObject *, int);
long ifhandler     (XplObject *, int);
long whilehandler  (XplObject *, int);
long untilhandler  (XplObject *, int);
long switchhandler (XplObject *, int);

void ExitXpl       (int);

Liste<XplObject*>          *constList, *objList, *globObjList;   // globals
Liste<Function*>           *funList;
Liste<Macro*>              *macList;
class GlobalList           *newgl;

Liste<Liste<XplObject*>*>   argumlist;         // global in xplore.y
Liste<XplObject*>           *tmparg, *seltmp, *csel;
Liste<const char*>	    *ctmp;
Liste<XplDouble*>	    *xpldoublelist;


extern "C" {
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <math.h>

int	yywrap  (void);		// Prototypes necessary here
void	yyerror (const char *);
int	yyparse (void);
int	mylex   (void);

%}


%union {
int		    Index;
double		    Real;
char                *Name;
class XplObject     *Obj;
class Variable      *Var;
class Function      *Fun;
class Macro         *Mac;
class MacroText	    *MacText;
Liste<Variable*>    *VarList;
Liste<XplObject*>   *ObjList;
Liste<const char*>  *CharList;
long		    Branch;
}
  /* for IF, ELSE, ENDIF, REPEAT, UNTIL */
  /*     WHILE, ENDWHILE */


%pure_parser
%start	program

%token	<Var>   VARIABLE
%token	<Obj>   CONSTVAR
%token	<Real>  CONST
%token	<Index> OP_PM OP_MD OP_TMD OP_P OP_REL OP_EQ OP_AND OP_OR OP_SEQ OP_CON SINGLE_OP
%token  <Index> DOT COMMENT INTERRUPT
%token  <Name>  NEWSTRING STRCONST REFMAC MFCAND
%token  <Fun>   FUNCTION 
%token  <Mac>   MACRO 

%token  <MacText>       PROC

%token	<Index> IF ELSE ENDIF WHILE ENDO REPEAT UNTIL ENDP
%token	<Index> SWITCH CASE BREAK DEFAULT ENDSW

%type   <Branch>        block         // probably not necessary

%type	<Obj>		ausdruck
%type   <ObjList>       arglist tmparglist tsellist sellist compsellist
%type   <CharList>      carglist parlist assignlist

%left	  OP_CON
%left	  OP_OR
%left	  OP_AND
%left	  OP_EQ
%left	  OP_REL
%nonassoc OP_SEQ
%left	  OP_PM
%left	  OP_MD
%left	  OP_TMD         //  y*x'*z    '* must have higher priority
%right	  OP_P
%left	  SINGLE_OP
%left	  SIGN


%%

program         :  /* empty */
                | program block
		| error { if (!errorToken)        // get the FIRST errTok !
			    {
#ifdef INETD
cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif
			    cerr << "Error: " << yychar << endl; 
			    errorToken = yychar;
			    }
			}
                  '\n'	{ yyerrok; PrintToken (errorToken); 
			  parselist.PrintCurrentToken ();
//globalError.ShowError ();
//			  mt.SetNewText();       // if error with proc def
			  errorToken = 0; }
                ;

block		: eingabe         {}  
		| INTERRUPT       { YYERROR; }
		| block eingabe   {}
		| IF     '(' ausdruck ')'      {  $$ = ifhandler     ($3, $1); }
		| WHILE  '(' ausdruck ')'      {  $$ = whilehandler  ($3, $1); }
		| UNTIL  '(' ausdruck ')'      {  $$ = untilhandler  ($3, $1); }
		| SWITCH '(' ausdruck ')'      {  $$ = switchhandler ($3, $1); }
		| CASE   '(' ausdruck ')'      {  $$ = casehandler   ($3, $1); }
		| PROC '(' parlist ')' '=' NEWSTRING '(' parlist ')'
		 { newgl = CopyTokenList();
		   if (GetXplObject (objList, $6) == NULL)
		     {
		     if (newgl != NULL)
		       {
		       mac = new Macro($6, $1->Text(), *$8,*$3, newgl);
		       macList->haengeAn (mac);
		       mac->PrepareToRun ();
		       delete $1;           // MacroText must be deleted
		       }
		     else
		       {
#ifdef INETD
cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif
		       cerr << "Error in Definition of " << $6 << endl;
                       }
		     }
		   else
		     {
#ifdef INETD
cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif
		     cerr << $6 << " is already Variable" << endl;
                     globalError.Set (ERR_INTERN, "Variable cannot be macro");
		     if (newgl != NULL)
		       delete newgl;           // what is with list elements?
		     }
		   }
		| PROC '(' parlist ')' '=' MACRO '(' parlist ')'
		 { $6->SetNewDefinition ($1->Text(), *$8, *$3, CopyTokenList());
		   delete $1;             // MacroText must be deleted
		   }
		;
eingabe		: ausdruck  { if ($1 != NULL)
				{
				$1->Print(); 
				if (!$1->isEmpty ())
				  {
				  if ($1->DefaultXplDouble () != NULL)
				    {
#if !defined (_GRAPHICS_) && !defined (INETD)
				    $1->DefaultXplDouble()->mdump("%lf", cout);
#endif
				    }
				  if ($1->DefaultComponent()->Type() != C_PLOT)
				    if (! $1->WriteProtection ())
				      delete $1;
				  }
			        }
			    }
                | '\n'
                | assignlist '=' ausdruck { if ($3 == NULL)
					      {
if (globalError.GetNumber () != ERR_INTR)      // error only if not interrupted
  {
#ifdef INETD
cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif
  cerr << "Error in list assignment: no Variables generated"<<endl;
  }
}
					    else
					      {
                                              assign ($1, $3);
					      if (!$3->WriteProtection ())
						delete $3;
					      }
//globalError.ShowError ();
                                          }
		| NEWSTRING '=' ausdruck  { if ($3 != NULL)
					      {
					      o = GetXplObject (objList, $1);
					      if (o != NULL)
						{
                      o->DelWriteProtection ();          // allow deletion of all components
                      o->TransferComponents ($3);
                      o->SetWriteProtection ();          // protect XplObject
                      GetXplObject(objList, $1)->SetDefaultXplDouble ();
		    }
					      else
						if ($3->WriteProtection ())
						  objList->haengeAn (new XplObject ($3, $1));
						else
						  {
						  $3->SetName ($1);
						  objList->haengeAn ($3);
						  }
					      }
					    else 
					      {
#ifdef INETD
cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif
cerr << "Error in assignment: Variable " << $1 << " not generated"<<endl;
                                              }
					  }
		| VARIABLE '=' ausdruck   { if ($3 != NULL)
					      {
                      o = $1->GetXplObject ();
                      o->DelWriteProtection ();          // allow deletion of all components
                      o->TransferComponents ($3);
                      o->SetWriteProtection ();          // protect XplObject
                      GetXplObject(objList, $1->Name())->SetDefaultXplDouble ();
                      if (! $3->WriteProtection ())
		        delete $3;
					      }
					    else
					      {
#ifdef INETD
cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif
cerr << "Error in assignment: Variable " << $1->Name() << " not changed"<<endl;
                                              }
					  }

		| VARIABLE  sellist '=' ausdruck  {
		    x = $1->GetXplObject()->DefaultXplDouble();
                    if (x == NULL)
                      globalError.Set (ERR_INTERN, "Var [] doen't exist ");
                    else
		      {                      
		      xplsel = GenerateXplSelection (x, nsel, MakeArglist($2));
		      if (xplsel == NULL)
		        {
#ifdef INETD
cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif
                        cerr << "ERROR: Selector for " << $1->Name () << endl;
                        }
		      else
		        x->assignment (nsel, xplsel, *($4->DefaultXplDouble ()));
		      $2->remove();
		      delete $2;
		      if (! $4->WriteProtection ())
		        delete $4;
                      }  }
		| VARIABLE  compsellist '=' ausdruck  {
		    obj = $1->GetXplObject ();
		    if (obj != NULL)
		      {
		      o = SelectComponentList (obj, $2);
		      AssignSelection (o, $4);
		      long i, length = $2->laenge();
		      for (i=0; i<length; i++)
			if (! (*$2)[i]->WriteProtection ())
			  delete (*$2)[i];
		      delete $2;
		      delete $4;
		      obj->SetDefaultXplDouble ();
		      }
		    else
                      globalError.Set (ERR_INTERN, "empty obejct in slc");
		    }
		;

ausdruck	: VARIABLE      { o = $1->GetXplObject ();
                                  if (o->isEmpty ())
                                    {
globalError.Set (ERR_INTERN, "empty Var ");
globalError.AppendText (ERR_INTERN, o->Name());
//                                   yyerror ("empty Var");
// cerr << "Couldn't access " << o->Name() << endl;
parselist.SetPos (parselist.GetTokenList()->laenge()-1);
                                    YYERROR;
                                    }
                                  $$ = o;  }
		| CONSTVAR      { $$ = $1; }
		| VARIABLE sellist    { 
		  x = $1->GetXplObject()->DefaultXplDouble();
                    if (x == NULL)
                      globalError.Set (ERR_INTERN, "Var [] doen't exist ");
                    else
		      {                      
		      x1 = new XplDouble;
		      xplsel = GenerateXplSelection (x, nsel, MakeArglist($2));
		      $2->remove();
		      delete $2;

		      if (xplsel == NULL)
		        {
#ifdef INETD
cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif
                        cerr << "ERROR: Selector for " << $1->Name () << endl;
		        $$ = NULL;
                        }
		      else
			{
			x1->selection (*x, nsel, xplsel);
			$$ = new XplObject (x1);
		        }
		      }
// werden alle Matrizen geloescht?
		}
		| ausdruck sellist    { 
		  if ($1 != NULL)
		    {
		    x = $1->DefaultXplDouble();
		    x1 = new XplDouble;
		    xplsel = GenerateXplSelection (x, nsel, MakeArglist($2));
		    $2->remove();
		    delete $2;

		    if (xplsel == NULL)
		      {
#ifdef INETD
cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif
		      cerr << "ERROR: Selector for term" << endl;
		      $$ = NULL;
		      }
		    else
		      {
		      x1->selection (*x, nsel, xplsel);
		      $$ = new XplObject (x1);
		      }
		    }
		  else
		    {
		    globalError.Set (ERR_INTERN, "selection of empty arg");
		    $$ = NULL;
		    }  }
		| ausdruck compsellist
                           { o = SelectComponentList ($1, $2);
			     if (o != NULL)
			       {
			       if (! $1->WriteProtection ())
				 {
				 $$ = new XplObject("csel");
				 $$->TransferXplObject (o);
				 delete $1;
			         }
			       else
				 $$ = new XplObject (o);
			       }
			     else
			       {
			       if (! $1->WriteProtection ())
				 delete $1;
		    globalError.Set (ERR_INTERN, "selection{} failed");
			       $$ = NULL;
			       }
			     long i, length = $2->laenge ();
			     for (i=0; i<length; i++)
			       if (! (*$2)[i]->WriteProtection ())
				 delete (*$2)[i];
			   }
		| VARIABLE compsellist
                           { obj = $1->GetXplObject ();
                             o   = SelectComponentList (obj, $2);
			     if (o != NULL)
			       {
			       if (! obj->WriteProtection ())
				 {
				 $$ = new XplObject("csel");
				 $$->TransferXplObject (o);
				 delete obj;
			         }
			       else
				 $$ = new XplObject (o);
			       }
			     else
			       {
		    globalError.Set (ERR_INTERN, "selection{} failed");
			       $$ = NULL;
			       }
			     long i, length = $2->laenge ();
			     for (i=0; i<length; i++)
			       if (! (*$2)[i]->WriteProtection ())
				 delete (*$2)[i];
			   }
		| CONST	   { x = new XplDouble (1);
				  (*x)(0) = $1;
				  $$ = new XplObject (x); }
		| STRCONST {
		           text = new char[strlen($1)];
			   sprintf (text, "%s", $1);
#ifdef _XPLCDEF_
			   xplc(0) = xplstring(text);
			   $$ = new XplObject (new Component (new XplChar (xplc)));
#else
                           $$ = new XplObject (new Component (text));
#endif
			   }
		| '(' ausdruck ')'		{ $$ = $2; }
		| OP_PM ausdruck  %prec SIGN	{ if ($1 == MINUS)
						    $$ = operate($2, $1);
						  else
						    $$ = $2; }
		| '!' ausdruck  %prec SIGN { $$ = operate($2,NOT); }
		| ausdruck SINGLE_OP       { $$ = operate($1,$2); }
		| ausdruck OP_CON ausdruck { $$ = operate($1,$3,$2); }
		| ausdruck OP_TMD ausdruck { $$ = operate($1,$3,$2); }
		| ausdruck OP_MD ausdruck  { $$ = operate($1,$3,$2); }
		| ausdruck OP_PM ausdruck  { $$ = operate($1,$3,$2); }
		| ausdruck OP_P  ausdruck  { $$ = operate($1,$3,$2); }
		| ausdruck OP_EQ ausdruck  { $$ = operate($1,$3,$2); }
		| ausdruck OP_REL ausdruck { $$ = operate($1,$3,$2); }
		| ausdruck OP_SEQ ausdruck { $$ = operate($1,$3,$2); }
		| ausdruck OP_AND ausdruck { $$ = operate($1,$3,$2); }
		| ausdruck OP_OR  ausdruck { $$ = operate($1,$3,$2); }
                | FUNCTION '('  { argumlist.fuegeEin (tmparg, 0);
			          tmparg = new Liste<XplObject*>; }
                               arglist ')' {
                                           GenerateTempXplObjects ($4);
				           $$ = ($1->Fun ()) ($4);
//globalError.ShowError ();
                                           long i, length = tmparg->laenge();
					   for (i=0; i<length; i++)
					     {
// cerr << "Fun: " << i << (*tmparg)[i]->WriteProtection () << endl;
                                             if (!(*tmparg)[i]->WriteProtection ())
					       delete (*tmparg)[i];
                                             }
					   delete tmparg;
                                           tmparg = argumlist[0];
					   argumlist.loesche(0); }
                | MACRO '('  { argumlist.fuegeEin (tmparg, 0);
			       tmparg = new Liste<XplObject*>; }
                   arglist ')'   {
		                 $$ = ParseMacro ($1, $4);
//globalError.ShowError ();
		                 long i, length = tmparg->laenge();
				 for (i=0; i<length; i++)
				   {
// cerr << "Mac: " << i << (*tmparg)[i]->WriteProtection () << endl;
                                   if (!(*tmparg)[i]->WriteProtection ())
				     delete (*tmparg)[i];
                                   }
				 delete tmparg;
				 tmparg = argumlist[0];
				 argumlist.loesche(0);
				 if ($$ != NULL)
				   $$->DelWriteProtection ();  }
//                | NEWSTRING '('  { argumlist.fuegeEin (tmparg, 0);
                | MFCAND '('  { argumlist.fuegeEin (tmparg, 0);
			       tmparg = new Liste<XplObject*>; }
                   arglist ')'   {
		                 if (NULL != (mac = GetMacro (macList, $1)))
				   {
				   $$ = ParseMacro (mac, $4);
//globalError.ShowError ();
				   long i, length = tmparg->laenge();
				   for (i=0; i<length; i++)
				     {
// cerr << "MFCAND MACRO\t" << $1 << endl;
// cerr << "Mac: " << i << (*tmparg)[i]->WriteProtection () << endl;
                                     if (!(*tmparg)[i]->WriteProtection ())
				       delete (*tmparg)[i];
                                     }
				   delete tmparg;
				   tmparg = argumlist[0];
				   argumlist.loesche(0);
				   if ($$ != NULL)
				     $$->DelWriteProtection ();
				 }
				 else {
globalError.Set (ERR_MAC, $1);
//globalError.ShowError ();
$$ = NULL;
                                 } }
                | REFMAC '('  { argumlist.fuegeEin (tmparg, 0);
			       tmparg = new Liste<XplObject*>; }
                   arglist ')'   {
o = GetXplObject (objList, $1);
if (o == NULL)
  {
#ifdef INETD
cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif
  cerr << "REFMAC Error: REF = " << $1 << endl;
  cerr << "exiting ..." << endl;
  exit (1);
  }
#ifdef _XPLCDEF_
text = (*o->DefaultComponent()->Content().xplchar)(0).tostr();
#else
text = o->DefaultComponent()->Content().text;
#endif
//  cerr << "REFMAC text = " << text << endl;
mac = NULL;
fun = NULL;
mac = GetMacro (macList, text);
fun = GetFunction (text);
		                 if ((fun != NULL) || (mac != NULL))
				   {
				   if (mac != NULL)
				     $$ = ParseMacro (mac, $4);
				   else
				     $$ = (fun->Fun ()) ($4);
//globalError.ShowError ();
				   long i, length = tmparg->laenge();
				   for (i=0; i<length; i++)
				     {
//cerr << "NEWSTRING MACRO\t" << $1 << endl;
// cerr << "Mac: " << i << (*tmparg)[i]->WriteProtection () << endl;
                                     if (!(*tmparg)[i]->WriteProtection ())
				       delete (*tmparg)[i];
                                     }
				   delete tmparg;
				   tmparg = argumlist[0];
				   argumlist.loesche(0);
				   if ($$ != NULL)
				     $$->DelWriteProtection ();
				 }
				 else {
globalError.Set (ERR_MAC, $1);
//globalError.ShowError ();
$$ = NULL;
                                 } }
                ;

tsellist	: ausdruck	      { seltmp->haengeAn ($1);   $$ = seltmp; }
		| ','		      { seltmp->haengeAn (NULL); $$ = seltmp; }
		| tsellist ausdruck   { seltmp->haengeAn ($2);   $$ = seltmp; }
		| tsellist ','	      { seltmp->haengeAn (NULL); $$ = seltmp; }
		;
sellist         : '[' tsellist ']' { $$ = $2; seltmp = new Liste<XplObject*>; }
                ;

compsellist     : '{' ausdruck '}'    { csel = new Liste <XplObject*>;
					csel->haengeAn ($2);
					$$ = csel; }
		| compsellist '{' ausdruck '}'  {  csel->haengeAn ($3);
						   $$ = csel; }
		;


tmparglist      : /* empty */       { $$ = tmparg; }
                | ausdruck	    { if ($1 != NULL)
				        tmparg->haengeAn ($1);   $$ = tmparg; }
		| tmparglist ',' ausdruck 
                                    { if ($3 != NULL)
				        tmparg->haengeAn ($3);   $$ = tmparg; }
                ;

arglist         : tmparglist              { $$ = $1; }
                ;

carglist	: /* empty */             { $$ = ctmp; }
                | NEWSTRING               { ctmp->haengeAn ($1);  $$ = ctmp; }
                | VARIABLE                { ctmp->haengeAn ($1->Name());
					    $$ = ctmp; }
		| carglist ',' NEWSTRING  { ctmp->haengeAn ($3);  $$ = ctmp; }
		| carglist ',' VARIABLE   { ctmp->haengeAn ($3->Name());
					    $$ = ctmp; }
                | MACRO                   { ctmp->haengeAn ($1->Name());
					    $$ = ctmp; }
		| carglist ',' MACRO      { ctmp->haengeAn ($3->Name());
					    $$ = ctmp; }
		;

parlist		: carglist	          { $$ = $1; ctmp = new Liste<const char*>; }
		;

assignlist	: '{' parlist '}'         { $$ = $2; }
		;

%%


#include "lex.yy.c"

}  // of extern C

      			
#include "xplpp.h"

#ifndef _GRAPHICS_
int main (int argc, char **argv)
{
  Liste<TokenElement> tl;
  char cmd[100000];

#ifdef YYDEBUG
  yydebug  = (argc >= 3);        // debug info if dummy parameter is present
#endif
#ifdef TTKDEBUG
  ttkdebug = (argc == 2) || (argc >= 4);
#endif

#ifdef INETD
cout << "Content-Type: multipart/mixed; boundary=\"-------=_aaaaaaa\"\n\n";
#endif


  InitializeInterpreter ();  
while (1)
  {
  fgets ( cmd, 100000, stdin);

/*  for (int i=0; i<strlen(cmd); i++)
    if (cmd[i] == '\r')
      cmd[i] = '\0';
*/

  parse (cmd);
  fflush (stdout);

/*
  BuildTokenList (&parselist);
  ScanFunsAndMacs (objList, &parselist);
  ScanVariables (objList, &parselist);
  InputBranch (&parselist, 0L);
  ShowTokenList (&parselist);
  ParseTokenList (parselist);
  DeleteVariables (&parselist);
  parselist.GetTokenList()->remove();
*/

  }

return (0);
}
#endif     // of _GRAPHICS_

long casehandler (XplObject *m, int jump)
{
  if (switchcount == 0)
    {
    globalError.Set (ERR_SYNTAX, "CASE without SWITCH");
    return (0L);
    }

  if (! (*(m->DefaultXplDouble()))(0,0))
    parselist.SetJump (jump);

  if (! m->WriteProtection ())
    delete m;
  return 0L;
}

long ifhandler (XplObject *m, int jump)
{
  if (! (*(m->DefaultXplDouble()))(0,0))
    parselist.SetJump (jump + 1);

  if (! m->WriteProtection ())
    delete m;
  return 0L;
}

long whilehandler (XplObject *m, int jump)
{
  if (! (*(m->DefaultXplDouble()))(0,0))
    parselist.SetJump (jump + 1);

  if (! m->WriteProtection ())
    delete m;
  return 0L;
}

long untilhandler (XplObject *m, int jump)
{
  if (! (*(m->DefaultXplDouble()))(0,0))
    parselist.SetJump (jump + 1);

  if (! m->WriteProtection ())
    delete m;
  return 0L;
}

long switchhandler (XplObject *m, int casetokpos)
{
  TokenElement current, casetok, consttok;
  long pos = parselist.Pos ();
  Liste<TokenElement> *tl = parselist.GetTokenList ();
  

  current    = (*tl)[pos];
  casetok    = (*tl)[casetokpos];
  consttok   = (*tl)[casetokpos+1];

  while (TRUE)
    {
    if (consttok.token != VARIABLE)
      {
#ifdef INETD
cout << "\"-------=_aaaaaaa\"\nContent-Type: text/plain\n\n";
#endif
      cerr << "Error:  switchhandler:  Missing VAR after CASE" << endl;
      return (0L);
      }
    if ((casetok.token == DEFAULT) || (casetok.token == ENDSW))
      {
      parselist.SetJump (casetokpos+1);   // DEFAULT
      if (! m->WriteProtection ())
	delete m;
      return (0L);
      }
    if (IsEqual (m, consttok.content.Var->GetXplObject ()))
      {
      parselist.SetJump (casetokpos+2);   // CASE CONST
      if (! m->WriteProtection ())
	delete m;
      return (0L);
      }
    casetokpos = casetok.content.Branch;
    casetok    = (*tl)[casetokpos];
    consttok   = (*tl)[casetokpos+1];
    }

  if (! m->WriteProtection ())
    delete m;
  return 0L;
}

void ExitXpl (int n)
{
  char cmd[256];

  sprintf (cmd, "rm /tmp/%s\n", xpl4parsefile);
  system (cmd);

  exit (n);
}
