/* $Header: /home/xplore/cvs/xpl4gr/xplorexa2.y,v 1.3 2002/08/16 12:32:02 sigbert Exp $ */

/*
Syntaktische Analyse fuer XploRe (Prototyp fuer XploRe 4 Interpreter):

by TTK 30.6.94  - 6.2.95 (repeat until)

Verarbeitung von Matrizen:

Folgende Operatoren werden unterstuetzt (in dieser Prioritaet):

	Operator	Bedeutung

	~ |		horizontale und vertikale Verkettung
	+ -		Addition, Subtraktion von Matrizen
	.* / *		elementweise Mult. und Division, Matrixmult.
	-		Vorzeichen
	:		Sequenz, nicht assoziativ
	< <= > >=	Relation
	== !=		Vergleich
	|| &&		OR / AND
	
Der Parser versteht folgende Regeln:

	Klammern (), Selektieren[], Funktion auf Matrixelemente,
	Zuweisung =, Vektorliste #(), Argumentliste (Selektor, Funktion)
*/
	
%{
#include <stream.h>
#include "liste.h"
#include "object.h"
#include "function.h"
#include "macro.h"
#include "xplcalc.h"
#include "xplpp.h"
#include "init.h"

#ifdef TTKDEBUG
extern int ttkdebug;
#endif

YYSTYPE yylval;		// needed for mylex !!

int errorToken = 0;     // for error state

int          nsel;
XplSelection *xplsel;
XplDouble    *x, *x1;
XplObject    *o;
Macro        *mac;

long ifhandler     (XplObject *, int);
long whilehandler  (XplObject *, int);
long untilhandler  (XplObject *, int);
long switchhandler (XplObject *, int);

void ExitXpl       (int);

Liste<XplObject*>          *objList;          // globals
Liste<Function*>           *funList;
Liste<Macro*>              *macList;
class GlobalList           *newgl;

Liste<Liste<XplObject*>*>   argumlist;         // global in xplore.y
Liste<XplObject*>           *tmparg, *seltmp;
Liste<const char*>	    *ctmp;
Liste<XplDouble*>	    *xpldoublelist;

Liste<double>		    tmp;


extern "C" {
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <math.h>

int	yywrap  (void);		// Prototypes necessary here
void	yyerror (const char *);
int	yyparse (void);
int	mylex   (void);

%}


%union {
int		    Index;
double		    Real;
char                *Name;
class XplObject     *Obj;
class Variable      *Var;
class Function      *Fun;
class Macro         *Mac;
class MacroText	    *MacText;
Liste<double>	    *List;
Liste<Variable*>    *VarList;
Liste<XplObject*>   *ObjList;
Liste<const char*>  *CharList;
long		    Branch;
}
  /* for IF, ELSE, ENDIF, REPEAT, UNTIL */
  /*     WHILE, ENDWHILE */


%pure_parser
%start	program

%token	<Var>   VARIABLE
%token	<Real>  CONST
%token	<Index> OP_PM OP_MD OP_P OP_REL OP_EQ OP_AND OP_OR OP_SEQ OP_CON SINGLE_OP
%token  <Index> DOT COMMENT
%token  <Name>  NEWSTRING STRCONST
%token  <Fun>   FUNCTION 
%token  <Mac>   MACRO 

%token  <MacText>       PROC

%token	<Index> IF ELSE ENDIF WHILE ENDO REPEAT UNTIL ENDP
%token	<Index> SWITCH CASE BREAK DEFAULT ENDSW

%type   <Branch>        block         // probably not necessary

%type	<Obj>		constvec ausdruck
%type   <List>          list
%type   <ObjList>       arglist tmparglist tsellist sellist
%type   <CharList>      carglist parlist assignlist

%left	  OP_CON
%left	  OP_OR
%left	  OP_AND
%left	  OP_EQ
%left	  OP_REL
%nonassoc OP_SEQ
%left	  OP_PM
%left	  OP_MD
%left	  OP_P
%left	  SINGLE_OP
%left	  SIGN


%%

program         :  /* empty */
                | program block
		| error { if (!errorToken)        // get the FIRST errTok !
			    {
			    cerr << "Error: " << yychar << endl; 
			    errorToken = yychar;
			    }
			}
                  '\n'	{ yyerrok; PrintToken (errorToken); 
			  parselist.PrintCurrentToken ();
globalError.ShowError ();
//			  mt.SetNewText();       // if error with proc def
                          tmp.remove ();         // if error in list
			  errorToken = 0; }
                ;

block		: eingabe         {}  
		| block eingabe   {}
		| IF     '(' ausdruck ')'      {  $$ = ifhandler     ($3, $1); }
		| WHILE  '(' ausdruck ')'      {  $$ = whilehandler  ($3, $1); }
		| UNTIL  '(' ausdruck ')'      {  $$ = untilhandler  ($3, $1); }
		| SWITCH '(' ausdruck ')'      {  $$ = switchhandler ($3, $1); }
		| PROC '(' parlist ')' '=' NEWSTRING '(' parlist ')'
		 { newgl = CopyTokenList();
		   if (GetXplObject (objList, $6) == NULL)
		     {
		     if (newgl != NULL)
		       {
		       mac = new Macro($6, $1->Text(), *$8,*$3, newgl);
		       macList->haengeAn (mac);
		       mac->PrepareToRun ();
		       delete $1;           // MacroText must be deleted
		       }
		     else
		       cerr << "Error in Definition of " << $6 << endl;
		     }
		   else
		     {
		     cerr << $6 << " is already Variable" << endl;
		     if (newgl != NULL)
		       delete newgl;           // what is with list elements?
		     }
		   }
		| PROC '(' parlist ')' '=' MACRO '(' parlist ')'
		 { $6->SetNewDefinition ($1->Text(), *$8, *$3, CopyTokenList());
		   delete $1;             // MacroText must be deleted
		   }
		;
eingabe		: ausdruck  { if ($1 != NULL)
				{
				$1->Print(); 
				if (!$1->isEmpty ())
				  {
				  if ($1->DefaultXplDouble () != NULL)
				    $1->DefaultXplDouble()->mdump("%lf", cout);
				  if ($1->DefaultComponent()->Type() != C_PLOT)
				    if (! $1->WriteProtection ())
				      delete $1;
				  }
			        }
			    }
                | '\n'
                | assignlist '=' ausdruck { if ($3 == NULL)
cerr << "Error in list assignment: no Variables generated"<<endl;
					    else
                                              assign ($1, $3);
globalError.ShowError ();
                                          }
		| NEWSTRING '=' ausdruck  { if ($3 != NULL)
					      {
					      if ($3->WriteProtection ())
					        objList->haengeAn (new XplObject ($3, $1));
					      else
					        {
					        $3->SetName ($1);
					        objList->haengeAn ($3);
					        }
					      }
					    else 
cerr << "Error in assignment: Variable " << $1 << " not generated"<<endl;
					  }
		| VARIABLE '=' ausdruck   { if ($3 != NULL)
					      {
                      o = $1->GetXplObject ();
                      o->DelWriteProtection ();          // allow deletion of all components
                      o->TransferComponents ($3);
                      o->SetWriteProtection ();          // protect XplObject
                      GetXplObject(objList, $1->Name())->SetDefaultXplDouble ();
                      if (! $3->WriteProtection ())
		        delete $3;
					      }
					    else
cerr << "Error in assignment: Variable " << $1->Name() << " not changed"<<endl;
					  }

		| VARIABLE  sellist '=' ausdruck  {
		    x = $1->GetXplObject()->DefaultXplDouble();
		    xplsel = GenerateXplSelection (x, nsel, MakeArglist($2));

		    if (xplsel == NULL)
                      cerr << "ERROR: Selector for " << $1->Name () << endl;
		    else
		      x->assignment (nsel, xplsel, *($4->DefaultXplDouble ()));

		    $2->remove();
		    delete $2;
		    if (! $4->WriteProtection ())
		      delete $4;  }
		;


list		: CONST			{ tmp.fuegeEin($1,0); $$ = &tmp; }
		| OP_PM CONST 		{ if ($1 == MINUS)
					    tmp.fuegeEin(-$2,0);
					  else
					    tmp.fuegeEin($2,0);
					  $$ = &tmp; }
		| list ',' CONST	{ tmp.fuegeEin($3,0); }
		| list ',' OP_PM CONST	{ if ($3 == MINUS)
					    tmp.fuegeEin(-$4,0);
					  else
					    tmp.fuegeEin($4,0);
					  $$ = &tmp; }
		;
		
constvec	: '#' '(' list ')'	{ x = new XplDouble ($3->laenge());
					  for (int j=0, i=$3->laenge(); i>0;)
					    (*x)(--i) = (*$3)[j++];
					  $3->remove();
                           		  $$ = new XplObject(x);
                                        }
                ;

ausdruck	: constvec		{ $$ = $1; }
		| VARIABLE      { o = $1->GetXplObject ();
                                  if (o->isEmpty ())
                                    {
                                    yyerror ("empty Var");
cerr << "Couldn't access " << o->Name() << endl;
parselist.SetPos (parselist.GetTokenList()->laenge()-1);
                                    YYERROR;
                                    }
                                  $$ = o;  }
		| VARIABLE sellist    { 
		  x = $1->GetXplObject()->DefaultXplDouble();
		  x1 = new XplDouble;
		  xplsel = GenerateXplSelection (x, nsel, MakeArglist($2));
		  $2->remove();
		  delete $2;

		  if (xplsel == NULL)
		    {
                    cerr << "ERROR: Selector for " << $1->Name () << endl;
		    $$ = NULL;
                    }
		  else
		    {
		    x1->selection (*x, nsel, xplsel);
		    $$ = new XplObject (x1);
		    }
// werden alle Matrizen geloescht?
}
		| ausdruck sellist    { 
		  x = $1->DefaultXplDouble();
		  x1 = new XplDouble;
		  xplsel = GenerateXplSelection (x, nsel, MakeArglist($2));
		  $2->remove();
		  delete $2;

		  if (xplsel == NULL)
		    {
		    cerr << "ERROR: Selector for term" << endl;
		    $$ = NULL;
		    }
		  else
		    {
		    x1->selection (*x, nsel, xplsel);
		    $$ = new XplObject (x1);
		    }  }
		| CONST		{ x = new XplDouble (1);
				  (*x)(0) = $1;
				  $$ = new XplObject (x); }
		| STRCONST	{ $$ = new XplObject (new Component ($1)); }
		| '(' ausdruck ')'		{ $$ = $2; }
		| OP_PM ausdruck  %prec SIGN	{ if ($1 == MINUS)
						    $$ = operate($2, $1);
						  else
						    $$ = $2; }
		| ausdruck SINGLE_OP       { $$ = operate($1,$2); }
		| ausdruck OP_CON ausdruck { $$ = operate($1,$3,$2); }
		| ausdruck OP_MD ausdruck  { $$ = operate($1,$3,$2); }
		| ausdruck OP_PM ausdruck  { $$ = operate($1,$3,$2); }
		| ausdruck OP_P  ausdruck  { $$ = operate($1,$3,$2); }
		| ausdruck OP_EQ ausdruck  { $$ = operate($1,$3,$2); }
		| ausdruck OP_REL ausdruck { $$ = operate($1,$3,$2); }
		| ausdruck OP_SEQ ausdruck { $$ = operate($1,$3,$2); }
		| ausdruck OP_AND ausdruck { $$ = operate($1,$3,$2); }
		| ausdruck OP_OR  ausdruck { $$ = operate($1,$3,$2); }
                | FUNCTION '('  { argumlist.fuegeEin (tmparg, 0);
			          tmparg = new Liste<XplObject*>; }
                               arglist ')' {
                                           GenerateTempXplObjects ($4);
				           $$ = ($1->Fun ()) ($4);
globalError.ShowError ();
                                           long i, length = tmparg->laenge();
					   for (i=0; i<length; i++)
					     {
// cerr << "Fun: " << i << (*tmparg)[i]->WriteProtection () << endl;
                                             if (!(*tmparg)[i]->WriteProtection ())
					       delete (*tmparg)[i];
                                             }
					   delete tmparg;
                                           tmparg = argumlist[0];
					   argumlist.loesche(0); }
                | MACRO '('  { argumlist.fuegeEin (tmparg, 0);
			       tmparg = new Liste<XplObject*>; }
                   arglist ')'   {
		                 $$ = ParseMacro ($1, $4);
globalError.ShowError ();
		                 long i, length = tmparg->laenge();
				 for (i=0; i<length; i++)
				   {
// cerr << "Mac: " << i << (*tmparg)[i]->WriteProtection () << endl;
                                   if (!(*tmparg)[i]->WriteProtection ())
				     delete (*tmparg)[i];
                                   }
				 delete tmparg;
				 tmparg = argumlist[0];
				 argumlist.loesche(0);
				 if ($$ != NULL)
				   $$->DelWriteProtection ();  }
                | NEWSTRING '('  { argumlist.fuegeEin (tmparg, 0);
			       tmparg = new Liste<XplObject*>; }
                   arglist ')'   {
		                 if (NULL != (mac = GetMacro (macList, $1)))
				   {
				   $$ = ParseMacro (mac, $4);
globalError.ShowError ();
				   long i, length = tmparg->laenge();
				   for (i=0; i<length; i++)
				     {
cerr << "NEWSTRING MACRO\t" << $1 << endl;
// cerr << "Mac: " << i << (*tmparg)[i]->WriteProtection () << endl;
                                     if (!(*tmparg)[i]->WriteProtection ())
				       delete (*tmparg)[i];
                                     }
				   delete tmparg;
				   tmparg = argumlist[0];
				   argumlist.loesche(0);
				   if ($$ != NULL)
				     $$->DelWriteProtection ();
                                   }
globalError.Set (ERR_MAC, $1);
globalError.ShowError ();
                                 }
                ;

tsellist	: ausdruck	      { seltmp->haengeAn ($1);   $$ = seltmp; }
		| ','		      { seltmp->haengeAn (NULL); $$ = seltmp; }
		| tsellist ausdruck   { seltmp->haengeAn ($2);   $$ = seltmp; }
		| tsellist ','	      { seltmp->haengeAn (NULL); $$ = seltmp; }
		;
sellist         : '[' tsellist ']' { $$ = $2; seltmp = new Liste<XplObject*>; }
                ;

tmparglist      : /* empty */       { $$ = tmparg; }
                | ausdruck	    { if ($1 != NULL)
				        tmparg->haengeAn ($1);   $$ = tmparg; }
		| tmparglist ',' ausdruck 
                                    { if ($3 != NULL)
				        tmparg->haengeAn ($3);   $$ = tmparg; }
                ;

arglist         : tmparglist              { $$ = $1; }
                ;

carglist	: /* empty */             { $$ = ctmp; }
                | NEWSTRING               { ctmp->haengeAn ($1);  $$ = ctmp; }
                | VARIABLE                { ctmp->haengeAn ($1->Name());
					    $$ = ctmp; }
		| carglist ',' NEWSTRING  { ctmp->haengeAn ($3);  $$ = ctmp; }
		| carglist ',' VARIABLE   { ctmp->haengeAn ($3->Name());
					    $$ = ctmp; }
                | MACRO                   { ctmp->haengeAn ($1->Name());
					    $$ = ctmp; }
		| carglist ',' MACRO      { ctmp->haengeAn ($3->Name());
					    $$ = ctmp; }
		;

parlist		: carglist	          { $$ = $1; ctmp = new Liste<const char*>; }
		;

assignlist	: '{' parlist '}'         { $$ = $2; }
		;

%%


#include "lex.yy.c"

}  // of extern C

      			
#include "xplpp.h"

#ifndef _GRAPHICS_
int main (int argc, char **argv)
{
  Liste<TokenElement> tl;

#ifdef YYDEBUG
  yydebug  = (argc >= 3);        // debug info if dummy parameter is present
#endif
#ifdef TTKDEBUG
  ttkdebug = (argc == 2) || (argc >= 4);
#endif

  InitializeInterpreter ();  

while (1)
  {

  BuildTokenList (&parselist);
  ScanFunsAndMacs (objList, &parselist);
  ScanVariables (objList, &parselist);
  InputBranch (&parselist, 0L);
  ShowTokenList (&parselist);
  ParseTokenList (parselist);
  DeleteVariables (&parselist);
  parselist.GetTokenList()->remove();
  }

return (0);
}
#endif     // of _GRAPHICS_

long ifhandler (XplObject *m, int jump)
{
  if (! (*(m->DefaultXplDouble()))(0,0))
    parselist.SetJump (jump + 1);

  if (! m->WriteProtection ())
    delete m;
  return 0L;
}

long whilehandler (XplObject *m, int jump)
{
  if (! (*(m->DefaultXplDouble()))(0,0))
    parselist.SetJump (jump + 1);

  if (! m->WriteProtection ())
    delete m;
  return 0L;
}

long untilhandler (XplObject *m, int jump)
{
  if (! (*(m->DefaultXplDouble()))(0,0))
    parselist.SetJump (jump + 1);

  if (! m->WriteProtection ())
    delete m;
  return 0L;
}

long switchhandler (XplObject *m, int casetokpos)
{
  TokenElement current, casetok, consttok;
  long pos = parselist.Pos ();
  Liste<TokenElement> *tl = parselist.GetTokenList ();
  

  current    = (*tl)[pos];
  casetok    = (*tl)[casetokpos];
  consttok   = (*tl)[casetokpos+1];

  while (TRUE)
    {
    if (consttok.token != VARIABLE)
      {
      cerr << "Error:  switchhandler:  Missing VAR after CASE" << endl;
      return (0L);
      }
    if ((casetok.token == DEFAULT) || (casetok.token == ENDSW))
      {
      parselist.SetJump (casetokpos+1);   // DEFAULT
      if (! m->WriteProtection ())
	delete m;
      return (0L);
      }
    if (IsEqual (m, consttok.content.Var->GetXplObject ()))
      {
      parselist.SetJump (casetokpos+2);   // CASE CONST
      if (! m->WriteProtection ())
	delete m;
      return (0L);
      }
    casetokpos = casetok.content.Branch;
    casetok    = (*tl)[casetokpos];
    consttok   = (*tl)[casetokpos+1];
    }

  if (! m->WriteProtection ())
    delete m;
  return 0L;
}

void ExitXpl (int n)
{
  char cmd[256];

  sprintf (cmd, "rm /tmp/%s\n", xpl4parsefile);
  system (cmd);

  exit (n);
}
