/* $Header: /home/xplore/cvs/xpl4gr/xplpp.h,v 1.3 2002/08/16 12:32:02 sigbert Exp $ */

#ifndef _PP_H_
#define _PP_H_

// #pragma interface

#ifdef MATRIX
#include "matrix.h"
typedef Matrix  XplDouble;
#else
#include "xreal.hpp"
#include "xint.hpp"
typedef XplReal<double> XplDouble;
#endif

#if defined(SOCKETSERVER) || defined(X4singleserver)
#include "xjprotocol.h"
extern class XJprotocol *protocol;
#endif

// Preprocessor for mdcalc
// Build up a token list with Lexer mylex

//#ifndef __BORLANDC__
#include "xplore.tab.h"
//#else
//#include "xpltab.h"
//#endif

#include <stdio.h>


typedef struct {
		YYSTYPE	content;
		int	token;
		long	SourceLine;
		long	jump;
		} TokenElement;

#include "macro.h"		// need decl. of Macro class now
#include "globlist.h"
#include "fileinfo.h"

static long frame = 0;

extern YYSTYPE yylval;
extern GlobalList parselist;

extern "C" {
int	checklex  (YYSTYPE *lvalp);

int	yywrap    (void);
void    yyrestart (FILE *);
int	yylex     (YYSTYPE *lvalp);
void	yyerror   (const char *);
int	yyparse   (void);

int	checkparse   (void);

int	mylex (void);
}  // of extern C

long CaseHandler   (XplObject *, int);
long IfHandler     (XplObject *, int);
long WhileHandler  (XplObject *, int);
long UntilHandler  (XplObject *, int);
long SwitchHandler (XplObject *, int);

GlobalList*	CopyTokenList (void);
void	BuildTokenList    (GlobalList *gl, int end = '\n');
int	ParseTokenList    (const GlobalList &gl);
void	CheckTokenList    (const GlobalList &gl);
XplObject *    ParseMacro (Macro *m, Liste<XplObject*> *args);
void	ShowTokenList     (const GlobalList *gl);
void	InputBranch       (const GlobalList *gl, long pos);
long	InputIfBranch     (const GlobalList *gl, long pos);
long	InputWhileBranch  (const GlobalList *gl, long pos);
long	InputUntilBranch  (const GlobalList *gl, long pos);
long	InputSwitchBranch (const GlobalList *gl, long pos);
void	PrintToken (int token);

void	PrintXplObjectList     (Liste<XplObject*> *);
void	PrintXplObjectNameList (Liste<XplObject*> *);
void	CopyXplObjectList      (Liste<XplObject*> *, Liste<XplObject*> *);

void AssignLocalVariables (const GlobalList *gl, Liste<XplObject*> *objList);
int     EvalString    (char *name, long &pos);

void    ScanFunsAndMacs    (GlobalList *gl);
void    ScanConstants      (GlobalList *gl);
void    ScanXplObjects     (Liste<XplObject*> *, const GlobalList *gl);
void    ScanMacroVariables (Liste<XplObject*> *, const GlobalList *gl);
int     ScanVariables      (Liste<XplObject*> *, const GlobalList *gl);
void    SetVariables2Null  (GlobalList *gl);
void    EvalXplObject      (Liste<XplObject*> *, Vector<TokenElement>*, long &pos);

void         DeleteVariables (GlobalList *gl);
Variable *   EvalVariable    (Vector<TokenElement>*, long &pos);
FileInfo *   GetFileInfo     (Vector<FileInfo*> &fileList, const char *n);
Function *   GetFunction     (const char *n);
Macro *      GetMacro        (Vector<Macro*> *, const char *n);
XplObject *  GetXplObject    (Liste<XplObject*> *, const char *n);
int          DeleteXplObject (Liste<XplObject*> *, const char *n);
XplObject *  GetXplObject    (Liste<XplObject*> *, const char *n, long &pos);
XplObject *  GetXplObject    (Liste<XplObject*> *, const long, long &pos);
XplObject *  GetXplObject    (Liste<XplObject*> *, const long);

void         TidyUp          (Liste<XplObject*> *);
XplObject *  GetContainerXplObject (Liste<XplObject*> *objList, const long id);

void    parse (const char *);
void    parse (FILE *);

int  GenerateTempXplObjects (Liste<XplObject*> *argl);

#ifndef MATRIX
#include "object.h"
XplSelection *  
   GenerateXplSelection (XplDouble *x, int &nsel, Liste<XplDouble*> *xd);
XplSelection *  
   GenerateXplSelection (XplIndex  d, int &nsel, Liste<XplDouble*> *xd);
#endif

XplObject * SelectComponents    (XplObject *root, XplObject *sel);
XplObject * SelectComponentList (XplObject *root, Liste<XplObject*> *sel);
void        AssignSelection     (XplObject *dest, XplObject *source);

void    ok_macro (char *t);          // interface to macro editor (ok button)

int         RemoveDisplay       (XplObject *o);  // for displayList and
                                                 // globObjList

#if defined (_GRAPHICS_) || defined (PSGEN) || defined (__BORLANDC__) //LOOK!
#include "object.h"
XplObject* GetXplObject (const class XplDisplay *d);
#endif

Liste<XplDouble*>* MakeArglist       (Liste<XplObject*> *mm, XplObject *o);

#ifdef _XPLCDEF_
XplDouble *        LabelSelection    (XplChar *sel, XplChar *base);
// const SK didn't it
XplChar *    GetDefaultXplChar (const XplObject *o);
// const SK didn't it
XplChar *    GetXplChar        (const XplObject *o, const char *name = NULL);
#endif

#ifdef RPC
void  StartRPCServer (int argc, char **argv);
#endif

#ifdef SOCKETSERVER
void  StartSocketServer (int argc, char **argv);
#endif

#if !defined (RPC) && !defined (SOCKETSERVER)
void  StartXploRe (int argc, char **argv);
#endif

void X4Start(int argc, char** argv);

#endif

