/* $Header: /home/xplore/cvs/xpl4gr/xstring.C,v 1.5 2003/02/23 17:17:07 cizek Exp $ */

#include "xstring.hpp"

int XplStringDatabase::search (const char *newtxt, long *mid)
{
    int i, cmp;
    long lo, hi;

    lo = 0;		// low limit
    hi = nstrings-1;	// high limit
    if(nstrings==0)	// empty database
    {	*mid = 0;
        return(0);
    }
    while (1)
    {
        *mid = (lo + hi) >> 1;				// middle index
        cmp = strcmp(newtxt,list[indices[*mid]].txt);
        if (cmp)
	     {
            if(cmp>0)		// newtxt lies right of the middle string
                lo = *mid+1;
            else		// newtxt lies left of the middle string
                hi = *mid-1;
        }
        else
            return(1);		 // string found

	if(lo>hi)		// set up mid to the following element
	{
            if(cmp>0)
                *mid = lo;
            else
                *mid = hi + 1;

	    return(0);
	}
    }
}

XplStringDatabase::XplStringDatabase (void)
{   unsigned i;
    list = (XplStringElement *) malloc(MAXSTRING*sizeof(XplStringElement));
    indices = (unsigned long *) malloc(MAXSTRING*sizeof(unsigned long));

    used = 0;
    nstrings = 0;
    empty = 1;
    size = MAXSTRING;
    emptystring = strdup(" ");

    if ((list == NULL) || (indices == NULL))
    {
       XAerror(NOMEM);
       return;
    }

    memset(list, 0, MAXSTRING*sizeof(XplStringElement));
    memset(indices, 0, MAXSTRING*sizeof(unsigned long));
}

XplStringDatabase::~XplStringDatabase ()
{
    if (list)
         free(list);
    if (indices)
         free(indices);
    free(emptystring);
}

unsigned long XplStringDatabase::entries (void)
{   return used; }

unsigned long XplStringDatabase::numstrings (void)
{   return nstrings; }

unsigned long XplStringDatabase::insert (const char *newtxt)
{   unsigned long i, j, pos;
    long ind = 0;
    int found = search (newtxt, &ind);
    XplStringElement *try1;
    unsigned long *try2;

    if (found)		// string found
    {
        pos = indices[ind];
        add(pos+1);	// 1 will be subtracted in add
	     return(pos+1);
    }
    else
    {
	// put new string
   while (empty >= size) 	// expand the database
	{
	    size+=STRINGSPLUS;
	    try1 = (XplStringElement *) realloc(list,size*sizeof(XplStringElement));
       try2 = (unsigned long *) realloc(indices,size*sizeof(unsigned long));
	    if(try1 && try2)
       {
       	  list = try1;
			  indices = try2;
	    }
	    else
            {   // error - not enough memory
               XAerror(NOMEM);
               return(0);
	    }
	}

        list[empty].txt = strdup(newtxt);
        list[empty].links = 1;
        pos = empty;

	if(empty>=used)
            used = empty + 1;

	// sort it
	i = ind;	// index where the new element belongs
	if (i<nstrings)
	    memmove(indices+i+1, indices+i, sizeof(unsigned long)*(nstrings-i));
	nstrings++;
	indices[i] = empty;

	// and find new empty block
	do
 	    empty++;
	while( (empty<used) && (list[empty].txt != emptystring) );

	   return(pos+1);
   }   // else - new string element
}

void XplStringDatabase::add (unsigned long no)
{
    if (0 < no)
         no--;
    else
    {
         XAerror(STRINGBAD);
         no = used;
    }

    if (no<used)
        list[no].links++;
}


void XplStringDatabase::kill (unsigned long no)
{
    XplStringElement *try1;
    unsigned long *try2;
    long i;

    if (0 < no)
         no--;
    else
    {
         //XAerror(STRINGBAD);
         no = used;
    }

    if (no>=used)
        return;
    if (list[no].links > 0)
        list[no].links--;

    // delete
    if (list[no].links==0)
    {
      long noind = 0;
      int found = search(list[no].txt, &noind);

	   if ((list[no].txt!=emptystring) && (list[no].txt))	// non empty
  	     free(list[no].txt);

	   list[no].txt = emptystring;	// point at emptystring so that no error occurs, if someone use deleted string

   // remove it from indices space and nstring
      if ((noind + 1 < nstrings) && found)
         memmove(indices+noind, indices+noind+1, sizeof(unsigned long)*(nstrings-noind-1));
	   nstrings--;


	if(no < empty)		// new free space
	    empty = no;

	if((no+1)==used) 	// last deleted, find as large block of deleted elements as possible
	{
	    for(i = used-1; (i >= 0) && (list[i].txt == emptystring); i--);
	    used = i + 1;
	    empty = used;
	}

	if ((used + STRINGSMINUS < size) && (size >= MAXSTRING + STRINGSMINUS ))	// chance for saving memory
	{
	    unsigned long newsize = (used + STRINGSPLUS > MAXSTRING) ? used + STRINGSPLUS : MAXSTRING;

	    try1 = (XplStringElement *) realloc(list,newsize*sizeof(XplStringElement));
       try2 = (unsigned long *) realloc(indices,newsize*sizeof(unsigned long));
	    if (try1)
                list = try1;
	    if (try2)
		indices = try2;
	    if (try1 || try2)
		size = newsize;
	}
    }   // deleting
}

void XplStringDatabase::dump (char *filename)
{   FILE *f = fopen (filename, "a");
    time_t t;
    unsigned l;

    ctime (&t);
    fprintf (f, "\n");
    fprintf (f, "Full Dump of Stringbase %s", ctime(&t));
  
    fprintf (f, "\n indices[1..%u] \n", nstrings);
    for(l = 0; l<nstrings; l++)
        fprintf (f, "%5u ", indices[l]);
    fprintf (f, "\n list[1..%u].links   list[1..%u].txt", used, used);
    for(l = 0; l<used; l++)
    {
        fprintf (f, "%5u ", list[l].links);
        fprintf (f, "%s\n", list[l].txt);
    }
    
    fclose(f);
}

//  void XplStringDatabase::sdump (char *filename, char *howtoopen = NULL)
void XplStringDatabase::sdump (char *filename, char *howtoopen)
{   FILE *f;
    time_t t;
    unsigned l;
  
    if (howtoopen==NULL)
        f = fopen (filename, "w");
    else
        f = fopen (filename, howtoopen);
  
    ctime (&t);
    fprintf (f, "Short Dump of Stringbase %s\n", ctime(&t));
    
    for(l = 0; l<nstrings; l++)
    { 
        fprintf (f, "%5u ", indices[l]);
        fprintf (f, "%5u ", list[indices[l]].links);
        fprintf (f, "%s\n", list[indices[l]].txt);
    }
    fclose(f);
}

const char* XplStringDatabase::operator[] (unsigned long no)
{
    if (0 < no)
         no--;
    else
    {
         XAerror(STRINGBAD);
         return emptystring;
    }

    return (list[no].txt);
}

static char buffer[2048], strtokbuffer[2048];

xplstring::xplstring () { strno = 0; }

xplstring::xplstring (const char *txt) { strno = xplorebase.insert(txt); }

xplstring::xplstring (const xplstring &txt) 
  { strno = txt.strno; 
    xplorebase.add(strno); 
  }

xplstring::~xplstring () { xplorebase.kill(strno); strno = 0; }

void xplstring::kill () { xplorebase.kill(strno); strno = 0; }

xplstring& xplstring::operator= (const xplstring &left)
  { if (this != &left)
    { xplorebase.kill(strno);
      strno = left.strno;
      xplorebase.add(strno);
    }
    return *this;
  }
  
void xplstring::setnull (void)
  { strno = 0; }

int xplstring::xplisnull(void) const { return (!strno); }

unsigned xplstring::xplstrlen (void) const { return (::strlen(xplorebase[strno])); }

xplstring xplstring::xplstrdup (void) const
  { xplstring res;
    res.strno = xplorebase.insert(xplorebase[strno]); 
    return (res);
  }

//  xplstring xplstring::xplstrlwr (void)
//  { xplstring res;
//   ::strcpy (buffer, xplorebase[strno]);
//    ::strlwr (buffer);
//    res.strno = xplorebase.insert(buffer);
//    return(res);
//  }

//  xplstring xplstring::xplstrupr (void)
//  { xplstring res;
//    ::strcpy (buffer, xplorebase[strno]);
//    ::strupr (buffer);
//    res.strno = xplorebase.insert(buffer);
//    return(res);
//  }

//  xplstring xplstring::xplstrrev (void)
//  { xplstring res;
//    ::strcpy (buffer, xplorebase[strno]);
//    ::strrev (buffer);
//    res.strno = xplorebase.insert(buffer);
//    return (res);
//  }

double xplstring::xplatof (void) const
  { ::strcpy (buffer, xplorebase[strno]);
    return (::x4atof(buffer));
  }

int xplstring::xplatoi (void) const
  { ::strcpy (buffer, xplorebase[strno]);
    return (::atoi(buffer));
  }

long xplstring::xplatol (void) const
  { ::strcpy (buffer, xplorebase[strno]);
    return (::atol(buffer));
  }

double xplstring::xplstrtod (void) const { return (xplatof()); }

xplstring xplstring::xplstrchr (int ch) const
  { char *ptr;
    xplstring res;
    ::strcpy (buffer, xplorebase[strno]);
    ptr = ::strchr (buffer, ch);
    res.strno = xplorebase.insert(ptr);
    return (res);
  }

xplstring xplstring::xplstrrchr (int ch) const
  { char *ptr;
    xplstring res;
    ::strcpy (buffer, xplorebase[strno]);
    ptr = ::strrchr (buffer, ch);
    res.strno = xplorebase.insert(ptr);
    return (res);
  }

//  xplstring xplstring::xplstrset (int ch)
//  { char *ptr;
//    xplstring res;
//    ::strcpy (buffer, xplorebase[strno]);
//    ::strset (buffer, ch);
//    res.strno = xplorebase.insert(buffer);
//    return (res);
//  }

//  xplstring xplstring::xplstrnset (int ch, unsigned n)
//  { char *ptr;
//    xplstring res;
//    ::strcpy (buffer, xplorebase[strno]);
//    ::strnset (buffer, ch, n);
//    res.strno = xplorebase.insert(buffer);
//    return (res);
//  }

long xplstring::xplstrtol (int base) const { return (xplatol()); }

unsigned long xplstring::xplstrtoul (int base) const
  { ::strcpy (buffer, xplorebase[strno]);
    return (::strtoul (buffer, NULL, base));
  }

//  xplstring xplstring::xplitoa (int val, int base)
//  { xplstring res;
//    ::itoa (val, buffer, base);
//    res.strno = xplorebase.insert(buffer); 
//    return(res);
//  }

//  xplstring xplstring::xplltoa (long val, int base)
//  { xplstring res;
//    ::ltoa (val, buffer, base);
//    res.strno = xplorebase.insert(buffer); 
//    return (res);
//  }

//  xplstring xplstring::xplultoa (unsigned long val, int base)
//  { xplstring res;
//    ::ultoa (val, buffer, base);
//    res.strno = xplorebase.insert(buffer); 
//    return (res);
//  }

float xplstring::xplstrtof (void) const { return (xplatof()); }

xplstring xplstring::xplstrcpy (const xplstring txt) const
  { xplstring res;
    res.strno = xplorebase.insert(xplorebase[txt.strno]);
    return (res);
  }

xplstring xplstring::xplstrcat (const xplstring txt) const
  { xplstring res;
    ::strcpy (buffer, xplorebase[strno]);
    ::strcat (buffer, xplorebase[txt.strno]);
    res.strno = xplorebase.insert(buffer);
    return (res);
  }

int xplstring::xplstrcmp (const xplstring txt) const
  { return (::strcmp(xplorebase[strno], xplorebase[txt.strno])); }

unsigned xplstring::xplstrcspn (const xplstring txt) const 
  { return (::strcspn(xplorebase[strno], xplorebase[txt.strno])); }

xplstring xplstring::xplstrpbrk (const xplstring txt) const
  { char *ptr;
    xplstring res;
    ::strcpy (buffer, xplorebase[strno]);
    ptr = ::strpbrk (buffer, xplorebase[txt.strno]); 
    res.strno = xplorebase.insert(ptr);
    return (res);
  }

unsigned xplstring::xplstrspn (const xplstring txt) const 
  { return (::strspn(xplorebase[strno], xplorebase[txt.strno])); }

xplstring xplstring::xplstrstr (const xplstring txt) const
  { char *ptr;
    xplstring res;
    ::strcpy (buffer, xplorebase[strno]);
    ptr = ::strstr (buffer, xplorebase[txt.strno]); 
    res.strno = xplorebase.insert(ptr);
    return (res);
  }

xplstring xplstring::xplstrtok (const xplstring txt) const
  { char *ptr;
    xplstring res;
    ::strcpy (strtokbuffer, xplorebase[strno]);
    ptr = ::strtok (strtokbuffer, xplorebase[txt.strno]);
    res.strno = xplorebase.insert(ptr);
    return (res);
  }

xplstring xplstring::xplstrtoknull (const xplstring txt) const
  { char *ptr;
    xplstring res;
    ptr = ::strtok (NULL, xplorebase[txt.strno]);
    res.strno = xplorebase.insert(ptr);
    return (res);
  }

//  int xplstring::xplstricmp (const xplstring txt)
//  { return (::stricmp(xplorebase[strno], xplorebase[txt.strno])); }

xplstring xplstring::xplstrcpy (const char *txt) const
  { xplstring res;
    ::strcpy (buffer, txt);
    res.strno = xplorebase.insert(buffer);
    return(res);
  }

xplstring xplstring::xplstrcat (const char *txt) const
  { xplstring res;
    ::strcpy (buffer, xplorebase[strno]);
    ::strcat (buffer, txt);
    res.strno = xplorebase.insert(buffer);
    return (res);
  }

int xplstring::xplstrcmp (const char *txt) const
  { return (::strcmp(xplorebase[strno], txt)); }

unsigned xplstring::xplstrcspn (const char *txt) const
  { return (::strcspn(xplorebase[strno], txt)); }

xplstring xplstring::xplstrpbrk (const char *txt) const
  { char *ptr;
    xplstring res;
    ::strcpy (buffer, xplorebase[strno]);
    ptr = ::strpbrk (buffer, txt); 
    res.strno = xplorebase.insert(ptr);
    return(res);
  }

unsigned xplstring::xplstrspn (const char *txt) const
  { return (::strspn(xplorebase[strno], txt)); }

xplstring xplstring::xplstrstr (const char *txt) const
  { char *ptr;
    xplstring res;
    ::strcpy (buffer, xplorebase[strno]);
    ptr = ::strstr (buffer, txt); 
    res.strno = xplorebase.insert(ptr);
    return(res);  
  }

xplstring xplstring::xplstrtok (const char *txt) const
  { char *ptr;
    xplstring res;
    ::strcpy (strtokbuffer, xplorebase[strno]);
    ptr = ::strtok (strtokbuffer, txt);
    res.strno = xplorebase.insert(ptr);
    return (res);
  }

xplstring xplstring::xplstrtoknull (const char *txt) const
  { char *ptr;
    xplstring res;
    ptr = ::strtok (NULL, txt);
    res.strno = xplorebase.insert(ptr);
    return (res);
  }

//  int xplstring::xplstricmp (const char *txt)
//  { return (::stricmp(xplorebase[strno], txt)); }

// PC MAC !!! to be completed
#ifdef __APPLE_CC__
#include <math.h>

char *ecvt(double number, int ndigits, int *decpt, int *sign)
{
	char fmtstr[60];
	static char resstring[60];
	if ((ndigits > 17) || (ndigits < 0))
		ndigits = 17;

	*sign = (number < 0.0) ? -1 : 1;
	*decpt = (number*number < 1.0) ? -1 : (int) log10(*sign * number) + 1;
	*sign = (number < 0.0) ? -1 : 0;

	sprintf(fmtstr, "%%%df", ndigits);
	sprintf(resstring, fmtstr, floor(number * pow(10,17-*decpt)));

	return resstring;
}

char *fcvt(double number, int ndigits, int *decpt, int *sign)
{
	char fmtstr[60];
	static char resstring[60];
	if ((ndigits > 17) || (ndigits < 0))
		ndigits = 17;

	*sign = (number < 0.0) ? -1 : 1;
	*decpt = (number*number < 1.0) ? -1 : (int) log10(*sign * number) + 1;
	*sign = (number < 0.0) ? -1 : 0;

	sprintf(fmtstr, "%%%df", ndigits);
	sprintf(resstring, fmtstr, floor(number * pow(10,17-*decpt)));

	return resstring;
}
#endif

xplstring xplstring::xplecvt (double val, int ndig, int *pdecpt, int *psign) const
  { xplstring res;
    res.strno = xplorebase.insert(::ecvt(val, ndig, pdecpt, psign));
    return (res);
  }

xplstring xplstring::xplfcvt (double val, int ndig, int *pdecpt, int *psign) const
  { xplstring res;
#ifndef __HPUXFUN__  // SK 02.10.97
    res.strno = xplorebase.insert(::fcvt(val, ndig, pdecpt, psign));
#else
    char *tmp;
    tmp = ::fcvt(val, ndig, pdecpt, psign);
    res.strno = xplorebase.insert(tmp);
//    delete[] tmp;  
#endif
    return (res);
  }

xplstring xplstring::xplstrncpy (const xplstring txt, unsigned n) const
  { xplstring res;
    ::strncpy(buffer, xplorebase[txt.strno], n);
    res.strno = xplorebase.insert(buffer);
    return(res);  
  }

xplstring xplstring::xplstrncat (const xplstring txt, unsigned n) const
  { xplstring res;
    ::strcpy(buffer, xplorebase[strno]);
    ::strncat(buffer, xplorebase[txt.strno], n);
    res.strno = xplorebase.insert(buffer);
    return(res);  
  }

int xplstring::xplstrncmp (const xplstring txt, unsigned n) const
  { return (::strncmp(xplorebase[strno], xplorebase[txt.strno], n));}

//unsigned xplstring::xplstrxfrm (const xplstring txt, unsigned n) const
//  { return (::strxfrm(xplorebase[strno], xplorebase[txt.strno], n));}

//  int xplstring::xplstrnicmp (const xplstring txt, unsigned n)
//  { return (::strnicmp(xplorebase[strno], xplorebase[txt.strno], n));}

xplstring xplstring::xplstrncpy (const char *txt, unsigned n) const
  { xplstring res;
    ::strncpy(buffer, txt, n);
    res.strno = xplorebase.insert(buffer);
    return(res);  
  }

xplstring xplstring::xplstrncat (const char *txt, unsigned n) const
  { xplstring res;
    ::strcpy(buffer, xplorebase[strno]);
    ::strncat(buffer, txt, n);
    res.strno = xplorebase.insert(buffer);
    return(res);  
  }

int xplstring::xplstrncmp (const char *txt, unsigned n) const
  { return (::strncmp(xplorebase[strno], txt, n));}

//unsigned xplstring::xplstrxfrm (const char *txt, unsigned n) const
//  { return (::strxfrm(xplorebase[strno], txt, n));}

int xplstring::operator< (const xplstring &r) const
  { int i = 0;
    while (1)
    { if ((xplorebase[strno])[i]==0)
      { if ((xplorebase[r.strno])[i]==0) return 0;
        return 1;
      }
      if ((xplorebase[r.strno])[i]==0) return 0;
      if (((xplorebase[strno])[i])<((xplorebase[r.strno])[i])) return 1;
      if (((xplorebase[strno])[i])>((xplorebase[r.strno])[i])) return 0;
      i++;
    }
    return 0;   
  }

int xplstring::operator> (const xplstring &r) const
  { int i = 0;
    while (1)
    { if ((xplorebase[strno])[i]==0)
      { if ((xplorebase[r.strno])[i]==0) return 0;
        return 0;
      }
      if ((xplorebase[r.strno])[i]==0) return 1;
      if (((xplorebase[strno])[i])<((xplorebase[r.strno])[i])) return 0;
      if (((xplorebase[strno])[i])>((xplorebase[r.strno])[i])) return 1;
      i++;
    }
    return 0;   
  }

int xplstring::operator== (const xplstring &r) const
  { return (strno == r.strno); }

int xplstring::operator!= (const xplstring &r) const
  { return (strno != r.strno); }

xplstring xplstring::operator+ (const xplstring &r) const
  { XAerrort (NOUSE, " xplstring operator+ (xplstring &)");
    xplstring ret; return (ret);
  }

xplstring xplstring::operator+ (long r) const
  { XAerrort (NOUSE, " xplstring operator+ (long)");
    xplstring ret; return (ret); 
  }

xplstring xplstring::operator* (const xplstring &r) const
  { XAerrort (NOUSE, " xplstring operator* (xplstring &)");
    xplstring ret; return (ret); 
  }

xplstring::xplstring (double d)
  { strno = 0; }

char* xplstring::tostr (void) const
  { if (strno>0)
      return (strdup(xplorebase[strno])); 
    XAerrort (STRINGBAD, "Dup from NULL");
    return (strdup(""));
  }

const char* xplstring::tocstr (void) const
  { return (xplorebase[strno]); }

//  xplstring xplstring::gcvt (double val, int ndig, char *)

void xplstring::print () const
  { printf ("%li ", strno);}

