/* $Header: /home/xplore/cvs/xpl4gr/xtime.C,v 1.3 2002/08/16 12:32:04 sigbert Exp $ */

#include "xtime.hpp"
#include "x4math2.h"

long _juldnj(struct tm *bdt, long Transition)
{	int ay, y, dy;
	int m;
	long jd_julian, jd_gregorian, jd;

	y = bdt->tm_year + 1900;
	m = bdt->tm_mon;

	dy = m/12;
	y += dy;
	m -= dy*12;
			/* assert( -11 <= m && m <= 11 ) */
	if(m < 0)
		{
		m += 12;
		y--;
		}
			/* assert( 0 <= m && m <= 11 ) */

	if ( m < 2 ) 
		{
		m += 1 + 12;
		y--;
		}
	else
		m++;

	ay = y + 4716;
	jd_julian = ((1461*(long)ay) >> 2) + (153*(m + 1)/5)
		   + (long)bdt->tm_mday - 1524;
	jd_gregorian = jd_julian + 2 - y/100 + y/400 - y/4000;
	if ( jd_gregorian >= Transition ) jd = jd_gregorian; 
	else jd = jd_julian;
	return jd;
}

struct tm* julcdj (long jd, long Transition)
{	long aa, ab, a;
	long b, d, ee;
	int ay;
	int em, m;
	int y;
	struct tm newyears;
	static struct tm date;

	memset(&date, 0, sizeof(date));

	if ( jd < Transition ) /* Julian Calendar */
		a = (long)(jd);
	else /* Gregorian Calendar */ 
		{
		aa = jd - 1721120L;
		ab = 31*(aa/1460969L); aa = aa % 1460969L;
		ab = ab + 3*(aa/146097L); aa = aa % 146097L;
		if ( aa == 146096L ) ab = ab + 3; 
		else ab = ab + aa/36524L;
		a = jd + (ab - 2);
		}
	b = a + 1524;
	ay = (int)((20*b - 2442)/7305);
	d = (1461*(long)(ay)) >> 2;
	ee = b - d;
	em = (int)(10000L*ee/306001L);

	date.tm_mday = (int)(ee - 306001L*em/10000L);

	m = em - 1;
	if(m > 12) m -= 12;
	date.tm_mon = m - 1;

	if ( m > 2 ) y = ay - 4716; 
	else y = ay - 4715;
	date.tm_year = y - 1900;

	date.tm_wday = (jd+1)%7;

	newyears = date;
	newyears.tm_mon = 0;
	newyears.tm_mday = 1;
	date.tm_yday = jd - _juldnj(&newyears, Transition);
	return &date;
}

long juldnj (struct tm *bdt, long Transition)
{	long jd;
	struct tm *normal;

	jd = _juldnj(bdt, Transition);

	normal = julcdj(jd, Transition);
	bdt->tm_year = normal->tm_year;
	bdt->tm_mon = normal->tm_mon;
	bdt->tm_mday = normal->tm_mday;
	bdt->tm_wday = normal->tm_wday;
	bdt->tm_yday = normal->tm_yday;

	return jd;
}


  XplTime::XplTime (const XplIndex newdim) : XplArray<long>(newdim)
  { granularity = Xsec; 
    zeroday     = 2440588L;
  };

  XplTime::XplTime (long d0, long d1, long d2, long d3, long d4, long d5, long d6, long d7) : XplArray<long>(d0, d1, d2, d3, d4, d5, d6, d7) 
  { granularity = Xsec;
    zeroday     = 2440588L;
  };

  XplTime::~XplTime () { };

/*  XplTime::XplTime (const XplTime& x) : XplArray<long>(x)
  { granularity = x.granularity;
    zeroday     = x.zeroday;
  }
  */
  XplTime::XplTime (const XplArray<long>& x) : XplArray<long>(x) 
  { granularity = Xsec; 
    zeroday     = 2440588L;
  }

//  XplTime& operator= (XplArray<long>& x)		
//  { assignment (x);
//    return (*this);
//  }  

  void XplTime::SetXplTimeGran (XGran newgran)
  { if ((newgran==Xweek) || (newgran==Xmonth) || (newgran==Xyear))
    { XAerrort (TIMEGRAN, "settimegran");
      return; 
    }
    granularity = newgran; 
  }

  void XplTime::SetXplTimeZero (long newzeroday)
  { zeroday = newzeroday; }

  XplTime& XplTime::time (const XplArray<long> &newtime, XGran newgran)
  { long *newtimeptr, i, s;
    if ((newgran==Xweek) || (newgran==Xmonth) || (newgran==Xyear))
    { XAerrort (TIMEGRAN, "time");
      return (*this); 
    }
    granularity = newgran;
    s = newtime.size();
    newtimeptr = newtime.getptr();
    if (ptr) delete[] ptr;
    dim = newtime.getdim();
    ptr = new long[s];
    for (i=0; i<s; i++)
    { // printf ("%i\n", i);
      *(ptr+i) = *(newtimeptr+i); 
    }
    return (*this); 
  }

  XplTime& XplTime::time (const XplArray<long> &newyear, const XplArray<long> *newmonth, 
                 const XplArray<long> *newday, const XplArray<long> *newhour,
                 const XplArray<long> *newminute, const XplArray<long> *newsecond)
  { long i, s, *yearptr, *monthptr, *dayptr, *hourptr, *minuteptr, *secondptr;
    struct tm tmp;
    for (i=0; i<MAX_XARRAY; i++)
    { if (newmonth)
      { if (newmonth->dim(i)!=newyear.dim(i))
        { XAerrort (DIMCON, "time: layer(year) <> layer(month)"); return (*this); }  
      }
      if (newday)
      { if (newday->dim(i)!=newyear.dim(i))
        { XAerrort (DIMCON, "time: layer(year) <> layer(day)"); return (*this); }
      }
      if (newhour)
      { if (newhour->dim(i)!=newyear.dim(i))
        { XAerrort (DIMCON, "time: layer(year) <> layer(hour)"); return (*this); }
      }
      if (newminute)
      { if (newminute->dim(i)!=newyear.dim(i))
        { XAerrort (DIMCON, "time: layer(year) <> layer(minute)"); return (*this); }
      }
      if (newsecond)
      { if (newsecond->dim(i)!=newyear.dim(i))
        { XAerrort (DIMCON, "time: layer(year) <> layer(second)"); return (*this); }
      }
    }
    dim = newyear.dim;
    s   = dim.prod();
    if (ptr) delete[] ptr;
    ptr = new long[s];
    yearptr   = newyear.getptr();
    monthptr  = (long *) (newmonth ? newmonth->getptr() : NULL);
    dayptr    = (long *) (newday ? newday->getptr() : NULL);
    hourptr   = (long *) (newhour ? newhour->getptr() : NULL);
    minuteptr = (long *) (newminute ? newminute->getptr() : NULL);
    secondptr = (long *) (newsecond ? newsecond->getptr() : NULL);
    granularity = Xday;
    if (newhour)   granularity = Xhour;
    if (newminute) granularity = Xminute;
    if (newsecond) granularity = Xsec;

    for (i=0; i<s; i++)
    { tmp.tm_sec   = (secondptr ? *(secondptr+i)%62 : 0);
      tmp.tm_min   = (minuteptr ? *(minuteptr+i)%60 : 0);
      tmp.tm_hour  = (hourptr   ? *(hourptr+i)%24   : 0);
      tmp.tm_mday  = (dayptr    ? 1+((*(dayptr+i)-1)%31) : 1);
      tmp.tm_mon   = (monthptr  ? (*(monthptr+i)-1)%12 : 0);
      tmp.tm_year  = *(yearptr+i) - 1900;
      *(ptr+i) = juldnj (&tmp) - zeroday;
      switch (granularity)
      { case Xsec    : *(ptr+i) = *(ptr+i) * 86400L + tmp.tm_hour * 3600L + 
                                  tmp.tm_min * 60L + tmp.tm_sec; break;
        case Xminute : *(ptr+i) = *(ptr+i) * 1440L + tmp.tm_hour * 60 +
                                  tmp.tm_min; break;
        case Xhour   : *(ptr+i) = *(ptr+i) * 24L + tmp.tm_hour; break;
      }
    }

    return (*this);
  }
  
  XplChar XplTime::year () const 
  { XplChar ret (dim);
    xplstring *retptr;
    long i, s, tmp;
    char buffer[128];
    struct tm *tmptr;
    s = dim.prod();
    retptr = ret.getptr();
    switch (granularity)
    { case Xsec:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 86400L + zeroday;
          if ((*(ptr+i)%86400L)<0) tmp--;
          tmptr = julcdj (tmp);
          sprintf (buffer, "%i", 1900+tmptr->tm_year);
          *(retptr+i) = xplstring(buffer);
        }
        break;
      case Xminute:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 1440L + zeroday;
          if ((*(ptr+i)%1440L)<0) tmp--;
          tmptr = julcdj (tmp);
          sprintf (buffer, "%i", 1900+tmptr->tm_year);
          *(retptr+i) = xplstring(buffer);
        }
        break;
      case Xhour:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 24L + zeroday;
          if ((*(ptr+i)%24L)<0) tmp--;
          tmptr = julcdj (tmp);
          sprintf (buffer, "%i", 1900+tmptr->tm_year);
          *(retptr+i) = xplstring(buffer);
        }
        break; 
      case Xday:
        for (i=0; i<s; i++)
        { tmptr = julcdj (*(ptr+i) + zeroday);
          sprintf (buffer, "%i", 1900+tmptr->tm_year);
          *(retptr+i) = xplstring(buffer);
        }
        break;
    }
    return (ret);
  }

  XplChar XplTime::month () const
  { XplChar ret (dim);
    xplstring *retptr;
    long i, s, tmp;
    char buffer[128];
    struct tm *tmptr;

    s = dim.prod();
    retptr = ret.getptr();
    switch (granularity)
    { case Xsec:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 86400L + zeroday;
          if ((*(ptr+i)%86400L)<0) tmp--;
          tmptr = julcdj (tmp);
          *(retptr+i) = xplstring(Smonth[tmptr->tm_mon]);
        }
        break;
      case Xminute:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 1440L + zeroday;
          if ((*(ptr+i)%1440L)<0) tmp--;
          tmptr = julcdj (tmp);
          *(retptr+i) = xplstring(Smonth[tmptr->tm_mon]);
        }
        break;
      case Xhour:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 24L + zeroday;
          if ((*(ptr+i)%24L)<0) tmp--;
          tmptr = julcdj (tmp);
          *(retptr+i) = xplstring(Smonth[tmptr->tm_mon]);
        }
        break; 
      case Xday:
        for (i=0; i<s; i++)
        { tmptr = julcdj (*(ptr+i) + zeroday);
          *(retptr+i) = xplstring(Smonth[tmptr->tm_mon]);
        }
        break;
    }
    return (ret);
  }

  XplChar XplTime::dayofmonth () const
  { XplChar ret (dim);
    xplstring *retptr;
    long i, s, tmp;
    char buffer[128];
    struct tm *tmptr;
    s = dim.prod();
    retptr = ret.getptr();
    switch (granularity)
    { case Xsec:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 86400L + zeroday;
          if ((*(ptr+i)%86400L)<0) tmp--;
          tmptr = julcdj (tmp);
          sprintf (buffer, "%02i", tmptr->tm_mday);
          *(retptr+i) = xplstring(buffer);
        }
        break;
      case Xminute:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 1440L + zeroday;
          if ((*(ptr+i)%1440L)<0) tmp--;
          tmptr = julcdj (tmp);
          sprintf (buffer, "%02i", tmptr->tm_mday);
          *(retptr+i) = xplstring(buffer);
        }
        break;
      case Xhour:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 24L + zeroday;
          if ((*(ptr+i)%24L)<0) tmp--;
          tmptr = julcdj (tmp);
          sprintf (buffer, "%02i", tmptr->tm_mday);
          *(retptr+i) = xplstring(buffer);
        }
        break; 
      case Xday:
        for (i=0; i<s; i++)
        { tmptr = julcdj (*(ptr+i) + zeroday);
          sprintf (buffer, "%02i", tmptr->tm_mday);
          *(retptr+i) = xplstring(buffer);
        }
        break;
    }
    return (ret);
  }

  XplChar XplTime::dayofweek () const
  { XplChar ret (dim);
    xplstring *retptr;
    long i, s, tmp;
    char buffer[128];
    struct tm *tmptr;
    s = dim.prod();
    retptr = ret.getptr();
    switch (granularity)
    { case Xsec:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 86400L + zeroday;
          if ((*(ptr+i)%86400L)<0) tmp--;
          tmptr = julcdj (tmp);
          *(retptr+i) = xplstring(Sday[tmptr->tm_wday]);
        }
        break;
      case Xminute:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 1440L + zeroday;
          if ((*(ptr+i)%1440L)<0) tmp--;
          tmptr = julcdj (tmp);
          *(retptr+i) = xplstring(Sday[tmptr->tm_wday]);
        }
        break;
      case Xhour:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 24L + zeroday;
          if ((*(ptr+i)%24L)<0) tmp--;
          tmptr = julcdj (tmp);
          *(retptr+i) = xplstring(Sday[tmptr->tm_wday]);
        }
        break; 
      case Xday:
        for (i=0; i<s; i++)
        { tmptr = julcdj (*(ptr+i) + zeroday);
          *(retptr+i) = xplstring(Sday[tmptr->tm_wday]);
        }
        break;
    }
    return (ret);
  }

  XplChar XplTime::hour () const
  { XplChar ret (dim);
    xplstring *retptr;
    long i, s, tmp;
    char buffer[128];
    struct tm *tmptr;

    s = dim.prod();
    retptr = ret.getptr();
    switch (granularity)
    { case Xsec:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 3600L;
          if ((*(ptr+i)%3600L)<0) tmp--;
          tmp %= 24L;
          if (tmp<0) tmp += 24L;
          sprintf (buffer, "%02i", tmp);
          *(retptr+i) = xplstring(buffer);
        }
        break;
      case Xminute:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 60L;
          if ((*(ptr+i)%60L)<0) tmp--;
          tmp %= 24L;
          if (tmp<0) tmp += 24L;
          sprintf (buffer, "%02i", tmp);
          *(retptr+i) = xplstring(buffer);
        }
        break;
      case Xhour:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) % 24L;
          if (tmp<0) tmp += 24L;
          sprintf (buffer, "%02i", tmp);
          *(retptr+i) = xplstring(buffer);
        }
        break; 
      case Xday:
        for (i=0; i<s; i++)
        { *(retptr+i) = xplstring("00"); }
        break;
    }
    return (ret);
  }

  XplChar XplTime::minute () const
  { XplChar ret (dim);
    xplstring *retptr;
    long i, s, tmp;
    char buffer[128];
    struct tm *tmptr;

    s = dim.prod();
    retptr = ret.getptr();
    switch (granularity)
    { case Xsec:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 60L;
          if ((*(ptr+i)%60L)<0) tmp--;
          tmp %= 60L;
          if (tmp<0) tmp += 60L;
          sprintf (buffer, "%02i", tmp);
          *(retptr+i) = xplstring(buffer);
        }
        break;
      case Xminute:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) % 60L;
          if (tmp<0) tmp += 60L;
          sprintf (buffer, "%02i", tmp);
          *(retptr+i) = xplstring(buffer);
        }
        break;
      case Xhour:
        for (i=0; i<s; i++)
        { *(retptr+i) = xplstring("00"); }
        break;
      case Xday:
        for (i=0; i<s; i++)
        { *(retptr+i) = xplstring("00"); }
        break;
    }
    return (ret);
  }

  XplChar XplTime::second () const
  { XplChar ret (dim);
    xplstring *retptr;
    long i, s, tmp;
    char buffer[128];
    struct tm *tmptr;

    s = dim.prod();
    retptr = ret.getptr();
    switch (granularity)
    { case Xsec:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) % 60L;
          if (tmp<0) tmp += 60L;
          sprintf (buffer, "%02i", tmp);
          *(retptr+i) = xplstring(buffer);
        }
        break;
      case Xminute:
        for (i=0; i<s; i++)
        { *(retptr+i) = xplstring("00"); }
        break;
      case Xhour:
        for (i=0; i<s; i++)
        { *(retptr+i) = xplstring("00"); }
        break;
      case Xday:
        for (i=0; i<s; i++)
        { *(retptr+i) = xplstring("00"); }
        break;
    }
    return (ret);
  }

  XplArray<long> XplTime::nyear () const
  { XplArray<long> ret (dim);
    long *retptr;
    long i, s, tmp;
    struct tm *tmptr;
    s = dim.prod();
    retptr = ret.getptr();
    switch (granularity)
    { case Xsec:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 86400L + zeroday;
          if ((*(ptr+i)%86400L)<0) tmp--;
          tmptr = julcdj (tmp);
          *(retptr+i) = 1900+tmptr->tm_year;
        }
        break;
      case Xminute:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 1440L + zeroday;
          if ((*(ptr+i)%1440L)<0) tmp--;
          tmptr = julcdj (tmp);
          *(retptr+i) = 1900+tmptr->tm_year;
        }
        break;
      case Xhour:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 24L + zeroday;
          if ((*(ptr+i)%24L)<0) tmp--;
          tmptr = julcdj (tmp);
          *(retptr+i) = 1900+tmptr->tm_year;
        }
        break; 
      case Xday:
        for (i=0; i<s; i++)
        { tmptr = julcdj (*(ptr+i) + zeroday);
          *(retptr+i) = 1900+tmptr->tm_year;
        }
        break;
    }
    return (ret);
  }

  XplArray<long> XplTime::nmonth () const
  { XplArray<long> ret (dim);
    long *retptr;
    long i, s, tmp;
    struct tm *tmptr;

    s = dim.prod();
    retptr = ret.getptr();
    switch (granularity)
    { case Xsec:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 86400L + zeroday;
          if ((*(ptr+i)%86400L)<0) tmp--;
          tmptr = julcdj (tmp);
          *(retptr+i) = tmptr->tm_mon+1;
        }
        break;
      case Xminute:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 1440L + zeroday;
          if ((*(ptr+i)%1440L)<0) tmp--;
          tmptr = julcdj (tmp);
          *(retptr+i) = tmptr->tm_mon+1;
        }
        break;
      case Xhour:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 24L + zeroday;
          if ((*(ptr+i)%24L)<0) tmp--;
          tmptr = julcdj (tmp);
          *(retptr+i) = tmptr->tm_mon+1;
        }
        break; 
      case Xday:
        for (i=0; i<s; i++)
        { tmptr = julcdj (*(ptr+i) + zeroday);
          *(retptr+i) = tmptr->tm_mon+1;
        }
        break;
    }
    return (ret);
  }

  XplArray<long> XplTime::ndayofmonth () const
  { XplArray<long> ret (dim);
    long *retptr;
    long i, s, tmp;
    struct tm *tmptr;
    s = dim.prod();
    retptr = ret.getptr();
    switch (granularity)
    { case Xsec:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 86400L + zeroday;
          if ((*(ptr+i)%86400L)<0) tmp--;
          tmptr = julcdj (tmp);
          *(retptr+i) = tmptr->tm_mday;
        }
        break;
      case Xminute:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 1440L + zeroday;
          if ((*(ptr+i)%1440L)<0) tmp--;
          tmptr = julcdj (tmp);
          *(retptr+i) = tmptr->tm_mday;
        }
        break;
      case Xhour:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 24L + zeroday;
          if ((*(ptr+i)%24L)<0) tmp--;
          tmptr = julcdj (tmp);
          *(retptr+i) = tmptr->tm_mday;
        }
        break; 
      case Xday:
        for (i=0; i<s; i++)
        { tmptr = julcdj (*(ptr+i) + zeroday);
          *(retptr+i) = tmptr->tm_mday;
        }
        break;
    }
    return (ret);
  }

  XplArray<long> XplTime::ndayofweek () const
  { XplArray<long> ret (dim);
    long *retptr;
    long i, s, tmp;
    struct tm *tmptr;
    s = dim.prod();
    retptr = ret.getptr();
    switch (granularity)
    { case Xsec:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 86400L + zeroday;
          if ((*(ptr+i)%86400L)<0) tmp--;
          tmptr = julcdj (tmp);
          *(retptr+i) = tmptr->tm_wday+1;
        }
        break;
      case Xminute:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 1440L + zeroday;
          if ((*(ptr+i)%1440L)<0) tmp--;
          tmptr = julcdj (tmp);
          *(retptr+i) = tmptr->tm_wday+1;
        }
        break;
      case Xhour:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 24L + zeroday;
          if ((*(ptr+i)%24L)<0) tmp--;
          tmptr = julcdj (tmp);
          *(retptr+i) = tmptr->tm_wday+1;
        }
        break; 
      case Xday:
        for (i=0; i<s; i++)
        { tmptr = julcdj (*(ptr+i) + zeroday);
          *(retptr+i) = tmptr->tm_wday+1;
        }
        break;
    }
    return (ret);
  }

  XplArray<long> XplTime::nhour () const 
  { XplArray<long> ret (dim);
    long *retptr;
    long i, s, tmp;
    struct tm *tmptr;

    s = dim.prod();
    retptr = ret.getptr();
    switch (granularity)
    { case Xsec:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 3600L;
          if ((*(ptr+i)%3600L)<0) tmp--;
          tmp %= 24L;
          if (tmp<0) tmp += 24L;
          *(retptr+i) = tmp;
        }
        break;
      case Xminute:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 60L;
          if ((*(ptr+i)%60L)<0) tmp--;
          tmp %= 24L;
          if (tmp<0) tmp += 24L;
          *(retptr+i) = tmp;
        }
        break;
      case Xhour:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) % 24L;
          if (tmp<0) tmp += 24L;
          *(retptr+i) = tmp;
        }
        break; 
      case Xday:
        for (i=0; i<s; i++)
        { *(retptr+i) = 0; }
        break;
    }
    return (ret);
  }

  XplArray<long> XplTime::nminute () const
  { XplArray<long> ret (dim);
    long *retptr;
    long i, s, tmp;
    struct tm *tmptr;

    s = dim.prod();
    retptr = ret.getptr();
    switch (granularity)
    { case Xsec:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) / 60L;
          if ((*(ptr+i)%60L)<0) tmp--;
          tmp %= 60L;
          if (tmp<0) tmp += 60L;
          *(retptr+i) = tmp;
        }
        break;
      case Xminute:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) % 60L;
          if (tmp<0) tmp += 60L;
          *(retptr+i) = tmp;
        }
        break;
      case Xhour:
        for (i=0; i<s; i++)
        { *(retptr+i) = 0; }
        break;
      case Xday:
        for (i=0; i<s; i++)
        { *(retptr+i) = 0; } 
        break;
    }
    return (ret);
  }

  XplArray<long> XplTime::nsecond () const
  { XplArray<long> ret (dim);
    long *retptr;
    long i, s, tmp;
    struct tm *tmptr;

    s = dim.prod();
    retptr = ret.getptr();
    switch (granularity)
    { case Xsec:
        for (i=0; i<s; i++)
        { tmp = *(ptr+i) % 60L;
          if (tmp<0) tmp += 60L;
          *(retptr+i) = tmp; 
        }
        break;
      case Xminute:
        for (i=0; i<s; i++)
        { *(retptr+i) = 0; } 
        break;
      case Xhour:
        for (i=0; i<s; i++)
        { *(retptr+i) = 0; } 
        break;
      case Xday:
        for (i=0; i<s; i++)
        { *(retptr+i) = 0; } 
        break;
    }
    return (ret);
  }

  XplArray<long> XplTime::getlong (void) const
  { XplArray<long> ret(dim);
    long *lptr;
    lptr = ret.getptr();
    ::memcopy (lptr, ptr, size());
    return (ret);
  }

  XplTime XplTime::fill (XGran stepgran, long elem, long step)
  { long i, r;
    XplTime ret(elem);
    struct tm *tmp;
    if (size()!=1)
    { XAerrort(FORMAT, "fill : size(origin)<>1"); return (ret); }
    else
    { ret.SetXplTimeGran (granularity);
      ret.SetXplTimeZero (zeroday);
      switch (granularity)
      { case Xsec    : tmp = julcdj (*ptr / 86400L + zeroday); 
                       r   = *ptr % 86400L;
                       tmp->tm_hour = r / 3600L;
                       r  %= 3600L;
                       tmp->tm_min = r / 60L;
                       r  %= 60L;
                       tmp->tm_sec = r;
                       break;
        case Xminute : tmp = julcdj (*ptr / 1440L + zeroday); 
                       r   = *ptr % 1440L;
                       tmp->tm_hour = r / 60L;
                       r  %= 60L;
                       tmp->tm_min = r;
                       tmp->tm_sec = 0;
                       break;
        case Xhour   : tmp = julcdj (*ptr / 24L + zeroday); 
                       r   = *ptr % 24L;
                       tmp->tm_hour = r;
                       tmp->tm_min = 0;
                       tmp->tm_sec = 0;
                       break;
        case Xday    : tmp = julcdj (*ptr + zeroday); 
                       tmp->tm_hour = 0;
                       tmp->tm_min = 0;
                       tmp->tm_sec = 0;
                       break;
      }
      for (i=0; i<elem; i++)
      { if (i)
        { switch (stepgran)
          { case Xsec    : tmp->tm_sec  += step; break;
            case Xminute : tmp->tm_min  += step; break;
            case Xhour   : tmp->tm_hour += step; break;
            case Xday    : tmp->tm_yday += step; break;
            case Xweek   : tmp->tm_yday += 7*step; break;
            case Xmonth  : tmp->tm_mon  += step; break;
            case Xyear   : tmp->tm_year += step; break;
          }
        }
        ret[i] = juldnj (tmp) - zeroday;
        switch (granularity)
        { case Xsec    : ret[i] = ret[i]*86400L+tmp->tm_hour*3600L+
                                   tmp->tm_min*60L+tmp->tm_sec; break;
          case Xminute : ret[i] = ret[i]*1440L+tmp->tm_hour*60L+
                                   tmp->tm_min; break;
          case Xhour   : ret[i] = ret[i]*24L+tmp->tm_hour; break;
        }
      }
    } 
    return (ret);
  }

  void XplTime::mdump (const char *fmt, Xplostream *stream, const char *varname,
              int pdim) const 
  { char out[256], *cgran;
    long *subptr;
    unsigned long count, maxcount;
    long i, j, k, l, s, m, d, starfmt;
    struct tm *tmptr;

    XplIndex tmp;
    Xplostream streammdump(stream);
    char 
      *outheadline = property.getProperty("outheadline"), 
      *outlineno = property.getProperty("outlineno"), 
      *outlayerline = property.getProperty("outlayerline");
    long outmaxdata = atol (property.getProperty("outmaxdata"));

    if (size()<=0) 
    { XAerrort (ARRAYBAD, "Arraysize <= 0");
      return;
    }
    l = dim(0)*dim(1);    
    s = dim.prod()/l;
    m = (long) (1.0+x4log10((double) (dim(0))));

    if (varname)
      sprintf (out, outheadline, varname);
    else
      sprintf (out, outheadline, "unnamed object");

    streammdump << out;
    starfmt = (strchr(outlineno, '*') != NULL);
    maxcount = (outmaxdata<1 ? 0xffffffff : outmaxdata); 
    count    = 0L;

    for (tmp=0, k=0; (k<s) && (count<maxcount); k++)
    { subptr = ptr+k*l;
      if (pdim)
      { if (s>1)
        { sprintf (out, outlayerline, 
                    tmp(2)+1,tmp(3)+1,tmp(4)+1,tmp(5)+1,tmp(6)+1,tmp(7)+1);
          streammdump << out;
        }
      }
      for (i=0; (i<dim(0)) && (count<maxcount); i++)
      { if (starfmt)
          sprintf (out, outlineno, m, i+1);
        else
          sprintf (out, outlineno, i+1);
        streammdump << out;

        for (j=0; (j<dim(1)) && (count<maxcount); j++, tmp.inc(dim))
        { d = *(subptr+j*dim(0)+i);
          switch (granularity)
          { case Xsec    : d = d/86400L + zeroday; break;
            case Xminute : d = d/1440L  + zeroday; break;
            case Xhour   : d = d/24L    + zeroday; break;
          }
          tmptr = julcdj (d);
          sprintf (out, "%5i ", tmptr->tm_year);
          streammdump << out;
          sprintf (out, "%2i ", 1+tmptr->tm_mon);
          streammdump << out;
          sprintf (out, "%2i ", tmptr->tm_mday);
          streammdump << out;
          d = *(subptr+j*dim(0)+i);
          switch (granularity)
          { case Xday    : d = 0L; break;
            case Xhour   : d %= 24;   d *= 3600L; break;
            case Xminute : d %= 1440; d *= 60L; break;
            case Xsec    : d %= 86400; break;
          }
          sprintf (out, "%02i ", d/1440);
          streammdump << out;
          d %= 1440;
          sprintf (out, "%02i ", d/60);
          streammdump << out;
          d %= 60;
          sprintf (out, "%02i ", d);
          streammdump << out;
        }
        streammdump << "\n";
      }
    } 
    if (count==maxcount)
      streammdump << "\r\nSorry, this object is too big !!\r\n";
  }

  void XplTime::now (XGran newgran, long newzeroday)
  { struct tm *yet_a_second_later;
    time_t tptr;
    dim = 1;
    if (ptr) delete[] ptr;
    granularity = newgran;
    zeroday     = newzeroday;
    ptr         = new long[1];
    *ptr        = 0;
    ::time(&tptr);
    yet_a_second_later = gmtime(&tptr);
    switch (granularity)
    { case Xsec : *ptr = 86400 *(juldnj(yet_a_second_later) - zeroday)+
                            3600 * yet_a_second_later->tm_hour +
                            60 * yet_a_second_later->tm_min +
                            yet_a_second_later->tm_sec;
                     break;
      case Xminute : *ptr = 1440 *(juldnj(yet_a_second_later) - zeroday)+
                            60 * yet_a_second_later->tm_hour +
                            yet_a_second_later->tm_min;
                     break;
      case Xhour   : *ptr = 24 *(juldnj(yet_a_second_later) - zeroday)+
                            yet_a_second_later->tm_hour;
                     break;
      case Xday    : *ptr = juldnj(yet_a_second_later) - zeroday;
                     break;
      default      : XAerrort (TIMEGRAN, "now");
                     break;
    }
  }

  void XplTime::nowlocal (XGran newgran, long newzeroday)
  { struct tm *yet_a_second_later;
    time_t tptr;
    dim = 1;
    if (ptr) delete[] ptr;
    granularity = newgran;
    zeroday     = newzeroday;
    ptr         = new long[1];
    *ptr        = 0;
    ::time(&tptr);
    yet_a_second_later = localtime(&tptr);
    switch (granularity)
    { case Xsec : *ptr = 86400 *(juldnj(yet_a_second_later) - zeroday)+
                            3600 * yet_a_second_later->tm_hour +
                            60 * yet_a_second_later->tm_min +
                            yet_a_second_later->tm_sec;
                     break;
      case Xminute : *ptr = 1440 *(juldnj(yet_a_second_later) - zeroday)+
                            60 * yet_a_second_later->tm_hour +
                            yet_a_second_later->tm_min;
                     break;
      case Xhour   : *ptr = 24 *(juldnj(yet_a_second_later) - zeroday)+
                            yet_a_second_later->tm_hour;
                     break;
      case Xday    : *ptr = juldnj(yet_a_second_later) - zeroday;
                     break;
      default      : XAerrort (TIMEGRAN, "now");
                     break;
    }
  }

  int XplTime::WriteXArray (FILE *f) const 
  { long i, s = size(), type = TYPE_TIME, md = MAX_XARRAY, adim;
    XplIndex d = getdim();
    fwrite (&type, sizeof(long), 1, f);
    fwrite (&md, sizeof(long), 1, f);
    for (i=0; i<MAX_XARRAY; i++)
    { adim = d(i);
      fwrite (&adim, sizeof(long), 1, f);
    }
    fwrite (&granularity, sizeof(XGran), 1, f);
    fwrite (&zeroday, sizeof(long), 1, f);
    fwrite (ptr, sizeof(long), s, f);
    return 0;
  }





	
