/* $Header: /home/xplore/cvs/xpl4gr/xreal.C,v 1.5 2003/05/15 08:27:10 cizek Exp $ */

#ifndef __BORLANDC__
#include "x4math2.h"
#endif
#include "xreal.hpp"
#include "xnan.h"

#include "chi2.h"
#include "eispack.h"
#include "xplore_strings.h"

#define EPSILON 1.0e-12

#ifdef HAS_STRCMPI
#define strcasecmp strcmpi
#endif

double x4acos (double x)
{ return (fabs(x)>1 ? sqrt(-1.0) : acos(x)); }

/**** Borland C++  */
#ifdef __BORLANDC__ 
  static double dnan = 0.0/0.0, dinf = 1.0/0.0, lgdnan = 0.0/0.0;
#endif

#ifdef __GNUC__
  static double lgdnan = 0.0/0.0, dinf = 1.0/0.0;
#endif

#ifdef __BORLANDC__
static double x4exp(double x) {
  if (XisNaN(x)) return(x);
  if (XisNegInf(x)) return(0);
  if (XisPosInf(x)) return(x);
  return(exp(x));
}

static double x4expm1(double x) {
  //  if (XisNaN(x)) return(x);
  if (XisNegInf(x)) return(-1);
  if (XisPosInf(x)) return(x);
  return(exp(x));
}

static double x4tanh(double x) {
  if (XisNaN(x)) return(x);
  if (XisPosInf(x)) return(1);
  if (XisNegInf(x)) return(-1);
  return(tanh(x));
}

static double x4cosh(double x) {
  if (XisNaN(x)) return(x);
  if (XisInf(x)) return(dinf);
  return(cosh(x));
}

static double x4asin(double x) {
  if (x<-1 || x>1) return (dnan);
  if (XisInf(x)) return (dnan);
  return(asin(x));
}

static double x4log(double x) {
  if (x<0) return (dnan);
  return (log(x));
}
#endif

//#ifdef HAS_SMPMATH

#include "x4math.h"

double gammal (double xx)
{ 
  if (xx<=0) return(lgdnan);
  if (xx==1) return(0);
  if (XisPosInf(xx)) return(xx);
  if (XisNegInf(xx)) return(lgdnan);
  // obige zeilen lschen um wieder originalcode zuerhalten
  double cof[6] = {76.18009173,    -86.50532033,  24.01409822,
             -1.234739516, 0.120858003e-2, -0.536382e-5},
      fact = 2.50662827465, x, result, temp, ser;
  int i;

  if (xx > 1.0)
  { x = xx - 1.0;
     result = 0.0;
  }                         /* for x>1 we have full accuracy  */
  else
  { x = xx;
     result = -x4log(xx);
  }                         /* x<1 use gamma(x+1)=x*gamma(x)  */
  temp = x + 5.5;
  temp = (x + 0.5) * x4log(temp) - temp;
  ser = 1.0;
  for (i = 0; i < 6; i++)
  { x = x + 1.0;
     ser = ser + cof[i] / x;
  }
  result = result + temp + x4log(fact * ser);
  return(result);
}


// um wieder lgamma zu erhalten:
//
// ifs raus
// PC MAC
#ifndef __APPLE_CC__
double lgamma (double xx) {
  /*  if (xx<=0) return(dnan);
  if (xx==1) return(0);
  if (XisPosInf(xx)) return(xx);
  if (XisNegInf(xx)) return(dnan);*/
  return (gammal(xx));
}
#endif
//#endif

#ifdef HAS_SMPMATH
double acosh (double x)
{ return(x4log(x+sqrt(x*x-1.0))); }

double asinh (double x) {
  if (XisNegInf(x)) return(x);
	return(x4log(x+sqrt(1.0+x*x))); }

double atanh (double x)
{ return(0.5*x4log((1.0+x)/(1.0-x))); }

//double rint (double x)
//{ return(ceil(x-0.5)); }

#define ITMAX 100
#define EPSSk 3.0e-7

void gser(double *gamser, double a, double x, double *gln)
{ int n;
  double sum,del,ap;

  *gln = gammal(a);
  if (x <= 0.0)
  { if (x < 0.0)
    { *gamser = sqrt(-1);
      return;
    }
    *gamser=0.0;
    return;
  }
  else
  { ap  = a;
    del = sum = 1.0/a;
    for (n=1;n<=ITMAX;n++)
    { ap  += 1.0;
      del *= x/ap;
      sum += del;
      if (fabs(del) < fabs(sum)*EPSSk)
      { *gamser = sum*x4exp(-x+a*x4log(x)-(*gln));
	return;
      }
    }
    *gamser = sqrt(-1);
    return;
  }
}

void gcf(double *gammcf, double a, double x, double *gln)
{ int n;
  double gold=0.0,g,fac=1.0,b1=1.0;
  double b0=0.0,anf,ana,an,a1,a0=1.0;

  *gln = gammal(a);
  a1=x;
  for (n=1;n<=ITMAX;n++)
  { an  = (double) n;
    ana = an-a;
    a0  = (a1+a0*ana)*fac;
    b0  = (b1+b0*ana)*fac;
    anf = an*fac;
    a1  = x*a0+anf*a1;
    b1  = x*b0+anf*b1;
    if (a1)
    { fac = 1.0/a1;
      g   = b1*fac;
      if (fabs((g-gold)/g) < EPSSk)
      { *gammcf = x4exp(-x+a*x4log(x)-(*gln))*g;
	 return;
      }
      gold=g;
    }
  }
  *gammcf = sqrt(-1);
}

#undef ITMAX
#undef EPSSk

double gammp(double a, double x)
{ double gamser, gammcf, gln;

  if (x < 0.0 || a <= 0.0) return (sqrt(-1));
  if (x < (a+1.0))
  { gser(&gamser,a,x,&gln);
    return (gamser);
  }
  else
  { gcf(&gammcf,a,x,&gln);
    return (1.0-gammcf);
  }
}

#ifndef __APPLE_CC__
double erf (double x)
{ return ((x<0.0) ? -gammp(0.5,x*x) : gammp(0.5,x*x)); }

double erfc (double x)
{ return (1-erf(x)); }
#endif

double expm1 (double x)
{ return (x4exp(x)-1); }

double log1p (double x)
{ return (x4log(1+x)); }

double rint (double x)
{ double flory = floor(x);
  return (x-flory<0.5 ? flory : flory+1);
}

double logb (double x)
{ return (x4log(x)/M_LN2); }

#endif

#ifndef HAS_GAMMA
double gamma (double x)
#ifdef HAS_SMPMATH
{ return (x4exp(gammal(x))); }
#else
{ return (x4exp(lgamma(x))); }
#endif
#endif

double suniform (double x)
{ if (fabs(x)>1.0) return 0.0;
  return (0.5);
}

double striangle (double x)
{ if (fabs(x)>1.0) return 0.0;
   return (1-fabs(x));
}

double sepanechnikov (double x)
{ if (fabs(x)>1.0) return 0.0;
  return (0.75*(1.0-x*x));
}

double squartic (double x)
{ if (fabs(x)>1.0) return 0.0;
  return (0.9375*(1.0-x*x)*(1.0-x*x));
}

double striweight (double x)
{ if (fabs(x)>1.0) return 0.0;
  return (35.0/32.0*(1.0-x*x)*(1.0-x*x)*(1.0-x*x));
}

void triangle (double f1, double f2, double f3,
                      double x1, double x2, double x3,
                      double y1, double y2, double y3,
                      double cut, double *lx, double *ly,
                      long *ln)
{ int ip = 0;
  double ratio;

  if (f1==f2)
  { if (f1==cut)
    { lx[*ln] = x1;  
      ly[*ln] = y1;  
      (*ln)++;
      ip++;
      lx[*ln] = x2;  
      ly[*ln] = y2;  
      (*ln)++;
      ip++;
    }
  }
  else
  { ratio = (cut-f1)/(f2-f1);
    if ((0.0<=ratio) && (ratio<1.0))
    { lx[*ln] = x1+ratio*(x2-x1);  
      ly[*ln] = y1+ratio*(y2-y1);  
      (*ln)++;
      ip++;
    }
  }

  if (f3==f2)
  { if (f3==cut)
    { lx[*ln] = x3;  
      ly[*ln] = y3;  
      (*ln)++;
      ip++;
      lx[*ln] = x2;  
      ly[*ln] = y2;  
      (*ln)++;
      ip++;
    }
  }
  else
  { ratio = (cut-f2)/(f3-f2);
    if ((0.0<=ratio) && (ratio<1.0))
    { lx[*ln] = x2+ratio*(x3-x2);  
      ly[*ln] = y2+ratio*(y3-y2);  
      (*ln)++;
      ip++;
    }
  }

  if (f3==f1)
  { if (f1==cut)
    { lx[*ln] = x3;  
      ly[*ln] = y3;  
      (*ln)++;
      ip++;
      lx[*ln] = x1;
      ly[*ln] = y1;  
      (*ln)++;
      ip++;
    }
  }
  else
  { ratio = (cut-f3)/(f1-f3);
    if ((0.0<=ratio) && (ratio<1.0))
    { lx[*ln] = x3+ratio*(x1-x3);  
      ly[*ln] = y3+ratio*(y1-y3);  
      (*ln)++;
      ip++;
    }
  }

  if (ip%2) (*ln)--;
}

int notpowof2 (long n)
{ long m = 1;
  while (m!=0)
  { if (m==n) return (0);
    m <<= 1;
  }
  return (1);
}


  int ludcmp (double *lu, long n, long *indx, double *d)
  { long i,imax,j,k;
    double aamax,dum,sum,temp;
    double *vv;

#ifndef MEMDEBUG
    vv = new double[n];
#else
    vv = (double *) xmem.newptr (sizeof(double), n, "XplReal", "ludcmp", "vv");
#endif  

    *d=1.0;

    for (i=0; i<n; i++)
      *(vv+i) = 0.0;
    for (i=0; i<n; i++)
    { for (j=0; j<n; j++)
        *(vv+j) = max1 (::fabs (*(lu+n*i+j)), *(vv+j));
    }
    for (i=0; i<n; i++)
    { // printf ("vv[%i]=%li\n", i, vv[i]);
      if (isZero(*(vv+i)))
      { 
#ifndef MEMDEBUG
        delete[] vv;
#else
        xmem.delptr (vv);
        vv = NULL;
#endif
        return (1);
      }
      *(vv+i) = 1.0/ *(vv+i);
    }
    for (j=0; j<n; j++)
    { for (i=0; i<j; i++)
      { sum = *(lu+n*j+i);
        for (k=0; k<i; k++)
          sum -= (*(lu+n*k+i) * *(lu+n*j+k));
        *(lu+n*j+i) = sum;
      }
      aamax = 0.0;
      for (imax=i=j; i<n; i++)
      { sum = *(lu+n*j+i);
        for (k=0; k<j; k++)
          sum -= (*(lu+n*k+i) * *(lu+n*j+k));
        *(lu+n*j+i) = sum;
        dum = *(vv+i) * ::fabs(sum);
        if (dum>=aamax)
        { imax  = i;
          aamax = dum;
        }
      }
      if (j!=imax)
      { for (k=0; k<n; k++)
        { dum            = *(lu+n*k+imax);
          *(lu+n*k+imax) = *(lu+n*k+j);
          *(lu+n*k+j)    = dum;
        }
        *d = -*d;
        *(vv+imax) = *(vv+j);
      }
      *(indx+j) = imax;
      if (isZero(*(lu+n*j+j)))
      { 
#ifndef MEMDEBUG
        delete[] vv;
#else
        xmem.delptr (vv);
        vv = NULL;
#endif
        return(1);
      }
      if (j!=(n-1))
      { dum = 1.0/ *(lu+n*j+j);
        for (i=j+1; i<n; i++)
          *(lu+n*j+i) *= dum;
      }
    }
#ifndef MEMDEBUG
    delete[] vv;
#else
    xmem.delptr (vv);
    vv = NULL;
#endif
    return (0);
  }

  void lubksb (double *lu, long n, long *indx, double *x)
  { long i, ii, ll, j;
    double sum;
   
    ii = -1;
    for (i=0; i<n; i++)
    { ll  = *(indx+i);
      sum = *(x+ll);
      *(x+ll) = *(x+i);
      if (ii!=-1)
      { for (j=ii; j<i; j++)
          sum -= (*(lu+n*j+i) * *(x+j));
      }
      else if (sum!=0.0)
        ii = i;
      *(x+i) = sum;
    }
    for (i=n-1; i>-1; i--)
    { sum = *(x+i);
      if (i<n-1)
      { for (j=i+1; j<n; j++)
          sum -= (*(lu+n*j+i) * *(x+j));
      }
      *(x+i) = sum/ *(lu+n*i+i);
    }
  }

  void matinv (double *tptr, long n, long p)
  { long i, j;
    double *lumat, *col, sum;
    long *indx;
#ifndef MEMDEBUG
    lumat = new double[n*p];
    indx  = new long[n];
#else
    lumat = (double *) xmem.newptr (sizeof(double), n*p, "XplReal", "matinv", "lumat");
    indx  = (long *) xmem.newptr (sizeof(long), n, "XplReal", "matinv", "indx");
#endif  

    memcpy (lumat, tptr, n*p*sizeof(double));
    for (i=0; i<n*p; i++)
      SetNaN (tptr+i);
    if (n!=p) 
    { //SK 02.09.97
#ifndef MEMDEBUG
      delete[] lumat;
      delete[] indx;
#else
      xmem.delptr (lumat);
      xmem.delptr (indx);
      indx = NULL;
      lumat = NULL;
#endif
      return;
    }
    if (ludcmp (lumat, n, indx, &sum))
    { //SK 02.09.97
#ifndef MEMDEBUG
      delete[] lumat;
      delete[] indx;
#else
      xmem.delptr (lumat);
      xmem.delptr (indx);
      indx = NULL;
      lumat = NULL;
#endif
      return;
    }
    for (i=0; i<n; i++)
    { // printf ("indx[%i]=%li\n", i, indx[i]);
      col = tptr+i*n;
      for (j=0; j<n; j++)
        *(col+j) = 0.0;
      *(col+i) = 1.0;
      lubksb (lumat, n, indx, col);
    }
#ifndef MEMDEBUG
    delete[] lumat;
    delete[] indx;
#else
    xmem.delptr (lumat);
    xmem.delptr (indx);
    indx = NULL;
    lumat = NULL;
#endif
  }

  void matlu (double *tptr, long n, long p, double *tptrl, double *tptru, long *indx)
  { // return:
    //  0     o.k.
    //  1     matrix singular
    long i, j, nind;
    double *lumat, *col, sum;
    long *vind;
#ifndef MEMDEBUG 
    lumat = new double[n*p];
    vind  = new long[n];
#else
    lumat = (double *) xmem.newptr (sizeof(double), n*p, "XplReal", "matlu", "lumat");
    vind  = (long *) xmem.newptr (sizeof(long), n, "XplReal", "matlu", "vind");
#endif
    memcpy (lumat, tptr, n*p*sizeof(double));
    for (i=0; i<n*p; i++)
    { SetNaN(tptrl+i);
      SetNaN(tptru+i);
      // printf ("%i: %f %f\n", i, *(lumat+i), *(tptr+i));
    }
    if (n!=p) 
    { //SK 02.09.97
#ifndef MEMDEBUG
      delete[] vind;
      delete[] lumat;
#else
      xmem.delptr (vind);
      xmem.delptr (lumat);
      vind = NULL;
      lumat = NULL;
#endif
      return;
    }
    if (ludcmp (lumat, n, indx, &sum)) 
    { //SK 02.09.97
#ifndef MEMDEBUG
      delete[] vind;
      delete[] lumat;
#else
      xmem.delptr (vind);
      xmem.delptr (lumat);
      vind = NULL;
      lumat = NULL;
#endif
      return;
    }
    for (i=0; i<n; ++i)
	{
	for (j=0; j<n; ++j)
		{
		if (i>j) 
			{
			tptru[i*n+j]=lumat[i*n+j];
			tptrl[i*n+j]=0;
			}
		if (i==j) 
			{
			tptru[i*n+j]=lumat[i*n+j];
			tptrl[i*n+j]=1.0;
			}
		if (i<j) 
			{
			tptrl[i*n+j]=lumat[i*n+j];
			tptru[i*n+j]=0;
			}
		}
	vind[i]=i+1;
	}
    for (i=n-1; i>=0; --i)
	{ 
	nind=vind[indx[i]];
	vind[indx[i]]=vind[i];
	vind[i]=nind;
	}
//    for (i=0; i<n; ++i)	indx[vind[i]-1]=i+1;
    for (i=0; i<n; ++i)	indx[i]=vind[i];

#ifndef MEMDEBUG
    delete[] vind;
    delete[] lumat;
#else
    xmem.delptr (vind);
    xmem.delptr (lumat);
    vind = NULL;
    lumat = NULL;
#endif
//    delete[] indx;
  } 

  void matdet (double *tptr, long n, long p, double *res)
  { // return:
    //  0     o.k.
    //  1     matrix singular
    long i, j;
    double *lumat, *col, sum;
    long *indx;
#ifndef MEMDEBUG 
    lumat = new double[n*p];
    indx  = new long[n];
#else
    lumat = (double *) xmem.newptr (sizeof(double), n*p, "XplReal", "matlu", "lumat");
    indx  = (long *) xmem.newptr (sizeof(long), n, "XplReal", "matlu", "indx");
#endif
    memcpy (lumat, tptr, n*p*sizeof(double));
    SetNaN (res);
    for (i=0; i<n*p; i++)
      SetNaN(tptr+i);
    if (n!=p) 
    { //SK 02.09.97
#ifndef MEMDEBUG
      delete[] lumat;
      delete[] indx;
#else
      xmem.delptr (lumat);
      xmem.delptr (indx);
      lumat = NULL;
      indx  = NULL;
#endif
      return;
    }
    ludcmp (lumat, n, indx, &sum);
    *res=sum;
    for (i=0; i<n; i++)
     *res *= lumat[i*n+i];
#ifndef MEMDEBUG
    delete[] lumat;
    delete[] indx;
#else
    xmem.delptr (lumat);
    xmem.delptr (indx);
    lumat = NULL;
    indx  = NULL;
#endif
  } 

  void initrandom(double start)
  { randomval = ::floor (fmod (::fabs(start), 62748517.0));
    nset   = 0;
  }

  double  uniformrandom()
  { randomval = fmod (27132.0 * randomval + 7.0, 62748517.0);
    return (randomval / 62748517.0);
  }

  double normalrandom()
  { double x, y, r;
    if (nset) { nset = 0; return(nrand); }
    do
    { x = 2.0*uniformrandom()-1.0;
      y = 2.0*uniformrandom()-1.0;
      r = x*x+y*y; 
    }
    while (r>=1.0);
    r = ::sqrt(-2.0*::x4log(r)/r);
    nrand = x*r;
    nset  = 1;
    return(y*r);
  }

  double ran0 (int *idum)
  {	double y,maxran,v[98];
	double dum;
	int iff=0;
	int j;
	unsigned i,k;

	if (*idum < 0 || iff == 0) {
		iff=1;
		i=2;
		do {
			k=i;
			i<<=1;
		} while (i);
		maxran=k;
		srand(*idum);
		*idum=1;
		for (j=1;j<=97;j++) dum=rand();
		for (j=1;j<=97;j++) v[j]=rand();
		y=rand();
	}
	j= (int) (1+97.0*y/maxran);
//	if (j > 97 || j < 1) XAerrort(NOCONV, "jacobi"));
	y=v[j];
	v[j]=rand();
	return y/maxran;
  }

  double ran0 (void)
  { int i=0;
    return ran0 (&i);
  }

  double ran1(int *idum)
  {	long ix1,ix2,ix3;
	double r[98];
	double temp;
	int iff=0;
	int j;

	if (*idum < 0 || iff == 0) {
		iff=1;
		ix1=(R1_IC1-(*idum)) % R1_M1;
		ix1=(R1_IA1*ix1+R1_IC1) % R1_M1;
		ix2=ix1 % R1_M2;
		ix1=(R1_IA1*ix1+R1_IC1) % R1_M1;
		ix3=ix1 % R1_M3;
		for (j=1;j<=97;j++) {
			ix1=(R1_IA1*ix1+R1_IC1) % R1_M1;
			ix2=(R1_IA2*ix2+R1_IC2) % R1_M2;
			r[j]=(ix1+ix2*R1_RM2)*R1_RM1;
		}
		*idum=1;
	}
	ix1=(R1_IA1*ix1+R1_IC1) % R1_M1;
	ix2=(R1_IA2*ix2+R1_IC2) % R1_M2;
	ix3=(R1_IA3*ix3+R1_IC3) % R1_M3;
	j=1 + ((97*ix3)/R1_M3);
//	if (j > 97 || j < 1) xplerror (0);
	temp=r[j];
	r[j]=(ix1+ix2*R1_RM2)*R1_RM1;
	return temp;
  }

  double ran1 (void)
  { int i=0;
    return ran1 (&i);
  }

  double ran2(long *idum)
  {	long iy,ir[98];
	int iff=0;
	int j;

	if (*idum < 0 || iff == 0) {
		iff=1;
		if ((*idum=(R2_IC-(*idum)) % R2_M) < 0) *idum = -(*idum);
		for (j=1;j<=97;j++) {
			*idum=(R2_IA*(*idum)+R2_IC) % R2_M;
			ir[j]=(*idum);
		}
		*idum=(R2_IA*(*idum)+R2_IC) % R2_M;
		iy=(*idum);
	}
	j= (int) (1 + 97.0*iy/R2_M);
//	if (j > 97 || j < 1) xplerror(0);
	iy=ir[j];
	*idum=(R2_IA*(*idum)+R2_IC) % R2_M;
	ir[j]=(*idum);
	return (double) iy/R2_M;
  }

  double ran2 (void)
  { long i=0;
    return ran2 (&i);
  }

  double ran3(int *idum)
  {	int inext,inextp;
	long ma[56];
	int iff=0;
	long mj,mk;
	int i,ii,k;

	if (*idum < 0 || iff == 0) {
		iff=1;
		mj=R3_MSEED-(*idum < 0 ? -*idum : *idum);
		mj %= R3_MBIG;
		ma[55]=mj;
		mk=1;
		for (i=1;i<=54;i++) {
			ii=(21*i) % 55;
			ma[ii]=mk;
			mk=mj-mk;
			if (mk < R3_MZ) mk += R3_MBIG;
			mj=ma[ii];
		}
		for (k=1;k<=4;k++)
			for (i=1;i<=55;i++) {
				ma[i] -= ma[1+(i+30) % 55];
				if (ma[i] < R3_MZ) ma[i] += R3_MBIG;
			}
		inext=0;
		inextp=31;
		*idum=1;
	}
	if (++inext == 56) inext=1;
	if (++inextp == 56) inextp=1;
	mj=ma[inext]-ma[inextp];
	if (mj < R3_MZ) mj += R3_MBIG;
	ma[inext]=mj;
	return mj*R3_FAC;
  }

  double ran3 (void)
  { int i=0;
    return ran3 (&i);
  }

  double expdev_base(double (*func) (void))
  {	double dum;
	do
		dum=func();
	while (isZero(dum));
	return -::x4log(dum);
  }

  double expdev(void)
  { return expdev_base (uniformrandom); }
  
  double expdev0(void)
  { return expdev_base (ran0); }

  double expdev1(void)
  { return expdev_base (ran1); }

  double expdev2(void)
  { return expdev_base (ran2); }

  double expdev3(void)
  { return expdev_base (ran3); }

  double gasdev_base(double (*func) (void))
  {	int iset=0;
	double gset;
	double fac,r,v1,v2;
	double ran1();

	if  (iset == 0) {
		do {
			v1=2.0*func()-1.0;
			v2=2.0*func()-1.0;
			r=v1*v1+v2*v2;
		} while (r >= 1.0 || isZero(r));
		fac=::sqrt(-2.0*::x4log(r)/r);
		gset=v1*fac;
		iset=1;
		return v2*fac;
	} else {
		iset=0;
		return gset;
	}
  }

  double gasdev(void)
  { return gasdev_base (uniformrandom); }

  double gasdev0(void)
  { return gasdev_base (ran0); }

  double gasdev1(void)
  { return gasdev_base (ran1); }
  
  double gasdev2(void)
  { return gasdev_base (ran2); }
  
  double gasdev3(void)
  { return gasdev_base (ran3); }

//SK 02.09.97
/*  void matdist (double *in, long n, long p, double *out, char *distm)
  { long i, j, k;
    double tmp, mtmp;
    if (strcasecmp (distm, "L1")==0)
    { for (i=0; i<n; i++)
      { *(out+n*i+i) = 0.0;
        for (j=0; j<i; j++)
        { tmp = 0.0;
          for (k=0; k<p; k++) 
            tmp += ::fabs((*(in+k*n+i) - *(in+k*n+j)));
          *(out+i*n+j) = *(out+j*n+i) = tmp;
        }
      }
    }
    else if (strcasecmp (distm, "L2")==0)
    { for (i=0; i<n; i++)
      { *(out+n*i+i) = 0.0;
        for (j=0; j<i; j++)
        { tmp = 0.0;
          for (k=0; k<p; k++) 
            tmp += (*(in+k*n+i) - *(in+k*n+j)) * (*(in+k*n+i) - *(in+k*n+j));
          *(out+i*n+j) = *(out+j*n+i) = ::sqrt(tmp);
        }
      }
    }
    else if (strcasecmp (distm, "MAXIMUM")==0)
    { for (i=0; i<n; i++)
      { *(out+n*i+i) = 0.0;
        for (j=0; j<i; j++)
        { mtmp = 0.0;
          for (k=0; k<p; k++) 
          { tmp = ::fabs((*(in+k*n+i) - *(in+k*n+j)));
            if (mtmp<tmp) mtmp = tmp;
          }
          *(out+i*n+j) = *(out+j*n+i) = mtmp;
        }
      }
    }
    else
    { for (i=0; i<n; i++)
      { SetNaN (out+n*i+i);
        for (j=0; j<i; j++)
        { SetNaN (out+i*n+j);
          SetNaN (out+j*n+i);
        }
      }
      XAerrort (PARAMFALSE, "L1, L2 or MAXIMUM expected");
    }
  }
*/

  double gammaln (double xx)
  { double cof[6] = {76.18009173,    -86.50532033,  24.01409822,
               -1.234739516, 0.120858003e-2, -0.536382e-5},
        fact = 2.50662827465, x, result, temp, ser;
    int i;

    if (xx > 1.0)
    { x = xx - 1.0;
       result = 0.0;
    }                         /* for x>1 we have full accuracy  */
    else
    { x = xx;
     result = -::x4log(xx);
    }                         /* x<1 use gamma(x+1)=x*gamma(x)  */
    temp = x + 5.5;
    temp = (x + 0.5) * ::x4log(temp) - temp;
    ser = 1.0;
    for (i = 0; i < 6; i++)
    { x = x + 1.0;
       ser = ser + cof[i] / x;
    }
    result = result + temp + ::x4log(fact * ser);
    return(result);
  }

  double spdfb(double x, double a, double b)
  { double bt;

    if( (x<=0.0) || (x>=1.0) || (a<=0.0) || (b<=0) )
       return( 0.0 );
    bt=::x4exp(gammaln(a+b)-gammaln(a)-gammaln(b)+(a-1.0)*::x4log(x)+(b-1.0)*::x4log(1.0-x));
    return( bt );
  }

  void matpdfb (double *resptr, long nr, long pr, 
                double *aptr, long naa, long paa, 
                double *bptr, long nbb, long pbb)
  { long i, s, as, bs;
    s = nr*pr;
    as = naa*paa;
    bs = nbb*pbb;

    for (i=0; i<s; i++)
      resptr[i] = spdfb (resptr[i], aptr[i%as], bptr[i%bs]); 
  }

  double spdfc(double x, double n)
  {  double h;
     if (x<=0.0)
       return( 0.0 );
     h=(n/2.0-1.0)*::x4log(x)-x/2.0-gammaln(n/2.0)-(n/2.0)*M_LN2;
     return( ::x4exp(h) );
  }

  void matpdfc (double *resptr, long nr, long pr, 
                double *aptr, long naa, long paa, 
                double *bptr, long nbb, long pbb)
  { long i, s, as, bs;
    s = nr*pr;
    as = naa*paa;

    for (i=0; i<s; i++)
      resptr[i] = spdfc (resptr[i], aptr[i%as]); 
  }

  double spdff(double x, double n1, double n2)
  {  double h;
     if( x<=0.0 )
       return( 0.0 );
     h=gammaln((n1+n2)/2.0)-gammaln(n1/2.0)-gammaln(n2/2.0)+n1/2.0*::x4log(n1/n2);
     h=h+(n1/2.0-1.0)*::x4log(x)-(n1+n2)/2.0*::x4log(1.0+n1*x/n2);
     return( ::x4exp(h) );
  }

  void matpdff (double *resptr, long nr, long pr, 
                double *aptr, long naa, long paa, 
                double *bptr, long nbb, long pbb)
  { long i, s, as, bs;
    s = nr*pr;
    as = naa*paa;
    bs = nbb*pbb;

    for (i=0; i<s; i++)
      resptr[i] = spdff (resptr[i], aptr[i%as], bptr[i%bs]); 
  }

  double spdfn (double x)
  { return (::x4exp(-x*x/2)*M_1_SQRTPI*M_SQRT_2); }

  void matpdfn (double *resptr, long nr, long pr, 
                double *aptr, long naa, long paa, 
                double *bptr, long nbb, long pbb)
  { long i, s, as, bs;
    s = nr*pr;
    as = naa*paa;
    bs = nbb*pbb;

    for (i=0; i<s; i++)
      resptr[i] = spdfn (resptr[i]);
  }

  double spdft(double x, double n)
  {  double h;
     h=gammaln((n+1.0)/2.0)-gammaln(n/2)-(n+1)/2.0*::x4log(1.0+x*x/n);
     h=::x4exp(h)/::sqrt(n*M_PI);
     return(h);
  }

  void matpdft (double *resptr, long nr, long pr, 
                double *aptr, long naa, long paa, 
                double *bptr, long nbb, long pbb)
  { long i, s, as, bs;
    s = nr*pr;
    as = naa*paa;

    for (i=0; i<s; i++)
      resptr[i] = spdft (resptr[i], aptr[i%as]); 
  }

  double gammaser (double x, double a, double *gln)
  { int itmax = 100, n;
    double eps = 3.e-10, ap, sum, del;

    *gln = gammaln(a);
    if (isZero(x))
       return(0);
    if (x < 0)
    { SetNaN (&del);
      return(del);
    }
    ap = a;
    del = sum = 1.0 / a;
    for (n = 0; n < itmax; n++)
    { ap++;
       del *= x / ap;
       sum += del;
       if (fabs(del) < (fabs(sum) * eps))
         break;
    }
    if (n == itmax)
    { SetNaN (&del);
      return(del);
    }
    return(sum * ::x4exp(-x + a * ::x4log(x) - *gln));
  }

  double gammacf (double x, double a, double *gln)
  { int itmax = 101, n;
    double eps = 3.e-7, gold, g, a0, a1, b0, b1, fac, an, ana, anf;

    *gln = gammaln(a);
    gold = 0.0;
    a0 = 1.0;
    a1 = x;
    b0 = 0.0;
    b1 = 1.0;
    fac = 1.0;
    for (n = 1; n < itmax; n++)
    { an = n;
      ana = an - a;
      a0 = (a1 + a0 * ana) * fac;
      b0 = (b1 + b0 * ana) * fac;
      anf = an * fac;
      a1 = x * a0 + anf * a1;
      b1 = x * b0 + anf * b1;
      if (a1 != 0.0)
      { fac = 1.0 / a1;
        g = b1 * fac;
        if (fabs((g - gold) / g) < eps )
          return(::x4exp(-x + a * ::x4log(x) - *gln) * g);
        gold = g;
      }
    }
    SetNaN (&ana);
    return(ana);
  }

  double stgamma(double x, double a)
  { double gln;
    if( (x<0.0) || (a<=0.0) )
    { SetNaN (&gln);
      return(gln);
    }
    if( x< a+1.0 )
      return( gammaser(x,a,&gln) );
    else
      return( 1.0-gammacf(x,a,&gln) );
  }

  double scdfn (double x)
  {
  // PC 20020711 - fixed error for x > 1e155
    if ( XisNegInf(x) || ((x < 0) && (XisPosInf(x*x))) ) 
      return 0.0;
    else if ( XisPosInf(x) || ((x > 0) && (XisPosInf(x*x))) )
      return 1.0;
    else if( x>=0.0)
      return( (1.0+stgamma(x*x/2.0,0.5))/2.0 );
    else
      return( (1.0-stgamma(x*x/2.0,0.5))/2.0 );
  }

  void matcdfn (double *resptr, long nr, long pr, 
                double *aptr, long naa, long paa, 
                double *bptr, long nbb, long pbb)
  { long i, s, as, bs;
    s = nr*pr;
    as = naa*paa;
    bs = nbb*pbb;

    for (i=0; i<s; i++)
      resptr[i] = scdfn (resptr[i]);
  }

  double scdft (double x, double n)
  { if ( x<0.0 )
      return( scdfb(n/(n+x*x),n/2.0,0.5)/2.0 );
    else
      return( 1.0-scdfb(n/(n+x*x),n/2.0,0.5)/2.0 );
  }

  void matcdft (double *resptr, long nr, long pr, 
                double *aptr, long naa, long paa, 
                double *bptr, long nbb, long pbb)
  { long i, s, as, bs;
    s = nr*pr;
    as = naa*paa;

    for (i=0; i<s; i++)
      resptr[i] = scdft (resptr[i], aptr[i%as]); 
  }

  double scdfc(double x, double n)
  { if (x<=0.0) return(0.0);
    return( stgamma(x/2.0, n/2.0));
  }

  void matcdfc (double *resptr, long nr, long pr, 
                double *aptr, long naa, long paa, 
                double *bptr, long nbb, long pbb)
  { long i, s, as, bs;
    s = nr*pr;
    as = naa*paa;

    for (i=0; i<s; i++)
      resptr[i] = scdfc (resptr[i], aptr[i%as]); 
  }

  double betacf (double x, double a, double b)
  { double eps = 3.e-7, am, bm, az, bz, ap, bp, app, bpp, aold,
      d, em, qab, qap, qam, tem;
    int itmax = 101, m;

    qab = a + b;
    qap = a + 1.0;
    qam = a - 1.0;
    am = 1.0;
    bm = 1.0;
    az = 1.0;
    bz = 1.0 - qab * x / qap;
    for (m = 1; m < itmax; m++)
    { em = m;
      tem = em + em;
      d = em * (b - em) * x / ((qam + tem) * (a + tem));
      ap = az + d * am;
      bp = bz + d * bm;
      d = -(a + em) * (qab + em) * x / ((a + tem) * (qap + tem));
      app = ap + d * az;
      bpp = bp + d * bz;
      aold = az;
      am = ap / bpp;
      bm = bp / bpp;
      az = app / bpp;
      bz = 1.0;
      if (fabs(az - aold) < eps * fabs(az))
        break;
    }
    if (m == itmax) ::SetNaN (&az);
    return(az);
  }

  double scdfb(double x, double a, double b)
  { double bt;

    if( (x<0.0) || (x>1.0) || (a<=0.0) || (b<=0) )
    { ::SetNaN (&bt);
      return (bt);
    }
    if( isZero(x) || isZero(x-1.0))
      bt=0.0;
    else
      bt=::x4exp(gammaln(a+b)-gammaln(a)-gammaln(b)+a*::x4log(x)+b*::x4log(1.0-x));
    if( x < (a+1.0)/(a+b+2.0) )
      return( bt*betacf(x,a,b)/a );
    else
      return( 1.0-bt*betacf(1.0-x,b,a)/b );
  }

  void matcdfb (double *resptr, long nr, long pr, 
                double *aptr, long naa, long paa, 
                double *bptr, long nbb, long pbb)
  { long i, s, as, bs;
    s = nr*pr;
    as = naa*paa;
    bs = nbb*pbb;

    for (i=0; i<s; i++)
      resptr[i] = scdfb (resptr[i], aptr[i%as], bptr[i%bs]); 
  }

  double scdff(double x, double n1, double n2)
  { if( x<=0.0 ) return(0.0);
    return( 1.0-scdfb(n2/(n2+n1*x),n2/2.0,n1/2.0) );
  }

  void matcdff (double *resptr, long nr, long pr, 
                double *aptr, long naa, long paa, 
                double *bptr, long nbb, long pbb)
  { long i, s, as, bs;
    s = nr*pr;
    as = naa*paa;
    bs = nbb*pbb;

    for (i=0; i<s; i++)
      resptr[i] = scdff (resptr[i], aptr[i%as], bptr[i%bs]); 
  }

  double sqfn (double k)
  { double lim, p0, p1, p2, p3, p4, p, y, q0, q1, q2, q3, q4, xp;

    if ((k>=0) && (k<EPSILON))
    { xp = -dinf;
      return (xp);
    }
    if ((1-k>=0) && (1-k<EPSILON))
    { xp = dinf;
      return (xp);
    }
    p   = k;
    lim = 10e-15;
    p0  = -0.322232431088;
    p1  = -1.0;
    p2  = -0.342242088547;
    p3  = -0.0204231210245;
    p4  = -0.453642210148e-4;
    q0  = 0.0993484626060;
    q1  = 0.588581570495;
    q2  = 0.531103462366;
    q3  = 0.103537752850;
    q4  = 0.38560700634e-2;
    if (p > 0.5)
      p = 1.0 - p;
    if (p < lim || k > (1.0 - lim))
    { SetNaN (&lim);
      return(lim);
    }
    if (isZero(p-0.5))
      return(0.0);
    y = ::sqrt(::x4log(1.0 / (p * p)));
    xp = y + ((((y * p4 + p3) * y + p2) * y + p1) * y + p0)
          / ((((y * q4 + q3) * y + q2) * y + q1) * y + q0);
    if (k < 0.5)
      xp = - xp;
    return(xp);
  }



  double sqft(double x, double df1)
  { double q, c, t, res,qcur, dx;
    int df = (int) df1,i;



    if( x <0.0 || x > 1.0)
    { SetNaN (&res);
      return (res);
    }
    if (df1<1.0)
    { SetNaN (&res);
      return (res);
    }
    if ((x<EPSILON)&&(x>=0))
    { res = -dinf;
      return (res);
    }
    if ((1-x>=0) && (1-x<EPSILON))
    { res = +dinf;
      return (res);
    }

    if( df == 1)
    { return (tan(M_PI*(x-0.5)));}
    if( df == 2)
    { t = 2.0*x -1.0;
      return (M_SQRT2*t /::sqrt(1.0-t*t));

    }
    q = sqfn(x);
    c = df - 2.0/3.0 + 1.0/(10.0*df);
    c = (df - 5.0/6.0) / (c*c);
    res = ::sqrt( df*::x4exp(c*q*q) - df );
    if ( x <= 0.5) res = -res;

    //SB improvement inspired by dll function   28072004
    // iterate
    #ifndef  NEWTON_EPS
    	#define NEWTON_EPS      1e-12
    #endif
    #ifndef  NEWTON_ITMAX
      #define NEWTON_ITMAX    100
    #endif


     qcur = res;
     for(i = 1; i <= NEWTON_ITMAX; i++)
     {
          dx = (scdft(qcur, df) - x) / spdft(qcur, df);
          qcur -= dx;
          if (abs(dx / qcur) < NEWTON_EPS)
               return qcur;
     }

     // iterations not successful
    return res;
  }




/*  double sqfc(double x, double df1)
  {  int c=0;
     double chi, h60, q, z;
      double h[15] = { -0.0118, -0.0067, -0.0033, -0.0010,  0.0001,
                          0.0006,  0.0006,  0.0002, -0.0003, -0.0006,
                         -0.0005,  0.0002,  0.0017,  0.0043,  0.0082};

     if( x < 0.0 || x >= 1.0)
     { SetNaN(&h60);
       return (h60); 
     }
     if (isZero(x)) return (x);
     q = sqfn(x);
     if ( (q < -3.5) || (q > 3.5) )
       h60 = 0.0;
     else
     { z = -3.5;
       while (z < q)
       { c++;
         z += 0.5;
       }
       h60 = h[c-1] + 2.0*(h[c]-h[c-1])*(q-z+0.5);
     }
     chi  = 1.0 - 2.0/(9.0*df1) + (q - 60.0/df1*h60)*::sqrt(2.0/(9.0*df1));
     chi *= chi*chi*df1;
     return (chi);
  }
  */
  double sqff(double x, double df1, double df2)
  { int f1 = (int) df1, f2 = (int) df2;
    double f, lam, h, y, w;

    if( x < 0.0 || x > 1.0)
    { SetNaN (&lam);
      return (lam); 
    }
    if ((1-x>=0) && (1-x<EPSILON))
    { lam = dinf;
      return (lam);
    }
    if((df1 < 1.0) || (df2<1.0))
    { SetNaN (&lam);
      return (lam); 
    }
/*
    if( isZero(x) ) return (x);
    if( f1 == 1)
    { f = sqft(0.5*(1.0+x),df2); 
      return (f*f);
    }
    if( f2 == 1) return (1.0/sqff(1.0-x,1,df1));
    y   = sqfn(x);
    lam = (y*y-3)/6;
    h   = 2.0/(1.0/(f1-1)+1.0/(f2-1));
    w   = y*::sqrt(h+lam)/h;
    w  -= (1.0/(f1-1)-1.0/(f2-1))*(lam+5.0/6.0-2.0/(3.0*h));
    return (::x4exp(2*w));
*/
    return (finv_(f1, f2, x));
  }

  void matqfn (double *resptr, long nr, long pr,
                double *aptr, long naa, long paa, 
                double *bptr, long nbb, long pbb)
  { long i, s, as, bs;
    s = nr*pr;
    as = naa*paa;
    bs = nbb*pbb;

    for (i=0; i<s; i++)
      resptr[i] = sqfn (resptr[i]);
  }

  void matqft (double *resptr, long nr, long pr, 
                double *aptr, long naa, long paa, 
                double *bptr, long nbb, long pbb)
  { long i, s, as, bs;
    s = nr*pr;
    as = naa*paa;

    for (i=0; i<s; i++)
      resptr[i] = sqft (resptr[i], aptr[i%as]); 
  }

  void matqfc (double *resptr, long nr, long pr, 
                double *aptr, long naa, long paa, 
                double *bptr, long nbb, long pbb)
  { long i, s, as, bs;
    s = nr*pr;
    as = naa*paa;

    for (i=0; i<s; i++)
      resptr[i] = ::sqfc (aptr[i%as], resptr[i]);
  }

  void matqff (double *resptr, long nr, long pr, 
                double *aptr, long naa, long paa, 
                double *bptr, long nbb, long pbb)
  { long i, s, as, bs;
    s = nr*pr;
    as = naa*paa;
    bs = nbb*pbb;

    for (i=0; i<s; i++)
      resptr[i] = sqff (resptr[i], aptr[i%as], bptr[i%bs]); 
  }

  void matatan2 (double *resptr, long nr, long pr, 
                        double *aptr, long naa, long paa, 
                        double *bptr, long nbb, long pbb)
  { long i, s, as, bs;
    s = nr*pr;
    as = naa*paa;

    for (i=0; i<s; i++)
      resptr[i] = ::atan2(resptr[i], aptr[i%as]); 
  }

  long circle(long x, long period)
  { return ((x>= 0) ? (x%period) : (period-1+(x+1)%period)); }

  void fwt_int (double *data, long length, long level, double *h, int hrange)
  { long i, k;
    long local=length;
    double    *temp, *temp_1;
    double    mult;
    long       i2,k2,time;
    long       center = length/2;

#ifndef MEMDEBUG
    temp   = new double[length];
    temp_1 = new double[length];
#else
    temp = (double *) xmem.newptr (sizeof(double), length, "XplReal", "ftw_int", "temp");
    temp_1 = (double *) xmem.newptr (sizeof(double), length, "XplReal", "ftw_int", "temp_1");
#endif  

    for(i=1; local > level;i++)
    { center = local/2;
      for (i2=0;i2< center;i2++)
      { time = 2*i2;
        temp[i2]   = 0.0;
        temp_1[i2] = 0.0;
        mult = 1.0;
        for(k2=0;k2<hrange;k2++)
        { mult =-mult;
          temp[i2]   +=  h[k2]*data[circle(time+k2,local)];
          temp_1[i2] +=  mult*h[k2]*data[circle(time+1-k2,local)];
        }
      }
      for(i2=0;i2<center; i2++)
      { data[i2]=temp[i2];
        data[i2+center] =temp_1[i2];
      }
      local = local/2;
    }
#ifndef MEMDEBUG
    delete[] temp;
    delete[] temp_1;
#else
    xmem.delptr (temp);
    xmem.delptr (temp_1);
    temp = NULL;
    temp_1 = NULL;
#endif
  }

  void filter_mother_father(double *x, long length, double* h, long hrange)
  { double *temp;
    double mult = -1.0;
    long i,k;

#ifndef MEMDEBUG
    temp   = new double[length];
#else
    temp = (double *) xmem.newptr (sizeof(double), length, "XplReal", "ftw_int", "temp");
#endif  

    for(i=0;i<=length/2-1; i++)
    { temp[2*i]=x[i];
      temp[2*i+1] =0.0;
    }
    for(i= 0;i< length; i++)
    { mult = -mult;
      x[i]=0.0;
      for(k=0;k<hrange;k++)
        x[i] +=   mult*h[k]*temp[circle(i-1+k,length)];
    }
#ifndef MEMDEBUG
    delete[] temp;
#else
    xmem.delptr (temp);
    temp = NULL;
#endif
  }

  void filter_father_father(double *x, long length, double* h, long hrange)
  { double *temp;
    long i,k;
#ifndef MEMDEBUG
    temp   = new double[length];
#else
    temp = (double *) xmem.newptr (sizeof(double), length, "XplReal", "ftw_int", "temp");
#endif  

    for(i=0;i<=length/2-1; i++){
      temp[2*i]=x[i];
      temp[2*i+1]=0.0;
    }
    for(i=0;i< length; i++){
      x[i] =  0.0;
      for(k=0;k<hrange;k++)
	x[i] += h[k]*temp[circle(i-k,length)];
    }
#ifndef MEMDEBUG
    delete[] temp;
#else
    xmem.delptr (temp);
    temp = NULL;
#endif
  }

  void invert_fwt(double *data, int length, int level, double* h, long hrange)
  { double *temp, *temp_1, *sum;
    long   mult =level;
    long   i,k;

#ifndef MEMDEBUG
    temp   = new double[length];
    temp_1 = new double[length];
    sum    = new double[length];
#else
    temp = (double *) xmem.newptr (sizeof(double), length, "XplReal", "invert_int", "temp");
    temp_1 = (double *) xmem.newptr (sizeof(double), length, "XplReal", "invert_int", "temp_1");
    sum    = (double *) xmem.newptr (sizeof(double), length, "XplReal", "invert_int", "sum");
#endif  

    for (i=0;i<=length-1;i++){
      sum[i]   = 0.0;
      temp[i]  = 0.0;
      temp_1[i]= 0.0;
    }
    for(k=0;k< level; k++)
      temp_1[k] = data[k];
    for (i=0;mult <=length/2; i++ ){
      for(k=0;k<=mult-1; k++)
      temp[k]=data[mult+k];

      filter_mother_father(temp,2*mult, h, hrange);
      filter_father_father(temp_1,2*mult, h, hrange);
      for(k=0;k<=length-1;k++)
        sum[k]=sum[k]+temp[k];
      mult=2*mult;
      if(mult <=length/2)
      filter_father_father(sum,2*mult, h, hrange);
    }

    for(k=0;k<=length-1;k++)
      data[k] = sum[k]+temp_1[k];

#ifndef MEMDEBUG
    delete[] temp;
    delete[] temp_1;
    delete[] sum;
#else
    xmem.delptr (temp);
    xmem.delptr (temp_1);
    xmem.delptr (sum);
    temp = NULL;
    temp_1 = NULL;
    sum = NULL;
#endif
  }

  void sisoreg (double *resvec, long n)
  { long i,j,k,m,nbl,monotone;
    long *bls, *ble;

#ifndef MEMDEBUG
    bls = new long[n];
    ble = new long[n];
#else
    bls = (long *) xmem.newptr (sizeof(long), n, "XplReal", "sisoreg", "bls");
    ble = (long *) xmem.newptr (sizeof(long), n, "XplReal", "sisoreg", "ble");
#endif  

    for (i = 0;i < n;i++)
    { bls[i] = i;
      ble[i] = i;
    }
    nbl = n;
    do
    { monotone = 1;
      k = 0;
      while (k < nbl-1)
      { if (*(resvec+bls[k]) > *(resvec+bls[k+1]))
        { monotone = 0;     /*  pool adjacent violators */
          *(resvec+bls[k]) = (*(resvec+bls[k])* (ble[k]-bls[k]+1) +
                              *(resvec+bls[k+1])*((ble[k+1])-bls[k+1]+1))
                             / (ble[k+1] - bls[k] + 1);
          ble[k] = ble[k+1];
          for (m = k+1;m < nbl;m++) /* push blockindex down */
          { bls[m] = bls[m+1];
            ble[m] = ble[m+1];
          }
          nbl--; /* decrease number of blocks */
        }
        else k++;
      }
      for (i = 0;i < nbl;i++)
       for (j = bls[i];j <= ble[i];j++)
         *(resvec+j) = *(resvec+bls[i]);
    }
    while (!monotone);
#ifndef MEMDEBUG
    delete[] bls;
    delete[] ble;
#else
    xmem.delptr (bls);
    xmem.delptr (ble);
    bls = NULL;
    ble = NULL;
#endif

  }

  void s_knn(double *x, double *y, long n, long k, double *sk)
  { long   k2, j, i, iup , ilo;
    double     mx, my, cxy, cxx, alpha, beta, v;

    k2 = k / 2 ;
    k  = 2*k2+1;
/*  determine starting mean at left boundary */
    mx  = 0.0;
    my  = 0.0;
    cxx = 0.0;
    cxy = 0.0;
    for(i=0; i<k; i++)
    { mx  += x[i] ;
      my  += y[i] ;
      cxy += x[i]*y[i] ;
      cxx += x[i]*x[i] ;
    }
    v = cxx-mx*mx/k ;
    if (v > 0.0)
      beta = (cxy-mx*my/k)/v;
    else
      beta = 0.0 ;
    alpha = (my - beta*mx)/k ;
/* left boundary */
   for(j=0; j<k2+1; j++)
     sk[j] = alpha + beta*x[j];

/* interior points */

   iup = k;
   ilo = 0;
   for(j=k2+1; j<n-k2; j++, iup++, ilo++)
   { mx  += x[iup] - x[ilo] ;
     my  += y[iup] - y[ilo] ;
     cxy += x[iup]*y[iup] - x[ilo]*y[ilo] ;
     cxx += x[iup]*x[iup] - x[ilo]*x[ilo] ;
     v = cxx-mx*mx/k ;
     if (v > 0.0)
       beta = (cxy-mx*my/k)/v;
     else
       beta = 0.0 ;
     alpha = (my - beta*mx)/k ;
     sk[j] = alpha + beta*x[j] ;
   }

/* right boundary */
   for(j=n-k2; j<n; j++)
     sk[j] = alpha + beta*x[j] ;
  }

  void swap(long l, long r, double *window, long *outlist, long *nrlist)
  { long nl, nr;
    double swap;

    swap        = window[l];
    nl          = nrlist[l];
    nr          = nrlist[r];

    window[l]   = window[r];
    nrlist[l]   = nr;
    outlist[nl] = r;

    window[r]   = swap;
    nrlist[r]   = nl;
    outlist[nr] = l;
  }

  void siftup(long l, long r, double *window, long *outlist, long *nrlist)
  { long i, j, nrold;
    double x;

    i     = l;
    j     = 2*i;
    x     = window[i];
    nrold = nrlist[i];
    while (j<=r)
    { if (j<r)
        if (window[j] < window[j+1])
          j++;
      if (x >= window[j])
        break;
      window[i]          = window[j];
      outlist[nrlist[j]] = i;
      nrlist[i]          = nrlist[j];
      i = j;
      j = 2*i;
    }
    window[i]      = x;
    outlist[nrold] = i;
    nrlist[i]      = nrold;
  }

  void heapsort(long low, long up, double* window, long *outlist, long *nrlist)
  { long l, u;

    l = (up/2) + 1;
    u = up;
    while(l>low)
    { l--;
      siftup(l, u, window, outlist, nrlist);
    }
    while(u>low)
    { swap(l, u, window, outlist, nrlist);
      u--;
      siftup(l, u, window, outlist, nrlist);
    }
  }


  void inittree(long n, long k, long k2, double *data, double *window, long *outlist, long *nrlist)
  { long i, k2p1;
    double big;

    for(i=1; i<=k; i++)
    { window[i] = data[i-1];
      nrlist[i] = outlist[i] = i;
    }

    heapsort(1, k, window, outlist, nrlist);

    big = fabs(window[k]);
    for(i=k+1; i<n; i++)
      if (big < fabs(data[i]))
        big=fabs(data[i]);

    for(i=k; i>0; i--)
    { window[i+k2] = window[i];
      nrlist[i+k2] = nrlist[i]-1;
    }

    for(i=0; i<k; i++)
      outlist[i]=outlist[i+1]+k2;

    k2p1 = k2+1;
    for(i=0; i<k2p1; i++)
    { window[i]        = -big;
      window[k+k2p1+i] = big;
    }
  }


  void toroot(long outvirt, long k, long nrnew, long outnext, double *data, double *window, long *outlist, long *nrlist)
  { long father;
    do
    { father                    = outvirt/2;
      window[outvirt+k]         = window[father+k];
      outlist[nrlist[father+k]] = outvirt+k;
      nrlist[outvirt+k]         = nrlist[father+k];
      outvirt                   = father;
    }
    while (father != 0);
    window[k]        = data[nrnew];
    outlist[outnext] = k;
    nrlist[k]        = outnext;
  }

  void downtoleave(long outvirt, long k, double *window, long *outlist, long *nrlist)
  { long childl, childr;

    for(;;)
    { childl = outvirt*2;
      childr = childl-1;
      if (window[childl+k] < window[childr+k])
        childl = childr;
      if (window[outvirt+k] >= window[childl+k])
        break;
      swap(outvirt+k, childl+k, window, outlist, nrlist);
      outvirt = childl;
    }
  }

  void uptoleave(long outvirt, long k,
                        double *window, long *outlist, long *nrlist)
  { long childl, childr;

    for(;;)
    { childl = outvirt*2;
      childr = childl+1;
      if (window[childl+k] > window[childr+k])
        childl = childr;
      if (window[outvirt+k] <= window[childl+k])
        break;
      swap(outvirt+k, childl+k, window, outlist, nrlist);
      outvirt = childl;
    }
  }
 
  void upperoutupperin(long outvirt, long k,
                              double *window, long *outlist, long *nrlist)
  { long father;

    uptoleave(outvirt, k, window, outlist, nrlist);
    father = outvirt/2;
    while (window[outvirt+k] < window[father+k])
    { swap(outvirt+k, father+k, window, outlist, nrlist);
      outvirt = father;
      father  = outvirt/2;
    }
  }

  void upperoutdownin(long outvirt, long k, long nrnew, long outnext,
                             double *data, double *window, long *outlist, long *nrlist)
  { toroot(outvirt, k, nrnew, outnext, data, window, outlist, nrlist);
    if(window[k] < window[k-1])
    { swap(k, k-1, window, outlist, nrlist);
      outvirt = -1;
      downtoleave(outvirt, k, window, outlist, nrlist);
    }
  }

  void downoutdownin(long outvirt, long k,
                            double *window, long *outlist, long *nrlist)
  { long father;

    downtoleave(outvirt, k, window, outlist, nrlist);
    father = outvirt/2;
    while (window[outvirt+k] > window[father+k])
    { swap(outvirt+k, father+k, window, outlist, nrlist);
      outvirt = father;
      father  = outvirt/2;
    }
  }

  void downoutupperin(long outvirt, long k, long nrnew, long outnext,
                             double *data, double *window, long *outlist, long *nrlist)
  { toroot(outvirt, k, nrnew, outnext, data, window, outlist, nrlist);
    if(window[k] > window[k+1])
    { swap(k, k+1, window, outlist, nrlist);
      outvirt = 1;
      uptoleave(outvirt, k, window, outlist, nrlist);
    }
  }

  void  wentoutone (long k, double *window, long *outlist, long *nrlist)
  { long outvirt;

    swap(k, k+1, window, outlist, nrlist);
    outvirt = 1;
    uptoleave(outvirt, k, window, outlist, nrlist);
  }

  void wentouttwo(long k, double *window, long *outlist, long *nrlist)
  { long outvirt;

    swap(k, k-1, window, outlist, nrlist);
    outvirt = -1;
    downtoleave(outvirt, k, window, outlist, nrlist);
  }

  void runmedint(long n, long k, long k2, double *data, double *median,
                        double *window, long *outlist, long *nrlist)
  { long i, nrnew, outnext, out, outvirt;

    for(i=0; i<=k2; median[i++]=window[k]);
    outnext = 0;
    for(i=k2+1; i<n-k2; i++)
    { out         = outlist[outnext];
      nrnew       = i+k2;
      window[out] = data[nrnew];
      outvirt     = out-k;
      if (out > k)
        if(data[nrnew] >= window[k])
          upperoutupperin(outvirt, k, window, outlist, nrlist);
        else
          upperoutdownin(outvirt, k, nrnew, outnext,
                         data, window, outlist, nrlist);
      else if(out < k)
        if(data[nrnew] < window[k])
          downoutdownin(outvirt, k, window, outlist, nrlist);
        else
          downoutupperin(outvirt, k, nrnew, outnext,
                         data, window, outlist, nrlist);
      else if(window[k] > window[k+1])
        wentoutone(k, window, outlist, nrlist);
      else if(window[k] < window[k-1])
        wentouttwo(k, window, outlist, nrlist);
      median[i] = window[k];
      outnext   = (outnext+1)%k;
    }
    for(i=n-k2; i<n; median[i++]=window[k]);
  }

  void runmed(long n, long k, double *data, double *median)
  { long k2, *outlist, *nrlist;
    double *window;

#ifndef MEMDEBUG
    outlist = new long[k+1];
    nrlist  = new long[2*k+1];
    window  = new double[2*k+1];
#else
    outlist = (long *) xmem.newptr (sizeof(long), k+1, "", "runmed", "outlist");
    nrlist = (long *) xmem.newptr (sizeof(long), 2*k+1, "", "runmed", "nrlist");
    window = (double *) xmem.newptr (sizeof(double), 2*k+1, "", "runmed", "window");
#endif
    k2 = k/2;
    k  = 2*k2+1;
    inittree(n, k, k2, data, window, outlist, nrlist);
    runmedint(n,k, k2, data, median, window, outlist, nrlist);
#ifndef MEMDEBUG
    delete[] outlist;
    delete[] nrlist;
    delete[] window;
#else
    xmem.delptr (outlist);
    xmem.delptr (nrlist);
    xmem.delptr (window);
    outlist = NULL;
    nrlist = NULL;
    window = NULL;
#endif
  }
/*
  void sl1line (long n, double *x, double *y, double *b)
  { long stop = 1,iter,ibas1,ibas2,flag,ready = 0,iout,iin,isave,more,i,j,rho,z;
    double a1,a2,a3,y1,y2,ahalf,aone,sum,d3,test,ratio,hold,a4,b3,b4,det,tot1,tot2,t;
    double *dev, *inext;
    a1 = x[0];
    y1 = y[0];
    for (i=1; i<n; i++)
    { // SK 24.04.99
      for (j=i+1; j<n; j++)
      { if (fabs(x[j]-x[i])<SMALL)
        { XAerrort (TIES, "l1line: ties found");
          b[0] = b[1] = ::sqrt(-1);
          return;
        }
      }
    }
#ifndef MEMDEBUG
    dev = new double[n];
    inext = new double[n];
#else
    dev = (double *) xmem.newptr (sizeof(double), n, "", "sl1line", "dev");
    inext = (double *) xmem.newptr (sizeof(double), n, "", "sl1line", "inext");
#endif
    iter = 0;
    ahalf = 0.5 + SMALL;
    aone = 2 * ahalf;
    a2 = x[i-1];
    y2 = y[i-1];
    ibas1 = 0;
    ibas2 = i - 1;
    det = 1 / (a2 - a1);

    a4 = (a2 * y1 - a1 * y2) * det;
    b4 = (y2 - y1) * det;
    for (i = 1;i < n;i++)
      dev[i]     = ( (y[i] >= (a4 + b4 * x[i])) ? 1.0 : -1.0);
    dev[ibas1] = 0.0;
    dev[ibas2] = -1.0;
    tot1 = 1.0; 
    tot2 = x[ibas2];
    for (i = 1;i < n;i++)
    { tot1 += dev[i];
      tot2 += dev[i] * x[i];
    }
    t = (a2 * tot1 - tot2) * det;
    flag = (fabs(t) > aone);
    while (!ready)
    { flag = !flag;
      if (flag)
      { t = (tot2 - a1 * tot1) * det;
        if (fabs(t) < aone) ready = 1;
        else
        { iout = ibas2;
          x[iout] = a1;
          a3 = a1; b3 = a2;
        }
      }
      else
      { if (!iter) det *= -1;
        else
        { t = (tot2 - a2*tot1) * det;
          if (fabs(t) < aone) ready = 1;
        }
        if (!ready)
        { a3 = a2; b3 = a1; iout = ibas1;
        }
      }
      if (!ready)
      { if (t > 0) rho = 1;
        else
        { rho = -1; t *= -1; det *= -1;
        }
        t *= 0.5;
        ratio = BIG; sum = ahalf;
        inext[ibas1] = ibas2;
     for ( i = 0;i < n;i++)
     { d3 = ( x[i] - a3) * det;
       if (d3 * dev[i] > SMALL)
       { test = (y[i] - (a4 + b4  * x[i])) / d3;
         if (test < ratio)
         { j = ibas1; sum += fabs(d3); more = 1; stop = 0;
           while (!stop)
           { isave = (int) fabs(inext[j]);
             if (test >= dev[isave])
             { more = 0; stop = 1;
              }
             else
             { if (sum < t) stop = 1;
               else
               { hold = sum - fabs(( x[isave] - a3) * det);
                 if (hold < t) stop = 1;
                 else
                 { dev[isave] = (inext[j] >= 0) ? 1 : -1;
                   sum = hold;
                   inext[j] = inext[isave];
                 }
               }
             }
           }
           if (more)
           { do
             { j = isave;
               isave = (int) fabs( inext[j]);
             }
             while(test < dev[isave]);
           }
           inext[i] = (int) inext[j];
           inext[j] = (dev[i] >= 0) ? i : -i;
           dev[i] = test;
           if ( sum >= t)
           { iin = (int) fabs( inext[ibas1]);
             ratio = dev[iin];
           }
         }
       }
     }
     iin = (int) fabs( inext[ibas1]);
     j = iin; stop = 0;
     while (!stop)
     { isave = (int) fabs( inext[j]);
       if (isave == ibas2) stop = 1;
       else
       { z = ( inext[j] >= 0) ? 1 : -1;
         dev[isave] = -z;
         z *= 2; j = isave;
         tot1 -= z; tot2 -= z * x[isave];
       }
     }

     z = ( (int) inext[ibas1] >= 0) ? 1 : -1;
     dev[iout] = -rho;
     tot1 -= (rho + z);
     tot2 -= (rho * b3 + z * x[iin]);
     iter++;
     if (flag)
     { x[ibas2] = a2;
       ibas2 = iin;
       a2 = x[iin];
       y2 = y[iin];
       det = 1.0 / (a1 - a2);
       dev[ibas2]= -1.0;
       a4 = (a1 * y2 - a2 * y1) * det;
       b4 = (y1 - y2) * det;
     }
     else
     { ibas1 = iin;
       a1 = x[iin];
       y1 = y[iin];
       det = 1.0 / (a2 - a1);
       dev[ibas1] = 0.0;
       a4 = (a2 * y1 - a1* y2) * det;
       b4 = (y2 - y1) * det;
     }
      }
    }
    for (i = 0;i < n;i++)
    { d3 = y[i] - (a4 + b4 * x[i]);
      dev[i] = d3;
    }
    b[0] = a4;
    b[1] = b4;
#ifndef MEMDEBUG
    delete[] inext;
    delete[] dev;
#else
    xmem.delptr (inext); inext = NULL;
    xmem.delptr (dev); dev = NULL;
#endif
  }
*/

/* definitions */
#define OK          0
#define NOMEMORY    -1

#define STOPRULE    0.01
#define EPSILON     1.0e-7

typedef unsigned long ULONG;
typedef double REAL;


/*
 * Auxiliary functions
 */

/* absolute value */
REAL rabs(REAL x)
{
     return( (x >= 0) ? (REAL) x : (REAL) -x );
}

/* signum */
REAL sign(REAL x)
{
     return( (x > 0.0) ? (REAL) 1.0 : (x < 0.0) ? (REAL) -1.0 : (REAL) 0.0 );
}

/*
 * Interface functions
 */

/* Median: computes median of the vector x (its length is n);
 *         vector x can be supposedly modified by the routine */
/*         should refer to Xplore routine, otherwise use qsort */
REAL median(REAL *x, ULONG n)
{  REAL *tmp = new REAL[n];
   memcpy (tmp, x, sizeof(REAL)*n);
   qsort (tmp, n, sizeof(REAL), less);
   if (n%2) return (tmp[n/2]);
   return (0.5*(tmp[n/2]+tmp[n/2-1]));
}

/* LSE: computes least squares estimate for linear model y = a + bx
 *      x and y are vectors with length n; results are returned via a and b pointers,
 *      stddev will contain std. deviation of parameter b
 *      no change occurs to x and y */
/*      should rather refer to Xplore routine (?) */
void lseline(REAL *x, REAL *y, ULONG n, REAL *a, REAL *b, REAL *stddev)
{
     ULONG i;
     REAL sumx = 0.0, sumx2 = 0.0, sumy = 0.0, sumxy = 0.0, sumdev = 0.0, temp;
     REAL aest, best, denom;

     for(i = 0; i < n; i++)
     {
          sumx += x[i];       sumx2 += x[i] * x[i];
          sumy += y[i];       sumxy += x[i] * y[i];
     }

     denom = n * sumx2 - sumx * sumx;

     /* results */
     aest = ( sumx2 * sumy - sumx * sumxy ) / denom;
     best = ( n * sumxy - sumx * sumy ) / denom;

     /* std. deviation */
     for(i = 0; i < n; i++)
     {
          temp = y[i] - aest - best * x[i];
          sumdev += temp * temp;
     }

     *a = aest;
     *b = best;
     *stddev = sqrt(sumdev / denom);
}

/*
 * Main routines
 */

/* sumxr: computes sum of x_i * sign(y_i - a - b x_i) for a given value of b
 *        a is computed as an implied estimate of b ... a = med( y_i - b x_i )
 *        computed a is returned back
 * parameters: vectors x and y of length n and the value of b for which the sum
 *             is evaluated; temp is a temporary vector for median routine of the
 *             same length as x and y are */
REAL sumxr(REAL *x, REAL *y, ULONG n, REAL b, REAL *temp, REAL *a)
{
     ULONG i;
     REAL aest, d, sum = 0.0;

     for(i = 0; i < n; i++)
          temp[i] = y[i] - b * x[i];
     aest = median(temp, n);

     for(i = 0; i < n; i++)
     {
          d = y[i] - aest - b * x[i];
          if (y[i] != 0) d /= rabs(y[i]);
          if (rabs(d) > EPSILON) sum += ((d >= 0.0) ? x[i] : -x[i]);
     }

     *a = aest;

     return(sum);
}

/* L1line: computes L1 estimate for regression model y_i = a + b x_i
 *         results are a, b
 * parameters: x and y are vectors of length n; a, b will contains results
 *             x and y will not be changed */
/* a = a(b) (see sumxr), b is found using bracketing and bisection */
/* return: OK (0) ... or error (!= 0) */
int l1sline(REAL *x, REAL *y, ULONG n, REAL *a, REAL *b)
{
     REAL f, f1, f2, b1, b2, stop;
     REAL aest, best, bdev;

     REAL *temp;

     /* temporary storage */
     temp = new REAL[n]; //malloc(n * sizeof(REAL));
     if (temp == NULL) return NOMEMORY;

     /* initialize starting interval - length 3*bdev */
     lseline(x, y, n, &aest, &best, &bdev);

     b1 = best;
     f1 = sumxr(x, y, n, b1, temp, &aest);
     /* guess for first bracket is interval 3*bdev in downhill direction of f1 */
     b2 = best - 3.0 * bdev * sign(f1);
     f2 = sumxr(x, y, n, b2, temp, &aest);

     /* do we have to iterate for a solution? */
     if (b1 == b2)
     {
          *a = aest;
          *b = best;

          delete[] temp;
          return OK;
     }

     /* bracketing: sumxr converges to negative / positive number if b goes to +/- infinity */
     while(f1 * f2 > 0.0)
     {
          best = b2 + 1.6 * (b2 - b1);
          b2 = b1;
          f2 = f1;
          b1 = best;
          f1 = sumxr(x, y, n, b1, temp, &aest);
     }

     /* bisection algorithm - refine until |b2-b1| < STOPRULE * bdev */
     stop = bdev * STOPRULE;
     while( rabs((REAL) (b2 - b1)) > stop )
     {
          best = 0.5 * (b1 + b2);
          if ( (best == b1) || (best == b2) ) break;
          f = sumxr(x, y, n, best, temp, &aest);

          if (f * f1 >= 0.0)
          {
               b1 = best;
               f1 = f;
          }
          else
          {
               b2 = best;
               f2 = f;
          }
     }

     /* return values */
     best = 0.5 * (b1 + b2);
     (void) sumxr(x, y, n, best, temp, &aest);
     *a = aest;
     *b = best;

     /* deallocation */
     delete[] temp;

     return OK;
  }

  void slowess (double *x_in, double *y_in, long n_in, double f_in,
                       long nsteps_in, double delta_in, double *ys_out)
  { /*
     PROGRAMMER : S. Klinke
     DAdoubleE       : October 88	(for XploRe 2.0)
                  April 92      (for XploRe 3.0)
                  February 97   (for XploRe 4.0)
     METHOD     :
     The fitted values are computed by using the nearest neighbor routine and
     robust locally weighted regession of degree 1 with the tricube weigth
     function. A few addtional features has added. Suppose r is FN truncated
     to an integer. Let h be the distance to the r-th nearest neighbor to X(i).
     All points within h of x(i) are used. Thus if the nearest neighbor is
     exactly the same distance as other points, more than r points can
     possibly be used for the smooth at x(i). There ar two cases where robust
     locally weigthed regression of degree 0 is actually used at x(i). One
     case occurs when the weigthed standard error of x(i) with respect to the
     weights w(j) is less than 0.001 times the range of the x(i), where w(j) is
     the weight assigned to the j-th point of x (the tricube weight times the
     robustness weight) divided by the sum of all weights. Finally, if the w(j)
     are all zero for the smooth at x(i), the fitted value is taken to be y(i).

     ARGUMENT DESCRIPTION :
     x_in       input  abscissas of the points on the scatterplot
	               the values in x must be orderd from smallest to largest
     y_in       input  ordinates of the points on the scatterplot
     n_in       input  dimension of x_in, y_in, ys_out
     f_in       input  f_in is the farction of points used to compute each fitted
				 value
     nsteps_in  input  the number of iterations in the robust fit;
				 if nsteps_in=0 the nonrobust fit is returned
     delta_in   input  nonnegative parameter which may be used to save
				 computations
     ys_out     output ys(i) is the fitted value at x(i)

     ADDITONAL TYPES AND PROCEDURES
     You have to define a global type "workarray" like this :

     TYPE workarray = ARRAY [1..1000] OF real;

     because Turbopascal 4.0 want to have a type-definiton in the parameter
     definition

     The programm calls 4 Functions and 1 Procedure :

     min  (a,b)       Minimum of two integers
     max  (a,b)       Maximum of two integers
     rmax (a,b)       Maximum of two reals
     pot  (x,n)       The n-th power of x  (n integer, x real)
     num_sort (a, a, a, n)
		      Sorting the array a (=workarray) with n elements
   */

    double *rw_out, *res_out;
#ifndef MEMDEBUG
    rw_out = new double[n_in];
    res_out = new double[n_in];
#else
    rw_out = (double *) xmem.newptr (sizeof(double), n_in, "", "slowess", "rw_out");
    res_out = (double *) xmem.newptr (sizeof(double), n_in, "", "slowess", "res_out");
#endif
    lowess_int (x_in, y_in, n_in, f_in, nsteps_in, delta_in,
		ys_out, rw_out, res_out);
#ifndef MEMDEBUG
    delete[] rw_out;
    delete[] res_out;
#else
    xmem.delptr (rw_out);
    xmem.delptr (res_out);
#endif
  }

  void lowest (double *x, double *y, long n, double xs, double *ys, long nleft, long nright, double *w, long userw, double *rw, long *ok)
  { long nrt, j;
    double a, b, c, h, r, h1, h9, rang;

    rang = X(n)-X(1);
    h    = max1(xs-X(nleft), X(nright)-xs);
    h9   = 0.999*h;
    h1   = 0.001*h;
    /* sum of weights */
    a = 0.0;
    FOR (j,nleft,n)
    { /* compute weights (pick up all ties on right) */
      W(j) = 0.0;
      r    = fabs(X(j)-xs);
      if (r>h9)
      { /* get out at first zero wt on right */
  	 if (X(j) > xs) goto label7;
      }
      else
      { if (r>h1) W(j) = TRI(1.0-TRI(r/h));
	/* small enough for non-zero weight */
	else W(j) = 1.0;
        if (userw) W(j) *= RW(j);
	a += W(j);
        /* rightmost pt (may be greater than nright because of ties) */
      }
    }

  label7:
    nrt = j-1;
    *ok = (a>0.0);
    if (*ok)
    { /* weighted least squares */
      FOR (j,nleft,nrt)  /* MAKE SUM OF W(J) == 1 */
	W(j) /= a;
      if (h>0.0)
      { a = 0.0;
	/* use linear fit */
	/* weighted center of x values */
	FOR (j,nleft,nrt) a += (W(j)*X(j));
	b = xs-a;
	c = 0.0;
	FOR (j,nleft,nrt) c += (W(j)*SQR(X(j)-a));
	if (::sqrt(c)>0.001*rang)
	{ b /= c;
	  /* points are spread out enough to compute slope */
	  FOR (j,nleft,nrt) W(j) *= (b*(X(j)-a)+1.0);
	}
      }
      *ys = 0.0;
      FOR (j,nleft,nrt) *ys += (W(j)*Y(j));
    }
  } /* lowest */

  int less (const void *a, const void *b)
  { if (*((double *) a) < *((double *) b)) return (-1);
    if (*((double *) a) > *((double *) b)) return (1);
    return (0);
  }

  void lowess_int (double *x, double *y, long n, double f, long nsteps,
			  double delta, double *ys, double *rw, double *res)
  { long nright, i, j, iter, last, m1, m2, ns, nleft, ok;
    double cut, cmad, r, d1, d2, c1, c9, alpha, denom;
    double *temp1, *temp2;

    if (n<2)
    { YS(1) = Y(1);
      return;
    }
    /* at least two, at most n points */
    ns = (long) max1(min1(floor(f*n), (double) n), 2.0);
    FOR (iter, 1, nsteps+1)
    { /* robustness iterations */
      nleft  = 1;
      nright = ns;             /* index of prev estimated point */
      last   = 0;              /* index of current point */
      i      = 1;
      do
      { /* printf("%i %i\n", d1, d2); getch(); d1>d2 */
	d1 = X(i)-X(nleft);
	d2 = X(nright+1)-X(i);
	while ((nright<n) && (d1>d2))
	{ /* move nleft, nright to right if radius decreases*/
	  d1 = X(i)-X(nleft);
	  /* if d1<=d2 with x(nright+1)==x(nright), lowest fixes*/
	  d2 = X(nright+1)-X(i);
	  if (d1>d2)
	  { /* radius will not decrease by move right */
            nleft++;
            nright++;
	  }
	}
	/* fitted value at x(i) */

	lowest(x,y,n,X(i),AYS(i),nleft,nright,res,iter>1,rw,&ok);

	if (!ok) YS(i) = Y(i);
	/* all weights zero - copy over value (all rw==0) */
	if (last<i-1)
	{ denom = X(i)-X(last);
	  /* skipped points -- interpolate non-zero - proof? */
	  FOR (j,last+1,i-1)
	  { alpha = (X(j)-X(last))/denom;
            YS(j) = alpha*YS(i)+(1.0-alpha)*YS(last);
	  }
	  /* last point actually estimated */
	}
	last = i;
	/* x coord of close points */
	cut = X(last)+delta;
	i = last+1;     /* find close points */
			/* i one    beyond last pt within cut */
	while ((i<=n) && (X(i)<=cut))
	{ if (X(i)==X(last))
	  { YS(i) = YS(last);
	    /* exact match in x */
	    last = i;
	  }
	  i++;
	}
	/* back 1 point so interpolation    within delta,
	   but always go forward */
	i = max1(last+1,i-1);
      }
      while (last<n);

      /* residuals */
      FOR (i,1,n) RES(i) = Y(i)-YS(i);
      if ((iter > nsteps)) return ;
      /* compute robustness weights except last time */
      FOR (i,1,n) RW(i) = fabs(RES(i));
      qsort (rw,n,sizeof(double),less);
      m1 = (n/2)+1;
      m2 = n-m1+1;
      /*6 median abs resid */
      cmad = 3.0*(RW(m1)+RW(m2));
      c9   = 0.999*cmad;
      c1   = 0.001*cmad;
      FOR (i,1,n)
      { r = fabs(RES(i));
	if (r>c1)
	{ if (r<=c9)
	    RW(i) = SQR(1.0-SQR(r/cmad));
	  else
	    RW(i) = 0.0;
	} /* near 1, avoid underflow */
	else
	{ RW(i) = 1.0;
	  /* near 0, avoid underflow */
	}
      }
    }
  } /* lowess_int */

// K-means clustering algorithms                                 02/94
//
// Exchange technique:  The centroids of the clusters are corrected
// (standard;npass>=0   immediately after shifting a point from one
//                      to another cluster (stop if the number of 
//                      shifted points is <= npass)
//            Criteria: Minimum of sum of within-clusters variances
//           Special 1: 10000<=npass<20000: Minimum distance criteria
//                      (stop if # of shifted points <= npass-10000)
//           Special 2: npass>=20000: Gradient technique, i.e. one (the
//                      best in minimum variance sense) shift per 
//                      iteration cycle only (npass-20000 cycles at most)
//
// Minimum distance technique: The centroids of the clusters are corrected
// (npass<0)                   after each iteration cycle (stop if the 
//                             number of shifted points <= |npass|-1)
//                             !!! hereby clusters can become empty - we 
//                             start with a new initial partition !!!            
//                   Criteria: Minimum distance
//                    Special: npass<=-20000: Gradient technique, i.e. one 
//                             (the minimum distance) shift per iteration 
//                             cycle only (|npass|-20000 cycles at most)  
//
// Special offer (npass=30000): no clustering, assignment of ungrouped 
//                              points into classes by minimum distance  
//
// author: H.-J. Mucha, Institute of Applied Analysis and Stochastics
//         Mohrenstr.39, D-10117 Berlin, email: mucha@iaas-berlin.d400.de

// SK 02.09.97
  void kmeans(long *ik, double *xx, double *cw, double *v, double *wk,
        double *we, double *xm, long n, long p, long nclu, long npass)
  {
    long i, j, k, k1, ki, kk, ks, kt, ns, kold, ij, nsh, gd, md, gk, gi;
    double d, x1, fc, ds, dss, fk1, fff, exl, ttt, xxm, gmax;

    /* Parameter adjustments */
    --we;
    --wk;
    --v;
    --xx;
    --ik;
    --cw;
    --xm;

     kt = 0;
     exl = .999999f;
     ki = 0;
     ks = 0;
     ns = 0;
     gmax = -1.f;
     md = 0;
     gd = 0;
     if(npass<=-20000 || npass>=20000) gd = 1;
     nsh = npass;
     if(npass>=20000) nsh=npass-20000;
     if(npass<=-20000) nsh=-npass-20000;     
     if(npass<0) md=1;
     if(npass<0 && gd!=1) nsh=-npass-1;
     if(npass>=10000 && gd!=1) md=-1;
     if(md==-1) nsh=npass-10000;
//printf ("nsh=%li\n", nsh);
/*        set the initial values of wk (masses of clusters),            */
/*        xm (cluster centroids), and v (within-clusters variances)     */
L82: for (k = 1; k <= nclu; ++k) {
     wk[k] = 0.f;
     for (j = 1; j <= p; ++j) {
         xm[(j-1)*nclu+k] = 0.f;
         v[(j-1)*nclu+k] = 0.f;
         }
     }
     if(gd==0) goto L22;
     if(gmax<-2.) goto L78;
     if(gmax<0.) goto L22;
     ik[gi]=gk;
     ns=1;
     gmax=-3.0f;
     if(kt>=nsh) goto L78;     

/*        compute wk and xm of the initial partition                    */
L22: for (i = 1; i <= n; ++i) {
     fff = cw[i];
     if (ik[i] <= 0 || ik[i] > nclu) ik[i] = 0;
        else {
          k1 = ik[i];
          xxm = wk[k1];
          wk[k1] += fff;
          for (j = 1; j <= p; ++j) {
          x1 = fff / wk[k1] * (xx[(j-1)*n+i] - xm[(j-1)*nclu+k1]);
          xm[(j-1)*nclu+k1] += x1;
          }
        }
     }
   if(npass==30000) goto L70;
   for (k=1; k<=nclu; ++k) {
       if(wk[k]==0.f) {
          for (i=1; i<=n; ++i) ik[i]=k;
          goto L82;
       }
   }
   if(kt==0) goto L71;
   goto L84;
/*                 classify ungrouped points into the clusters using  */
/*                 the minimal distance criteria                      */
L70: for (i = 1; i <= n; ++i) {
      ttt = 1.E36;
      kk = ik[i];
        if (kk >= 1) goto L80;
      for (k = 1; k <= nclu; ++k)  {
          ds = 0.0;
          for (j = 1; j <= p; ++j) {
           dss = xx[(j-1)*n+i] - xm[(j-1)*nclu+k];
           ds += dss * dss * we[j];
          }
          if (ds <= ttt) {
           ttt = ds;
           kk = k;
          }
      }
L80:      ik[i] = kk;
     }
     goto L78;

/*          md= 0 & gd=0: Standard (Exchange technique, variance criteria) */
/*          md=-1 & gd=0: Exchange technique, distance criteria            */ 
/*          md= 1 & gd=0: Minimum distance technique                       */
/*                        iteration process stops if the number of shifted */   
/*                        points <= nsh                                    */
/*               gd=1   : gradient technique                               */   
/*                        nsh iteration cycles at most                     */

L71: ++ks;
       if (ks - n > 0) goto L78;
L72: ++ki;
       if (ki - n <= 0) goto L73;
     ++kt;
     if(gd==1) goto L82;
     if(ns<=nsh) goto L78;
     if(md==1) goto L82;
L84: ki = 1;
     ns = 0;
L73: k1 = ik[ki];
     fc = 1.E36;
     fk1 = 0.f;
     ttt = 1.E36;
     d = cw[ki];

/*        look for the minimum variance contribution of point ki        */
     // printf ("kmeans: %li\n", ki);
     for (k = 1; k <= nclu; ++k)  {
        ds = 0.0;
        for (j = 1; j <= p; ++j) {
          dss = xx[(j-1)*n+ki] - xm[(j-1)*nclu+k];
          ds += dss * dss * we[j];
        }
          if (k - k1 != 0) goto L75;
          if(wk[k]* exl - d > 0.f)  fk1= wk[k]* d * ds / (wk[k]- d);
        ttt = fk1;
        if(md!=0) ttt=ds;
        goto L76;
L75:    fff = wk[k] * d * ds / (wk[k] + d);
        if(md!=0) fff=ds;
        if (fff - fc >= 0.f) goto L76;
     fc = fff;
     kk = k;
L76: ;
     }

/*        shift point ki into cluster kk; change the centroids          */
       if (fc - exl * ttt >= 0.f) goto L71;
     ks = 0;
     if(gd==1) { if(gmax>=ttt-fc) goto L72;
                 gmax=ttt-fc;
                 gi=ki;
                 gk=kk;
                 goto L72;}
     ik[ki] = kk;
     ++ns;
     if(md==1) goto L72;
     kold = k1;
       if( k1 -1 < 0 ) k1 = 1;
       if ( wk[k1]* exl - d > 0 && kold != 0) wk[k1] = wk[k1] - d;
     wk[kk] += d;
     for (j = 1; j <= p; ++j) {
      ij = (j-1) * nclu;
      if (kold !=0) xm[ij+k1] -= d / wk[k1]* (xx[(j-1)*n+ki] - xm[ij+k1]);
      xm[ij+kk] += d / wk[kk] * (xx[(j-1)*n+ki] - xm[ij+kk]);
     }
     goto L72;

/*        compute the statistics of the final partition                 */
L78: for (k = 1; k <= nclu; ++k) {
      wk[k] = 0.f;
      for (j = 1; j <= p; ++j) {
          xm[(j-1)*nclu+k] = 0.f;
      }
     }
     for (i = 1; i <= n; ++i) {
      fff = cw[i];
      k1 = ik[i];
      xxm = wk[k1];
      wk[k1] += fff;
#ifdef WIN32
// VC++ 5.0 produziert: fatal error C1001: INTERNER COMPILER- FEHLER
// nur beim Optimieren
      for (j = 1; j <= p; ++j) {
int wa = (j-1)*nclu+k1;
          x1 = fff / wk[k1] * (xx[(j-1)*n+i] - xm[wa]);
          v[wa] += xxm * wk[k1] / fff * x1 * x1;
          xm[wa] += x1;
#else
      for (j = 1; j <= p; ++j) {
          x1 = fff / wk[k1] * (xx[(j-1)*n+i] - xm[(j-1)*nclu+k1]);
          v[(j-1)*nclu+k1] += xxm * wk[k1] / fff * x1 * x1;
          xm[(j-1)*nclu+k1] += x1;
#endif
      }
     }
     for (k = 1; k <= nclu; ++k) {
      for (j = 1; j <= p; ++j) v[(j-1)*nclu+k]=v[(j-1)*nclu+k]/wk[k];
     }
     return;
  }


//  Hierarchical cluster analysis, 8 methods (agglomerative)     12/92
//
//  author: H.-J. Mucha, Institute of Applied Analysis and Stochastics
//          Mohrenstr.39, D-10117 Berlin
// SK 02.09.97

  void agglom(long *ka, long *kb, double *h, long *kp, double *q, double *x,
  // neu!: y: Clusterdistanzen (wie x gespeichert); wy: Masse der Cluster;
  // neu!: knr: Anzahl der Cluster (1>knr>n: y und wy werden nicht erstellt)
          long n, long mtd, double beta, double *y, double *wy, long knr)
  {
    double d, f, v, w, hi, hj, xi, xj;
    int i, j, k, m1, ic, jc, ii, jj, ind;
    int indk, indw;

    /* Parameter adjustments */
    --x;
    --q;
    --h;
    --kb;
    --ka;
    --kp;
    --y;
    --wy;

    m1 = n - 1;
    v = 1.f - beta;
    w = v / 2.f;
    for (i = 1; i <= n; ++i) kp[i] = 1;
    k = 0;
    indk = 1;
    indw = 1;

/*        change the distances in the case of Ward's method             */
    if (mtd == 7)  {
     for (i = 2; i <= n; ++i) {
         jj = i - 1;
         ind = jj;
         for (j = 1; j <= jj; ++j) {
            x[ind] = x[ind] * q[i] * q[j] / (q[i] + q[j]);
            ind = ind + n - j - 1;
         }
     }
    }

/*        Iteration: collaps the two most similar                       */
/*                   clusters at each step                              */
L48:++k;
    if (k == n) return;
    SetMax (&d);
    ic = -1;
    jj = 1;
    for (i = 1; i <= m1; ++i) {
     if (kp[i] >= 0) goto L32;
     jj = jj + n - i;
     goto L24;
L32:     ii = i + 1;
    if (n-k == knr-1) {
       wy[indw]=q[i];
       indw = indw + 1;
    }
     for (j = ii; j <= n; ++j) {
         if (kp[j] < 0) goto L23;
           if (n-k == knr-1) {
       y[indk] =x[jj];
       indk=indk+1;
    } 
         if (x[jj] - d > 0.f) goto L23;
         ic = i;
         jc = j;
         d = x[jj];
L23:         ++jj;
     }
L24:     ;
    }
    kp[jc] = -1;
    if (ic==-1) // SK 04.05.99
    { XAerror (INTERN);
      return;
    }
    hi = q[ic];
    q[ic] += q[jc];
    ka[k] = ic;
    kb[k] = jc;
    h[k] = d;
    hj = q[jc];
    f = 1.f / (hi + hj);
    ii = ic - 1;
    jj = jc - 1;
    for (i = 1; i <= n; ++i) {
     j = m1 - i - 1;
     if (kp[i] <  0) goto L29;
     if (i - ic == 0) goto L29;
     xi = x[ii];
     xj = x[jj];

/*        compute the distances between the new created                 */
/*        cluster and the remaining ones                                */
     switch (mtd) {
         case 1:  goto L1;
         case 2:  goto L2;
         case 3:  goto L3;
         case 4:  goto L5;
         case 5:  goto L6;
         case 6:  goto L7;
         case 7:  goto L8;
         case 8:  goto L9;
     }
L1:     x[ii] = xi;
     if ( xi > xj ) x[ii] = xj;
     goto L29;
L2:     x[ii] = xj;
     if ( xi > xj) x[ii]= xi;
     goto L29;
L3:     x[ii] = (xi + xj) * .5f;
     goto L29;
L5:     x[ii] = (xi + xj) * .5f - d * .25f;
     goto L29;
L6:     x[ii] = f * (hi * xi + hj * xj);
     goto L29;
L7:     x[ii] = f * (hi * xi + hj * xj - hi * hj * f * d);
     goto L29;
L8:     x[ii]=((hi + q[i])* xi +(hj + q[i])* xj - q[i]* d)/(q[i] + hi + hj);
     goto L29;
L9:     x[ii] = w * (xi + xj) + beta * d;
L29:     ++ii;
     ++jj;
     if (i < ic) ii += j;
     if (i < jc) jj += j;
    }
    goto L48;
}

//  cut a hierarchy, and save the corresponding partition        10/93
//
//  author: H.-J. Mucha, Institute of Applied Analysis and Stochastics
//          Mohrenstr.39, D-10117 Berlin

  void parti(long *ka, long *kb, double *h, long *kc, long *kq, long *kd,
        long *kf, long *kt, long *jx, double *d, long *np, long ll, long n)
/* setting kc from ka, kb and h, also np = partition of datapoint */
  {
    long i, j, k, k1, k2, m1, ic, jc, ii, ik, ij, kh;
    double s;

    /* Parameter adjustments */
    --np;
    --d;
    --jx;
    --kt;
    --kf;
    --kd;
    --kq;
    --kc;
    --h;
    --kb;
    --ka;

    kh = n;
    if ( kh > ll ) kh = ll;
    m1 = n - 1;
    h[n] = 1.E36;
    for (i = 1; i <= m1; ++i) {
     if (h[i + 1] - h[i] < 0.f) {
         for (j = i + 1; j <= n; ++j) h[j] *= 1.000001f;
     }
    }
    j = 1;
    for (i = 1; i <= n; ++i) {
     kq[i] = 0;
     jx[i] = 0;
     np[i] = 0;
    }

/*        compute the cut level for kh clusters in a hierarchy          */
    k1 = n - kh;
    s = h[1] * .9999f;
      if (kh < n) s = (h[k1] + h[k1 + 1]) / 2.f;
    k = 1;
    ic = 1;
    kq[1] = 1;
    kc[1] = 1;

/*        investigate a hierarchy in order to obtain kh clusters        */
L42:
    jx[ic] = -1;
    ++k;
      if (k - n > 0) goto L46;
L43:
    for (i = 1; i <= m1; ++i) {
     jc = kb[i];
       if (jx[jc] < 0 || ka[i] != ic) goto L44;
     ic = jc;
     kq[k] = ic;
     kc[k] = ic;
     d[k - 1] = h[i];
     j = k;
     goto L42;
L44:
     ;
    }
    kq[j] = 0;
    k1 = j - 1;
    i = k1;
    for (k2 = 1; k2 <= k1; ++k2) {
     if (kq[i] == 0) goto L45;
     j = i;
     ic = kq[j];
     goto L43;
L45:
     --i;
    }
L46:
    j = 1;
    d[n] = 0.f;
    for (i = 1; i <= m1; ++i) {
     k2 = kc[i];
     kq[k2] = j;
       if (d[i] > s) ++j;
    }
    k2 = kc[n];
    kq[k2] = j;
    h[n] = 0.f;
    jc = n - kh + 1;
    k = 1;
    jx[1] = ka[jc];
    for (i = jc; i <= m1; ++i) {
     for (ii = 1; ii <= k; ++ii) {
         if (ka[i] - jx[ii] == 0) goto L4;
         ;
     }
     ++k;
     jx[k] = ka[i];
L4:
     for (ii = 1; ii <= k; ++ii) {
         if (kb[i] - jx[ii] == 0) goto L2;
         ;
     }
     ++k;
     jx[k] = kb[i];
L2:
     ;
    }
    k = 0;
    for (i = 1; i <= n; ++i) {
     for (j = 1; j <= kh; ++j) {
         if (kc[i] - jx[j] != 0) goto L8;
         ++k;
         kd[k] = kc[i];
L8:
         ;
     }
    }
    for (i = jc; i <= m1; ++i) {
     k2 = i - jc + 1;
     k = 0;
     ik = 1;
     jx[ik] = kb[i];
L16:
     ++k;
     for (j = jc; j <= i; ++j) {
         if (jx[k] - ka[j] != 0) goto L15;
         ++ik;
         jx[ik] = kb[j];
L15:
         ;
     }
       if (k - ik < 0) goto L16;
     for (j = jc; j <= n; ++j) {
         ic = j - jc + 1;
            if (ka[i] == kd[ic]) kf[k2] = ic;
         for (ij = 1; ij <= ik; ++ij) {
          if (jx[ij] == kd[ic]) kt[k2] = ic;
         }
     }
    }

/*        jx is reserved for a future connection with kmeans clustering */
/*        np contains the partition of n points into kh clusters        */
      for (i = 1; i <= n; ++i)      np[i] = 0;
      for (i = 1; i <= n; ++i)      jx[i] = i;
    for (i = 1; i <= n; ++i)
    {     k1 = jx[i];
     np[k1] = kq[i];
    }
    return;
  }

//  dendrogram of a hierarchy, saved as x-y - coordinates        12/92
//  drawing only dendrogram
//
//  author: H.-J. Mucha, Institute of Applied Analysis and Stochastics
//          Mohrenstr.39, D-10117 Berlin

//SK 02.09.97
  void dendro(long *ka, long *kb, double *h, 
         long *kc, long *ke, long *kp, long *kq,
         long *kr, long *ks, long *kt, long *ku, long *kv, long *kw,
         long *coor, long n, long *ld)
/* Object ka(i) link with object kb(i) on level h(i)  i=1,...,n-1
   Object kc(j)
   Pointwise output for dendrogram  coor (nx2)
   all others working areas
*/
  {
    long jj, izei, i, j, k, i1, j2, j1, m1, i3, nd, ii, kk;
    long ip, iq, io[73], o[73], p[73], ir, iw, ir1, ldd, s;
    double x, y, zz;

    /* Parameter adjustments */
    --kw;
    --kv;
    --ku;
    --kt;
    --ks;
    --kr;
    --kq;
    --kp;
    --ke;
    --kc;
    --h;
    --kb;
    --ka;

    ldd = 0;
    *ld = 0;
    jj = 72;
    x = jj - 1.5f;
    m1 = n - 1;

/*        rescale the distance levels in order to draw a                */
/*        well-informative dendrogram also in the case of Ward's method */
      for (i = 1; i <= n; ++i) {
      if(h[m1] < 1) h[i] = sqrt(sqrt(h[i]));
      h[i] = x4log(h[i] + 1.f);
      }
    y = (h[m1] - h[1]) / x;
    for (i = 1; i <= n; ++i) {
     iq = kc[i];
     kw[iq] = i;
    }

/*        compose the dendrogram structure and set initial values       */
    for (i = 1; i <= m1; ++i) {
     iq = ka[i];
     ir = kb[i];
     zz = (h[i] - h[1]) / y;
     if (zz > x) zz = x;
     kt[i] = (long) (zz + 2.f); // SK (long)
     ip = kw[ir] - kw[iq];
     j2 = i - 1;
     j = j2;
     if (j2 == 0) goto L77;
     for (k = 1; k <= j2; ++k) {
         if (ir != ka[j] && ir != kb[j]) goto L7;
         ip += ku[j];
         goto L12;
L7:
         --j;
     }
L77:
     ++ip;
L12:
     ku[i] = ip;
       if (ip - 2 != 0) goto L9;
     kp[i] = iq;
     kq[i] = ir;
     kr[i] = 0;
     ks[i] = 0;
     goto L14;
L9:
     j = j2;
       if (j2 == 0) goto L78;
     for (k = 1; k <= j2; ++k) {
           if (iq != ka[j] && iq != kb[j]) goto L21;
         kp[i] = kv[j];
         kr[i] = kt[j];
         goto L13;
L21:
         --j;
     }
L78:
     kp[i] = iq;
     kr[i] = 0;
L13:
     j = j2;
     if (j2 == 0) goto L79;
     for (k = 1; k <= j2; ++k) {
            if (ir != ka[j] && ir != kb[j]) goto L24;
         kq[i] = kv[j];
         ks[i] = kt[j];
         goto L14;
L24:
         --j;
     }
L79:
     kq[i] = ir;
     ks[i] = 0;
L14:
       if (ip == 2) kv[i] = iq;
     iq = kw[iq] + 1;
       if (ip == 3) kv[i] = kc[iq];
       if (ip == 2 || ip == 3) goto L6;
     iq = kq[i];
     ir = kp[i];
     k = (long) ((kw[iq] + kw[ir]) * .5f); // SK (long)
     kv[i] = kc[k];
L6:
     ;
    }

/*        compose one row (i.e., one terminal root) of a dendrogram     */
/*        in detail                                                     */
    for (i = 1; i <= n; ++i) {
       for (k = 1; k <= jj; ++k) io[k - 1] = 0;
     izei = 1;
     j1 = 0;
     ip = 0;
     iq = kc[i];
     iw = 0;
     for (j = 1; j <= m1; ++j) {
         ii = kp[j];
         kk = kq[j];
           if (ii == iq || kk == iq) goto L15;
           if (kw[ii] >= i || kw[kk] <= i) goto L32;
         ir = kt[j] - ip - 1;
           if (j1 == 0 && iw > 0) j1 = kt[j];
         ip = kt[j];
           if (ir < 0) goto L32;
         izei = izei + ir + 1;
         io[izei - 2] = 1;
         goto L32;
L15:
         i1 = kt[j];
         k = i1 - ip;
           if (k <= 0) goto L32;
L86:
         ir1 = ip + 1;
         i3 = i1 - 1;
           if (i3 - ir1 < 0) goto L38;
         for (ir = ir1; ir <= i3; ++ir) {
          io[izei - 1] = 2;
          ++izei;
         }
L38:
         io[izei - 1] = 1;
         ++izei;
           if (ip != 0) goto L40;
         iw = kt[j];
L40:
         ip = kt[j];
L32:
         ;
     }

/*        save the pixels of one row of the dendrogram as               */
/*        ordinary x-y - coordinates                                    */
     k = kc[i];
     for (j = 0; j < jj; ++j) {
         if (io[j] != 0) {
            iw = 2 * (n - i) + 1;
            coor[ldd] = j;
            ++(*ld);
            ++ldd;
            coor[ldd] = iw;
            ++ldd;

/*        compute an additional dendrogram line                         */
/*            iq = *ld;                              */
           if (i == 1) goto L1;
           if (io[j] != 1 || o[j] !=1) goto L1;
           if (i == 2) goto L2;
           if (j == 0 || j == jj-1) goto L2;
           s = j - 1;
           if (io[j] == 1 && o[j] == 1 && p[j] == 1 &&
               io[s] == 2 && o[s] == 2 && p[s] == 2 &&
               p[j+1] == 2) goto L1;
           if (io[j] == 1 && o[j] == 1 && p[j] == 1 &&
               io[s] == 2 && o[s] == 2 && p[j+1] == 2) goto L1;
L2:               coor[ldd] = j;
               ++(*ld);
               ++ldd;
               coor[ldd] = iw + 1;
               ++ldd;
            }
L1:     ;
     }
     for (j = 0; j < jj; ++j) {
         if (i > 1) p[j] = o[j];
         o[j] = io[j];
     }
    }
    return;
  }


  void dendro2 (long *ka, long *kb, long *kc, double *h, long len,
              double *x, double *y, double *z, long *l)
  { long i, *cluno, *hnorm, j, k, inmerge, ismerge;

#ifndef MEMDEBUG
  cluno = new long[len];
  hnorm = new long[len];
#else
  cluno = (long *) xmem.newptr (sizeof(long), len, "", "dendro2", "cluno");
  hnorm = (long *) xmem.newptr (sizeof(long), len, "", "dendro2", "hnorm");
#endif
  for (i=0; i<len; i++)
  { cluno[i] = kc[i];
    if (i<len-1) 
      hnorm[i] = (long) (1.0 + 70.0 * (h[i]-h[0])/(h[len-2]-h[0]));
  }

  k = *l = 0;
  for (i=0; i<72; i++)
  { // merging required
    ismerge = 0;
    while (hnorm[k] == i)
    { inmerge = 0;
      ismerge = 1;
      for (j=0; (j<len) && (inmerge<2); j++)
      { switch (inmerge)
        { case 0 : if (cluno[j] == ka[k])
                   { cluno[j] = -1;
                     *(x+(*l)) = i;
                     *(y+(*l)) = j+1;
                     *(z+(*l)) = 1;
                     (*l)++;
                     *(x+(*l)) = i;
                     *(y+(*l)) = j+1.5;
                     *(z+(*l)) = 0;
                     (*l)++;
                     inmerge = 1;
                   }
                   break;
          case 1 : if (cluno[j] != kb[k])
                   { *(x+(*l)) = i;
                     *(y+(*l)) = j+1;
                     *(z+(*l)) = 0;
                     (*l)++;
                     *(x+(*l)) = i;
                     *(y+(*l)) = j+1.5;
                     *(z+(*l)) = 0;
                     (*l)++;
                   }
                   else
                   { cluno[j] = ka[k];
                     inmerge = 2;
                   }  
        }
      }
      k++; 
    }
    // go on with points
    for (j=0; j<len; j++)
    { if (cluno[j]>0)
      { *(x+(*l)) = i;
        *(y+(*l)) = j+1;
        *(z+(*l)) = ismerge;
        (*l)++;
      } 
    }
  }
#ifndef MEMDEBUG
  delete[] hnorm;
  delete[] cluno;
#else
  xmem.delptr (hnorm);
  xmem.delptr (cluno);
#endif
}


  CartSplit::CartSplit (int maxlen)
  { nr = new int[maxlen];
    nl = new int[maxlen]; 
    dr = new doublep[maxlen];
    dl = new doublep[maxlen];
    i  = 0;
  }

  CartSplit::~CartSplit()
  { int j;
    delete[] nr;
    delete[] nl;
    for (j=0; j<i; j++)
    { if (dl+j) delete[] (*(dl+j));
      if (dr+j) delete[] (*(dr+j));
    }
    delete[] dl;
    delete[] dr;
  }
  
  void CartSplit::append (int nli, double *pli, int nri, double *pri)
  { nl[i] = nli;
    nr[i] = nri;
    dl[i] = pli;
    dr[i] = pri;
    i++;
  }

  int CartSplit::length ()
  { return i; }

  int CartSplit::operator() (int v, int ni)
  { return (v ? nl[ni] : nr[ni]); }

  double CartSplit::operator() (int v, int ni, int nj)
  { return (v ? dl[ni][nj] : dr[ni][nj]); }

  double cut, *tmp;
  long (*cmpop) (double, double), treeform, lineform;

// SK 02.09.97
  long build_tree (double *vec, long i, long *pos, double *x, double *y,
                   double *linex, double *liney, double *ind, 
                   long *linen, long m, long *no)
  { if (i>=m)
    { XAerrort (PARAMFALSE, "tree: end of tree expected");
      return -1;
    }  
//    printf ("build_tree %li\n", i);
    if (XisNaN(*(vec+i)) || cmpop(*(vec+i),cut))
    { /* end knot reached */ 
      if (*pos==m) 
      { XAerrort (PARAMFALSE, "tree: end of tree expected");
        return -1;
      }  
      *x = *pos;
      (*pos)++;
      *y = cut;
      *no = i;
      return i;
    }
    else
    { long j, li, ri;
      double lx, ly, rx, ry, ylevel = *(vec+i);
    /* go to left knot */
      li = build_tree (vec, i+1, pos, &lx, &ly, linex, liney, ind, linen, m, &j);
    /* go to right knot */
      ri = build_tree (vec, j+1, pos, &rx, &ry, linex, liney, ind, linen, m, &j);
      *x = lx + 0.5 * treeform * (rx-lx);
      *y = ylevel;
      if (lineform)
      { *(linex+*linen) = lx;
        *(liney+*linen) = ly;
        *(ind+*linen) = li;
        (*linen)++;
        *(linex+*linen) = 0.5*(lx+rx);
        *(liney+*linen) = ylevel;
        *(ind+*linen) = 0;
        (*linen)++;
        *(linex+*linen) = rx;
        *(liney+*linen) = ry;
        *(ind+*linen) = ri;
        (*linen)++;
        *(linex+*linen) = ::sqrt(-1);
        *(liney+*linen) = ::sqrt(-1);
        *(ind+*linen) = 0;
        (*linen)++;
      }
      else
      { *(linex+*linen) = lx;
        *(liney+*linen) = ly;
        *(ind+*linen) = li;
        (*linen)++;
        *(linex+*linen) = lx;
        *(liney+*linen) = ylevel;
        *(ind+*linen) = 0;
        (*linen)++;
        *(linex+*linen) = rx;
        *(liney+*linen) = ylevel;
        *(ind+*linen) = 0;
        (*linen)++;
        *(linex+*linen) = rx;
        *(liney+*linen) = ry;
        *(ind+*linen) = ri;
        (*linen)++;
        *(linex+*linen) = ::sqrt(-1);
        *(liney+*linen) = ::sqrt(-1);
        *(ind+*linen) = 0;
        (*linen)++;
      }
      *no = j;
      return i;
    }
  }
  
  long greatertree (double l, double r)
  { return (l>r); }

  long smallertree (double l, double r)
  { return (l<r); }


// SK 02.09.97
void distance(double *x, int mtc, double *w, int n, int p, double *d)
{
    int i, j, kk, ind1, mm, kr, n1;
    double s, t, u, si, sj, t1, t2, t3, t4, t5, t6, *rj, xh, xy;
    double xii, xjj, xdii, xdjj;

    /* Parameter adjustments  */
    --d;
    --w;
    --x;

    n1 = n - 1;
    ind1 = 1;
    xy = 0.0;
    xh = 0.0;
#ifndef MEMDEBUG
    rj = new double[p+1];
#else
    rj = (double *) xmem.newptr(sizeof(double), p+1, "", "distance", "rj");
#endif
/*        compute the marginal frequencies (for Chisquare distance)     */
    for (j = 1; j <= p; ++j) {
       xy = xy + w[j];
       rj[j] = 0.0;
       for (i = 1; i <= n; ++i) {
            xh = xh + x[(j-1)*n+i];
            rj[j]= rj[j] + x[(j-1)*n+i];
       }
    }

/*        compute the distance between two row points each              */
    for (i = 1; i <= n1; ++i) {
       si = 0.0;
       for (mm =1; mm <= p; ++mm)
            si = si + x[(mm-1)*n+i];
       kk = i + 1;
       for (j =kk; j <= n; ++j) {
         t1 = 0.0;
         t2 = 0.0;
         t3 = 0.0;
         t4 = 0.0;
         t5 = 0.0;
         t6 = 0.0;
         sj = 0.0;
         for (mm =1; mm <= p; ++mm)
            sj = sj + x[(mm-1)*n+j];

/*        consider each column at one time                              */
         for (mm = 1; mm <= p; ++mm) {
          xii = x[(mm-1)*n+i];
          xjj = x[(mm-1)*n+j];
          if (mtc - 5 > 0) goto L40;
          t =  xii - xjj;
          if (t < 0) t = -t;
          u = t * w[mm];
          switch (mtc) {
              case 1:  goto L31;
              case 2:  goto L32;
              case 3:  goto L33;
              case 4:  goto L30;
              case 5:  goto L35;
          }
L31:          t1 = t1 + t * u;
          goto L2;
L32:          t1 = t1 + u;
          goto L2;
L33:          if (t1 < u) t1 = u;
          goto L2;
L30:          t1 = t1 + xii * xjj;
          t2 = t2 + xii * xii;
          t3 = t3 + xjj * xjj;
          goto L2;
L35:          t2 = xii / si - xjj / sj;
          t1 = t1 + t2 * t2 / rj[mm] * xh;
          goto L2;

/*        dichotomize the data in the case of distance measures         */
/*        for 0-1 - data values                                         */
L40:            xdii = 0.0;
            if(xii >= 1.0) xdii = 1.0;
          xdjj = 0.0;
            if(xjj >= 1.0) xdjj = 1.0;
          t5 = t5 + xdii * w[mm];
          t6 = t6 + xdjj * w[mm];
          if (xdii < xdjj) t4 = t4 + xdii * w[mm];
              else t4 = t4 + xdjj * w[mm];
L2:         ;
         }

/*        special treatment of the sum of absolute differences          */
         s = t1;
         switch (mtc) {
          case 1:  goto L45;
          case 2:  goto L5;
          case 3:  goto L5;
          case 4:  goto L39;
          case 5:  goto L1;
          case 6:  goto L48;
          case 7:  goto L49;
          case 8:  goto L50;
         }
L45:         s = sqrt(s);
         goto L5;
L39:         t2 = t2 * t3;
         s = -1.f;
         if (t2 >= 1.E-17) s = t1 / sqrt(t2);
         s = 2.f - s - s;
         goto L5;
L48:         t = t5 + t6 - t4 - t4;
         s = t * 2 / (xy + t);
         goto L5;
L49:         t = t5 + t6 - t4;
         s = 1.0;
         if (t >= 1.E-17) s = (t - t4) / t;
         goto L5;
L50:        s = (t5 + t6 - t4 - t4) / xy;
         goto L5;
L1:         s = s * si * sj / xh / xh / (si / xh + sj / xh );
L5:
         d[ind1] = s;
         ++ind1;
     }
L4:     ;
    }
#ifndef MEMDEBUG
    delete[] rj;
#else
    xmem.delptr (rj); rj = NULL;
#endif
    return;
}

  double *lessvec;

  int lessind (const void *a, const void *b)
  { if (lessvec[*((long *) a)] < lessvec[*((long *) b)]) return (-1);
    if (lessvec[*((long *) a)] > lessvec[*((long *) b)]) return (1);
    return (0);
  }


  int split (double *x, long n, long p, double *y, double *t, long *grp, long nmaxgrp, double vmaxgrp, long actgrp,
             double *vec, CartSplit *val, long *nextgrp, long *kk, double mg, double dg, long ng,
             double *mean, double *var, double *nelem, long nmingrp, double *endpoint)
  { long i, j, k, l, m, nd, nf, *ind, nl, nr, nextgrpb, imin, jmin, *nyl, *indy,
         nlmin, nrmin, *indmin, *xind, ci, cj, *indyy;
    double *dfit, *dvec, ml, mr, dr, dl, dmin, vmin, *ymean, mlmin, mrmin, drmin,
           dlmin, *dv, *dvr, *memchk, *index;
    FILE *f;

/*    f = fopen ("cart.log", "a");
    fprintf (f, "kk=%i, actgrp=%i, ng=%i, n=%i\n", *kk, actgrp, ng, 2*n);
    fclose(f);
*/
    mean[*kk]  = mg;
    var[*kk]   = dg;
    nelem[*kk] = ng;

    if ((dg<=vmaxgrp) || (ng<nmaxgrp))
    { // if y_var = 0 then break
      SetNaN(&vec[*kk]);
      dv = new double[1];
      SetNaN(dv);
      val->append (1, dv, 0, NULL);
      endpoint[*kk] = (*kk)+1;
      (*kk)++;
      return 0;
    }

    // check for memory
    memchk = new double[8*n];
    if (memchk==NULL)
    { // if y_var = 0 then break
/*    f = fopen ("cart.log", "a");
    fprintf (f, "Memcheck failed!\n");
    fclose(f);
*/
      dv = new double[1];
      SetNaN(dv);
      val->append (1, dv, 0, NULL);
      endpoint[*kk] = (*kk)+1;
      (*kk)++;
      XAwarnt (NOMEM, "cartsplit: incomplete tree");
      return 0;
    }
    delete[] memchk;

    ind    = new long[n];
    indmin = new long[n];
    xind   = new long[n];
    ymean  = new double[n];
    nyl    = new long[n];
    indy   = new long[n];
    index  = new double[n];
    indyy  = new long[n];

/*    f = fopen ("cart.log", "a");
    fprintf (f, "New: %p %p %p %p %p %p\n", ind, indmin, xind, ymean, nyl, indy);
    fclose(f);
*/    dmin = 1.0e38;
    jmin = -1;
    imin = -1;
    SetNaN (&vmin);

    for (k=j=0;j<p; j++)
    { if (*(t+(j%p))) // continuous variable
      { for (i=0; i<n; i++)
	       ind[i] = i;

	     lessvec = (x+j*n);
        qsort (ind, n, sizeof(long), lessind);

	     for (i=0; i<n-1; i++)
	     { if (grp[ind[i]]==actgrp)
	       { // check if same value can be found again !!
            for (l=i+1; l<n; l++)
	         { if ((grp[ind[l]]==actgrp) && (lessvec[ind[i]]==lessvec[ind[l]]))
                i = l;
            }
            if (i==(n-1)) break;

            // now start computing
            ml = mr = 0.0;
	         nl = nr = 0;
	         for (l=0; l<n; l++)
	         { if (grp[ind[l]]==actgrp)
	           { if (l<=i)
		          { ml += *(y+ind[l]);
                  nl++;
                }
	             else
	             { mr += *(y+ind[l]);
                  nr++;
                }
              }
            }
            mr /= nr;
            ml /= nl;
            dl = dr = 0.0;
            for (l=0; l<n; l++)
	         { if (grp[ind[l]]==actgrp)
	           { if (l<=i)
	               dl += (*(y+ind[l])-ml) * (*(y+ind[l])-ml);
	             else
	               dr += (*(y+ind[l])-mr) * (*(y+ind[l])-mr);
              }
            }
            if (dl+dr<dmin)
	         { dmin = dl+dr;
              imin = i;
              jmin = j;
              vmin = *(x+j*n+ind[i]);
              nrmin = nr;
              nlmin = nl;
              drmin = dr;
              dlmin = dl;
              mrmin = mr;
              mlmin = ml;
            }
          }
        }
      }
      else // categorial variable
      { // printf ("cartsplit for categorial variables is not completely yet implemented!\n");
        // step 1: compute mean value for each realization
        for (i=0; i<n; i++)
  	     { ind[i]   = i;
          ymean[i] = 0;
          nyl[i]   = 0;
          xind[i]  = -1;
        }

        lessvec = (x+j*n);
        qsort (ind, n, sizeof(long), lessind);
        /* for (m=0; m<n; m++)
          printf ("o%5i %10g %5i\n",  m, *(x+j*n+m), ind[m]);
*/
        nyl[0] = 0;
        m = 0;
        for (l=i=0; i<n; i++)
	     { if (grp[ind[i]]==actgrp)
	       { if (m && (lessvec[ind[i]]!=lessvec[ind[m-1]]))
            { *(ymean+l) /= nyl[l];
              l++;
              m=i+1;
            }
            if (m==0) m=i+1;
            xind[ind[i]] = l;
            *(ymean+l) += *(y+ind[i]);
            nyl[l]++;
          }
        }
        *(ymean+l) /= nyl[l];
        l++;
/*
    f = fopen ("cart.log", "a");
    fprintf (f, "groups=%i\n", l);
         for (m=0; m<l; m++)
          fprintf (f, "y%1i %10g %5i\n", m, ymean[m], nyl[m]);
//        for (m=0; m<n; m++)
//          if (xind[m]!=-1) fprintf (f, "o%5i %10g %5i\n",  m, *(x+j*n+m), xind[m]);



   fclose(f);
*/
        if (l>1)  // more than one realization found
        {
          // step 2: sort y
          for (i=0; i<l; i++)
            indy[i] = i;
          lessvec = ymean;
          qsort (indy, l, sizeof(long), lessind);
          for (i=0; i<l; i++)
          { indyy[i] = i;
            index[i] = indy[i];
          }
          lessvec = index;
          qsort (indyy, l, sizeof(long), lessind);

 /*
    f = fopen ("cart.log", "a");
    fprintf (f, "groups: %i\n", l);
          for (m=0; m<l; m++)
            printf ("s%1i %10g %5i %5i\n", m, ymean[m], indy[m], nyl[m]);
fclose(f);
*/
          // step 3: compute partitions

	       for (i=0; i<l-1; i++)
	       { // if (grp[i]==actgrp)
            // SK 08.05.99
            /*for (k=i+1; k<l; k++)
	           if (ymean[indy[i]]!=ymean[indy[k]]) break;
            i = k-1;
            if (i==(l-1)) break;
              */
	         { ml = mr = 0.0;
	           nl = nr = 0;
	           for (m=0; m<l; m++)
	           { if (m<=i)
	             { ml += *(ymean+indy[m]) * nyl[indy[m]];
                  nl += nyl[indy[m]];
                }
	             else
	             { mr += *(ymean+indy[m]) * nyl[indy[m]];
                  nr += nyl[indy[m]];
                }
              }
              mr /= nr;
              ml /= nl;
              dl = dr = 0.0;
              for (m=0; m<n; m++)
              { if (grp[m]==actgrp)
	             { if (indyy[xind[m]]<=i)
	               { dl += (y[m]-ml) * (y[m]-ml);
		            }
                  else
	               { dr += (y[m]-mr) * (y[m]-mr);
                  }
                }
              }
//              printf ("split%2i %10g %10g %10g %10g\n", i, ml, dl, mr, dr);
              if (dl+dr<dmin)
	           { dmin = dl+dr;
                imin = i;
                jmin = j;
                vmin = *(ymean+indy[i]);
                nrmin = nr;
                nlmin = nl;
                drmin = dr;
                dlmin = dl;
                mrmin = mr;
                mlmin = ml;


/*    f = fopen ("cart.log", "a");
    fprintf (f, "ng=%i nl=%i, nr=%i\n", ng, nl, nr);
    fclose(f);
*/
/*if (*kk>400)
{                // SK 08.05.99
    f = fopen ("cart.log", "a");
    fprintf (f, "groups: %i %i %f\n", l, i, vmin);
          for (m=0; m<l; m++)
            fprintf (f, "y%i %5i %5i %10f %5i\n", m, indy[m], indyy[m], ymean[indy[m]], nyl[indy[m]]);
    fclose(f);
}
*/
                for (nl=nr=m=0; m<n; m++)
	             { if (grp[m]==actgrp)
                  { if (indyy[xind[m]]<=i)
//                    if (ymean[xind[m]] <= vmin)
                   // if (indy[xind[m]] <= i)
                    { indmin[m] =  0;
                        nl++;
                    }
                    else
                    { indmin[m] = 1;
                      nr++;
                    }
/*
if (*kk>400)
{    f = fopen ("cart.log", "a");
   fprintf (f, "x%i %10f %5i %5i %10f %5i %10f\n", m, *(x+jmin*n+m), ind[m], indmin[m], y[m], xind[m], ymean[xind[m]]);
            fclose(f);
}
*/
                  }
                }

/*    f = fopen ("cart.log", "a");
    fprintf (f, "nlmin=%i, nl=%i, nrmin=%i, nr=%i\n", nlmin, nl, nrmin, nr);
fclose(f);
*/
              }
            }
          }
        }
/*        for (i=0; i<n; i++)
	{ printf ("o%5i: %10g %10g %5i %5i %5i\n", i, *(x+j*n+i), *(y+i), grp[i], xind[i], indmin[i]);
        }
*/      }
    }

//    printf ("split %i %i\n", jmin, imin);

    if ((imin>-1) && (jmin>-1) && (nlmin>=nmingrp) && (nrmin>=nmingrp))
    { vec[*kk] = jmin+1;
//      val[*kk] = vmin;
      (*kk)++;

      nextgrpb = *nextgrp;
      if (*(t+(jmin%p))) // continuous variable
      {
        dv = new double[1];
        dv[0] = vmin;
        val->append (1, dv, 0, NULL);

        for (i=0; i<n; i++)
          ind[i] = i;
        lessvec = (x+jmin*n);
        qsort (ind, n, sizeof(long), lessind);

        for (l=0; l<n; l++)
        { if (grp[ind[l]]==actgrp)
	       { if (l<=imin)
              grp[ind[l]] = nextgrpb;
            else
              grp[ind[l]] = nextgrpb+1;
          }
        }
      }
      else // categorial variable
      { dv = new double[n];
        dvr = new double[n];
        for (l=j=i=0; i<n; i++)
        { if (grp[i]==actgrp)
	       { grp[i] = nextgrpb+indmin[i];
            if (indmin[i]==0)
            { dv[j] = *(x+jmin*n+i);
              for (k=j-1; k>-1; k--)
                if (dv[j]==dv[k]) break;
//              printf ("categorial: %f %i\n", dv[j], k);
              if (k<0) j++;
	         }
            else
            { dvr[l] = *(x+jmin*n+i);
              for (k=l-1; k>-1; k--)
                if (dvr[l]==dvr[k]) break;
//              printf ("categorial: %f %i\n", dvr[j], l);
              if (k<0) l++;
	         }
          }
        }
        if (j==0) j++;
        if (l==0) l++;
        val->append (j, dv, l, dvr);
/*        if (j)
	  val->append (j, dv);
        else
        { SetNaN(dv);
          val->append (1, dv);
        }
*/    }

      (*nextgrp) += 2;
      m = (*kk)-1;
/*
    f = fopen ("cart.log", "a");
        for (m=0; m<n; m++)
          fprintf (f, "o%5i %5i\n",  m, grp[m]);
   fclose(f);
*/
      split (x, n, p, y, t, grp, nmaxgrp, vmaxgrp, nextgrpb, vec, val, nextgrp, kk, mlmin, dlmin, nlmin, mean, var, nelem, nmingrp, endpoint);
      split (x, n, p, y, t, grp, nmaxgrp, vmaxgrp, nextgrpb+1, vec, val, nextgrp, kk, mrmin, drmin, nrmin, mean, var, nelem, nmingrp, endpoint);
      endpoint[m] = (*kk);
    }
    else
    { SetNaN(&vec[*kk]);
      dv = new double[1];
      SetNaN(dv);
      val->append (1, dv, 0, NULL);
      // SetNaN(&val[*kk]);
      endpoint[*kk] = (*kk)+1;
      (*kk)++;
    }

/*    f = fopen ("cart.log", "a");
    fprintf (f, "Del: %p %p %p %p %p %p\n", ind, indmin, xind, ymean, nyl, indy);
    fclose(f);
*/
    delete[] indmin;
    delete[] xind;
    delete[] ymean;
    delete[] nyl;
    delete[] ind;
    delete[] indy;
    delete[] indyy;
    delete[] index;

    return (1);
  }

  double select(long k, long n, double *arr) // PC 20010221
  { 
     unsigned long i,ir,j,l,mid;
     double a, temp;

     l=0;
     ir=n-1;
     for(;;) {
          if(ir<=l+1) { 	         	                //active partition contains 1 or 2 elements
               if((ir == (l+1))&&(arr[ir] < arr[l])) {	   //case of 2 elements
                    SWAP(arr[l], arr[ir], temp)
               } //if
               return(arr[k-1]);
          }
          else {
               mid = (l+ir) >> 1;				//choose median of left, center and right element
               SWAP(arr[mid], arr[l+1], temp)		//as partitioning element a
               if (arr[l] > arr[ir]) {
                    SWAP(arr[l], arr[ir], temp)
               }
               if (arr[l+1] > arr[ir]) {
                    SWAP(arr[l+1], arr[ir], temp)
               }
               if (arr[l] > arr[l+1]) {
                    SWAP(arr[l], arr[l+1], temp)
               }
               i=l+1;
               j=ir;
               a=arr[l+1];                    //the partitioning element
               for(;;) {
                    do i++; while (arr[i] < a);
                    do j--; while (arr[j] > a);
                    if (j==i) j--;							//if arr[i]==a
                    if (j<i) break;						//pointers crossed - partitioning complete
                    SWAP(arr[i], arr[j], temp)
               } //for
               arr[l+1]=arr[j];
               arr[j]=a;
               if(j>=k) ir=j-1;	   		 	//keep active the partition that
               if(i<=k) l=i;				 		//contains the k-th element
          } //else
     } //for(;;)
  }  //select


// template!


  template<class T> XplReal<T>::XplReal (const XplIndex newdim) : XplNumber<T>(newdim) 
  { }

  template<class T> XplReal<T>::XplReal ()
  { ptr = NULL; }

  template<class T> XplReal<T>::XplReal (long d0, long d1, long d2, long d3, long d4, long d5, long d6, long d7) : XplNumber<T>(d0, d1, d2, d3, d4, d5, d6, d7) 
  { }

  template<class T> XplReal<T>::XplReal (XplNumber<T> x) : XplNumber<T>(x) 
  { }

  template<class T> XplReal<T>::XplReal (XplArray<T> x) : XplNumber<T>(x) 
  { }

/*  template<class T> XplReal<T>::XplReal (const XplReal<T> &x) 
  { // assignment(x);
    dim = x.dim;
#ifndef MEMDEBUG
    ptr = new T[x.size()];
#else
    ptr = (T *) xmem.newptr (sizeof(T), x.size(), "XplReal", "XplReal&", "ptr");
#endif  
    memcpy (ptr, x.ptr, x.size()*sizeof(T));
  }  
*/
  template<class T> XplReal<T>::XplReal(const XplReal<T> *x)  //: XplNumber<T>(*x) 
  { dim = x->dim;
#ifndef MEMDEBUG
    ptr = new T[x->size()];
#else
    ptr = (T *) xmem.newptr (sizeof(T), x->size(), "XplReal",
    "XplReal*", "ptr");
#endif
    memcpy (ptr, x->ptr, x->size()*sizeof(T));
  }

  template<class T> XplReal<T>::~XplReal () { };

  template<class T> XplReal<T>& XplReal<T>::operator= (const XplReal<T> &x) 
  { assignment (x);
    return (*this);
  }

  template<class T> XplReal<T>& XplReal<T>::operator= (const XplNumber<T> &x)
  { assignment (x);
    return (*this);
  }

  template<class T> XplReal<T>& XplReal<T>::operator= (const XplArray<T> &x) 
  { assignment (x);
    return (*this);
  }
/*
  template<class T> XplReal<T> XplReal<T>::acos (void)     
  { return (singlefunc (*this, ::acos)); }

  template<class T> XplReal<T> XplReal<T>::acosh (void)    
  { return (singlefunc (*this, ::acosh)); }

  template<class T> XplReal<T> XplReal<T>::x4asin (void)
  { return (singlefunc (*this, ::asin)); }

  template<class T> XplReal<T> XplReal<T>::asinh (void)    
  { return (singlefunc (*this, ::asinh)); }

  template<class T> XplReal<T> XplReal<T>::atan (void)     
  { return (singlefunc (*this, ::atan)); }

  template<class T> XplReal<T> XplReal<T>::atanh (void)    
  { return (singlefunc (*this, ::atanh)); }

  template<class T> XplReal<T> XplReal<T>::ceil (void)     
  { return (singlefunc (*this, ::ceil)); }

  template<class T> XplReal<T> XplReal<T>::cos (void)      
  { return (singlefunc (*this, ::cos)); }

  template<class T> XplReal<T> XplReal<T>::x4cosh (void)
  { return (singlefunc (*this, ::x4cosh)); }

  template<class T> XplReal<T> XplReal<T>::erf (void)      
  { return (singlefunc (*this, ::erf)); }

  template<class T> XplReal<T> XplReal<T>::erfc (void)     
  { return (singlefunc (*this, ::erfc)); }

  template<class T> XplReal<T> XplReal<T>::exp (void)
  { return (singlefunc (*this, ::x4exp)); }

  template<class T> XplReal<T> XplReal<T>::expm1 (void)    
  { return (singlefunc (*this, ::expm1)); }

  template<class T> XplReal<T> XplReal<T>::fabs (void)     
  { return (singlefunc (*this, ::fabs)); }

  template<class T> XplReal<T> XplReal<T>::floor (void)    
  { return (singlefunc (*this, ::floor)); }

  template<class T> XplReal<T> XplReal<T>::gamma (void)    
  { return (singlefunc (*this, ::gamma));  }

  template<class T> XplReal<T> XplReal<T>::lgamma (void)   
#if (defined(__BORLANDC__) || defined(_LINUX_))
  { return (singlefunc (*this, ::gammal)); };
#else
  { return (singlefunc (*this, ::lgamma)); }
#endif

  template<class T> XplReal<T> XplReal<T>::log (void)      
  { return (singlefunc (*this, ::log)); }

  template<class T> XplReal<T> XplReal<T>::log10 (void)    
  { return (singlefunc (*this, ::log10)); }

  template<class T> XplReal<T> XplReal<T>::log1p (void)    
  { return (singlefunc (*this, ::log1p)); }

  template<class T> XplReal<T> XplReal<T>::logb (void)     
  { return (singlefunc (*this, ::logb)); }

  template<class T> XplReal<T> XplReal<T>::rint (void)     
  { return (singlefunc (*this, ::rint)); }

  template<class T> XplReal<T> XplReal<T>::sin (void)      
  { return (singlefunc (*this, ::sin)); }

  template<class T> XplReal<T> XplReal<T>::sinh (void)     
  { return (singlefunc (*this, ::sinh)); }

  template<class T> XplReal<T> XplReal<T>::sqrt (void)     
  { return (singlefunc (*this, ::sqrt)); }

  template<class T> XplReal<T> XplReal<T>::tan (void)      
  { return (singlefunc (*this, ::tan)); }

  template<class T> XplReal<T> XplReal<T>::x4tanh (void)
  { return (singlefunc (*this, ::x4tanh)); }
*/
//double  copysign(double,double);
//double  frexp(double, int*);
//double  hypot(double,double);
//double  infnan(int);
//int     isinf(double);
//int     isnan(double);
//double  j0(double);
//double  j1(double);
//double  jn(int, double);
//double  ldexp(double, int);
//double  modf(double, double*);
//double  pow(double, double);
//double  y0(double);
//double  y1(double);
//double  yn(int, double);
//double aint(double);
//double anint(double);
//int irint(double);
//int nint(double);


  template<class T> XplReal<T> XplReal<T>::inv (const XplReal<T> &x)
  { *this = x; layernpnp (matinv); return (*this); }

  template<class T> void XplReal<T>::lu (XplReal<T> &l, XplReal<T> &u, XplNumber<long> &ix) const      
  { XplReal<T> lumat = *this;
    XplIndex ixdim; long *ixptr;
    l = *this;
    u = *this;
    ixdim = dim;
    ixdim.setdim (1,1);
    ix.setdim(ixdim);
#ifndef MEMDEBUG
    ixptr = new long[ixdim.prod()];
#else
    ixptr = (long *) xmem.newptr (sizeof(long), ixdim.prod(), "XplReal", "lu", "ixptr");
#endif  
    ix.setptr(ixptr);
    lumat.layernpnp2 (matlu, l, u, ix); 
  }

  template<class T> XplArray<T> XplReal<T>::det (void) const     
  { XplReal<T> lumat = *this;
    return(lumat.layernpnp1 (matdet));
  }

  template<class T> XplReal<T>& XplReal<T>::randomize (int no, T init)
  { int i = (int) -::fabs(init);
    long l = (long) -::fabs(init);
    switch (no)
    { case 1 : initrandom (init); break;
      case 2 : ran0(&i); break;
      case 3 : ran1(&i); break;
      case 4 : ran2(&l); break;
      case 5 : ran3(&i); break;
      case 6 : initrandom (init); break;
    }
    return (*this);
  }
   
  template<class T> XplReal<T>& XplReal<T>::random (int no)
  { T (*func) (void);
    long i, maxelem;
    switch (no)
    { case  1 : func = uniformrandom; break;
      case  2 : func = ran0; break;
      case  3 : func = ran1; break;
      case  4 : func = ran2; break;
      case  5 : func = ran3; break;
      case  6 : func = normalrandom; break;
      case 11 : func = expdev; break;
      case 12 : func = expdev0; break;
      case 13 : func = expdev1; break;
      case 14 : func = expdev2; break;
      case 15 : func = expdev3; break;
      case 21 : func = gasdev; break;
      case 22 : func = gasdev0; break;
      case 23 : func = gasdev1; break;
      case 24 : func = gasdev2; break;
      case 25 : func = gasdev3; break;
    }
    maxelem = dim.prod();
    for (i=0; i<maxelem; i++)
      *(ptr+i) = func();
    return (*this);
  }
  
  template<class T> XplReal<T> XplReal<T>::fft (void) const     
  { XplReal<T> ret (dim);
    long il, s;
    T *retptr, *data;
    long n, mmax, m, j, istep, i, nn, k;
    short isign;
    T wtemp, wr, wpr, wpi, wi, theta, tempr, tempi;

    
    if (dim(1)!=2) { XAerrort(DIMCON, "fft : cols(x) <> 2"); return (ret); }
    if (notpowof2(dim(0))) { XAerrort(DIMCON, "fft : rows(x) not power of 2"); return (ret); }
    nn = dim(0);
    s = size() / (2*nn);
    
    retptr = ret.getptr();
    memcpy(retptr, ptr, size()*sizeof(T));
    isign = 1;
  
    for (il=0; il<s; il++)
    { data = retptr+2*il*nn;
      
// From XploRe 3.2
      n = nn << 1;
      j = 1;
      for (i = 1; i < n; i += 2)
      { if (j > i) 
        { SWAP2(DATAXR(j), DATAXR(i))
          SWAP2(DATAXR(j + 1), DATAXR(i + 1))
        }
        m = n >> 1;
        while (m >= 2 && j > m)
        { j -= m;
          m >>= 1;
        }
        j += m;

      }
      mmax = 2;
      while (n > mmax)
      { istep = 2 * mmax;
        theta = 6.28318530717959 / (isign * mmax);
        wtemp = ::sin(0.5 * theta);
        wpr = -2.0 * wtemp * wtemp;
        wpi = ::sin(theta);
        wr = 1.0;
        wi = 0.0;
        for (m = 1; m < mmax; m += 2)
        { for (i = m; i <= n; i += istep)
          { j = i + mmax;
            tempr = wr * DATAXR(j) - wi * DATAXR(j + 1);
            tempi = wr * DATAXR(j + 1) + wi * DATAXR(j);
            DATAXR(j) = DATAXR(i) - tempr;
            DATAXR(j + 1) = DATAXR(i + 1) - tempi;
            DATAXR(i) += tempr;
            DATAXR(i + 1) += tempi;
          }
          wtemp = wr;
          wr = wr * wpr - wi * wpi + wr;
          wi = wi * wpr + wtemp * wpi + wi;
        }
        mmax = istep;
      }

      if (isign < 0)
        for (i = 1; i <= n; i++)
          DATAXR(i) /= nn;
    }
    return (ret);
  }

  template<class T> XplReal<T> XplReal<T>::invfft (void) const     
  { XplReal<T> ret (dim);
    long il, s;
    T *retptr, *data;
    long n, mmax, m, j, istep, i, nn, k;
    short isign;
    T wtemp, wr, wpr, wpi, wi, theta, tempr, tempi;

    
    if (dim(1)!=2) { XAerrort(DIMCON, "invfft : cols(x) <> 2"); return (ret); }
    if (notpowof2(dim(0))) { XAerrort(DIMCON, "invfft : rows(x) not power of 2"); return (ret); }
    nn = dim(0);
    s = size() / (2*nn);
    
    retptr = ret.getptr();
    memcpy(retptr, ptr, size()*sizeof(T));
    isign = -1;
  
    for (il=0; il<s; il++)
    { data = retptr+2*il*nn;
      
// From XploRe 3.2
      n = nn << 1;
      j = 1;
      for (i = 1; i < n; i += 2)
      { if (j > i) 
        { SWAP2(DATAXR(j), DATAXR(i))
          SWAP2(DATAXR(j + 1), DATAXR(i + 1))
        }
        m = n >> 1;
        while (m >= 2 && j > m)
        { j -= m;
          m >>= 1;
        }
        j += m;

      }
      mmax = 2;
      while (n > mmax)
      { istep = 2 * mmax;
        theta = 6.28318530717959 / (isign * mmax);
        wtemp = ::sin(0.5 * theta);
        wpr = -2.0 * wtemp * wtemp;
        wpi = ::sin(theta);
        wr = 1.0;
        wi = 0.0;
        for (m = 1; m < mmax; m += 2)
        { for (i = m; i <= n; i += istep)
          { j = i + mmax;
            tempr = wr * DATAXR(j) - wi * DATAXR(j + 1);
            tempi = wr * DATAXR(j + 1) + wi * DATAXR(j);
            DATAXR(j) = DATAXR(i) - tempr;
            DATAXR(j + 1) = DATAXR(i + 1) - tempi;
            DATAXR(i) += tempr;
            DATAXR(i + 1) += tempi;
          }
          wtemp = wr;
          wr = wr * wpr - wi * wpi + wr;
          wi = wi * wpr + wtemp * wpi + wi;
        }
        mmax = istep;
      }

      if (isign < 0)
        for (i = 1; i <= n; i++)
          DATAXR(i) /= nn;
    }
    return (ret);
  }

  template<class T> void XplReal<T>::eigsm (XplNumber<T> *xb, XplNumber<T> *yb) const
  { long i, j, iq, ip, n, loop, loopi;
    T tresh, theta, tau, t, sm, s, h, g, c, *b, *z, *xbptr, *ybptr, *a, *v, *d;
    XplIndex xbd, ybd;

    if (dim(0)!=dim(1)) { XAerrort (DIMCON, "eigsm : cols(x) <> rows(x)"); return; }
    n = dim(0);

    xbd = ybd = dim;
    ybd.setdim (1, 1);

#ifndef MEMDEBUG    
    b = new T[n];
    z = new T[n];
    xbptr = new T[xbd.prod()];
    ybptr = new T[ybd.prod()];
#else
    b = (T *) xmem.newptr (sizeof(T), n, "XplReal", "eigsm", "b");
    z = (T *) xmem.newptr (sizeof(T), n, "XplReal", "eigsm", "z");
    xbptr = (T *) xmem.newptr (sizeof(T), xbd.prod(), "XplReal", "eigsm", "xbptr");
    ybptr = (T *) xmem.newptr (sizeof(T), ybd.prod(), "XplReal", "eigsm", "ybptr");
#endif  

    loop = size()/(n*n); 

    for (loopi=0; loopi<loop; loopi++)
    { a = (ptr+loopi*n*n);
      v = (xbptr+loopi*n*n);
      d = (ybptr+loopi*n);

// From XploRe 3.2
      for (ip = 0; ip < n; ip++)
      { for (iq = 0; iq < n; iq++)
          *(v + n * iq + ip) = 0.0;
        *(v + n * ip + ip) = 1.0;
      }
      for (ip = 0; ip < n; ip++)
      { d[ip] = b[ip] = *(a + n * ip + ip);
        z[ip] = 0.0;
      }
      for (i = 0; i < MAXJACLOOP; i++)
      { sm = 0.0;
        for (ip = 0; ip < (n - 1); ip++)
          for (iq = (ip + 1); iq < n; iq++)
            sm += ::fabs(*(a + n * iq + ip));
        if (isZero(sm)) break;
        if (i < 3)
          tresh = 0.2 * sm / ::pow(n, 2);
        else
          tresh = 0.0;
        for (ip = 0; ip < (n - 1); ip++)
        { for (iq = (ip + 1); iq < n; iq++)
          { g = 100.0 * ::fabs(*(a + n * iq + ip));
            if ((i > 3) && ((::fabs(d[ip]) + g) == ::fabs(d[ip]))
                        && ((::fabs(d[iq]) + g) == ::fabs(d[iq])))
              *(a + n * iq + ip) = 0.0;
            else if (::fabs(*(a + n * iq + ip)) > tresh)
            { h = d[iq] - d[ip];
              if ((::fabs(h) + g) == ::fabs(h))
                t = *(a + n * iq + ip) / h;
              else
              { theta = 0.5 * h / *(a + n * iq + ip);
                t = 1.0 / (::fabs(theta) + ::sqrt(1.0 + ::pow(theta, 2)));
                if (theta < 0.0)
                  t = -t;
              }
              c = 1.0 / ::sqrt(1.0 + ::pow(t, 2));
              s = t * c;
              tau = s / (1.0 + c);
              h = t * (*(a + n * iq + ip));
              z[ip] -= h;
              z[iq] += h;
              d[ip] -= h;
              d[iq] += h;
              *(a + n * iq + ip) = 0.0;
              for (j = 0; j <= (ip - 1); j++)
              { g = *(a + n * ip + j);
                h = *(a + n * iq + j);
                *(a + n * ip + j) = g - s * (h + g * tau);
                *(a + n * iq + j) = h + s * (g - h * tau);
              }
              for (j = (ip + 1); j <= (iq - 1); j++)
              { g = *(a + n * j + ip);
                h = *(a + n * iq + j);
                *(a + n * j + ip) = g - s * (h + g * tau);
                *(a + n * iq + j) = h + s * (g - h * tau);
              }
              for (j = (iq + 1); j < n; j++)
              { g = *(a + n * j + ip);
                h = *(a + n * j + iq);
                *(a + n * j + ip) = g - s * (h + g * tau);
                *(a + n * j + iq) = h + s * (g - h * tau);
              }
              for (j = 0; j < n; j++)
              { g = *(v + n * ip + j);
                h = *(v + n * iq + j);
                *(v + n * ip + j) = g - s * (h + g * tau);
                *(v + n * iq + j) = h + s * (g - h * tau);
              }
            }
          }
        }
      }
      if (i==MAXJACLOOP)
      { for (i = 0; i < n; i++)
        { b[i] += z[i];
          d[i] = b[i];
          z[i] = 0.0;
        }
        XAerror (NOCONV);
      }
    }
#ifndef MEMDEBUG  
    delete[] b;
    delete[] z;
#else
    xmem.delptr (b);
    xmem.delptr (z);
    b = NULL;
    z = NULL;
#endif
    xb->setdim(xbd);
    xb->setptr(xbptr);
    yb->setdim(ybd);
    yb->setptr(ybptr);
  }

  // PC 20010222
  template<class T> XplReal<T> XplReal<T>::mseq (long n, const XplReal<T> *d) const
  { XplReal<T> ret(1);
    XplIndex retbd;
    long i;
    T *retptr, *step;

// n is the length:
    if (n < 1) { XAerrort (ERR_VAL, "mseq : n has to be a positive integer"); return(ret); }

#ifndef MEMDEBUG
    retptr = new T[n];
#else
    retptr = (T *) xmem.newptr (sizeof(T), n, "XplReal", "mseq", "retptr");
#endif

    step = d->getptr();
    *retptr = *ptr;
    for (i=1; i<n; i++)
      *(retptr+i) = *(retptr + i - 1) * *step;

    retbd = dim;
    retbd.setdim (0, n);

    ret.setdim(retbd);
    ret.setptr(retptr);

    return (ret);
  }     //mseq

  // PC 20010222
  template<class T> XplReal<T> XplReal<T>::aseq (long n, const XplReal<T> *d) const
  { XplReal<T> ret(1);
    XplIndex retbd;
    long i;
    T *retptr, *step;

// n is the length:
    if (n < 1)
    {
      XAerrort (ERR_VAL, "aseq : n has to be a positive integer");
      return(ret);
    }

#ifndef MEMDEBUG
    retptr = new T[n];
#else
    retptr = (T *) xmem.newptr (sizeof(T), n, "XplReal", "aseq", "retptr");
#endif

    step = d->getptr();
    *retptr = *ptr;
    for (i=1; i<n; i++)
      *(retptr+i) = (*step * i) + *(retptr);


    retbd = dim;
    retbd.setdim (0, n);

    ret.setdim(retbd);
    ret.setptr(retptr);

    return (ret);
  }             //aseq

  // PC 20010221
  template<class T> XplReal<T> XplReal<T>::median (char f)
  { XplReal<T> ret(1);
    XplIndex retbd, tmp;
    char even, nbr;
    long i, s, n, mi;
    T *xp, *retptr, *retp;

    n    = dim(0);							//# of rows
    mi   = (n+1) >> 1;						//the middle index
    even = ((n & 1) == 0);             //is n even?

    if (n == 1) return(this);

    if(f && (n < 100)) f = 0;				//faster median but small sample

    retbd = dim;
    retbd.setdim(0, 1);
    s = retbd.prod();                  //# of elements in the input matrix
#ifndef MEMDEBUG
    retptr   = new T[s];
#else
    retptr   = (T *) xmem.newptr (sizeof(T), s, "XplReal", "median", "retptr");
#endif
    for (i=0; i<s; i++)
      *(retptr+i) = 0;

    tmp  = 0;
    do {
      xp   = ptr+dim.value(tmp);				//set up the indices
      retp  = retptr+retbd.value(tmp);

      nbr = 1;										//was not breaked yet
      for(i=0;i<n;i++)   						//any NaN returns NaN
          if(XisNaN(*(xp+i))) {
               *retp = *(xp+i);
               nbr = 0;                        //NaN found
               break;
          }

      if (nbr) {
          *retp = select(mi,n,xp);
          if (even && (!f)) {					  	//n is even, counting exact
               *retp += select(mi+1,n,xp);
               *retp /= 2;
          }
      }
    } while (tmp.inc(dim, 0));

    ret.setdim(retbd);
    ret.setptr(retptr);

    return (ret);
  }	//median

  // PC 20010221
  template<class T> XplReal<T> XplReal<T>::medianNaN (char f)
  { XplReal<T> ret(1);
    XplIndex retbd, tmp;
    char even;
    double temp;                       //for the macro SWAP
    long i, s, l, n, mi;
    T *xp, *retptr, *retp;

    n    = dim(0);							//# of rows

    retbd = dim;
    retbd.setdim(0, 1);
    s = retbd.prod();                  //# of elements in the input matrix
#ifndef MEMDEBUG
    retptr   = new T[s];
#else
    retptr   = (T *) xmem.newptr (sizeof(T), s, "XplReal", "medianNaN", "retptr");
#endif
    for (i=0; i<s; i++)
      *(retptr+i) = 0;

    tmp  = 0;
    do {
      xp   = ptr+dim.value(tmp);						  //set up the indeces
      retp  = retptr+retbd.value(tmp);

      i=0; l=n-1;
      while((XisNaN(*(xp+l)))&&(l>i)) { l--; }       //set up l to the last not NaN
      while(i<l) {							           //move NaN's to the end
        if(XisNaN(*(xp+i))) {
          SWAP(*(xp+i),*(xp+l), temp);
          while((XisNaN(*(xp+l)))&&(l>i)) { l--; }   //set up l to the last not NaN
        }
        i++;
      }

      l++;
      if (l == 1)
        *retp = *xp;
      else {
        mi   = (l+1) >> 1;									//the middle index
        even = ((l & 1) == 0);  		               //is l even?
        if(f && (l < 100)) f = 0;						//faster median but small sample
        *retp = select(mi,l,xp);
        if (even && (!f)) {								//l is even, counting exact
          *retp += select(mi+1,l,xp);
          *retp /= 2;
        }
      }
    } while (tmp.inc(dim, 0));

    ret.setdim(retbd);
    ret.setptr(retptr);

    return (ret);
  }	//medianNaN

  template<class T> XplReal<T> XplReal<T>::mean (long d, const XplNumber<long> *cond) const
  { XplReal<T> ret(1);
    XplIndex retbd, tmp, tmp2, condbd;
    long i, s, ci, ca, *condptr, jmp, j, *condp, *obs, *obsp;
    T *xp, *retptr, *retp;  
 
    if ((d<0) || (d>=MAX_XARRAY))
    { XAerrort (RANGE, "mean : dim out of range"); return(ret); }
    if (cond)
    { condbd = cond->dim;
      for (i=0; i<MAX_XARRAY; i++)
        if (dim(i)!=condbd(i))
        { XAerrort(DIMCON, "mean : layer(x)<>layer(cond)"); return(ret); }
      s = (cond->dim).prod();
      condptr = cond->getptr();
      ci = ca = *condptr;
      for (i=1; i<s; i++)
      { if (*(condptr+i) > ca) ca = *(condptr+i);
        if (*(condptr+i) < ci) ci = *(condptr+i);
      }
    }
    else
    { condptr = NULL;
      ci = ca = 0;
    }

    if (ci<0)
    { XAerrort(RANGE, "mean : min(cond) < 0"); return(ret); }
    
    retbd = dim;
    retbd.setdim (d, 1);
    if (MAXLONG / retbd.prod() < ca+1)
    { XAerrort (RANGE, "mean : result too large"); return (ret); }
    retbd.setdim(d, ca+1);
    s        = retbd.prod();
#ifndef MEMDEBUG
    retptr   = new T[s];
    obs      = new long[s];
#else
    retptr   = (T *) xmem.newptr (sizeof(T), s, "XplReal", "mean", "retptr");
    obs      = (long *) xmem.newptr (sizeof(long), s, "XplReal", "mean", "obs");
#endif  

    for (i=0; i<s; i++)
    { *(retptr+i) = 0;
      *(obs+i)    = 0;
    }

    jmp = 1;
    for (i=0; i<d; i++)
      jmp *= dim(i);

    tmp  = 0;

    if (condptr)
    { do
      { xp    = ptr+dim.value(tmp);
        retp  = retptr+retbd.value(tmp);
        obsp  = obs+retbd.value(tmp);
        condp = condptr+condbd.value(tmp);

        for (j=0; j<dim(d); j++, xp += jmp, condp += jmp)
        { *(retp+jmp* *condp) += *xp;
          (*(obsp+jmp* *condp))++;
        }
      }
      while (tmp.inc(dim, d));

      for (i=0; i<s; i++)
        *(retptr+i) /= *(obs+i);
    }
    else
    { do
      { xp    = ptr+dim.value(tmp);
        retp  = retptr+retbd.value(tmp);
        for (j=0; j<dim(d); j++, xp += jmp)
          *retp += *xp;
      }
      while (tmp.inc(dim, d));

      for (i=0; i<s; i++)
        *(retptr+i) /= dim(d);
    }
#ifndef MEMDEBUG
    delete []obs;
#else
    xmem.delptr (obs);
    obs = NULL;
#endif

    ret.setdim(retbd);
    ret.setptr(retptr);

    return (ret);
  }

  // PC 20010125
  template<class T> XplReal<T> XplReal<T>::meannan (long d) const
  { XplReal<T> ret(1);
    XplIndex retbd, tmp, tmp2;
    long i, s, jmp, j, len;
    T *xp, *retptr, *retp;  
 
    if ((d<0) || (d>=MAX_XARRAY))
    { XAerrort (RANGE, "meannan : dim out of range"); return(ret); }
    
    retbd = dim;
    retbd.setdim (d, 1);
    if (MAXLONG / retbd.prod() < 1)
    { XAerrort (RANGE, "meannan : result too large"); return (ret); }
    s        = retbd.prod();
#ifndef MEMDEBUG
    retptr   = new T[s];
#else
    retptr   = (T *) xmem.newptr (sizeof(T), s, "XplReal", "meannan", "retptr");
#endif  

    for (i=0; i<s; i++)
      *(retptr+i) = 0;

    jmp = 1;
    for (i=0; i<d; i++)
      jmp *= dim(i);

    tmp  = 0;

    do
    { 
        xp    = ptr+dim.value(tmp);
        retp  = retptr+retbd.value(tmp);
        len = 0;
        
        for (j=0; j<dim(d); j++, xp += jmp)
          if (!isNaN(*xp))
          {
               *retp += *xp;
               len++;
          }
        
        if (len > 0)
          *retp /= len;
        else 
          SetNaN(retp); 
    }
    while (tmp.inc(dim, d));

    ret.setdim(retbd);
    ret.setptr(retptr);

    return (ret);
  }

  template<class T> XplReal<T> XplReal<T>::var (long d, const XplNumber<long> *cond) const
  { XplReal<T> ret(1);
    XplIndex retbd, tmp, tmp2, condbd;
    long i, s, ci, ca, *condptr, jmp, j, *condp, *obs, *obsp;
    T *xp, *retptr, *retp, *mp, *m, tmpd;  
 
    if ((d<0) || (d>=MAX_XARRAY))
    { XAerrort (RANGE, "var : dim out of range"); return (ret); }
    if (cond)
    { condbd = cond->dim;
      for (i=0; i<MAX_XARRAY; i++)
        if (dim(i)!=condbd(i))
        { XAerrort(DIMCON, "var : layer(x)<>layer(cond)"); return (ret); }
      s = (cond->dim).prod();
      condptr = cond->getptr();
      ci = ca = *condptr;
      for (i=1; i<s; i++)
      { if (*(condptr+i) > ca) ca = *(condptr+i);
        if (*(condptr+i) < ci) ci = *(condptr+i);
      }
    }
    else
    { condptr = NULL;
      ci = ca = 0;
    }

    if (ci<0)
    { XAerrort(RANGE, "var : min(cond) < 0"); return (ret); }
    
    retbd = dim;
    retbd.setdim(d, 1);
    if (MAXLONG / retbd.prod() < ca+1)
    { XAerrort (RANGE, "var : result too large"); return (ret); }
    retbd.setdim(d, ca+1);
    s        = retbd.prod();
#ifndef MEMDEBUG
    retptr   = new T[s];
    obs      = new long[s];
    m        = new T[s];
#else
    retptr   = (T *) xmem.newptr (sizeof(T), s, "XplReal", "var", "retptr");
    obs      = (long *) xmem.newptr (sizeof(long), s, "XplReal", "var", "obs");
    m        = (T *) xmem.newptr (sizeof(T), s, "XplReal", "var", "m");
#endif  
    
    for (i=0; i<s; i++)
    { *(retptr+i) = 0;
      *(obs+i)    = 0;
      *(m+i)      = 0;
    }

    jmp = 1;
    for (i=0; i<d; i++)
      jmp *= dim(i);

    tmp  = 0;

    if (condptr)
    { do
      { xp    = ptr+dim.value(tmp);
        mp    = m+retbd.value(tmp);
        obsp  = obs+retbd.value(tmp);
        condp = condptr+condbd.value(tmp);

        for (j=0; j<dim(d); j++, xp += jmp, condp += jmp)
        { *(mp+jmp* *condp) += *xp;
          (*(obsp+jmp* *condp))++;
        }
      }
      while (tmp.inc(dim, d));

      for (i=0; i<s; i++)
        *(m+i) /= *(obs+i);

      tmp = 0;
      do
      { xp    = ptr+dim.value(tmp);
        retp  = retptr+retbd.value(tmp);
        mp    = m+retbd.value(tmp);
        obsp  = obs+retbd.value(tmp);
        condp = condptr+condbd.value(tmp);

        for (j=0; j<dim(d); j++, xp += jmp, condp += jmp)
        { tmpd = *xp-*(mp+jmp* *condp);
          *(retp+jmp* *condp) += tmpd*tmpd;
        }
      }
      while (tmp.inc(dim, d));

      for (i=0; i<s; i++)
        *(retptr+i) /= (*(obs+i)-1);

    }
    else
    { do
      { xp   = ptr+dim.value(tmp);
        tmpd = 0;

        for (j=0; j<dim(d); j++, xp += jmp)
          tmpd += *xp;
        tmpd /= dim(d);
       
        xp   = ptr+dim.value(tmp);
        retp  = retptr+retbd.value(tmp);
        for (j=0; j<dim(d); j++, xp += jmp)
          *retp += (*xp - tmpd) * (*xp - tmpd);

      }
      while (tmp.inc(dim, d));

      for (i=0; i<s; i++)
        *(retptr+i) /= (dim(d)-1);
    }
#ifndef MEMDEBUG
    delete []obs;
    delete []m;
#else
    xmem.delptr (obs);
    xmem.delptr (m);
    obs = NULL;
    m = NULL;
#endif

    ret.setdim(retbd);
    ret.setptr(retptr);

    return (ret);
  }

  // PC 20010125
  template<class T> XplReal<T> XplReal<T>::varnan (long d) const
  { XplReal<T> ret(1);
    XplIndex retbd, tmp, tmp2, condbd;
    long i, s, jmp, j, len;
    T *xp, *retptr, *retp, tmpd;  
 
    if ((d<0) || (d>=MAX_XARRAY))
    { XAerrort (RANGE, "varnan : dim out of range"); return (ret); }
    
    retbd = dim;
    retbd.setdim(d, 1);
    if (MAXLONG / retbd.prod() < 1)
    { XAerrort (RANGE, "varnan : result too large"); return (ret); }
    s        = retbd.prod();
#ifndef MEMDEBUG
    retptr   = new T[s];
#else
    retptr   = (T *) xmem.newptr (sizeof(T), s, "XplReal", "varnan", "retptr");
#endif  
    
    for (i=0; i<s; i++)
      *(retptr+i) = 0;

    jmp = 1;
    for (i=0; i<d; i++)
      jmp *= dim(i);

    tmp  = 0;

    do
    { 
        xp   = ptr+dim.value(tmp);
        tmpd = 0;
        len = 0;

        //mean
        for (j=0; j<dim(d); j++, xp += jmp)
          if (!isNaN(*xp))
          {
               tmpd += *xp;
               len++;
          }
        if (len > 0)
          tmpd /= len;
        else
          tmpd = 0;
       
        // var
        xp   = ptr+dim.value(tmp);
        retp  = retptr+retbd.value(tmp);
        for (j=0; j<dim(d); j++, xp += jmp)
           if (!isNaN(*xp))
              *retp += (*xp - tmpd) * (*xp - tmpd);

        if (len > 1)
          *retp /= len - 1;
        else
          SetNaN(retp);
    }
    while (tmp.inc(dim, d));

    ret.setdim(retbd);
    ret.setptr(retptr);

    return (ret);
  }

  // PC 20010124: covariance
  // PC 20010205: added parameter condl
  template<class T> XplReal<T> XplReal<T>::cov (const XplNumber<long> *cond) const
  { XplReal<T> ret(1);
    XplIndex retbd;                 //indeX of the returned value
    XplIndex condbd;
    XplIndex coord;                 //coordinates in the covariance matrix
    XplIndex ind1, ind2;            //2 tmp indeces
    long i, j, tmp;                 //3 tmp's
    long nt, n = dim(0); 			   //# of obs in vector having zeros in cond, length of the first column
    long *condptr, *condp, *condp2;
    long s;                         //total # of elements in ret
    T *xp, *xp2;                    //vectors X1,X2 for E(X1-EX1)'(X2-EX2)
    T *retptr, *retp;               //returned object; tmp object
    T ev, ev2, summ;                //expected values EX1, EX2; the sum of (X1-EX1)'(X2-EX2)

    if (cond)
    { condbd = cond->dim;
      for (i=0; i<MAX_XARRAY; i++)
        if (dim(i)!=condbd(i))
        { XAerrort(DIMCON, "cov : layer(x)<>layer(cond)"); return (ret); }
      condptr = cond->getptr();
    }  //if(cond)
    else { condptr = NULL;}

    retbd = dim;
    retbd.setdim(0, dim(1));
    s        = retbd.prod();
#ifndef MEMDEBUG
    retptr   = new T[s];
#else
    retptr   = (T *) xmem.newptr (sizeof(T), s, "XplReal", "cov", "retptr");
#endif

    for (i=0; i<s; i++)
    { *(retptr+i) = 0;
    }

    ind1  = 0;
    if (condptr)
    { do
      { ind2 = ind1;
        do
        {
          // mean
          xp  = ptr+dim.value(ind1);
          xp2 = ptr+dim.value(ind2);
          condp = condptr+condbd.value(ind1);
          condp2 = condptr+condbd.value(ind2);
          ev = 0;
          ev2 = 0;
          nt = 0;
          for (j=0; j<n; j++, xp++, condp++, xp2++, condp2++)
            if((*condp==0)&&(*condp2==0))
            { ev += *xp;
              ev2 += *xp2;
              nt++;
            }
          ev /= nt;		//case of nt=0 is solved above the part "//save it"
          ev2 /= nt;

          // sum of (X1-EX1)'(X2-EX2)
          xp  = ptr+dim.value(ind1);
          xp2 = ptr+dim.value(ind2);
          condp = condptr+condbd.value(ind1);
          condp2 = condptr+condbd.value(ind2);
          summ = 0;
          for (j=0; j<n; j++, xp++, xp2++, condp++, condp2++)
            if((*condp==0)&&(*condp2==0))
              summ += (*xp - ev) * (*xp2 - ev2);
          summ /= nt-1;

          if(nt==0) summ /= nt;			//no observations available
          // save it
          coord = ind2;
          coord.setdim(0, ind1(1));

          retp  = retptr + retbd.value(coord);
          *retp = summ;

          // the symmetric part
          coord.setdim(0, ind2(1));
          coord.setdim(1, ind1(1));

          retp  = retptr+retbd.value(coord);
          *retp = summ;
        } while ( (ind2.inc(dim, 0)) ? ind1.equal(ind2, 1) : 0 );
      } while (ind1.inc(dim, 0));
    }		//if(condptr)
    else
    { do
      { ind2 = ind1;

        // mean
        xp = ptr+dim.value(ind1);
        ev = 0;
        for (j=0; j<n; j++, xp++)
          ev += *xp;
        ev /= n;

        do
        {
          // 2nd mean
          xp2 = ptr+dim.value(ind2);
          ev2 = 0;
          for (j=0; j<n; j++, xp2++)
            ev2 += *xp2;
          ev2 /= n;

          // sum of (X1-EX1)'(X2-EX2)
          xp  = ptr+dim.value(ind1);
          xp2 = ptr+dim.value(ind2);
          summ = 0;
          for (j=0; j<n; j++, xp++, xp2++)
            summ += (*xp - ev) * (*xp2 - ev2);
          summ /= n-1;

          // save it
          coord = ind2;
          coord.setdim(0, ind1(1));

          retp  = retptr + retbd.value(coord);
          *retp = summ;

          // the symmetric part
          coord.setdim(0, ind2(1));
          coord.setdim(1, ind1(1));

          retp  = retptr+retbd.value(coord);
          *retp = summ;
        } while ( (ind2.inc(dim, 0)) ? ind1.equal(ind2, 1) : 0 );
      } while (ind1.inc(dim, 0));
    }		//else

    ret.setdim(retbd);
    ret.setptr(retptr);
    return (ret);
  }     //cov

  // PC 20010125
  template<class T> XplReal<T> XplReal<T>::covnan () const
  { XplReal<T> ret(1);               //returned value
    XplIndex retbd;                      //indeX of the returned value
    XplIndex coord;                      //coordinates in the covariance matrix
    XplIndex ind1, ind2;                 //2 tmp indeces
    long i, j, tmp, n = dim(0);          //3 tmp's, length of the first column
    long s, len;                            //total # of elements in ret
    T *xp, *xp2;                            //vectors X1,X2 for E(X1-EX1)'(X2-EX2)
    T *retptr, *retp;                    //returned object; tmp object
    T ev, ev2, summ;                //expected values EX1, EX2; the sum of (X1-EX1)'(X2-EX2)

    retbd = dim;
    retbd.setdim(0, dim(1));
    s     = retbd.prod();

#ifndef MEMDEBUG
    retptr   = new T[s];
#else
    retptr   = (T *) xmem.newptr (sizeof(T), s, "XplReal", "cov", "retptr");
#endif

    for (i=0; i<s; i++)
      *(retptr+i) = 0;

    ind1  = 0;
    do
    { 
      ind2 = ind1;

      do
      { 
        // means
        xp = ptr+dim.value(ind1);
        ev = 0;
        xp2 = ptr+dim.value(ind2);
        ev2 = 0;
        len = 0;
        
        for (j=0; j<n; j++, xp++, xp2++)
             if (!isNaN(*xp) && !isNaN(*xp2))
             {      
               ev += *xp;
               ev2 += *xp2;
               len++;
             }
        
        if (len > 0)
        {
          ev /= len;  ev2 /= len;
        }
        else 
        {
          ev = 0; ev2 = 0;
        }

        // sum of (X1-EX1)'(X2-EX2)
        xp  = ptr+dim.value(ind1);
        xp2 = ptr+dim.value(ind2);
        summ = 0;
        for (j=0; j<n; j++, xp++, xp2++)  
          if (!isNaN(*xp) && !isNaN(*xp2))
             summ += (*xp - ev) * (*xp2 - ev2);

        if (len > 1)
          summ /= len-1;
        else
          SetNaN(&summ);

        // save it
        coord = ind2;
        coord.setdim(0, ind1(1));
        
        retp  = retptr + retbd.value(coord);
        *retp = summ;

        // the symmetric part
        coord.setdim(0, ind2(1));
        coord.setdim(1, ind1(1));

        retp  = retptr+retbd.value(coord);
        *retp = summ;
      }
      while ( (ind2.inc(dim, 0)) ? ind1.equal(ind2, 1) : 0 );
    }
    while (ind1.inc(dim, 0));

    ret.setdim(retbd);
    ret.setptr(retptr);

    return (ret);
  } // cov

  template<class T> XplReal<T> XplReal<T>::sker (const XplReal<T>& bandwidth,
                     const char *kernel,
                     const XplReal<T>& weight,
                     XplReal<T> *xest) const
  { XplIndex rbd;
    T *xesto, *res, *h, *y, *xptr, *bptr, *hptr, *xestptr, *wptr,
      *xo, cor, *rptr, maxh, pdiffmaxh;
    void *id;
    long dimen, len, xestlen, i, j, k, wdim, l, start, translated = 0, 
         m, hlen;
    XplIndex tmp, tmp2, dbd, xdim, bdim, hdim, xestdim, wdimg, loop;
    XplReal<T> *pdiff, *pkern;
    double (*intkern) (double) = NULL;
    char cmd[256];

    if (dimension(1)!=bandwidth.dimension(1))
      XAerrort(DIMCON, "sker : cols(x)<>cols(bandwidth)");
    if (weight.dimension(0)!=dimension(0))
      XAerrort(DIMCON, "sker : rows(weight)<>rows(x)");
    if (xest)
    { if ((bandwidth.dimension(0)!=1) && 
          (bandwidth.dimension(0)!=xest->dimension(0)))
        XAerrort(DIMCON, "sker : rows(bandwidth) not conformable");
      if (dimension(1)!=xest->dimension(1))
        XAerrort(DIMCON, "sker : cols(x)<>cols(xest)");
      rbd.setdim (0, xest->dimension(0));
    }
    else
    { if ((bandwidth.dimension(0)!=1) && 
          (bandwidth.dimension(0)!=dimension(0)))
        XAerrort(DIMCON, "sker : rows(bandwidth) not conformable");
      rbd.setdim(0, dimension(0));
    }
    rbd.setdim(1, weight.dimension(1)); 
    hlen   = bandwidth.dimension(0);

    for (i=2; i<MAX_XARRAY; i++)
    { rbd.setdim (i, dimension(i));
      if (bandwidth.dimension(i)>1)
      { if (rbd(i)==1)
          rbd.setdim (i, bandwidth.dimension(i));
        else
          XAerrort(DIMCON, "sker: layer(bandwidth)<>1"); 
      }

      if (weight.dimension(i)>1)
      { if (rbd(i)==1)
          rbd.setdim (i, weight.dimension(i));
        else
          XAerrort(DIMCON, "sker: layer(weight)<>1"); 
      }
    }
    
    if (strcmp(kernel, "uni")==0)   intkern = suniform;
    if (strcmp(kernel, "trian")==0) intkern = striangle;
    if (strcmp(kernel, "epa")==0)   intkern = sepanechnikov;
    if (strcmp(kernel, "qua")==0)   intkern = squartic;
    if (strcmp(kernel, "tri")==0)   intkern = striweight;
   
    XplReal<T> ret(rbd), tmpkern(1);
    dimen = dimension(1);
    len   = dimension(0);
    wdim  = weight.dimension(1);
    if (xest)
      xestlen = xest->dimension(0);
    else
      xestlen = len;
    
    dbd = 1;
    dbd.setdim (0, len); 
    dbd.setdim (1, dimen); 
    XplReal<T> diff(dbd), kern(1);
    dbd.setdim(1, 2);
    XplNumber<long> index(dbd);

    tmp = tmp2 = 0;
    loop = rbd;
    loop.setdim(0, 1);
    loop.setdim(1, 1);

    xptr = getptr();
    xdim = getdim();
    bptr = bandwidth.getptr();
    bdim = bandwidth.getdim();
    wptr = weight.getptr();
    wdimg = weight.getdim();
    if (xest)
    { xestptr = xest->getptr();
      xestdim = xest->getdim();
    }
    else
    { xestptr = getptr();
      xestdim = getdim();
    }
    rptr = ret.getptr();
    for (i=0; i<ret.size(); i++)
      rptr[i] = (T) 0.0;

    pdiff = &diff;
    pkern = &kern;
    id = StartExecution();
    GenerateXplObject ("x", (void *) pdiff, id, TRUE);
    GenerateXplObject ("y", (void *) pkern, id, TRUE);

    do
    { tmp2=tmp<<2;
      tmp2=tmp2%xdim; xo = xptr+xdim.value(tmp2); 

      tmp2=tmp<<2;
      tmp2=tmp2%bdim; h = bptr+bdim.value(tmp2); 

      tmp2=tmp<<2;
      tmp2=tmp2%wdimg; y = wptr+wdimg.value(tmp2); 

      if (xest)
      { tmp2=tmp<<2;
        tmp2=tmp2%xestdim; xesto = xestptr+xestdim.value(tmp2); 
      }
      else
        xesto = xo;

      tmp2=tmp<<2;
      tmp2=tmp2%rbd; res = rptr+rbd.value(tmp2); 


// here comes XploRe 3.2 
// void kernelest (double *res, double *res2, double *xo, double *h, double *y, 
//                int dim, int len, double *xest, int xestlen)

      start = 0;
      l     = 0;
      if (intkern==NULL)
        pdiff = (XplReal<T> *) GetXplObject ("x", id);        
      maxh = *h; // Supp in X4 for local bandwidth
      for (i=1; i<hlen; i++)  
      { if (maxh < *(h+i))
          maxh = *(h+i);
      }

      for (i=0; i<xestlen; i++)
      { for (j=start; j<len; j++)
        { (*pdiff)(l,0) = ((*(xo+j) - *(xesto+i)) / *(h+i%hlen));
          pdiffmaxh = (*(xo+j) - *(xesto+i)) / maxh;
          if (pdiffmaxh < -1.0)
            start++;
          else if (pdiffmaxh > 1.0)
            break;
          else
          { for (k=1; k<dimen; k++)
              (*pdiff)(l,k) = (*(xo+k*len+j) - *(xesto+k*xestlen+i)) /
                               *(h+hlen*k+i%hlen);
            index(l,0) = i;
            index(l,1) = j;
            l++;
            if (l==len) // matrix is filled
            { 
              if (intkern==NULL) 
              { sprintf (cmd, "y = %s (x)", kernel);
                if (Execute (cmd, id, translated))
                  XAerrort (FUNCCALL, cmd);
                 pkern = (XplReal<T> *) GetXplObject ("y", id);
              }
              else
              { kern = singlefunc(diff, intkern);
                tmpkern = kern.mul(1);
                pkern = &tmpkern;
              }
              if (pkern->dimension(0)!=l)
                XAerrort (DIMCON, "sker : rows(kern) is wrong");
              for (k=1; k<MAX_XARRAY; k++)
                if (pkern->dimension(k)!=1)
                  XAerrort (DIMCON, "sker : layer(kern) is wrong");
              for (k=0; k<len; k++)
              { for (m=0; m<wdim; m++)
                  *(res+m*xestlen+index(k,0)) += *(y+m*len+index(k,1)) * (*pkern)(k);
              }
              l = 0;
            }
          }
        }
      }
        
      if (l) // if any rest left
      { 
        if (intkern==NULL)
        { sprintf (cmd, "y = %s (x)", kernel);
          if (Execute (cmd, id, translated))
            XAerrort (FUNCCALL, cmd);
          pkern = (XplReal<T> *) GetXplObject ("y", id);
        }
        else
        { kern = singlefunc(diff, intkern);
          tmpkern = kern.mul(1);
          pkern = &tmpkern;
        }
//        if (pkern->dimension(0)!=l)
//          XAerrort (DIMCON, "sker : rows(kern) is wrong");
        for (k=1; k<MAX_XARRAY; k++)
          if (pkern->dimension(k)!=1)
            XAerrort (DIMCON, "sker : layer(kern) is wrong");
        for (k=0; k<l; k++)
        { for (m=0; m<wdim; m++)
            *(res+m*xestlen+index(k,0)) += *(y+m*len+index(k,1)) * (*pkern)(k);

        }
      }
        
      for (k=0; k<xestlen; k++)
      { cor = len;
        for (j=0; j<dimen; j++)
          cor *= *(h+j*hlen+k%hlen);
        for (j=0; j<wdim; j++)
         *(res+j*xestlen+k) /= cor;
      }

    }
    while (tmp.inc(loop));
    EndExecution (id);

    return (ret);
  }

// SK 02.09.97
/*  template<class T> XplReal<T> XplReal<T>::distance (const char *dist = "L2")
  { XplReal<T> ret;
    ret = layernpnpdist(matdist, dist);
    return (ret);
  }
*/
  template<class T> XplReal<T> XplReal<T>::pdfb (const XplReal<T> &a, const XplReal<T> &b) const
  { long i;

    if ((a.dim(0)!=dim(0)) && (a.dim(0)!=1))
      XAerrort(DIMCON, "pdfb: rows(a) == rows(x) or 1");
    if ((a.dim(1)!=dim(1)) && (a.dim(1)!=1))
      XAerrort(DIMCON, "pdfb: rows(a) == rows(x) or 1");

    if ((b.dim(0)!=dim(0)) && (b.dim(0)!=1))
      XAerrort(DIMCON, "pdfb: rows(b) == rows(x) or 1");
    if ((b.dim(1)!=dim(1)) && (b.dim(1)!=1))
      XAerrort(DIMCON, "pdfb: rows(b) == rows(x) or 1");

    for (i=2; i<MAX_XARRAY; i++)
    { if (dim(i)!=a.dim(i)) 
        XAerrort(DIMCON, "pdfb: layer(x) != layer(a)");
      if (dim(i)!=b.dim(i)) 
        XAerrort(DIMCON, "pdfb: layer(x) != layer(b)");
    }    

    return (layernpnp3 (matpdfb, a, b));

  }

  template<class T> XplReal<T> XplReal<T>::pdfc (const XplReal<T> &a) const
  { long i;

    if ((a.dim(0)!=dim(0)) && (a.dim(0)!=1))
      XAerrort(DIMCON, "pdfc: rows(a) == rows(x) or 1");
    if ((a.dim(1)!=dim(1)) && (a.dim(1)!=1))
      XAerrort(DIMCON, "pdfc: rows(a) == rows(x) or 1");

    for (i=2; i<MAX_XARRAY; i++)
    { if (dim(i)!=a.dim(i)) 
        XAerrort(DIMCON, "pdfc: layer(x) != layer(a)");
    }    

    return (layernpnp3 (matpdfc, a, a));

  }

  template<class T> XplReal<T> XplReal<T>::pdff (const XplReal<T> &a, const XplReal<T> &b) const
  { long i;

    if ((a.dim(0)!=dim(0)) && (a.dim(0)!=1))
      XAerrort(DIMCON, "pdff: rows(a) == rows(x) or 1");
    if ((a.dim(1)!=dim(1)) && (a.dim(1)!=1))
      XAerrort(DIMCON, "pdff: rows(a) == rows(x) or 1");

    if ((b.dim(0)!=dim(0)) && (b.dim(0)!=1))
      XAerrort(DIMCON, "pdff: rows(b) == rows(x) or 1");
    if ((b.dim(1)!=dim(1)) && (b.dim(1)!=1))
      XAerrort(DIMCON, "pdff: rows(b) == rows(x) or 1");

    for (i=2; i<MAX_XARRAY; i++)
    { if (dim(i)!=a.dim(i)) 
        XAerrort(DIMCON, "pdff: layer(x) != layer(a)");
      if (dim(i)!=b.dim(i)) 
        XAerrort(DIMCON, "pdff: layer(x) != layer(b)");
    }    

    return (layernpnp3 (matpdff, a, b));
  }

  template<class T> XplReal<T> XplReal<T>::pdfn (void) const
  { long i;
    XplReal<T> dummy;

    return (layernpnp3 (matpdfn, dummy, dummy));
  }

  template<class T> XplReal<T> XplReal<T>::pdft (const XplReal<T> &a) const
  { long i;

    if ((a.dim(0)!=dim(0)) && (a.dim(0)!=1))
      XAerrort(DIMCON, "pdft: rows(a) == rows(x) or 1");
    if ((a.dim(1)!=dim(1)) && (a.dim(1)!=1))
      XAerrort(DIMCON, "pdft: rows(a) == rows(x) or 1");

    for (i=2; i<MAX_XARRAY; i++)
    { if (dim(i)!=a.dim(i)) 
        XAerrort(DIMCON, "pdft: layer(x) != layer(a)");
    }    

    return (layernpnp3 (matpdft, a, a));
  }

  template<class T> XplReal<T> XplReal<T>::cdfn (void) const
  { long i;
    XplReal<T> dummy(1);

    return (layernpnp3 (matcdfn, dummy, dummy));
  }

  template<class T> XplReal<T> XplReal<T>::cdft (const XplReal<T> &a) const
  { long i;

    if ((a.dim(0)!=dim(0)) && (a.dim(0)!=1))
      XAerrort(DIMCON, "cdft: rows(a) == rows(x) or 1");
    if ((a.dim(1)!=dim(1)) && (a.dim(1)!=1))
      XAerrort(DIMCON, "cdft: rows(a) == rows(x) or 1");

    for (i=2; i<MAX_XARRAY; i++)
    { if (dim(i)!=a.dim(i)) 
        XAerrort(DIMCON, "cdft: layer(x) != layer(a)");
    }    

    return (layernpnp3 (matcdft, a, a));
  }

  template<class T> XplReal<T> XplReal<T>::cdfc (const XplReal<T> &a) const
  { long i;

    if ((a.dim(0)!=dim(0)) && (a.dim(0)!=1))
      XAerrort(DIMCON, "cdfc: rows(a) == rows(x) or 1");
    if ((a.dim(1)!=dim(1)) && (a.dim(1)!=1))
      XAerrort(DIMCON, "cdfc: rows(a) == rows(x) or 1");

    for (i=2; i<MAX_XARRAY; i++)
    { if (dim(i)!=a.dim(i)) 
        XAerrort(DIMCON, "cdfc: layer(x) != layer(a)");
    }    

    return (layernpnp3 (matcdfc, a, a));
  }

  template<class T> XplReal<T> XplReal<T>::cdfb (const XplReal<T> &a, const XplReal<T> &b) const
  { long i;

    if ((a.dim(0)!=dim(0)) && (a.dim(0)!=1))
      XAerrort(DIMCON, "cdfb: rows(a) == rows(x) or 1");
    if ((a.dim(1)!=dim(1)) && (a.dim(1)!=1))
      XAerrort(DIMCON, "cdfb: rows(a) == rows(x) or 1");

    if ((b.dim(0)!=dim(0)) && (b.dim(0)!=1))
      XAerrort(DIMCON, "cdfb: rows(b) == rows(x) or 1");
    if ((b.dim(1)!=dim(1)) && (b.dim(1)!=1))
      XAerrort(DIMCON, "cdfb: rows(b) == rows(x) or 1");

    for (i=2; i<MAX_XARRAY; i++)
    { if (dim(i)!=a.dim(i)) 
        XAerrort(DIMCON, "cdfb: layer(x) != layer(a)");
      if (dim(i)!=b.dim(i)) 
        XAerrort(DIMCON, "cdfb: layer(x) != layer(b)");
    }    

    return (layernpnp3 (matcdfb, a, b));
  }

  template<class T> XplReal<T> XplReal<T>::cdff (const XplReal<T> &a, const XplReal<T> &b) const
  { long i;

    if ((a.dim(0)!=dim(0)) && (a.dim(0)!=1))
      XAerrort(DIMCON, "cdff: rows(a) == rows(x) or 1");
    if ((a.dim(1)!=dim(1)) && (a.dim(1)!=1))
      XAerrort(DIMCON, "cdff: rows(a) == rows(x) or 1");

    if ((b.dim(0)!=dim(0)) && (b.dim(0)!=1))
      XAerrort(DIMCON, "cdff: rows(b) == rows(x) or 1");
    if ((b.dim(1)!=dim(1)) && (b.dim(1)!=1))
      XAerrort(DIMCON, "cdff: rows(b) == rows(x) or 1");

    for (i=2; i<MAX_XARRAY; i++)
    { if (dim(i)!=a.dim(i)) 
        XAerrort(DIMCON, "cdff: layer(x) != layer(a)");
      if (dim(i)!=b.dim(i)) 
        XAerrort(DIMCON, "cdff: layer(x) != layer(b)");
    }    

    return (layernpnp3 (matcdff, a, b));
  }

  template<class T> XplReal<T> XplReal<T>::qfn (void) const
  { long i;
    XplReal<T> dummy(1);

    return (layernpnp3 (matqfn, dummy, dummy));
  }

  template<class T> XplReal<T> XplReal<T>::qft (const XplReal<T> &a) const
  { long i;

    if ((a.dim(0)!=dim(0)) && (a.dim(0)!=1))
      XAerrort(DIMCON, "qft: rows(a) == rows(x) or 1");
    if ((a.dim(1)!=dim(1)) && (a.dim(1)!=1))
      XAerrort(DIMCON, "qft: rows(a) == rows(x) or 1");

    for (i=2; i<MAX_XARRAY; i++)
    { if (dim(i)!=a.dim(i)) 
        XAerrort(DIMCON, "qft: layer(x) != layer(a)");
    }    

    return (layernpnp3 (matqft, a, a));
  }

  template<class T> XplReal<T> XplReal<T>::qfc (const XplReal<T> &a) const
  { long i;

    if ((a.dim(0)!=dim(0)) && (a.dim(0)!=1))
      XAerrort(DIMCON, "qfc: rows(a) == rows(x) or 1");
    if ((a.dim(1)!=dim(1)) && (a.dim(1)!=1))
      XAerrort(DIMCON, "qfc: rows(a) == rows(x) or 1");

    for (i=2; i<MAX_XARRAY; i++)
    { if (dim(i)!=a.dim(i)) 
        XAerrort(DIMCON, "qfc: layer(x) != layer(a)");
    }    

    return (layernpnp3 (matqfc, a, a));
  }

  template<class T> XplReal<T> XplReal<T>::qff (const XplReal<T> &a, const XplReal<T> &b) const
  { long i;

    if ((a.dim(0)!=dim(0)) && (a.dim(0)!=1))
      XAerrort(DIMCON, "qff: rows(a) == rows(x) or 1");
    if ((a.dim(1)!=dim(1)) && (a.dim(1)!=1))
      XAerrort(DIMCON, "qff: rows(a) == rows(x) or 1");

    if ((b.dim(0)!=dim(0)) && (b.dim(0)!=1))
      XAerrort(DIMCON, "qff: rows(b) == rows(x) or 1");
    if ((b.dim(1)!=dim(1)) && (b.dim(1)!=1))
      XAerrort(DIMCON, "qff: rows(b) == rows(x) or 1");

    for (i=2; i<MAX_XARRAY; i++)
    { if (dim(i)!=a.dim(i)) 
        XAerrort(DIMCON, "qff: layer(x) != layer(a)");
      if (dim(i)!=b.dim(i)) 
        XAerrort(DIMCON, "qff: layer(x) != layer(b)");
    }    

    return (layernpnp3 (matqff, a, b));
  }

  template<class T> XplReal<T> XplReal<T>::atan2 (const XplReal<T> &a) const
  { long i;

    if ((a.dim(0)!=dim(0)) && (a.dim(0)!=1))
      XAerrort(DIMCON, "atan2: rows(x) == rows(y) or 1");
    if ((a.dim(1)!=dim(1)) && (a.dim(1)!=1))
      XAerrort(DIMCON, "atan2: rows(x) == rows(y) or 1");

    for (i=2; i<MAX_XARRAY; i++)
    { if (dim(i)!=a.dim(i)) 
        XAerrort(DIMCON, "atan2: layer(y) != layer(x)");
    }    

    return (layernpnp3 (matatan2, a, a));
  }

  template<class T> void XplReal<T>::fwt (const XplNumber<long> &level, const XplReal<T> &h, XplReal<T> &a, XplReal<T> &b)
  { long i, j, k, l, n, p, s, ll, hrange, alen, blen;
    T *xptr, *aptr, *bptr, *alptr, *blptr, *hptr;
    XplIndex adim, bdim;

    n = dim(0);
    p = dim(1);
    l = level[0];

    if (notpowof2(n)) { XAerrort (DIMCON, "fwt : rows(x) != 2^n"); return; }
    if (p!=1)         { XAerrort (DIMCON, "fwt : cols(x) != 1"); return; }
    if (notpowof2(l)) { XAerrort (DIMCON, "fwt : level   != 2^n"); return; }
    if (l>=n)         { XAerrort (DIMCON, "fwt : level   >= rows(x)"); return; }
    if (h.dim(1)!=1)  { XAerrort (DIMCON, "fwt : cols(h) != 1"); return; }

    for (i=2; i<MAX_XARRAY; i++)
      if (h.dim(i)!=1) { XAerrort (DIMCON, "fwt : layer(h) != 1"); return; }
      
    adim = bdim = dim;
    adim.setdim(0, l); 
    adim.setdim(1, 2);
    bdim.setdim(0, n-l); 
    bdim.setdim(1, 3);
    alen = adim(0)*adim(1);
    blen = bdim(0)*bdim(1);

    hptr   = h.getptr();
    hrange = h.size();

#ifndef MEMDEBUG
    xptr = new T[n];
    aptr = new T[adim.prod()];
    bptr = new T[bdim.prod()];
#else
    xptr = (T *) xmem.newptr (sizeof(T), n, "XplReal", "fwt", "xptr");
    aptr = (T *) xmem.newptr (sizeof(T), adim.prod(), "XplReal", "fwt", "aptr");
    bptr = (T *) xmem.newptr (sizeof(T), bdim.prod(), "XplReal", "fwt", "bptr");
#endif  

    s = size()/n;
    alptr = aptr; blptr = bptr;
    for (i=0; i<s; i++, alptr += alen, blptr += blen)
    { memcopy (xptr, ptr+i*n, n);
      fwt_int (xptr, n, l, hptr, hrange);
      for (j=0; j<l; j++)
      { *(alptr+j) = j;
        *(alptr+l+j) = *(xptr+j) / ::sqrt(n);
      }
      k  = j;
      ll = 0;
      for (j=0; j<n-l; j++, ll++)
      { if (ll>=k) { k*=2; ll=0; }
        *(blptr+j)     = (int) (::x4log(k)/LN2);
        *(blptr+n-l+j)   = ll;    
        *(blptr+2*(n-l)+j) = *(xptr+l+j) / ::sqrt(n);
      }
    }
#ifndef MEMDEBUG
    delete[] xptr;
#else
    xmem.delptr (xptr);
    xptr = NULL;
#endif
    a.setdim(adim);
    a.setptr(aptr);
    b.setdim(bdim);
    b.setptr(bptr);

  }

  template<class T> void XplReal<T>::invfwt (const XplReal<T> &a, const XplReal<T> &b, const XplNumber<long> &n, const XplNumber<long> &l, const XplReal<T> &h)
  { long i, j, k, ma, sa, la, ll, alen, blen, hrange, level, s, length;
    T *xptr, *xlptr, *aptr, *bptr, *alptr, *blptr, *hptr;
    XplIndex xdim;

    if (n.dim(0)!=1) { XAerrort(DIMCON, "invfwt : n has to a scalar"); return; }
    if (n.dim(1)!=1) { XAerrort(DIMCON, "invfwt : n has to a scalar"); return; }
    if (l.dim(0)!=1) { XAerrort(DIMCON, "invfwt : l has to a scalar"); return; }
    if (l.dim(1)!=1) { XAerrort(DIMCON, "invfwt : l has to a scalar"); return; }
    if (l(0)>=n(0))  { XAerrort(DIMCON, "invfwt : l >= n"); return; }
    if (a.dim(1)!=2) { XAerrort(DIMCON, "invfwt : 2 columns for a expected"); return; }
    if (b.dim(1)!=3) { XAerrort(DIMCON, "invfwt : 3 columns for b expected"); return; }

    length = n[0];
    xdim.setdim (0, length);
    xdim.setdim (1, 1);
    for (i=2; i<MAX_XARRAY; i++)
    { if (n.dim(i)!=1) { XAerrort(DIMCON, "invfwt : n has to be a scalar"); return; }
      if (l.dim(i)!=1) { XAerrort(DIMCON, "invfwt : l has to be a scalar"); return; }
      if (a.dim(i)!=b.dim(i)) { XAerrort (DIMCON, "invfwt : layer(a) == layer(b)"); return; }
      if (h.dim(i)!=1) { XAerrort(DIMCON, "invfwt : h has to be a vector"); return; }
      xdim.setdim(i, a.dim(i));
    }
   
    s = b.size()/(b.dim(0)*b.dim(1));
    alen = a.dim(0);
    blen = b.dim(0);
#ifndef MEMDEBUG
    xlptr = xptr = new T[xdim.prod()];
#else
    xlptr = xptr = (T *) xmem.newptr (sizeof(T), xdim.prod(), "XplReal", "invfwt", "xlptr/xptr");
#endif  

    alptr  = aptr = a.getptr();
    blptr  = bptr = b.getptr();
    hptr   = h.getptr();
    hrange = h.dim(0)*h.dim(1);
    level  = l[0];

    for (i=0; i<s; i++, alptr += (2*alen), blptr += (3*blen), xlptr += length)
    { for (j=0; j<length; j++) *(xlptr+j) = 0.0;
      for (j=0; j<level; j++)
      { for (k=0; k<alen; k++)
          if (((long) *(alptr+k)) == j) 
            *(xlptr+j) = *(alptr+alen+k) * ::sqrt(length);
      }
      ll = level;
      ma = (long) (::x4log(length)/LN2);
      for (la = (long) (::x4log(ll)/LN2); la<ma; la++, ll *=2)
      { for (sa = 0; sa<ll; sa++)
        { for (k=0; k<blen; k++)
          if (((long) (*(blptr+k)) == la) &&
              (((long) *(blptr+blen+k)) == sa))
            *(xlptr+ll+sa) = *(blptr+2*blen+k) * ::sqrt(length);
        }
      }
      invert_fwt (xlptr, length, level, hptr, hrange);
    }

    if (ptr) 
#ifndef MEMDEBUG
      delete[] ptr;
#else
      xmem.delptr (ptr);
      ptr = NULL;
#endif
    dim = xdim;
    ptr = xptr;
  }

  template<class T> XplReal<T> XplReal<T>::ons (const char *base, const XplNumber<long>& order)
  { XplReal<T> ret(1);
    XplIndex retdim;    
    T *retptr, factor;
    long i, n, p, s, k, ord;

    for (i=1; i<MAX_XARRAY; i++)
    { if (dim(i)!=1) XAerrort (DIMCON, "ons : layer(x)<>1");
      if (order.dim(i)!=1) XAerrort (DIMCON, "ons : layer(o)<>1");    
    }

    retdim = dim;
    p = order.dim(0);
    retdim.setdim(1, p);
    n = retdim(0);
    s = retdim.prod();
#ifndef MEMDEBUG
    retptr = new T[s];
#else
    retptr = (T *) xmem.newptr (sizeof(T), s, "XplReal", "ons", "retptr");
#endif  
    
    if (strcasecmp(base, "fourier"))
    { for (k=0; k<p; k++)
      { ord = order(k);
        if (ord<0)
        { for (i=0; i<n; i++) 
            SetNaN (retptr+k*n+i);
        }
        else if (ord==0)
        { for (i=0; i<n; i++) 
            *(retptr+k*n+i) = M_2_PI;
        }
        else if (ord%2)
        { factor = 1+(ord-1)/2;
          for (i=0; i<n; i++) 
            *(retptr+k*n+i) = ::cos (factor* *(ptr+i));          
        }
        else 
        { factor = 1+(ord-1)/2;
          for (i=0; i<n; i++) 
            *(retptr+k*n+i) = ::sin (factor* *(ptr+i));          
        }
      }
    }
    else
    { for (i=0; i<s; i++) *(retptr+i) = 0.0; }

    ret.setdim (retdim);
    ret.setptr (retptr);

    return (ret);
  } 

  template<class T> void XplReal<T>::contour2 (const XplReal<T> &cuta, XplNumber<long> *retl, XplReal<T> *retp)
  { long i, j, k, nx, ny, n, len, oldlen;
    T xl, xr, yl, yr, xm, ym, f1, f2, f3, f4, fm, ratio, *x, *y, *f,
      cut, *lx, *ly;
     
    n = dim(0);
    if (dim(1)!=3)
    { XAerrort (DIMBIG, "contour2 : cols(x)<>3");
      return;
    }
    if ((cuta.dim(0)!=1) || (cuta.dim(1)!=1))
    { XAerrort (DIMBIG, "contour2 : cut scalar expected");
      return;
    }

    for (i=2; i<MAX_XARRAY; i++)
    { if (dim(i)!=1)
      { XAerrort (DIMBIG, "contour2 : layer(x)<>1");
        return;
      }
      if (cuta.dim(i)!=1)
      { XAerrort (DIMBIG, "contour2 : cut scalar expected");
        return;
      }
    }
    cut = cuta();

    x = ptr;
    y = (ptr+n);
    f = (ptr+2*n);
    ny = 1;
    for (i=0; i<=n/2; i++, ny++)
      if (*(x+i)!=*(x+i+1)) break;
    if (n%ny)
    { XAerrort (DIMCON, "contour2 : rows(x)<> #xgrid * #ygrid");
      return;
    }
    nx = n/ny;

    len = 24*(nx-1)*(ny-1);
#ifndef MEMDEBUG
    lx  = new T[len];
    ly  = new T[len];
#else
    lx  = (T *) xmem.newptr (sizeof(T), len, "XplReal", "contour2", "lx");
    ly  = (T *) xmem.newptr (sizeof(T), len, "XplReal", "contour2", "ly");
#endif
    oldlen = 0;
    
    for (i=0; i<nx-1; i++)
    { xl = x[i*ny];
      xr = x[i*ny+ny];
      xm = 0.5*(xl+xr);
      for (j=0; j<ny-1; j++)    
      { yl = y[j];
        yr = y[j+1];
        ym = 0.5*(yr+yl);
        f1 = f[i*ny+j];
        f2 = f[i*ny+ny+j];
        f3 = f[i*ny+j+1];
        f4 = f[i*ny+ny+j+1];
        fm = 0.25*(f1+f2+f3+f4);
        triangle (fm, f1, f2, xm, xl, xr, ym, yl, yl, cut, lx, ly, &oldlen);
        triangle (fm, f2, f4, xm, xr, xr, ym, yl, yr, cut, lx, ly, &oldlen);
        triangle (fm, f4, f3, xm, xr, xl, ym, yr, yr, cut, lx, ly, &oldlen);
        triangle (fm, f3, f1, xm, xl, xl, ym, yr, yl, cut, lx, ly, &oldlen);
      }
    }

    if (oldlen==0) {
#ifndef MEMDEBUG
      delete[] lx;
      delete[] ly;
#else
      xmem.delptr (lx);
      xmem.delptr (ly);
      lx = NULL;
      ly = NULL;
#endif
      XAerrort (NULMEM, XMSG_NULMEM_CONT_STR);
      return;
   }

    long *retlptr;
    T *retpptr;
    XplIndex retdim;

#ifndef MEMDEBUG
    retlptr = new long[oldlen];
    retpptr = new T[2*oldlen];
#else
    retlptr = (long *) xmem.newptr (sizeof(long), oldlen, "XplReal", "contour2", "retlptr");
    retpptr = (T *) xmem.newptr (sizeof(T), 2*oldlen, "XplReal", "contour2", "retpptr");
#endif
    for (i=0; i<oldlen; i++)
    { *(retpptr+i) = *(lx+i);
      *(retpptr+oldlen+i) = *(ly+i);
      if (i<oldlen/2)
      { *(retlptr+i) = 2*i+1;
        *(retlptr+oldlen/2+i) = 2*i+2; 
      }
    } 

    retdim = 1;
    retdim.setdim (0, oldlen/2);
    retdim.setdim (1, 2);
    retl->setptr (retlptr);
    retl->setdim (retdim);

    retdim.setdim (0, oldlen);
    retp->setptr (retpptr);
    retp->setdim (retdim);
#ifndef MEMDEBUG
    delete[] lx;
    delete[] ly;
#else
    xmem.delptr (lx);
    xmem.delptr (ly);
    lx = NULL;
    ly = NULL;
#endif
  }

  template<class T> void XplReal<T>::contour3 (const XplReal<T> &cuta, XplNumber<long> *retl, XplReal<T> *retp)
  { long i, j, k, nx, ny, nz, n, len, oldlen;
    T xl, xr, yl, yr, xm, ym, f1, f2, f3, f4, fm, ratio, *x, *y, *z, *f,
      cut, *lx, *ly, *lz;

    n = dim(0);
    if (dim(1)!=4)
    { XAerrort (DIMBIG, "contour3 : cols(x)<>4");
      return;
    }
    if ((cuta.dim(0)!=1) || (cuta.dim(1)!=1))
    { XAerrort (DIMBIG, "contour3 : cut scalar expected");
      return;
    }

    for (i=2; i<MAX_XARRAY; i++)
    { if (dim(i)!=1)
      { XAerrort (DIMBIG, "contour3 : layer(x)<>1");
        return;
      }
      if (cuta.dim(i)!=1)
      { XAerrort (DIMBIG, "contour3 : cut scalar expected");
        return;
      }
    }
    cut = cuta();

    x = ptr;
    y = (ptr+n);
    z = (ptr+2*n);
    f = (ptr+3*n);
    nz = 1;
    for (i=0; i<=n/2; i++, nz++)
      if (*(z+i)>*(z+i+1)) break;
    if (n%nz)
    { XAerrort (DIMCON, "contour3 : rows(x) mod #zgrid <> 0");
      return;
    }
    ny = 1;
    for (i=0; i<=n/2; i++, ny++)
      if (*(y+i)>*(y+i+1)) break;
    if (ny%nz)
    { XAerrort (DIMCON, "contour3 : rows(x) mod #ygrid <> 0");
      return;
    }
    ny /= nz;
    if (n%(ny*nz))
    { XAerrort (DIMCON, "contour3 : rows(x) mod #xgrid <> 0");
      return;
    }
    nx = n/(ny*nz);

    len = 24*((nx-1)*(ny-1)*nz+(nx-1)*ny*(nz-1)+nx*(ny-1)*(nz-1));
#ifndef MEMDEBUG
    lx  = new T[len];
    ly  = new T[len];
    lz  = new T[len];
#else
    lx  = (T *) xmem.newptr (sizeof(T), len, "XplReal", "contour3", "lx");
    ly  = (T *) xmem.newptr (sizeof(T), len, "XplReal", "contour3", "ly");
    lz  = (T *) xmem.newptr (sizeof(T), len, "XplReal", "contour3", "lz");
#endif
    oldlen = 0;

    for (k=0; k<nz; k++)
    { for (i=oldlen; i<len; i++)
        *(lz+i) = z[k];
      for (i=0; i<nx-1; i++)
      { xl = x[i*ny*nz];
        xr = x[(i+1)*ny*nz];
        xm = 0.5*(xl+xr);
        for (j=0; j<ny-1; j++)
        { yl = y[j*nz];
          yr = y[(j+1)*nz];
          ym = 0.5*(yr+yl);
          f1 = f[i*ny*nz+j*nz+k];
          f2 = f[(i+1)*ny*nz+j*nz+k];
          f3 = f[i*ny*nz+(j+1)*nz+k];
          f4 = f[(i+1)*ny*nz+(j+1)*nz+k];
          fm = 0.25*(f1+f2+f3+f4);
          triangle (fm, f1, f2, xm, xl, xr, ym, yl, yl, cut, lx, ly, &oldlen);
          triangle (fm, f2, f4, xm, xr, xr, ym, yl, yr, cut, lx, ly, &oldlen);
          triangle (fm, f4, f3, xm, xr, xl, ym, yr, yr, cut, lx, ly, &oldlen);
          triangle (fm, f3, f1, xm, xl, xl, ym, yr, yl, cut, lx, ly, &oldlen);
        }
      }
    }

    for (k=0; k<ny; k++)
    { for (i=oldlen; i<len; i++)
        *(ly+i) = y[k*nz];
      for (i=0; i<nx-1; i++)
      { xl = x[i*ny*nz];
        xr = x[(i+1)*ny*nz];
        xm = 0.5*(xl+xr);
        for (j=0; j<nz-1; j++)    
        { yl = z[j];
          yr = z[j+1];
          ym = 0.5*(yr+yl);
          f1 = f[i*ny*nz+k*nz+j];
          f2 = f[(i+1)*ny*nz+k*nz+j];
          f3 = f[i*ny*nz+k*nz+j+1];
          f4 = f[(i+1)*ny*nz+k*nz+j+1];
          fm = 0.25*(f1+f2+f3+f4);
          triangle (fm, f1, f2, xm, xl, xr, ym, yl, yl, cut, lx, lz, &oldlen);
          triangle (fm, f2, f4, xm, xr, xr, ym, yl, yr, cut, lx, lz, &oldlen);
          triangle (fm, f4, f3, xm, xr, xl, ym, yr, yr, cut, lx, lz, &oldlen);
          triangle (fm, f3, f1, xm, xl, xl, ym, yr, yl, cut, lx, lz, &oldlen);
        }
      }
    }

    for (k=0; k<nx; k++)
    { for (i=oldlen; i<len; i++)
        *(lx+i) = x[k*nz*ny];
      for (i=0; i<ny-1; i++)
      { xl = y[i*nz];
        xr = y[(i+1)*nz];
        xm = 0.5*(xl+xr);
        for (j=0; j<nz-1; j++)    
        { yl = z[j];
          yr = z[j+1];
          ym = 0.5*(yr+yl);
          f1 = f[k*ny*nz+i*nz+j];
          f2 = f[k*ny*nz+(i+1)*nz+j];
          f3 = f[k*ny*nz+i*nz+j+1];
          f4 = f[k*ny*nz+(i+1)*nz+j+1];
          fm = 0.25*(f1+f2+f3+f4);
          triangle (fm, f1, f2, xm, xl, xr, ym, yl, yl, cut, ly, lz, &oldlen);
          triangle (fm, f2, f4, xm, xr, xr, ym, yl, yr, cut, ly, lz, &oldlen);
          triangle (fm, f4, f3, xm, xr, xl, ym, yr, yr, cut, ly, lz, &oldlen);
          triangle (fm, f3, f1, xm, xl, xl, ym, yr, yl, cut, ly, lz, &oldlen);
        }
      }
    }

    if (oldlen==0) {
#ifndef MEMDEBUG
      delete[] lx;
      delete[] ly;
      delete[] lz;
#else
      xmem.delptr (lx);
      xmem.delptr (ly);
      xmem.delptr (lz);
      lx = NULL;
      ly = NULL;
      lz = NULL;
#endif
      XAerrort(NULMEM, XMSG_NULMEM_CONT_STR);
      return;
            }

    long *retlptr;
    T *retpptr;
    XplIndex retdim;

#ifndef MEMDEBUG
    retlptr = new long[oldlen];
    retpptr = new T[3*oldlen];
#else
    retlptr = (long *) xmem.newptr (sizeof(long), oldlen, "XplReal", "contour3", "retlptr");
    retpptr = (T *) xmem.newptr (sizeof(T), 3*oldlen, "XplReal", "contour3", "retpptr");
#endif

    for (i=0; i<oldlen; i++)
    { *(retpptr+i) = *(lx+i);
      *(retpptr+oldlen+i) = *(ly+i);
      *(retpptr+2*oldlen+i) = *(lz+i);
      if (i<oldlen/2)
      { *(retlptr+i) = 2*i+1;
        *(retlptr+oldlen/2+i) = 2*i+2; 
      }
    } 

    retdim = 1;
    retdim.setdim (0, oldlen/2);
    retdim.setdim (1, 2);
    retl->setptr (retlptr);
    retl->setdim (retdim);

    retdim.setdim (0, oldlen);
    retdim.setdim (1, 3);
    retp->setptr (retpptr);
    retp->setdim (retdim);
#ifndef MEMDEBUG
    delete[] lx;
    delete[] ly;
    delete[] lz;
#else
    xmem.delptr (lx);
    xmem.delptr (ly);
    xmem.delptr (lz);
    lx = NULL;
    ly = NULL;
    lz = NULL;
#endif
  }

  template<class T> void XplReal<T>::nninit (const XplReal<T> &unit, XplNumber<long> *sortweiarray)
  { long i, j, tmp, exch = 0, lenwei, *sortwei;
    T *wei;
    XplIndex udim;

    udim = unit.getdim();

    if (dim(1)!=3)
    { XAerrort (DIMCON, "nninit : cols(w) <> 3");
      return;
    }
    if (udim(1)!=2)
    { XAerrort (DIMCON, "nninit : cols(u) <> 2");
      return;
    }
    for (i=2; i<MAX_XARRAY; i++)
    { if (udim(i)!=1)
      { XAerrort (DIMCON, "nninit : layer(u) <> 1");
        return;
      }
      if (dim(i)!=1)
      { XAerrort (DIMCON, "nninit : layer(w) <> 1");
        return;
      }
    }
      
    lenwei  = dim(0);
#ifndef MEMDEBUG
    sortwei = new long[lenwei];
#else
    sortwei = (long *) xmem.newptr (sizeof(long), lenwei, "XplReal", "nninit", "sortwei");
#endif
    for (i=0; i<lenwei; i++)
      sortwei[i] = i;

    for (i=0; i<lenwei; i++)
    { if (*(ptr+sortwei[i])!=INUNIT)
      { for (j=i+1; j<lenwei; j++)
        { if (*(ptr+lenwei+sortwei[j])==*(ptr+sortwei[i]))
          { tmp = sortwei[i];
            sortwei[i] = sortwei[j];
            sortwei[j] = tmp;
            i--;
            break;
          }
        }
        if (j==lenwei) 
          exch = 0;
        else
          exch++;
      }
      if (exch > lenwei-i)
      { XAerrort (NN, "No feed forward network");
        return;
      }
    }
    for (i=0; i<lenwei; i++)
      sortwei[i]++;

    udim.setdim (0, lenwei);
    udim.setdim (1, 1);
    sortweiarray->setptr(sortwei);
    sortweiarray->setdim(udim);
    return;
  }

  template<class T> void XplReal<T>::nnfunc (const XplReal<T> &unitarray, const XplReal<T> &xarray, XplReal<T> *yarray)
  { long i, j, k, out = 0, minout, from, to, in, actf, n, p, lenwei, lenunit;
    XplIndex udim;
    T *result, tmp, *y, wft, *wei, *unit, *x;    

    // printf ("Enter nnfunc\n");

    udim   = xarray.getdim();
    n      = udim(0);
    p      = udim(1);
    lenwei = dim(0);
    udim   = unitarray.getdim();
    lenunit = udim(0);

    wei  = ptr;
    unit = unitarray.getptr();
    x    = xarray.getptr();

    if (dim(1)!=3)
    { XAerrort (NN, "nnfunc : cols(w)<>3");
      return;
    }
    if (udim(1)!=2)
    { XAerrort (NN, "nnfunc : cols(u)<>2");
      return;
    }
    for (i=2; i<MAX_XARRAY; i++)
    { if (dim(i)!=1)
      { XAerrort (NN, "nnfunc : layer(w)<>1");
        return;
      }
      if (udim(i)!=1)
      { XAerrort (NN, "nnfunc : layer(u)<>1");
        return;
      }
    }
 
#ifndef MEMDEBUG
    result = new T[2*lenwei];
#else
    result = (T *) xmem.newptr (sizeof(T), 2*lenwei, "XplReal", "nnfunc", "result");
#endif
    for (i=0; i<lenunit; i++)
    { if (*(unit+i)==INUNIT)
        in = i;
      if (*(unit+i)==OUTUNIT)
      { minout = i;
        out = lenunit-i;
        break;
      }
    }

    if (in>p) 
    { XAerrort (NN, "nnfunc : Number of input units > cols(x)");
      return;
    }

    if (out==0) 
    { XAerrort (NN, "nnfunc : No output units found");
      return;
    }
#ifndef MEMDEBUG
    y = new T[n*out];  
#else
    y = (T *) xmem.newptr (sizeof(T), n*out, "XplReal", "nnfunc", "y");
#endif  
    for (k=0; k<n; k++)
    { // printf ("X %i\n", k);
      for (i=0; i<lenwei; i++)
      { from = (long) (*(wei+i)-1);
        to   = (long) (*(wei+lenwei+i)-1);
        wft  = *(wei+2*lenwei+i);
        // printf ("%i -> %i : %f\n", from, to, wft);
        if (*(unit+from)==INUNIT)
        { tmp = *(x+from*n+k);
          actf = (long) *(unit+lenunit+from);
          switch (actf)
          { case 1 : *(result+i) = (tmp>0.0 ? 1.0 : 0.0); break;
            case 2 : *(result+i) = 1.0/(1.0+::x4exp(-tmp)); break;
            case 3 : *(result+i) = ::x4tanh(tmp); break;
            case 4 : *(result+i) = ::atan(tmp); break;
            default: *(result+i) = tmp; 
          }
          *(result+lenwei+i) = wft* *(result+i);
        }
        else
        { tmp = 0.0;
          for (j=i-1; j>-1; j--)
          { if (from == *(wei+j)-1)
            { *(result+i) = *(result+j);
              *(result+lenwei+i) = wft* *(result+i);
              break;
            }
            if (from == *(wei+lenwei+j)-1)
              tmp += *(result+lenwei+j);
          }
          if (j<0)
          { actf = (long) *(unit+lenunit+from);
            switch (actf)
            { case 1 : *(result+i) = (tmp>0.0 ? 1.0 : 0.0); break;
              case 2 : *(result+i) = 1.0/(1.0+::x4exp(-tmp)); break;
              case 3 : *(result+i) = ::x4tanh(tmp); break;
              case 4 : *(result+i) = ::atan(tmp); break;
              default: *(result+i) = tmp;
            }
            *(result+lenwei+i) =wft* *(result+i);
          }
        }
      }
      for (j=0; j<out; j++)
        *(y+j*n+k) = 0.0;
      for (j=0; j<lenwei; j++)
      { to  = (long) *(wei+lenwei+j)-1;
        if (to>=minout)
          *(y+(to-minout)*n+k) += *(result+lenwei+j);
      }
      for (j=0; j<out; j++)
      { actf = (long) *(unit+lenunit+minout+j);
        switch (actf)
        { case 1 : *(y+j*n+k) = (*(y+j*n+k) >0.0 ? 1.0 : 0.0); break;
          case 2 : *(y+j*n+k) = 1.0/(1.0+::x4exp(- *(y+j*n+k))); break;
          case 3 : *(y+j*n+k) = ::x4tanh(*(y+j*n+k)); break;
          case 4 : *(y+j*n+k) = ::atan(*(y+j*n+k)); break;
        }
      }  
    }
#ifndef MEMDEBUG
    delete[] result; 
#else
    xmem.delptr (result);
    result = NULL;
#endif    
    udim.setdim (0, n);
    udim.setdim (1, out);
    yarray->setdim(udim);
    yarray->setptr(y);
  }

  template<class T> void XplReal<T>::nnvisu (const XplReal<T> &xvarray, XplReal<T> *xvdarray, XplReal<T> *verrarray)
  { T t, *xvd, l, d, bdis, adis, xmin, ymin, xmax, ymax, *dis, *xv, *verr, *u; 
    long i, j, k, m, dislen, xvlen, xvwid;
    XplIndex xvdim, verrdim;
    
    dislen = dim(0);
    dis    = ptr;
    
    xv    = xvarray.getptr();
    xvdim = xvarray.getdim();
    xvlen = xvdim(0);
    xvwid = xvdim(1);

    if (dim(1)!=3)
    { XAerrort (NN, "nnvisu : cols(d)<>3");
      return;
    }
/*    if (xvdim(1)!=2)
    { XAerrort (NN, "nnvisu : cols(p)<>2");
      return;
    }
*/
    for (i=2; i<MAX_XARRAY; i++)
    { if (dim(i)!=1)
      { XAerrort (NN, "nnvisu : layer(d)<>1");
        return;
      }
      if (xvdim(i)!=1)
      { XAerrort (NN, "nnvisu : layer(p)<>1");
        return;
      }
    }

#ifndef MEMDEBUG 
    xvd  = new T[xvwid*xvlen];
#else
    xvd  = (T *) xmem.newptr (sizeof(T), xvwid*xvlen, "XplReal", "nnvisu", "xvd");
#endif
    memcpy (xvd, xv, xvlen*xvwid*sizeof(T));

    verrdim = 1;
    verrdim.setdim (0, 1+dim(0));
#ifndef MEMDEBUG 
    u    = new T[verrdim(0)];
    verr = new T[verrdim(0)];
    xv  = new T[xvwid*xvlen];
#else
    u    = (T *) xmem.newptr (sizeof(T), verrdim(0), "XplReal", "nnvisu", "u");
    verr = (T *) xmem.newptr (sizeof(T), verrdim(0), "XplReal", "nnvisu", "verr");
    xv   = (T *) xmem.newptr (sizeof(T), xvwid*xvlen, "XplReal", "nnvisu", "xv");
#endif

    memcpy (xv, xvd, xvlen*xvwid*sizeof(T));

    u[dislen] = l = 0.0;
    for (i=0; i<dislen; i++)
    { j = (long) (*(dis+i) - 1);
      k = (long) (*(dis+dislen+i) - 1);
      //      if ((j<0) || (k<0) || (j>=dislen) || (k>=dislen)) replaced 030513 by SK
      if ((j<0) || (k<0) || (j>=xvlen) || (k>=xvlen))
      { 
	XAerrort (NN, "nnvisu : distance index out of range");
#ifndef MEMDEBUG    
	delete[] xvd;
	delete[] u;
#else
	xmem.delptr (xvd);
	xmem.delptr (u);
	xvd = NULL;
	u = NULL;
#endif
	return;  
      }   
      d = 0.0;
      for (m=0; m<xvwid; m++)
        d = d + (*(xvd+m*xvlen+j) - *(xvd+m*xvlen+k)) * (*(xvd+m*xvlen+j) - *(xvd+m*xvlen+k));
      l += d;
      u[i] = (*(dis+2*dislen+i) - ::sqrt(d)) * (*(dis+2*dislen+i) - ::sqrt(d));
//      printf ("%i %f %f %f\n", i+1, *(dis+2*dislen+i), ::sqrt(d), u[i]);
      u[dislen] += u[i];
      u[i] /= d;
    }
    bdis = u[dislen]/l;
    memcpy (verr, u, (dislen+1)*sizeof(T));
    
    t = 1.0;
    while (t>0.001)
    { for (i=0; i<xvlen; i++)
        for (m=0; m<xvwid; m++)
          *(xvd+m*xvlen+i) = *(xv+m*xvlen+i) + t*normalrandom();
      u[dislen] = l = 0.0;
      for (i=0; i<dislen; i++)
      { j = (long) (*(dis+i) - 1);
        k = (long) (*(dis+dislen+i) - 1);
        d = 0.0;
        for (m=0; m<xvwid; m++)
          d = d + (*(xvd+m*xvlen+j) - *(xvd+m*xvlen+k)) * (*(xvd+m*xvlen+j) - *(xvd+m*xvlen+k));
        l += d;
        u[i] = (*(dis+2*dislen+i) - ::sqrt(d)) * (*(dis+2*dislen+i) - ::sqrt(d));
        u[dislen] += u[i];
        u[i] /= d;
       }
      adis = u[dislen]/l;
      if (bdis>adis)
      { bdis = adis;
        memcpy (xv, xvd, xvwid*xvlen*sizeof(T));
        memcpy (verr, u, (dislen+1)*sizeof(T));
      }
      t *= 0.99;
    }  
// just center !!
    for (m=0; m<xvwid; m++)
    { xmin = xmax = *(xv+m*xvlen);
      for (i=1; i<xvlen; i++)
      { if (xmin>*(xv+m*xvlen+i)) xmin = *(xv+m*xvlen+i); 
        if (xmax<*(xv+m*xvlen+i)) xmax = *(xv+m*xvlen+i); 
      }
      l = 0.5*(xmax+xmin);
      for (i=0; i<xvlen; i++)
        *(xv+m*xvlen+i) -= l;
    }
#ifndef MEMDEBUG    
    delete[] xvd;
    delete[] u;
#else
    xmem.delptr (xvd);
    xmem.delptr (u);
    xvd = NULL;
    u = NULL;
#endif

    xvdarray->setdim(xvdim);
    xvdarray->setptr(xv);

    verrarray->setdim(verrdim);
    verrarray->setptr(verr);    
  }

  template<class T> XplReal<T> XplReal<T>::isoreg (void) const
  { XplReal<T> ret;
    ret = (*this);
    ret.layernn (sisoreg);
    return (ret);
  }

  template<class T> XplReal<T> XplReal<T>::sknn (const XplReal<T> &yarray, const XplNumber<long> &karray) const
  { T *y, *sm;
    long k; int i;
    XplIndex ydim;
    XplReal<T> ret(1);

    ydim = yarray.getdim();

    if (karray.size()!=1)
    { XAerrort (DIMCON, "sknn: k is not scalar");
      return (ret);
    }
    if (dim(0)!=ydim(0))
    { XAerrort (DIMCON, "sknn: rows(x)<>rows(y)");
      return (ret);
    }
    for (i=1; i<MAX_XARRAY; i++)
    { if (dim(1)!=1)
      { XAerrort (DIMCON, "sknn: layer_i(x)<>1");
        return (ret);
      }
      if (ydim(1)!=1)
      { XAerrort (DIMCON, "sknn: layer_i(y)<>1");
        return (ret);
      }
    }
    k = karray[0];
    if (k%2==0)
    { XAerrort (UNEXVAL, "sknn: odd k expected");
      return (ret);
    }
    if (k>dim(0))
    { XAerrort (DIMELEM, "sknn: k > rows(x)");
      return (ret);
    }
    // PC 20011004: unnecessary transformation removed:
    // k = 1+k/2;

    y = yarray.getptr();
#ifndef MEMDEBUG
    sm = new T[dim(0)];
#else
    sm = (T *) xmem.newptr (sizeof(T), dim(0), "XplReal", "sknn", "sm");
#endif
    s_knn (ptr, y, dim(0), k, sm);

    ydim = 1;
    ydim.setdim (0, dim(0));

    ret.setptr (sm);
    ret.setdim (ydim);
    return (ret);
  }

  template<class T> XplReal<T> XplReal<T>::rmed (const XplNumber<long> &karray) const
  { XplReal<T> ret (dim);
    T *retptr, *fromptr;
    XplIndex kdim, resdim, tmp;
    long i, k, n, s;
    kdim = karray.getdim();
    if (kdim(0)!=1)
    { XAerrort (DIMCON, "rmed: rows(k)<>1");
      return (ret);
    }
    for (i=1; i<MAX_XARRAY; i++)
    { if ((kdim(i)!=dim(i)) && (kdim(i)!=1))
      { XAerrort (DIMCON, "rmed: layer(x)<>layer(k)");
        return ret;
      }
    }
    n = dim(0);
    resdim = ret.getdim();
    resdim.setdim (0, 1);
    tmp = 0L;
    fromptr = ptr;
    retptr = ret.getptr();
    do
    { k = karray[kdim.value(tmp % kdim)];
      if ((k<1) || (k%2==0))
      { XAerrort (DIMCON, "rmed: k has to be an positive odd number");
        return (ret);
      }
      runmed (n, k, fromptr, retptr);    
      fromptr += n;
      retptr  += n;
    }
    while (tmp.inc(resdim));
    return (ret);
  }

  template<class T> XplReal<T> XplReal<T>::l1line (const XplReal<T> &yarray) const
  { long n, i, tind, rind, lind;
    XplReal<T> ret(1);
    XplIndex tmp, rdim, retdim, resdim;
    T *lptr, *rptr, *retptr, *xtmp;
    
    n = dim(0);
    rdim = yarray.getdim();
    if (rdim(0)!=dim(0))
    { XAerrort (DIMCON, "l1line: rows(x)<>rows(y)");
      return ret;
    }
    retdim.setdim(0, 2);
    for (i=1; i<MAX_XARRAY; i++)
    { retdim.setdim(i, dim(i));
      if (retdim(i)==1)
        retdim.setdim(i, rdim(i));
      else if (rdim(i)>1)
      { if (rdim(i)!=retdim(i))
        { XAerrort (DIMCON, "l1line: layer(x)<>layer(y)");
          return ret;
        }
      }
    }
#ifndef MEMDEBUG
    retptr = new T[retdim.prod()];
    xtmp   = new T[n];
#else
    retptr = (T *) xmem.newptr (sizeof(T), retdim.prod(), "XplReal", "l1line", "retptr");
    xtmp = (T *) xmem.newptr (sizeof(T), n, "XplReal", "l1line", "xtmp");
#endif  
    resdim = retdim;
    resdim.setdim(0, 1);
    rptr = yarray.getptr();
    tmp = 0L;
    do
    { tind = retdim.value(tmp % retdim);
      rind = rdim.value(tmp % rdim);
      lind = dim.value(tmp % dim);
      memcopy (xtmp, ptr+lind, n);
//      sl1line(n, xtmp, (rptr+rind), (retptr+tind));
      l1sline(xtmp, (rptr+rind), n, (retptr+tind), (retptr+tind+1));
    }
    while (tmp.inc(resdim));
#ifndef MEMDEBUG
    delete[] xtmp;
#else
    xmem.delptr (xtmp); xtmp = NULL;
#endif
    ret.setdim(retdim);
    ret.setptr(retptr);
    return (ret);
  }

  template<class T> XplReal<T> XplReal<T>::lowess (const XplReal<T> &yarray, const XplReal<T> *farray) const
  { XplIndex retdim, ydim, fdim, resdim, tmp;
    XplReal<T> ret;
    T *yptr, *fptr, *retptr;
    long i, n, rind, lind, tind, find;

    ydim = yarray.getdim();
    yptr = yarray.getptr();
    if (farray!=NULL)
    { fdim = farray->getdim();
      fptr = farray->getptr();
    }
    else // default f=0.5
    { fdim = 1;
#ifndef MEMDEBUG
      fptr = new T[1];
#else
      fptr = (T *) xmem.newptr (sizeof(T), 1, "XplReal", "lowess", "fptr");
#endif  
      *fptr = 0.5;
    }
    
    if (dim(0)!=ydim(0))
    { XAerrort (DIMCON, "lowess: rows(x)<>rows(y)");
#ifndef MEMDEBUG
      if (farray==NULL)
        delete[] fptr;
#else
      if (farray==NULL)
      { xmem.delptr (fptr); fptr = NULL; }
#endif
      return (ret);
    }    
    n = dim(0);
    retdim = dim;
//    retdim.setdim(0, n);
    if (fdim(0)!=1)
    { XAerrort (DIMCON, "lowess: rows(f)<>1");
#ifndef MEMDEBUG
      if (farray==NULL) 
        delete[] fptr;
#else
      if (farray==NULL) 
      { xmem.delptr (fptr); fptr = NULL; }
#endif
      return (ret);
    }    
    for (i=1; i<MAX_XARRAY; i++)
    { // retdim.setdim(i) = dim(i);
      if (retdim(i)>1)
      { if ((ydim(i)>1) && (ydim(i)!=retdim(i)))
        { XAerrort (DIMCON, "lowess: layer(x)<>layer(y)");
#ifndef MEMDEBUG
          if (farray==NULL) 
            delete[] fptr;
#else
          if (farray==NULL) 
          { xmem.delptr (fptr); fptr = NULL; }
#endif
          return (ret);
        }    
        if ((fdim(i)>1) && (fdim(i)!=retdim(i)))
        { XAerrort (DIMCON, "lowess: layer(x)<>layer(f)");
#ifndef MEMDEBUG
          if (farray==NULL) 
            delete[] fptr;
#else
          if (farray==NULL) 
          { xmem.delptr (fptr); fptr = NULL; }
#endif
          return (ret);
        }    
      }
      else 
      { retdim.setdim(i, ydim(i));
        if ((fdim(i)>1) && (fdim(i)!=retdim(i)))
        { XAerrort (DIMCON, "lowess: layer(y)<>layer(f)");
#ifndef MEMDEBUG
          if (farray==NULL) 
            delete[] fptr;
#else
          if (farray==NULL) 
          { xmem.delptr (fptr); fptr = NULL; }
#endif
          return (ret);
        }    
        else
          retdim.setdim(i, fdim(i));
      }  
    }

#ifndef MEMDEBUG
    retptr = new T[retdim.prod()];
#else
    retptr = (T *) xmem.newptr (sizeof(T), retdim.prod(), "XplReal", "lowess", "retptr");
#endif  
    resdim = retdim;
    resdim.setdim (0, 1);
    tmp = 0L;
    do
    { tind = retdim.value(tmp % retdim);
      rind = ydim.value(tmp % ydim);
      lind = dim.value(tmp % dim);
      find = fdim.value(tmp % fdim);
      slowess ((ptr+lind), (yptr+rind), n, *(fptr+find), 1, 0.0, (retptr+tind));
    }
    while (tmp.inc(resdim));
        
#ifndef MEMDEBUG
    if (farray==NULL) 
      delete[] fptr;
#else
    if (farray==NULL) 
    { xmem.delptr (fptr); fptr = NULL; }
#endif
    ret.setdim(retdim);
    ret.setptr(retptr);
    return (ret);
  }

  template<class T> void XplReal<T>::fwtin (const XplNumber<long> &deptharray, const XplReal<T> &hvec, 
              XplReal<T> *ti) const
  { long i, j, k;
    XplIndex tsdim, pmdim, hdim;
    T *datap, *hvecp;
    long hlen, depth, length;

    length = dim(0);
    pmdim  = deptharray.getdim();
    hdim   = hvec.getdim();

    if (pmdim(0)!=1)
    { XAerrort (DIMCON, "fwtin: rows(depth<>1");
        return;
    }

    for (i=1; i<MAX_XARRAY; i++)
    { if (dim(i)!=1)
      { XAerrort (DIMCON, "fwtin: layer(x)<>1");
        return;
      }
      if (pmdim(i)!=1)
      { XAerrort (DIMCON, "fwtin: layer(depth)<>1");
        return;
      }
      if (hdim(i)!=1)
      { XAerrort (DIMCON, "fwtin: layer(h)<>1");
        return;
      }
    }

    hlen  = hdim.prod();
    hvecp = hvec.getptr();
    depth = deptharray(0);
    if ((depth<2) || (depth>::x4log(length)/LOG2)) 
    { XAerrort (RANGE, "fwtin: 1<depth<::log2(n)");
      return;
    }
    setwc (hlen, hvecp);
    double **TI = ::wvmatrix(depth,length);
    fwt_TI(ptr,TI,length,depth);

#ifndef MEMDEBUG
    hvecp = new T[depth*length];
#else
    hvecp = (T *) xmem.newptr(sizeof(T), depth*length, "XplReal", "fwtin", "hvecp");
#endif
    for (i=k=0; i<depth; i++)
    { for (j=0; j<length; j++, k++)
        hvecp[k] = TI[i][j];
    }
    delete_matrix(TI,depth);

    hdim    = 1;
    hdim.setdim (0, length);
    hdim.setdim (1, depth);   
    
    ti->setdim(hdim);
    ti->setptr(hvecp);
  }

  template<class T> void XplReal<T>::invfwtin (const XplNumber<long> &deptharray, const XplReal<T> &hvec, 
                 XplReal<T> *ti) const
  { long i, j, k;
    XplIndex tsdim, pmdim, hdim;
    T *datap, *hvecp;
    long hlen, depth, length;

    length = dim(0);
    pmdim  = deptharray.getdim();
    hdim   = hvec.getdim();

    if (pmdim(0)!=1)
    { XAerrort (DIMCON, "invfwtin: rows(depth<>1");
        return;
    }

    for (i=1; i<MAX_XARRAY; i++)
    { if (i==1) 
      { if (dim(i)!=deptharray[0])
        { XAerrort (DIMCON, "invfwtin: cols(x)<>depth");
          return;
        }
      }
      else if (dim(i)!=1)
      { XAerrort (DIMCON, "invfwtin: layer(x)<>1");
        return;
      }
      if (pmdim(i)!=1)
      { XAerrort (DIMCON, "invfwtin: layer(depth)<>1");
        return;
      }
      if (hdim(i)!=1)
      { XAerrort (DIMCON, "invfwtin: layer(h)<>1");
        return;
      }
    }

    hlen  = hdim.prod();
    hvecp = hvec.getptr();
    depth = deptharray(0);
    if ((depth<2) || (depth>::x4log(length)/LOG2)) 
    { XAerrort (RANGE, "invfwtin: 1<depth<::log2(n)");
      return;
    }
    setwc (hlen, hvecp);
    double **TI = ::wvmatrix(depth,length);

    for (i=k=0; i<depth; i++)
    { for (j=0; j<length; j++, k++)
        TI[i][j] = ptr[k];
    }

#ifndef MEMDEBUG
    hvecp = new T[depth*length];
#else
    hvecp = (T *) xmem.newptr(sizeof(T), depth*length, "XplReal", "fwtin", "hvecp");
#endif

    invert_TI(TI,length,depth,hvecp);

    delete_matrix(TI, depth);

    hdim    = 1;
    hdim.setdim (0, length);
    
    ti->setdim(hdim);
    ti->setptr(hvecp);
  }

  template<class T> void XplReal<T>::fwtin_shift (const XplNumber<long> &param, XplReal<T> *ti) const
  { long i, j, k, shift;
    XplIndex tsdim, hdim;
    T *datap, *hvecp;
    long depth, length;

    length = dim(0);
    hdim  = param.getdim();

    if (hdim(0)!=2)
    { XAerrort (DIMCON, "fwtin_shift: rows(param)<>2");
        return;
    }

    for (i=1; i<MAX_XARRAY; i++)
    { if (i==1) 
      { if (dim(i)!=param(0))
        { XAerrort (DIMCON, "fwtin_shift: cols(x)<>depth");
          return;
        }
      }
      else if (dim(i)!=1)
      { XAerrort (DIMCON, "fwtin_shift: layer(x)<>1");
        return;
      }
      if (hdim(i)!=1)
      { XAerrort (DIMCON, "fwtin_shift: layer(param)<>1");
        return;
      }
    }

    depth = param(0);
    if ((depth<2) || (depth>::x4log(length)/LOG2)) 
    { XAerrort (RANGE, "fwtin_shift: 1<depth<::log2(n)");
      return;
    }

    double **TI = ::wvmatrix(depth,length);

    for (i=k=0; i<depth; i++)
    { for (j=0; j<length; j++, k++)
        TI[i][j] = ptr[k];
    }

#ifndef MEMDEBUG
    hvecp = new T[depth*length];
#else
    hvecp = (T *) xmem.newptr(sizeof(T), depth*length, "XplReal", "fwtin", "hvecp");
#endif
    shift = param(1);
    retrieve(TI, shift, length, depth, hvecp);

//    for (i=k=0; i<depth; i++)
//    delete[] (TI[0]);
//    delete[] TI;
//    deletematrixptr();
    delete_matrix (TI, depth);    

    hdim    = 1;
    hdim.setdim (0, length);
    
    ti->setdim(hdim);
    ti->setptr(hvecp);
  }

  template<class T> void XplReal<T>::nnrsetnet (const XplReal<T>& param, const XplReal<T> &nconn, const XplReal<T> &conn) const
  { long i;

    if (param.dim(0)!=7)
    { XAerrort (NNR, "nnrsetnet: rows(param) <> 7");
      return;
    } 
    if (param.dim(1)!=1)
    { XAerrort (NNR, "nnrsetnet: cols(param) <> 1");
      return;
    } 
    if (nconn.dim(1)!=1)
    { XAerrort (NNR, "nnrsetnet: cols(nconn) <> 1");
      return;
    } 
    if (conn.dim(1)!=1)
    { XAerrort (NNR, "nnrsetnet: cols(conn) <> 1");
      return;
    } 

    for (i=2; i<MAX_XARRAY; i++)
    { if (param.dim(i)!=1)
      { XAerrort (NNR, "nnrsetnet: layer(param) <> 1");
        return;
      } 
      if (nconn.dim(i)!=1)
      { XAerrort (NNR, "nnrsetnet: layer(nconn) <> 1");
        return;
      } 
      if (conn.dim(i)!=1)
      { XAerrort (NNR, "nnrsetnet: layer(conn) <> 1");
        return;
      } 
    }

    long *n, *nconnp, *connp, nsunits, entropy, softmax;
    double decay;

#ifndef MEMDEBUG
    n       = new long[3];
    nconnp  = new long[nconn.size()];
    connp   = new long[conn.size()];
#else
    n       = (long *) xmem.newptr(sizeof(long), 3, "XplReal", "nnrsetnet", "n");
    nconnp  = (long *) xmem.newptr(sizeof(long), nconn.size(),"XplReal", "nnrsetnet", "nconnp");
    connp   = (long *) xmem.newptr(sizeof(long), conn.size(),"XplReal", "nnrsetnet", "connp");
#endif    
    n[0]    = (long) param(0);
    n[1]    = (long) param(1);
    n[2]    = (long) param(2);
    nsunits = (long) param(3);
    entropy = (long) param(4);
    softmax = (long) param(5);
    decay   = param(6);
    for (i=0; i<nconn.size(); i++)
      nconnp[i] = (long) nconn(i);
    for (i=0; i<conn.size(); i++)
      connp[i] = (long) conn(i);
    ::set_net (n, nconnp, connp, &nsunits, &entropy, &softmax, &decay);

#ifndef MEMDEBUG  
    delete[] n;
    delete[] nconnp;
    delete[] connp;
#else
    xmem.delptr (n); n = NULL;
    xmem.delptr (nconnp); nconnp = NULL;
    xmem.delptr (connp); connp = NULL;
#endif
  }

  template<class T> void XplReal<T>::nnrsettrain (const XplReal<T> &ntr, const XplReal<T> &xy, const XplReal<T> &weights) const
  { long i, j;

    if (ntr.dim(0)!=1)
    { XAerrort (NNR, "nnrsettrain: rows(ntr)<>1");
      return;
    }
    if (ntr.dim(1)!=1)
    { XAerrort (NNR, "nnrsettrain: cols(ntr)<>1");
      return;
    }
    if (weights.dim(1)!=1)
    { XAerrort (NNR, "nnrsettrain: cols(weights)<>1");
      return;
    }
    
    for (i=2; i<MAX_XARRAY; i++)
    { if (ntr.dim(i)!=1)
      { XAerrort (NNR, "nnrsettrain: layer(ntr) <> 1");
        return;
      } 
      if (xy.dim(i)!=1)
      { XAerrort (NNR, "nnrsettrain: layer(xy) <> 1");
        return;
      } 
      if (weights.dim(i)!=1)
      { XAerrort (NNR, "nnrsettrain: layer(weights) <> 1");
        return;
      } 
    }

    long ntrv;
    double *trainp, *weightsp;
  
    ntrv = (long) ntr(0);

#ifndef MEMDEBUG
    trainp = new double[xy.size()];
    weightsp = new double[weights.size()];
#else
    trainp = (double *) xmem.newptr(sizeof(double), xy.size(), "XplReal", "nnrsettrain", "trainp");
    weightsp = (double *) xmem.newptr(sizeof(double), weights.size(), "XplReal", "nnrsettrain", "weightsp");
#endif
    for (i=0; i<xy.size(); i++)
      trainp[i] = xy[i];

    for (i=0; i<weights.size(); i++)
      weightsp[i] = weights[i];

    ::set_train (&ntrv, trainp, weightsp);

#ifndef MEMDEBUG
    delete[] trainp;
    delete[] weightsp;
#else
    xmem.delptr (trainp); trainp = NULL;
    xmem.delptr (weightsp); weightsp = NULL;
#endif
  }

  template<class T> void XplReal<T>::nnrdovm (const XplReal<T> &param, const XplReal<T> &wts, XplReal<T> *nwts, XplReal<T> *nval) const
  { long i;

    if (param.dim(0)!=3)
    { XAerrort (NNR, "nnrdovm: rows(param) <> 3");
      return;
    } 
    if (param.dim(1)!=1)
    { XAerrort (NNR, "nnrdovm: cols(param) <> 1");
      return;
    } 
    if (wts.dim(1)!=1)
    { XAerrort (NNR, "nnrdovm: cols(wts) <> 1");
      return;
    } 
    for (i=2; i<MAX_XARRAY; i++)
    { if (param.dim(i)!=1)
      { XAerrort (NNR, "nnrdovm: layer(param) <> 1");
        return;
      } 
      if (wts.dim(i)!=1)
      { XAerrort (NNR, "nnrdovm: layer(wts) <> 1");
        return;
      }
    }

    long nweights, maxit, trace;
    double *wtsp, *fmin;

    nweights = (long) param(0);
    maxit    = (long) param(1);
    trace    = (long) param(2);

#ifndef MEMDEBUG
    wtsp = new double[wts.size()];
    fmin = new double[1];
#else
    wtsp = (double *) xmem.newptr(sizeof(double), wts.size(), "XplReal", "nnrdovm", "wtsp");
    fmin = (double *) xmem.newptr(sizeof(double), 1, "XplReal", "nnrdovm", "fmin");
#endif
    for (i=0; i<wts.size(); i++)
      wtsp[i] = wts[i];

    ::dovm (&nweights, wtsp, fmin, &maxit, &trace);

    nwts->setptr(wtsp);
    nwts->setdim(wts.getdim());

    nval->setptr(fmin);
    nval->dim = 1; 

  }

  template<class T> void XplReal<T>::nnrtest (const XplReal<T> &param, const XplReal<T> &xy, const XplReal<T> &wts, XplReal<T> *yh, XplReal<T> *he) const
  { long i;

    if (param.dim(0)!=2)
    { XAerrort (NNR, "nnrtest: rows(param) <> 2");
      return;
    } 
    if (param.dim(1)!=1)
    { XAerrort (NNR, "nnrtest: cols(param) <> 1");
      return;
    } 
    if (wts.dim(1)!=1)
    { XAerrort (NNR, "nnrtest: cols(wts) <> 1");
      return;
    } 

    for (i=2; i<MAX_XARRAY; i++)
    { if (param.dim(i)!=1)
      { XAerrort (NNR, "nnrtest: layer(param) <> 1");
        return;
      } 
      if (wts.dim(i)!=1)
      { XAerrort (NNR, "nnrtest: layer(wts) <> 1");
        return;
      }
      if (xy.dim(i)!=1)
      { XAerrort (NNR, "nnrtest: layer(xy) <> 1");
        return;
      }
    }

    long ntest, np;
    double *test, *result, *inwts, *hess;

    ntest = (long) param(0);
    np    = (long) param(1);

#ifndef MEMDEBUG
    test = new double[xy.size()];
    result = new double[xy.dim(0)*np];
    inwts  = new double[wts.size()];
    hess = new double[wts.dim(0)*wts.dim(0)];
#else
    test = (double *) xmem.newptr(sizeof(double), xy.size(), "XplReal", "nnrtest", "test");
    result = (double *) xmem.newptr(sizeof(double), xy.dim(0)*np, "XplReal", "nnrtest", "result");
    inwts  = (double *) xmem.newptr(sizeof(double), wts.size(), "XplReal", "nnrtest", "inwts");
    hess = (double *) xmem.newptr(sizeof(double), wts.dim(0)*wts.dim(0) ,"XplReal","nnrtest","hess");
#endif
    for (i=0; i<xy.size(); i++)
      test[i] = xy[i];

     for (i=0; i<wts.size(); i++)
      inwts[i] = wts[i];

    ::nntest (&ntest, test, result, inwts);
    ::Hessian(inwts, hess);
    yh->setptr(result);
    yh->dim = 1;
    (yh->dim).setdim(0, xy.dim(0));
    (yh->dim).setdim(1, np);

    he->setptr(hess);
    he->dim = 1;
    (he->dim).setdim(0, wts.dim(0));
    (he->dim).setdim(1, wts.dim(0));
#ifndef MEMDEBUG
    delete[] test;
    delete[] inwts;
#else
    xmem.delptr (test); test = NULL;
    xmem.delptr (inwts); inwts = NULL;
#endif    

  }

  template<class T> int XplReal<T>::svd (XplReal<T> *u, XplReal<T> *l, XplReal<T> *v) const
  { int n, p, i, j, one, ierr;
    XplIndex udim, ldim, vdim;
    double *uptr, *lptr, *vptr, *rv1;

    for (i=2; i<MAX_XARRAY; i++)
    { if (dim(i)>1) 
      { XAerrort (DIMCON, "svd : layer(x) larger than 1");
        return 1;
      }
    }

    p = dim(1);
    n = dim(0);

    if ((n<0) || (p<0) || (n<p))
    { XAerrort (DIMCON, "svd : rows(x) > cols(x) or cols(x)/rows(x) less than 1");
      return 1;
    }

#ifndef MEMDEBUG    
    uptr = new T[n*p];
    vptr = new T[n*p];
    lptr = new T[p];
    rv1  = new T[n];
#else
    uptr = (T *) xmem.newptr(sizeof(T), n*p, "XplReal", "svd", "uptr");
    vptr = (T *) xmem.newptr(sizeof(T), n*p, "XplReal", "svd", "vptr");
    lptr = (T *) xmem.newptr(sizeof(T), p, "XplReal", "svd", "lptr");
    rv1  = (T *) xmem.newptr(sizeof(T), n, "XplReal", "svd", "rv1");
#endif

    one  = 1; 

    ::svd_ (&n, &n, &p, ptr, lptr, &one, uptr, &one, vptr, &ierr, rv1);

#ifndef MEMDEBUG
    delete[] rv1;
#else
    xmem.delptr (rv1);
#endif

    if (ierr)
    { XAerrort (NOCONV, "svd : decomposition failed");
#ifndef MEMDEBUG
      delete[] uptr;
      delete[] lptr;
      delete[] vptr;
#else
      xmem.delptr (uptr); uptr = NULL;
      xmem.delptr (lptr); lptr = NULL;
      xmem.delptr (vptr); vptr = NULL;
#endif
      return 1;
    }

    udim = 1;
    udim.setdim(0, n);
    udim.setdim(1, p);

    ldim = 1;
    ldim.setdim(0, p);
    
    vdim = 1;
    vdim.setdim(0, p);
    vdim.setdim(1, p);

#ifndef MEMDEBUG
    rv1 = new T[p*p];
#else
    rv1 = (T *) xmem.newptr (sizeof(T), p*p, "XplReal", "svd", "rv1");
#endif
    for (i=0; i<p; i++)
    { for (j=0; j<p; j++)
        *(rv1+j*p+i) = *(vptr+j*n+i);
    }
    delete []vptr;

    u->setdim (udim);
    u->setptr (uptr);
  
    l->setdim (ldim);
    l->setptr (lptr);

    v->setdim (vdim);
    v->setptr (rv1);
    return 0;
  }

// SK 02.09.97
  template<class T> void XplReal<T>::xplkmeans 
                 (const XplNumber<long> &bmat, XplNumber<long> *rmat, 
                  XplNumber<long> *gmat, XplReal<T> *cmat, 
                  XplReal<T> *vmat, XplReal<T> *smat,
                  XplReal<T> *wmat, XplReal<T> *mmat) const
  { long size, len, wsize, i, j, l, nclus, npass, r;
    T *x = NULL, *m = NULL, *w = NULL, *c = NULL, *v = NULL, *s = NULL; 
    long *bt = NULL, *g=NULL, *b = NULL;

/*      ( g     c     v     s ) = KMEANS ( x     b    r   (w)   (m) )       */
/*      (n,1) (k,p) (k,p) (k,1)          (n,p) (n,1)     (p,1) (n,1)        */
/*    n=len=# of rows   p=size=# of columns   k=nclus=# of clusters         */
/* argc=2-5 argv[0]=x argv[1]=b argv[2]=r arg[3]=m argv[4]=w arg[5]=KMEANS  */

//  (this->dim).dump();
  for (i=2; i<MAX_XARRAY; i++)
  { if (dim(i)!=1)
      { XAerrort (DIMCON, "kmeans: layer(x)<>1");
        return;
      }
    if (bmat.dim(i)!=1)
      { XAerrort (DIMCON, "kmeans: layer(b)<>1");
        return;
      }
    if (rmat)
    { if (rmat->dim(i)!=1)
      { XAerrort (DIMCON, "kmeans: layer(r)<>1");
        return;
      }
    }
    if (wmat)
    { if (wmat->dim(i)!=1)
      { XAerrort (DIMCON, "kmeans: layer(w)<>1");
        return;
      }
    }
    if (mmat)
    { if (mmat->dim(i)!=1)
      { XAerrort (DIMCON, "kmeans: layer(m)<>1");
        return;
      }
    }
  }

  size = dim(1);
  len = dim(0);
  if (len < 3) 
  { XAerrort (PARAMFALSE, "kmeans: rows(x)<3");
    return;
  }

/*                                      partition vector                  */
  if (bmat.dim(0)!=len) 
  { XAerrort (DIMCON, "kmeans: rows(x)<>rows(b)");
    return;
  }

  wsize = 1;
  if (wmat)
  { if (wmat->dim(1) !=wsize ) 
    { XAerrort (DIMCON, "kmeans: cols(w)<>1");
      return;
    }
  }

  bt = bmat.getptr();
#ifndef MEMDEBUG 
  b  = new long[len];
#else
  b = (long *) xmem.newptr(sizeof(long), len, "XplReal", "xplkmeans", "b");
#endif

  nclus = 0;
  for (i = 0; i < len; ++i)
  { b[i] = bt[i];
    if (nclus - b[i] < 0) nclus = b[i];
  }
  
  if ( nclus <=1 )
  { XAerrort (PARAMFALSE, "kmeans: No cluster found");
    return;
  }  

  for (j = 1; j <= nclus; ++j) {
     l = 0;
     for (i = 0; i < len; ++i)
	if ( b[i] - j == 0 ) l = 1;
     if ( l < 1 ) 
     { XAerrort (PARAMFALSE, "kmeans: Invalid partition");
       return;
     }

  }
/*        r = control parameter (for example, the number of shifts at least */

  npass = 100;
  if (rmat)
    npass = (*rmat)(0);
  r=npass;

  if (!wmat)
  { 
#ifndef MEMDEBUG
    w = new T[wsize*size];
#else
    w = (T *) xmem.newptr(sizeof(T), wsize*wsize, "XplReal", "xplkmeans", "w");
#endif
    for ( i=0; i < size; ++i)
      w[i] = 1;
  }
  else
  { 
    if (wmat->dim(0)!=size) 
      { XAerrort (DIMCON, "kmeans: rows(x)<>rows(w)");
        return;
      }
    if (wmat->dim(1) !=wsize ) 
      { XAerrort (DIMCON, "kmeans: cols(w)<>1");
        return;
      }
    w = wmat->getptr();
    for (i = 0; i < size; ++i)
	if(w[i] <= 0.0) 
	  { XAerrort (PARAMFALSE, "kmeans: negative weight");
            return;
          }
  }

  if (!mmat)
  { 
#ifndef MEMDEBUG
    m = new T[wsize*len];
#else
    m = (T *) xmem.newptr(sizeof(T), wsize*len, "XplReal", "xplkmeans", "m");
#endif
    for ( i=0; i < len; ++i)
      m[i] = 1;
  }
  else
  { if (mmat->dim(0)!=len)
    { XAerrort (DIMCON, "kmeans: rows(m)<>rows(x)");
      return;
    }
    if (mmat->dim(1) !=wsize ) 
    { XAerrort (DIMCON, "kmeans: cols(m)<>cols(w)");
      return;
    }
    m = mmat->getptr();
    for (i = 0; i < len; ++i)
	if(m[i] < 0.0) 
        { XAerrort (PARAMFALSE, "negative weight");
          return;
        }
  }

  x = ptr;
#ifndef MEMDEBUG
  c = new T[nclus*size];
  v = new T[nclus*size];
  s = new T[nclus];
  g = new long[len];
#else
  c = (T *) xmem.newptr(sizeof(T), nclus*size, "XplReal", "xplkmeans", "c");
  v = (T *) xmem.newptr(sizeof(T), nclus*size, "XplReal", "xplkmeans", "v");
  s = (T *) xmem.newptr(sizeof(T), nclus, "XplReal", "xplkmeans", "s");
  g = (long *) xmem.newptr(sizeof(long), len, "XplReal", "xplkmeans", "g");
#endif
  for (i = 0; i < len; ++i)
    g[i] = b[i];

//  (this->dim).dump();
  ::kmeans (g, x, m, v, s, w, c, len, size, nclus, r);

#ifndef MEMDEBUG
  if (!mmat) delete[] m;
  if (!wmat) delete[] w;
#else
  if (!mmat) xmem.delptr (m);
  if (!wmat) xmem.delptr (w);
  m = NULL;
  w = NULL;
#endif

  XplIndex gdim (len, wsize), cdim(nclus, size), 
           vdim(nclus, size), sdim(nclus, wsize);

  gmat->setdim (gdim);
  gmat->setptr (g);

  cmat->setdim (cdim);
  cmat->setptr (c);
  
  vmat->setdim (vdim);
  vmat->setptr (v);

  smat->setdim (sdim);
  smat->setptr (s);

}

//SK 02.09.97
  template<class T> void XplReal<T>::xplagglom (int imtd, XplNumber<long> lmat, 
                                                XplReal<T> *pmat, XplReal<T> *tmat, XplReal<T> *gmat,
                                                XplReal<T> *pd, XplReal<T> *dmat, XplReal<T> *opt) const
/*   (  p      t       g   pd d) = AGGLOM (   d    mtd    l   (b)    (w) )    */
/*    (n,1) (2l,2)  (2n,2)           (len,1)                  (n,1)     */
/*      n=# of objects   len=# of distances   l=# of terminal clusters    */
/*      argc=2-5: argv[0]=d  argv[1]=mtd  argv[2]=l  argv[3]=b            */
/*                argv[4]=w  argv[5]=AGGLOM                               */
  { struct matrix *dmatrix = NULL, *wmatrix = NULL, *mat = NULL;
    long size, len, wsize, i, j, k, k1, nold, mtd, nlen, l, ll;
    T beta;
    long index, istep, *ibit, *ka, *kb, *kc, *kd, *np, *jx, *kf, *kq, *kt,
         knr;
    T *x = NULL, *w = NULL,  *d = NULL, *t = NULL, *h, *dbit, *g = NULL, 
      *y = NULL, *wy = NULL;

    size = dim(1);
    if (size==1) 
    { len = dim(0);
      if (len <= 2)
      { XAerrort (PARAMFALSE, "agglom: rows(x)<3");
        return;
      }
#ifndef MEMDEBUG
      x = new T[len];
#else
      x = (T *) xmem.newptr (sizeof(T), len, "XplReal", "", "x");
#endif
      for (i=0; i<len; i++) 
      { x[i] = ptr[i];
        if(x[i] < 0.0)
        { XAerrort (PARAMFALSE, "agglom: Negative distances in x");
            return;
        }
      }
    }
    else if (size==dim(0))
    { len = dim(0)*(dim(0)-1)/2;
      x = new T[len];
      for (k=i=0; i<size; i++)
      { for (j=i+1; j<size; j++, k++)
        { //printf ("aggl: %i = (%i,%i)\n", k, i, j);
          x[k] = (*this)(i,j);
          if(x[k] < 0.0)
          { XAerrort (PARAMFALSE, "agglom: Negative distances in x");
	    // printf ("%i %i %i %f\n", i, j, k, x[k]);
            return;
          }
        }
      }
    }
    else
    { XAerrort (PARAMFALSE, "agglom: cols(x)<>1 and cols(x)<>rows(x)");
      return;
    }

    nlen = (long) (::sqrt(len *2) +1);
//printf ("%li\n", imtd);
    if ((imtd<1) || (imtd>7))
    { XAerrort (PARAMFALSE, "agglom: Unknown clustering method");
      return;
    }

/*                                    numbers of clusters at most        */
/*    ll   = 20;  SK */
    ll   = lmat(0);
    if(ll < 2) ll = 2;
    if(ll > nlen) ll=nlen;   
    knr = ll;   

/*                                    Lance&Williams method parameter    */   
/*                                    weights (masses) of row points    */
   beta  = -0.15;
   wsize = 1;

   switch (imtd)
   { case 8 : /* LANCE */
       if (opt)
         beta = (*opt)(0);
       break;
     case 5 : /* AVERAGE */
     case 6 : /* CENTROID */
     case 7 : /* WARD */
       if (opt==NULL)
       { 
#ifndef MEMDEBUG
         w = new T[wsize*nlen];
#else
         w = (T *) xmem.newptr (sizeof(T), wsize*nlen, "XplReal", "xplagglom", "w");
#endif 
         for ( i=0; i < nlen; ++i)
           w[i] = 1;
       }
       else
       { w = opt->getptr();
         for (i=0; i < nlen; ++i)
         { if (w[i]<0)
           { XAerrort (PARAMFALSE, "agglom: Negative weights in opt");
             return;
           }
           // printf ("w[%i]=%f\n", i, w[i]);
         }
       }
       break;
     default:
       if (opt)
       { XAerrort (PARAMFALSE, "agglom: Unexpected optional parameter");
         return;
       }
#ifndef MEMDEBUG
       w = new T[wsize*nlen];
#else
       w = (T *) xmem.newptr (sizeof(T), wsize*nlen, "XplReal", "xplagglom", "w");
#endif 
       for ( i=0; i < nlen; ++i)
         w[i] = 1;
   } 

   if(beta < -1.2) beta = -0.15;
   if(beta > 1) beta = -0.15;

#ifndef MEMDEBUG
   ka = new long[nlen];
   kb = new long[nlen];
   kc = new long[nlen];
   h  = new T[nlen];
   y  = new T[knr*(knr-1)/2];
   wy = new T[nlen];
#else
   ka = (long *) xmem.newptr (sizeof(long), nlen, "XplReal", "xplagglom", "ka");
   kb = (long *) xmem.newptr (sizeof(long), nlen, "XplReal", "xplagglom", "kb");
   kc = (long *) xmem.newptr (sizeof(long), nlen, "XplReal", "xplagglom", "kc");
   h  = (T *) xmem.newptr (sizeof(T), nlen, "XplReal", "xplagglom", "h");
   y  = (T *) xmem.newptr (sizeof(T),  knr*(knr-1)/2, "XplReal", "xplagglom", "y");
   wy = (T *) xmem.newptr (sizeof(T), nlen, "XplReal", "xplagglom", "wy");
#endif

//   printf ("agglom nlen = %li beta=%f\n", nlen, beta);
   agglom (ka, kb, h, kc, w, x, nlen, imtd, beta, y, wy, knr);
//   printf ("agglom out\n");

#ifndef MEMDEBUG
   delete[] x;
   if (opt==NULL) delete[] w;
#else
   xmem.delptr (x);
   x = NULL;
   if (w) xmem.delptr(w);
   w = NULL;
#endif
   l=0;

#ifndef MEMDEBUG
   kq = new long[nlen];
   kd = new long[nlen];
   kf = new long[nlen];
   kt = new long[nlen];
   jx = new long[nlen];
   d  = new T[nlen];
   np = new long[nlen];
#else
   kq = (long *) xmem.newptr (sizeof (long), nlen, "XplReal", "xplagglom", "kq");
   kd = (long *) xmem.newptr (sizeof (long), nlen, "XplReal", "xplagglom", "kd");
   kf = (long *) xmem.newptr (sizeof (long), nlen, "XplReal", "xplagglom", "kf");
   kt = (long *) xmem.newptr (sizeof (long), nlen, "XplReal", "xplagglom", "kt");
   jx = (long *) xmem.newptr (sizeof (long), nlen, "XplReal", "xplagglom", "jx");
   d  = (T *) xmem.newptr (sizeof (T), nlen, "XplReal", "xplagglom", "d");
   np = (long *) xmem.newptr (sizeof (long), nlen, "XplReal", "xplagglom", "np");
#endif
   parti (ka, kb, h, kc, kq, kd, kf, kt, jx, d, np, knr, nlen);
   for (i = 0; i < nlen; ++i) wy[i] = np[i];

   parti (ka, kb, h, kc, kq, kd, kf, kt, jx, d, np, ll, nlen);

/*                                    save the partition vector       */
   for (i = 0; i < nlen; ++i) 
   { d[i] = np[i];
     // printf ("d[%i]=%f\n", i, d[i]);
   }   

   XplIndex ddim(nlen,wsize);
   pmat->setdim(ddim);
   pmat->setptr(d); 

   XplIndex wydim(nlen);
   pd->setdim(wydim);
   pd->setptr(wy);
   
   XplIndex ydim(knr*(knr-1)/2);
   dmat->setdim(ydim);
   dmat->setptr(y);
  
/*                         if l=0 then save the full binary tree      */
/*                         if l>0 then save the cluster dendrogram    */
L2:
   dbit = new T[nlen*2];
   ibit = new long[nlen*2];

   istep = -1;
   index = 1;

L1:
   for (i=nlen-2; i>=0; --i)
   {   if(ka[i]==index)
       {  istep=istep+1;
          dbit[istep]=h[i];
          ibit[istep]=kb[i];
       }
   }
   istep=istep+1;
   dbit[istep]=::sqrt(-1);
   ibit[istep]=index;
   index=0;
   for (i=istep-1; i>=0; --i)
   {   if(!XisNaN(dbit[i]) && (ibit[i]!=0))
       {  index=ibit[i];
          ibit[i]=0;
          break;
       }
   }
   if(index!=0) goto L1;
   dbit[istep+1]=::sqrt(-1);
   ibit[istep+1]=0;
   if (l)
   {
#ifndef MEMDEBUG
     t = new T[nlen*4];
#else
     t = (T *) xmem.newptr (sizeof(T), nlen*4, "XplReal", "xplagglom", "t");
#endif
     for (i=0; i<nlen*2; ++i)
     { t[i]=dbit[i];
       t[2*nlen+i]=ibit[i];
     }
/*     for (j=0; j<4*nlen; j++)
       printf ("Tptr[%i]==%f\n", j, t[j]); 
*/     j=2;
     XplIndex tdim(2*nlen, j);
     tmat->setdim(tdim);
     tmat->setptr(t);
/*     for (j=0; j<gmat->size(); j++)
       printf ("T[%i]=%f\n", j, tmat[j]);   
*/   }
   else
   {
#ifndef MEMDEBUG
      g = new T[nlen*4];
#else
      g = (T *) xmem.newptr (sizeof(T), nlen*4, "XplReal", "xplagglom", "g");
#endif
     for (i=0; i<nlen*2; ++i)
     { g[i]=dbit[i];
       g[2*nlen+i]=ibit[i];
     }
/*     for (j=0; j<4*nlen; j++)
       printf ("Gptr[%i]=%f\n", j, g[j]); 
*/     j=2;
     XplIndex gdim(2*nlen, j);
     gmat->setdim(gdim);
     gmat->setptr(g);
/*     for (j=0; j<gmat->size(); j++)
       printf ("G[%i]=%f\n", j, gmat[j]); 
*/   }
#ifndef MEMDEBUG
   delete[] dbit;
   delete[] ibit;
#else
   xmem.delptr (dbit);
   xmem.delptr (ibit);
   dbit = NULL;
   ibit = NULL;
#endif
   if(l!=0) goto L3;
       
/*                                    cut the hierarchy             */
   nold = nlen;
   nlen = ll;
   l=nlen;
   if (nlen == nold) goto L2;
   k1 = nold - nlen +1;
   d = new T[nold]; 

   for (i = 0; i < nold; ++i) {
      d[i] = h[i];
      k = kc[i];
      for (j = k1 - 1; j < nold-1; ++j) {
	  if (ka[j] == k) kd[j] = np[k-1];
	  if (kb[j] == k) kf[j] = np[k-1];
      }
   }
//printf ("del 4 in\n");
#ifndef MEMDEBUG
   delete[] ka;
   delete[] kb;
   delete[] kc;
   delete[] h; 
   ka = new long[nlen];
   kb = new long[nlen];
   kc = new long[nlen];
   h = new T[nlen];
#else
   xmem.delptr (ka);
   xmem.delptr (kb);
   xmem.delptr (kc);
   xmem.delptr (h);
   ka = (long *) xmem.newptr (sizeof(long), nlen, "XplReal", "xplagglom", "ka");
   kb = (long *) xmem.newptr (sizeof(long), nlen, "XplReal", "xplagglom", "kb");
   kc = (long *) xmem.newptr (sizeof(long), nlen, "XplReal", "xplagglom", "kc");
   h  = (T *) xmem.newptr (sizeof(T), nlen, "XplReal", "xplagglom", "h");
#endif

   k = 0;
   for (j = k1-1; j < nold-1; ++j) {
      ka[k] = kd[j];
      kb[k] = kf[j];
      h[k] = d[j];
      kc[k] = k+1;
      ++k;
   }
   ka[nlen-1] = 0;
   kb[nlen-1] = 0;
   kc[nlen-1] = nlen;
   h[nlen-1] = 1.E36;

//printf ("del 7 in\n");
#ifndef MEMDEBUG
   delete[] jx;
   delete[] kt;
   delete[] kq;
   delete[] d;
   delete[] kd;
   delete[] kf;
   delete[] np;
#else
   xmem.delptr (jx); jx = NULL;
   xmem.delptr (kt); kt = NULL;
   xmem.delptr (kq); kq = NULL;
   xmem.delptr (d);  d = NULL;
   xmem.delptr (kd); kd = NULL;
   xmem.delptr (kf); kf = NULL;
   xmem.delptr (np); np = NULL;
#endif
//printf ("del 7 out\n");
   goto L2;


L3:
#ifndef MEMDEBUG
   delete[] ka;
   delete[] kb;
   delete[] h;
#else
   xmem.delptr (ka); ka = NULL;
   xmem.delptr (kb); kb = NULL;
   xmem.delptr (h);  h  = NULL;
#endif

/*   for (i = 0; i < nlen; ++i) 
     printf ("pmat(%i)=%f\n", i, (*pmat)(i));
*/}

//SK 02.09.97
  template<class T> void XplReal<T>::xpltree (double cut, long opt,
                                              XplReal<T> *g, XplReal<T> *mind) const
  { long rot = 0, pos = 0, n = 0, i;
    T *x, *y, *v, t, px, py, *gptr, *ind;

    for (i=2; i<MAX_XARRAY; i++)
    { if (dim(i)!=1) 
      { XAerrort (PARAMFALSE, "tree: layer(g)<>1");
        return;
      }
    } 

    cmpop    = ((opt & 0x00000010) ? greatertree : smallertree);
    treeform = opt & 0x00000003; 
    rot      = opt & 0x00000004;
    lineform = opt & 0x00000008;

#ifndef MEMDEBUG
    x = new T[5*dim(0)];
    y = new T[5*dim(0)];
    ind = new T[5*dim(0)];
#else
    x   = (T *) xmem.newptr (sizeof(T), 5*dim(0), "XplReal", "xpltree", "x");
    y   = (T *) xmem.newptr (sizeof(T), 5*dim(0), "XplReal", "xpltree", "y");
    ind = (T *) xmem.newptr (sizeof(T), 5*dim(0), "XplReal", "xpltree", "ind");
#endif

//    build_tree (ptr, 0, &pos, &px, &py, x, y, &n, dim(0), fvec, tvec, &i);
    build_tree (ptr, 0, &pos, &px, &py, x, y, ind, &n, dim(0), &i);

    if (n<1)
    { XAerrort (ARRAYBAD, "rows(t)<1");
      return;
    }

    gptr = new T[2*n];
    if (rot)
    { for (i=0; i<n; i++)
      { gptr[i]   = *(y+i);
        gptr[i+n] = *(x+i);
      }
    }
    else
    { for (i=0; i<n; i++)
      { gptr[i]   = *(x+i);
        gptr[i+n] = *(y+i);
      }
    }

#ifndef MEMDEBUG  
    delete[] x;
    delete[] y;
#else
    xmem.delptr (x);
    xmem.delptr (y);
#endif

    XplIndex gdim (n, 2);
    g->setdim (gdim);
    g->setptr (gptr);  

    XplIndex mdim (n,1);
    mind->setdim(mdim);
    mind->setptr(ind);

  }

// SK 02.09.97
  template<class T> void XplReal<T>::xpldistance (int mtc, XplReal<T> *d,
                                                  XplReal<T> *w) const
  { long i, j, k, n, p;
    T *dptr, *ddptr, *wptr;

    n = dim(0);
    p = dim(1);

    for (i=2; i<MAX_XARRAY; i++)
    { if (dim(i)!=1)
      { XAerrort (DIMCON, "distance: layer(x)<>1");
        return;
      }
      if (w)
      { if (w->dim(i)!=1)
        { XAerrort (DIMCON, "distance: layer(w)<>1");
          return;
        }
      }
    }
    if (w)
    { if (w->dim(1)!=1)
      { XAerrort (DIMCON, "distance: cols(w)<>1");
        return;
      }
      if (w->dim(0)!=dim(1))
      { XAerrort (DIMCON, "distance: rows(w)<>cols(x)");
        return;
      }
      wptr = w->getptr();
    }
    else
    { wptr = new T[p];
      for (i=0; i<p; i++)
        wptr[i] = 1;
    }    
 
    XplIndex ddim (n,n);
#ifndef MEMDEBUG
    dptr  = new T[n*n];
    ddptr = new T[n*(n-1)/2];
#else
    dptr  = (T *) xmem.newptr (sizeof(T), n*n, "XplReal", "xpldistance", "dptr");
    ddptr = (T *) xmem.newptr (sizeof(T), n*(n-1)/2, "XplReal", "xpldistance", "ddptr");
#endif

    ::distance (ptr, mtc, wptr, n, p, ddptr);

    for (k=i=0; i<n; i++)
    { for (j=i+1; j<n; j++, k++)
      { //printf ("dist: (%i,%i) = %i\n", j, i, k);
        *(dptr+i*n+j) = ddptr[k];
        *(dptr+j*n+i) = ddptr[k];
      }
      *(dptr+i*n+i) = 0.0;
    }

#ifndef MEMDEBUG
    delete[] ddptr;
    if (w==NULL) delete[] wptr;
#else
    xmem.delptr (ddptr); ddptr = NULL;
    if (w==NULL) xmem.delptr (wptr);
    wptr = NULL;
#endif
    d->setdim (ddim);
    d->setptr (dptr);

  }

// SK 08.09.97
  template<class T> void XplReal<T>::fwt2 (XplReal<T> *res,
                                           XplReal<T> *hvec,           
                                           long level,
					   long asm_index) const
  { T *resptr, *hvecp;;
    XplIndex resdim, hdim;
    long i, j, k, hlen;

    if (dim(0)!=dim(1))
    { XAerrort (DIMCON, "fwt2: rows(x)<>cols(x)");
      return;
    }

    if (notpowof2(dim(0)))
    { XAerrort (DIMCON, "fwt2: rows(x) not power of 2");
      return;
    }

    if (level<1)
    { XAerrort (DIMCON, "fwt2: level < 1");
      return;
    }

    hdim = hvec->getdim();
    if (hdim(1)!=1)
    { XAerrort (DIMCON, "fwt2: cols(h)<>1");
      return;
    }

    for (i=2; i<MAX_XARRAY ; i++)
    { if (dim(i)!=1)
      { XAerrort (DIMCON, "fwt2: layer(x)<>1");
        return;
      }
      if (hdim(i)!=1)
      { XAerrort (DIMCON, "fwt2: layer(h)<>1");
        return;
      }
    }
   
    hvecp = hvec->getptr();
    hlen  = hdim.prod();
    setwc (hlen, hvecp);

    double **TI = ::wvmatrix(dim(1),dim(0));

    for (i=k=0; i<dim(1); i++)
    { for (j=0; j<dim(0); j++, k++)
        TI[i][j] = ptr[k];
    }
    
    ::fwt2(TI, dim(0), level, asm_index);


    resdim = dim;
    resptr = new T[size()];
    for (i=k=0; i<dim(1); i++)
    { for (j=0; j<dim(0); j++, k++)
        resptr[k] = TI[i][j];
    }
    
    delete_matrix(TI,dim(1));

    res->setdim (resdim);
    res->setptr (resptr);

  }

// SK 08.09.97
  template<class T> void XplReal<T>::invfwt2 (XplReal<T> *res,
                                           XplReal<T> *hvec,           
                                           long level,
					   long asm_index) const
  { T *resptr, *hvecp;
    XplIndex resdim, hdim;
    long i, j, k, hlen;

    if (dim(0)!=dim(1))
    { XAerrort (DIMCON, "invfwt2: rows(x)<>cols(x)");
      return;
    }

    if (notpowof2(dim(0)))
    { XAerrort (DIMCON, "invfwt2: rows(x) not power of 2");
      return;
    }

    if (level<1)
    { XAerrort (DIMCON, "invfwt2: level > 0");
      return;
    }

    hdim = hvec->getdim();
    if (hdim(1)!=1)
    { XAerrort (DIMCON, "fwt2: cols(h)<>1");
      return;
    }

    for (i=2; i<MAX_XARRAY ; i++)
    { if (dim(i)!=1)
      { XAerrort (DIMCON, "invfwt2: layer(x)<>1");
        return;
      }
      if (hdim(i)!=1)
      { XAerrort (DIMCON, "fwt2: layer(h)<>1");
        return;
      }
    }
   
    hvecp = hvec->getptr();
    hlen  = hdim.prod();
    setwc (hlen, hvecp);

    double **TI = ::wvmatrix(dim(1),dim(0));

    for (i=k=0; i<dim(1); i++)
    { for (j=0; j<dim(0); j++, k++)
        TI[i][j] = ptr[k];
    }
    
    ::invert_fwt2(TI, dim(0), level, asm_index);


    resdim = dim;
    resptr = new T[size()];
    for (i=k=0; i<dim(1); i++)
    { for (j=0; j<dim(0); j++, k++)
        resptr[k] = TI[i][j];
    }
    
    delete_matrix(TI,dim(1));

    res->setdim (resdim);
    res->setptr (resptr);

  }

// SK 29.09.97
  template<class T> void XplReal<T>::dwt (const XplNumber<long> &level, const XplReal<T> &h, XplReal<T> &d) const
  { long i, j, k, l, n, p, s, ll, hrange, alen, blen;
    T *xptr, *dptr, *hptr;
    XplIndex adim, bdim;

    n = dim(0);
    p = dim(1);
    l = level[0];

    if (notpowof2(n)) { XAerrort (DIMCON, "dwt : rows(x) != 2^n"); return; }
    if (p!=1)         { XAerrort (DIMCON, "dwt : cols(x) != 1"); return; }
    if (notpowof2(l)) { XAerrort (DIMCON, "dwt : level   != 2^n"); return; }
    if (l>=n)         { XAerrort (DIMCON, "dwt : level   >= rows(x)"); return; }
    if (h.dim(1)!=1)  { XAerrort (DIMCON, "dwt : cols(h) != 1"); return; }

    for (i=2; i<MAX_XARRAY; i++)
      if (h.dim(i)!=1) { XAerrort (DIMCON, "dwt : layer(h) != 1"); return; }
      
    d.setdim(dim);    

    hptr   = h.getptr();
    hrange = h.size();

#ifndef MEMDEBUG
    dptr = new T[n];
#else
    dptr = (T *) xmem.newptr (sizeof(T), n, "XplReal", "dwt", "dptr");
#endif  

    memcopy (dptr, ptr, n);
    fwt_int (dptr, n, l, hptr, hrange);
    d.setptr(dptr);    
  }

// SK 29.09.97
  template<class T> void XplReal<T>::invdwt (XplReal<T> &x, const XplNumber<long> &l, const XplReal<T> &h) const
  { long i, j, k, ma, sa, la, ll, alen, blen, hrange, level, s, length;
    T *xptr, *xlptr, *aptr, *bptr, *alptr, *blptr, *hptr;
    XplIndex xdim;

    if (l.dim(0)!=1) { XAerrort(DIMCON, "invdwt : l has to a scalar"); return; }
    if (l.dim(1)!=1) { XAerrort(DIMCON, "invdwt : l has to a scalar"); return; }
    if (l(0)>=dim(0))  { XAerrort(DIMCON, "invdwt : l >= n"); return; }
    if (dim(1)!=1) { XAerrort(DIMCON, "invdwt : vector for d expected"); return; }

    length = dim(0);
    xdim.setdim (0, length);
    xdim.setdim (1, 1);
    for (i=2; i<MAX_XARRAY; i++)
    { if (l.dim(i)!=1) { XAerrort(DIMCON, "invdwt : l has to be a scalar"); return; }
      if (dim(i)!=1)   { XAerrort (DIMCON, "invdwt : layer(d) <> 1"); return; }
      if (h.dim(i)!=1) { XAerrort(DIMCON, "invdwt : h has to be a vector"); return; }
      xdim.setdim(i,dim(i));
    }
   
#ifndef MEMDEBUG
    xptr = new T[xdim.prod()];
#else
    xptr = (T *) xmem.newptr (sizeof(T), xdim.prod(), "XplReal", "invdwt", "xlptr/xptr");
#endif  

    hptr   = h.getptr();
    hrange = h.dim(0)*h.dim(1);
    level  = l[0];

    memcopy (xptr, ptr, length);
    invert_fwt (xptr, length, level, hptr, hrange);

    x.setdim(xdim);
    x.setptr(xptr);
  }

// SK 21.10.97
  template<class T> void XplReal<T>::cartsplit (const XplReal<T> &y, const XplReal<T> &d,
                                                const XplReal<T> &t, const XplReal<T> &nd, 
                                                XplReal<T> *splitvec, CartSplit *splitval,
                                                XplReal<T> *splitmean, XplReal<T> *splitvar,
                                                XplReal<T> *splitn, 
						long nmaxgrp, long nmingrp,
                                                T vmaxgrp, XplReal<T> *endpoint)  const
  { 
    T *nptr;
    XplIndex dvec  (dim(0)),  //, dim(1)), 
             dval  (dim(0)),
             dmean (dim(0)),
             dvar  (dim(0)),
             dn    (dim(0)),
             dend  (dim(0));
    T mw, dw, *yptr, *tptr;

    long *sgrp, i, nextgrp, kk;


    if (nmaxgrp<1)
    { XAerrort(DIMCON, "cartsplit : n<1");
      return;
    }
    if (nmingrp<1)
    { XAerrort(DIMCON, "cartsplit : n<1");
      return;
    }

    if (dim(0)!=y.dim(0))
    { XAerrort(DIMCON, "cartsplit : rows(x)<>rows(y)");
      return;
    }
    if ((t.dim(0)!=1) && (t.dim(0)!=dim(1)))
    { XAerrort(DIMCON, "cartsplit : rows(t)<>1 and rows(t)<>cols(x)");
      return;
    }

    if (y.dim(1)!=1)
    { XAerrort(DIMCON, "cartsplit : cols(y)<>1");
      return;
    }
    if (t.dim(1)!=1)
    { XAerrort(DIMCON, "cartsplit : cols(t)<>1");
      return;
    }

    for (i=2; i<MAX_XARRAY; i++)
    { if (dim(i)!=1)
      { XAerrort(DIMCON, "cartsplit : layer(x)<>1");
	return;
      }
      if (y.dim(i)!=1)
      { XAerrort(DIMCON, "cartsplit : layer(y)<>1");
	return;
      }
      if (t.dim(i)!=1)
      { XAerrort(DIMCON, "cartsplit : layer(t)<>1");
	return;
      }
    } 

    sgrp = new long[2*dim(0)];
    for (i=0; i<dim(0); i++)
      sgrp[i] = 0;

    nptr = new T[2*dim(0)];  // later: dim(0)*dim(1)
    for (i=0; i<dim(0); i++)
      SetNaN (&nptr[i]);
    splitvec->setptr (nptr);
    splitvec->setdim (dvec);

/*    nptr = new T[2*dim(0)];
    for (i=0; i<dim(0); i++)
      SetNaN (&nptr[i]);
    splitval->setptr (nptr);
    splitval->setdim (dval);
*/ 
    nptr = new T[2*dim(0)];
    for (i=0; i<dim(0); i++)
      SetNaN (&nptr[i]);
    splitmean->setptr (nptr);
    splitmean->setdim (dmean);

    nptr = new T[2*dim(0)];
    for (i=0; i<dim(0); i++)
      SetNaN (&nptr[i]);
    splitvar->setptr (nptr);
    splitvar->setdim (dvar);

    nptr = new T[2*dim(0)];
    for (i=0; i<dim(0); i++)
      SetNaN (&nptr[i]);
    splitn->setptr (nptr);
    splitn->setdim (dn);

    nptr = new T[2*dim(0)];
    for (i=0; i<dim(0); i++)
      SetNaN (&nptr[i]);
    endpoint->setptr (nptr);
    endpoint->setdim (dn);

    nextgrp = 1;
    kk = 0;

    yptr = y.getptr();
    tptr = t.getptr();
    mw = 0.0;
    for (i=0; i<dim(0); i++)
      mw += *(yptr+i);
    mw /= dim(0);
    dw = 0.0;
    for (i=0; i<dim(0); i++)
      dw += (*(yptr+i)-mw)*(*(yptr+i)-mw);

    split (ptr, dim(0), dim(1), yptr, tptr, sgrp, nmaxgrp, vmaxgrp, 0, splitvec->getptr(), splitval, &nextgrp, &kk, mw, dw, dim(0),
	    splitmean->getptr(),  splitvar->getptr(),  splitn->getptr(), nmingrp, endpoint->getptr());
    (*endpoint)[0] = kk;

    dvec.setdim (0, kk);
    dval.setdim (0, kk);
    dmean.setdim (0, kk);
    dvar.setdim (0, kk);
    dn.setdim (0, kk);
    dend.setdim (0, kk);
    splitvec->setdim (dvec);
//    splitval->setdim (dval);
    splitmean->setdim (dval);
    splitvar->setdim (dval);
    splitn->setdim (dval);
    endpoint->setdim (dend);
  }

// SK 24.02.98
  template<class T> void XplReal<T>::eiggn (XplReal<T> *vecr, XplReal<T> *veci,
					    XplReal<T> *valr, XplReal<T> *vali,
                                            XplReal<T> *imag)
  { long i, j;
    int size;

    size = dim(0);
    if (dim(0)!=dim(1))
    { XAerrort (DIMCON, "eiggn : cols(x)<>rows(x)");
      return;
    }
    if (imag) 
    { if (imag->dim(0)!=imag->dim(1))
      { XAerrort (DIMCON, "eiggn : cols(x.im)<>rows(x.im)");
        return;
      }
      if (imag->dim(0)!=dim(0))
      { XAerrort (DIMCON, "eiggn : rows(x.re)<>rows(x.im)");
        return;
      }
    } 

    for (i=2; i<MAX_XARRAY; i++)
    { if (dim(i)!=1)
      { XAerrort (DIMCON, "eiggn : layer(x)<>1");
        return;
      }
      if (imag && imag->dim(i)!=1)
      { XAerrort (DIMCON, "eiggn : layer(x.im)<>1");
        return;
      }
    }

// XploRe 3.2

  T *scale, *ar, *ai, *wr, *wi, *vr, *vi, *hr, *hi;
  int *in, *in2, low = 0, igh = 0, ierr=0;
  
  ai = new T[size*size];
  if (imag)
  { ar = imag->getptr();
    for (i=0; i<size*size; i++)
      ai[i] = ar[i];
  }
  else
  { for (i=0; i<size*size; i++)
      ai[i] = 0;
  }

  ar = new T[size*size];
  for (i=0; i<size*size; i++)
    ar[i] = ptr[i];
 
  scale = new T[size];
  for (i=0; i<size; i++) scale[i] = 1.0;

  cbal_ (&size, &size, ar, ai, &low, &igh, scale);

  in  = new int[size];
  in2 = new int[size];
  for (i=0; i<size; i++) in[i] = 0;

  comhes_ (&size, &size, &low, &igh, ar, ai, in);

  wr = new T[size];
  wi = new T[size];
  vr = new T[size*size];
  vi = new T[size*size];
  hr = new T[size*size];
  hi = new T[size*size];

  for (i=0; i<size; i++)
  { for (j=0; j<size; j++)
    { if (i>j+1)
      { 
        hr[j*size+i] = hi[j*size+i] = 0.0;
      }
      else
      { hr[j*size+i] = ar[j*size+i];
        hi[j*size+i] = ai[j*size+i];
      }
    }
    if ((i>=low-1) && (i<=igh-1))
      in2[i]=i+1;
  }

  comlr2_ (&size, &size, &low, &igh, in2, hr, hi, wr, wi, vr, vi, &ierr);

  if (ierr) 
  { XAerrort (NOCONV, "eiggn : eigenvalue computation does not converge");
    delete[] ai;
    delete[] ar;
    delete[] scale;
    delete[] in;
    delete[] in2;
    delete[] wr;
    delete[] wi;
    delete[] vr;
    delete[] vi;
    delete[] hr;
    delete[] hi;
    return;
  }

  combak_ (&size, &low, &igh, ar, ai, in, &size, vr, vi);

  cbabk2_ (&size, &size, &low, &igh, scale, &size, vr, vi);

  XplIndex vecdim (size, size), valdim(size);

  vecr->setdim (vecdim);
  vecr->setptr (vr);
  veci->setdim (vecdim);
  veci->setptr (vi); 
  valr->setdim (valdim);
  valr->setptr (wr);
  vali->setdim (valdim);
  vali->setptr (wi); 

  delete[] ai;
  delete[] ar;
  delete[] hi;
  delete[] hr;
  delete[] in;
  delete[] in2;
  delete[] scale;

}


template class XplReal<double>;

template<class T> XplReal<T> singlefunc (XplReal<T> &x, double (*func) (double l))
  { XplReal<T> ret(x.getdim());
    long i, n = x.size();
    for (i=0; i<n; i++)
      ret[i] = func (x[i]);
    return (ret);
  }

#ifndef __BORLANDC__
template class XplReal<double> singlefunc (XplReal<double> &x, double (*func) (double l));
#else
XplReal<double> singlefunc (XplReal<double> &x, double (*func) (double l));
#endif






